//
//  Agent.java
//  MathSwingJava
//
//  Created by Jason Alexander on Mon Sep 15 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.math.*;
import java.util.*;
import java.awt.image.*;
import com.wolfram.jlink.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.lang.reflect.Array;

public class Agent {
	//public static Random randomGenerator = new Random();

    int strategy;
	int nextStrategy;
	
    PayoffMatrix payoffMatrix;
    Color colors[];
	
	ArrayList interactionNeighborhood;
	ArrayList updateNeighborhood;
	LinkedList changeListeners;

	boolean useProbabilisticUpdating;
    int encounteredStrategies[];
	
	LearningRule javaLearningRule;
	Expr learningRule;
	Expr data;
	
	Model parentModel;

    double score;
    public double getScore() { return score; }

    String updateRule;
    public void setUpdateRule(String s) { updateRule = s; }
    
    int numberOfStrategies;
	int strategiesIveSeen[];
   
	public void sawStrategy(int s) {
		strategiesIveSeen[s]++;
	}
	
    public Agent(Model m) {
		parentModel = m;
		learningRule = null;
		javaLearningRule = new ImitateBestNeighbor();
		useProbabilisticUpdating = false;
		changeListeners = new LinkedList();
        colors = new Color[] { Color.red, Color.green };
		interactionNeighborhood = new ArrayList();
		updateNeighborhood = new ArrayList();
		
        if (parentModel.random.nextDouble() > .5)
			strategy = 0;
		else
			strategy = 1;

        score = 0.0;
    }
	
	public int[] strategiesIveSeen() {
		return strategiesIveSeen;
	}
	
	public void setJavaLearningRule(LearningRule lr) {
		javaLearningRule = lr;
	}

	public LearningRule getJavaLearningRule() {
		return javaLearningRule;
	}
	
	public void setLearningRule(Expr e) {
		learningRule = e;
	}
	
	public Expr getLearningRule() {
		if (learningRule != null)
			return learningRule;
		else {
			Expr e = new Expr( Expr.SYMBOL,
							   javaLearningRule.getClass().getName()
							   );
			return e;
		}
	}
	
	public void setData(Expr d) {
		data = d;
	}
	
	public Expr getData() {
		return data;
	}
    
	public void setProbabilisticUpdating(boolean b) {
		useProbabilisticUpdating = b;
	}
		
	public void addChangeListener(ChangeListener cl) {
		changeListeners.add(cl);
	}
	
	public void fireStateChanged() {
		Iterator itr = changeListeners.iterator();
		while (itr.hasNext()) {
			ChangeListener cl = (ChangeListener) itr.next();
			cl.stateChanged(new ChangeEvent(this) );
		}
	}
	
    public void setPayoffMatrix(PayoffMatrix pm) { 
		payoffMatrix = pm; 
		strategiesIveSeen = new int[ pm.getNumberOfStrategies() ];
	}
    
	public PayoffMatrix getPayoffMatrix() { return payoffMatrix; }
    
	public void deleteAllNeighbors() {
		interactionNeighborhood.clear();
		updateNeighborhood.clear();
	}
	
	public void addInteractionNeighbor(Agent n) {
		interactionNeighborhood.add(n);
	}
	
	public void addUpdateNeighbor(Object n) {
		updateNeighborhood.add(n);
	}
	
	public Agent[] getInteractionNeighbors() {
		Agent[] a = new Agent[ interactionNeighborhood.size() ];
		return (Agent[]) interactionNeighborhood.toArray(a);
	}
	
	public Collection getInteractionNeighborsAsCollection() {
		return interactionNeighborhood;
	}
	
	public Object[] getUpdateNeighbors() {
		return updateNeighborhood.toArray();
	}
	
	public Collection getUpdateNeighborsAsCollection() {
		return updateNeighborhood;
	}
	
    public void setColors(Color c[]) {
        colors = c;
    }
    
	public void setNextStrategy(int s) {
		nextStrategy = s;
	}
    
	public void setStrategy(int s) {
        strategy = s;
		fireStateChanged();
    }
	
	public void setStrategyQuietly(int s) {
		strategy = s;
	}
	
    public int getStrategy() { return strategy; }
    
    public Color getColor() {
        return colors[strategy];
    }
	
	public void getColorAsString(StringBuffer sb) {
		Color c = colors[strategy];
		sb.append("RGBColor[");
		sb.append(c.getRed()/255.0); sb.append(",");
		sb.append(c.getGreen()/255.0); sb.append(",");
		sb.append(c.getBlue()/255.0);
		sb.append("]");
	}
	
    public void reset() {
        score = 0.0;
		int i;
		for (i=0; i<Array.getLength(strategiesIveSeen); i++) {
			strategiesIveSeen[i] = 0;
		}
    }
	
	public void incrementScore(double d) {
		score += d;
	}

	Collection constructUpdateNeighborhood() {
		if (useProbabilisticUpdating == true) {
			LinkedList l = new LinkedList();
			Iterator itr = updateNeighborhood.iterator();
			while (itr.hasNext()) {
				Object o = itr.next();
				if (o.getClass().getName().equals("PointProbability")) {
					PointProbability pp = (PointProbability) o;
					Object obj;
					if ( (obj = pp.objectWithProbability()) != null ) {
						l.add(obj);
					}
				}
				else { // In this case, it's an Agent
					l.add(o);
				}
			}
			return l;
		}
		else
			return updateNeighborhood;
	}
	
	void useCustomLearningRule() {
		//System.out.println("Called custom learning rule");
		
		KernelLink kl = StdLink.getLink();
		
		try {
			//System.out.println("Trying to enable object references...");
			//kl.enableObjectReferences();
			//System.out.println("Enabled object references");
			kl.putFunction("EvaluatePacket", 1);
			//System.out.println("Put evaluate packet");
			kl.putFunction(learningRule.toString(), 1);
			//System.out.println("Called: " + learningRule.toString());	
			kl.putReference(this);
			//System.out.println("Put reference");		
			kl.endPacket();
			//System.out.println("Ended packet");		
			kl.discardAnswer();
			//System.out.println("Discarded answer");		
		}
		catch (MathLinkException e) {
			kl.clearError();
			System.out.println(e.getMessage());
			return;
		}		
	}
	
	public void determineNextStrategy() {
		if (learningRule != null) {
			useCustomLearningRule();
			return;
		}
		
		if (javaLearningRule != null) {
			nextStrategy = javaLearningRule.nextStrategy(this);
			return;
		}
	}
	
    public void adoptNewStrategy() {
        strategy = nextStrategy;
    }    
       
}
