//
//  JLinkAssistant.java
//  LocalInteractions
//
//  Created by Jason on Mon Apr 19 2004.
//  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
//

/*
 The sole purpose of this class is for offloading some computations to Java which can
 be done much faster than in Mathematica.
 */
import java.util.*;
import com.wolfram.jlink.*;
import java.lang.reflect.*;

public class JLinkAssistant {

	public static void latticeConstructor(int w, int h, int offsets[][]) throws MathLinkException {
		latticeConstructor(false, w, h, offsets);
	}
		
	public static void latticeConstructor(boolean wrap, int w, int h, int offsets[][]) 
	throws MathLinkException {
		int array[][] = new int[w][h];
		int row,col;
		int index=1;
		
		class edge {
			public int one;
			public int two;
			public edge(int v1, int v2) {
				if (v1<v2) {
					one = v1;
					two = v2;					
				}
				else {
					two = v1;
					one = v2;
				}
			}
			public boolean equals(Object obj) {
				if (obj==null) return false;
				edge p = (edge) obj;
				if (this.one == p.one && this.two == p.two)
					return true;
				else
					return false;
			}
			public int hashCode() {
				return one + 10000*two;
			}
		}
		
		// assign indexes
		for (row=0; row<h; row++) {
			for (col=0; col<w; col++) {
				array[row][col] = index++;
			}
		}
		
		// create edges
		int size = Array.getLength(offsets);
		HashSet l = new HashSet(w*h*size);
		if (wrap == false) {
			for (row=0; row<h; row++) {
				for (col=0; col<w; col++) {
					for(int k=0; k<size; k++) {
						int nbrX = col + offsets[k][0];
						int nbrY = row + offsets[k][1];
						if (0<=nbrX && nbrX<w && 0<=nbrY && nbrY<h) {
							l.add(new edge(array[row][col], array[nbrY][nbrX]));
						}
					}
				}
			}
		}
		else { // wrap == true
			for (row=0; row<h; row++) {
				for (col=0; col<w; col++) {
					for(int k=0; k<size; k++) {
						int nbrX = ((col + offsets[k][0]) + w) % w;
						int nbrY = ((row + offsets[k][1]) + h) % h;
						l.add(new edge(array[row][col], array[nbrY][nbrX]));
					}
				}
			}
		}
		
		KernelLink kl = StdLink.getLink();
		kl.beginManual();
		kl.putFunction("Graph",2);
		{
			// Write out the edges
			kl.putFunction("List", l.size());
			Iterator itr = l.iterator();
			while(itr.hasNext()) {
				edge e = (edge) itr.next();
				kl.putFunction("List",1);
				{
					kl.putFunction("List",2);
					kl.put(e.one);
					kl.put(e.two);
				}
			}
			// Write out the vertices
			// assign indexes
			kl.putFunction("List", w*h);
			for (row=0; row<h; row++) {
				for (col=0; col<w; col++) {
					kl.putFunction("List",1);
					{
						kl.putFunction("List",2);
						{
							kl.put((double)col);
							kl.put((double)row);
						}
					}
				}
			}
		}
	}
}
