//
//  Model.java
//  MathSwingJava
//
//  Created by Jason Alexander on Mon Sep 15 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import com.wolfram.jlink.*;
import com.wolfram.jlink.ui.*;
import java.lang.reflect.*;

public class Model {
	ArrayList agentList;
	LinkedList edgeList;
	
	int numberOfStrategies;
	int numberOfAgents;
    Parameters parameters;
	View view;
	
	public Random random;
	long randomSeed;
	public long getRandomSeed() { return randomSeed; }
	public void setRandomSeed(long l) {
		randomSeed = l;
		if ( random.getClass().getName().equals("MathematicaRandom") ) {
			random.setSeed( randomSeed );
		}
		else
			random = new Random ( randomSeed );
	}
	
	CombinatoricaGraph interactionNeighborhood;
	public void setInteractionNeighborhood(CombinatoricaGraph g) { 
		interactionNeighborhood = g;
	}
	
	CombinatoricaGraph updateNeighborhood;
	public void setUpdateNeighborhood(CombinatoricaGraph g) { 
		updateNeighborhood = g;
	}
	
	CombinatoricaGraph displayGraph;
	Expr cachedDisplayGraphExpr;
	public void setDisplayGraph( CombinatoricaGraph g ) {
		displayGraph = g;
		if (cachedDisplayGraphExpr != null)
			cachedDisplayGraphExpr.dispose();
		cachedDisplayGraphExpr = g.toExpr();
	}
	public Expr getDisplayGraph() {
		return cachedDisplayGraphExpr;
	}
	
	/*
	 public Expr getGraphAsExpr() {
		return view.drawPanel.graph.graphToExpr();
	}
	
	 public Expr getGraph() {
		return view.getGraphAsDisplayed();
	}
	*/
	
    int updateRule;
    String custom_update_rule_name;
    
    /*
    boolean useTraditionalFictPlay;
    public boolean useTraditionalFictitiousPlay() {
        if (updateRule == TraditionalFictitousPlay)
            return true;
        else
            return false;
    }
     */
    
	public boolean viewExists() {
		if (view == null)
			return false;
		else
			return true;
	}
	
    public void hide() {
		if (view != null) 
			view.setVisible(false);
    }
    
	public Model(int n, boolean useMathematicaGenerator) {
		super();
		
		numberOfAgents = n;
		
		edgeList = new LinkedList();
		agentList = new ArrayList( numberOfAgents );
	
		// Initialise random generator and save seed
		randomSeed = System.currentTimeMillis();
		if (useMathematicaGenerator)
			random = new MathematicaRandom( randomSeed );
		else
			random = new Random( randomSeed );
	}
	
	public void createAgents() {
		int i;
		for (i=0; i<numberOfAgents; i++) {
			Agent a = new Agent(this);
			agentList.add( a );
		}
	}
		
	public void deleteAllEdges() {
		edgeList = new LinkedList();
		Iterator itr = agentList.iterator();
		while(itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.deleteAllNeighbors();
		}
		if (view != null) {
			if (view.type == View.FAST_JAVA_GRAPHICS) {
				view.drawPanel.graph.deleteAllEdges();
			}
		}

		// Need to make sure interaction and update neighborhoods are valid.
	
	}
	
	/*
	 addEdgeToInteractionNeighborhood and addEdgesToInteractionNeighborhood only 
	 affect the Model class - i.e., the underlying data object representing the model.
	 It does not add an edge to the GEGraph object which displays the graph. This is
	 because the most frequent use of these functions is during model creation, where
	 for reasons of efficiency we create the edges in the GEGraph class a different way
	 (i.e., by passing in a string representation of the Mathematica graph class).  Hence,
	 if the user attempts to programmatically add an edge to the graph from the 
	 Mathematica front end, we need to call addEdgeToGraphDisplay to also add the edge
	 to the display engine.																				
	 */
	public void addEdgeToInteractionNeighborhood(double edge[]) {
		Agent one = (Agent) agentList.get( (int) edge[0] );
		Agent two = (Agent) agentList.get( (int) edge[1] );
		double weight = edge[2];
		boolean undirected = (edge[3] == 1.0)? true: false;
		boolean useWeights = (edge[4] == 1.0)? true: false;
		Edge e = new Edge(one, two, weight, undirected, useWeights);
		edgeList.add( e );
		if (e.isUndirected()) {
			one.addInteractionNeighbor( two );
			two.addInteractionNeighbor( one );
		}
		else {
			one.addInteractionNeighbor( two );
		}
	}
	
	public void addEdgesToInteractionNeighborhood(double edges[][]) {
		int i;
		for (i=0; i<Array.getLength( edges ); i++) {
			// the array contains the following information:
			//
			// agent one, agent two,weight, undirected, useWeights
			//
			// we code it this way for efficiency from Mathematica
			addEdgeToInteractionNeighborhood(edges[i]);
			/*
			 Agent one = (Agent) agentList.get( (int) edges[i][0] );
			Agent two = (Agent) agentList.get( (int) edges[i][1] );
			double weight = edges[i][2];
			boolean undirected = (edges[i][3] == 1.0)? true: false;
			boolean useWeights = (edges[i][4] == 1.0)? true: false;
			Edge e = new Edge(one, two, weight, undirected, useWeights);
			edgeList.add( e );
			if (e.isUndirected()) {
				one.addInteractionNeighbor( two );
				two.addInteractionNeighbor( one );
			}
			else {
				one.addInteractionNeighbor( two );
			}
			*/
		}
	}
	
	public void wireInteractionNeighborhood(int edges[][], boolean undirected) {
		int i;
		for (i=0; i < Array.getLength(edges); i++) {
			Agent one = (Agent) agentList.get( edges[i][0] );
			Agent two = (Agent) agentList.get( edges[i][1] ) ;
			edgeList.add( new Edge(one, two, undirected) );
			if (undirected) {
				one.addInteractionNeighbor( two );
				two.addInteractionNeighbor( one );
			}
			else {
				one.addInteractionNeighbor( two );
			}
		}		
	}
	
	public void NEWWireInteractionNeighborhood(int edges[][], 
											   boolean isDirected[],
											   boolean isWeighted[],
											   double weights[]) 
	{
		for (int i=0; i < Array.getLength(edges); i++) {
			Agent one = (Agent) agentList.get( edges[i][0] );
			Agent two = (Agent) agentList.get( edges[i][1] );
			boolean undirected = (isDirected[i])? false: true;
			boolean weighted = isWeighted[i];
			Edge e = new Edge( one, two, weights[i], undirected, weighted );
			edgeList.add( e );
		}
	}
	
	public void addEdgesToGraphDisplay( int edges[][], Color c) {
		if (view != null) {
			if (view.type == View.FAST_JAVA_GRAPHICS) {
				view.drawPanel.graph.addEdges(edges, c);
			}
		}
	}
	
	public void addEdgeToGraphDisplay(int one, int two) {
		if (view != null) {
			if (view.type == View.FAST_JAVA_GRAPHICS) {
				view.drawPanel.graph.addEdge(one, two);
				view.repaint();
			}
		}
	}
	
	public void addEdgeToUpdateNeighborhood(int one, int two) {
		Agent a = (Agent) agentList.get( one );
		a.addUpdateNeighbor( (Agent) agentList.get( two ) );
	}
	
	public void addEdgeToUpdateNeighborhood(int one, int two, double prob) {
		Agent a = (Agent) agentList.get( one );
		Agent b = (Agent) agentList.get( two );
		PointProbability pp = new PointProbability(b, prob, random);
		a.addUpdateNeighbor( pp );
	}
	
	public void NEWWireUpdateNeighborhood(int edges[][], 
											   boolean isDirected[],
											   boolean isWeighted[],
											   double weights[]) 
	{
		for (int i=0; i < Array.getLength(edges); i++) {
			Agent one = (Agent) agentList.get( edges[i][0] );
			Agent two = (Agent) agentList.get( edges[i][1] );
			boolean undirected = (isDirected[i])? false: true;
			boolean weighted = isWeighted[i];
			
			if (undirected == true) {
				if (weighted == false) {
					one.addUpdateNeighbor( two );
					two.addUpdateNeighbor( one );
				}
				else {
					one.addUpdateNeighbor(
										  new PointProbability(two, weights[i], random)
										  );
					two.addUpdateNeighbor(
										  new PointProbability(one, weights[i], random)
										  );
					one.setProbabilisticUpdating( true );
					two.setProbabilisticUpdating( true );
				}
			}
			else {
				if (weighted == false) {
					one.addUpdateNeighbor( two );
				}
				else {
					one.addUpdateNeighbor(
										  new PointProbability(two, weights[i], random)
										  );
					one.setProbabilisticUpdating( true );
				}
			}
		}
	}
	
	public void wireUpdateNeighborhood(int edges[][]) {
		int i;
		for (i=0; i < Array.getLength(edges); i++) {
			Agent a = (Agent) agentList.get( edges[i][0] );
			a.addUpdateNeighbor( (Agent) agentList.get( edges[i][1] ) );
		}
	}
	
	public void wireUpdateNeighborhoodWithProbabilities(double edges[][]) {
		/*
		 The array consists of the following:
		 Agent one index, Agent two index, probability of interacting
		 */
		int i;
		
		for (i=0; i < Array.getLength(edges); i++) {
			Agent one = (Agent) agentList.get( (int) edges[i][0] );
			Agent two = (Agent) agentList.get( (int) edges[i][1] );
			double pr = edges[i][2];
			PointProbability pp = new PointProbability(two, pr, random);
			one.addUpdateNeighbor( pp );
			one.setProbabilisticUpdating( true );
		}
	}
	
	public void setGraphAndReparse(CombinatoricaGraph g) {
		displayGraph = g;
		if (view != null && view.type == View.FAST_JAVA_GRAPHICS) {
			Expr expr = displayGraph.toExpr();
			view.drawPanel.graph.readGraphFromExpr(expr);
			expr.dispose();
		}
	}
	
	
	public void createMainFrame(Expr graphicsType) {
		if (graphicsType.toString().equals("Mathematica")) {
			view = new View(this, View.MATHEMATICA_GRAPHICS);
		}
		else {
			view = new View(this, View.FAST_JAVA_GRAPHICS);
		}
		
		Iterator itr = agentList.iterator();
		while( itr.hasNext() ) {
			Agent a = (Agent) itr.next();
			a.addChangeListener(view);
		}
		
		view.layoutView();
		//System.out.println("Laid out View successfully.");
	}
	
    public void showWindow() { 
		view.updateGraphDisplay();
		view.setVisible(true);
	}
	
	public void resize() {
		view.invalidate();
		view.validate();
		view.updateGraphDisplay();
	}

    public void setLabel(String s) { view.setTitle(s); }
	
	public void setLearningRule(LearningRule lr) {
		Iterator itr = agentList.iterator();
		while (itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.setJavaLearningRule(lr);
		}
	}

	public void setLearningRule(Expr e) {
		Iterator itr = agentList.iterator();
		while (itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.setLearningRule(e);
		}
	}
	
	public Agent getAgent(int i) {
		return (Agent) agentList.get(i);
	}
	
    public Agent[] getAgents() {
		Agent l[] = new Agent[ numberOfAgents ];
		int i;
		for (i=0; i<numberOfAgents; i++) 
			l[i] = (Agent) agentList.get(i);
        return l;
    }
	
    public int getNumberOfAgents() {
        return agentList.size();
    }

    public void setColors(Color c[]) {
        Iterator i = agentList.iterator();
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            a.setColors(c);
        }
    }
    
    public void setStrategies(int l[]) {
        Iterator i = agentList.iterator();
        int j=0;
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            a.setStrategyQuietly(l[j++]);
        }
    }

	public void setNumberOfStrategies(int i) {
		numberOfStrategies = i;
	}
	
	public int getNumberOfStrategies() {
		return numberOfStrategies;
	}
	
    public void setPayoffMatrices(double pm[][][][]) throws MathLinkException {
		int numberOfMatrices = Array.getLength(pm);
		if (numberOfMatrices != agentList.size()) {
			throw new MathLinkException(0, "The number of payoff matrices given does not equal the number of agents.");
		}
        
		Iterator i = agentList.iterator();
        int j=0;
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            a.setPayoffMatrix(new PayoffMatrix(pm[j++]));
        }
    }
    
    public void interact() {
		resetAgents();        
		calculateAgentScores();		
		updateStrategies();
		if (view != null)
			view.updateGraphDisplay();
    }
	
	public void interact(boolean b) {
		resetAgents();        
		calculateAgentScores();		
		updateStrategies();
		if (b && view != null)
			view.updateGraphDisplay();
    }
	
	public void updateDisplay() {
		if (view != null)
			view.updateGraphDisplay();
	}
	
	void resetAgents() {
		Iterator itr = agentList.iterator();
		while (itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.reset();
		}
	}

	public void showVertexNumbers() { view.showVertexNumbers(); }
	public void hideVertexNumbers() { view.hideVertexNumbers(); }
	
	
	void calculateAgentScores() {
		int total_edges=0, unused_edges=0;
		Iterator itr = edgeList.iterator();
        while (itr.hasNext()) {
            Edge e = (Edge) itr.next();
			Agent one = e.one;
			Agent two = e.two;
			PayoffMatrix pmOne = one.getPayoffMatrix();
			PayoffMatrix pmTwo = two.getPayoffMatrix();
			
			double ran = random.nextDouble();
			total_edges++;
			
			if ( !e.isWeighted() || (e.isWeighted() && ran < e.getWeight()) ) {
				// Flip a coin to see who is row. If <.5, player one is; otherwise, player two is.
				if (random.nextDouble() < .5) {
					// Player one is row.
					if (e.isUndirected()) {
						double playerOnePayoff = pmOne.getPayoff(one.getStrategy(), two.getStrategy(), PayoffMatrix.ROW);
						double playerTwoPayoff = pmTwo.getPayoff(one.getStrategy(), two.getStrategy(), PayoffMatrix.COLUMN);
						one.incrementScore( playerOnePayoff );
						two.incrementScore( playerTwoPayoff );
						one.sawStrategy( two.getStrategy() );
						two.sawStrategy( one.getStrategy() );
					}
					else {
						double playerOnePayoff = pmOne.getPayoff(one.getStrategy(), two.getStrategy(), PayoffMatrix.ROW);
						one.incrementScore( playerOnePayoff );
						one.sawStrategy( two.getStrategy() );
					}
				}
				else {
					// Player two is row
					if (e.isUndirected()) {
						double playerOnePayoff = pmOne.getPayoff(two.getStrategy(), one.getStrategy(), PayoffMatrix.COLUMN);
						double playerTwoPayoff = pmTwo.getPayoff(two.getStrategy(), one.getStrategy(), PayoffMatrix.ROW);
						one.incrementScore( playerOnePayoff );
						two.incrementScore( playerTwoPayoff );					
						one.sawStrategy( two.getStrategy() );
						two.sawStrategy( one.getStrategy() );
					}
					else {
						double playerOnePayoff = pmOne.getPayoff(two.getStrategy(), one.getStrategy(), PayoffMatrix.COLUMN);
						one.incrementScore( playerOnePayoff );
						one.sawStrategy( two.getStrategy() );
					}
				}
			}
			else {
				unused_edges++;
			}
		}
		//System.out.println("Of " + total_edges + " total edges, " + unused_edges + " were unused.");
	}
	
	void updateStrategies() {
		Iterator itr;
		
		itr = agentList.iterator();
		while (itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.determineNextStrategy();
		}
		
		itr = agentList.iterator();
		while (itr.hasNext()) {
			Agent a = (Agent) itr.next();
			a.adoptNewStrategy();
		}
	}
	
	public String getColorsAsString() {
		int size = agentList.size();
		String s = "RGBColor[.77777777,.77777777,.77777777]";
		StringBuffer sb = new StringBuffer( s.length() * size + size + 2 );
		sb.append("{");
		
		Iterator itr = agentList.iterator();
		int j=0;
		while (itr.hasNext()) {
			Agent a = (Agent)itr.next();
			a.getColorAsString(sb);

			if (j!=size-1)
				sb.append(",");
			
			j++;
		}
		sb.append("}");
		return sb.toString();
	}
	
    public int[] getStrategies() {
        Iterator i = agentList.iterator();
        int array[] = new int[agentList.size()];
        int j=0;
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            array[j++] = a.strategy;
        }
        return array;
    }

	public int[] getStrategyTotals() {
        Iterator i = agentList.iterator();
        int array[] = new int[numberOfStrategies];
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            array[a.strategy]++ ;
        }
        return array;		
	}
	
    public double[] getFrequencyDistribution() {
        Iterator i = agentList.iterator();
        int size = agentList.size();
        double array[] = new double[numberOfStrategies];
        while (i.hasNext()) {
            Agent a = (Agent) i.next();
            array[a.strategy]++;
        }

        int j;
        for (j=0; j<array.length; j++) {
            array[j] = array[j]/size;
        }
        
        return array;
    }

	/*
	public String parseEdges(Expr edges) throws ExprFormatException {
		int i;
		int totalEdges = 0;
		for (i=1; i<= edges.length(); i++) {
			parseEdge( edges.part(i) );
			totalEdges++;
		}
		
		edges.dispose();
		
		return ("Total edges parsed: " + totalEdges);
	}
	
	final Expr Rule = new Expr(Expr.SYMBOL, "Rule");
	final Expr EdgeWeight = new Expr(Expr.SYMBOL, "EdgeWeight");
	
	protected Edge parseEdge(Expr edge) throws ExprFormatException {
		final boolean printStatements = false;
		//
		// An edge has the form {{v1,v2}, Rule[Foo1, Bar1], Rule[Foo2, Bar2],..., Rule[FooN, BarN]}
		//
		int one = edge.part( new int[] {1,1} ).asInt() - 1;
		int two = edge.part( new int[] {1,2} ).asInt() - 1;
			
		if (printStatements)
				System.out.println("Edge connecting vertex " + one + " to vertex " + two);
			
		int j;
		double weight = 0.0;
		boolean weighted = false;
		// Now parse any addition Rules which apply to the edge
		for (j=2; j <= edge.length(); j++) {
			Expr subExpr = edge.part(j); // grab the jth subexpression (typically a rule)
			Expr head = subExpr.head();
			if (subExpr.head().equals(Rule)) {
				Expr symbol = subExpr.part(1); // the symbol of the rule
				if (symbol.equals(EdgeWeight)) {
					// process edgeweight
					weight = subExpr.part(2).asDouble();
					weighted = true;
					if (printStatements)
						System.out.println("	Edge has weight " + weight);
				}
			}
			subExpr.dispose();
			head.dispose();
		}
		
		edge.dispose();
		//Agent agentOne = (Agent) agentList.get(one);
		//Agent agentTwo = (Agent) agentList.get(two);
		
		Edge theEdge = new Edge((Agent)agentList.get(1), (Agent)agentList.get(1));
		if (weighted)
			theEdge.setWeight(weight);
		
		return theEdge;
	}
	 */
		
	/*
	 MATHEMATICA FRONT END HELPER FUNCTIONS
	 
	 The functions defined after this section are purely helper functions for interfacing with the 
	 Mathematica front end.
	 */
	public void extractGraph() throws MathLinkException {
		KernelLink kl = StdLink.getLink();
		kl.beginManual();
		if ( viewExists() && view.type == View.FAST_JAVA_GRAPHICS ) {
			view.drawPanel.sendGraphToMathematica(kl);
		}
		else {
			displayGraph.sendToMathematica(kl);
		}
		kl.endPacket();
	}
	
	public void sendInteractionNeighborhoodToMathematica() throws MathLinkException {
		KernelLink kl = StdLink.getLink();
		int i;
		kl.beginManual();
		interactionNeighborhood.sendToMathematica(kl);
		kl.endPacket();
	}
	
	public void sendUpdateNeighborhoodToMathematica() throws MathLinkException {
		KernelLink kl = StdLink.getLink();
		int i;
		kl.beginManual();
		updateNeighborhood.sendToMathematica(kl);
		kl.endPacket();
	}

	public void sendDisplayNeighborhoodToMathematica() throws MathLinkException {
		KernelLink kl = StdLink.getLink();
		int i;
		kl.beginManual();
		updateNeighborhood.sendToMathematica(kl);
		kl.endPacket();
	}
	
	public void sendAgentColorsToMathematica() throws MathLinkException {
		KernelLink kl = StdLink.getLink();
		kl.beginManual();
		kl.putFunction("List", agentList.size());
		{
			Iterator itr = agentList.iterator();
			while(itr.hasNext()) {
				Agent a = (Agent) itr.next();
				Color c = a.getColor();
				kl.putFunction("RGBColor", 3);
				{
					kl.put(c.getRed()/255.0);
					kl.put(c.getGreen()/255.0);
					kl.put(c.getBlue()/255.0);
				}
			}
		}
		kl.endPacket();
	}
}
