(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    308430,       6345]*)
(*NotebookOutlinePosition[    564415,      15241]*)
(*  CellTagsIndexPosition[    564327,      15235]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Agents", "ObjectName",
  CellTags->"LocalInteractions:2"],

Cell[TextData[{
  "\[Bullet] ",
  StyleBox["Agents", "MR"],
  "[ ",
  StyleBox["model ",
    FontSlant->"Italic"],
  "] gives a list of the agents in ",
  StyleBox["model",
    FontSlant->"Italic"],
  ".\n",
  "\[Bullet] ",
  StyleBox["Agents", "MR"],
  "[ ",
  StyleBox["model, list ",
    FontSlant->"Italic"],
  "] gives the agents in ",
  StyleBox["model ",
    FontSlant->"Italic"],
  "whose indexes are included in ",
  StyleBox["list.",
    FontSlant->"Italic"]
}], "Usage",
  CellTags->"LocalInteractions:2"],

Cell[TextData[{
  "\[FilledSmallSquare] The first time ",
  StyleBox["Agents", "MR"],
  " is called will be slow as the Agent Java objects are loaded into the \
kernel. After that, each call will be quite fast."
}], "Notes",
  CellTags->"LocalInteractions:2"],

Cell[TextData[{
  "\[FilledSmallSquare] If you call ",
  StyleBox["Agents", "MR"],
  " on a model, you will need to explicitly release the agent objects after \
deleting the model, as ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " will keep a hold on them."
}], "Notes",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell["Example", "ExampleSection",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[{
    \(\(m = LocalInteractionModel[];\)\), "\[IndentingNewLine]", 
    \(\(agentlist = Agents[m];\)\)}], "Input",
  CellLabel->"In[67]:=",
  CellTags->"LocalInteractions:2"],

Cell["\<\
All of the agents initially have strategy 1 (red) in the default \
model.\
\>", "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowLabeledGraph[m];\)\)], "Input",
  CellLabel->"In[69]:=",
  CellTags->"LocalInteractions:2"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.04545 .04545 m
.14646 .04545 L
s
.04545 .04545 m
.04545 .14646 L
s
.14646 .04545 m
.24747 .04545 L
s
.14646 .04545 m
.14646 .14646 L
s
.24747 .04545 m
.34848 .04545 L
s
.24747 .04545 m
.24747 .14646 L
s
.34848 .04545 m
.44949 .04545 L
s
.34848 .04545 m
.34848 .14646 L
s
.44949 .04545 m
.55051 .04545 L
s
.44949 .04545 m
.44949 .14646 L
s
.55051 .04545 m
.65152 .04545 L
s
.55051 .04545 m
.55051 .14646 L
s
.65152 .04545 m
.75253 .04545 L
s
.65152 .04545 m
.65152 .14646 L
s
.75253 .04545 m
.85354 .04545 L
s
.75253 .04545 m
.75253 .14646 L
s
.85354 .04545 m
.95455 .04545 L
s
.85354 .04545 m
.85354 .14646 L
s
.95455 .04545 m
.95455 .14646 L
s
.04545 .14646 m
.14646 .14646 L
s
.04545 .14646 m
.04545 .24747 L
s
.14646 .14646 m
.24747 .14646 L
s
.14646 .14646 m
.14646 .24747 L
s
.24747 .14646 m
.34848 .14646 L
s
.24747 .14646 m
.24747 .24747 L
s
.34848 .14646 m
.44949 .14646 L
s
.34848 .14646 m
.34848 .24747 L
s
.44949 .14646 m
.55051 .14646 L
s
.44949 .14646 m
.44949 .24747 L
s
.55051 .14646 m
.65152 .14646 L
s
.55051 .14646 m
.55051 .24747 L
s
.65152 .14646 m
.75253 .14646 L
s
.65152 .14646 m
.65152 .24747 L
s
.75253 .14646 m
.85354 .14646 L
s
.75253 .14646 m
.75253 .24747 L
s
.85354 .14646 m
.95455 .14646 L
s
.85354 .14646 m
.85354 .24747 L
s
.95455 .14646 m
.95455 .24747 L
s
.04545 .24747 m
.14646 .24747 L
s
.04545 .24747 m
.04545 .34848 L
s
.14646 .24747 m
.24747 .24747 L
s
.14646 .24747 m
.14646 .34848 L
s
.24747 .24747 m
.34848 .24747 L
s
.24747 .24747 m
.24747 .34848 L
s
.34848 .24747 m
.44949 .24747 L
s
.34848 .24747 m
.34848 .34848 L
s
.44949 .24747 m
.55051 .24747 L
s
.44949 .24747 m
.44949 .34848 L
s
.55051 .24747 m
.65152 .24747 L
s
.55051 .24747 m
.55051 .34848 L
s
.65152 .24747 m
.75253 .24747 L
s
.65152 .24747 m
.65152 .34848 L
s
.75253 .24747 m
.85354 .24747 L
s
.75253 .24747 m
.75253 .34848 L
s
.85354 .24747 m
.95455 .24747 L
s
.85354 .24747 m
.85354 .34848 L
s
.95455 .24747 m
.95455 .34848 L
s
.04545 .34848 m
.14646 .34848 L
s
.04545 .34848 m
.04545 .44949 L
s
.14646 .34848 m
.24747 .34848 L
s
.14646 .34848 m
.14646 .44949 L
s
.24747 .34848 m
.34848 .34848 L
s
.24747 .34848 m
.24747 .44949 L
s
.34848 .34848 m
.44949 .34848 L
s
.34848 .34848 m
.34848 .44949 L
s
.44949 .34848 m
.55051 .34848 L
s
.44949 .34848 m
.44949 .44949 L
s
.55051 .34848 m
.65152 .34848 L
s
.55051 .34848 m
.55051 .44949 L
s
.65152 .34848 m
.75253 .34848 L
s
.65152 .34848 m
.65152 .44949 L
s
.75253 .34848 m
.85354 .34848 L
s
.75253 .34848 m
.75253 .44949 L
s
.85354 .34848 m
.95455 .34848 L
s
.85354 .34848 m
.85354 .44949 L
s
.95455 .34848 m
.95455 .44949 L
s
.04545 .44949 m
.14646 .44949 L
s
.04545 .44949 m
.04545 .55051 L
s
.14646 .44949 m
.24747 .44949 L
s
.14646 .44949 m
.14646 .55051 L
s
.24747 .44949 m
.34848 .44949 L
s
.24747 .44949 m
.24747 .55051 L
s
.34848 .44949 m
.44949 .44949 L
s
.34848 .44949 m
.34848 .55051 L
s
.44949 .44949 m
.55051 .44949 L
s
.44949 .44949 m
.44949 .55051 L
s
.55051 .44949 m
.65152 .44949 L
s
.55051 .44949 m
.55051 .55051 L
s
.65152 .44949 m
.75253 .44949 L
s
.65152 .44949 m
.65152 .55051 L
s
.75253 .44949 m
.85354 .44949 L
s
.75253 .44949 m
.75253 .55051 L
s
.85354 .44949 m
.95455 .44949 L
s
.85354 .44949 m
.85354 .55051 L
s
.95455 .44949 m
.95455 .55051 L
s
.04545 .55051 m
.14646 .55051 L
s
.04545 .55051 m
.04545 .65152 L
s
.14646 .55051 m
.24747 .55051 L
s
.14646 .55051 m
.14646 .65152 L
s
.24747 .55051 m
.34848 .55051 L
s
.24747 .55051 m
.24747 .65152 L
s
.34848 .55051 m
.44949 .55051 L
s
.34848 .55051 m
.34848 .65152 L
s
.44949 .55051 m
.55051 .55051 L
s
.44949 .55051 m
.44949 .65152 L
s
.55051 .55051 m
.65152 .55051 L
s
.55051 .55051 m
.55051 .65152 L
s
.65152 .55051 m
.75253 .55051 L
s
.65152 .55051 m
.65152 .65152 L
s
.75253 .55051 m
.85354 .55051 L
s
.75253 .55051 m
.75253 .65152 L
s
.85354 .55051 m
.95455 .55051 L
s
.85354 .55051 m
.85354 .65152 L
s
.95455 .55051 m
.95455 .65152 L
s
.04545 .65152 m
.14646 .65152 L
s
.04545 .65152 m
.04545 .75253 L
s
.14646 .65152 m
.24747 .65152 L
s
.14646 .65152 m
.14646 .75253 L
s
.24747 .65152 m
.34848 .65152 L
s
.24747 .65152 m
.24747 .75253 L
s
.34848 .65152 m
.44949 .65152 L
s
.34848 .65152 m
.34848 .75253 L
s
.44949 .65152 m
.55051 .65152 L
s
.44949 .65152 m
.44949 .75253 L
s
.55051 .65152 m
.65152 .65152 L
s
.55051 .65152 m
.55051 .75253 L
s
.65152 .65152 m
.75253 .65152 L
s
.65152 .65152 m
.65152 .75253 L
s
.75253 .65152 m
.85354 .65152 L
s
.75253 .65152 m
.75253 .75253 L
s
.85354 .65152 m
.95455 .65152 L
s
.85354 .65152 m
.85354 .75253 L
s
.95455 .65152 m
.95455 .75253 L
s
.04545 .75253 m
.14646 .75253 L
s
.04545 .75253 m
.04545 .85354 L
s
.14646 .75253 m
.24747 .75253 L
s
.14646 .75253 m
.14646 .85354 L
s
.24747 .75253 m
.34848 .75253 L
s
.24747 .75253 m
.24747 .85354 L
s
.34848 .75253 m
.44949 .75253 L
s
.34848 .75253 m
.34848 .85354 L
s
.44949 .75253 m
.55051 .75253 L
s
.44949 .75253 m
.44949 .85354 L
s
.55051 .75253 m
.65152 .75253 L
s
.55051 .75253 m
.55051 .85354 L
s
.65152 .75253 m
.75253 .75253 L
s
.65152 .75253 m
.65152 .85354 L
s
.75253 .75253 m
.85354 .75253 L
s
.75253 .75253 m
.75253 .85354 L
s
.85354 .75253 m
.95455 .75253 L
s
.85354 .75253 m
.85354 .85354 L
s
.95455 .75253 m
.95455 .85354 L
s
.04545 .85354 m
.14646 .85354 L
s
.04545 .85354 m
.04545 .95455 L
s
.14646 .85354 m
.24747 .85354 L
s
.14646 .85354 m
.14646 .95455 L
s
.24747 .85354 m
.34848 .85354 L
s
.24747 .85354 m
.24747 .95455 L
s
.34848 .85354 m
.44949 .85354 L
s
.34848 .85354 m
.34848 .95455 L
s
.44949 .85354 m
.55051 .85354 L
s
.44949 .85354 m
.44949 .95455 L
s
.55051 .85354 m
.65152 .85354 L
s
.55051 .85354 m
.55051 .95455 L
s
.65152 .85354 m
.75253 .85354 L
s
.65152 .85354 m
.65152 .95455 L
s
.75253 .85354 m
.85354 .85354 L
s
.75253 .85354 m
.75253 .95455 L
s
.85354 .85354 m
.95455 .85354 L
s
.85354 .85354 m
.85354 .95455 L
s
.95455 .85354 m
.95455 .95455 L
s
.04545 .95455 m
.14646 .95455 L
s
.14646 .95455 m
.24747 .95455 L
s
.24747 .95455 m
.34848 .95455 L
s
.34848 .95455 m
.44949 .95455 L
s
.44949 .95455 m
.55051 .95455 L
s
.55051 .95455 m
.65152 .95455 L
s
.65152 .95455 m
.75253 .95455 L
s
.75253 .95455 m
.85354 .95455 L
s
.85354 .95455 m
.95455 .95455 L
s
1 0 0 r
.025 w
.04545 .04545 Mdot
0 0 0 r
[(1)] .02545 .02545 1 0 Mshowa
1 0 0 r
.14646 .04545 Mdot
0 0 0 r
[(2)] .12646 .02545 1 0 Mshowa
1 0 0 r
.24747 .04545 Mdot
0 0 0 r
[(3)] .22747 .02545 1 0 Mshowa
1 0 0 r
.34848 .04545 Mdot
0 0 0 r
[(4)] .32848 .02545 1 0 Mshowa
1 0 0 r
.44949 .04545 Mdot
0 0 0 r
[(5)] .42949 .02545 1 0 Mshowa
1 0 0 r
.55051 .04545 Mdot
0 0 0 r
[(6)] .53051 .02545 1 0 Mshowa
1 0 0 r
.65152 .04545 Mdot
0 0 0 r
[(7)] .63152 .02545 1 0 Mshowa
1 0 0 r
.75253 .04545 Mdot
0 0 0 r
[(8)] .73253 .02545 1 0 Mshowa
1 0 0 r
.85354 .04545 Mdot
0 0 0 r
[(9)] .83354 .02545 1 0 Mshowa
1 0 0 r
.95455 .04545 Mdot
0 0 0 r
[(10)] .93455 .02545 1 0 Mshowa
1 0 0 r
.04545 .14646 Mdot
0 0 0 r
[(11)] .02545 .12646 1 0 Mshowa
1 0 0 r
.14646 .14646 Mdot
0 0 0 r
[(12)] .12646 .12646 1 0 Mshowa
1 0 0 r
.24747 .14646 Mdot
0 0 0 r
[(13)] .22747 .12646 1 0 Mshowa
1 0 0 r
.34848 .14646 Mdot
0 0 0 r
[(14)] .32848 .12646 1 0 Mshowa
1 0 0 r
.44949 .14646 Mdot
0 0 0 r
[(15)] .42949 .12646 1 0 Mshowa
1 0 0 r
.55051 .14646 Mdot
0 0 0 r
[(16)] .53051 .12646 1 0 Mshowa
1 0 0 r
.65152 .14646 Mdot
0 0 0 r
[(17)] .63152 .12646 1 0 Mshowa
1 0 0 r
.75253 .14646 Mdot
0 0 0 r
[(18)] .73253 .12646 1 0 Mshowa
1 0 0 r
.85354 .14646 Mdot
0 0 0 r
[(19)] .83354 .12646 1 0 Mshowa
1 0 0 r
.95455 .14646 Mdot
0 0 0 r
[(20)] .93455 .12646 1 0 Mshowa
1 0 0 r
.04545 .24747 Mdot
0 0 0 r
[(21)] .02545 .22747 1 0 Mshowa
1 0 0 r
.14646 .24747 Mdot
0 0 0 r
[(22)] .12646 .22747 1 0 Mshowa
1 0 0 r
.24747 .24747 Mdot
0 0 0 r
[(23)] .22747 .22747 1 0 Mshowa
1 0 0 r
.34848 .24747 Mdot
0 0 0 r
[(24)] .32848 .22747 1 0 Mshowa
1 0 0 r
.44949 .24747 Mdot
0 0 0 r
[(25)] .42949 .22747 1 0 Mshowa
1 0 0 r
.55051 .24747 Mdot
0 0 0 r
[(26)] .53051 .22747 1 0 Mshowa
1 0 0 r
.65152 .24747 Mdot
0 0 0 r
[(27)] .63152 .22747 1 0 Mshowa
1 0 0 r
.75253 .24747 Mdot
0 0 0 r
[(28)] .73253 .22747 1 0 Mshowa
1 0 0 r
.85354 .24747 Mdot
0 0 0 r
[(29)] .83354 .22747 1 0 Mshowa
1 0 0 r
.95455 .24747 Mdot
0 0 0 r
[(30)] .93455 .22747 1 0 Mshowa
1 0 0 r
.04545 .34848 Mdot
0 0 0 r
[(31)] .02545 .32848 1 0 Mshowa
1 0 0 r
.14646 .34848 Mdot
0 0 0 r
[(32)] .12646 .32848 1 0 Mshowa
1 0 0 r
.24747 .34848 Mdot
0 0 0 r
[(33)] .22747 .32848 1 0 Mshowa
1 0 0 r
.34848 .34848 Mdot
0 0 0 r
[(34)] .32848 .32848 1 0 Mshowa
1 0 0 r
.44949 .34848 Mdot
0 0 0 r
[(35)] .42949 .32848 1 0 Mshowa
1 0 0 r
.55051 .34848 Mdot
0 0 0 r
[(36)] .53051 .32848 1 0 Mshowa
1 0 0 r
.65152 .34848 Mdot
0 0 0 r
[(37)] .63152 .32848 1 0 Mshowa
1 0 0 r
.75253 .34848 Mdot
0 0 0 r
[(38)] .73253 .32848 1 0 Mshowa
1 0 0 r
.85354 .34848 Mdot
0 0 0 r
[(39)] .83354 .32848 1 0 Mshowa
1 0 0 r
.95455 .34848 Mdot
0 0 0 r
[(40)] .93455 .32848 1 0 Mshowa
1 0 0 r
.04545 .44949 Mdot
0 0 0 r
[(41)] .02545 .42949 1 0 Mshowa
1 0 0 r
.14646 .44949 Mdot
0 0 0 r
[(42)] .12646 .42949 1 0 Mshowa
1 0 0 r
.24747 .44949 Mdot
0 0 0 r
[(43)] .22747 .42949 1 0 Mshowa
1 0 0 r
.34848 .44949 Mdot
0 0 0 r
[(44)] .32848 .42949 1 0 Mshowa
1 0 0 r
.44949 .44949 Mdot
0 0 0 r
[(45)] .42949 .42949 1 0 Mshowa
1 0 0 r
.55051 .44949 Mdot
0 0 0 r
[(46)] .53051 .42949 1 0 Mshowa
1 0 0 r
.65152 .44949 Mdot
0 0 0 r
[(47)] .63152 .42949 1 0 Mshowa
1 0 0 r
.75253 .44949 Mdot
0 0 0 r
[(48)] .73253 .42949 1 0 Mshowa
1 0 0 r
.85354 .44949 Mdot
0 0 0 r
[(49)] .83354 .42949 1 0 Mshowa
1 0 0 r
.95455 .44949 Mdot
0 0 0 r
[(50)] .93455 .42949 1 0 Mshowa
1 0 0 r
.04545 .55051 Mdot
0 0 0 r
[(51)] .02545 .53051 1 0 Mshowa
1 0 0 r
.14646 .55051 Mdot
0 0 0 r
[(52)] .12646 .53051 1 0 Mshowa
1 0 0 r
.24747 .55051 Mdot
0 0 0 r
[(53)] .22747 .53051 1 0 Mshowa
1 0 0 r
.34848 .55051 Mdot
0 0 0 r
[(54)] .32848 .53051 1 0 Mshowa
1 0 0 r
.44949 .55051 Mdot
0 0 0 r
[(55)] .42949 .53051 1 0 Mshowa
1 0 0 r
.55051 .55051 Mdot
0 0 0 r
[(56)] .53051 .53051 1 0 Mshowa
1 0 0 r
.65152 .55051 Mdot
0 0 0 r
[(57)] .63152 .53051 1 0 Mshowa
1 0 0 r
.75253 .55051 Mdot
0 0 0 r
[(58)] .73253 .53051 1 0 Mshowa
1 0 0 r
.85354 .55051 Mdot
0 0 0 r
[(59)] .83354 .53051 1 0 Mshowa
1 0 0 r
.95455 .55051 Mdot
0 0 0 r
[(60)] .93455 .53051 1 0 Mshowa
1 0 0 r
.04545 .65152 Mdot
0 0 0 r
[(61)] .02545 .63152 1 0 Mshowa
1 0 0 r
.14646 .65152 Mdot
0 0 0 r
[(62)] .12646 .63152 1 0 Mshowa
1 0 0 r
.24747 .65152 Mdot
0 0 0 r
[(63)] .22747 .63152 1 0 Mshowa
1 0 0 r
.34848 .65152 Mdot
0 0 0 r
[(64)] .32848 .63152 1 0 Mshowa
1 0 0 r
.44949 .65152 Mdot
0 0 0 r
[(65)] .42949 .63152 1 0 Mshowa
1 0 0 r
.55051 .65152 Mdot
0 0 0 r
[(66)] .53051 .63152 1 0 Mshowa
1 0 0 r
.65152 .65152 Mdot
0 0 0 r
[(67)] .63152 .63152 1 0 Mshowa
1 0 0 r
.75253 .65152 Mdot
0 0 0 r
[(68)] .73253 .63152 1 0 Mshowa
1 0 0 r
.85354 .65152 Mdot
0 0 0 r
[(69)] .83354 .63152 1 0 Mshowa
1 0 0 r
.95455 .65152 Mdot
0 0 0 r
[(70)] .93455 .63152 1 0 Mshowa
1 0 0 r
.04545 .75253 Mdot
0 0 0 r
[(71)] .02545 .73253 1 0 Mshowa
1 0 0 r
.14646 .75253 Mdot
0 0 0 r
[(72)] .12646 .73253 1 0 Mshowa
1 0 0 r
.24747 .75253 Mdot
0 0 0 r
[(73)] .22747 .73253 1 0 Mshowa
1 0 0 r
.34848 .75253 Mdot
0 0 0 r
[(74)] .32848 .73253 1 0 Mshowa
1 0 0 r
.44949 .75253 Mdot
0 0 0 r
[(75)] .42949 .73253 1 0 Mshowa
1 0 0 r
.55051 .75253 Mdot
0 0 0 r
[(76)] .53051 .73253 1 0 Mshowa
1 0 0 r
.65152 .75253 Mdot
0 0 0 r
[(77)] .63152 .73253 1 0 Mshowa
1 0 0 r
.75253 .75253 Mdot
0 0 0 r
[(78)] .73253 .73253 1 0 Mshowa
1 0 0 r
.85354 .75253 Mdot
0 0 0 r
[(79)] .83354 .73253 1 0 Mshowa
1 0 0 r
.95455 .75253 Mdot
0 0 0 r
[(80)] .93455 .73253 1 0 Mshowa
1 0 0 r
.04545 .85354 Mdot
0 0 0 r
[(81)] .02545 .83354 1 0 Mshowa
1 0 0 r
.14646 .85354 Mdot
0 0 0 r
[(82)] .12646 .83354 1 0 Mshowa
1 0 0 r
.24747 .85354 Mdot
0 0 0 r
[(83)] .22747 .83354 1 0 Mshowa
1 0 0 r
.34848 .85354 Mdot
0 0 0 r
[(84)] .32848 .83354 1 0 Mshowa
1 0 0 r
.44949 .85354 Mdot
0 0 0 r
[(85)] .42949 .83354 1 0 Mshowa
1 0 0 r
.55051 .85354 Mdot
0 0 0 r
[(86)] .53051 .83354 1 0 Mshowa
1 0 0 r
.65152 .85354 Mdot
0 0 0 r
[(87)] .63152 .83354 1 0 Mshowa
1 0 0 r
.75253 .85354 Mdot
0 0 0 r
[(88)] .73253 .83354 1 0 Mshowa
1 0 0 r
.85354 .85354 Mdot
0 0 0 r
[(89)] .83354 .83354 1 0 Mshowa
1 0 0 r
.95455 .85354 Mdot
0 0 0 r
[(90)] .93455 .83354 1 0 Mshowa
1 0 0 r
.04545 .95455 Mdot
0 0 0 r
[(91)] .02545 .93455 1 0 Mshowa
1 0 0 r
.14646 .95455 Mdot
0 0 0 r
[(92)] .12646 .93455 1 0 Mshowa
1 0 0 r
.24747 .95455 Mdot
0 0 0 r
[(93)] .22747 .93455 1 0 Mshowa
1 0 0 r
.34848 .95455 Mdot
0 0 0 r
[(94)] .32848 .93455 1 0 Mshowa
1 0 0 r
.44949 .95455 Mdot
0 0 0 r
[(95)] .42949 .93455 1 0 Mshowa
1 0 0 r
.55051 .95455 Mdot
0 0 0 r
[(96)] .53051 .93455 1 0 Mshowa
1 0 0 r
.65152 .95455 Mdot
0 0 0 r
[(97)] .63152 .93455 1 0 Mshowa
1 0 0 r
.75253 .95455 Mdot
0 0 0 r
[(98)] .73253 .93455 1 0 Mshowa
1 0 0 r
.85354 .95455 Mdot
0 0 0 r
[(99)] .83354 .93455 1 0 Mshowa
1 0 0 r
.95455 .95455 Mdot
0 0 0 r
[(100)] .93455 .93455 1 0 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[69]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`000_ooool3
o`0001Wooooo1Ol0000Iooooo`?o00006oooool3o`0001Wooooo0ol0000Jooooo`?o00006oooool0
0ol0003oooooooooo`0Iooooo`?o00006_ooool2o`0001Gooooo0ol00004ooooo`;o00005_ooool0
00?ooooo00?o0000ooooooooool06Oooool01?l0003oooooooooool0000Hooooo`05o`000?oooooo
ooooooooool000006oooool00ol0003oooooooooo`0Gooooo`05o`000?ooooooooooooooool00000
6?ooool01Ol0003ooooooooooooooooo000001[ooooo00?o0000ooooooooool06?ooool01Ol0003o
oooooooooooooooo000001_ooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`02oooo
o`04o`000?ooooooooooo`0001Gooooo0003ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool07?ooool00ol0003oooooooooo`0Fooooo`Go00007?ooool00ol0003oooooooooo`0Foooo
o`05o`000?ooooooooooooooool000006oooool00ol0003oooooooooo`0Gooooo`05o`000?oooooo
ooooooooool000007?ooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo00;ooooo00Co
0000oooooooooooo00005Oooool000?ooooo00?o0000ooooooooool06oooool00ol0003ooooooooo
o`0Iooooo`;o00006Oooool01?l0003oooooooooool0000Mooooo`03o`000?oooooooooo01Kooooo
1?l0000Looooo`03o`000?oooooooooo01Sooooo0ol0000Jooooo`Co00005?ooool00ol0003ooooo
ooooo`02ooooo`04o`000?ooooooooooo`0001Gooooo0003ooooo`03o`000?oooooooooo01Sooooo
00Go0000ooooooooooooooooo`00000Looooo`03o`000?oooooooooo01Oooooo00?o0000ooooool0
00006Oooool4o`0001Wooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`0Fooooo`05
o`000?ooooooooooooooool000006?ooool01Ol0003ooooooooooooooooo000001Cooooo00?o0000
ooooooooool00_ooool01?l0003oooooooooool0000Eooooo`000_ooool2o`0001[ooooo00Go0000
ooooooooooooooooo`00000Hooooo`05o`000?ooooooooooooooool000006_ooool2o`0001Wooooo
00?o0000ooooooooool06oooool00ol0003oooooooooo`0Iooooo`05o`000?ooooooooooooooool0
00006?ooool01Ol0003ooooooooooooooooo000001Sooooo00Go0000ooooooooooooooooo`00000C
ooooo`;o00001?ooool01?l0003oooooooooool0000Eooooo`000oooool00ol0003oooooooooo`05
ooooo`?oo`004Oooool3o`0000Kooooo0ooo000Aooooo`?o00001_ooool3ool001?ooooo00?o0000
ooooooooool01?ooool3ool0013ooooo1Ol00005ooooo`?oo`004_ooool2o`0000Kooooo0ooo000@
ooooo`Go00001Oooool3ool0017ooooo0ol00006ooooo`?oo`004Oooool3o`0000Kooooo0ooo000<
ooooo`03o`000?oooooooooo00?ooooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Oo
o`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo
o`005_ooool7ool000_ooooo0009ooooo`Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
2oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002_ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool0
01[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool0
01[ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000;ooooo0ol00008ooooo`03o`00
0?oooooooooo00Wooooo0ol00002ooooo`Go00001oooool00ol0003oooooooooo`09ooooo`?o0000
0oooool3o`0000Sooooo00?o0000ooooooooool02Oooool3o`0000Cooooo0ol00007ooooo`03o`00
0?oooooooooo00Wooooo0ol00003ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o0000
0oooool3o`0000Oooooo00?o0000ooooooooool02_ooool3o`0000Cooooo00?o0000ooooooooool0
1_ooool00ol0003oooooooooo`0:ooooo`?o00000oooool3o`0000Oooooo00?o0000ooooooooool0
2_ooool3o`0000?ooooo0_l00008ooooo`03o`000?oooooooooo00Wooooo1Ol00003ooooo`;o0000
1oooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool00_ooool01?l0003oooooooooool00007ooooo`03o`00
0?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Go
oooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Ko
oooo00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000000oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02_ooool01?l0003ooooooooo
ool00002ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0:ooooo`04o`000?ooooooooooooooo`Go
00001oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000000oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`000003
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02
ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00?ooooo0_l00008
ooooo`03o`000?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000002ooooo`03o`000?oo
oooooooo00Kooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02_ooool01?l0003oooooooooooooool4o`0000Oooooo
00?o0000ooooooooool02oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`?o00001oooool00ol0003ooooo
ooooo`0;ooooo`03o`000?oooooooooo00;ooooo1?l00006ooooo`03o`000?oooooooooo00cooooo
00Go0000ooooooooooooooooo`000002ooooo`03o`000?oooooooooo00Cooooo00?o0000oooooooo
ool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`05o`00
0?ooooooooooooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool00_ooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00[o
oooo00Co0000oooooooooooooooo1?l00008ooooo`03o`000?oooooooooo00[ooooo00Go0000oooo
ooooooooooooo`00000:ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool01Ol0003ooooooooo
oooooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05o`00
0?ooooooooooooooool000000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo
00cooooo0002ooooo`;o00002Oooool00ol0003oooooooooo`09ooooo`;o00000oooool01Ol0003o
oooooooooooooooo000000Oooooo00?o0000ooooooooool02Oooool2o`0000?ooooo00Go0000oooo
ooooooooooooo`000007ooooo`03o`000?oooooooooo00Wooooo0_l00005ooooo`;o00002?ooool0
0ol0003oooooooooo`09ooooo`;o00000oooool00ol0003oooooooooo`09ooooo`03o`000?oooooo
oooo00Wooooo0_l00004ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02_ooool2
o`0000?ooooo00Go0000ooooooooooooooooo`000006ooooo`03o`000?oooooooooo00[ooooo0_l0
0003ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003oooooooooo`0:ooooo`;o0000
0oooool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool01Ol0003o
oooooooooooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000oooool00ol0003oooooooooo`05ooooo`?oo`003?ooool00ol0003oooooooooo`02oooo
o`?o00001_ooool3ool000cooooo00?o0000ooooooooool00_ooool3o`0000Kooooo0ooo000<oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool3ool000cooooo00Co0000oooo
oooooooooooo1Ol00005ooooo`?oo`003?ooool00ol0003oooooooooo`03ooooo`;o00001_ooool3
ool000cooooo00Co0000oooooooooooooooo1Ol00005ooooo`?oo`003?ooool00ol0003ooooooooo
o`02ooooo`?o00001_ooool3ool000cooooo00?o0000ooooooooool00_ooool3o`0000Kooooo0ooo
000;ooooo`?o00001?ooool2o`0000Kooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000H
ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000H
ooooo`Goo`006?ooool5ool000cooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Foooo
o`Ooo`002oooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Oo
o`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`00
2Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`00
5_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3
ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3
ool001[ooooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000oooooooo
ool02?ooool5o`000003ooooool0003o000000?o00001oooool00ol0003oooooooooo`08ooooo`Go
00000_ooool3o`0000Sooooo00?o0000ooooooooool02?ooool5o`0000?ooooo0ol00007ooooo`03
o`000?oooooooooo00Sooooo1Ol00002ooooo`?o00002?ooool00ol0003oooooooooo`08ooooo`Go
00000_ooool3o`0000Oooooo00?o0000ooooooooool02Oooool5o`0000?ooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`09ooooo`Go00000_ooool3o`0000Oooooo00?o0000oooooooo
ool02Oooool5o`0000;ooooo0_l00008ooooo`03o`000?oooooooooo00[ooooo0ol00004ooooo`;o
00001oooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`00
0?oooooooooo00Wooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00007oooo
o`03o`000?oooooooooo00Wooooo00Ko0000oooooooooooo0000ooooool00003ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02Oooool01?l0003oooooooooool00004ooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool02Oooool01_l0003oooooooooool0003oooooo`00
00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`09ooooo`06o`000?oooooooooo
o`000?oooooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00[ooooo00Co
0000oooooooooooo00000oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00[ooooo
00Ko0000oooooooooooo0000ooooool00003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02_ooool01?l0003oooooooooool00004ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool02Oooool01Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo0000
1_ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00[ooooo00Co0000oooooooooooooooo1Ol00007ooooo`03o`000?oooooooooo00[ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[ooooo
00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooooooo
ool02oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool02oooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0=ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool0
0ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo
00_ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo
00_ooooo00?o0000ooooooooool00_ooool2o`0000Sooooo00?o0000ooooooooool02oooool01?l0
003oooooooooool00002ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02oooool0
0ol0003oooooooooo`04o`0000Oooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01?l0003ooooooooooooo
ool3o`0000Oooooo00?o0000ooooooooool03?ooool01?l0003oooooooooooooool4o`0000Kooooo
00?o0000ooooooooool02oooool2o`0000?ooooo00Co0000oooooooooooo00001_ooool00ol0003o
ooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Sooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`08ooooo`05o`000?ooooooooooooooool000001Oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000002ooooo`03o`000?oo
oooo000000Sooooo00?o0000ooooooooool02?ooool01_l0003ooooooooooooooooo0000ooooo`Co
00002?ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`00000:
ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000005ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool0
00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000oooooooo
ooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=
ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`0<ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooo
oooo0000ooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000
ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000002oooool00ol0003ooooo
ooooo`08ooooo`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooo
oooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?ooooooooooo`0000Kooooo00?o
0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01Oooool3ool000_ooooo0ol00003
ooooo`?o00001_ooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3ool000_ooooo0ol00005
ooooo`03o`000?oooooooooo00Cooooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;
ooooo`?o00001?ooool2o`0000Kooooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;
ooooo`?o00000oooool3o`0000Kooooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000;
ooooo`?o00001?ooool2o`0000Kooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool000cooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo
o`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo
o`002oooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`002Ooo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3ool0
01[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool0
01[ooooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000ooooooooool0
2Oooool3o`0000;ooooo1Ol00007ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o0000
2?ooool00ol0003oooooooooo`09ooooo`?o00001?ooool3o`0000Oooooo00?o0000ooooooooool0
2Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o0000
1oooool00ol0003oooooooooo`0:ooooo`?o00001?ooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:ooooo`?o0000
0oooool2o`0000Sooooo00?o0000ooooooooool02oooool3o`0000?ooooo0_l00007ooooo`03o`00
0?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?oo
ool01Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003o
ooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003o
ooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003o
oooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004oooo
o`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01Ol0003ooooooooooooooooo
000000;ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0
003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Go0000ooooooooooooooooo`000009
ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000001?l00007ooooo`03o`000?oooooo
oooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00cooooo00?o0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo1Ol00002ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?oo
ool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`;o00001Oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00[ooooo0_l00004ooooo`;o00002?ooool0
0ol0003oooooooooo`0:ooooo`;o00000_ooool01?l0003oooooooooool00008ooooo`03o`000?oo
oooooooo00[ooooo0_l00006ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02_oo
ool2o`0000;ooooo1?l00007ooooo`03o`000?oooooooooo00_ooooo0_l00005ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02oooool2o`0000?ooooo0ol00007ooooo`03o`000?oo
oooooooo00_ooooo0_l00003ooooo`Co00001_ooool00ol0003oooooooooo`09ooooo`04o`000?oo
ooooooooo`0000?ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`00
0oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool00000
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00Ko0000oooooooooooo
0000ooooool00008ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000ol00008oooo
o`03o`000?oooooooooo00cooooo00?o0000ooooool000002_ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooo000000?ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooooooo
0000ooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo
0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000oooo
ooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000002oooool00ol0003ooooooooo
o`08ooooo`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03o`00
0?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003oooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooo
oooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_o
oooo0_l00003ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?o
oooo00?o0000ooooooooool01Oooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3ool000_o
oooo0ol00003ooooo`?o00001_ooool3ool000_ooooo0ol00005ooooo`03o`000?oooooooooo00Co
oooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;ooooo`?o00001?ooool2o`0000Ko
oooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;ooooo`?o00000oooool3o`0000Ko
oooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000=ooooo`03o`000?oooooooooo00;o
oooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000<ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo0009
ooooo`Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000F
o`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`002oooool000Wooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`00
6_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`00
3Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000;ooooo0ol00008ooooo`03o`000?oooooooooo00[ooooo0ol00000
0ooooooo0000o`000003o`0000Oooooo00?o0000ooooooooool02_ooool3o`0000;ooooo0ol00008
ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:
ooooo`?o00000_ooool3o`0000Sooooo00?o0000ooooooooool02_ooool3o`0000;ooooo0ol00007
ooooo`03o`000?oooooooooo00_ooooo0ol00003ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool02oooool3o`0000;ooooo0ol00007ooooo`03o`000?oooooooooo00_ooooo0ol00002
ooooo`;o00002?ooool00ol0003oooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000
ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0;oooo
o`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00_ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`05ooooo`03o`00
0?oooooooooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`04ooooo`03
o`000?oooooooooo00cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo00cooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`04oooo
o`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?oooooo
ooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`Go00000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooo
oooo00Sooooo1Ol00005ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool5
o`000003ooooool0003o000000?o00001oooool00ol0003oooooooooo`08ooooo`Go00001Oooool0
0ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo1Ol000000ooooooo0000ooooo`02
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool5o`0000Cooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`09ooooo`Go000000?oooooo`000?ooool00_ooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo1Ol00005ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool03Oooool01?l0003oooooooooool00002ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`04o`000?ooooooooooo`0000Gooooo00?o0000ooooooooool01_oo
ool00ol0003oooooooooo`08ooooo`04o`000?ooooooooooo`0000Cooooo0_l00008ooooo`03o`00
0?oooooooooo00Sooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00008oooo
o`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00000_ooool4o`0000Oooooo00?o
0000ooooooooool02Oooool01?l0003oooooooooool00005ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02Oooool01?l0003oooooooooool00003ooooo`?o00001oooool00ol0003o
ooooooooo`09ooooo`04o`000?ooooooooooo`0000?ooooo1?l00006ooooo`03o`000?oooooooooo
00gooooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool00ol0
003oooooo`000002ooooo`05o`000?ooooooooooooooool000001oooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000Kooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000?ooooo00?o0000ooooool000002?ooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000;ooooo1?l00008ooooo`03o`000?oooooooooo00Wooooo00?o0000
ooooool000000_ooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooool000001_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooool000000_ooool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02_oo
ool00ol0003oooooo`000002ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003ooooo
ooooo`09ooooo`Co00000oooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo00co
oooo0002ooooo`;o00002Oooool00ol0003oooooooooo`0:ooooo`;o00000_ooool01Ol0003ooooo
oooooooooooo000000Oooooo00?o0000ooooooooool02_ooool2o`0000;ooooo00Go0000oooooooo
ooooooooo`000007ooooo`03o`000?oooooooooo00[ooooo0_l00004ooooo`;o00002?ooool00ol0
003oooooooooo`0:ooooo`;o00000_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo
00[ooooo0_l00003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02oooool2o`00
00;ooooo00Go0000ooooooooooooooooo`000006ooooo`03o`000?oooooooooo00_ooooo0_l00002
ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003oooooooooo`0;ooooo`;o00000_oo
ool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool00ol0003ooooo
ooooo`04ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01Oooool3ool000gooooo00Co0000oooooooooooooooo0ol00006ooooo`?o
o`003Oooool01?l0003oooooooooooooool3o`0000Kooooo0ooo000=ooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool01?ooool3ool000gooooo00?o0000ooooooooool01Ol00005oooo
o`?oo`003Oooool00ol0003oooooooooo`02ooooo`;o00001_ooool3ool000gooooo00?o0000oooo
ooooool01Ol00005ooooo`?oo`003Oooool01?l0003oooooooooooooool3o`0000Kooooo0ooo000=
ooooo`04o`000?ooooooooooooooo`?o00001_ooool3ool000[ooooo1Ol00003ooooo`;o00001_oo
ool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`003?ooool0
00Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002Oooool7ool001Ko
00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko
00001ooo000Fo`0000Ooo`005_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo
1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo
1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo
000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo
000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000J
ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool000gooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00co
oooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o00000_ooool5o`0000Oo
oooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wo
oooo0ol00004ooooo`?o00001oooool00ol0003oooooooooo`09ooooo`?o00000oooool3o`0000So
oooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00[o
oooo0ol00004ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02_ooool3o`0000?o
oooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`;o00002?ooool00ol0003o
ooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool0
00000_ooool01?l0003oooooooooool00007ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`08ooooo`05o`000?ooooooooooooooool000001?ooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooo
oooooooo000000?ooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`09ooooo`07o`00
0?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02Oooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003ooooooooo
ool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool03?ooool01Ol0003ooooooooooooooooo000000Wooooo00?o0000ooooooooool0
3?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0
3?ooool00ol0003oooooo`000004o`0000Oooooo00?o0000ooooooooool03?ooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool00ol0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool00ol0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01Ol0003ooooo
oooooooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<oooo
o`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00?o0000oooooooo
ool00_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00Co0000oooooooo
oooooooo0_l00008ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000_ooool00ol0
003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000ol00007oooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool00ol00007ooooo`03o`000?oooooooooo
00gooooo00?o0000ooooooooool01?l00006ooooo`03o`000?oooooooooo00Wooooo1?l00003oooo
o`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`08ooooo`Co00000_ooool01Ol0003ooooooooooooooooo
000000Oooooo00?o0000ooooooooool02?ooool4o`0000Kooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`08ooooo`Co00000oooool00ol0003oooooo`000008ooooo`03o`000?oooooo
oooo00Sooooo1?l00002ooooo`Co00002?ooool00ol0003oooooooooo`08ooooo`Co00000_ooool0
0ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Wooooo1?l00006ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool02Oooool4o`0000;ooooo00Go0000ooooooooooooooooo`00
0006ooooo`03o`000?oooooooooo00Wooooo1?l00002ooooo`05o`000?ooooooooooooooool00000
1_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Cooooo00Co0000oooooooooooo
00001_ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool0
2?ooool00ol0003oooooooooo`03ooooo`05o`000?ooooooooooooooool000001oooool00ol0003o
ooooooooo`08ooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`000007oooo
o`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01Oooool2o`0000Sooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Wooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Oooooo00?o0000oooo
ooooool02Oooool00ol0003oooooooooo`03ooooo`05o`000?ooooooooooooooool000001_ooool0
0ol0003oooooooooo`09ooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`00
0006ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00oooool01Ol0003ooooooooo
oooooooo000000Kooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`03ooooo`04o`00
0?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool0
1Oooool3ool000[ooooo1Ol00002ooooo`?o00001_ooool3ool000[ooooo1Ol00002ooooo`?o0000
1_ooool3ool000[ooooo1Ol00004ooooo`03o`000?oooooooooo00Cooooo0ooo000:ooooo`Go0000
00?oooooo`000?l000000ol00005ooooo`?oo`002_ooool5o`0000?ooooo0_l00006ooooo`?oo`00
2_ooool5o`000003ooooool0003o000000?o00001Oooool3ool000[ooooo1Ol00002ooooo`?o0000
1_ooool3ool000[ooooo1Ol00002ooooo`?o00001_ooool3ool000cooooo0_l00004ooooo`;o0000
1_ooool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`003?oo
ool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007
ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`002Oooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool000gooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o00000_ooool5o`00
00Oooooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo
00Wooooo0ol00004ooooo`?o00001oooool00ol0003oooooooooo`09ooooo`?o00000oooool3o`00
00Sooooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo
00[ooooo0ol00004ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02_ooool3o`00
00?ooooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`;o00002?ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00?ooooo0_l00007ooooo`03o`000?oooooooooo
00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool01Ol0003o
oooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003oooooooooo`08
ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooo
oooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo
00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o0000oooooooo
ool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04
ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`04o`000?oo
ooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01ooo
ool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000005ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01_l0003ooooooooooooooooo0000
ooooo`Go00001oooool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool000001Ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05
o`000?ooooooooooooooool000001?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo
00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000001Oooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00cooooo00Go0000ooooooooooooooooo`000002ooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`08ooooo`Co00001Oooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00Sooooo1?l00004ooooo`;o00002?ooool00ol0003oooooooooo`08ooooo`Co0000
0_ooool01?l0003oooooooooool00008ooooo`03o`000?oooooooooo00Sooooo1?l00006ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool4o`0000;ooooo1?l00007ooooo`03
o`000?oooooooooo00Wooooo1?l00005ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool02Oooool4o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00Wooooo1?l00003ooooo`Co
00001_ooool00ol0003oooooooooo`0<ooooo`05o`000?ooooooooooooooool000000_ooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`03ooooo`05o`000?oooooooooooooo
ool000001oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Oooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ool000002?ooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00?ooooo1?l00008oooo
o`03o`000?oooooooooo00Sooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`08oooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`04oooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00oooool01Ol0003ooooooooooooooooo
000000Kooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`03ooooo`05o`000?oooooo
ooooooooool000001_ooool00ol0003oooooooooo`0=ooooo`04o`000?ooooooooooo`0000;ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`0000Wooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`02ooooo`05o`000?ooooooooooooooool00000
1oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00;ooooo00Go0000oooooooooooo
ooooo`000007ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01?ooool2o`0000So
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Wo
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`02ooooo`05o`000?oooooooooooooo
ool000001_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00;ooooo00Go0000oooo
ooooooooooooo`000006ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool00_ooool0
1Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooo
oooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`00
0oooool00ol0003oooooooooo`05ooooo`?oo`003?ooool2o`0000?ooooo0ol00006ooooo`?oo`00
3?ooool2o`0000?ooooo0ol00006ooooo`?oo`003?ooool2o`0000Gooooo00?o0000ooooooooool0
1?ooool3ool000cooooo0_l00002ooooo`Go00001Oooool3ool000cooooo0_l00004ooooo`;o0000
1_ooool3ool000cooooo0_l00002ooooo`Go00001Oooool3ool000cooooo0_l00003ooooo`?o0000
1_ooool3ool000cooooo0_l00003ooooo`?o00001_ooool3ool000[ooooo1Ol00003ooooo`;o0000
1_ooool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`003?oo
ool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007
ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`002Oooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool000gooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`0:ooooo`04o`000?oooooooooo
ooooo`Go00001oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00;ooooo0ol00008
ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool00oooool3o`0000Oooooo00?o0000
ooooooooool02_ooool00ol0003oooooooooo`02ooooo`?o00002?ooool00ol0003oooooooooo`0:
ooooo`03o`000?oooooooooo00;ooooo0ol00007ooooo`03o`000?oooooooooo00_ooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00?o0000
ooooooooool00_ooool3o`0000Oooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02
ooooo`;o00002?ooool00ol0003oooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000
ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooooooo00;ooooo00Co0000oooooooooooo00001oooool00ol0003oooooooooo`0:
ooooo`05o`000?ooooooooooooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00[ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000oooooooo
ool01?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`0;ooooo`05o`000?ooooooooooooooool000000oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002
ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Go00001oooool0
0ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`0;ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`0<ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003o
ooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Ko
oooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`;o00002?ooool00ol0003o
ooooooooo`0;ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01_ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Co00001oooool00ol0003oooooooooo`0<oooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<oooo
o`04o`000?ooooooooooooooo`?o00001oooool00ol0003oooooooooo`0<ooooo`04o`000?oooooo
ooooooooo`Co00001_ooool00ol0003oooooooooo`0:ooooo`?o00000oooool01?l0003ooooooooo
ool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool03?ooool00ol0003oooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool03?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool03?ooool01_l0003oooooooooool0003oooooo`0000Sooooo00?o0000oooooooo
ool03?ooool00ol0003oooooo`000003o`0000Sooooo00?o0000ooooooooool03?ooool00ol0003o
ooooo`00000:ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool00oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00gooooo00?o0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00gooooo00?o0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002oooo
o`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000;ooooo0_l00009oooo
o`03o`000?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o
0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003o
ooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003o
oooooooooooooooo000000?ooooo0_l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000_ooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooo
oooooooo000000;ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`09ooooo`07o`00
0?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?o
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05o`000?oooooooooooooo
ool000000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003oooo
o`03o`000?oooooooooo00Gooooo0ooo000:ooooo`Go00000_ooool3o`0000Kooooo0ooo000:oooo
o`Go00000_ooool3o`0000Kooooo0ooo000:ooooo`Go00001?ooool00ol0003oooooooooo`04oooo
o`?oo`002_ooool5o`000003ooooool0003o000000?o00001Oooool3ool000[ooooo1Ol00003oooo
o`;o00001_ooool3ool000[ooooo1Ol000000ooooooo0000o`000003o`0000Gooooo0ooo000:oooo
o`Go00000_ooool3o`0000Kooooo0ooo000:ooooo`Go00000_ooool3o`0000Kooooo0ooo000;oooo
o`?o00001?ooool2o`0000Kooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool000cooooo0009ooooo`Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
2oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002Oooool7
ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7
ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`006?ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3ool001[o
oooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[o
oooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000ooooooooool02Ooo
ool3o`0000;ooooo1Ol00007ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o00002?oo
ool00ol0003oooooooooo`09ooooo`?o00001?ooool3o`0000Oooooo00?o0000ooooooooool02Ooo
ool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o00001ooo
ool00ol0003oooooooooo`0:ooooo`?o00001?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:ooooo`?o00000ooo
ool2o`0000Sooooo00?o0000ooooooooool02_ooool2o`0000Gooooo0_l00007ooooo`03o`000?oo
oooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool0
1Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003ooooo
ooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003ooooo
o`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003ooooo
oooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00Wooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01Ol0003ooooooooooooooooo0000
00;ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000003oooo
o`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo
000000Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`06o`000?oooooo
ooooooooool0003ooooo1Ol00007ooooo`03o`000?oooooooooo00Sooooo00Go0000oooooooooooo
ooooo`000005ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003o
oooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000005oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool01?l0003oooooooooool00002
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`09ooooo`?o00001Oooool00ol0003oooooooooo`06oooo
o`03o`000?oooooooooo00Wooooo0ol00004ooooo`;o00002?ooool00ol0003oooooooooo`09oooo
o`?o00000_ooool01?l0003oooooooooool00008ooooo`03o`000?oooooooooo00Wooooo0ol00006
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02Oooool3o`0000;ooooo1?l00007
ooooo`03o`000?oooooooooo00[ooooo0ol00005ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool02_ooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003
ooooo`Co00001_ooool00ol0003oooooooooo`0:ooooo`Co00000_ooool01?l0003oooooooooool0
0006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000
ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooo
ooooo`05ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000005ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo0000
00;ooooo00?o0000ooooool000002?ooool00ol0003oooooooooo`08ooooo`06o`000?oooooooooo
ooooool0003ooooo1?l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000[ooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooooooooooo0000
00Gooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?oooooooooo
ooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Ooo
ool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03
o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?oooooooooo
o`0000Kooooo00?o0000ooooooooool03?ooool000;ooooo0_l00009ooooo`03o`000?oooooooooo
00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo0000
00?ooooo0_l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oo
oooo000000_ooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000;ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0
003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0
003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oo
oooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooooooo
ool01?ooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0
003oooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo
00Gooooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000;ooooo`?o00000oooool3o`00
00Kooooo0ooo000;ooooo`?o00001Oooool00ol0003oooooooooo`04ooooo`?oo`002oooool3o`00
00;ooooo1Ol00005ooooo`?oo`002oooool3o`0000Cooooo0_l00006ooooo`?oo`002oooool3o`00
00;ooooo1Ol00005ooooo`?oo`002oooool3o`0000?ooooo0ol00006ooooo`?oo`002oooool3o`00
00?ooooo0ol00006ooooo`?oo`002oooool3o`0000Cooooo0_l00006ooooo`?oo`003Oooool000[o
oooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001So
oooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Ko0000
1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko0000
1ooo000Fo`0000Ooo`005_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`002oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000F
ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;
ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool000cooooo000;ooooo`?o
o`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?o
o`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000;ooooo0ol0
0008ooooo`03o`000?oooooooooo00Wooooo0_l00003ooooo`Go00001oooool00ol0003ooooooooo
o`09ooooo`;o00001?ooool3o`0000Sooooo00?o0000ooooooooool02Oooool2o`0000Gooooo0ol0
0007ooooo`03o`000?oooooooooo00Wooooo0_l00004ooooo`?o00002?ooool00ol0003ooooooooo
o`09ooooo`;o00001?ooool3o`0000Oooooo00?o0000ooooooooool02_ooool2o`0000Gooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`0:ooooo`;o00001?ooool3o`0000Oooooo00?o
0000ooooooooool02_ooool2o`0000Cooooo0_l00008ooooo`03o`000?oooooooooo00Cooooo0ol0
0004ooooo`;o00001?ooool2o`0000Oooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`05o`000?ooooooooooooooool000000_oo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo0000
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo
00000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooo
oooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00Co0000oooo
oooooooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool00_ooool01?l0003oooooooooool00002ooooo`04o`000?ooooooooooo`0000Kooooo
00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003ooooooooo
o`0<ooooo`05o`000?ooooooooooooooool000002Oooool00ol0003oooooooooo`0<ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<ooooo`03o`00
0?oooooo000000Co00001oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooo000000?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00;ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooo000000?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00;ooooo
00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo
00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool4o`0000Co
oooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`09ooooo`Co00000oooool2o`0000So
oooo00?o0000ooooooooool02Oooool4o`000005ooooool0003oooooooooool000002?ooool00ol0
003oooooooooo`09ooooo`Co00001Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo
00Wooooo1?l000000ooooooo0000o`000002o`0000Oooooo00?o0000ooooooooool02_ooool4o`00
00Cooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0:ooooo`Co00000_ooool3o`00
00Oooooo00?o0000ooooooooool02_ooool4o`0000;ooooo1?l00006ooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool00_ooool01?l0003oooooooooool00002ooooo`04o`000?oooooo
ooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`00
0?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Go
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`05o`000?oooooooooooooo
ool000000_ooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00Sooooo00Ko0000oooo
ooooooooooooo`000?ooool4o`0000Sooooo00?o0000ooooooooool02?ooool01ol0003ooooooooo
oooooooo0000ooooool000002_ooool00ol0003oooooooooo`09ooooo`05o`000?oooooooooooooo
ool000001Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`09ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Co
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`00
00;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`00
00Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000000ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`05
o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000ooooooooool02?ooool01ol0
003ooooooooooooooooo0000ooooool000002oooool00ol0003oooooooooo`08ooooo`05o`000?oo
ooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Wooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool0
00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo0_l00004ooooo`04
o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000oooool00ol0003oooooooooo`05ooooo`?oo`002oooool3o`0000?ooooo0ol00006oooo
o`?oo`002oooool3o`0000?ooooo0ol00006ooooo`?oo`002oooool3o`0000Gooooo00?o0000oooo
ooooool01?ooool3ool000_ooooo0ol00002ooooo`Go00001Oooool3ool000_ooooo0ol00004oooo
o`;o00001_ooool3ool000_ooooo0ol00002ooooo`Go00001Oooool3ool000_ooooo0ol00003oooo
o`?o00001_ooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3ool000Kooooo00?o0000oooo
ooooool00oooool2o`0000Cooooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`002ooo
ool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7
ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002_ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[o
oooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[o
oooo0ooo000Jooooo`?oo`003Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500054, -0.0500054, \
0.00383279, 0.00383279}},
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell["Here is a subset of 20 agents:", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[{
    \(\(someindexes = 
        RandomKSubset[Range[100], 20];\)\), "\[IndentingNewLine]", 
    \(\(someagents = Agents[m, someindexes];\)\)}], "Input",
  CellLabel->"In[73]:=",
  CellTags->"LocalInteractions:2"],

Cell["We can now change their strategy to 2.", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(SetStrategies[
      someagents, \[IndentingNewLine]Table[
        2, {20}]\[IndentingNewLine]]\)], "Input",
  CellLabel->"In[75]:=",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowLabeledGraph[m];\)\)], "Input",
  CellLabel->"In[76]:=",
  CellTags->"LocalInteractions:2"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.04545 .04545 m
.14646 .04545 L
s
.04545 .04545 m
.04545 .14646 L
s
.14646 .04545 m
.24747 .04545 L
s
.14646 .04545 m
.14646 .14646 L
s
.24747 .04545 m
.34848 .04545 L
s
.24747 .04545 m
.24747 .14646 L
s
.34848 .04545 m
.44949 .04545 L
s
.34848 .04545 m
.34848 .14646 L
s
.44949 .04545 m
.55051 .04545 L
s
.44949 .04545 m
.44949 .14646 L
s
.55051 .04545 m
.65152 .04545 L
s
.55051 .04545 m
.55051 .14646 L
s
.65152 .04545 m
.75253 .04545 L
s
.65152 .04545 m
.65152 .14646 L
s
.75253 .04545 m
.85354 .04545 L
s
.75253 .04545 m
.75253 .14646 L
s
.85354 .04545 m
.95455 .04545 L
s
.85354 .04545 m
.85354 .14646 L
s
.95455 .04545 m
.95455 .14646 L
s
.04545 .14646 m
.14646 .14646 L
s
.04545 .14646 m
.04545 .24747 L
s
.14646 .14646 m
.24747 .14646 L
s
.14646 .14646 m
.14646 .24747 L
s
.24747 .14646 m
.34848 .14646 L
s
.24747 .14646 m
.24747 .24747 L
s
.34848 .14646 m
.44949 .14646 L
s
.34848 .14646 m
.34848 .24747 L
s
.44949 .14646 m
.55051 .14646 L
s
.44949 .14646 m
.44949 .24747 L
s
.55051 .14646 m
.65152 .14646 L
s
.55051 .14646 m
.55051 .24747 L
s
.65152 .14646 m
.75253 .14646 L
s
.65152 .14646 m
.65152 .24747 L
s
.75253 .14646 m
.85354 .14646 L
s
.75253 .14646 m
.75253 .24747 L
s
.85354 .14646 m
.95455 .14646 L
s
.85354 .14646 m
.85354 .24747 L
s
.95455 .14646 m
.95455 .24747 L
s
.04545 .24747 m
.14646 .24747 L
s
.04545 .24747 m
.04545 .34848 L
s
.14646 .24747 m
.24747 .24747 L
s
.14646 .24747 m
.14646 .34848 L
s
.24747 .24747 m
.34848 .24747 L
s
.24747 .24747 m
.24747 .34848 L
s
.34848 .24747 m
.44949 .24747 L
s
.34848 .24747 m
.34848 .34848 L
s
.44949 .24747 m
.55051 .24747 L
s
.44949 .24747 m
.44949 .34848 L
s
.55051 .24747 m
.65152 .24747 L
s
.55051 .24747 m
.55051 .34848 L
s
.65152 .24747 m
.75253 .24747 L
s
.65152 .24747 m
.65152 .34848 L
s
.75253 .24747 m
.85354 .24747 L
s
.75253 .24747 m
.75253 .34848 L
s
.85354 .24747 m
.95455 .24747 L
s
.85354 .24747 m
.85354 .34848 L
s
.95455 .24747 m
.95455 .34848 L
s
.04545 .34848 m
.14646 .34848 L
s
.04545 .34848 m
.04545 .44949 L
s
.14646 .34848 m
.24747 .34848 L
s
.14646 .34848 m
.14646 .44949 L
s
.24747 .34848 m
.34848 .34848 L
s
.24747 .34848 m
.24747 .44949 L
s
.34848 .34848 m
.44949 .34848 L
s
.34848 .34848 m
.34848 .44949 L
s
.44949 .34848 m
.55051 .34848 L
s
.44949 .34848 m
.44949 .44949 L
s
.55051 .34848 m
.65152 .34848 L
s
.55051 .34848 m
.55051 .44949 L
s
.65152 .34848 m
.75253 .34848 L
s
.65152 .34848 m
.65152 .44949 L
s
.75253 .34848 m
.85354 .34848 L
s
.75253 .34848 m
.75253 .44949 L
s
.85354 .34848 m
.95455 .34848 L
s
.85354 .34848 m
.85354 .44949 L
s
.95455 .34848 m
.95455 .44949 L
s
.04545 .44949 m
.14646 .44949 L
s
.04545 .44949 m
.04545 .55051 L
s
.14646 .44949 m
.24747 .44949 L
s
.14646 .44949 m
.14646 .55051 L
s
.24747 .44949 m
.34848 .44949 L
s
.24747 .44949 m
.24747 .55051 L
s
.34848 .44949 m
.44949 .44949 L
s
.34848 .44949 m
.34848 .55051 L
s
.44949 .44949 m
.55051 .44949 L
s
.44949 .44949 m
.44949 .55051 L
s
.55051 .44949 m
.65152 .44949 L
s
.55051 .44949 m
.55051 .55051 L
s
.65152 .44949 m
.75253 .44949 L
s
.65152 .44949 m
.65152 .55051 L
s
.75253 .44949 m
.85354 .44949 L
s
.75253 .44949 m
.75253 .55051 L
s
.85354 .44949 m
.95455 .44949 L
s
.85354 .44949 m
.85354 .55051 L
s
.95455 .44949 m
.95455 .55051 L
s
.04545 .55051 m
.14646 .55051 L
s
.04545 .55051 m
.04545 .65152 L
s
.14646 .55051 m
.24747 .55051 L
s
.14646 .55051 m
.14646 .65152 L
s
.24747 .55051 m
.34848 .55051 L
s
.24747 .55051 m
.24747 .65152 L
s
.34848 .55051 m
.44949 .55051 L
s
.34848 .55051 m
.34848 .65152 L
s
.44949 .55051 m
.55051 .55051 L
s
.44949 .55051 m
.44949 .65152 L
s
.55051 .55051 m
.65152 .55051 L
s
.55051 .55051 m
.55051 .65152 L
s
.65152 .55051 m
.75253 .55051 L
s
.65152 .55051 m
.65152 .65152 L
s
.75253 .55051 m
.85354 .55051 L
s
.75253 .55051 m
.75253 .65152 L
s
.85354 .55051 m
.95455 .55051 L
s
.85354 .55051 m
.85354 .65152 L
s
.95455 .55051 m
.95455 .65152 L
s
.04545 .65152 m
.14646 .65152 L
s
.04545 .65152 m
.04545 .75253 L
s
.14646 .65152 m
.24747 .65152 L
s
.14646 .65152 m
.14646 .75253 L
s
.24747 .65152 m
.34848 .65152 L
s
.24747 .65152 m
.24747 .75253 L
s
.34848 .65152 m
.44949 .65152 L
s
.34848 .65152 m
.34848 .75253 L
s
.44949 .65152 m
.55051 .65152 L
s
.44949 .65152 m
.44949 .75253 L
s
.55051 .65152 m
.65152 .65152 L
s
.55051 .65152 m
.55051 .75253 L
s
.65152 .65152 m
.75253 .65152 L
s
.65152 .65152 m
.65152 .75253 L
s
.75253 .65152 m
.85354 .65152 L
s
.75253 .65152 m
.75253 .75253 L
s
.85354 .65152 m
.95455 .65152 L
s
.85354 .65152 m
.85354 .75253 L
s
.95455 .65152 m
.95455 .75253 L
s
.04545 .75253 m
.14646 .75253 L
s
.04545 .75253 m
.04545 .85354 L
s
.14646 .75253 m
.24747 .75253 L
s
.14646 .75253 m
.14646 .85354 L
s
.24747 .75253 m
.34848 .75253 L
s
.24747 .75253 m
.24747 .85354 L
s
.34848 .75253 m
.44949 .75253 L
s
.34848 .75253 m
.34848 .85354 L
s
.44949 .75253 m
.55051 .75253 L
s
.44949 .75253 m
.44949 .85354 L
s
.55051 .75253 m
.65152 .75253 L
s
.55051 .75253 m
.55051 .85354 L
s
.65152 .75253 m
.75253 .75253 L
s
.65152 .75253 m
.65152 .85354 L
s
.75253 .75253 m
.85354 .75253 L
s
.75253 .75253 m
.75253 .85354 L
s
.85354 .75253 m
.95455 .75253 L
s
.85354 .75253 m
.85354 .85354 L
s
.95455 .75253 m
.95455 .85354 L
s
.04545 .85354 m
.14646 .85354 L
s
.04545 .85354 m
.04545 .95455 L
s
.14646 .85354 m
.24747 .85354 L
s
.14646 .85354 m
.14646 .95455 L
s
.24747 .85354 m
.34848 .85354 L
s
.24747 .85354 m
.24747 .95455 L
s
.34848 .85354 m
.44949 .85354 L
s
.34848 .85354 m
.34848 .95455 L
s
.44949 .85354 m
.55051 .85354 L
s
.44949 .85354 m
.44949 .95455 L
s
.55051 .85354 m
.65152 .85354 L
s
.55051 .85354 m
.55051 .95455 L
s
.65152 .85354 m
.75253 .85354 L
s
.65152 .85354 m
.65152 .95455 L
s
.75253 .85354 m
.85354 .85354 L
s
.75253 .85354 m
.75253 .95455 L
s
.85354 .85354 m
.95455 .85354 L
s
.85354 .85354 m
.85354 .95455 L
s
.95455 .85354 m
.95455 .95455 L
s
.04545 .95455 m
.14646 .95455 L
s
.14646 .95455 m
.24747 .95455 L
s
.24747 .95455 m
.34848 .95455 L
s
.34848 .95455 m
.44949 .95455 L
s
.44949 .95455 m
.55051 .95455 L
s
.55051 .95455 m
.65152 .95455 L
s
.65152 .95455 m
.75253 .95455 L
s
.75253 .95455 m
.85354 .95455 L
s
.85354 .95455 m
.95455 .95455 L
s
0 1 0 r
.025 w
.04545 .04545 Mdot
0 0 0 r
[(1)] .02545 .02545 1 0 Mshowa
1 0 0 r
.14646 .04545 Mdot
0 0 0 r
[(2)] .12646 .02545 1 0 Mshowa
0 1 0 r
.24747 .04545 Mdot
0 0 0 r
[(3)] .22747 .02545 1 0 Mshowa
1 0 0 r
.34848 .04545 Mdot
0 0 0 r
[(4)] .32848 .02545 1 0 Mshowa
1 0 0 r
.44949 .04545 Mdot
0 0 0 r
[(5)] .42949 .02545 1 0 Mshowa
0 1 0 r
.55051 .04545 Mdot
0 0 0 r
[(6)] .53051 .02545 1 0 Mshowa
1 0 0 r
.65152 .04545 Mdot
0 0 0 r
[(7)] .63152 .02545 1 0 Mshowa
0 1 0 r
.75253 .04545 Mdot
0 0 0 r
[(8)] .73253 .02545 1 0 Mshowa
1 0 0 r
.85354 .04545 Mdot
0 0 0 r
[(9)] .83354 .02545 1 0 Mshowa
1 0 0 r
.95455 .04545 Mdot
0 0 0 r
[(10)] .93455 .02545 1 0 Mshowa
1 0 0 r
.04545 .14646 Mdot
0 0 0 r
[(11)] .02545 .12646 1 0 Mshowa
1 0 0 r
.14646 .14646 Mdot
0 0 0 r
[(12)] .12646 .12646 1 0 Mshowa
1 0 0 r
.24747 .14646 Mdot
0 0 0 r
[(13)] .22747 .12646 1 0 Mshowa
1 0 0 r
.34848 .14646 Mdot
0 0 0 r
[(14)] .32848 .12646 1 0 Mshowa
1 0 0 r
.44949 .14646 Mdot
0 0 0 r
[(15)] .42949 .12646 1 0 Mshowa
1 0 0 r
.55051 .14646 Mdot
0 0 0 r
[(16)] .53051 .12646 1 0 Mshowa
0 1 0 r
.65152 .14646 Mdot
0 0 0 r
[(17)] .63152 .12646 1 0 Mshowa
0 1 0 r
.75253 .14646 Mdot
0 0 0 r
[(18)] .73253 .12646 1 0 Mshowa
1 0 0 r
.85354 .14646 Mdot
0 0 0 r
[(19)] .83354 .12646 1 0 Mshowa
0 1 0 r
.95455 .14646 Mdot
0 0 0 r
[(20)] .93455 .12646 1 0 Mshowa
1 0 0 r
.04545 .24747 Mdot
0 0 0 r
[(21)] .02545 .22747 1 0 Mshowa
1 0 0 r
.14646 .24747 Mdot
0 0 0 r
[(22)] .12646 .22747 1 0 Mshowa
1 0 0 r
.24747 .24747 Mdot
0 0 0 r
[(23)] .22747 .22747 1 0 Mshowa
1 0 0 r
.34848 .24747 Mdot
0 0 0 r
[(24)] .32848 .22747 1 0 Mshowa
1 0 0 r
.44949 .24747 Mdot
0 0 0 r
[(25)] .42949 .22747 1 0 Mshowa
1 0 0 r
.55051 .24747 Mdot
0 0 0 r
[(26)] .53051 .22747 1 0 Mshowa
0 1 0 r
.65152 .24747 Mdot
0 0 0 r
[(27)] .63152 .22747 1 0 Mshowa
1 0 0 r
.75253 .24747 Mdot
0 0 0 r
[(28)] .73253 .22747 1 0 Mshowa
1 0 0 r
.85354 .24747 Mdot
0 0 0 r
[(29)] .83354 .22747 1 0 Mshowa
1 0 0 r
.95455 .24747 Mdot
0 0 0 r
[(30)] .93455 .22747 1 0 Mshowa
1 0 0 r
.04545 .34848 Mdot
0 0 0 r
[(31)] .02545 .32848 1 0 Mshowa
1 0 0 r
.14646 .34848 Mdot
0 0 0 r
[(32)] .12646 .32848 1 0 Mshowa
1 0 0 r
.24747 .34848 Mdot
0 0 0 r
[(33)] .22747 .32848 1 0 Mshowa
1 0 0 r
.34848 .34848 Mdot
0 0 0 r
[(34)] .32848 .32848 1 0 Mshowa
1 0 0 r
.44949 .34848 Mdot
0 0 0 r
[(35)] .42949 .32848 1 0 Mshowa
1 0 0 r
.55051 .34848 Mdot
0 0 0 r
[(36)] .53051 .32848 1 0 Mshowa
1 0 0 r
.65152 .34848 Mdot
0 0 0 r
[(37)] .63152 .32848 1 0 Mshowa
1 0 0 r
.75253 .34848 Mdot
0 0 0 r
[(38)] .73253 .32848 1 0 Mshowa
1 0 0 r
.85354 .34848 Mdot
0 0 0 r
[(39)] .83354 .32848 1 0 Mshowa
1 0 0 r
.95455 .34848 Mdot
0 0 0 r
[(40)] .93455 .32848 1 0 Mshowa
1 0 0 r
.04545 .44949 Mdot
0 0 0 r
[(41)] .02545 .42949 1 0 Mshowa
1 0 0 r
.14646 .44949 Mdot
0 0 0 r
[(42)] .12646 .42949 1 0 Mshowa
1 0 0 r
.24747 .44949 Mdot
0 0 0 r
[(43)] .22747 .42949 1 0 Mshowa
1 0 0 r
.34848 .44949 Mdot
0 0 0 r
[(44)] .32848 .42949 1 0 Mshowa
1 0 0 r
.44949 .44949 Mdot
0 0 0 r
[(45)] .42949 .42949 1 0 Mshowa
1 0 0 r
.55051 .44949 Mdot
0 0 0 r
[(46)] .53051 .42949 1 0 Mshowa
1 0 0 r
.65152 .44949 Mdot
0 0 0 r
[(47)] .63152 .42949 1 0 Mshowa
0 1 0 r
.75253 .44949 Mdot
0 0 0 r
[(48)] .73253 .42949 1 0 Mshowa
1 0 0 r
.85354 .44949 Mdot
0 0 0 r
[(49)] .83354 .42949 1 0 Mshowa
1 0 0 r
.95455 .44949 Mdot
0 0 0 r
[(50)] .93455 .42949 1 0 Mshowa
1 0 0 r
.04545 .55051 Mdot
0 0 0 r
[(51)] .02545 .53051 1 0 Mshowa
1 0 0 r
.14646 .55051 Mdot
0 0 0 r
[(52)] .12646 .53051 1 0 Mshowa
1 0 0 r
.24747 .55051 Mdot
0 0 0 r
[(53)] .22747 .53051 1 0 Mshowa
1 0 0 r
.34848 .55051 Mdot
0 0 0 r
[(54)] .32848 .53051 1 0 Mshowa
1 0 0 r
.44949 .55051 Mdot
0 0 0 r
[(55)] .42949 .53051 1 0 Mshowa
0 1 0 r
.55051 .55051 Mdot
0 0 0 r
[(56)] .53051 .53051 1 0 Mshowa
1 0 0 r
.65152 .55051 Mdot
0 0 0 r
[(57)] .63152 .53051 1 0 Mshowa
0 1 0 r
.75253 .55051 Mdot
0 0 0 r
[(58)] .73253 .53051 1 0 Mshowa
1 0 0 r
.85354 .55051 Mdot
0 0 0 r
[(59)] .83354 .53051 1 0 Mshowa
1 0 0 r
.95455 .55051 Mdot
0 0 0 r
[(60)] .93455 .53051 1 0 Mshowa
1 0 0 r
.04545 .65152 Mdot
0 0 0 r
[(61)] .02545 .63152 1 0 Mshowa
1 0 0 r
.14646 .65152 Mdot
0 0 0 r
[(62)] .12646 .63152 1 0 Mshowa
1 0 0 r
.24747 .65152 Mdot
0 0 0 r
[(63)] .22747 .63152 1 0 Mshowa
1 0 0 r
.34848 .65152 Mdot
0 0 0 r
[(64)] .32848 .63152 1 0 Mshowa
1 0 0 r
.44949 .65152 Mdot
0 0 0 r
[(65)] .42949 .63152 1 0 Mshowa
1 0 0 r
.55051 .65152 Mdot
0 0 0 r
[(66)] .53051 .63152 1 0 Mshowa
0 1 0 r
.65152 .65152 Mdot
0 0 0 r
[(67)] .63152 .63152 1 0 Mshowa
0 1 0 r
.75253 .65152 Mdot
0 0 0 r
[(68)] .73253 .63152 1 0 Mshowa
1 0 0 r
.85354 .65152 Mdot
0 0 0 r
[(69)] .83354 .63152 1 0 Mshowa
1 0 0 r
.95455 .65152 Mdot
0 0 0 r
[(70)] .93455 .63152 1 0 Mshowa
1 0 0 r
.04545 .75253 Mdot
0 0 0 r
[(71)] .02545 .73253 1 0 Mshowa
1 0 0 r
.14646 .75253 Mdot
0 0 0 r
[(72)] .12646 .73253 1 0 Mshowa
0 1 0 r
.24747 .75253 Mdot
0 0 0 r
[(73)] .22747 .73253 1 0 Mshowa
1 0 0 r
.34848 .75253 Mdot
0 0 0 r
[(74)] .32848 .73253 1 0 Mshowa
1 0 0 r
.44949 .75253 Mdot
0 0 0 r
[(75)] .42949 .73253 1 0 Mshowa
1 0 0 r
.55051 .75253 Mdot
0 0 0 r
[(76)] .53051 .73253 1 0 Mshowa
1 0 0 r
.65152 .75253 Mdot
0 0 0 r
[(77)] .63152 .73253 1 0 Mshowa
1 0 0 r
.75253 .75253 Mdot
0 0 0 r
[(78)] .73253 .73253 1 0 Mshowa
0 1 0 r
.85354 .75253 Mdot
0 0 0 r
[(79)] .83354 .73253 1 0 Mshowa
1 0 0 r
.95455 .75253 Mdot
0 0 0 r
[(80)] .93455 .73253 1 0 Mshowa
1 0 0 r
.04545 .85354 Mdot
0 0 0 r
[(81)] .02545 .83354 1 0 Mshowa
1 0 0 r
.14646 .85354 Mdot
0 0 0 r
[(82)] .12646 .83354 1 0 Mshowa
1 0 0 r
.24747 .85354 Mdot
0 0 0 r
[(83)] .22747 .83354 1 0 Mshowa
0 1 0 r
.34848 .85354 Mdot
0 0 0 r
[(84)] .32848 .83354 1 0 Mshowa
1 0 0 r
.44949 .85354 Mdot
0 0 0 r
[(85)] .42949 .83354 1 0 Mshowa
0 1 0 r
.55051 .85354 Mdot
0 0 0 r
[(86)] .53051 .83354 1 0 Mshowa
0 1 0 r
.65152 .85354 Mdot
0 0 0 r
[(87)] .63152 .83354 1 0 Mshowa
1 0 0 r
.75253 .85354 Mdot
0 0 0 r
[(88)] .73253 .83354 1 0 Mshowa
1 0 0 r
.85354 .85354 Mdot
0 0 0 r
[(89)] .83354 .83354 1 0 Mshowa
1 0 0 r
.95455 .85354 Mdot
0 0 0 r
[(90)] .93455 .83354 1 0 Mshowa
1 0 0 r
.04545 .95455 Mdot
0 0 0 r
[(91)] .02545 .93455 1 0 Mshowa
1 0 0 r
.14646 .95455 Mdot
0 0 0 r
[(92)] .12646 .93455 1 0 Mshowa
1 0 0 r
.24747 .95455 Mdot
0 0 0 r
[(93)] .22747 .93455 1 0 Mshowa
1 0 0 r
.34848 .95455 Mdot
0 0 0 r
[(94)] .32848 .93455 1 0 Mshowa
1 0 0 r
.44949 .95455 Mdot
0 0 0 r
[(95)] .42949 .93455 1 0 Mshowa
0 1 0 r
.55051 .95455 Mdot
0 0 0 r
[(96)] .53051 .93455 1 0 Mshowa
1 0 0 r
.65152 .95455 Mdot
0 0 0 r
[(97)] .63152 .93455 1 0 Mshowa
1 0 0 r
.75253 .95455 Mdot
0 0 0 r
[(98)] .73253 .93455 1 0 Mshowa
0 1 0 r
.85354 .95455 Mdot
0 0 0 r
[(99)] .83354 .93455 1 0 Mshowa
1 0 0 r
.95455 .95455 Mdot
0 0 0 r
[(100)] .93455 .93455 1 0 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[76]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`000_ooool3
o`0001Wooooo1Ol0000Iooooo`?o00006oooool3o`0001Wooooo0ol0000Jooooo`?o00006oooool0
0ol0003oooooooooo`0Iooooo`?o00006_ooool2o`0001Gooooo0ol00004ooooo`;o00005_ooool0
00?ooooo00?o0000ooooooooool06Oooool01?l0003oooooooooool0000Hooooo`05o`000?oooooo
ooooooooool000006oooool00ol0003oooooooooo`0Gooooo`05o`000?ooooooooooooooool00000
6?ooool01Ol0003ooooooooooooooooo000001[ooooo00?o0000ooooooooool06?ooool01Ol0003o
oooooooooooooooo000001_ooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`02oooo
o`04o`000?ooooooooooo`0001Gooooo0003ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool07?ooool00ol0003oooooooooo`0Fooooo`Go00007?ooool00ol0003oooooooooo`0Foooo
o`05o`000?ooooooooooooooool000006oooool00ol0003oooooooooo`0Gooooo`05o`000?oooooo
ooooooooool000007?ooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo00;ooooo00Co
0000oooooooooooo00005Oooool000?ooooo00?o0000ooooooooool06oooool00ol0003ooooooooo
o`0Iooooo`;o00006Oooool01?l0003oooooooooool0000Mooooo`03o`000?oooooooooo01Kooooo
1?l0000Looooo`03o`000?oooooooooo01Sooooo0ol0000Jooooo`Co00005?ooool00ol0003ooooo
ooooo`02ooooo`04o`000?ooooooooooo`0001Gooooo0003ooooo`03o`000?oooooooooo01Sooooo
00Go0000ooooooooooooooooo`00000Looooo`03o`000?oooooooooo01Oooooo00?o0000ooooool0
00006Oooool4o`0001Wooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`0Fooooo`05
o`000?ooooooooooooooool000006?ooool01Ol0003ooooooooooooooooo000001Cooooo00?o0000
ooooooooool00_ooool01?l0003oooooooooool0000Eooooo`000_ooool2o`0001[ooooo00Go0000
ooooooooooooooooo`00000Hooooo`05o`000?ooooooooooooooool000006_ooool2o`0001Wooooo
00?o0000ooooooooool06oooool00ol0003oooooooooo`0Iooooo`05o`000?ooooooooooooooool0
00006?ooool01Ol0003ooooooooooooooooo000001Sooooo00Go0000ooooooooooooooooo`00000C
ooooo`;o00001?ooool01?l0003oooooooooool0000Eooooo`000oooool00ol0003oooooooooo`05
ooooo`?o0?l04Oooool3o`0000Kooooo0ooo000Aooooo`?o00001_ooool3o`3o01?ooooo00?o0000
ooooooooool01?ooool3ool0013ooooo1Ol00005ooooo`?oo`004_ooool2o`0000Kooooo0ol0o`0@
ooooo`Go00001Oooool3ool0017ooooo0ol00006ooooo`?o0?l04Oooool3o`0000Kooooo0ooo000<
ooooo`03o`000?oooooooooo00?ooooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ol0o`0Hoooo
o`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hoooo
o`Go0?l06?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7o`3o01Kooooo1ooo000Fooooo`Oo
0?l05_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Oo
o`005_ooool7ool000_ooooo0009ooooo`Oo0?l05_l00007ool001Ko00001ol0o`0Fo`0000Ooo`00
5_l00007ool001Ko00001ol0o`0Fo`0000Ooo`005_l00007o`3o01Ko00001ooo000Fo`0000Ooo`00
2oooool000Wooooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Ooo`005_oo
ool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo000;ooooo`002_ooool5
o`3o01Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5
ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5ool000cooooo000;ooooo`?o0?l06_ooool3ool0
01[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3o`3o
01[ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000;ooooo0ol00008ooooo`03o`00
0?oooooooooo00Wooooo0ol00002ooooo`Go00001oooool00ol0003oooooooooo`09ooooo`?o0000
0oooool3o`0000Sooooo00?o0000ooooooooool02Oooool3o`0000Cooooo0ol00007ooooo`03o`00
0?oooooooooo00Wooooo0ol00003ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o0000
0oooool3o`0000Oooooo00?o0000ooooooooool02_ooool3o`0000Cooooo00?o0000ooooooooool0
1_ooool00ol0003oooooooooo`0:ooooo`?o00000oooool3o`0000Oooooo00?o0000ooooooooool0
2_ooool3o`0000?ooooo0_l00008ooooo`03o`000?oooooooooo00Wooooo1Ol00003ooooo`;o0000
1oooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool00_ooool01?l0003oooooooooool00007ooooo`03o`00
0?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Go
oooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Ko
oooo00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000000oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02_ooool01?l0003ooooooooo
ool00002ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0:ooooo`04o`000?ooooooooooooooo`Go
00001oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000000oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`000003
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02
ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00?ooooo0_l00008
ooooo`03o`000?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000002ooooo`03o`000?oo
oooooooo00Kooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02_ooool01?l0003oooooooooooooool4o`0000Oooooo
00?o0000ooooooooool02oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`?o00001oooool00ol0003ooooo
ooooo`0;ooooo`03o`000?oooooooooo00;ooooo1?l00006ooooo`03o`000?oooooooooo00cooooo
00Go0000ooooooooooooooooo`000002ooooo`03o`000?oooooooooo00Cooooo00?o0000oooooooo
ool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`05o`00
0?ooooooooooooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool00_ooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00[o
oooo00Co0000oooooooooooooooo1?l00008ooooo`03o`000?oooooooooo00[ooooo00Go0000oooo
ooooooooooooo`00000:ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00Go0000ooooooooooooooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool01Ol0003ooooooooo
oooooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05o`00
0?ooooooooooooooool000000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo
00cooooo0002ooooo`;o00002Oooool00ol0003oooooooooo`09ooooo`;o00000oooool01Ol0003o
oooooooooooooooo000000Oooooo00?o0000ooooooooool02Oooool2o`0000?ooooo00Go0000oooo
ooooooooooooo`000007ooooo`03o`000?oooooooooo00Wooooo0_l00005ooooo`;o00002?ooool0
0ol0003oooooooooo`09ooooo`;o00000oooool00ol0003oooooooooo`09ooooo`03o`000?oooooo
oooo00Wooooo0_l00004ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02_ooool2
o`0000?ooooo00Go0000ooooooooooooooooo`000006ooooo`03o`000?oooooooooo00[ooooo0_l0
0003ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003oooooooooo`0:ooooo`;o0000
0oooool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool01Ol0003o
oooooooooooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000oooool00ol0003oooooooooo`05ooooo`?oo`003?ooool00ol0003oooooooooo`02oooo
o`?o00001_ooool3ool000cooooo00?o0000ooooooooool00_ooool3o`0000Kooooo0ooo000<oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool3ool000cooooo00Co0000oooo
oooooooooooo1Ol00005ooooo`?oo`003?ooool00ol0003oooooooooo`03ooooo`;o00001_ooool3
ool000cooooo00Co0000oooooooooooooooo1Ol00005ooooo`?o0?l03?ooool00ol0003ooooooooo
o`02ooooo`?o00001_ooool3o`3o00cooooo00?o0000ooooooooool00_ooool3o`0000Kooooo0ooo
000;ooooo`?o00001?ooool2o`0000Kooooo0ol0o`0=ooooo`002_ooool5ool001Sooooo1Ooo000H
ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0H
ooooo`Goo`006?ooool5o`3o00cooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o01Kooooo1ooo000Foooo
o`Oo0?l02oooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Oo
o`005_l00007ool001Ko00001ol0o`0Fo`0000Oo0?l05_l00007ool001Ko00001ol0o`0;ooooo`00
2Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`00
5_ooool7o`3o01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o00_ooooo000:ooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?oo
ool5o`3o01Sooooo1Ooo000Hooooo`Go0?l03?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3
ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?o0?l06_ooool3
ool001[ooooo0ol0o`0=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000oooooooo
ool02?ooool5o`000003ooooool0003o000000?o00001oooool00ol0003oooooooooo`08ooooo`Go
00000_ooool3o`0000Sooooo00?o0000ooooooooool02?ooool5o`0000?ooooo0ol00007ooooo`03
o`000?oooooooooo00Sooooo1Ol00002ooooo`?o00002?ooool00ol0003oooooooooo`08ooooo`Go
00000_ooool3o`0000Oooooo00?o0000ooooooooool02Oooool5o`0000?ooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`09ooooo`Go00000_ooool3o`0000Oooooo00?o0000oooooooo
ool02Oooool5o`0000;ooooo0_l00008ooooo`03o`000?oooooooooo00[ooooo0ol00004ooooo`;o
00001oooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`00
0?oooooooooo00Wooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00007oooo
o`03o`000?oooooooooo00Wooooo00Ko0000oooooooooooo0000ooooool00003ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02Oooool01?l0003oooooooooool00004ooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool02Oooool01_l0003oooooooooool0003oooooo`00
00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`09ooooo`06o`000?oooooooooo
o`000?oooooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00[ooooo00Co
0000oooooooooooo00000oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00[ooooo
00Ko0000oooooooooooo0000ooooool00003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02_ooool01?l0003oooooooooool00004ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool02Oooool01Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo0000
1_ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`07ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00[ooooo00Co0000oooooooooooooooo1Ol00007ooooo`03o`000?oooooooooo00[ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[ooooo
00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooooooo
ool02oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool02oooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0=ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool0
0ol0003oooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo
00_ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo
00_ooooo00?o0000ooooooooool00_ooool2o`0000Sooooo00?o0000ooooooooool02oooool01?l0
003oooooooooool00002ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02oooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02oooool0
0ol0003oooooooooo`04o`0000Oooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01?l0003ooooooooooooo
ool3o`0000Oooooo00?o0000ooooooooool03?ooool01?l0003oooooooooooooool4o`0000Kooooo
00?o0000ooooooooool02oooool2o`0000?ooooo00Co0000oooooooooooo00001_ooool00ol0003o
ooooooooo`0<ooooo`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Sooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`08ooooo`05o`000?ooooooooooooooool000001Oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000002ooooo`03o`000?oo
oooo000000Sooooo00?o0000ooooooooool02?ooool01_l0003ooooooooooooooooo0000ooooo`Co
00002?ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`00000:
ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000005ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool0
00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000oooooooo
ooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=
ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`0<ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooo
oooo0000ooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000
ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000002oooool00ol0003ooooo
ooooo`08ooooo`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooo
oooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?ooooooooooo`0000Kooooo00?o
0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01Oooool3ool000_ooooo0ol00003
ooooo`?o00001_ooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3ool000_ooooo0ol00005
ooooo`03o`000?oooooooooo00Cooooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;
ooooo`?o00001?ooool2o`0000Kooooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ol0o`0;
ooooo`?o00000oooool3o`0000Kooooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000;
ooooo`?o00001?ooool2o`0000Kooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hoooo
o`Goo`006?ooool5ool000cooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo
o`005_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo000Fooooo`Oo
o`002oooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ol0o`0Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`002Ooo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7o`3o01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3ool0
01[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool0
01[ooooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000ooooooooool0
2Oooool3o`0000;ooooo1Ol00007ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o0000
2?ooool00ol0003oooooooooo`09ooooo`?o00001?ooool3o`0000Oooooo00?o0000ooooooooool0
2Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o0000
1oooool00ol0003oooooooooo`0:ooooo`?o00001?ooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:ooooo`?o0000
0oooool2o`0000Sooooo00?o0000ooooooooool02oooool3o`0000?ooooo0_l00007ooooo`03o`00
0?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?oo
ool01Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003o
ooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003o
ooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003o
oooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004oooo
o`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01Ol0003ooooooooooooooooo
000000;ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0
003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Go0000ooooooooooooooooo`000009
ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000001?l00007ooooo`03o`000?oooooo
oooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00cooooo00?o0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo1Ol00002ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?oo
ool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`;o00001Oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00[ooooo0_l00004ooooo`;o00002?ooool0
0ol0003oooooooooo`0:ooooo`;o00000_ooool01?l0003oooooooooool00008ooooo`03o`000?oo
oooooooo00[ooooo0_l00006ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02_oo
ool2o`0000;ooooo1?l00007ooooo`03o`000?oooooooooo00_ooooo0_l00005ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02oooool2o`0000?ooooo0ol00007ooooo`03o`000?oo
oooooooo00_ooooo0_l00003ooooo`Co00001_ooool00ol0003oooooooooo`09ooooo`04o`000?oo
ooooooooo`0000?ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`00
0oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool00000
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00Ko0000oooooooooooo
0000ooooool00008ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000ol00008oooo
o`03o`000?oooooooooo00cooooo00?o0000ooooool000002_ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooo000000?ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooooooo
0000ooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo
0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000oooo
ooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000002oooool00ol0003ooooooooo
o`08ooooo`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03o`00
0?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003oooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooo
oooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00_o
oooo0_l00003ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?o
oooo00?o0000ooooooooool01Oooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3ool000_o
oooo0ol00003ooooo`?o00001_ooool3ool000_ooooo0ol00005ooooo`03o`000?oooooooooo00Co
oooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;ooooo`?o00001?ooool2o`0000Ko
oooo0ooo000;ooooo`?o00000_ooool5o`0000Gooooo0ooo000;ooooo`?o00000oooool3o`0000Ko
oooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000=ooooo`03o`000?oooooooooo00;o
oooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000<ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo0009
ooooo`Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000F
o`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`002oooool000Wooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000;ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go
o`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`00
6_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`00
3Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000;ooooo0ol00008ooooo`03o`000?oooooooooo00[ooooo0ol00000
0ooooooo0000o`000003o`0000Oooooo00?o0000ooooooooool02_ooool3o`0000;ooooo0ol00008
ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:
ooooo`?o00000_ooool3o`0000Sooooo00?o0000ooooooooool02_ooool3o`0000;ooooo0ol00007
ooooo`03o`000?oooooooooo00_ooooo0ol00003ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool02oooool3o`0000;ooooo0ol00007ooooo`03o`000?oooooooooo00_ooooo0ol00002
ooooo`;o00002?ooool00ol0003oooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000
ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0;oooo
o`05o`000?ooooooooooooooool000000_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00_ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`05ooooo`03o`00
0?oooooooooo00_ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`04ooooo`03
o`000?oooooooooo00cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo00cooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`04oooo
o`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?oooooo
ooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`Go00000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooo
oooo00Sooooo1Ol00005ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool5
o`000003ooooool0003o000000?o00001oooool00ol0003oooooooooo`08ooooo`Go00001Oooool0
0ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo1Ol000000ooooooo0000ooooo`02
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool5o`0000Cooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`09ooooo`Go000000?oooooo`000?ooool00_ooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo1Ol00005ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool03Oooool01?l0003oooooooooool00002ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`04o`000?ooooooooooo`0000Gooooo00?o0000ooooooooool01_oo
ool00ol0003oooooooooo`08ooooo`04o`000?ooooooooooo`0000Cooooo0_l00008ooooo`03o`00
0?oooooooooo00Sooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00008oooo
o`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00000_ooool4o`0000Oooooo00?o
0000ooooooooool02Oooool01?l0003oooooooooool00005ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02Oooool01?l0003oooooooooool00003ooooo`?o00001oooool00ol0003o
ooooooooo`09ooooo`04o`000?ooooooooooo`0000?ooooo1?l00006ooooo`03o`000?oooooooooo
00gooooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool00ol0
003oooooo`000002ooooo`05o`000?ooooooooooooooool000001oooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000Kooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000?ooooo00?o0000ooooool000002?ooool00ol0003oooooooooo`09
ooooo`03o`000?oooooo000000;ooooo1?l00008ooooo`03o`000?oooooooooo00Wooooo00?o0000
ooooool000000_ooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooool000001_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooool000000_ooool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02_oo
ool00ol0003oooooo`000002ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003ooooo
ooooo`09ooooo`Co00000oooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo00co
oooo0002ooooo`;o00002Oooool00ol0003oooooooooo`0:ooooo`;o00000_ooool01Ol0003ooooo
oooooooooooo000000Oooooo00?o0000ooooooooool02_ooool2o`0000;ooooo00Go0000oooooooo
ooooooooo`000007ooooo`03o`000?oooooooooo00[ooooo0_l00004ooooo`;o00002?ooool00ol0
003oooooooooo`0:ooooo`;o00000_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo
00[ooooo0_l00003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02oooool2o`00
00;ooooo00Go0000ooooooooooooooooo`000006ooooo`03o`000?oooooooooo00_ooooo0_l00002
ooooo`05o`000?ooooooooooooooool000001_ooool00ol0003oooooooooo`0;ooooo`;o00000_oo
ool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool00ol0003ooooo
ooooo`04ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01Oooool3ool000gooooo00Co0000oooooooooooooooo0ol00006ooooo`?o
o`003Oooool01?l0003oooooooooooooool3o`0000Kooooo0ooo000=ooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool01?ooool3ool000gooooo00?o0000ooooooooool01Ol00005oooo
o`?oo`003Oooool00ol0003oooooooooo`02ooooo`;o00001_ooool3ool000gooooo00?o0000oooo
ooooool01Ol00005ooooo`?oo`003Oooool01?l0003oooooooooooooool3o`0000Kooooo0ol0o`0=
ooooo`04o`000?ooooooooooooooo`?o00001_ooool3ool000[ooooo1Ol00003ooooo`;o00001_oo
ool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`003?ooool0
00Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo000;ooooo`002Oooool7ool001Ko
00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko
00001ol0o`0Fo`0000Ooo`005_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo
1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo
1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo
000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo1Ooo
000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000J
ooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool000gooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00co
oooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o00000_ooool5o`0000Oo
oooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wo
oooo0ol00004ooooo`?o00001oooool00ol0003oooooooooo`09ooooo`?o00000oooool3o`0000So
oooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00[o
oooo0ol00004ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02_ooool3o`0000?o
oooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`;o00002?ooool00ol0003o
ooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000ooooooooool03?ooool000?ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool0
00000_ooool01?l0003oooooooooool00007ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`08ooooo`05o`000?ooooooooooooooool000001?ooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`00
0003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooo
oooooooo000000?ooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`09ooooo`07o`00
0?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02Oooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003ooooooooo
ool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool03?ooool01Ol0003ooooooooooooooooo000000Wooooo00?o0000ooooooooool0
3?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0
3?ooool00ol0003oooooo`000004o`0000Oooooo00?o0000ooooooooool03?ooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool00ol0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool00ol0003ooooo
o`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01Ol0003ooooo
oooooooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<oooo
o`000oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00?o0000oooooooo
ool00_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00Co0000oooooooo
oooooooo0_l00008ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000_ooool00ol0
003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00cooooo00?o0000ooooool000000ol00007oooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool00ol00007ooooo`03o`000?oooooooooo
00gooooo00?o0000ooooooooool01?l00006ooooo`03o`000?oooooooooo00Wooooo1?l00003oooo
o`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`08ooooo`Co00000_ooool01Ol0003ooooooooooooooooo
000000Oooooo00?o0000ooooooooool02?ooool4o`0000Kooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`08ooooo`Co00000oooool00ol0003oooooo`000008ooooo`03o`000?oooooo
oooo00Sooooo1?l00002ooooo`Co00002?ooool00ol0003oooooooooo`08ooooo`Co00000_ooool0
0ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Wooooo1?l00006ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool02Oooool4o`0000;ooooo00Go0000ooooooooooooooooo`00
0006ooooo`03o`000?oooooooooo00Wooooo1?l00002ooooo`05o`000?ooooooooooooooool00000
1_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Cooooo00Co0000oooooooooooo
00001_ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`0000Wooooo00?o0000ooooooooool0
2?ooool00ol0003oooooooooo`03ooooo`05o`000?ooooooooooooooool000001oooool00ol0003o
ooooooooo`08ooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`000007oooo
o`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01Oooool2o`0000Sooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Wooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Oooooo00?o0000oooo
ooooool02Oooool00ol0003oooooooooo`03ooooo`05o`000?ooooooooooooooool000001_ooool0
0ol0003oooooooooo`09ooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`00
0006ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00oooool01Ol0003ooooooooo
oooooooo000000Kooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`03ooooo`04o`00
0?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool0
1Oooool3ool000[ooooo1Ol00002ooooo`?o00001_ooool3ool000[ooooo1Ol00002ooooo`?o0000
1_ooool3ool000[ooooo1Ol00004ooooo`03o`000?oooooooooo00Cooooo0ooo000:ooooo`Go0000
00?oooooo`000?l000000ol00005ooooo`?oo`002_ooool5o`0000?ooooo0_l00006ooooo`?o0?l0
2_ooool5o`000003ooooool0003o000000?o00001Oooool3ool000[ooooo1Ol00002ooooo`?o0000
1_ooool3o`3o00[ooooo1Ol00002ooooo`?o00001_ooool3ool000cooooo0_l00004ooooo`;o0000
1_ooool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`003?oo
ool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007
o`3o01Ko00001ooo000Fo`0000Oo0?l05_l00007ool001Ko00001ooo000;ooooo`002Oooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool0
01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o01Ko
oooo1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo
1Ooo000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool000gooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`09ooooo`?o00000_ooool5o`00
00Oooooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo
00Wooooo0ol00004ooooo`?o00001oooool00ol0003oooooooooo`09ooooo`?o00000oooool3o`00
00Sooooo00?o0000ooooooooool02Oooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo
00[ooooo0ol00004ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02_ooool3o`00
00?ooooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`;o00002?ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00?ooooo0_l00007ooooo`03o`000?oooooooooo
00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool01Ol0003o
oooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003oooooooooo`08
ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooo
oooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo
00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o0000oooooooo
ool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04
ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`04o`000?oo
ooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01ooo
ool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool000000oooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000005ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01_l0003ooooooooooooooooo0000
ooooo`Go00001oooool00ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool000001Ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05
o`000?ooooooooooooooool000001?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo
00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000001Oooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00cooooo00Go0000ooooooooooooooooo`000002ooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`08ooooo`Co00001Oooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00Sooooo1?l00004ooooo`;o00002?ooool00ol0003oooooooooo`08ooooo`Co0000
0_ooool01?l0003oooooooooool00008ooooo`03o`000?oooooooooo00Sooooo1?l00006ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool4o`0000;ooooo1?l00007ooooo`03
o`000?oooooooooo00Wooooo1?l00005ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool02Oooool4o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00Wooooo1?l00003ooooo`Co
00001_ooool00ol0003oooooooooo`0<ooooo`05o`000?ooooooooooooooool000000_ooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`03ooooo`05o`000?oooooooooooooo
ool000001oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Oooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ool000002?ooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00?ooooo1?l00008oooo
o`03o`000?oooooooooo00Sooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`08oooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`04oooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00oooool01Ol0003ooooooooooooooooo
000000Kooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`03ooooo`05o`000?oooooo
ooooooooool000001_ooool00ol0003oooooooooo`0=ooooo`04o`000?ooooooooooo`0000;ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`0000Wooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`02ooooo`05o`000?ooooooooooooooool00000
1oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00;ooooo00Go0000oooooooooooo
ooooo`000007ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01?ooool2o`0000So
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Wo
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`02ooooo`05o`000?oooooooooooooo
ool000001_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00;ooooo00Go0000oooo
ooooooooooooo`000006ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool00_ooool0
1Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooo
oooooooo000000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`00
0oooool00ol0003oooooooooo`05ooooo`?oo`003?ooool2o`0000?ooooo0ol00006ooooo`?oo`00
3?ooool2o`0000?ooooo0ol00006ooooo`?oo`003?ooool2o`0000Gooooo00?o0000ooooooooool0
1?ooool3ool000cooooo0_l00002ooooo`Go00001Oooool3ool000cooooo0_l00004ooooo`;o0000
1_ooool3ool000cooooo0_l00002ooooo`Go00001Oooool3o`3o00cooooo0_l00003ooooo`?o0000
1_ooool3o`3o00cooooo0_l00003ooooo`?o00001_ooool3ool000[ooooo1Ol00003ooooo`;o0000
1_ooool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`003?oo
ool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007
ool001Ko00001ol0o`0Fo`0000Oo0?l05_l00007ool001Ko00001ooo000;ooooo`002Oooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o
01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o01Ko
oooo1ooo000Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5ool001Sooooo
1Ooo000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?oo`006_ooool3o`3o01[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool000gooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo0002ooooo`?o00002?ooool00ol0003oooooooooo`0:ooooo`04o`000?oooooooooo
ooooo`Go00001oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00;ooooo0ol00008
ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool00oooool3o`0000Oooooo00?o0000
ooooooooool02_ooool00ol0003oooooooooo`02ooooo`?o00002?ooool00ol0003oooooooooo`0:
ooooo`03o`000?oooooooooo00;ooooo0ol00007ooooo`03o`000?oooooooooo00_ooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00?o0000
ooooooooool00_ooool3o`0000Oooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02
ooooo`;o00002?ooool00ol0003oooooooooo`0:ooooo`?o00001?ooool2o`0000Oooooo00?o0000
ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooooooo00;ooooo00Co0000oooooooooooo00001oooool00ol0003oooooooooo`0:
ooooo`05o`000?ooooooooooooooool000000oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00[ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00[ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000oooooooo
ool01?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`0;ooooo`05o`000?ooooooooooooooool000000oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002
ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Go00001oooool0
0ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`0;ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`0<ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003o
ooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Ko
oooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`;o00002?ooool00ol0003o
ooooooooo`0;ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01_ooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`0;ooooo`03o`000?oooooooooo00Co00001oooool00ol0003oooooooooo`0<oooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<oooo
o`04o`000?ooooooooooooooo`?o00001oooool00ol0003oooooooooo`0<ooooo`04o`000?oooooo
ooooooooo`Co00001_ooool00ol0003oooooooooo`0:ooooo`?o00000oooool01?l0003ooooooooo
ool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool03?ooool00ol0003oooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool03?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool03?ooool01_l0003oooooooooool0003oooooo`0000Sooooo00?o0000oooooooo
ool03?ooool00ol0003oooooo`000003o`0000Sooooo00?o0000ooooooooool03?ooool00ol0003o
ooooo`00000:ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool00oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00gooooo00?o0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00gooooo00?o0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002oooo
o`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool03?ooool000;ooooo0_l00009oooo
o`03o`000?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o
0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003o
ooooo`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003o
oooooooooooooooo000000?ooooo0_l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000_ooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooo
oooooooo000000;ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`09ooooo`07o`00
0?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000oooo
ooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?o
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`05o`000?oooooooooooooo
ool000000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003oooo
o`03o`000?oooooooooo00Gooooo0ooo000:ooooo`Go00000_ooool3o`0000Kooooo0ooo000:oooo
o`Go00000_ooool3o`0000Kooooo0ol0o`0:ooooo`Go00001?ooool00ol0003oooooooooo`04oooo
o`?oo`002_ooool5o`000003ooooool0003o000000?o00001Oooool3ool000[ooooo1Ol00003oooo
o`;o00001_ooool3ool000[ooooo1Ol000000ooooooo0000o`000003o`0000Gooooo0ooo000:oooo
o`Go00000_ooool3o`0000Kooooo0ooo000:ooooo`Go00000_ooool3o`0000Kooooo0ol0o`0;oooo
o`?o00001?ooool2o`0000Kooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Go
0?l06?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go
0?l06?ooool5ool000cooooo0009ooooo`Ooo`005_l00007ool001Ko00001ol0o`0Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ol0o`0Fo`0000Ooo`00
2oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000;ooooo`002Oooool7
ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7
ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool000_ooooo000:ooooo`Goo`006?ooool5ool0
01Sooooo1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0
01Sooooo1Ol0o`0Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3o`3o01[o
oooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3o`3o01[o
oooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`000_ooool3o`0000Sooooo00?o0000ooooooooool02Ooo
ool3o`0000;ooooo1Ol00007ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o00002?oo
ool00ol0003oooooooooo`09ooooo`?o00001?ooool3o`0000Oooooo00?o0000ooooooooool02Ooo
ool3o`0000?ooooo0ol00008ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`?o00001ooo
ool00ol0003oooooooooo`0:ooooo`?o00001?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00[ooooo0ol00003ooooo`?o00001oooool00ol0003oooooooooo`0:ooooo`?o00000ooo
ool2o`0000Sooooo00?o0000ooooooooool02_ooool2o`0000Gooooo0_l00007ooooo`03o`000?oo
oooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool0
1Ol0003ooooooooooooooooo000000;ooooo00Co0000oooooooooooo00001oooool00ol0003ooooo
ooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003ooooo
o`000003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003ooooo
oooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00Wooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool03?ooool01Ol0003ooooooooooooooooo0000
00;ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`000oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000003oooo
o`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo
000000Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`06o`000?oooooo
ooooooooool0003ooooo1Ol00007ooooo`03o`000?oooooooooo00Sooooo00Go0000oooooooooooo
ooooo`000005ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01ol0003o
oooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00Go0000ooooooooooooooooo`000004ooooo`03o`000?oooooooooo00Gooooo00?o
0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000005oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool01?l0003oooooooooool00002
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`09ooooo`?o00001Oooool00ol0003oooooooooo`06oooo
o`03o`000?oooooooooo00Wooooo0ol00004ooooo`;o00002?ooool00ol0003oooooooooo`09oooo
o`?o00000_ooool01?l0003oooooooooool00008ooooo`03o`000?oooooooooo00Wooooo0ol00006
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool02Oooool3o`0000;ooooo1?l00007
ooooo`03o`000?oooooooooo00[ooooo0ol00005ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool02_ooool3o`0000?ooooo0ol00007ooooo`03o`000?oooooooooo00[ooooo0ol00003
ooooo`Co00001_ooool00ol0003oooooooooo`0:ooooo`Co00000_ooool01?l0003oooooooooool0
0006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000
ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003ooooo
ooooo`05ooooo`03o`000?oooooooooo00Sooooo00Go0000ooooooooooooooooo`000005ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo0000
00;ooooo00?o0000ooooool000002?ooool00ol0003oooooooooo`08ooooo`06o`000?oooooooooo
ooooool0003ooooo1?l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000[ooooo00?o0000ooooooooool02Oooool01Ol0003ooooooooooooooooo0000
00Gooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`07o`000?oooooooooo
ooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Ooo
ool01ol0003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03
o`000?oooooooooo00Wooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?oooooooooo
o`0000Kooooo00?o0000ooooooooool03?ooool000;ooooo0_l00009ooooo`03o`000?oooooooooo
00Sooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo0000
00?ooooo0_l00008ooooo`03o`000?oooooooooo00Sooooo00Oo0000ooooooooooooooooo`000?oo
oooo000000_ooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000;ooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`09ooooo`07o`000?ooooooooooooooool0
003oooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0
003ooooooooooooooooo0000ooooool000000oooool00ol0003oooooooooo`04ooooo`03o`000?oo
oooooooo00Wooooo00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000oooooooo
ool01?ooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool000000_ooool01?l0
003oooooooooool00006ooooo`03o`000?oooooooooo00cooooo0003ooooo`03o`000?oooooooooo
00Gooooo0ooo000;ooooo`?o00000oooool3o`0000Kooooo0ooo000;ooooo`?o00000oooool3o`00
00Kooooo0ooo000;ooooo`?o00001Oooool00ol0003oooooooooo`04ooooo`?o0?l02oooool3o`00
00;ooooo1Ol00005ooooo`?oo`002oooool3o`0000Cooooo0_l00006ooooo`?o0?l02oooool3o`00
00;ooooo1Ol00005ooooo`?o0?l02oooool3o`0000?ooooo0ol00006ooooo`?oo`002oooool3o`00
00?ooooo0ol00006ooooo`?oo`002oooool3o`0000Cooooo0_l00006ooooo`?oo`003Oooool000[o
oooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01So
oooo1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Ko0000
1ooo000Fo`0000Ooo`005_l00007o`3o01Ko00001ooo000Fo`0000Oo0?l05_l00007o`3o01Ko0000
1ooo000Fo`0000Ooo`005_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`002oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0F
ooooo`Ooo`005_ooool7o`3o01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000;
ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hoooo
o`Go0?l06?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool000cooooo000;ooooo`?o
o`006_ooool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0o`0Jooooo`?o
0?l06_ooool3ool001[ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000;ooooo0ol0
0008ooooo`03o`000?oooooooooo00Wooooo0_l00003ooooo`Go00001oooool00ol0003ooooooooo
o`09ooooo`;o00001?ooool3o`0000Sooooo00?o0000ooooooooool02Oooool2o`0000Gooooo0ol0
0007ooooo`03o`000?oooooooooo00Wooooo0_l00004ooooo`?o00002?ooool00ol0003ooooooooo
o`09ooooo`;o00001?ooool3o`0000Oooooo00?o0000ooooooooool02_ooool2o`0000Gooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`0:ooooo`;o00001?ooool3o`0000Oooooo00?o
0000ooooooooool02_ooool2o`0000Cooooo0_l00008ooooo`03o`000?oooooooooo00Cooooo0ol0
0004ooooo`;o00001?ooool2o`0000Oooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0;ooooo`05o`000?ooooooooooooooool000000_oo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo0000
0oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo
00000oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooo
oooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00cooooo00Co0000oooo
oooooooo00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00cooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool00_ooool01?l0003oooooooooool00002ooooo`04o`000?ooooooooooo`0000Kooooo
00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool00ol0003ooooooooo
o`0<ooooo`05o`000?ooooooooooooooool000002Oooool00ol0003oooooooooo`0<ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<ooooo`03o`00
0?oooooo000000Co00001oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooo000000?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00;ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooo000000?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00;ooooo
00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo
00cooooo0003ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool4o`0000Co
oooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`09ooooo`Co00000oooool2o`0000So
oooo00?o0000ooooooooool02Oooool4o`000005ooooool0003oooooooooool000002?ooool00ol0
003oooooooooo`09ooooo`Co00001Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo
00Wooooo1?l000000ooooooo0000o`000002o`0000Oooooo00?o0000ooooooooool02_ooool4o`00
00Cooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0:ooooo`Co00000_ooool3o`00
00Oooooo00?o0000ooooooooool02_ooool4o`0000;ooooo1?l00006ooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool00_ooool01?l0003oooooooooool00002ooooo`04o`000?oooooo
ooooo`0000Kooooo00?o0000ooooooooool03?ooool000?ooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`08ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`00
0?oooooooooo00Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooooooooooooooo000000Go
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`05o`000?oooooooooooooo
ool000000_ooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00Sooooo00Ko0000oooo
ooooooooooooo`000?ooool4o`0000Sooooo00?o0000ooooooooool02?ooool01ol0003ooooooooo
oooooooo0000ooooool000002_ooool00ol0003oooooooooo`09ooooo`05o`000?oooooooooooooo
ool000001Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Wooooo00Oo0000oooo
ooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`09ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00Co
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`00
00;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<ooooo`000_ooool2o`00
00Wooooo00?o0000ooooooooool02?ooool01ol0003ooooooooooooooooo0000ooooool000000ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00Oo0000oooooooooooooooo
o`000?oooooo000000?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`08ooooo`05
o`000?ooooooooooooooool000000oooool2o`0000Sooooo00?o0000ooooooooool02?ooool01ol0
003ooooooooooooooooo0000ooooool000002oooool00ol0003oooooooooo`08ooooo`05o`000?oo
ooooooooooooool000000_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Wooooo
00Oo0000ooooooooooooooooo`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`09ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool02Oooool01ol0003ooooooooooooooooo0000ooooool0
00000oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo0_l00004ooooo`04
o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0<
ooooo`000oooool00ol0003oooooooooo`05ooooo`?oo`002oooool3o`0000?ooooo0ol00006oooo
o`?oo`002oooool3o`0000?ooooo0ol00006ooooo`?oo`002oooool3o`0000Gooooo00?o0000oooo
ooooool01?ooool3ool000_ooooo0ol00002ooooo`Go00001Oooool3ool000_ooooo0ol00004oooo
o`;o00001_ooool3o`3o00_ooooo0ol00002ooooo`Go00001Oooool3ool000_ooooo0ol00003oooo
o`?o00001_ooool3ool000_ooooo0ol00003ooooo`?o00001_ooool3o`3o00Kooooo00?o0000oooo
ooooool00oooool2o`0000Cooooo0_l00006ooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Go
o`006?ooool5o`3o01Sooooo1Ooo000<ooooo`002Oooool7ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000Fo`0000Oo0?l05_l00007ool001Ko00001ooo000Fo`0000Oo0?l0
5_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`002ooo
ool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7
o`3o01Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000;ooooo`002_ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool0
01Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[o
oooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool001[o
oooo0ol0o`0Jooooo`?oo`003Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500054, -0.0500054, \
0.00383279, 0.00383279}},
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell[TextData[{
  "Since we saved the list of indexes, we can highlight those vertices using \
",
  StyleBox["Combinatorica's",
    FontSlant->"Italic"],
  " ",
  StyleBox["Highlight",
    FontWeight->"Bold"],
  " command."
}], "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Highlight[ExtractGraph[m], 
          someindexes, \[IndentingNewLine]HighlightedVertexColors \[Rule] \
{Green}\[IndentingNewLine]] // ShowGraph;\)\)], "Input",
  CellLabel->"In[81]:=",
  CellTags->"LocalInteractions:2"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.04545 .04545 m
.14646 .04545 L
s
.04545 .04545 m
.04545 .14646 L
s
.14646 .04545 m
.24747 .04545 L
s
.14646 .04545 m
.14646 .14646 L
s
.24747 .04545 m
.34848 .04545 L
s
.24747 .04545 m
.24747 .14646 L
s
.34848 .04545 m
.44949 .04545 L
s
.34848 .04545 m
.34848 .14646 L
s
.44949 .04545 m
.55051 .04545 L
s
.44949 .04545 m
.44949 .14646 L
s
.55051 .04545 m
.65152 .04545 L
s
.55051 .04545 m
.55051 .14646 L
s
.65152 .04545 m
.75253 .04545 L
s
.65152 .04545 m
.65152 .14646 L
s
.75253 .04545 m
.85354 .04545 L
s
.75253 .04545 m
.75253 .14646 L
s
.85354 .04545 m
.95455 .04545 L
s
.85354 .04545 m
.85354 .14646 L
s
.95455 .04545 m
.95455 .14646 L
s
.04545 .14646 m
.14646 .14646 L
s
.04545 .14646 m
.04545 .24747 L
s
.14646 .14646 m
.24747 .14646 L
s
.14646 .14646 m
.14646 .24747 L
s
.24747 .14646 m
.34848 .14646 L
s
.24747 .14646 m
.24747 .24747 L
s
.34848 .14646 m
.44949 .14646 L
s
.34848 .14646 m
.34848 .24747 L
s
.44949 .14646 m
.55051 .14646 L
s
.44949 .14646 m
.44949 .24747 L
s
.55051 .14646 m
.65152 .14646 L
s
.55051 .14646 m
.55051 .24747 L
s
.65152 .14646 m
.75253 .14646 L
s
.65152 .14646 m
.65152 .24747 L
s
.75253 .14646 m
.85354 .14646 L
s
.75253 .14646 m
.75253 .24747 L
s
.85354 .14646 m
.95455 .14646 L
s
.85354 .14646 m
.85354 .24747 L
s
.95455 .14646 m
.95455 .24747 L
s
.04545 .24747 m
.14646 .24747 L
s
.04545 .24747 m
.04545 .34848 L
s
.14646 .24747 m
.24747 .24747 L
s
.14646 .24747 m
.14646 .34848 L
s
.24747 .24747 m
.34848 .24747 L
s
.24747 .24747 m
.24747 .34848 L
s
.34848 .24747 m
.44949 .24747 L
s
.34848 .24747 m
.34848 .34848 L
s
.44949 .24747 m
.55051 .24747 L
s
.44949 .24747 m
.44949 .34848 L
s
.55051 .24747 m
.65152 .24747 L
s
.55051 .24747 m
.55051 .34848 L
s
.65152 .24747 m
.75253 .24747 L
s
.65152 .24747 m
.65152 .34848 L
s
.75253 .24747 m
.85354 .24747 L
s
.75253 .24747 m
.75253 .34848 L
s
.85354 .24747 m
.95455 .24747 L
s
.85354 .24747 m
.85354 .34848 L
s
.95455 .24747 m
.95455 .34848 L
s
.04545 .34848 m
.14646 .34848 L
s
.04545 .34848 m
.04545 .44949 L
s
.14646 .34848 m
.24747 .34848 L
s
.14646 .34848 m
.14646 .44949 L
s
.24747 .34848 m
.34848 .34848 L
s
.24747 .34848 m
.24747 .44949 L
s
.34848 .34848 m
.44949 .34848 L
s
.34848 .34848 m
.34848 .44949 L
s
.44949 .34848 m
.55051 .34848 L
s
.44949 .34848 m
.44949 .44949 L
s
.55051 .34848 m
.65152 .34848 L
s
.55051 .34848 m
.55051 .44949 L
s
.65152 .34848 m
.75253 .34848 L
s
.65152 .34848 m
.65152 .44949 L
s
.75253 .34848 m
.85354 .34848 L
s
.75253 .34848 m
.75253 .44949 L
s
.85354 .34848 m
.95455 .34848 L
s
.85354 .34848 m
.85354 .44949 L
s
.95455 .34848 m
.95455 .44949 L
s
.04545 .44949 m
.14646 .44949 L
s
.04545 .44949 m
.04545 .55051 L
s
.14646 .44949 m
.24747 .44949 L
s
.14646 .44949 m
.14646 .55051 L
s
.24747 .44949 m
.34848 .44949 L
s
.24747 .44949 m
.24747 .55051 L
s
.34848 .44949 m
.44949 .44949 L
s
.34848 .44949 m
.34848 .55051 L
s
.44949 .44949 m
.55051 .44949 L
s
.44949 .44949 m
.44949 .55051 L
s
.55051 .44949 m
.65152 .44949 L
s
.55051 .44949 m
.55051 .55051 L
s
.65152 .44949 m
.75253 .44949 L
s
.65152 .44949 m
.65152 .55051 L
s
.75253 .44949 m
.85354 .44949 L
s
.75253 .44949 m
.75253 .55051 L
s
.85354 .44949 m
.95455 .44949 L
s
.85354 .44949 m
.85354 .55051 L
s
.95455 .44949 m
.95455 .55051 L
s
.04545 .55051 m
.14646 .55051 L
s
.04545 .55051 m
.04545 .65152 L
s
.14646 .55051 m
.24747 .55051 L
s
.14646 .55051 m
.14646 .65152 L
s
.24747 .55051 m
.34848 .55051 L
s
.24747 .55051 m
.24747 .65152 L
s
.34848 .55051 m
.44949 .55051 L
s
.34848 .55051 m
.34848 .65152 L
s
.44949 .55051 m
.55051 .55051 L
s
.44949 .55051 m
.44949 .65152 L
s
.55051 .55051 m
.65152 .55051 L
s
.55051 .55051 m
.55051 .65152 L
s
.65152 .55051 m
.75253 .55051 L
s
.65152 .55051 m
.65152 .65152 L
s
.75253 .55051 m
.85354 .55051 L
s
.75253 .55051 m
.75253 .65152 L
s
.85354 .55051 m
.95455 .55051 L
s
.85354 .55051 m
.85354 .65152 L
s
.95455 .55051 m
.95455 .65152 L
s
.04545 .65152 m
.14646 .65152 L
s
.04545 .65152 m
.04545 .75253 L
s
.14646 .65152 m
.24747 .65152 L
s
.14646 .65152 m
.14646 .75253 L
s
.24747 .65152 m
.34848 .65152 L
s
.24747 .65152 m
.24747 .75253 L
s
.34848 .65152 m
.44949 .65152 L
s
.34848 .65152 m
.34848 .75253 L
s
.44949 .65152 m
.55051 .65152 L
s
.44949 .65152 m
.44949 .75253 L
s
.55051 .65152 m
.65152 .65152 L
s
.55051 .65152 m
.55051 .75253 L
s
.65152 .65152 m
.75253 .65152 L
s
.65152 .65152 m
.65152 .75253 L
s
.75253 .65152 m
.85354 .65152 L
s
.75253 .65152 m
.75253 .75253 L
s
.85354 .65152 m
.95455 .65152 L
s
.85354 .65152 m
.85354 .75253 L
s
.95455 .65152 m
.95455 .75253 L
s
.04545 .75253 m
.14646 .75253 L
s
.04545 .75253 m
.04545 .85354 L
s
.14646 .75253 m
.24747 .75253 L
s
.14646 .75253 m
.14646 .85354 L
s
.24747 .75253 m
.34848 .75253 L
s
.24747 .75253 m
.24747 .85354 L
s
.34848 .75253 m
.44949 .75253 L
s
.34848 .75253 m
.34848 .85354 L
s
.44949 .75253 m
.55051 .75253 L
s
.44949 .75253 m
.44949 .85354 L
s
.55051 .75253 m
.65152 .75253 L
s
.55051 .75253 m
.55051 .85354 L
s
.65152 .75253 m
.75253 .75253 L
s
.65152 .75253 m
.65152 .85354 L
s
.75253 .75253 m
.85354 .75253 L
s
.75253 .75253 m
.75253 .85354 L
s
.85354 .75253 m
.95455 .75253 L
s
.85354 .75253 m
.85354 .85354 L
s
.95455 .75253 m
.95455 .85354 L
s
.04545 .85354 m
.14646 .85354 L
s
.04545 .85354 m
.04545 .95455 L
s
.14646 .85354 m
.24747 .85354 L
s
.14646 .85354 m
.14646 .95455 L
s
.24747 .85354 m
.34848 .85354 L
s
.24747 .85354 m
.24747 .95455 L
s
.34848 .85354 m
.44949 .85354 L
s
.34848 .85354 m
.34848 .95455 L
s
.44949 .85354 m
.55051 .85354 L
s
.44949 .85354 m
.44949 .95455 L
s
.55051 .85354 m
.65152 .85354 L
s
.55051 .85354 m
.55051 .95455 L
s
.65152 .85354 m
.75253 .85354 L
s
.65152 .85354 m
.65152 .95455 L
s
.75253 .85354 m
.85354 .85354 L
s
.75253 .85354 m
.75253 .95455 L
s
.85354 .85354 m
.95455 .85354 L
s
.85354 .85354 m
.85354 .95455 L
s
.95455 .85354 m
.95455 .95455 L
s
.04545 .95455 m
.14646 .95455 L
s
.14646 .95455 m
.24747 .95455 L
s
.24747 .95455 m
.34848 .95455 L
s
.34848 .95455 m
.44949 .95455 L
s
.44949 .95455 m
.55051 .95455 L
s
.55051 .95455 m
.65152 .95455 L
s
.65152 .95455 m
.75253 .95455 L
s
.75253 .95455 m
.85354 .95455 L
s
.85354 .95455 m
.95455 .95455 L
s
0 1 0 r
.07 w
.04545 .04545 Mdot
1 0 0 r
.025 w
.14646 .04545 Mdot
0 1 0 r
.07 w
.24747 .04545 Mdot
1 0 0 r
.025 w
.34848 .04545 Mdot
.44949 .04545 Mdot
0 1 0 r
.07 w
.55051 .04545 Mdot
1 0 0 r
.025 w
.65152 .04545 Mdot
0 1 0 r
.07 w
.75253 .04545 Mdot
1 0 0 r
.025 w
.85354 .04545 Mdot
.95455 .04545 Mdot
.04545 .14646 Mdot
.14646 .14646 Mdot
.24747 .14646 Mdot
.34848 .14646 Mdot
.44949 .14646 Mdot
.55051 .14646 Mdot
0 1 0 r
.07 w
.65152 .14646 Mdot
.75253 .14646 Mdot
1 0 0 r
.025 w
.85354 .14646 Mdot
0 1 0 r
.07 w
.95455 .14646 Mdot
1 0 0 r
.025 w
.04545 .24747 Mdot
.14646 .24747 Mdot
.24747 .24747 Mdot
.34848 .24747 Mdot
.44949 .24747 Mdot
.55051 .24747 Mdot
0 1 0 r
.07 w
.65152 .24747 Mdot
1 0 0 r
.025 w
.75253 .24747 Mdot
.85354 .24747 Mdot
.95455 .24747 Mdot
.04545 .34848 Mdot
.14646 .34848 Mdot
.24747 .34848 Mdot
.34848 .34848 Mdot
.44949 .34848 Mdot
.55051 .34848 Mdot
.65152 .34848 Mdot
.75253 .34848 Mdot
.85354 .34848 Mdot
.95455 .34848 Mdot
.04545 .44949 Mdot
.14646 .44949 Mdot
.24747 .44949 Mdot
.34848 .44949 Mdot
.44949 .44949 Mdot
.55051 .44949 Mdot
.65152 .44949 Mdot
0 1 0 r
.07 w
.75253 .44949 Mdot
1 0 0 r
.025 w
.85354 .44949 Mdot
.95455 .44949 Mdot
.04545 .55051 Mdot
.14646 .55051 Mdot
.24747 .55051 Mdot
.34848 .55051 Mdot
.44949 .55051 Mdot
0 1 0 r
.07 w
.55051 .55051 Mdot
1 0 0 r
.025 w
.65152 .55051 Mdot
0 1 0 r
.07 w
.75253 .55051 Mdot
1 0 0 r
.025 w
.85354 .55051 Mdot
.95455 .55051 Mdot
.04545 .65152 Mdot
.14646 .65152 Mdot
.24747 .65152 Mdot
.34848 .65152 Mdot
.44949 .65152 Mdot
.55051 .65152 Mdot
0 1 0 r
.07 w
.65152 .65152 Mdot
.75253 .65152 Mdot
1 0 0 r
.025 w
.85354 .65152 Mdot
.95455 .65152 Mdot
.04545 .75253 Mdot
.14646 .75253 Mdot
0 1 0 r
.07 w
.24747 .75253 Mdot
1 0 0 r
.025 w
.34848 .75253 Mdot
.44949 .75253 Mdot
.55051 .75253 Mdot
.65152 .75253 Mdot
.75253 .75253 Mdot
0 1 0 r
.07 w
.85354 .75253 Mdot
1 0 0 r
.025 w
.95455 .75253 Mdot
.04545 .85354 Mdot
.14646 .85354 Mdot
.24747 .85354 Mdot
0 1 0 r
.07 w
.34848 .85354 Mdot
1 0 0 r
.025 w
.44949 .85354 Mdot
0 1 0 r
.07 w
.55051 .85354 Mdot
.65152 .85354 Mdot
1 0 0 r
.025 w
.75253 .85354 Mdot
.85354 .85354 Mdot
.95455 .85354 Mdot
.04545 .95455 Mdot
.14646 .95455 Mdot
.24747 .95455 Mdot
.34848 .95455 Mdot
.44949 .95455 Mdot
0 1 0 r
.07 w
.55051 .95455 Mdot
1 0 0 r
.025 w
.65152 .95455 Mdot
.75253 .95455 Mdot
0 1 0 r
.07 w
.85354 .95455 Mdot
1 0 0 r
.025 w
.95455 .95455 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[81]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`002_ooool6o`3o03Cooooo1_l0
o`1Aooooo`Ko0?l0=?ooool6o`3o04Gooooo0008ooooo`[o0?l0<?ooool:o`3o04gooooo2_l0o`0`
ooooo`[o0?l0@oooool000Kooooo3_l0o`0/ooooo`ko0?l0BOooool>o`3o02cooooo3_l0o`11oooo
o`001Oooool@o`3o02[ooooo4?l0o`17oooooa3o0?l0:_ooool@o`3o043ooooo0005oooooa3o0?l0
:_ooool@o`3o04Oooooo4?l0o`0Zoooooa3o0?l0@?ooool000Cooooo4_l0o`0Xoooooa;o0?l0AOoo
oolBo`3o02Sooooo4_l0o`0oooooo`001?oooolBo`3o02Sooooo4_l0o`15oooooa;o0?l0:?oooolB
o`3o03oooooo0003oooooaCo0?l04Oooool3ool001;ooooo5?l0o`0Aooooo`?oo`006_ooool3ool0
01;ooooo5?l0o`0Aooooo`?oo`004_oooolDo`3o017ooooo0ooo000Jooooo`?oo`003Oooool000?o
oooo5?l0o`0@ooooo`Goo`004OoooolDo`3o013ooooo1Ooo000Hooooo`Goo`004OoooolDo`3o013o
oooo1Ooo000AoooooaCo0?l04?ooool5ool001Sooooo1Ooo000<ooooo`000ooooolDo`3o00oooooo
1ooo000@oooooaCo0?l03oooool7ool001Kooooo1ooo000@oooooaCo0?l03oooool7ool0013ooooo
5?l0o`0?ooooo`Ooo`005_ooool7ool000_ooooo0003oooooaCo0?l03ol00007ool0013o00005?l0
o`0?o`0000Ooo`005_l00007ool0013o00005?l0o`0?o`0000Ooo`004?l0000Do`3o00oo00001ooo
000Fo`0000Ooo`002oooool000?ooooo5?l0o`0?ooooo`Ooo`004?oooolDo`3o00oooooo1ooo000F
ooooo`Ooo`004?oooolDo`3o00oooooo1ooo000@oooooaCo0?l03oooool7ool001Kooooo1ooo000;
ooooo`000ooooolDo`3o013ooooo1Ooo000AoooooaCo0?l04?ooool5ool001Sooooo1Ooo000Aoooo
oaCo0?l04?ooool5ool0017ooooo5?l0o`0@ooooo`Goo`006?ooool5ool000cooooo0004oooooa;o
0?l04_ooool3ool001?ooooo4_l0o`0Booooo`?oo`006_ooool3ool001?ooooo4_l0o`0Booooo`?o
o`004ooooolBo`3o01;ooooo0ooo000Jooooo`?oo`003Oooool000Cooooo4_l0o`0Dooooo`03o`00
0?oooooooooo017ooooo4_l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
4OoooolBo`3o01?ooooo00?o0000ooooooooool04_oooolBo`3o01?ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`001Oooool@o`3o01Gooooo00?o0000ooooooooool04_oo
ool@o`3o01Gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Boooooa3o0?l05?oo
ool00ol0003oooooooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo0005oooooa3o0?l05Oooool00ol0003oooooooooo`0Boooooa3o0?l05Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01;ooooo4?l0o`0Dooooo`03o`000?oooooo
oooo01?ooooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00Kooooo3_l0o`0Fooooo`03o`000?oooooooooo01?ooooo3_l0o`0Fooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool04oooool>o`3o01Gooooo00?o0000ooooooooool05?ooool>o`3o
01Gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`002?ooool:o`3o01So
oooo00?o0000ooooooooool05Oooool:o`3o01Sooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Eooooo`[o0?l05oooool00ol0003oooooooooo`0Fooooo`[o0?l05oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000:ooooo`Ko0?l06_ooool00ol0003ooooo
ooooo`0Gooooo`Ko0?l06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Oooooo
1_l0o`0Iooooo`03o`000?oooooooooo01Sooooo1_l0o`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01Sooooo1_l0o`0Gooooo`Ko0?l06Oooool00ol0003ooooo
ooooo`0Hooooo`Ko0?l02oooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01Kooooo2_l0o`0Cooooo`[o0?l05oooool00ol0003ooooooooo
o`0Fooooo`[o0?l02Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01Cooooo3_l0o`0?ooooo`ko0?l05Oooool00ol0003oooooooooo`0D
ooooo`ko0?l01oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01?ooooo4?l0o`0=oooooa3o0?l05?ooool00ol0003oooooooooo`0Coooo
oa3o0?l01_ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01?ooooo4?l0o`0=oooooa3o0?l05?ooool00ol0003oooooooooo`0Coooooa3o
0?l01_ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01;ooooo4_l0o`0;oooooa;o0?l04oooool00ol0003oooooooooo`0Boooooa;o0?l0
1Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01;ooooo4_l0o`0;oooooa;o0?l04oooool00ol0003oooooooooo`0Boooooa;o0?l01Ooo
ool000_ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3
ool001;ooooo5?l0o`09oooooaCo0?l04Oooool3ool001;ooooo5?l0o`04ooooo`002_ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`004OoooolDo`3o
00Wooooo5?l0o`0@ooooo`Goo`004OoooolDo`3o00Cooooo0009ooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000@oooooaCo0?l02OoooolDo`3o00oo
oooo1ooo000@oooooaCo0?l01?ooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko0000
1ooo000Fo`0000Ooo`005_l00007ool0013o00005?l0o`09o`0001Co0?l03ol00007ool0013o0000
5?l0o`04ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`004?oooolDo`3o00Wooooo5?l0o`0?ooooo`Ooo`004?oooolDo`3o00Cooooo000:
ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000A
oooooaCo0?l02OoooolDo`3o013ooooo1Ooo000AoooooaCo0?l01?ooool000_ooooo0ooo000Joooo
o`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001?ooooo4_l0o`0;oooo
oa;o0?l04_ooool3ool001?ooooo4_l0o`05ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool04_oooolBo`3o00_ooooo4_l0o`0Cooooo`03
o`000?oooooooooo01;ooooo4_l0o`05ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool04oooool@o`3o00gooooo4?l0o`0Dooooo`03o`00
0?oooooooooo01?ooooo4?l0o`06ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool04oooool@o`3o00gooooo4?l0o`0Dooooo`03o`000?oo
oooooooo01?ooooo4?l0o`06ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool05?ooool>o`3o00oooooo3_l0o`0Eooooo`03o`000?oooooo
oooo01Cooooo3_l0o`07ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool05_ooool:o`3o01?ooooo2_l0o`0Gooooo`03o`000?oooooooooo
01Kooooo2_l0o`09ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06?ooool6o`3o01Oooooo1_l0o`0Iooooo`03o`000?oooooooooo01So
oooo1_l0o`0;ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06?ooool6o`3o01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Fooooo`[o0?l05oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01Cooooo3_l0
o`0Eooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool04oooool@o`3o01Cooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01;ooooo4_l0o`0Cooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
4_oooolBo`3o01?ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000;ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_oo
ool3ool001[ooooo0ooo000BoooooaCo0?l04Oooool3ool001[ooooo0ooo000Jooooo`?oo`003Ooo
ool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5
ool0017ooooo5?l0o`0@ooooo`Goo`006?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`004?oooolDo`3o
00oooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo0009ooooo`Ooo`005_l00007ool001Ko
00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000@o`0001Co0?l03ol00007ool001Ko
00001ooo000Fo`0000Ooo`002oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo
1ooo000Fooooo`Ooo`005_ooool7ool0013ooooo5?l0o`0?ooooo`Ooo`005_ooool7ool001Kooooo
1ooo000;ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo
000Hooooo`Goo`004OoooolDo`3o013ooooo1Ooo000Hooooo`Goo`006?ooool5ool000cooooo000;
ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000C
oooooa;o0?l04_ooool3ool001[ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01;ooooo4_l0o`0Coooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool04oooool@o`3o01Cooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01Cooooo3_l0o`0Eooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool05_ooool:
o`3o01Oooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Hooooo`Ko0?l06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000_ooooo0ooo000Jooooo`?oo`00
6_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`00
6_ooool3ool001[ooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool000cooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7
ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`002oooool0
00Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool0
01Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000;ooooo`002Oooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Ko
oooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo
1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Hooooo`Ko0?l06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Kooooo2_l0o`0Gooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool05?ooool>o`3o01Go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01?ooooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool04_oooolBo`3o01?ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Boooo
oa;o0?l04oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000;ooooo`?o
o`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?o
o`004_oooolDo`3o017ooooo0ooo000Jooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Goo`00
6?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000AoooooaCo0?l0
4?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0013ooooo5?l0o`0?ooooo`Ooo`005_oo
ool7ool000_ooooo0009ooooo`Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007
ool001Ko00001ooo000Fo`0000Ooo`004?l0000Do`3o00oo00001ooo000Fo`0000Ooo`002oooool0
00Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000@oooooaCo0?l03oooool7ool001Kooooo1ooo000;ooooo`002_ooool5ool001So
oooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0017o
oooo5?l0o`0@ooooo`Goo`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[ooooo
0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`004ooooolBo`3o01;ooooo
0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool04_oooolBo`3o01?ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0C
oooooa3o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01?ooooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool05?ooool>o`3o01Gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Fooooo`[o0?l0
5oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Sooooo1_l0o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Gooooo`Ko0?l06Ooo
ool00ol0003oooooooooo`0Hooooo`Ko0?l06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool05Oooool:
o`3o01Oooooo00?o0000ooooooooool05_ooool:o`3o01Oooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01?ooooo3_l0o`0Eooooo`03o`000?oooooooooo01Cooooo3_l0o`0Eooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Boooooa3o0?l05?ooool00ol0003oooooooooo`0Coooooa3o0?l05?ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool04_ooool@o`3o01Cooooo00?o0000ooooooooool04oooool@o`3o01Cooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo017ooooo4_l0o`0Cooooo`03o`000?oooooooooo01;ooooo4_l0
o`0Cooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Aoooooa;o0?l04oooool00ol0003oooooooooo`0B
oooooa;o0?l04oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000;oooo
o`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001;ooooo5?l0o`0Aoooo
o`?oo`004_oooolDo`3o017ooooo0ooo000Jooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`004OoooolDo`3o013ooooo1Ooo000AoooooaCo
0?l04?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Ko00001ooo000Fo`0000Ooo`00
5_l00007ool001Ko00001ooo000@o`0001Co0?l03ol00007ool0013o00005?l0o`0?o`0000Ooo`00
5_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool0013ooooo5?l0o`0?ooooo`Ooo`004?oooolDo`3o00oooooo1ooo000Fooooo`Ooo`002ooo
ool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`004?oooolD
o`3o00oooooo1ooo000@oooooaCo0?l03oooool7ool001Kooooo1ooo000;ooooo`002_ooool5ool0
01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000AoooooaCo0?l04?ooool5ool0
017ooooo5?l0o`0@ooooo`Goo`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool001[o
oooo0ooo000Jooooo`?oo`006_ooool3ool001?ooooo4_l0o`0Booooo`?oo`004ooooolBo`3o01;o
oooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Aoooooa;o0?l04oooool00ol0003oooooooooo`0Boooooa;o0?l04oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool04_ooool@o`3o01Cooooo00?o0000ooooooooool04oooool@o`3o01Cooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01;ooooo4?l0o`0Dooooo`03o`000?oooooooooo01?ooooo4?l0o`0D
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Cooooo`ko0?l05Oooool00ol0003oooooooooo`0Doooo
o`ko0?l05Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool05Oooool:o`3o01Oooooo00?o0000oooooooo
ool05_ooool:o`3o01Oooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Oooooo1_l0o`0Iooooo`03o`00
0?oooooooooo01Sooooo1_l0o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01Sooooo1_l0o`0Gooooo`Ko0?l06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Fooooo`[o0?l04oooool:o`3o01Oooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool05?ooool>o`3o00oooooo3_l0o`0Eooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01?ooooo4?l0o`0=oooooa3o0?l05?ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Coooooa3o0?l03Oooool@o`3o01Co
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool04_oooolBo`3o00_ooooo
4_l0o`0Cooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01;ooooo4_l0
o`0;oooooa;o0?l04oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000;
ooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000B
oooooaCo0?l02OoooolDo`3o017ooooo0ooo000Jooooo`?oo`003Oooool000[ooooo1Ooo000Hoooo
o`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0017ooooo5?l0o`09oooo
oaCo0?l04?ooool5ool001Sooooo1Ooo000<ooooo`002Oooool7ool001Ko00001ooo000Fo`0000Oo
o`005_l00007ool001Ko00001ooo000Fo`0000Ooo`004?l0000Do`3o00Wo00005?l0o`0?o`0000Oo
o`005_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`00
5_ooool7ool001Kooooo1ooo000@oooooaCo0?l02OoooolDo`3o00oooooo1ooo000Fooooo`Ooo`00
2oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_oo
ool7ool0013ooooo5?l0o`09oooooaCo0?l03oooool7ool001Kooooo1ooo000;ooooo`002_ooool5
ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`004OoooolD
o`3o00Wooooo5?l0o`0@ooooo`Goo`006?ooool5ool000cooooo000;ooooo`?oo`006_ooool3ool0
01[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Coooooa;o0?l02ooooolBo`3o
01;ooooo0ooo000Jooooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01;ooooo4_l0o`0;oooooa;o0?l04oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Coooooa3o0?l03Oooool@o`3o01Cooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool04oooool@o`3o00gooooo4?l0
o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01Cooooo3_l0o`0?
ooooo`ko0?l05Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Foooo
o`[o0?l04oooool:o`3o01Oooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06?ooool6o`3o01Oooooo1_l0o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Gooooo`Ko
0?l06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Sooooo1_l0o`0Iooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
5Oooool:o`3o01Sooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Fooooo`[o0?l0
5oooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01?ooooo3_l0o`0Fooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool05?oo
ool>o`3o01Gooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Boooooa3o0?l05Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01?ooooo4?l0o`0Dooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool04_ooool@o`3o01Gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo017ooooo4_l0o`0Dooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool04_oooolBo`3o01?ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Aoooooa;o0?l05?ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01;ooooo4_l0o`0Cooooo`03o`000?oooooooooo00cooooo
000;ooooo`?oo`006_ooool3ool001;ooooo5?l0o`0Aooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?oo`006_ooool3ool001;ooooo5?l0o`0Aooooo`?oo`003Oooool000[ooooo1Ooo000H
ooooo`Goo`004OoooolDo`3o013ooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000H
ooooo`Goo`004OoooolDo`3o013ooooo1Ooo000<ooooo`002Oooool7ool001Ko00001ooo000@o`00
01Co0?l03ol00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000@o`00
01Co0?l03ol00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool0013ooooo5?l0o`0?ooooo`Oo
o`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool0013ooooo5?l0o`0?ooooo`Oo
o`002oooool000Wooooo1ooo000Fooooo`Ooo`004?oooolDo`3o00oooooo1ooo000Fooooo`Ooo`00
5_ooool7ool001Kooooo1ooo000Fooooo`Ooo`004?oooolDo`3o00oooooo1ooo000;ooooo`002_oo
ool5ool001Sooooo1Ooo000AoooooaCo0?l04?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5ool001Sooooo1Ooo000AoooooaCo0?l04?ooool5ool000cooooo000;ooooo`?oo`006_ooool3
ool001?ooooo4_l0o`0Booooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3
ool001?ooooo4_l0o`0Booooo`?oo`003Oooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Aoooooa;o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01;ooooo4_l0o`0Cooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool04_ooool@o`3o01Gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Coooooa3o0?l05?ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01;ooooo4?l0o`0Eooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool04oooool@o`3o01Cooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Cooooo`ko0?l05_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Cooooo3_l0o`0Eooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool05Oooool:o`3o01Sooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Fooooo`[o0?l05oooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Oooooo1_l0o`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06?ooool6o`3o01Wooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Oooooo1_l0o`0Jooooo`03o`000?oooooooooo01Oooooo1_l0o`0Goooo
o`Ko0?l06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Gooooo2_l0o`0Hooooo`03o`000?oooooooooo01Gooooo2_l0o`0Cooooo`[o
0?l05oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01?ooooo3_l0o`0Fooooo`03o`000?oooooooooo01?ooooo3_l0o`0?ooooo`ko0?l0
5Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01;ooooo4?l0o`0Eooooo`03o`000?oooooooooo01;ooooo4?l0o`0=oooooa3o0?l05?oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01;ooooo4?l0o`0Eooooo`03o`000?oooooooooo01;ooooo4?l0o`0=oooooa3o0?l05?ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
017ooooo4_l0o`0Dooooo`03o`000?oooooooooo017ooooo4_l0o`0;oooooa;o0?l04oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo017o
oooo4_l0o`0Dooooo`03o`000?oooooooooo017ooooo4_l0o`0;oooooa;o0?l04oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000_ooooo
0ooo000Jooooo`?oo`006_ooool3ool001;ooooo5?l0o`0Aooooo`?oo`004_oooolDo`3o00Wooooo
5?l0o`0Aooooo`?oo`006_ooool3ool001[ooooo0ooo000=ooooo`002_ooool5ool001Sooooo1Ooo
000Hooooo`Goo`004OoooolDo`3o013ooooo1Ooo000AoooooaCo0?l02OoooolDo`3o013ooooo1Ooo
000Hooooo`Goo`006?ooool5ool000cooooo0009ooooo`Ooo`005_l00007ool001Ko00001ooo000@
o`0001Co0?l03ol00007ool0013o00005?l0o`09o`0001Co0?l03ol00007ool001Ko00001ooo000F
o`0000Ooo`002oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool0013ooooo5?l0o`0?oooo
o`Ooo`004?oooolDo`3o00Wooooo5?l0o`0?ooooo`Ooo`005_ooool7ool001Kooooo1ooo000;oooo
o`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`004?oooolDo`3o00oooooo1ooo000@oooooaCo
0?l02OoooolDo`3o00oooooo1ooo000Fooooo`Ooo`005_ooool7ool000_ooooo000:ooooo`Goo`00
6?ooool5ool001Sooooo1Ooo000AoooooaCo0?l04?ooool5ool0017ooooo5?l0o`09oooooaCo0?l0
4?ooool5ool001Sooooo1Ooo000Hooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`006_oo
ool3ool001?ooooo4_l0o`0Booooo`?oo`004ooooolBo`3o00_ooooo4_l0o`0Booooo`?oo`006_oo
ool3ool001[ooooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool04OoooolBo`3o01Cooooo00?o0000ooooooooool04OoooolB
o`3o00_ooooo4_l0o`0Cooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool04_ooool@o`3o01Gooooo00?o0000ooooooooool04_ooool@o`3o
00gooooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool04_ooool@o`3o01Gooooo00?o0000ooooooooool04_ooool@o`3o00go
oooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool04oooool>o`3o01Kooooo00?o0000ooooooooool04oooool>o`3o00oooooo
3_l0o`0Eooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool05Oooool:o`3o01Sooooo00?o0000ooooooooool05Oooool:o`3o01?ooooo2_l0
o`0Gooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool05oooool6o`3o01[ooooo00?o0000ooooooooool05oooool6o`3o01Oooooo1_l0o`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Oooooo1_l0o`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06?ooool6o`3o01Wooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Eoooo
o`[o0?l05oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Kooooo2_l0o`0Goooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool04oooool>o`3o01Gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Dooooo`ko
0?l05Oooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01;ooooo4?l0o`0Dooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
4oooool@o`3o01Cooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Boooooa3o0?l05?ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01?ooooo4?l0o`0Dooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool04OoooolBo`3o01?ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Boooooa;o0?l04oooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo017ooooo4_l0o`0Cooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool04_oooolBo`3o01?ooooo00?o0000ooooooooool03?ooool000_o
oooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?oo`004_oooolDo`3o017o
oooo0ooo000Jooooo`?oo`004_oooolDo`3o017ooooo0ooo000=ooooo`002_ooool5ool001Sooooo
1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000AoooooaCo0?l04?ooool5ool001Sooooo
1Ooo000AoooooaCo0?l04?ooool5ool000cooooo0009ooooo`Ooo`005_l00007ool001Ko00001ooo
000Fo`0000Ooo`005_l00007ool0013o00005?l0o`0?o`0000Ooo`005_l00007ool0013o00005?l0
o`0?o`0000Ooo`002oooool000Wooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000F
ooooo`Ooo`004?oooolDo`3o00oooooo1ooo000Fooooo`Ooo`004?oooolDo`3o00oooooo1ooo000;
ooooo`002Oooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000@oooo
oaCo0?l03oooool7ool001Kooooo1ooo000@oooooaCo0?l03oooool7ool000_ooooo000:ooooo`Go
o`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0017ooooo5?l0o`0@ooooo`Go
o`006?ooool5ool0017ooooo5?l0o`0@ooooo`Goo`003?ooool000_ooooo0ooo000Jooooo`?oo`00
6_ooool3ool001[ooooo0ooo000Jooooo`?oo`004ooooolBo`3o01;ooooo0ooo000Jooooo`?oo`00
4ooooolBo`3o01;ooooo0ooo000=ooooo`00UOoooolBo`3o04Gooooo4_l0o`0Rooooo`00U_ooool@
o`3o04Oooooo4?l0o`0Sooooo`00U_ooool@o`3o04Oooooo4?l0o`0Sooooo`00Uoooool>o`3o04Wo
oooo3_l0o`0Tooooo`00VOooool:o`3o04gooooo2_l0o`0Vooooo`00Voooool6o`3o057ooooo1_l0
o`0Xooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500054, -0.0500054, \
0.00383279, 0.00383279}},
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell[BoxData[
    \(DeleteModel[m]\)], "Input",
  CellLabel->"In[82]:=",
  CellTags->"LocalInteractions:2"],

Cell["Perform some necessary clean up.", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(Scan[ReleaseJavaObject, PeekObjects[]]\)], "Input",
  CellLabel->"In[84]:=",
  CellTags->"LocalInteractions:2"],

Cell["\<\
The reason why the above cleanup work was necessary is the \
following.  Begin with a simple model:\
\>", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(\(m = 
        LocalInteractionModel[\[IndentingNewLine]InteractionNeighborhood \
\[Rule] RandomGraph[20, 
              0.3], \[IndentingNewLine]InitialStrategyDistribution \[Rule] 
            Table[1, {20}]\[IndentingNewLine]];\)\)], "Input",
  CellLabel->"In[44]:=",
  CellTags->"LocalInteractions:2"],

Cell["Get the list of agents.", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(\(agentlist = Agents[m];\)\)], "Input",
  CellLabel->"In[45]:=",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Length[agentlist]\)], "Input",
  CellLabel->"In[46]:=",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(20\)], "Output",
  CellLabel->"Out[46]=",
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell["\<\
Peeking at the Java objects current loaded into the kernel shows \
that, in addition to the model, there are the 20 agents.\
\>", "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(PeekObjects[]\)], "Input",
  CellLabel->"In[49]:=",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject114572312707073], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject113437518594049], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject266203415183361], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject54984607531009], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject152855923130369], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject149088515391489], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject27626722820097], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject146609413292033], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject170153534816257], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject46677050261505], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject21544008941569], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject279487967133697], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject30628888182785], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject32616216526849], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject95134934892545], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Model"] \[RightGuillemet]\),
          JLink`Objects`JavaObject107705750519809], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject43996923559937], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject243497332375553], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject40537998491649], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject99969004470273], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject36224676921345]}], "}"}]], "Output",
  CellLabel->"Out[49]=",
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(PeekObjects[] // Length\)], "Input",
  CellLabel->"In[50]:=",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(21\)], "Output",
  CellLabel->"Out[50]=",
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell[BoxData[
    \(DeleteModel[m]\)], "Input",
  CellLabel->"In[51]:=",
  CellTags->"LocalInteractions:2"],

Cell["All of the agents remain in the kernel.", "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(PeekObjects[]\)], "Input",
  CellLabel->"In[53]:=",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject114572312707073], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject113437518594049], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject266203415183361], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject54984607531009], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject152855923130369], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject149088515391489], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject27626722820097], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject146609413292033], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject170153534816257], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject46677050261505], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject21544008941569], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject279487967133697], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject30628888182785], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject32616216526849], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject95134934892545], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject43996923559937], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject243497332375553], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject40537998491649], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject99969004470273], ",", 
        InterpretationBox[\(\[LeftGuillemet]JavaObject[
              "Agent"] \[RightGuillemet]\),
          JLink`Objects`JavaObject36224676921345]}], "}"}]], "Output",
  CellLabel->"Out[53]=",
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell["Free them as follows:", "Text",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \(Scan[ReleaseJavaObject, PeekObjects[]]\)], "Input",
  CellLabel->"In[54]:=",
  CellTags->"LocalInteractions:2"],

Cell["No Java objects remain.", "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{

Cell[BoxData[
    \(PeekObjects[]\)], "Input",
  CellLabel->"In[55]:=",
  CellTags->"LocalInteractions:2"],

Cell[BoxData[
    \({}\)], "Output",
  CellLabel->"Out[55]=",
  CellTags->"LocalInteractions:2"]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1152}, {0, 701}},
WindowSize->{601, 572},
WindowMargins->{{Automatic, 79}, {51, Automatic}},
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions for On-line Help", "Title"],
  
  Cell["\<\
These styles are carefully tuned for all aspects of the online \
help. It's recommended that they remain unchanged.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ShowCellBracket->False,
    CellBracketOptions->{"Color"->RGBColor[0.269993, 0.308507, 0.6]},
    ShowAutoStyles->False,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Brackets"],
    PageWidth->WindowWidth,
    ShowCellBracket->True,
    CellBracketOptions->{"Color"->RGBColor[0.269993, 0.308507, 0.6]},
    ShowAutoStyles->False,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ShowCellBracket->False,
    CellBracketOptions->{"Color"->RGBColor[0.269993, 0.308507, 0.6]},
    ShowAutoStyles->False,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "SlideShow"],
    PageWidth->WindowWidth,
    ScrollingOptions->{"PagewiseDisplay"->True,
    "VerticalScrollRange"->Fit},
    ShowCellBracket->False,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ShowCellBracket->False,
    ShowAutoStyles->False,
    ScriptMinSize->5,
    PrivateFontOptions->{"FontType"->"Outline"}],
  
  Cell[StyleData[All, "EnhancedPrintout"],
    PageWidth->PaperWidth,
    ShowCellBracket->True,
    ScriptMinSize->5,
    FontFamily->"Palatino",
    PrivateFontOptions->{"FontType"->"Outline"}],
  
  Cell[StyleData[All, "EnhancedPrintoutNonGray"],
    PageWidth->PaperWidth,
    ShowCellBracket->False,
    ScriptMinSize->5,
    FontFamily->"Palatino",
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
  The options defined for the style below will be used at the \
Notebook level.  \
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {"Printed from the ", 
              StyleBox[ "Mathematica", FontSlant -> "Italic"], 
              " Help Browser"}], "PageHeader"]}, {Cell[ 
            TextData[ {"Printed from the ", 
              StyleBox[ "Mathematica", FontSlant -> "Italic"], 
              " Help Browser"}], "PageHeader"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{None, Cell[ 
            
            TextData[ {
              "\[Copyright]1988-2003 Wolfram Research, Inc. All rights \
reserved."}], "PageFooter"], None}, {None, Cell[ 
            
            TextData[ {
              "\[Copyright]1988-2003 Wolfram Research, Inc. All rights \
reserved."}], "PageFooter"], None}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FacingPages"->True},
    CellLabelAutoDelete->False,
    CellFrameLabelMargins->6,
    StyleMenuListing->None],
  
  Cell[StyleData["Notebook", "Printout"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {"Printed from the ", 
              StyleBox[ "Mathematica", FontSlant -> "Italic"], 
              " Help Browser"}], "PageHeader"]}, {Cell[ 
            TextData[ {"Printed from the ", 
              StyleBox[ "Mathematica", FontSlant -> "Italic"], 
              " Help Browser"}], "PageHeader"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{None, Cell[ 
            TextData[ {
              "\[Copyright]1988-2003 Wolfram Research, Inc. All rights \
reserved."}], "PageFooter"], None}, {None, Cell[ 
            
            TextData[ {
              "\[Copyright]1988-2003 Wolfram Research, Inc. All rights \
reserved."}], "PageFooter"], None}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FacingPages"->True}],
  
  Cell[StyleData["Notebook", "EnhancedPrintout"],
    PageFooters->{{None, None, None}, {None, None, None}},
    PageHeaderLines->{True, True},
    PageFooterLines->{False, False},
    PrintingOptions->{"PrintingMargins"->{{54, 54}, {36, 72}},
    "PageSize"->{594, 657},
    "PrintRegistrationMarks"->True,
    "FirstPageHeader"->False,
    "FacingPages"->True,
    "FirstPageFace"->Right}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Chapter"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{27, 27}, {10, 40}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], BoxMargins -> 
                {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Chapter",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->24,
    FontWeight->"Plain",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["Chapter", "Presentation"],
    FontSize->36],
  
  Cell[StyleData["Chapter", "Printout"],
    CellMargins->{{2, 10}, {6, 30}},
    FontSize->24],
  
  Cell[StyleData["Chapter", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {6, 30}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->24,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Chapter", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {6, 30}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->24]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellMargins->{{27, 27}, {9, 30}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->20,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["Section", "Presentation"],
    FontSize->30],
  
  Cell[StyleData["Section", "SlideShow"],
    CellFrame->{{0, 0}, {0, 0.5}},
    ShowCellBracket->False,
    CellMargins->{{0, 0}, {5, 0}},
    PageBreakBelow->False,
    CellFrameMargins->{{36, 36}, {Inherited, 15}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.408011, 0.440726, 0.8]],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{2, 0}, {0, 18}},
    PageBreakBelow->False,
    FontSize->18],
  
  Cell[StyleData["Section", "EnhancedPrintout"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{2, 2}, {5, 30}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->8,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->16,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Section", "EnhancedPrintoutNonGray"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{2, Inherited}, {5, 30}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->8,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->16,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellDingbat->None,
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CellFrameLabelMargins->6,
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["Subsection", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["Subsection", "SlideShow"],
    CellDingbat->None,
    CellMargins->{{70, 0}, {4, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{2, 0}, {5, 24}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {5, 20}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->12,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Subsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{9, 0}, {5, 20}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->12,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontSize->10],
  
  Cell[StyleData["Subsubsection", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Subsubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsubsectionColor"],
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 50},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->RGBColor[0.297993, 0, 0.454994]],
  
  Cell[StyleData["SubsubsectionColor", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["SubsubsectionColor", "Printout"],
    CellMargins->{{2, 0}, {0, 18}},
    PageBreakBelow->False,
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsectionIcon"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{46, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->8},
    CellGroupingRules->{"SectionGrouping", 40},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    CounterIncrements->"Subsection",
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["SubsectionIcon", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["SubsectionIcon", "Printout"],
    CellDingbat->"\[FilledSquare]",
    ShowGroupOpenCloseIcon->False,
    CellMargins->{{16, 0}, {0, 18}},
    PageBreakBelow->False,
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsubsectionIcon"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{46, 27}, {0, 8}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellGroupingRules->{"SectionGrouping", 50},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    CounterIncrements->"Subsubsection",
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["SubsubsectionIcon", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["SubsubsectionIcon", "Printout"],
    ShowGroupOpenCloseIcon->False,
    CellMargins->{{2, 0}, {0, 18}},
    PageBreakBelow->False,
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{27, 27}, {0, 8}},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Text",
    FontSize->12],
  
  Cell[StyleData["Text", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Text", "SlideShow"],
    CellMargins->{{36, 24}, {4, 4}},
    Hyphenation->False,
    FontSize->14],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["Text", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["Text", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CompactText"],
    CellMargins->{{27, 27}, {0, 8}},
    ParagraphSpacing->{0, 0},
    StyleMenuListing->None,
    FontSize->10],
  
  Cell[StyleData["CompactText", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["CompactText", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontSize->8],
  
  Cell[StyleData["CompactText", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->8],
  
  Cell[StyleData["CompactText", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Compact12Text"],
    CellMargins->{{27, 27}, {0, 8}},
    ParagraphSpacing->{0, 0},
    StyleMenuListing->None],
  
  Cell[StyleData["Compact12Text", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Compact12Text", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["Compact12Text", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["Compact12Text", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TextSans"],
    CellMargins->{{27, 27}, {0, 8}},
    CellHorizontalScrolling->True,
    ParagraphSpacing->{0, 8},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10],
  
  Cell[StyleData["TextSans", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["TextSans", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontSize->8],
  
  Cell[StyleData["TextSans", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Frutiger 55 Roman",
    FontSize->8],
  
  Cell[StyleData["TextSans", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontFamily->"Frutiger 55 Roman",
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedText"],
    CellMargins->{{66, 4}, {4, 4}},
    LineSpacing->{1, 3},
    ParagraphIndent->-22,
    CounterIncrements->"ItemizedText",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedText", "Printout"],
    CellMargins->{{16, 4}, {2, 4}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ItemizedText", "EnhancedPrintout"],
    CellMargins->{{16, 4}, {2, 4}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedText", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    LineSpacing->{1, 1},
    ParagraphIndent->-18,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTextNote"],
    CellMargins->{{88, 4}, {4, 0}},
    LineSpacing->{1, 2},
    CounterIncrements->"ItemizedTextNote",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedTextNote", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedTextNote", "Printout"],
    CellMargins->{{34, 4}, {4, 1}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ItemizedTextNote", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {4, 1}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedTextNote", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 4}, {4, 1}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IndentedText"],
    CellMargins->{{66, 4}, {0, 6}},
    LineSpacing->{1, 3},
    CounterIncrements->"IndentedText",
    StyleMenuListing->None],
  
  Cell[StyleData["IndentedText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["IndentedText", "SlideShow"],
    CellMargins->{{80, 24}, {5, 5}},
    Hyphenation->False,
    FontSize->13],
  
  Cell[StyleData["IndentedText", "Printout"],
    CellMargins->{{16, 4}, {2, 4}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["IndentedText", "EnhancedPrintout"],
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["IndentedText", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedList"],
    CellMargins->{{58, 4}, {4, 4}},
    CellFrameLabels->{{Cell[ 
            TextData[ {
              StyleBox[ 
                CounterBox[ "ItemizedList"], "TB"], 
              StyleBox[ ".", "TB"]}], CellBaseline -> Baseline, CellSize -> {
            24, Inherited}, TextAlignment -> Right], Inherited}, {
        Inherited, Inherited}},
    LineSpacing->{1, 3},
    CounterIncrements->"ItemizedList",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedList", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedList", "Printout"],
    CellMargins->{{9, 4}, {2, 4}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ItemizedList", "EnhancedPrintout"],
    CellMargins->{{9, 4}, {2, 4}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedList", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedSubList"],
    CellMargins->{{92, 4}, {4, 4}},
    CellFrameLabels->{{Cell[ 
            TextData[ {
              StyleBox[ 
                CounterBox[ "ItemizedSubList"], "TB"], 
              StyleBox[ ".", "TB"]}], CellBaseline -> Baseline, CellSize -> {
            24, Inherited}, TextAlignment -> Right], Inherited}, {
        Inherited, Inherited}},
    CounterIncrements->"ItemizedSubList",
    StyleMenuListing->None,
    CounterBoxOptions->{CounterFunction:>(Part[ 
        CharacterRange[ "a", "z"], #]&)}],
  
  Cell[StyleData["ItemizedSubList", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedSubList", "Printout"],
    CellMargins->{{34, 4}, {4, 1}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ItemizedSubList", "EnhancedPrintout"],
    CellMargins->{{26, 4}, {2, 4}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedSubList", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedSubListNote"],
    CellMargins->{{110, 4}, {4, 4}},
    CounterIncrements->"ItemizedSubList",
    StyleMenuListing->None,
    CounterBoxOptions->{CounterFunction:>(Part[ 
        CharacterRange[ "a", "z"], #]&)}],
  
  Cell[StyleData["ItemizedSubListNote", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedSubListNote", "Printout"],
    CellMargins->{{34, 4}, {4, 1}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ItemizedSubListNote", "EnhancedPrintout"],
    CellMargins->{{56, 4}, {2, 4}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedSubListNote", "EnhancedPrintoutNonGray"],
    CellMargins->{{48, 4}, {2, 4}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BulletedList"],
    CellMargins->{{76, 4}, {4, 4}},
    CellFrameLabels->{{
      Cell[ "\[Bullet]", "BulletedList", CellBaseline -> Baseline], 
          Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 3},
    CounterIncrements->"BulletText",
    StyleMenuListing->None],
  
  Cell[StyleData["BulletedList", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["BulletedList", "Printout"],
    CellMargins->{{16, 4}, {4, 1}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["BulletedList", "EnhancedPrintout"],
    CellMargins->{{26, 4}, {2, 4}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["BulletedList", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ListNote"],
    CellMargins->{{88, 4}, {4, 0}},
    LineSpacing->{1, 2},
    CounterIncrements->"ListNote",
    StyleMenuListing->None],
  
  Cell[StyleData["ListNote", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ListNote", "Printout"],
    CellMargins->{{34, 4}, {4, 1}},
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["ListNote", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {4, 1}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ListNote", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 4}, {4, 1}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MathCaption"],
    CellFrame->{{0, 0}, {0, 0.5}},
    CellMargins->{{66, 12}, {2, 20}},
    PageBreakBelow->False,
    CellFrameMargins->{{8, 8}, {8, 2}},
    CellFrameColor->GrayLevel[0.700008],
    CellFrameLabelMargins->4,
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 8},
    FontColor->GrayLevel[0.2]],
  
  Cell[StyleData["MathCaption", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["MathCaption", "SlideShow"],
    ShowCellBracket->False,
    CellMargins->{{36, 36}, {6, 6}}],
  
  Cell[StyleData["MathCaption", "Printout"],
    CellMargins->{{9, 0}, {0, 14}},
    Hyphenation->True,
    FontSize->9,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["MathCaption", "EnhancedPrintout"],
    CellMargins->{{9, 0}, {0, 14}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->9,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["MathCaption", "EnhancedPrintoutNonGray"],
    CellMargins->{{9, 0}, {0, 14}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{27, 27}, {0, 8}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0.7, 0},
    CounterIncrements->"SmallText",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{2, 2}, {5, 5}},
    TextJustification->1,
    Hyphenation->True,
    FontSize->7],
  
  Cell[StyleData["SmallText", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {5, 5}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->7],
  
  Cell[StyleData["SmallText", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {5, 5}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->7]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.    \
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {3, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    ShowSpecialCharacters->Automatic,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->"Mathematica",
    FormatType->StandardForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    FontWeight->"Bold"],
  
  Cell[StyleData["Input", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Input", "SlideShow"],
    ShowCellBracket->True,
    CellMargins->{{100, 50}, {0, 6}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellFrameMargins->{{8, 8}, {4, 8}},
    Background->RGBColor[1, 1, 0.8]],
  
  Cell[StyleData["Input", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {4, 6}},
    FontSize->9],
  
  Cell[StyleData["Input", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {4, 6}},
    FontSize->9],
  
  Cell[StyleData["Input", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {4, 6}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {8, 2}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->None,
    FormatType->StandardForm],
  
  Cell[StyleData["Output", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Output", "SlideShow"],
    ShowCellBracket->True,
    CellMargins->{{100, 50}, {6, 0}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellFrameMargins->{{8, 8}, {8, 4}},
    Background->RGBColor[1, 1, 0.8]],
  
  Cell[StyleData["Output", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 4}},
    FontSize->9],
  
  Cell[StyleData["Output", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 4}},
    FontSize->9],
  
  Cell[StyleData["Output", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 4}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InputOnly"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    ShowSpecialCharacters->Automatic,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->"Mathematica",
    FormatType->StandardForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    StyleMenuListing->None,
    FontWeight->"Bold"],
  
  Cell[StyleData["InputOnly", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["InputOnly", "SlideShow"],
    ShowCellBracket->True,
    CellMargins->{{100, 50}, {6, 6}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    Background->RGBColor[1, 1, 0.8]],
  
  Cell[StyleData["InputOnly", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 6}},
    FontSize->9],
  
  Cell[StyleData["InputOnly", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 6}},
    FontSize->9],
  
  Cell[StyleData["InputOnly", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{39, 0}, {6, 6}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {4, 2}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    AutoStyleOptions->{"UnmatchedBracketStyle"->None},
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->None,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontColor->RGBColor[0.6, 0.100008, 0.100008]],
  
  Cell[StyleData["Message", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Message", "SlideShow"],
    CellMargins->{{100, 50}, {6, 6}},
    FontSize->10,
    FontColor->RGBColor[0.699992, 0.4, 0]],
  
  Cell[StyleData["Message", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    FontSize->9,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Message", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    FontFamily->"Courier",
    FontSize->9,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Message", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    FontFamily->"Courier",
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {4, 2}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->None,
    FormatType->InputForm,
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Print", "SlideShow"],
    CellMargins->{{100, 50}, {6, 6}}],
  
  Cell[StyleData["Print", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{57, Inherited}, {Inherited, Inherited}},
    FontSize->9],
  
  Cell[StyleData["Print", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{57, Inherited}, {Inherited, Inherited}},
    FontFamily->"Courier Std",
    FontSize->9],
  
  Cell[StyleData["Print", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{57, Inherited}, {Inherited, Inherited}},
    FontFamily->"Courier Std",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    ShowCellBracket->True,
    CellMargins->{{66, 4}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultFormatTypeForOutput,
    LanguageCategory->None,
    FormatType->InputForm,
    StyleMenuListing->None,
    FontWeight->"Plain"],
  
  Cell[StyleData["Graphics", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Graphics", "SlideShow"],
    ShowCellBracket->True,
    CellMargins->{{100, 50}, {6, 0}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellFrameMargins->{{8, 8}, {4, 4}},
    Background->RGBColor[1, 1, 0.8]],
  
  Cell[StyleData["Graphics", "Printout"],
    ShowCellBracket->False,
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250},
    FontSize->9],
  
  Cell[StyleData["Graphics", "EnhancedPrintout"],
    ShowCellBracket->False,
    CellMargins->{{32, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250},
    FontFamily->"Courier Std",
    FontSize->9],
  
  Cell[StyleData["Graphics", "EnhancedPrintoutNonGray"],
    ShowCellBracket->False,
    CellMargins->{{32, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250},
    FontFamily->"Courier Std",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    LanguageCategory->None,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Plain",
    FontColor->RGBColor[0.269993, 0.308507, 0.6]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["CellLabel", "SlideShow"],
    FontColor->RGBColor[0.699992, 0.4, 0]],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontFamily->"Courier",
    FontSize->8,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["CellLabel", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSize->8,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["CellLabel", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSize->8,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Name"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Plain",
    FontSlant->"Italic",
    FontColor->RGBColor[0.269993, 0.308507, 0.6]],
  
  Cell[StyleData["Name", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["Name", "Printout"],
    FontSize->7,
    FontSlant->"Oblique",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Name", "EnhancedPrintout"],
    FontFamily->"Frutiger 56 Italic",
    FontSize->7,
    FontSlant->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Name", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 56 Italic",
    FontSize->7,
    FontSlant->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlinks", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks. The others point to the various sections of the help \
browser. These styles will modify individual words or letters in a cell \
exclusive of the cell tag.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Hyperlink", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Hyperlink", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MainBookLink", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MainBookLink", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLinkMR"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLinkMR", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MainBookLinkMR", "EnhancedPrintout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MainBookLinkMR", "EnhancedPrintoutNonGray"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["AddOnsLink", "EnhancedPrintout"],
    FontFamily->"WRICourier",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["AddOnsLink", "EnhancedPrintoutNonGray"],
    FontFamily->"WRICourier",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLinkText"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Times",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLinkText", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["AddOnsLinkText", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["AddOnsLinkText", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuide", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["RefGuideLink", "EnhancedPrintout"],
    FontFamily->"WRICourier",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["RefGuideLink", "EnhancedPrintoutNonGray"],
    FontFamily->"WRICourier",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["GettingStartedLink", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["GettingStartedLink", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DemosLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "Demos", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["DemosLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["DemosLink", "EnhancedPrintout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["DemosLink", "EnhancedPrintoutNonGray"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TourLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "Tour", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["TourLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["TourLink", "EnhancedPrintout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["TourLink", "EnhancedPrintoutNonGray"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["OtherInformationLink", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["OtherInformationLink", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLinkMR"],
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLinkMR", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["OtherInformationLinkMR", "EnhancedPrintout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["OtherInformationLinkMR", "EnhancedPrintoutNonGray"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MasterIndexLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MasterIndex", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MasterIndexLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MasterIndexLink", "EnhancedPrintout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["MasterIndexLink", "EnhancedPrintoutNonGray"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NotebookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Times",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["NotebookLink", "Printout"],
    FontSize->11,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["NotebookLink", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["NotebookLink", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontSize->10,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NotebookLinkMR"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["NotebookLinkMR", "Printout"],
    FontSize->11,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["NotebookLinkMR", "EnhancedPrintout"],
    FontSize->10,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["NotebookLinkMR", "EnhancedPrintoutNonGray"],
    FontSize->10,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Palette Styles", "Section"],
  
  Cell["\<\
The cells below define styles that define standard \
ButtonFunctions, for use in palette buttons.\
\>", "Text"],
  
  Cell[StyleData["Paste"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, After]}]&)}],
  
  Cell[StyleData["Evaluate"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionEvaluate[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["EvaluateCell"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionMove[ 
          FrontEnd`InputNotebook[ ], All, Cell, 1], 
        FrontEnd`SelectionEvaluateCreateCell[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["CopyEvaluate"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`SelectionCreateCell[ 
          FrontEnd`InputNotebook[ ], All], 
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionEvaluate[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["CopyEvaluateCell"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`SelectionCreateCell[ 
          FrontEnd`InputNotebook[ ], All], 
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionEvaluateCreateCell[ 
          FrontEnd`InputNotebook[ ], All]}]&)}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Inline Formatting", "Section"],
  
  Cell["\<\
These styles are for modifying individual words or letters in a \
cell exclusive of the cell tag.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["JT"],
    StyleMenuListing->None,
    FontFamily->"JFontText"],
  
  Cell[StyleData["JT", "Printout"]],
  
  Cell[StyleData["JT", "EnhancedPrintout"]],
  
  Cell[StyleData["JT", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["JS"],
    StyleMenuListing->None,
    FontFamily->"JFontSans"],
  
  Cell[StyleData["JS", "Printout"]],
  
  Cell[StyleData["JS", "EnhancedPrintout"]],
  
  Cell[StyleData["JS", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["JSB"],
    StyleMenuListing->None,
    FontFamily->"JFontSansBold"],
  
  Cell[StyleData["JSB", "Printout"]],
  
  Cell[StyleData["JSB", "EnhancedPrintout"]],
  
  Cell[StyleData["JSB", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RM"],
    StyleMenuListing->None,
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["RM", "Printout"]],
  
  Cell[StyleData["RM", "EnhancedPrintout"]],
  
  Cell[StyleData["RM", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BF"],
    StyleMenuListing->None,
    FontWeight->"Bold"],
  
  Cell[StyleData["BF", "Printout"]],
  
  Cell[StyleData["BF", "EnhancedPrintout"]],
  
  Cell[StyleData["BF", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IT"],
    StyleMenuListing->None,
    FontSlant->"Italic"],
  
  Cell[StyleData["IT", "Printout"]],
  
  Cell[StyleData["IT", "EnhancedPrintout"]],
  
  Cell[StyleData["IT", "EnhancedPrintoutNonGray"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TR"],
    StyleMenuListing->None,
    FontFamily->"Times",
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["TR", "Printout"]],
  
  Cell[StyleData["TR", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["TR", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TI"],
    StyleMenuListing->None,
    FontFamily->"Times",
    FontSlant->"Italic"],
  
  Cell[StyleData["TI", "Printout"]],
  
  Cell[StyleData["TI", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["TI", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TB"],
    StyleMenuListing->None,
    FontFamily->"Times",
    FontWeight->"Bold",
    FontSlant->"Plain"],
  
  Cell[StyleData["TB", "Printout"]],
  
  Cell[StyleData["TB", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["TB", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TBI"],
    StyleMenuListing->None,
    FontFamily->"Times",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  
  Cell[StyleData["TBI", "Printout"]],
  
  Cell[StyleData["TBI", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["TBI", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MR"],
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["MR", "Printout"]],
  
  Cell[StyleData["MR", "EnhancedPrintout"],
    FontFamily->"Courier Std"],
  
  Cell[StyleData["MR", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MO"],
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  
  Cell[StyleData["MO", "Printout"],
    FontSlant->"Oblique"],
  
  Cell[StyleData["MO", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSlant->"Oblique"],
  
  Cell[StyleData["MO", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSlant->"Oblique"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MB"],
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain"],
  
  Cell[StyleData["MB", "Printout"]],
  
  Cell[StyleData["MB", "EnhancedPrintout"],
    FontFamily->"Courier Std"],
  
  Cell[StyleData["MB", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MBO"],
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  
  Cell[StyleData["MBO", "Printout"],
    FontSlant->"Oblique"],
  
  Cell[StyleData["MBO", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSlant->"Oblique"],
  
  Cell[StyleData["MBO", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSlant->"Oblique"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SR"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["SR", "Printout"]],
  
  Cell[StyleData["SR", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman"],
  
  Cell[StyleData["SR", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SO"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  
  Cell[StyleData["SO", "Printout"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  
  Cell[StyleData["SO", "EnhancedPrintout"],
    FontFamily->"Frutiger 56 Italic",
    FontSlant->"Plain"],
  
  Cell[StyleData["SO", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 56 Italic",
    FontSlant->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SB"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain"],
  
  Cell[StyleData["SB", "Printout"],
    StyleMenuListing->None],
  
  Cell[StyleData["SB", "EnhancedPrintout"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"],
  
  Cell[StyleData["SB", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SBColor"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain"],
  
  Cell[StyleData["SBColor", "Printout"],
    StyleMenuListing->None],
  
  Cell[StyleData["SBColor", "EnhancedPrintout"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"],
  
  Cell[StyleData["SBColor", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SBO"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  
  Cell[StyleData["SBO", "Printout"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  
  Cell[StyleData["SBO", "EnhancedPrintout"],
    FontFamily->"Frutiger 66 BoldItalic",
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["SBO", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 66 BoldItalic",
    FontWeight->"Plain",
    FontSlant->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["M4"],
    ShowCellBracket->True,
    StyleMenuListing->None,
    FontFamily->"Mathematica4",
    CharacterEncoding->"Mathematica4"],
  
  Cell[StyleData["M4", "Printout"],
    ShowCellBracket->True,
    StyleMenuListing->None,
    FontFamily->"Mathematica4",
    CharacterEncoding->"Mathematica4"],
  
  Cell[StyleData["M4", "EnhancedPrintout"],
    ShowCellBracket->True,
    StyleMenuListing->None,
    FontFamily->"Mathematica4",
    CharacterEncoding->"Mathematica4"],
  
  Cell[StyleData["M4", "EnhancedPrintoutNonGray"],
    ShowCellBracket->True,
    StyleMenuListing->None,
    FontFamily->"Mathematica4",
    CharacterEncoding->"Mathematica4"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SO10"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Plain",
    FontSlant->"Italic"],
  
  Cell[StyleData["SO10", "Printout"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->8,
    FontWeight->"Plain",
    FontSlant->"Italic"],
  
  Cell[StyleData["SO10", "EnhancedPrintout"],
    StyleMenuListing->None,
    FontFamily->"Frutiger 56 Italic",
    FontSize->8,
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["SO10", "EnhancedPrintoutNonGray"],
    StyleMenuListing->None,
    FontFamily->"Frutiger 56 Italic",
    FontSize->8,
    FontWeight->"Plain",
    FontSlant->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["S9"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Plain"],
  
  Cell[StyleData["S9", "Printout"],
    FontSize->9,
    FontWeight->"Plain"],
  
  Cell[StyleData["S9", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman",
    FontSize->9,
    FontWeight->"Plain"],
  
  Cell[StyleData["S9", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman",
    FontSize->9,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["S10"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Plain"],
  
  Cell[StyleData["S10", "Printout"],
    FontSize->10,
    FontWeight->"Plain"],
  
  Cell[StyleData["S10", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman",
    FontSize->10,
    FontWeight->"Plain"],
  
  Cell[StyleData["S10", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman",
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SB9"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold",
    PrivateFontOptions->{"OperatorSubstitution"->False}],
  
  Cell[StyleData["SB9", "Printout"],
    FontSize->7],
  
  Cell[StyleData["SB9", "EnhancedPrintout"],
    FontFamily->"Frutiger 65 Bold",
    FontSize->7,
    FontWeight->"Plain"],
  
  Cell[StyleData["SB9", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 65 Bold",
    FontSize->7,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SB10"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Bold"],
  
  Cell[StyleData["SB10", "Printout"],
    FontSize->8],
  
  Cell[StyleData["SB10", "EnhancedPrintout"],
    FontFamily->"Frutiger 65 Bold",
    FontSize->8,
    FontWeight->"Plain"],
  
  Cell[StyleData["SB10", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 65 Bold",
    FontSize->8,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MR10"],
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    SingleLetterItalics->False,
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->10,
    FontWeight->"Plain",
    FontSlant->"Plain"],
  
  Cell[StyleData["MR10", "Printout"],
    FontSize->7],
  
  Cell[StyleData["MR10", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSize->7],
  
  Cell[StyleData["MR10", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["C1"],
    StyleMenuListing->None,
    FontColor->RGBColor[0.4, 0, 1]],
  
  Cell[StyleData["C1", "Printout"],
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["C1", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["C1", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["C1MR"],
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontColor->RGBColor[0.4, 0, 1]],
  
  Cell[StyleData["C1MR", "Printout"],
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["C1MR", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["C1MR", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallCaps"],
    StyleMenuListing->None,
    Magnification->0.8],
  
  Cell[StyleData["SmallCaps", "Printout"]],
  
  Cell[StyleData["SmallCaps", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["SmallCaps", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["KeyboardChars"],
    StyleMenuListing->None,
    FontSize->12],
  
  Cell[StyleData["KeyboardChars", "Printout"],
    FontSize->10],
  
  Cell[StyleData["KeyboardChars", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["KeyboardChars", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Emphasis Boxes and Pictures", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Box"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnSpacings->1}],
  
  Cell[StyleData["Box", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Box", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["Box", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["Box", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DoubleBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnSpacings->2,
    RowAlignments->Top}],
  
  Cell[StyleData["DoubleBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DoubleBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["DoubleBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["DoubleBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["1ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnSpacings->1}],
  
  Cell[StyleData["1ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["1ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["1ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["1ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    SingleLetterItalics->False,
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->{0.31, 0.67}}],
  
  Cell[StyleData["2ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["2ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnEvenBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->0.46}],
  
  Cell[StyleData["2ColumnEvenBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["2ColumnEvenBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnEvenBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnEvenBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnSmallBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnSpacings->1.5,
    ColumnWidths->0.35,
    ColumnAlignments->{Right, Left}}],
  
  Cell[StyleData["2ColumnSmallBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["2ColumnSmallBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnSmallBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["2ColumnSmallBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->0.32}],
  
  Cell[StyleData["3ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["3ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["3ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["3ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnSmallBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnSpacings->1.5,
    ColumnWidths->0.24,
    ColumnAlignments->{Right, Center, Left}}],
  
  Cell[StyleData["3ColumnSmallBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["3ColumnSmallBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["3ColumnSmallBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["3ColumnSmallBox", "EnhancedPrintoutNonGray"],
    CellFrame->0.5,
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["4ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    SingleLetterItalics->False,
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->{0.13, 0.35, 0.13, 0.35}}],
  
  Cell[StyleData["4ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["4ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["4ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["4ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["5ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->0.202}],
  
  Cell[StyleData["5ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["5ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["5ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["5ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->9,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["6ColumnBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxFrame->0.5,
    BoxMargins->True},
    GridBoxOptions->{ColumnWidths->{0.12, 0.22, 0.12, 0.12, 0.22, 0.12}}],
  
  Cell[StyleData["6ColumnBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["6ColumnBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["6ColumnBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["6ColumnBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FramedBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["FramedBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FramedBox", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["FramedBox", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["FramedBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DefinitionBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.4, 0.6},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["DefinitionBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DefinitionBox", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DefinitionBox3Col"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.35, 0.2, 0.45},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["DefinitionBox3Col", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DefinitionBox3Col", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox3Col", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox3Col", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DefinitionBox4Col"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.15, 0.35, 0.15, 0.35},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["DefinitionBox4Col", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DefinitionBox4Col", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox4Col", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox4Col", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DefinitionBox5Col"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->0.2,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["DefinitionBox5Col", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DefinitionBox5Col", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox5Col", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox5Col", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DefinitionBox6Col"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.13, 0.24, 0.13, 0.13, 0.24, 0.13},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["DefinitionBox6Col", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DefinitionBox6Col", "Printout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox6Col", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["DefinitionBox6Col", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TopBox"],
    CellFrame->{{0.5, 0.5}, {0, 0.5}},
    CellMargins->{{27, 12}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["TopBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TopBox", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    Background->GrayLevel[1]],
  
  Cell[StyleData["TopBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1]],
  
  Cell[StyleData["TopBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MiddleBox"],
    CellFrame->{{0.5, 0.5}, {0, 0}},
    CellMargins->{{27, 12}, {0, -7}},
    CellHorizontalScrolling->True,
    PageBreakAbove->False,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["MiddleBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["MiddleBox", "Printout"],
    CellMargins->{{2, 0}, {0, 2}},
    CellFrameColor->GrayLevel[0],
    Background->GrayLevel[1]],
  
  Cell[StyleData["MiddleBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 4}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1]],
  
  Cell[StyleData["MiddleBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 4}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BottomBox"],
    CellFrame->{{0.5, 0.5}, {0.5, 0}},
    CellMargins->{{27, 12}, {0, -7}},
    CellHorizontalScrolling->True,
    PageBreakAbove->False,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["BottomBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["BottomBox", "Printout"],
    CellMargins->{{2, 0}, {0, -5}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->2.2}],
  
  Cell[StyleData["BottomBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, -5}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->2.2}],
  
  Cell[StyleData["BottomBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, -5}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->2.2}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TopSpanBox"],
    CellFrame->{{0.5, 0.5}, {0, 0.5}},
    CellMargins->{{27, 12}, {-2, 8}},
    CellHorizontalScrolling->True,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.9, 0.03},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["TopSpanBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TopSpanBox", "Printout"],
    CellMargins->{{2, 0}, {-2, 8}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["TopSpanBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {-4, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]],
  
  Cell[StyleData["TopSpanBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {-4, 8}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MiddleSpanBox"],
    CellFrame->{{0.5, 0.5}, {0, 0}},
    CellMargins->{{27, 12}, {0, 0}},
    CellHorizontalScrolling->True,
    PageBreakAbove->False,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.9, 0.03},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["MiddleSpanBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["MiddleSpanBox", "Printout"],
    CellMargins->{{2, 0}, {-5, 0}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}],
  
  Cell[StyleData["MiddleSpanBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {-7, 0}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}],
  
  Cell[StyleData["MiddleSpanBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {-7, 0}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    FontSize->10,
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Picture"],
    CellMargins->{{27, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["Picture", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Picture", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.65],
  
  Cell[StyleData["Picture", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.65],
  
  Cell[StyleData["Picture", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.65]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OpenCloseItemizedPicture"],
    CellMargins->{{88, 4}, {4, 4}},
    PrivateCellOptions->{"DefaultCellGroupOpen"->False},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["OpenCloseItemizedPicture", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["OpenCloseItemizedPicture", "Printout"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False],
  
  Cell[StyleData["OpenCloseItemizedPicture", "EnhancedPrintout"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False],
  
  Cell[StyleData["OpenCloseItemizedPicture", "EnhancedPrintoutNonGray"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedPicture"],
    CellMargins->{{88, 4}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedPicture", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedPicture", "Printout"],
    CellMargins->{{77, 2}, {4, -4}},
    Magnification->0.5],
  
  Cell[StyleData["ItemizedPicture", "EnhancedPrintout"],
    CellMargins->{{78, 2}, {12, -6}},
    Magnification->0.5],
  
  Cell[StyleData["ItemizedPicture", "EnhancedPrintoutNonGray"],
    CellMargins->{{75, 2}, {4, -6}},
    Magnification->0.5]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ListGraphic"],
    CellMargins->{{88, 4}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ListGraphic", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ListGraphic", "Printout"],
    CellMargins->{{77, 2}, {4, -4}},
    Magnification->0.5],
  
  Cell[StyleData["ListGraphic", "EnhancedPrintout"],
    CellMargins->{{78, 2}, {12, -6}},
    Magnification->0.5],
  
  Cell[StyleData["ListGraphic", "EnhancedPrintoutNonGray"],
    CellMargins->{{75, 2}, {4, -6}},
    Magnification->0.5]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ListNoteBox"],
    CellFrame->0.5,
    CellMargins->{{88, 12}, {8, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxMargins->{{1, 1}, {1.5, 1.5}}},
    GridBoxOptions->{ColumnSpacings->1}],
  
  Cell[StyleData["ListNoteBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ListNoteBox", "Printout"],
    CellMargins->{{77, 4}, {6, 2}},
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ListNoteBox", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ListNoteBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{77, 4}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureGroup"],
    CellMargins->{{41, 4}, {0, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureGroup", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureGroup", "Printout"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False],
  
  Cell[StyleData["PictureGroup", "EnhancedPrintout"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False],
  
  Cell[StyleData["PictureGroup", "EnhancedPrintoutNonGray"],
    CellMargins->{{76, 2}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->1,
    "ClosedCellHeight"->0},
    CellOpen->False]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Sound"],
    ShowCellBracket->True,
    CellMargins->{{27, Inherited}, {0, 8}},
    StyleMenuListing->None],
  
  Cell[StyleData["Sound", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Sound", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["Sound", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["Sound", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caption"],
    CellMargins->{{27, 4}, {0, 3}},
    PageBreakAbove->False,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Caption", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["Caption", "Printout"],
    CellMargins->{{2, 4}, {2, 0}},
    FontSize->7],
  
  Cell[StyleData["Caption", "EnhancedPrintout"],
    CellMargins->{{2, 4}, {2, 2}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->7],
  
  Cell[StyleData["Caption", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {2, -2}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->7]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Tables", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnTable"],
    CellMargins->{{35, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->{0.34, 0.64},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["2ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["2ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["2ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["2ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["2ColumnEvenTable"],
    CellMargins->{{35, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->0.49,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["2ColumnEvenTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["2ColumnEvenTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["2ColumnEvenTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["2ColumnEvenTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnTable"],
    CellMargins->{{35, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->{0.28, 0.28, 0.43},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["3ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["3ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["3ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9],
  
  Cell[StyleData["3ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InlineFormula"],
    CellMargins->{{27, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    LanguageCategory->"Formula",
    AutoSpacing->True,
    ScriptLevel->1,
    SingleLetterItalics->True,
    SpanMaxSize->Infinity,
    StyleMenuListing->None,
    GridBoxOptions->{ColumnWidths->Automatic}],
  
  Cell[StyleData["InlineFormula", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["InlineFormula", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["InlineFormula", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InlineOutput"],
    CellMargins->{{27, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    LanguageCategory->None,
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["InlineOutput", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9.5],
  
  Cell[StyleData["InlineOutput", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9.5],
  
  Cell[StyleData["InlineOutput", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}},
    FontSize->9.5]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{66, Inherited}, {Inherited, 10}},
    CellHorizontalScrolling->True,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    LanguageCategory->"Formula",
    ScriptLevel->0,
    SingleLetterItalics->True,
    SpanMaxSize->Infinity,
    StyleMenuListing->None,
    UnderoverscriptBoxOptions->{LimitsPositioning->True},
    GridBoxOptions->{ColumnWidths->Automatic}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DisplayFormula", "Printout"]],
  
  Cell[StyleData["DisplayFormula", "EnhancedPrintout"],
    FontFamily->"Palatino"],
  
  Cell[StyleData["DisplayFormula", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayMath"],
    CellMargins->{{66, 10}, {7, 7}},
    Evaluatable->False,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    HyphenationOptions->{"HyphenationCharacter"->"\[Continuation]"},
    AutoItalicWords->{},
    LanguageCategory->"Formula",
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input"],
  
  Cell[StyleData["DisplayMath", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["DisplayMath", "Printout"],
    CellMargins->{{39, 0}, {6, 6}},
    FontSize->9],
  
  Cell[StyleData["DisplayMath", "EnhancedPrintout"],
    CellMargins->{{39, 0}, {4, 6}},
    FontFamily->"WRICourier",
    FontSize->9],
  
  Cell[StyleData["DisplayMath", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 0}, {4, 6}},
    FontFamily->"WRICourier",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{66, 27}, {7, 7}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    LanguageCategory->"Formula",
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{39, 55}, {0, 10}},
    FontSize->10],
  
  Cell[StyleData["NumberedEquation", "EnhancedPrintout"],
    CellMargins->{{39, 55}, {0, 10}},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["NumberedEquation", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 55}, {0, 10}},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["EquationNumber"],
    CellMargins->{{66, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1.83, 0},
    LanguageCategory->"Formula"],
  
  Cell[StyleData["EquationNumber", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["EquationNumber", "Printout"],
    LineSpacing->{1.7, 0}],
  
  Cell[StyleData["EquationNumber", "EnhancedPrintout"],
    LineSpacing->{1.7, 0},
    FontFamily->"Palatino"],
  
  Cell[StyleData["EquationNumber", "EnhancedPrintoutNonGray"],
    LineSpacing->{1.7, 0},
    FontFamily->"Palatino"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{66, Inherited}, {Inherited, 10}},
    CellHorizontalScrolling->True,
    ShowSpecialCharacters->Automatic,
    AutoItalicWords->{},
    LanguageCategory->"Formula",
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{42, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "EnhancedPrintout"],
    CellMargins->{{42, 30}, {Inherited, 4}},
    FontFamily->"WRICourier",
    FontSize->9.5],
  
  Cell[StyleData["Program", "EnhancedPrintoutNonGray"],
    CellMargins->{{42, 30}, {Inherited, 4}},
    FontFamily->"WRICourier",
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Help Browser Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell["\<\
Standard Styles
\
\>", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InspectorLevel1"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 10}, {2, 7}},
    CellGroupingRules->{"SectionGrouping", 60},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphIndent->-38,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold"],
  
  Cell[StyleData["InspectorLevel1", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["InspectorLevel1", "SlideShow"],
    HyphenationOptions->{"HyphenationCharacter"->"-"}],
  
  Cell[StyleData["InspectorLevel1", "Printout"],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InspectorLevel2"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 14}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->14,
    FontWeight->"Bold"],
  
  Cell[StyleData["InspectorLevel2", "Presentation"],
    FontSize->21],
  
  Cell[StyleData["InspectorLevel2", "SlideShow"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{48, 10}, {2, 12}},
    CellGroupingRules->{"SectionGrouping", 70},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphIndent->-27,
    FontFamily->"Helvetica",
    FontSize->14,
    FontWeight->"Bold"],
  
  Cell[StyleData["InspectorLevel2", "Printout"],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ObjectName"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->{"SectionGrouping", 50},
    CellFrameMargins->{{27, 12}, {3, 15}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->18,
    FontWeight->"Bold",
    Background->RGBColor[0.960784, 0.878431, 0.666667]],
  
  Cell[StyleData["ObjectName", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["ObjectName", "SlideShow"],
    CellMargins->{{72, 4}, {1, 14}},
    CellGroupingRules->{"SectionGrouping", 90},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["ObjectName", "Printout"],
    CellMargins->{{0, 0}, {0, 10}},
    CellFrameMargins->{{24, 8}, {Inherited, Inherited}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->12]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubObjectNameClosed"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {1, 10}},
    PrivateCellOptions->{"DefaultCellGroupOpen"->False},
    CellGroupingRules->{"SectionGrouping", 90},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->14,
    FontWeight->"Bold"],
  
  Cell[StyleData["SubObjectNameClosed", "Presentation"],
    FontSize->21],
  
  Cell[StyleData["SubObjectNameClosed", "SlideShow"],
    CellMargins->{{72, 4}, {1, 14}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["SubObjectNameClosed", "Printout"],
    CellMargins->{{2, 0}, {1, 10}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubObjectName"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {1, 10}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->18,
    FontWeight->"Bold"],
  
  Cell[StyleData["SubObjectName", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["SubObjectName", "SlideShow"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{90, 4}, {1, 14}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontFamily->"Courier",
    FontSize->12],
  
  Cell[StyleData["SubObjectName", "Printout"],
    CellMargins->{{2, 0}, {1, 10}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontFamily->"Courier",
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GroupObjectName"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 14}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->18,
    FontWeight->"Bold"],
  
  Cell[StyleData["GroupObjectName", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["GroupObjectName", "SlideShow"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{72, 10}, {2, 12}},
    CellGroupingRules->{"SectionGrouping", 70},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphIndent->-27,
    CounterIncrements->"Outline2",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["GroupObjectName", "Printout"],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GroupObjectNameClosed"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 14}},
    PrivateCellOptions->{"DefaultCellGroupOpen"->False},
    CellGroupingRules->{"SectionGrouping", 80},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->18,
    FontWeight->"Bold"],
  
  Cell[StyleData["GroupObjectNameClosed", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["GroupObjectNameClosed", "SlideShow"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{72, 10}, {2, 12}},
    CellGroupingRules->{"SectionGrouping", 70},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphIndent->-27,
    CounterIncrements->"Outline2",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["GroupObjectNameClosed", "Printout"],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ObjectNameWarningMessages"],
    CellMargins->{{27, 4}, {1, 10}},
    CellGroupingRules->{"SectionGrouping", 65},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->18,
    FontWeight->"Bold"],
  
  Cell[StyleData["ObjectNameWarningMessages", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["ObjectNameWarningMessages", "SlideShow"],
    CellMargins->{{10, 4}, {1, 10}},
    CellGroupingRules->{"SectionGrouping", 65},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontFamily->"Courier",
    FontSize->16,
    FontWeight->"Bold"],
  
  Cell[StyleData["ObjectNameWarningMessages", "Printout"],
    CellMargins->{{2, 0}, {1, 10}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Command"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{0, 0}, {0, 0}},
    CellFrameMargins->{{27, 12}, {3, 15}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    CounterIncrements->"Section",
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->18,
    FontWeight->"Bold",
    Background->RGBColor[0.960784, 0.878431, 0.666667]],
  
  Cell[StyleData["Command", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["Command", "Printout"],
    CellMargins->{{0, 0}, {0, 14}},
    PageBreakBelow->False,
    CellFrameMargins->{{24, 8}, {Inherited, Inherited}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CharacterName"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{0, 0}, {0, 0}},
    CellFrameMargins->{{27, 27}, {8, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216]],
  
  Cell[StyleData["CharacterName", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["CharacterName", "SlideShow"],
    CellMargins->{{72, 4}, {1, 14}},
    CellGroupingRules->{"SectionGrouping", 90},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["CharacterName", "Printout"],
    CellMargins->{{2, 0}, {0, 4}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Usage"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{0, 0}, {0, 0}},
    CellFrameMargins->{{27, 27}, {8, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    Hyphenation->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 6},
    ParagraphIndent->-10,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216]],
  
  Cell[StyleData["Usage", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Usage", "SlideShow"],
    CellFrame->{{0, 0}, {0.5, 0.5}},
    CellMargins->{{72, 4}, {0, 4}},
    CellFrameMargins->5,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 6},
    ParagraphIndent->-10,
    StyleMenuListing->None],
  
  Cell[StyleData["Usage", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    CellFrameMargins->{{24, 8}, {Inherited, Inherited}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ErrorMessage"],
    CellFrame->{{0, 0}, {0.5, 0.5}},
    ShowCellBracket->True,
    CellMargins->{{27, 4}, {0, 4}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{14, Inherited}, {Inherited, Inherited}},
    CellFrameMargins->5,
    DefaultFormatType->DefaultOutputFormatType,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FormatType->InputForm,
    StyleMenuListing->None,
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["ErrorMessage", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ErrorMessage", "SlideShow"],
    CellMargins->{{72, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None],
  
  Cell[StyleData["ErrorMessage", "Printout"],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Notes"],
    CellMargins->{{27, 27}, {0, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 6},
    ParagraphIndent->-10,
    StyleMenuListing->None,
    FontFamily->"Times"],
  
  Cell[StyleData["Notes", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Notes", "SlideShow"],
    CellMargins->{{72, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 6},
    ParagraphIndent->-10,
    StyleMenuListing->None,
    FontFamily->"Times"],
  
  Cell[StyleData["Notes", "Printout"],
    CellMargins->{{24, 0}, {0, 8}},
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NotesList"],
    CellMargins->{{20, 4}, {2, 8}},
    CellHorizontalScrolling->True,
    ScriptLevel->1,
    SingleLetterItalics->True,
    StyleMenuListing->None,
    FontFamily->"Times"],
  
  Cell[StyleData["NotesList", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NotesList", "SlideShow"],
    CellMargins->{{82, 4}, {2, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ScriptLevel->1,
    SingleLetterItalics->True,
    StyleMenuListing->None,
    FontFamily->"Times"],
  
  Cell[StyleData["NotesList", "Printout"],
    CellMargins->{{12, 0}, {2, 8}},
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ReferenceSection"],
    CellFrame->{{0, 0}, {0, 0.5}},
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 10}},
    CellGroupingRules->{"SectionGrouping", 30},
    CellFrameMargins->12,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold"],
  
  Cell[StyleData["ReferenceSection", "Presentation"],
    FontSize->20],
  
  Cell[StyleData["ReferenceSection", "Printout"],
    CellMargins->{{3, 0}, {0, 10}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Reference"],
    CellMargins->{{27, 27}, {0, 8}},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"Reference",
    FontSize->12],
  
  Cell[StyleData["Reference", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Reference", "SlideShow"],
    CellMargins->{{27, 27}, {4, 4}},
    Hyphenation->False,
    FontSize->14],
  
  Cell[StyleData["Reference", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["Reference", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["Reference", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExampleSection"],
    CellFrame->{{0, 0}, {0, 0.5}},
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 10}},
    CellGroupingRules->{"SectionGrouping", 70},
    CellFrameMargins->12,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold"],
  
  Cell[StyleData["ExampleSection", "Presentation"],
    FontSize->20],
  
  Cell[StyleData["ExampleSection", "Printout"],
    CellMargins->{{3, 0}, {0, 10}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExampleText"],
    CellMargins->{{27, 40}, {0, 18}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 6},
    StyleMenuListing->None,
    FontColor->RGBColor[0, 0, 0.4]],
  
  Cell[StyleData["ExampleText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ExampleText", "Printout"],
    CellMargins->{{2, 80}, {0, 8}},
    PageBreakBelow->False,
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExampleSubsection"],
    ShowCellBracket->True,
    CellMargins->{{27, 4}, {0, 16}},
    CellGroupingRules->{"SectionGrouping", 80},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Bold"],
  
  Cell[StyleData["ExampleSubsection", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["ExampleSubsection", "Printout"],
    CellMargins->{{12, 0}, {0, 16}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MessagesSection"],
    CellFrame->{{0, 0}, {0, 0.5}},
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{27, 4}, {0, 10}},
    CellGroupingRules->{"SectionGrouping", 70},
    CellFrameMargins->12,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold"],
  
  Cell[StyleData["MessagesSection", "Presentation"],
    FontSize->20],
  
  Cell[StyleData["MessagesSection", "Printout"],
    CellMargins->{{3, 0}, {0, 10}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TemplateSection"],
    CellFrame->{{0, 0}, {0, 0.5}},
    CellMargins->{{27, 4}, {0, 10}},
    PrivateCellOptions->{"DefaultCellGroupOpen"->True},
    CellGroupingRules->{"SectionGrouping", 70},
    CellFrameMargins->12,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontWeight->"Bold"],
  
  Cell[StyleData["TemplateSection", "Presentation"],
    FontSize->20],
  
  Cell[StyleData["TemplateSection", "Printout"],
    CellMargins->{{3, 0}, {0, 10}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["LineHead"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{27, 4}, {8, 18}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["LineHead", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["LineHead", "Printout"],
    CellMargins->{{2, 0}, {8, 18}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FlushInput"],
    CellMargins->{{27, 4}, {1, 4}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    ShowSpecialCharacters->Automatic,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FormatType->StandardForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    StyleMenuListing->None,
    FontWeight->"Bold"],
  
  Cell[StyleData["FlushInput", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FlushInput", "Printout"],
    CellMargins->{{2, 0}, {1, 4}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FlushOutput"],
    CellMargins->{{27, Inherited}, {Inherited, Inherited}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelAutoDelete->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FormatType->StandardForm,
    StyleMenuListing->None],
  
  Cell[StyleData["FlushOutput", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FlushOutput", "Printout"],
    CellMargins->{{2, Inherited}, {Inherited, Inherited}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DialogItem"],
    CellDingbat->"\[EmptyDiamond]",
    CellMargins->{{27, 4}, {0, 12}},
    CellGroupingRules->{"SectionGrouping", 70},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Bold"],
  
  Cell[StyleData["DialogItem", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["DialogItem", "Printout"],
    CellMargins->{{12, 0}, {0, 12}},
    PageBreakBelow->False,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GroupDivider"],
    CellMargins->{{27, Inherited}, {Inherited, Inherited}},
    CellEditDuplicate->True,
    CellGroupingRules->{"TitleGrouping", 0},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None],
  
  Cell[StyleData["GroupDivider", "Presentation"],
    FontSize->18]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Highlight"],
    CellFrame->0.5,
    CellMargins->{{27, 4}, {0, 18}},
    CellFrameColor->GrayLevel[0.800015],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    Background->GrayLevel[0.950011]],
  
  Cell[StyleData["Highlight", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Highlight", "Printout"],
    CellMargins->{{2, 0}, {0, 18}},
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 3},
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["UserNotes"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{27, 4}, {0, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    ParagraphIndent->-7,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10],
  
  Cell[StyleData["UserNotes", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["UserNotes", "Printout"],
    CellMargins->{{2, 0}, {0, 4}},
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 3},
    FontSize->8],
  
  Cell[StyleData["UserNotes", "EnhancedPrintout"],
    CellMargins->{{0, 0}, {0, 4}},
    Hyphenation->True,
    LineSpacing->{1, 3},
    FontFamily->"Frutiger 55 Roman",
    FontSize->8],
  
  Cell[StyleData["UserNotes", "EnhancedPrintoutNonGray"],
    CellMargins->{{0, 0}, {0, 4}},
    Hyphenation->True,
    LineSpacing->{1, 3},
    FontFamily->"Frutiger 55 Roman",
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MoreInfo"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    ParagraphIndent->-7,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    Background->RGBColor[0.964706, 0.929412, 0.839216]],
  
  Cell[StyleData["MoreInfo", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["MoreInfo", "Printout"],
    CellMargins->{{2, 0}, {0, 4}},
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 3},
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["UserTips"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {0, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    ParagraphIndent->-7,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10,
    Background->RGBColor[0.964706, 0.929412, 0.839216]],
  
  Cell[StyleData["UserTips", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["UserTips", "Printout"],
    CellMargins->{{2, 0}, {0, 4}},
    CellFrameColor->GrayLevel[0.800015],
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 3},
    FontSize->8,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["UserTips", "EnhancedPrintout"],
    CellMargins->{{0, 0}, {0, 4}},
    CellFrameColor->GrayLevel[0],
    Hyphenation->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 3},
    FontFamily->"Frutiger 55 Roman",
    FontSize->8,
    Background->GrayLevel[0.700008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{4, 0}, {0, 0}},
    CellMargins->{{27, 62}, {0, 14}},
    CellFrameMargins->4,
    CellFrameColor->RGBColor[0, 0, 0.8],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 8}],
  
  Cell[StyleData["Theorem", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{2, 0}, {0, 14}},
    CellFrameMargins->3,
    CellFrameColor->GrayLevel[0.8],
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->9,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Commentary"],
    CellFrame->{{2, 0}, {0, 0}},
    CellMargins->{{36, 10}, {7, 7}},
    PageBreakBelow->False,
    CellFrameMargins->8,
    CellFrameColor->RGBColor[0, 0.2, 1],
    LineSpacing->{1, 3},
    FontSlant->"Italic"],
  
  Cell[StyleData["Commentary", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Commentary", "Printout"],
    CellMargins->{{36, 0}, {6, 6}},
    CellFrameColor->GrayLevel[0.8],
    FontSize->10],
  
  Cell[StyleData["Commentary", "EnhancedPrintout"],
    CellMargins->{{36, 0}, {6, 6}},
    CellFrameColor->GrayLevel[0.8],
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["Commentary", "EnhancedPrintoutNonGray"],
    CellMargins->{{36, 0}, {6, 6}},
    CellFrameColor->GrayLevel[0.8],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Level1Icon"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{48, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 8},
    StyleMenuListing->None],
  
  Cell[StyleData["Level1Icon", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Level1Icon", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Level2Icon"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{48, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    ParagraphSpacing->{0, 8},
    StyleMenuListing->None],
  
  Cell[StyleData["Level2Icon", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Level2Icon", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    HyphenationOptions->{"HyphenationCharacter"->"-"},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CopyrightHead"],
    CellMargins->{{27, 4}, {0, 8}},
    PageBreakAbove->False,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["CopyrightHead", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["CopyrightHead", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{27, 4}, {0, 8}},
    PageBreakAbove->False,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->10],
  
  Cell[StyleData["Copyright", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->8],
  
  Cell[StyleData["Copyright", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Frutiger 55 Roman",
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RegisterSubsection"],
    CellDingbat->None,
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 45},
    CellFrameLabelMargins->6,
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    FontFamily->"Helvetica",
    FontSize->16],
  
  Cell[StyleData["RegisterSubsection", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["RegisterSubsection", "SlideShow"],
    CellDingbat->None,
    CellMargins->{{70, 0}, {4, 12}},
    FontSize->12],
  
  Cell[StyleData["RegisterSubsection", "Printout"],
    CellMargins->{{9, 0}, {5, 24}},
    FontSize->12],
  
  Cell[StyleData["RegisterSubsection", "EnhancedPrintout"],
    CellMargins->{{9, 0}, {5, 20}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->12],
  
  Cell[StyleData["RegisterSubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{9, 0}, {5, 20}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RegisterSubsubsection"],
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 50},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["RegisterSubsubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["RegisterSubsubsection", "Printout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontSize->10],
  
  Cell[StyleData["RegisterSubsubsection", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["RegisterSubsubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RegisterText"],
    CellMargins->{{27, 4}, {0, 8}},
    PageBreakAbove->False,
    StyleMenuListing->None,
    FontSize->10],
  
  Cell[StyleData["RegisterText", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["RegisterText", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->8],
  
  Cell[StyleData["RegisterText", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Palatino",
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CaptionHead"],
    CellMargins->{{27, 4}, {0, 8}},
    PageBreakAbove->False,
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["CaptionHead", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["CaptionHead", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->8],
  
  Cell[StyleData["CaptionHead", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Palatino",
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CaptionText"],
    CellMargins->{{66, 4}, {0, 8}},
    PageBreakAbove->False,
    StyleMenuListing->None],
  
  Cell[StyleData["CaptionText", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["CaptionText", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->8],
  
  Cell[StyleData["CaptionText", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Palatino",
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BulletText"],
    CellMargins->{{76, 4}, {4, 4}},
    CellFrameLabels->{{
      Cell[ "\[Bullet]", "BulletText", CellBaseline -> Baseline], 
          Inherited}, {Inherited, Inherited}},
    Hyphenation->False,
    LineSpacing->{1, 3},
    CounterIncrements->"BulletText",
    StyleMenuListing->None],
  
  Cell[StyleData["BulletText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["BulletText", "Printout"],
    FontSize->10],
  
  Cell[StyleData["BulletText", "EnhancedPrintout"],
    CellMargins->{{16, 4}, {2, 4}},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["BulletText", "EnhancedPrintoutNonGray"],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["HelpLabels"],
    CellMargins->{{88, 4}, {0, 8}},
    StyleMenuListing->None],
  
  Cell[StyleData["HelpLabels", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["HelpLabels", "Printout"],
    CellMargins->{{34, 2}, {6, 6}},
    FontSize->10],
  
  Cell[StyleData["HelpLabels", "EnhancedPrintout"],
    CellMargins->{{34, 2}, {6, 6}},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Manual.nb Merge", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell["Function List", "Subsubsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionName"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{24, 10}, {10, 12}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->14,
    FontWeight->"Bold",
    CellTags->"FunctionName"],
  
  Cell[StyleData["FunctionName", "Presentation"],
    FontSize->21],
  
  Cell[StyleData["FunctionName", "Printout"],
    CellMargins->{{9, 0}, {10, 10}},
    FontSize->12],
  
  Cell[StyleData["FunctionName", "EnhancedPrintout"],
    CellMargins->{{9, 0}, {10, 10}},
    FontFamily->"Courier",
    FontSize->12],
  
  Cell[StyleData["FunctionName", "EnhancedPrintoutNonGray"],
    CellMargins->{{9, 0}, {10, 10}},
    FontFamily->"Courier",
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionSubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{36, 10}, {0, 6}},
    CellGroupingRules->{"SectionGrouping", 60},
    PageBreakBelow->False,
    StyleMenuListing->None,
    FontWeight->"Bold",
    CellTags->"FunctionSubsection"],
  
  Cell[StyleData["FunctionSubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionSubsection", "Printout"],
    CellMargins->{{30, 0}, {0, 6}}],
  
  Cell[StyleData["FunctionSubsection", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {0, 6}},
    FontFamily->"Palatino",
    FontSize->11],
  
  Cell[StyleData["FunctionSubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {0, 6}},
    FontFamily->"Palatino",
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionDescription"],
    CellMargins->{{36, 10}, {4, 0}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0.7, 0},
    CounterIncrements->"FunctionDescription",
    StyleMenuListing->None,
    FontWeight->"Plain"],
  
  Cell[StyleData["FunctionDescription", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionDescription", "Printout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontSize->11],
  
  Cell[StyleData["FunctionDescription", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["FunctionDescription", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionDescriptionRuled"],
    CellFrame->{{0, 0}, {0.25, 0.25}},
    CellMargins->{{36, 10}, {4, 0}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0.7, 0},
    CounterIncrements->"FunctionDescriptionRuled",
    StyleMenuListing->None],
  
  Cell[StyleData["FunctionDescriptionRuled", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionDescriptionRuled", "Printout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontSize->11],
  
  Cell[StyleData["FunctionDescriptionRuled", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["FunctionDescriptionRuled", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionDescriptionRuleTop"],
    CellFrame->{{0, 0}, {0, 0.25}},
    CellMargins->{{36, 10}, {4, 0}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0.7, 0},
    CounterIncrements->"FunctionDescriptionRuleTop",
    StyleMenuListing->None],
  
  Cell[StyleData["FunctionDescriptionRuleTop", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionDescriptionRuleTop", "Printout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontSize->11],
  
  Cell[StyleData["FunctionDescriptionRuleTop", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["FunctionDescriptionRuleTop", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionDescriptionRuleBottom"],
    CellFrame->{{0, 0}, {0.25, 0}},
    CellMargins->{{36, 10}, {4, 0}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0.7, 0},
    CounterIncrements->"FunctionDescriptionRuleBottom",
    StyleMenuListing->None],
  
  Cell[StyleData["FunctionDescriptionRuleBottom", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionDescriptionRuleBottom", "Printout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontSize->11],
  
  Cell[StyleData["FunctionDescriptionRuleBottom", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["FunctionDescriptionRuleBottom", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {4, 0}},
    TextJustification->0,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionExample"],
    CellMargins->{{36, 10}, {4, 0}},
    StyleMenuListing->None,
    FontSize->10],
  
  Cell[StyleData["FunctionExample", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["FunctionExample", "Printout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontSize->9],
  
  Cell[StyleData["FunctionExample", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {4, 0}},
    FontFamily->"Palatino",
    FontSize->9],
  
  Cell[StyleData["FunctionExample", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {4, 0}},
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["FunctionTextNote"],
    CellMargins->{{36, 4}, {0, 4}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 4},
    LineIndent->0,
    StyleMenuListing->None],
  
  Cell[StyleData["FunctionTextNote", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["FunctionTextNote", "Printout"],
    CellMargins->{{36, 4}, {0, 4}},
    FontSize->9],
  
  Cell[StyleData["FunctionTextNote", "EnhancedPrintout"],
    CellMargins->{{36, 4}, {0, 4}},
    FontFamily->"Palatino",
    FontSize->8],
  
  Cell[StyleData["FunctionTextNote", "EnhancedPrintoutNonGray"],
    CellMargins->{{36, 4}, {0, 4}},
    FontFamily->"Palatino",
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["3ColumnFunctionTable"],
    CellMargins->{{36, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->{0.28, 0.28, 0.43},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["3ColumnFunctionTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["3ColumnFunctionTable", "Printout"],
    CellMargins->{{30, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["3ColumnFunctionTable", "EnhancedPrintout"],
    CellMargins->{{30, 0}, {0, 8}},
    FontSize->10],
  
  Cell[StyleData["3ColumnFunctionTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{30, 0}, {0, 8}},
    FontSize->10]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Browser Emulation", "Subsubsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TextNormal"],
    CellMargins->{{12, 10}, {7, 7}},
    LineSpacing->{1, 3},
    CounterIncrements->"TextNormal",
    StyleMenuListing->None,
    FontFamily->"Times",
    FontSize->12],
  
  Cell[StyleData["TextNormal", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TextNormal", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BrowserEmulation"],
    CellMargins->{{12, 1}, {7, 24}},
    CellGroupingRules->{"SectionGrouping", 70},
    LineSpacing->{1, 3},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    GridBoxOptions->{ColumnWidths->0.25,
    ColumnAlignments->{Left},
    GridFrame->True,
    GridFrameMargins->0.5,
    ColumnLines->True}],
  
  Cell[StyleData["BrowserEmulation", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["BrowserEmulation", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CategoryDefs"],
    CellMargins->{{300, 10}, {7, 7}},
    LineSpacing->{1, 3},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    Background->RGBColor[1, 0.8, 0.6],
    GridBoxOptions->{ColumnAlignments->{Left},
    GridFrame->False,
    RowLines->True,
    ColumnLines->True}],
  
  Cell[StyleData["CategoryDefs", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["CategoryDefs", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Category1"],
    ShowCellBracket->False,
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{12, 10}, {7, 7}},
    CellGroupingRules->{"SectionGrouping", 50},
    StyleMenuListing->None,
    FontSize->16,
    FontWeight->"Bold"],
  
  Cell[StyleData["Category1", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["Category1", "Printout"],
    CellMargins->{{2, 2}, {6, 6}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Category2"],
    ShowCellBracket->False,
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{85, 10}, {7, 7}},
    CellGroupingRules->{"SectionGrouping", 60},
    StyleMenuListing->None,
    FontSize->14,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.500008]],
  
  Cell[StyleData["Category2", "Presentation"],
    FontSize->21],
  
  Cell[StyleData["Category2", "Printout"],
    CellMargins->{{2, 2}, {6, 6}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Category3"],
    ShowCellBracket->False,
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{157, 10}, {7, 7}},
    CellGroupingRules->{"SectionGrouping", 70},
    StyleMenuListing->None,
    FontSize->12,
    FontWeight->"Bold",
    FontColor->RGBColor[0.2, 0, 0.4]],
  
  Cell[StyleData["Category3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Category3", "Printout"],
    CellMargins->{{2, 2}, {6, 6}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Category4"],
    ShowCellBracket->False,
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{230, 10}, {7, 7}},
    CellGroupingRules->{"SectionGrouping", 80},
    StyleMenuListing->None,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0.300008, 0.4]],
  
  Cell[StyleData["Category4", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Category4", "Printout"],
    CellMargins->{{2, 2}, {6, 6}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Misc", "Subsubsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["4ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->0.25,
    ColumnAlignments->{Left, Center}}],
  
  Cell[StyleData["4ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["4ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["4ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["4ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["5ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->0.2,
    ColumnAlignments->{Left, Center}}],
  
  Cell[StyleData["5ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["5ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["5ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["5ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["6ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->0.16,
    ColumnAlignments->{Left, Center}}],
  
  Cell[StyleData["6ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["6ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["6ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["6ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["7ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->0.14,
    ColumnAlignments->{Left, Center}}],
  
  Cell[StyleData["7ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["7ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["7ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["7ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["8ColumnTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnWidths->0.12,
    ColumnAlignments->{Left, Center}}],
  
  Cell[StyleData["8ColumnTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["8ColumnTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["8ColumnTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["8ColumnTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ABItemizedText"],
    CellMargins->{{20, 4}, {0, 5}},
    LineSpacing->{1, 3},
    ParagraphIndent->-21,
    CounterIncrements->"ItemizedText",
    StyleMenuListing->None],
  
  Cell[StyleData["ABItemizedText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ABItemizedText", "Printout"],
    Hyphenation->True,
    ParagraphIndent->-21,
    FontSize->11],
  
  Cell[StyleData["ABItemizedText", "EnhancedPrintout"],
    Hyphenation->True,
    ParagraphIndent->-21,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ABItemizedText", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    ParagraphIndent->-21,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChapNumber"],
    CellMargins->{{27, Inherited}, {20, 40}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->42,
    FontWeight->"Plain",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["ChapNumber", "Presentation"],
    FontSize->63],
  
  Cell[StyleData["ChapNumber", "Printout"],
    CellMargins->{{2, 10}, {15, 30}},
    FontSize->36],
  
  Cell[StyleData["ChapNumber", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {15, 30}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->36,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["ChapNumber", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {15, 30}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->36,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChapTitle"],
    CellMargins->{{27, Inherited}, {80, 10}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->36,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["ChapTitle", "Presentation"],
    FontSize->54],
  
  Cell[StyleData["ChapTitle", "Printout"],
    CellMargins->{{2, 10}, {80, 6}},
    FontSize->30],
  
  Cell[StyleData["ChapTitle", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {80, 6}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->30,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["ChapTitle", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {80, 6}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->30,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemDingbat"],
    CellSize->{22, 14},
    TextAlignment->Right,
    StyleMenuListing->None,
    FontWeight->"Bold"],
  
  Cell[StyleData["ItemDingbat", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemDingbat", "Printout"],
    CellSize->{22, 13.125}],
  
  Cell[StyleData["ItemDingbat", "EnhancedPrintout"],
    CellSize->{22, 11}],
  
  Cell[StyleData["ItemDingbat", "EnhancedPrintoutNonGray"],
    CellSize->{22, 11}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item1"],
    CellMargins->{{19, 4}, {4, 4}},
    ParagraphIndent->-22,
    CounterIncrements->"Item1",
    StyleMenuListing->None],
  
  Cell[StyleData["Item1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item1", "Printout"],
    CellMargins->{{12, 0}, {2, 4}},
    FontSize->10],
  
  Cell[StyleData["Item1", "EnhancedPrintout"],
    CellMargins->{{16, 0}, {2, 4}},
    FontSize->10],
  
  Cell[StyleData["Item1", "EnhancedPrintoutNonGray"],
    CellMargins->{{16, 0}, {2, 4}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item1Continued"],
    CellMargins->{{41, 4}, {4, 1}},
    CounterIncrements->"Item1Continued",
    StyleMenuListing->None],
  
  Cell[StyleData["Item1Continued", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item1Continued", "Printout"],
    CellMargins->{{34, 0}, {4, 1}},
    FontSize->10],
  
  Cell[StyleData["Item1Continued", "EnhancedPrintout"],
    CellMargins->{{39, 0}, {4, 1}},
    FontSize->10],
  
  Cell[StyleData["Item1Continued", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 0}, {4, 1}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item2"],
    CellMargins->{{32, 4}, {4, 4}},
    ParagraphIndent->-22,
    CounterIncrements->"Item2",
    StyleMenuListing->None],
  
  Cell[StyleData["Item2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item2", "Printout"],
    CellMargins->{{26, 0}, {4, 4}},
    ParagraphIndent->-22,
    FontSize->10],
  
  Cell[StyleData["Item2", "EnhancedPrintout"],
    CellMargins->{{29, 0}, {2, 4}},
    ParagraphIndent->-22,
    FontSize->10],
  
  Cell[StyleData["Item2", "EnhancedPrintoutNonGray"],
    CellMargins->{{29, 0}, {2, 4}},
    ParagraphIndent->-22,
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item2Continued"],
    CellMargins->{{54, 4}, {4, 1}},
    CounterIncrements->"Item2Continued",
    StyleMenuListing->None],
  
  Cell[StyleData["Item2Continued", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item2Continued", "Printout"],
    CellMargins->{{27, 0}, {4, 1}}],
  
  Cell[StyleData["Item2Continued", "EnhancedPrintout"],
    CellMargins->{{52, 0}, {4, 1}},
    FontSize->10],
  
  Cell[StyleData["Item2Continued", "EnhancedPrintoutNonGray"],
    CellMargins->{{52, 0}, {4, 1}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item3"],
    CellMargins->{{45, 4}, {4, 4}},
    ParagraphIndent->-23,
    CounterIncrements->"Item3",
    StyleMenuListing->None],
  
  Cell[StyleData["Item3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item3", "Printout"],
    CellMargins->{{27, 0}, {2, 4}}],
  
  Cell[StyleData["Item3", "EnhancedPrintout"],
    CellMargins->{{42, 0}, {2, 4}},
    FontSize->10],
  
  Cell[StyleData["Item3", "EnhancedPrintoutNonGray"],
    CellMargins->{{42, 0}, {2, 4}},
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Item3Continued"],
    CellMargins->{{67, 4}, {4, 1}},
    CounterIncrements->"Item3Continued",
    StyleMenuListing->None],
  
  Cell[StyleData["Item3Continued", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Item3Continued", "Printout"],
    CellMargins->{{27, 0}, {4, 1}}],
  
  Cell[StyleData["Item3Continued", "EnhancedPrintout"],
    CellMargins->{{65, 0}, {4, 1}},
    FontSize->10],
  
  Cell[StyleData["Item3Continued", "EnhancedPrintoutNonGray"],
    CellMargins->{{65, 0}, {4, 1}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedBulletText"],
    CellMargins->{{19, 4}, {0, 5}},
    LineSpacing->{1, 3},
    ParagraphIndent->-22,
    CounterIncrements->"ItemizedBulletText",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedBulletText", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedBulletText", "Printout"],
    Hyphenation->True,
    ParagraphIndent->-20,
    FontSize->11],
  
  Cell[StyleData["ItemizedBulletText", "EnhancedPrintout"],
    Hyphenation->True,
    ParagraphIndent->-19,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedBulletText", "EnhancedPrintoutNonGray"],
    Hyphenation->True,
    ParagraphIndent->-19,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NoteBox"],
    CellFrame->0.5,
    CellMargins->{{27, 12}, {8, 8}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxMargins->{{1, 1}, {1.5, 1.5}}},
    GridBoxOptions->{ColumnSpacings->1}],
  
  Cell[StyleData["NoteBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NoteBox", "Printout"],
    CellMargins->{{2, 4}, {6, 2}},
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["NoteBox", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["NoteBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 4}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedNoteBox"],
    CellFrame->0.5,
    CellMargins->{{41, 4}, {8, 2}},
    CellHorizontalScrolling->True,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    FrameBoxOptions->{BoxMargins->{{1, 1}, {1.5, 1.5}}},
    GridBoxOptions->{ColumnSpacings->1}],
  
  Cell[StyleData["ItemizedNoteBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedNoteBox", "Printout"],
    CellMargins->{{39, 4}, {6, 2}},
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedNoteBox", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedNoteBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 4}, {6, 2}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTopBox"],
    CellFrame->{{0.5, 0.5}, {0, 0.5}},
    CellMargins->{{37, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    CellFrameMargins->{{8, 8}, {0, 8}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["ItemizedTopBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedTopBox", "Printout"],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedTopBox", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {0, 6}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedTopBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 4}, {0, 6}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedBottomBox"],
    CellFrame->{{0.5, 0.5}, {0.5, 0}},
    CellMargins->{{37, 4}, {4, 0}},
    CellHorizontalScrolling->True,
    CellFrameMargins->{{8, 8}, {8, 0}},
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    LineIndent->0,
    StyleMenuListing->None,
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["ItemizedBottomBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedBottomBox", "Printout"],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedBottomBox", "EnhancedPrintout"],
    CellMargins->{{39, 4}, {6, 0}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["ItemizedBottomBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{39, 4}, {6, 0}},
    CellFrameColor->GrayLevel[0],
    FontSize->10,
    Background->GrayLevel[1]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MiddleSpacedBox"],
    CellFrame->{{0.5, 0.5}, {0, 0}},
    CellMargins->{{27, 12}, {0, -7}},
    CellHorizontalScrolling->True,
    PageBreakAbove->False,
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.74902, 0.694118, 0.552941],
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    FontWeight->"Plain",
    Background->RGBColor[0.964706, 0.929412, 0.839216],
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->1,
    ColumnWidths->{0.31, 0.62},
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["MiddleSpacedBox", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["MiddleSpacedBox", "Printout"],
    CellMargins->{{2, 0}, {0, 0}},
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}],
  
  Cell[StyleData["MiddleSpacedBox", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 0}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}],
  
  Cell[StyleData["MiddleSpacedBox", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 0}},
    CellFrameColor->GrayLevel[0],
    FontFamily->"Palatino",
    Background->GrayLevel[1],
    GridBoxOptions->{RowMinHeight->1.8}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedNote"],
    CellMargins->{{41, 4}, {0, 12}},
    LineSpacing->{1, 2},
    CounterIncrements->"ItemizedNote",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedNote", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedNote", "Printout"],
    CellMargins->{{38, 4}, {0, 9}},
    Hyphenation->True,
    FontSize->11],
  
  Cell[StyleData["ItemizedNote", "EnhancedPrintout"],
    CellMargins->{{38, 4}, {0, 0}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedNote", "EnhancedPrintoutNonGray"],
    CellMargins->{{38, 4}, {0, 0}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedNoteBelow"],
    CellMargins->{{41, 4}, {0, 8}},
    LineSpacing->{1, 2},
    CounterIncrements->"ItemizedNote",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedNoteBelow", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedNoteBelow", "Printout"],
    CellMargins->{{38, 4}, {0, 12}},
    Hyphenation->True,
    FontSize->11],
  
  Cell[StyleData["ItemizedNoteBelow", "EnhancedPrintout"],
    CellMargins->{{38, 4}, {0, 8}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedNoteBelow", "EnhancedPrintoutNonGray"],
    CellMargins->{{38, 4}, {0, 8}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTextBelowPic"],
    CellMargins->{{19, 4}, {0, 8}},
    LineSpacing->{1, 3},
    ParagraphIndent->-22,
    CounterIncrements->"ItemizedText",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedTextBelowPic", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedTextBelowPic", "Printout"],
    CellMargins->{{20, 4}, {2, 12}},
    Hyphenation->True,
    ParagraphIndent->-18,
    FontSize->11],
  
  Cell[StyleData["ItemizedTextBelowPic", "EnhancedPrintout"],
    CellMargins->{{20, 4}, {2, 12}},
    Hyphenation->True,
    ParagraphIndent->-17,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedTextBelowPic", "EnhancedPrintoutNonGray"],
    CellMargins->{{20, 4}, {2, 12}},
    Hyphenation->True,
    ParagraphIndent->-17,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemizedTextNote2"],
    CellMargins->{{41, 4}, {0, 4}},
    LineSpacing->{1, 2},
    CounterIncrements->"ItemizedNote",
    StyleMenuListing->None],
  
  Cell[StyleData["ItemizedTextNote2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemizedTextNote2", "Printout"],
    CellMargins->{{38, 4}, {0, 2}},
    Hyphenation->True,
    FontSize->11],
  
  Cell[StyleData["ItemizedTextNote2", "EnhancedPrintout"],
    CellMargins->{{38, 4}, {0, 2}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["ItemizedTextNote2", "EnhancedPrintoutNonGray"],
    CellMargins->{{38, 4}, {0, 2}},
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedItem1"],
    CellMargins->{{48, 4}, {4, 4}},
    CellGroupingRules->{"SectionGrouping", 15},
    ParagraphIndent->-23,
    CounterIncrements->"NumberedItem1",
    StyleMenuListing->None],
  
  Cell[StyleData["NumberedItem1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NumberedItem1", "Printout"],
    CellMargins->{{12, 0}, {0, 8}}],
  
  Cell[StyleData["NumberedItem1", "EnhancedPrintout"],
    CellMargins->{{12, 0}, {0, 8}}],
  
  Cell[StyleData["NumberedItem1", "EnhancedPrintoutNonGray"],
    CellMargins->{{12, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedItem2"],
    CellMargins->{{71, 4}, {3, 3}},
    CellGroupingRules->{"SectionGrouping", 15},
    ParagraphIndent->-21,
    CounterIncrements->"NumberedItem2",
    StyleMenuListing->None],
  
  Cell[StyleData["NumberedItem2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NumberedItem2", "Printout"],
    CellMargins->{{27, 0}, {0, 4}}],
  
  Cell[StyleData["NumberedItem2", "EnhancedPrintout"],
    CellMargins->{{27, 0}, {0, 4}}],
  
  Cell[StyleData["NumberedItem2", "EnhancedPrintoutNonGray"],
    CellMargins->{{27, 0}, {0, 4}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedItem3"],
    CellMargins->{{94, 4}, {2, 2}},
    CellGroupingRules->{"SectionGrouping", 15},
    ParagraphIndent->-23,
    CounterIncrements->"NumberedItem3",
    StyleMenuListing->None],
  
  Cell[StyleData["NumberedItem3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["NumberedItem3", "Printout"],
    CellMargins->{{27, 0}, {0, 4}}],
  
  Cell[StyleData["NumberedItem3", "EnhancedPrintout"],
    CellMargins->{{27, 0}, {0, 4}}],
  
  Cell[StyleData["NumberedItem3", "EnhancedPrintoutNonGray"],
    CellMargins->{{27, 0}, {0, 4}},
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PageHeaderBold"],
    CellMargins->{{0, 0}, {2, 0}},
    StyleMenuListing->None,
    FontSize->8,
    FontWeight->"Bold",
    FontSlant->"Italic"],
  
  Cell[StyleData["PageHeaderBold", "Printout"]],
  
  Cell[StyleData["PageHeaderBold", "EnhancedPrintout"],
    FontFamily->"Frutiger"],
  
  Cell[StyleData["PageHeaderBold", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SingleRowTable"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    AutoIndent->False,
    AutoSpacing->False,
    LineIndent->0,
    StyleMenuListing->None,
    GridBoxOptions->{RowSpacings->1.5,
    ColumnSpacings->2,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["SingleRowTable", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["SingleRowTable", "Printout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["SingleRowTable", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}}],
  
  Cell[StyleData["SingleRowTable", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {0, 8}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsubsection"],
    CellMargins->{{27, 4}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    StyleMenuListing->None,
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Subsubsubsection", "Printout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontSize->10],
  
  Cell[StyleData["Subsubsubsection", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["Subsubsubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {3, 8}},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    CellMargins->{{27, Inherited}, {20, 15}},
    CellGroupingRules->{"TitleGrouping", 20},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    FontSize->21],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    CellMargins->{{2, 10}, {8, 10}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {8, 10}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {8, 10}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{27, Inherited}, {20, 15}},
    CellGroupingRules->{"TitleGrouping", 10},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->24],
  
  Cell[StyleData["Subtitle", "Presentation"],
    FontSize->36],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{2, 10}, {10, 15}},
    FontSize->24],
  
  Cell[StyleData["Subtitle", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {10, 15}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->24],
  
  Cell[StyleData["Subtitle", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {10, 15}},
    FontFamily->"Frutiger 55 Roman",
    FontSize->24]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TextBelow"],
    CellMargins->{{27, 4}, {0, 14}},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    ParagraphSpacing->{0, 8},
    StyleMenuListing->None,
    FontSize->12],
  
  Cell[StyleData["TextBelow", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TextBelow", "Printout"],
    CellMargins->{{2, 2}, {6, 12}},
    TextJustification->1,
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["TextBelow", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 12}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["TextBelow", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {6, 12}},
    TextJustification->1,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    CellMargins->{{27, Inherited}, {20, 40}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    DefaultNewInlineCellStyle->"None",
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    LanguageCategory->"NaturalLanguage",
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->36,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["Title", "Presentation"],
    FontSize->54],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{2, 10}, {15, 30}},
    FontSize->30],
  
  Cell[StyleData["Title", "EnhancedPrintout"],
    CellMargins->{{2, 10}, {15, 30}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->30.5,
    FontWeight->"Plain"],
  
  Cell[StyleData["Title", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 10}, {15, 30}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->30.5,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Undersubsection"],
    CellMargins->{{10, 4}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm]},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["Undersubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Undersubsection", "Printout"],
    CellMargins->{{2, 0}, {7, 10}},
    FontSize->10],
  
  Cell[StyleData["Undersubsection", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {7, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain"],
  
  Cell[StyleData["Undersubsection", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {7, 8}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SectionFirst"],
    CellMargins->{{27, 27}, {12, 18}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CounterIncrements->"Section",
    FontFamily->"Helvetica",
    FontSize->20,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["SectionFirst", "Presentation"],
    FontSize->30],
  
  Cell[StyleData["SectionFirst", "Printout"],
    CellMargins->{{2, 0}, {0, 6}},
    PageBreakBelow->False,
    FontSize->18],
  
  Cell[StyleData["SectionFirst", "EnhancedPrintout"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{2, Inherited}, {5, 6}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->8,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->16,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["SectionFirst", "EnhancedPrintoutNonGray"],
    CellFrame->{{0, 0}, {0.5, 0}},
    CellMargins->{{2, Inherited}, {5, 6}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->8,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Frutiger 65 Bold",
    FontSize->16,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsectionFirst"],
    CellDingbat->None,
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["SubsectionFirst", "Presentation"],
    FontSize->24],
  
  Cell[StyleData["SubsectionFirst", "SlideShow"],
    CellDingbat->None,
    CellMargins->{{70, 0}, {4, 6}},
    FontSize->12],
  
  Cell[StyleData["SubsectionFirst", "Printout"],
    CellMargins->{{9, 0}, {5, 6}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontSize->12],
  
  Cell[StyleData["SubsectionFirst", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {5, 6}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontFamily->"Frutiger 65 Bold",
    FontSize->12,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["SubsectionFirst", "EnhancedPrintoutNonGray"],
    CellMargins->{{9, 0}, {5, 6}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontFamily->"Frutiger 65 Bold",
    FontSize->12,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsubsectionFirst"],
    CellMargins->{{27, 27}, {0, 18}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["SubsubsectionFirst", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["SubsubsectionFirst", "Printout"],
    CellMargins->{{2, 0}, {3, 3}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontSize->10],
  
  Cell[StyleData["SubsubsectionFirst", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {3, 3}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["SubsubsectionFirst", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {3, 3}},
    CellBracketOptions->{"Color"->RGBColor[1, 0, 0]},
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SubsubsubsectionFirst"],
    CellMargins->{{10, 4}, {0, 12}},
    CellElementSpacings->{"ClosedGroupTopMargin"->18},
    StyleMenuListing->None,
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["SubsubsubsectionFirst", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["SubsubsubsectionFirst", "Printout"],
    CellMargins->{{2, 0}, {3, 3}},
    FontSize->10],
  
  Cell[StyleData["SubsubsubsectionFirst", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {3, 3}},
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["SubsubsubsectionFirst", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 0}, {3, 3}},
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell["Spacer Cell", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Spacer"],
    CellMargins->{{10, 10}, {0, 0}},
    CellElementSpacings->{"CellMinHeight"->2,
    "ClosedCellHeight"->1},
    CellOpen->False,
    StyleMenuListing->None],
  
  Cell[StyleData["Spacer", "Printout"],
    CellOpen->True,
    FontSize->3]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DiscretionaryHyphen"],
    StyleMenuListing->None,
    FontSize->0.0625,
    FontColor->GrayLevel[1]],
  
  Cell[StyleData["DiscretionaryHyphen", "Printout"],
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["DiscretionaryHyphen", "EnhancedPrintout"],
    FontSize->10,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["DiscretionaryHyphen", "EnhancedPrintoutNonGray"],
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller1"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller1", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.63],
  
  Cell[StyleData["PictureSmaller1", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.63],
  
  Cell[StyleData["PictureSmaller1", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.63]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller2"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller2", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.61],
  
  Cell[StyleData["PictureSmaller2", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.61],
  
  Cell[StyleData["PictureSmaller2", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.61]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller3"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller3", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.59],
  
  Cell[StyleData["PictureSmaller3", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.59],
  
  Cell[StyleData["PictureSmaller3", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.59]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller4"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller4", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller4", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.57],
  
  Cell[StyleData["PictureSmaller4", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.57],
  
  Cell[StyleData["PictureSmaller4", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.57]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller5"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller5", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller5", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.55],
  
  Cell[StyleData["PictureSmaller5", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.55],
  
  Cell[StyleData["PictureSmaller5", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.55]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureSmaller6"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureSmaller6", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureSmaller6", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.53],
  
  Cell[StyleData["PictureSmaller6", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.53],
  
  Cell[StyleData["PictureSmaller6", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.53]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPicture"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPicture", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPicture", "Printout"],
    CellMargins->{{55, Inherited}, {8, -4}},
    Magnification->0.65],
  
  Cell[StyleData["ItemPicture", "EnhancedPrintout"],
    CellMargins->{{60, Inherited}, {10, -3}},
    Magnification->0.65],
  
  Cell[StyleData["ItemPicture", "EnhancedPrintoutNonGray"],
    CellMargins->{{60, Inherited}, {10, -3}},
    Magnification->0.65]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller1"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller1", "Printout"],
    CellMargins->{{56, Inherited}, {9, -3}},
    Magnification->0.63],
  
  Cell[StyleData["ItemPictureSmaller1", "EnhancedPrintout"],
    CellMargins->{{63, Inherited}, {10, -3}},
    Magnification->0.63],
  
  Cell[StyleData["ItemPictureSmaller1", "EnhancedPrintoutNonGray"],
    CellMargins->{{63, Inherited}, {10, -3}},
    Magnification->0.63]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller2"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller2", "Printout"],
    CellMargins->{{58, Inherited}, {8, -4}},
    Magnification->0.61],
  
  Cell[StyleData["ItemPictureSmaller2", "EnhancedPrintout"],
    CellMargins->{{65, Inherited}, {11, -3}},
    Magnification->0.61],
  
  Cell[StyleData["ItemPictureSmaller2", "EnhancedPrintoutNonGray"],
    CellMargins->{{65, Inherited}, {11, -3}},
    Magnification->0.61]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller3"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller3", "Printout"],
    CellMargins->{{61, Inherited}, {10, -2}},
    Magnification->0.59],
  
  Cell[StyleData["ItemPictureSmaller3", "EnhancedPrintout"],
    CellMargins->{{67, Inherited}, {11, -3}},
    Magnification->0.59],
  
  Cell[StyleData["ItemPictureSmaller3", "EnhancedPrintoutNonGray"],
    CellMargins->{{67, Inherited}, {11, -3}},
    Magnification->0.59]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller4"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller4", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller4", "Printout"],
    CellMargins->{{63, Inherited}, {10, -2}},
    Magnification->0.57],
  
  Cell[StyleData["ItemPictureSmaller4", "EnhancedPrintout"],
    CellMargins->{{69, Inherited}, {12, -4}},
    Magnification->0.57],
  
  Cell[StyleData["ItemPictureSmaller4", "EnhancedPrintoutNonGray"],
    CellMargins->{{69, Inherited}, {12, -4}},
    Magnification->0.57]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller5"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller5", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller5", "Printout"],
    CellMargins->{{65, Inherited}, {11, -2}},
    Magnification->0.55],
  
  Cell[StyleData["ItemPictureSmaller5", "EnhancedPrintout"],
    CellMargins->{{71, Inherited}, {13, -2}},
    Magnification->0.55],
  
  Cell[StyleData["ItemPictureSmaller5", "EnhancedPrintoutNonGray"],
    CellMargins->{{71, Inherited}, {13, -2}},
    Magnification->0.55]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureSmaller6"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureSmaller6", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureSmaller6", "Printout"],
    CellMargins->{{67, Inherited}, {12, -2}},
    Magnification->0.53],
  
  Cell[StyleData["ItemPictureSmaller6", "EnhancedPrintout"],
    CellMargins->{{74, Inherited}, {13, -3}},
    Magnification->0.53],
  
  Cell[StyleData["ItemPictureSmaller6", "EnhancedPrintoutNonGray"],
    CellMargins->{{74, Inherited}, {13, -3}},
    Magnification->0.53]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger1"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger1", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.68],
  
  Cell[StyleData["PictureBigger1", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.68],
  
  Cell[StyleData["PictureBigger1", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.68]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger2"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger2", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.71],
  
  Cell[StyleData["PictureBigger2", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.71],
  
  Cell[StyleData["PictureBigger2", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.71]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger3"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger3", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.74],
  
  Cell[StyleData["PictureBigger3", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.74],
  
  Cell[StyleData["PictureBigger3", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.74]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger4"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger4", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger4", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.77],
  
  Cell[StyleData["PictureBigger4", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.77],
  
  Cell[StyleData["PictureBigger4", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.77]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger5"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger5", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger5", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.8],
  
  Cell[StyleData["PictureBigger5", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.8],
  
  Cell[StyleData["PictureBigger5", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PictureBigger6"],
    CellMargins->{{10, Inherited}, {4, 4}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PictureBigger6", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PictureBigger6", "Printout"],
    CellMargins->{{2, Inherited}, {4, 4}},
    Magnification->0.83],
  
  Cell[StyleData["PictureBigger6", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.83],
  
  Cell[StyleData["PictureBigger6", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {4, 2}},
    Magnification->0.83]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger1"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger1", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger1", "Printout"],
    CellMargins->{{50, Inherited}, {9, -3}},
    Magnification->0.68],
  
  Cell[StyleData["ItemPictureBigger1", "EnhancedPrintout"],
    CellMargins->{{57, Inherited}, {10, -3}},
    Magnification->0.68],
  
  Cell[StyleData["ItemPictureBigger1", "EnhancedPrintoutNonGray"],
    CellMargins->{{57, Inherited}, {10, -3}},
    Magnification->0.68]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger2"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger2", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger2", "Printout"],
    CellMargins->{{48, Inherited}, {8, -4}},
    Magnification->0.71],
  
  Cell[StyleData["ItemPictureBigger2", "EnhancedPrintout"],
    CellMargins->{{54, Inherited}, {10, -3}},
    Magnification->0.71],
  
  Cell[StyleData["ItemPictureBigger2", "EnhancedPrintoutNonGray"],
    CellMargins->{{54, Inherited}, {10, -3}},
    Magnification->0.71]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger3"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger3", "Printout"],
    CellMargins->{{46, Inherited}, {10, -2}},
    Magnification->0.74],
  
  Cell[StyleData["ItemPictureBigger3", "EnhancedPrintout"],
    CellMargins->{{52, Inherited}, {10, -3}},
    Magnification->0.74],
  
  Cell[StyleData["ItemPictureBigger3", "EnhancedPrintoutNonGray"],
    CellMargins->{{52, Inherited}, {10, -3}},
    Magnification->0.74]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger4"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger4", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger4", "Printout"],
    CellMargins->{{45, Inherited}, {10, -2}},
    Magnification->0.77],
  
  Cell[StyleData["ItemPictureBigger4", "EnhancedPrintout"],
    CellMargins->{{50, Inherited}, {10, -3}},
    Magnification->0.77],
  
  Cell[StyleData["ItemPictureBigger4", "EnhancedPrintoutNonGray"],
    CellMargins->{{50, Inherited}, {10, -3}},
    Magnification->0.77]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger5"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger5", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger5", "Printout"],
    CellMargins->{{43, Inherited}, {11, -2}},
    Magnification->0.8],
  
  Cell[StyleData["ItemPictureBigger5", "EnhancedPrintout"],
    CellMargins->{{49, Inherited}, {10, -3}},
    Magnification->0.8],
  
  Cell[StyleData["ItemPictureBigger5", "EnhancedPrintoutNonGray"],
    CellMargins->{{49, Inherited}, {10, -3}},
    Magnification->0.8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ItemPictureBigger6"],
    CellMargins->{{41, Inherited}, {4, 2}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ItemPictureBigger6", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ItemPictureBigger6", "Printout"],
    CellMargins->{{43, Inherited}, {12, -2}},
    Magnification->0.83],
  
  Cell[StyleData["ItemPictureBigger6", "EnhancedPrintout"],
    CellMargins->{{46, Inherited}, {8, -2}},
    Magnification->0.83],
  
  Cell[StyleData["ItemPictureBigger6", "EnhancedPrintoutNonGray"],
    CellMargins->{{46, Inherited}, {8, -2}},
    Magnification->0.83]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ScreenGraphic"],
    CellMargins->{{12, Inherited}, {0, 8}},
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["ScreenGraphic", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ScreenGraphic", "Printout"],
    CellMargins->{{2, Inherited}, {0, 8}}],
  
  Cell[StyleData["ScreenGraphic", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->2,
    "ClosedCellHeight"->2},
    CellOpen->False],
  
  Cell[StyleData["ScreenGraphic", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->2,
    "ClosedCellHeight"->2},
    CellOpen->False]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PrintGraphic"],
    CellMargins->{{12, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->2,
    "ClosedCellHeight"->2},
    CellOpen->False,
    CellHorizontalScrolling->True,
    StyleMenuListing->None],
  
  Cell[StyleData["PrintGraphic", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["PrintGraphic", "Printout"],
    CellMargins->{{2, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->2,
    "ClosedCellHeight"->2},
    CellOpen->False],
  
  Cell[StyleData["PrintGraphic", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->12},
    CellOpen->True],
  
  Cell[StyleData["PrintGraphic", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, Inherited}, {0, 8}},
    CellElementSpacings->{"CellMinHeight"->12},
    CellOpen->True]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallTextSans"],
    CellMargins->{{10, 4}, {8, 8}},
    TextJustification->1,
    LineSpacing->{1, 2},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallTextSans", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["SmallTextSans", "Printout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    FontSize->7],
  
  Cell[StyleData["SmallTextSans", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {0, 8}},
    Hyphenation->True,
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, -2},
    FontFamily->"Frutiger 55 Roman",
    FontSize->7]
  }, Closed]]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["AuthorTools", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["BilateralCell"],
    CellMargins->{{27, 4}, {0, 8}},
    StyleMenuListing->None,
    FontSize->12,
    GridBoxOptions->{ColumnWidths->{0.33, 0.6},
    RowAlignments->Top,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["BilateralCell", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["BilateralCell", "SlideShow"],
    CellMargins->{{36, 24}, {4, 4}},
    Background->GrayLevel[0.900008]],
  
  Cell[StyleData["BilateralCell", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontSize->10],
  
  Cell[StyleData["BilateralCell", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Palatino",
    FontSize->10],
  
  Cell[StyleData["BilateralCell", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {6, 6}},
    TextJustification->1,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RuledPageBreak"],
    CellFrame->{{0, 0}, {0.25, 0}},
    CellMargins->{{27, 4}, {12, 0}},
    CellFrameMargins->{{0, 0}, {1, 0}},
    FontSize->8],
  
  Cell[StyleData["RuledPageBreak", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["RuledPageBreak", "SlideShow"],
    CellMargins->{{36, 24}, {4, 4}}],
  
  Cell[StyleData["RuledPageBreak", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    PageBreakBelow->True,
    FontSize->1],
  
  Cell[StyleData["RuledPageBreak", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {6, 6}},
    PageBreakBelow->True],
  
  Cell[StyleData["RuledPageBreak", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {6, 6}},
    PageBreakBelow->True]
  }, Closed]],
  
  Cell[StyleData["PaletteTitle"],
    CellFrame->{{0, 0}, {1, 0}},
    CellMargins->{{4, 0}, {Inherited, 0}},
    StyleMenuListing->None,
    ButtonStyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->14,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0, 0.32549, 0.537255],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #2]}]&),
    ButtonEvaluator->None,
    ButtonStyle->"PaletteButton",
    ButtonMargins->3}],
  
  Cell[StyleData["PaletteButton"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    ButtonBoxOptions->{ButtonEvaluator->"Local",
    Active->True,
    ButtonMargins->6,
    Background->RGBColor[0.329412, 0.584314, 0.694118]}],
  
  Cell[StyleData["PaletteButton2"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    ButtonBoxOptions->{ButtonEvaluator->"Local",
    Active->True,
    ButtonMargins->4,
    Background->RGBColor[0.537255, 0.72549, 0.843137]}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Outline Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Outline1"],
    CellMargins->{{27, 10}, {7, 7}},
    CellGroupingRules->{"SectionGrouping", 50},
    ParagraphIndent->-38,
    CounterIncrements->"Outline1",
    CounterAssignments->{{"Outline2", 0}, {"Outline3", 0}, {"Outline4", 0}},
    FontSize->18,
    FontWeight->"Bold",
    CounterBoxOptions->{CounterFunction:>CapitalRomanNumeral}],
  
  Cell[StyleData["Outline1", "Presentation"],
    FontSize->27],
  
  Cell[StyleData["Outline1", "Printout"],
    CounterBoxOptions->{CounterFunction:>CapitalRomanNumeral}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Outline2"],
    CellMargins->{{57, 10}, {6, 6}},
    CellGroupingRules->{"SectionGrouping", 60},
    ParagraphIndent->-27,
    CounterIncrements->"Outline2",
    CounterAssignments->{{"Outline3", 0}, {"Outline4", 0}},
    FontSize->15,
    FontWeight->"Bold",
    CounterBoxOptions->{CounterFunction:>(Part[ 
        CharacterRange[ "A", "Z"], #]&)}],
  
  Cell[StyleData["Outline2", "Presentation"],
    FontSize->23],
  
  Cell[StyleData["Outline2", "Printout"],
    CounterBoxOptions->{CounterFunction:>(Part[ 
        CharacterRange[ "A", "Z"], #]&)}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Outline3"],
    CellMargins->{{87, 10}, {5, 5}},
    CellGroupingRules->{"SectionGrouping", 70},
    ParagraphIndent->-21,
    CounterIncrements->"Outline3",
    CounterAssignments->{{"Outline4", 0}},
    FontSize->12,
    CounterBoxOptions->{CounterFunction:>Identity}],
  
  Cell[StyleData["Outline3", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Outline3", "Printout"],
    CounterBoxOptions->{CounterFunction:>Identity}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Outline4"],
    CellMargins->{{117, 10}, {4, 4}},
    CellGroupingRules->{"SectionGrouping", 80},
    ParagraphIndent->-18,
    CounterIncrements->"Outline4",
    FontSize->10,
    CounterBoxOptions->{CounterFunction:>(Part[ 
        CharacterRange[ "a", "z"], #]&)}],
  
  Cell[StyleData["Outline4", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["Outline4", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Slide Show Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SlideShowNavigationBar"],
    Editable->False,
    CellFrame->True,
    CellMargins->{{0, 0}, {3, 3}},
    CellElementSpacings->{"CellMinHeight"->0.8125},
    CellGroupingRules->{"SectionGrouping", 30},
    CellFrameMargins->False,
    CellFrameColor->GrayLevel[1],
    CellFrameLabelMargins->False,
    TextAlignment->Center,
    CounterIncrements->"SlideShowNavigationBar",
    StyleMenuListing->None,
    FontSize->10,
    Background->GrayLevel[0.8],
    Magnification->1,
    GridBoxOptions->{GridBaseline->Center,
    RowSpacings->0,
    ColumnSpacings->0,
    ColumnWidths->{3.5, 3.5, 3.5, 3.5, 13, 5, 4},
    RowAlignments->Baseline,
    ColumnAlignments->{
      Center, Center, Center, Center, Center, Center, Right, Center}}],
  
  Cell[StyleData["SlideShowNavigationBar", "SlideShow"],
    Deletable->False,
    ShowCellBracket->False,
    CellMargins->{{-1, -1}, {-1, -1}},
    PageBreakAbove->True,
    CellFrameMargins->{{1, 1}, {0, 0}}],
  
  Cell[StyleData["SlideShowNavigationBar", "Printout"],
    CellMargins->{{18, 4}, {4, 4}},
    LineSpacing->{1, 3},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SlideShowSection"],
    CellFrame->{{0, 0}, {0, 0.5}},
    CellMargins->{{0, 0}, {10, 0}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CellFrameMargins->{{12, 4}, {6, 12}},
    InputAutoReplacements->{"TeX"->StyleBox[ 
            RowBox[ {"T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "LaTeX"->StyleBox[ 
            RowBox[ {"L", 
              StyleBox[ 
                AdjustmentBox[ 
                "A", BoxMargins -> {{-0.35999999999999999, \
-0.10000000000000001}, {0, 0}}, BoxBaselineShift -> -0.20000000000000001], 
                FontSize -> Smaller], "T", 
              AdjustmentBox[ 
              "E", BoxMargins -> {{-0.074999999999999997, \
-0.085000000000000006}, {0, 0}}, BoxBaselineShift -> 0.5], "X"}]], 
        "mma"->"Mathematica", "Mma"->"Mathematica", "MMA"->"Mathematica", 
        "gridMathematica"->FormBox[ 
            RowBox[ {"grid", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.175, 0}, {0, 0}}]}], TextForm], 
        "webMathematica"->FormBox[ 
            RowBox[ {"web", 
              AdjustmentBox[ 
                StyleBox[ "Mathematica", FontSlant -> "Italic"], 
                BoxMargins -> {{-0.17499999999999999, 0}, {0, 0}}]}], 
            TextForm], Inherited},
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->18,
    FontWeight->"Plain",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.408011, 0.440726, 0.8]],
  
  Cell[StyleData["SlideShowSection", "Presentation"],
    FontSize->27],
  
  Cell[StyleData["SlideShowSection", "SlideShow"],
    ShowCellBracket->False,
    PageBreakAbove->True],
  
  Cell[StyleData["SlideShowSection", "Printout"],
    CellMargins->{{18, 30}, {0, 30}},
    CellFrameMargins->5,
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SlideHyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontSize->26,
    FontColor->GrayLevel[0.400015],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonMinHeight->0.85,
    ButtonMargins->0.5,
    ButtonNote->None}],
  
  Cell[StyleData["SlideHyperlink", "SlideShow"]],
  
  Cell[StyleData["SlideHyperlink", "Printout"],
    FontSize->10,
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SlideTOCLink"],
    CellMargins->{{24, Inherited}, {Inherited, Inherited}},
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Helvetica",
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonMargins->1.5,
    ButtonNote->ButtonData}],
  
  Cell[StyleData["SlideTOCLink", "SlideShow"]],
  
  Cell[StyleData["SlideTOCLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SlideTOC"],
    CellDingbat->"\[Bullet]",
    CellMargins->{{18, Inherited}, {Inherited, Inherited}},
    StyleMenuListing->None,
    FontFamily->"Helvetica"],
  
  Cell[StyleData["SlideTOC", "SlideShow"],
    FontSize->14],
  
  Cell[StyleData["SlideTOC", "Printout"],
    FontSize->10,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Title Page, Contents, and Index", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ContentsTitle"],
    CellMargins->{{21, 4}, {12, 24}},
    FontFamily->"Helvetica",
    FontSize->24,
    FontWeight->"Bold",
    FontColor->RGBColor[0.592157, 0.313726, 0.313726]],
  
  Cell[StyleData["ContentsTitle", "Presentation"],
    FontSize->36],
  
  Cell[StyleData["ContentsTitle", "Printout"],
    CellMargins->{{12, 0}, {102, 89}},
    PageBreakBelow->False,
    FontSize->30.5,
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["ContentsTitle", "EnhancedPrintout"],
    CellMargins->{{2, Inherited}, {15, 30}},
    PageBreakBelow->False,
    FontFamily->"Frutiger 65 Bold",
    FontSize->30.5,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ContentsSection"],
    ShowGroupOpenCloseIcon->True,
    CellMargins->{{35, 4}, {3, 18}},
    CellGroupingRules->{"SectionGrouping", 70},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["ContentsSection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["ContentsSection", "Printout"],
    CellMargins->{{20, 0}, {3, 18}},
    PageBreakBelow->False,
    FontSize->11],
  
  Cell[StyleData["ContentsSection", "EnhancedPrintout"],
    CellMargins->{{20, 0}, {3, 18}},
    PageBreakBelow->False,
    FontFamily->"Frutiger 65 Bold",
    FontSize->11,
    FontWeight->"Plain"],
  
  Cell[StyleData["ContentsSection", "EnhancedPrintoutNonGray"],
    CellMargins->{{20, 0}, {3, 18}},
    PageBreakBelow->False,
    FontFamily->"Frutiger 65 Bold",
    FontSize->11,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IndexSection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{20, 4}, {5, 20}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["IndexSection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["IndexSection", "Printout"],
    CellMargins->{{12, 0}, {5, 20}},
    PageBreakBelow->False,
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Contents"],
    ShowGroupOpenCloseIcon->False,
    CellMargins->{{50, 4}, {0, 5}},
    CellGroupingRules->{"SectionGrouping", 90},
    StyleMenuListing->None],
  
  Cell[StyleData["Contents", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Contents", "Printout"],
    CellMargins->{{40, 4}, {0, 5}},
    FontSize->11],
  
  Cell[StyleData["Contents", "EnhancedPrintout"],
    CellMargins->{{40, 5}, {0, 5}},
    FontFamily->"Palatino",
    FontSize->11],
  
  Cell[StyleData["Contents", "EnhancedPrintoutNonGray"],
    CellMargins->{{40, 5}, {0, 5}},
    FontFamily->"Palatino",
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Index"],
    CellMargins->{{21, 4}, {0, 0}},
    ParagraphIndent->-48,
    StyleMenuListing->None],
  
  Cell[StyleData["Index", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["Index", "Printout"],
    CellMargins->{{13, 0}, {0, 0}},
    FontSize->9],
  
  Cell[StyleData["Index", "EnhancedPrintout"],
    CellMargins->{{13, 0}, {0, 0}},
    FontFamily->"Palatino",
    FontSize->9],
  
  Cell[StyleData["Index", "EnhancedPrintoutNonGray"],
    CellMargins->{{13, 0}, {0, 0}},
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["IndexSubentry"],
    CellMargins->{{36, 4}, {0, 0}},
    ParagraphIndent->-48],
  
  Cell[StyleData["IndexSubentry", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["IndexSubentry", "Printout"],
    CellMargins->{{24, 0}, {0, 0}},
    FontSize->9],
  
  Cell[StyleData["IndexSubentry", "EnhancedPrintout"],
    CellMargins->{{24, 0}, {0, 0}},
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Index2Column"],
    CellMargins->{{21, 4}, {0, 8}},
    ParagraphIndent->-20,
    TabSpacings->1,
    StyleMenuListing->None,
    FontSize->10,
    GridBoxOptions->{ColumnWidths->0.5,
    ColumnAlignments->{Left}}],
  
  Cell[StyleData["Index2Column", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["Index2Column", "Printout"],
    CellMargins->{{2, 2}, {6, 6}},
    FontSize->8],
  
  Cell[StyleData["Index2Column", "EnhancedPrintout"],
    CellMargins->{{2, 2}, {1, 1}},
    LineSpacing->{1, 0.2},
    FontFamily->"Palatino",
    FontSize->8],
  
  Cell[StyleData["Index2Column", "EnhancedPrintoutNonGray"],
    CellMargins->{{2, 2}, {1, 1}},
    LineSpacing->{1, 0.2},
    FontFamily->"Palatino",
    FontSize->8]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Leader"],
    TabFilling->".",
    StyleMenuListing->None,
    FontFamily->"Times",
    FontSize->12,
    FontWeight->"Plain"],
  
  Cell[StyleData["Leader", "Printout"],
    FontSize->10],
  
  Cell[StyleData["Leader", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["LeaderNoDots"],
    TabFilling->" ",
    StyleMenuListing->None,
    FontFamily->"Times",
    FontSize->12,
    FontWeight->"Plain"],
  
  Cell[StyleData["LeaderNoDots", "Printout"],
    FontSize->10],
  
  Cell[StyleData["LeaderNoDots", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PageLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0.269993, 0.308507, 0.6],
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["PageLink", "Printout"],
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["PageLink", "EnhancedPrintout"],
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCPage"],
    CellSize->{22, Inherited},
    TextAlignment->Right,
    StyleMenuListing->None,
    FontSize->12,
    FontWeight->"Plain"],
  
  Cell[StyleData["TOCPage", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCPage", "Printout"],
    FontSize->10],
  
  Cell[StyleData["TOCPage", "EnhancedPrintout"],
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCTitle"],
    CellMargins->{{20, 0}, {1, 8}},
    CellGroupingRules->{"SectionGrouping", 70},
    TabSpacings->46,
    CounterIncrements->"TOCTitle",
    CounterAssignments->{{"TOCSection", 0}},
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["TOCTitle", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCTitle", "Printout"],
    CellMargins->{{12, 0}, {1, 8}},
    PageBreakBelow->False,
    TabSpacings->47.4,
    FontSize->11],
  
  Cell[StyleData["TOCTitle", "EnhancedPrintout"],
    CellMargins->{{2, 0}, {2, 10}},
    PageBreakBelow->False,
    TabSpacings->49.3,
    FontFamily->"Frutiger 65 Bold",
    FontSize->10,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSection"],
    CellMargins->{{48, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 75},
    TabSpacings->42.9,
    CounterIncrements->"TOCSection",
    CounterAssignments->{{"TOCSubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->12],
  
  Cell[StyleData["TOCSection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCSection", "Printout"],
    CellMargins->{{30, 0}, {1, 1}},
    PageBreakBelow->False,
    FontSize->12],
  
  Cell[StyleData["TOCSection", "EnhancedPrintout"],
    CellMargins->{{26, 0}, {1, 3}},
    PageBreakBelow->False,
    TabSpacings->46.45,
    FontFamily->"Frutiger",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSubsection"],
    CellMargins->{{76, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 80},
    TabSpacings->39.8,
    CounterIncrements->"TOCSubsection",
    CounterAssignments->{{"TOCSubsubsection", 0}},
    FontSize->12],
  
  Cell[StyleData["TOCSubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCSubsection", "Printout"],
    CellMargins->{{70, 0}, {1, 1}},
    PageBreakBelow->False,
    FontSize->10],
  
  Cell[StyleData["TOCSubsection", "EnhancedPrintout"],
    CellMargins->{{38, 0}, {1, 1}},
    PageBreakBelow->False,
    TabSpacings->45.15,
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSubsubsection"],
    CellMargins->{{104, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 85},
    TabSpacings->36.7,
    CounterIncrements->"TOCSubsubsection",
    FontSize->12],
  
  Cell[StyleData["TOCSubsubsection", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCSubsubsection", "Printout"],
    CellMargins->{{96, 0}, {1, 1}},
    PageBreakBelow->False,
    FontSize->9],
  
  Cell[StyleData["TOCSubsubsection", "EnhancedPrintout"],
    CellMargins->{{96, 0}, {1, 1}},
    PageBreakBelow->False,
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCTitleDingbat"],
    CellSize->{34, Inherited},
    TextAlignment->Right],
  
  Cell[StyleData["TOCTitleDingbat", "EnhancedPrintout"],
    FontFamily->"Frutiger",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSectionDingbat"],
    CellSize->{38, Inherited},
    TextAlignment->Right],
  
  Cell[StyleData["TOCSectionDingbat", "EnhancedPrintout"],
    FontFamily->"Frutiger",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSubsectionDingbat"],
    CellSize->{54, Inherited},
    TextAlignment->Right],
  
  Cell[StyleData["TOCSubsectionDingbat", "EnhancedPrintout"],
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCTitleCondensed"],
    CellMargins->{{20, 0}, {1, 8}},
    CellGroupingRules->{"SectionGrouping", 70},
    TabSpacings->50.4,
    CounterIncrements->"TOCTitleCondensed",
    CounterAssignments->{{"TOCSectionCondensed", 0}, {
        "TOCSubsectionCondensed", 0}},
    FontFamily->"Helvetica",
    FontSize->12,
    FontWeight->"Bold"],
  
  Cell[StyleData["TOCTitleCondensed", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCTitleCondensed", "Printout"],
    CellMargins->{{12, 0}, {1, 8}},
    PageBreakBelow->False,
    FontSize->11],
  
  Cell[StyleData["TOCTitleCondensed", "EnhancedPrintout"],
    CellMargins->{{12, 0}, {1, 8}},
    PageBreakBelow->False,
    FontFamily->"Frutiger 65 Bold",
    FontSize->11,
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSectionCondensed"],
    CellMargins->{{48, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 75},
    TabSpacings->47.4,
    CounterIncrements->"TOCSectionCondensed",
    CounterAssignments->{{"TOCSubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->10,
    FontWeight->"Bold",
    FontColor->GrayLevel[0.2]],
  
  Cell[StyleData["TOCSectionCondensed", "Presentation"],
    FontSize->15],
  
  Cell[StyleData["TOCSectionCondensed", "Printout"],
    CellMargins->{{36, 0}, {1, 1}},
    PageBreakBelow->False,
    FontSize->11],
  
  Cell[StyleData["TOCSectionCondensed", "EnhancedPrintout"],
    CellMargins->{{36, 0}, {1, 1}},
    PageBreakBelow->False,
    FontFamily->"Palatino",
    FontSize->11]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSubsectionCondensed"],
    PageWidth->470,
    CellMargins->{{76, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 80},
    TextJustification->1,
    FontSize->12],
  
  Cell[StyleData["TOCSubsectionCondensed", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCSubsectionCondensed", "Printout"],
    CellMargins->{{72, 0}, {1, 1}},
    PageBreakBelow->False,
    Hyphenation->True,
    FontSize->10],
  
  Cell[StyleData["TOCSubsectionCondensed", "EnhancedPrintout"],
    CellMargins->{{72, 0}, {1, 1}},
    PageBreakBelow->False,
    Hyphenation->True,
    FontFamily->"Palatino",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TOCSubsubsectionCondensed"],
    CellMargins->{{104, 0}, {1, 1}},
    CellGroupingRules->{"SectionGrouping", 85},
    TabSpacings->39.3,
    CounterIncrements->"TOCSubsubsectionCondensed",
    FontSize->12],
  
  Cell[StyleData["TOCSubsubsectionCondensed", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TOCSubsubsectionCondensed", "Printout"],
    CellMargins->{{96, 0}, {1, 1}},
    PageBreakBelow->False,
    FontSize->9],
  
  Cell[StyleData["TOCSubsubsectionCondensed", "EnhancedPrintout"],
    CellMargins->{{96, 0}, {1, 1}},
    PageBreakBelow->False,
    FontFamily->"Palatino",
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TitlePageGraphic"],
    CellMargins->{{0, 0}, {100, 20}},
    Evaluatable->False,
    PageBreakBelow->False,
    TextAlignment->Center],
  
  Cell[StyleData["TitlePageGraphic", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TitlePageGraphic", "Printout"],
    CellMargins->{{0, 0}, {230, 88}}],
  
  Cell[StyleData["TitlePageGraphic", "EnhancedPrintout"],
    CellMargins->{{0, 0}, {230, 88}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["TitlePageLogo"],
    CellMargins->{{0, 0}, {0, 0}}],
  
  Cell[StyleData["TitlePageLogo", "Presentation"],
    FontSize->18],
  
  Cell[StyleData["TitlePageLogo", "Printout"]],
  
  Cell[StyleData["TitlePageLogo", "EnhancedPrintout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Horizontal Lines", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Line1"],
    CellFrame->{{0, 0}, {0, 0.5}},
    CellMargins->{{10, 4}, {-16, 9}},
    StyleMenuListing->None,
    FontSize->2],
  
  Cell[StyleData["Line1", "Presentation"],
    FontSize->3],
  
  Cell[StyleData["Line1", "Printout"],
    CellMargins->{{2, 0}, {-16, 9}}]
  }, Closed]],
  
  Cell["\<\

The item above is horizontal 0.5 pt line\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ProgramTop"],
    CellFrame->{{0.5, 0.5}, {0, 0.5}},
    CellMargins->{{27, 4}, {-18, 18}},
    LineSpacing->{1, -4},
    StyleMenuListing->None,
    FontSize->4],
  
  Cell[StyleData["ProgramTop", "Presentation"],
    FontSize->6],
  
  Cell[StyleData["ProgramTop", "Printout"],
    CellMargins->{{2, 0}, {-18, 18}}]
  }, Closed]],
  
  Cell["\<\

The item above is a ProgramTop line
\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ProgramBottom"],
    CellFrame->{{0.5, 0.5}, {0.5, 0}},
    CellMargins->{{27, 4}, {-10, -2}},
    LineSpacing->{1, -4},
    StyleMenuListing->None,
    FontSize->4],
  
  Cell[StyleData["ProgramBottom", "Presentation"],
    FontSize->6],
  
  Cell[StyleData["ProgramBottom", "Printout"],
    CellMargins->{{2, 0}, {-10, -2}}]
  }, Closed]],
  
  Cell["\<\

The item above is a ProgramBottom line
\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SessionLabel"],
    CellFrame->{{0.5, 0}, {0, 0.5}},
    CellMargins->{{27, 4}, {-4, 18}},
    LineSpacing->{1, 0},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SessionLabel", "Presentation"],
    FontSize->14],
  
  Cell[StyleData["SessionLabel", "Printout"],
    CellMargins->{{2, 0}, {-4, 18}},
    FontSize->7]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Page Headers and Footers", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PageHeader"],
    CellMargins->{{0, 0}, {2, 0}},
    StyleMenuListing->None,
    FontSize->8,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageHeader", "Printout"]],
  
  Cell[StyleData["PageHeader", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman"],
  
  Cell[StyleData["PageHeader", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {2, 0}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->8,
    FontWeight->"Bold"],
  
  Cell[StyleData["PageNumber", "Printout"]],
  
  Cell[StyleData["PageNumber", "EnhancedPrintout"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"],
  
  Cell[StyleData["PageNumber", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 65 Bold",
    FontWeight->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["PageFooter"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->6],
  
  Cell[StyleData["PageFooter", "Printout"]],
  
  Cell[StyleData["PageFooter", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman"],
  
  Cell[StyleData["PageFooter", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["For compatibility", "Subsection"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {2, 0}},
    StyleMenuListing->None,
    FontSize->8,
    FontSlant->"Italic"],
  
  Cell[StyleData["Header", "Printout"]],
  
  Cell[StyleData["Header", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman",
    FontSlant->"Plain"],
  
  Cell[StyleData["Header", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman",
    FontSlant->"Plain"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Footer"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->6],
  
  Cell[StyleData["Footer", "Printout"]],
  
  Cell[StyleData["Footer", "EnhancedPrintout"],
    FontFamily->"Frutiger 55 Roman"],
  
  Cell[StyleData["Footer", "EnhancedPrintoutNonGray"],
    FontFamily->"Frutiger 55 Roman"]
  }, Closed]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["FormatType Styles", "Section"],
  
  Cell["\<\
The cells below define styles that are mixed in with the styles \
of most cells.  If a cell's FormatType matches the name of one of the styles \
defined below, then that style is applied between the cell's style and its \
own options. This is particularly true of Input and Output.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellExpression"],
    PageWidth->Infinity,
    CellMargins->{{6, Inherited}, {Inherited, Inherited}},
    ShowCellLabel->False,
    ShowSpecialCharacters->False,
    AllowInlineCells->False,
    Hyphenation->False,
    AutoItalicWords->{},
    StyleMenuListing->None,
    FontFamily->"Courier",
    FontSize->12,
    Background->GrayLevel[1]],
  
  Cell[StyleData["CellExpression", "EnhancedPrintout"],
    FontFamily->"Courier Std",
    FontSize->10],
  
  Cell[StyleData["CellExpression", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InputForm"],
    InputAutoReplacements->{},
    AllowInlineCells->False,
    Hyphenation->False,
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["InputForm", "EnhancedPrintout"],
    FontFamily->"Courier Std"],
  
  Cell[StyleData["InputForm", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OutputForm"],
    PageWidth->Infinity,
    TextAlignment->Left,
    LineSpacing->{0.6, 1},
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["OutputForm", "EnhancedPrintout"],
    FontFamily->"Courier Std"],
  
  Cell[StyleData["OutputForm", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["StandardForm"],
    InputAutoReplacements->{
      "->"->"\[Rule]", ":>"->"\[RuleDelayed]", "<="->"\[LessEqual]", 
        ">="->"\[GreaterEqual]", "!="->"\[NotEqual]", "=="->"\[Equal]", 
        Inherited},
    "TwoByteSyntaxCharacterAutoReplacement"->True,
    LineSpacing->{1.25, 0},
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["StandardForm", "EnhancedPrintout"],
    FontFamily->"Courier Std"],
  
  Cell[StyleData["StandardForm", "EnhancedPrintoutNonGray"],
    FontFamily->"Courier Std"]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Automatic Styles", "Section"],
  
  Cell["\<\
The cells below define styles that are used to affect the display \
of certain types of objects in typeset expressions.  For example, \
\"UnmatchedBracket\" style defines how unmatched bracket, curly bracket, and \
parenthesis characters are displayed (typically by coloring them to make them \
stand out).\
\>", "Text"],
  
  Cell[StyleData["UnmatchedBracket"],
    StyleMenuListing->None]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "LocalInteractions:2"->{
    Cell[1776, 53, 63, 1, 38, "ObjectName",
      CellTags->"LocalInteractions:2"],
    Cell[1842, 56, 516, 22, 54, "Usage",
      CellTags->"LocalInteractions:2"],
    Cell[2361, 80, 259, 6, 40, "Notes",
      CellTags->"LocalInteractions:2"],
    Cell[2623, 88, 317, 9, 40, "Notes",
      CellTags->"LocalInteractions:2"],
    Cell[2965, 101, 68, 1, 38, "ExampleSection",
      CellTags->"LocalInteractions:2"],
    Cell[3036, 104, 187, 4, 42, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[3226, 110, 132, 4, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[3383, 118, 117, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[3503, 123, 110707, 2187, 294, 13426, 980, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"LocalInteractions:2"],
    Cell[114225, 2313, 81, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[114309, 2316, 226, 5, 42, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[114538, 2323, 89, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[114630, 2326, 188, 5, 58, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[114843, 2335, 117, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[114963, 2340, 110707, 2187, 294, 13426, 980, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"LocalInteractions:2"],
    Cell[225685, 4530, 270, 10, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[225980, 4544, 242, 5, 58, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[226225, 4551, 72745, 1538, 294, 9120, 747, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"LocalInteractions:2"],
    Cell[298985, 6092, 107, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[299095, 6097, 83, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[299181, 6100, 131, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[299315, 6105, 158, 4, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[299476, 6111, 325, 7, 74, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[299804, 6120, 74, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[299881, 6123, 119, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[300025, 6130, 110, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[300138, 6135, 96, 3, 25, "Output",
      CellTags->"LocalInteractions:2"],
    Cell[300249, 6141, 182, 4, 40, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[300456, 6149, 106, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[300565, 6154, 3427, 67, 121, "Output",
      CellTags->"LocalInteractions:2"],
    Cell[304029, 6226, 116, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[304148, 6231, 96, 3, 25, "Output",
      CellTags->"LocalInteractions:2"],
    Cell[304259, 6237, 107, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[304369, 6242, 90, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[304484, 6247, 106, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[304593, 6252, 3269, 64, 121, "Output",
      CellTags->"LocalInteractions:2"],
    Cell[307877, 6319, 72, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[307952, 6322, 131, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[308086, 6327, 74, 1, 24, "Text",
      CellTags->"LocalInteractions:2"],
    Cell[308185, 6332, 106, 3, 26, "Input",
      CellTags->"LocalInteractions:2"],
    Cell[308294, 6337, 96, 3, 25, "Output",
      CellTags->"LocalInteractions:2"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"LocalInteractions:2", 560715, 15145}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 63, 1, 38, "ObjectName",
  CellTags->"LocalInteractions:2"],
Cell[1842, 56, 516, 22, 54, "Usage",
  CellTags->"LocalInteractions:2"],
Cell[2361, 80, 259, 6, 40, "Notes",
  CellTags->"LocalInteractions:2"],
Cell[2623, 88, 317, 9, 40, "Notes",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[2965, 101, 68, 1, 38, "ExampleSection",
  CellTags->"LocalInteractions:2"],
Cell[3036, 104, 187, 4, 42, "Input",
  CellTags->"LocalInteractions:2"],
Cell[3226, 110, 132, 4, 24, "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[3383, 118, 117, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[3503, 123, 110707, 2187, 294, 13426, 980, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[114225, 2313, 81, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[114309, 2316, 226, 5, 42, "Input",
  CellTags->"LocalInteractions:2"],
Cell[114538, 2323, 89, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[114630, 2326, 188, 5, 58, "Input",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[114843, 2335, 117, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[114963, 2340, 110707, 2187, 294, 13426, 980, "GraphicsData", \
"PostScript", "Graphics",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[225685, 4530, 270, 10, 24, "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[225980, 4544, 242, 5, 58, "Input",
  CellTags->"LocalInteractions:2"],
Cell[226225, 4551, 72745, 1538, 294, 9120, 747, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[298985, 6092, 107, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[299095, 6097, 83, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[299181, 6100, 131, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[299315, 6105, 158, 4, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[299476, 6111, 325, 7, 74, "Input",
  CellTags->"LocalInteractions:2"],
Cell[299804, 6120, 74, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[299881, 6123, 119, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[300025, 6130, 110, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[300138, 6135, 96, 3, 25, "Output",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[300249, 6141, 182, 4, 40, "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[300456, 6149, 106, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[300565, 6154, 3427, 67, 121, "Output",
  CellTags->"LocalInteractions:2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[304029, 6226, 116, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[304148, 6231, 96, 3, 25, "Output",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[304259, 6237, 107, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[304369, 6242, 90, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[304484, 6247, 106, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[304593, 6252, 3269, 64, 121, "Output",
  CellTags->"LocalInteractions:2"]
}, Open  ]],
Cell[307877, 6319, 72, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],
Cell[307952, 6322, 131, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[308086, 6327, 74, 1, 24, "Text",
  CellTags->"LocalInteractions:2"],

Cell[CellGroupData[{
Cell[308185, 6332, 106, 3, 26, "Input",
  CellTags->"LocalInteractions:2"],
Cell[308294, 6337, 96, 3, 25, "Output",
  CellTags->"LocalInteractions:2"]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

