(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    112042,       1706]*)
(*NotebookOutlinePosition[    115244,       1787]*)
(*  CellTagsIndexPosition[    115132,       1780]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["BoundedDegreeNetwork", "ObjectName",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[TextData[{
  "\[FilledSmallSquare] BoundedDegreeNetwork[",
  StyleBox["size, ",
    FontSlant->"Italic"],
  Cell[BoxData[
      \(TraditionalForm\`k\_min, \ k\_max\)]],
  "] creates, if possible, a graph where every vertex has degree between ",
  Cell[BoxData[
      \(TraditionalForm\`k\_min\ and\ \(\(k\_max\)\(.\)\)\)]]
}], "Usage",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[TextData[{
  "\[FilledSmallSquare] The degree sequence for the graph is randomly \
generated using a uniform distribution over the integers between ",
  Cell[BoxData[
      \(TraditionalForm\`\([k\_min, \ k\_max]\)\)]],
  ".  Sometimes no possible graph can be found for the specified constraints, \
in which case the function returns ",
  StyleBox["Null",
    FontFamily->"Courier"],
  " rather than a ",
  StyleBox["Graph",
    FontFamily->"Courier"],
  " object."
}], "Notes",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{

Cell["Examples", "ExampleSection",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(BoundedDegreeNetwork[23, 2, 5] // ShowGraph;\)\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0424116 0.911989 0.0455028 0.910056 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.93747 .62291 m
.34499 .92926 L
s
.93747 .62291 m
.0456 .43797 L
s
.88794 .7367 m
.46672 .9545 L
s
.88794 .7367 m
.2346 .87218 L
s
.88794 .7367 m
.9544 .5 L
s
.80948 .83294 m
.70791 .90448 L
s
.80948 .83294 m
.46672 .9545 L
s
.80948 .83294 m
.14374 .7875 L
s
.80948 .83294 m
.0456 .56203 L
s
.80948 .83294 m
.0456 .43797 L
s
.70791 .90448 m
.34499 .07074 L
s
.59076 .94603 m
.34499 .92926 L
s
.59076 .94603 m
.88794 .2633 L
s
.46672 .9545 m
.88794 .2633 L
s
.46672 .9545 m
.9544 .5 L
s
.34499 .92926 m
.2346 .87218 L
s
.34499 .92926 m
.07914 .3185 L
s
.34499 .92926 m
.93747 .37709 L
s
.2346 .87218 m
.0456 .56203 L
s
.2346 .87218 m
.34499 .07074 L
s
.2346 .87218 m
.59076 .05397 L
s
.14374 .7875 m
.80948 .16706 L
s
.07914 .6815 m
.34499 .07074 L
s
.07914 .6815 m
.80948 .16706 L
s
.07914 .6815 m
.9544 .5 L
s
.0456 .56203 m
.0456 .43797 L
s
.0456 .56203 m
.88794 .2633 L
s
.0456 .43797 m
.34499 .07074 L
s
.0456 .43797 m
.70791 .09552 L
s
.07914 .3185 m
.34499 .07074 L
s
.07914 .3185 m
.9544 .5 L
s
.14374 .2125 m
.2346 .12782 L
s
.14374 .2125 m
.80948 .16706 L
s
.2346 .12782 m
.46672 .0455 L
s
.2346 .12782 m
.59076 .05397 L
s
.2346 .12782 m
.70791 .09552 L
s
.46672 .0455 m
.70791 .09552 L
s
.46672 .0455 m
.88794 .2633 L
s
.46672 .0455 m
.9544 .5 L
s
.59076 .05397 m
.80948 .16706 L
s
.59076 .05397 m
.93747 .37709 L
s
.70791 .09552 m
.88794 .2633 L
s
.80948 .16706 m
.93747 .37709 L
s
.025 w
.93747 .62291 Mdot
.88794 .7367 Mdot
.80948 .83294 Mdot
.70791 .90448 Mdot
.59076 .94603 Mdot
.46672 .9545 Mdot
.34499 .92926 Mdot
.2346 .87218 Mdot
.14374 .7875 Mdot
.07914 .6815 Mdot
.0456 .56203 Mdot
.0456 .43797 Mdot
.07914 .3185 Mdot
.14374 .2125 Mdot
.2346 .12782 Mdot
.34499 .07074 Mdot
.46672 .0455 Mdot
.59076 .05397 Mdot
.70791 .09552 Mdot
.80948 .16706 Mdot
.88794 .2633 Mdot
.93747 .37709 Mdot
.9544 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[10]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00Q?ooool3o`0009Wooooo0023ooooo`Go0000V?ooool008;ooooo1ol0000Oooooo`?o0000
MOooool008;ooooo1ol0000Nooooo`Go0000M?ooool0087ooooo3Ol0000Hooooo`Oo0000Loooool0
07kooooo0ol00002ooooo`So00000oooool5o`0001?ooooo1ol0001cooooo`00Noooool3o`0000Ko
oooo0ol00002ooooo`04o`000?oooooo0000o`0000Kooooo1Ol0000:ooooo`_o0000Loooool0067o
oooo0ol0000Dooooo`?o00003oooool00ol0003oooooooooo`02o`0000Wooooo2_l00005ooooo`So
0000LOooool0063ooooo1Ol0000@ooooo`?o00004oooool01?l0003oooooooooooooool2o`0000Oo
oooo2_l00006ooooo`?o00000_ooool01?l0003oooooo`000?l0001_ooooo`00Goooool7o`0000co
oooo0ol0000Gooooo`03o`000?oooooooooo00;ooooo1ol0000:ooooo`Ko00001_ooool00ol0003o
ooooooooo`02o`0006gooooo001Oooooo`Oo00002Oooool3o`0001_ooooo2?l0000Booooo`Go0000
0oooool01?l0003oooooooooooooool2o`0006_ooooo001Oooooo`Oo00001oooool2o`0001[ooooo
1_l00006ooooo`;o00004?ooool00ol0003oooooooooo`02ooooo`Go00001?ooool2o`0006Wooooo
001Pooooo`Go00001Oooool3o`0001Sooooo1?l00006ooooo`03o`000?oooooooooo00Gooooo0_l0
000=ooooo`03o`000?oooooooooo00Sooooo1Ol000000ooooooo0000o`00001Wooooo`00Goooool0
0ol0003oooooo`000003o`0000;ooooo0ol0000Fooooo`Go00002oooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02Oooool01?l0003ooooooooooooo
ool5o`0006Gooooo001Nooooo`05o`000?oooooo0000o`000?ooool01?l0000Dooooo`Go00004Ooo
ool00ol0003oooooooooo`06ooooo`;o00002Oooool00ol0003oooooooooo`0;ooooo`03o`000?oo
oooooooo00Gooooo1Ol0000:ooooo`?o0000Doooool005gooooo00Co0000ooooool0003ooooo0ol0
00000ooooooo0000ooooo`0?ooooo`Go00005oooool00ol0003oooooooooo`07ooooo`;o00001ooo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Kooooo0_l000000ooooooo0000o`00
0003o`0000Cooooo1Ol0001Booooo`00G?ooool00ol0003oooooo`000004o`0000;ooooo00?o0000
ooooooooool02Oooool5o`0001gooooo0_l00009ooooo`;o00001?ooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo00Oooooo0_l00004ooooo`[o0000DOooool005_ooooo1?l000001?oo
oooo0000ooooool00003ooooo`03o`000?oooooooooo00?ooooo1Ol0000Tooooo`03o`000?oooooo
oooo00Sooooo0_l00002ooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool02?ooool>
o`00057ooooo001Iooooo`;o00000_ooool01_l0003oooooooooool0003oooooo`0000?ooooo1_l0
000Zooooo`03o`000?oooooooooo00Wooooo0ol0000?oooooa7o00000oooool7o`00057ooooo001F
ooooo`Co00000_ooool01?l0003oooooooooool00002ooooo`Ko0000;oooool00ol0003ooooooooo
o`09oooooa7o00001Oooool00ol0003oooooooooo`09ooooo`?o000000?oooooo`000?l000001?l0
001Aooooo`00Doooool3o`0000;ooooo00Co0000oooooooooooo00000_ooool5o`0000Cooooo00?o
0000ooooooooool0;?ooool?o`000003ooooool0003o000001;ooooo00?o0000ooooooooool01ooo
ool2o`000004ooooool0003o0000ooooo`?o00000_ooool00ol0003oooooooooo`1>ooooo`00D?oo
ool3o`0000Cooooo00?o0000ooooool000001Ol00003ooooo`03o`000?oooooooooo00?ooooo00?o
0000ooooooooool07?ooool?o`0000?ooooo00?o0000ooooooooool01_ooool00ol0003ooooooooo
o`03ooooo`;o00004Oooool00ol0003oooooooooo`04ooooo`;o00001Oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo04gooooo001=ooooo`?o00001?ooool5o`000003ooooool0003o
oooo00;ooooo00Co0000oooooooooooo00001_ooool00ol0003oooooooooo`0=ooooo`oo00004ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Gooooo0_l0000@ooooo`04o`000?oo
ooooooooooooo`;o00002?ooool2o`0000Gooooo00?o0000ooooooooool0C?ooool004;ooooo0ol0
0005ooooo`?o00000_ooool6o`0000Cooooo00Go0000ooooooooooooooooo`000003ooooo`03o`00
0?oooooooooo00Cooooo4?l0000Sooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0
2?ooool2o`0000oooooo0ol0000<ooooo`;o00001?ooool00ol0003oooooooooo`1;ooooo`00@Ooo
ool5o`000003ooooool0003o000000Ko00001?ooool00ol0003oooooooooo`02ooooo`03o`000?oo
oooo000000oo0000<_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00[ooooo0_l0
000<ooooo`?o00003oooool2o`0000?ooooo00?o0000ooooooooool0B_ooool0043ooooo2_l00000
0ooooooo0000o`00000=o`0000?ooooo00Go0000ooooooooooooooooo`000007ooooo`03o`000?oo
oooooooo037ooooo00Go0000ooooooooooooooooo`00000>ooooo`;o00002Oooool01Ol0003ooooo
oooooooooooo0000013ooooo0_l00002ooooo`03o`000?oooooooooo04Wooooo0010ooooo`_o0000
1_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0
0_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo037ooooo00?o0000ooooool00000
4Oooool2o`0000Gooooo0_l00005ooooo`03o`000?oooooooooo00oooooo0_l000000ooooooo0000
ooooo`19ooooo`00@?ooool7o`0000Wooooo00?o0000ooooooooool00_ooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01_ooool00ol0003ooooooooo
o`0booooo`;o00004oooool2o`000003ooooool0003o000000Sooooo00?o0000ooooooooool04?oo
ool3o`0004Wooooo0010ooooo`Ko00002Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooo
oooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`07ooooo`03o`000?oooooo
oooo037ooooo0_l0000Dooooo`?o00002_ooool00ol0003oooooooooo`0Aooooo`;o0000B?ooool0
03oooooo00?o0000ooooooooool00ol00009ooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Sooooo00?o0000oooo
ooooool0<Oooool01?l0003oooooo`000?l0000@ooooo`;o00000oooool2o`0000Wooooo00?o0000
ooooooooool04_ooool2o`0004Kooooo000nooooo`03o`000?oooooooooo00cooooo00?o0000oooo
ooooool00oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`0`ooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool02oooool2o`0000Oooooo0_l00008ooooo`03o`000?oooooooooo01;ooooo0ol00014oooo
o`00?Oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Wooooo00?o0000oooooooo
ool0;oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Sooooo0_l0000;ooooo`;o
00001oooool00ol0003oooooooooo`0Booooo`04o`000?oooooo0000o`0004;ooooo000looooo`03
o`000?oooooooooo00cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo00?ooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0^ooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool01Oooool2o`0000oooooo0_l00006ooooo`;o
00004oooool00ol0003oooooooooo`02o`00043ooooo000kooooo`03o`000?oooooooooo00cooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool02_ooool00ol0003oooooooooo`0]ooooo`03o`000?oooooooooo00Kooooo
00?o0000ooooooooool00_ooool2o`0001?ooooo0_l00006ooooo`03o`000?oooooooooo017ooooo
0_l00002ooooo`;o0000?_ooool003[ooooo00?o0000ooooooooool02oooool2o`0000Kooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00[ooooo00?o
0000ooooooooool0;Oooool00ol0003oooooooooo`07ooooo`04o`000?oooooo0000o`0001Oooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Booooo`04o`000?ooooooooooooooo`;o
0000??ooool003Wooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`05ooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0;ooooo`03o`00
0?oooooooooo02_ooooo00?o0000ooooooooool02?ooool2o`0001[ooooo0_l00005ooooo`03o`00
0?oooooooooo01;ooooo00?o0000ooooooooool00_ooool2o`000004ooooool0003o0000o`0003Ko
oooo000hooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01Oooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03?ooool00ol0003ooooo
ooooo`0[ooooo`03o`000?oooooooooo00Kooooo0_l00002ooooo`03o`000?oooooooooo01Wooooo
0_l00004ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool00oooool5o`0003Gooooo
000gooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`07ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool00ol0003ooooooooo
o`0Yooooo`03o`000?oooooooooo00Gooooo0_l00005ooooo`03o`000?oooooooooo01[ooooo0_l0
0003ooooo`03o`000?oooooooooo01;ooooo2ol0000dooooo`00=Oooool2o`0000gooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool03Oooool00ol0003oooooooooo`0Yooooo`03o`000?oooooooooo00?ooooo0_l00008
ooooo`03o`000?oooooooooo01_ooooo0_l00002ooooo`03o`000?oooooooooo00;ooooo3ol00000
1?oooooo0000ooooooooool7o`0003Cooooo000dooooo`03o`000?oooooooooo00cooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool03Oooool00ol0003oooooooooo`0Xooooo`03o`000?oooooooooo00;ooooo0_l0000;
ooooo`03o`000?oooooooooo01Cooooo3ol0000Aooooo`Wo0000=?ooool003?ooooo00?o0000oooo
ooooool03?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Sooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo02Oooooo00?o0000oooo
ooooool00_l0000>ooooo`03o`000?oooooooooo00Gooooo3_l0000:ooooo`?o00004oooool2o`00
0003ooooool0003o000000?o0000=Oooool003;ooooo00?o0000ooooooooool03?ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`0>ooooo`03o`000?oooooooooo02Oooooo0ol00009ooooo`oo00006_ooool2o`00017o
oooo0_l000000ooooooo0000ooooo`04o`0003Gooooo000aooooo`03o`000?oooooooooo00cooooo
00?o0000ooooooooool01_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Oooooo
00?o0000ooooooooool03oooool00ol0003oooooooooo`0Sooooo`oo00002Oooool2o`00023ooooo
0_l0000=ooooo`?o00000oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo03;ooooo
000`ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool01_ooool00ol0003ooooooooo
o`0:ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`0Dooooo`oo000000?oooooo`000?l000005oooool00ol0003oooooooooo`0Nooooo`?o00002_oo
ool00ol0003oooooo`000005ooooo`05o`000?ooooooooooooooool00000=?ooool002oooooo00?o
0000ooooooooool03?ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Wooooo00?o
0000ooooooooool01oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00Gooooo3_l0
000>ooooo`?o00006Oooool00ol0003oooooooooo`0Nooooo`04o`000?oooooo0000o`0000Kooooo
0_l000000ooooooo0000ooooo`06ooooo`05o`000?ooooooooooooooool00000<oooool002kooooo
00?o0000ooooooooool02oooool2o`0000Wooooo00?o0000ooooooooool02_ooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo00[ooooo3ol0000Jooooo`;o00000_ooool00ol0003ooooo
ooooo`0Hooooo`03o`000?oooooooooo01kooooo0_l000000ooooooo0000o`000003ooooo`04o`00
0?ooooooooooo`0000Wooooo00Go0000ooooooooooooooooo`00000booooo`00;Oooool00ol0003o
ooooooooo`0;ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool02_ooool00ol0003o
ooooooooo`06ooooo`oo00001oooool00ol0003oooooooooo`0Mooooo`;o00000oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool00ol00002ooooo`03
o`000?oooooooooo00Wooooo00Co0000oooooooooooo0000<_ooool002cooooo00?o0000oooooooo
ool02oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Kooooo3_l00002ooooo`03
o`000?oooooooooo01;ooooo00?o0000ooooooooool06_ooool2o`0000Gooooo00?o0000oooooooo
ool06oooool00ol0003oooooooooo`0Oooooo`03o`000?oooooo000000Co00003Oooool01?l0003o
ooooooooool0000aooooo`00:?ooool4o`0000gooooo00?o0000ooooooooool00oooool?o`0000Co
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo01So
oooo0_l00007ooooo`03o`000?oooooooooo01cooooo00?o0000ooooooooool07oooool01?l0003o
ooooooooooooool3o`0000gooooo00?o0000ooooool00000<Oooool002Oooooo1Ol00004ooooo`oo
00001?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Wooooo00?o0000oooooooo
ool04oooool00ol0003oooooooooo`0Eooooo`;o00002?ooool00ol0003oooooooooo`0Nooooo`03
o`000?oooooooooo01cooooo0_l000001?oooooo0000ooooool00003ooooo`;o00003?ooool3o`00
033ooooo000Vooooo`[o00001oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00co
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo01?o
oooo0_l0000:ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`02ooooo`03o`000?oooooooooo00Cooooo0_l0000<ooooo`;o0000;oooool002Kooooo
1ol00009ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool03?ooool00ol0003ooooo
ooooo`09ooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool04?ooool2o`0000_ooooo
00?o0000ooooooooool08Oooool00ol0003oooooooooo`0Gooooo`;o00001?ooool00ol0003ooooo
o`000007ooooo`;o00002oooool00ol0003oooooooooo`0]ooooo`009_ooool7o`0000Sooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00Wooooo00?o
0000ooooooooool05Oooool00ol0003oooooooooo`0>ooooo`;o00003Oooool00ol0003ooooooooo
o`0Rooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool00_ooool2o`0000?ooooo00?o
0000ooooooooool01_ooool2o`0000[ooooo00?o0000ooooooooool0;?ooool002Oooooo1Ol00008
ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0:
ooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool03?ooool2o`0000kooooo00?o0000
ooooooooool09?ooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`07ooooo`;o00002?ooool2o`0002gooooo000Xoooo
o`?o00002?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00kooooo00?o0000oooo
ooooool02_ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo00Wooooo0_l0000@oooo
o`03o`000?oooooooooo02Gooooo00?o0000ooooooooool04?ooool2o`0000Cooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`08ooooo`;o00001oooool2o`0002cooooo000booooo`03
o`000?oooooooooo00[ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0;ooooo`03
o`000?oooooooooo01Kooooo00?o0000ooooooooool01oooool2o`0001;ooooo00?o0000oooooooo
ool09_ooool2o`0000oooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`08ooooo`03
o`000?oooooooooo00Wooooo00?o0000ooooooooool01Oooool2o`0002_ooooo000aooooo`03o`00
0?oooooooooo00[ooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`0;ooooo`03o`00
0?oooooooooo01Oooooo00?o0000ooooooooool01?ooool2o`0001?ooooo00?o0000ooooooooool0
:Oooool00ol0003oooooooooo`0:ooooo`;o00001Oooool00ol0003oooooooooo`0:ooooo`03o`00
0?oooooooooo00Wooooo0_l00005ooooo`03o`000?oooooo000002[ooooo000_ooooo`;o00003?oo
ool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool05ooo
ool00ol0003oooooooooo`02ooooo`;o00005Oooool00ol0003oooooooooo`0Zooooo`03o`000?oo
oooooooo00Sooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`03o`000?oo
oooooooo00[ooooo0_l00004ooooo`03o`000?oooooo000002Wooooo000^ooooo`03o`000?oooooo
oooo00_ooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooo
oooo01Wooooo00Co0000oooooooooooo00005_ooool00ol0003oooooooooo`0/ooooo`03o`000?oo
oooooooo00Gooooo0_l00006ooooo`03o`000?oooooooooo00kooooo0_l0000<ooooo`;o00000_oo
ool01?l0003oooooooooool0000Xooooo`00;Oooool00ol0003oooooooooo`0;ooooo`03o`000?oo
oooooooo013ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0Iooooo`?o00005ooo
ool00ol0003oooooooooo`0]ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00_ooooo0_l000001Ooooooo0000oooo
oooooooo000002Oooooo000/ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool04?oo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo01Sooooo0ol0000Gooooo`03o`000?oo
oooooooo02oooooo00Go0000ooooooooooooooooo`000007ooooo`03o`000?oooooooooo01?ooooo
00?o0000ooooooooool03?ooool3o`0000;ooooo00?o0000ooooooooool09?ooool002_ooooo00?o
0000ooooooooool03?ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo00cooooo00?o
0000ooooooooool05_ooool2o`0000;ooooo00?o0000ooooooooool05Oooool00ol0003ooooooooo
o`0`ooooo`?o00001oooool00ol0003oooooooooo`0Eooooo`03o`000?oooooooooo00gooooo0_l0
0002ooooo`03o`000?oooooooooo02?ooooo000Zooooo`03o`000?oooooooooo00cooooo00?o0000
ooooooooool04?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo01Cooooo0_l00004
ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool0<Oooool2o`0000Oooooo00?o0000
ooooooooool05oooool00ol0003oooooooooo`0=ooooo`?o000000?oooooo`000?ooool00ol0000P
ooooo`00:Oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo017ooooo00?o0000oooo
ooooool03Oooool00ol0003oooooooooo`0Booooo`;o00001oooool00ol0003oooooooooo`0Coooo
o`03o`000?oooooooooo02oooooo0_l00002ooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Oo00007oooool002So
oooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00ko
oooo00?o0000ooooooooool04?ooool2o`0000Wooooo00?o0000ooooooooool04oooool00ol0003o
ooooooooo`0^ooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`00000Moooo
o`03o`000?oooooooooo00cooooo00?o0000ooooooooool01ol0000Nooooo`009oooool00ol0003o
ooooooooo`0<ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool03Oooool00ol0003o
ooooooooo`0?ooooo`;o00003?ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo02ko
oooo00?o0000ooooooooool01Oooool00ol0003oooooo`00000Oooooo`03o`000?oooooooooo00co
oooo00?o0000ooooool000001_l0000Nooooo`009_ooool00ol0003oooooooooo`0=ooooo`03o`00
0?oooooooooo01;ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0=ooooo`;o0000
3_ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo02cooooo0_l00009ooooo`03o`00
0?oooooooooo01oooooo00?o0000ooooooooool02oooool9o`0001kooooo000Uooooo`03o`000?oo
oooooooo00gooooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00_ooooo0_l0000Aooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool0;?oo
ool00ol0003oooooooooo`07ooooo`;o000000?oooooo`000?ooool08?ooool00ol0003ooooooooo
o`08ooooo`Co000000?oooooo`000?l000000ol0000Oooooo`009?ooool00ol0003oooooooooo`0=
ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`09
ooooo`;o00004oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo02[ooooo0_l00009
ooooo`03o`000?oooooooooo00;ooooo0_l0000Pooooo`03o`000?oooooooooo00Cooooo0ol00004
ooooo`Go00008?ooool002;ooooo0_l0000?ooooo`03o`000?oooooooooo01?ooooo00?o0000oooo
ooooool03oooool00ol0003oooooooooo`07ooooo`;o00005Oooool00ol0003oooooooooo`0>oooo
o`03o`000?oooooooooo02[ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo01kooooo00?o0000ooooooooool00ol00007ooooo`;o00008oooool0027o
oooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`0Dooooo`03o`000?oooooooooo00oo
oooo00?o0000ooooooooool01Oooool2o`0001Sooooo00?o0000ooooooooool03Oooool00ol0003o
ooooooooo`0Xooooo`;o00002_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo01ko
oooo0_l0000:ooooo`;o00008oooool0023ooooo00?o0000ooooooooool03_ooool00ol0003ooooo
ooooo`0Dooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool00oooool2o`0001[ooooo
00?o0000ooooooooool03Oooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo00Wooooo
00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Jooooo`?o000000?oooooo`000?l00000
2Oooool2o`0002?ooooo000Oooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool05?oo
ool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00;ooooo0_l0000Mooooo`03o`000?oo
oooooooo00_ooooo00?o0000ooooooooool09oooool00ol0003oooooooooo`09ooooo`03o`000?oo
oooooooo00_ooooo00?o0000ooooooooool05_ooool3o`0000Kooooo00?o0000ooooooooool01Ooo
ool2o`000003ooooool0003ooooo027ooooo000Nooooo`03o`000?oooooooooo00oooooo00?o0000
ooooooooool05?ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo00;o00007oooool0
0ol0003oooooooooo`0;ooooo`03o`000?oooooooooo02Gooooo0_l0000;ooooo`03o`000?oooooo
oooo00gooooo00?o0000ooooooooool04_ooool3o`0000[ooooo00?o0000ooooooooool00oooool0
0ol0003oooooo`000002ooooo`03o`000?oooooooooo01oooooo000Mooooo`03o`000?oooooooooo
00oooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooo0000
02;ooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Uooooo`03o`000?oooooooooo
00[ooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0>ooooo`?o00003_ooool00ol0
003oooooooooo`02ooooo`;o00000oooool00ol0003oooooooooo`0Oooooo`007?ooool00ol0003o
ooooooooo`0?ooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool04?ooool2o`0002?o
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Sooooo`;o00003?ooool00ol0003o
ooooooooo`0Aooooo`03o`000?oooooooooo00_ooooo0_l0000Booooo`06o`000?ooooooooooo`00
0?oooooo00001?ooool00ol0003oooooooooo`0Nooooo`006oooool00ol0003oooooooooo`0?oooo
o`03o`000?oooooooooo01Kooooo00?o0000ooooooooool03oooool3o`0002Gooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`0Sooooo`03o`000?oooooooooo00_ooooo00?o0000oooo
ooooool04oooool00ol0003oooooooooo`07ooooo`?o00005Oooool01?l0003oooooo`000?l00005
ooooo`03o`000?oooooooooo01kooooo000Jooooo`03o`000?oooooooooo013ooooo00?o0000oooo
ooooool05_ooool00ol0003oooooooooo`0=ooooo`;o00000_ooool00ol0003oooooooooo`0Soooo
o`03o`000?oooooooooo00Oooooo00?o0000ooooooooool08_ooool00ol0003oooooooooo`0;oooo
o`03o`000?oooooooooo01Gooooo00?o0000ooooooooool00oooool3o`0001Wooooo00?o0000oooo
ool000001_ooool00ol0003oooooooooo`0Mooooo`005Oooool3o`000003ooooool0003ooooo017o
oooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`0<ooooo`;o00001?ooool00ol0003o
ooooooooo`0Sooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool08?ooool2o`0000go
oooo00?o0000ooooooooool05oooool00ol0003oooooo`000002o`0001_ooooo00Co0000ooooool0
003o00001oooool00ol0003oooooooooo`0Looooo`005?ooool5o`0001;ooooo00?o0000oooooooo
ool05oooool00ol0003oooooooooo`0:ooooo`;o00001Oooool00ol0003oooooooooo`0Uooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooooooool08?ooool00ol0003oooooooooo`0<ooooo`03
o`000?oooooooooo01Oooooo0ol0000Nooooo`04o`000?oooooo0000o`0000Oooooo00?o0000oooo
ooooool07?ooool001?ooooo1ol0000@ooooo`03o`000?oooooooooo01Oooooo00?o0000oooooooo
ool02Oooool2o`0000Oooooo00?o0000ooooooooool09Oooool00ol0003oooooooooo`05ooooo`03
o`000?oooooooooo01kooooo0_l0000>ooooo`03o`000?oooooooooo01Kooooo0_l00003ooooo`03
o`000?oooooooooo01[ooooo00Ko0000oooooooooooo0000ooooool00007ooooo`03o`000?oooooo
oooo01_ooooo000Cooooo`Oo00003oooool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo
00Oooooo0_l00009ooooo`03o`000?oooooooooo02Kooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`0Nooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool05?ooool3o`00
00Kooooo0_l0000Jooooo`03o`000?oooooo000000?ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0Kooooo`004oooool;o`0000[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`05ooooo`;o00002oooool00ol0003oooooooooo`0Vooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool07?ooool2o`0000kooooo0_l0000Dooooo`?o00002oooool00ol0003o
ooooooooo`0Fooooo`04o`000?ooooooooooo`0000Cooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0Jooooo`005?ooool5o`0000Gooooo1Ol00005ooooo`03o`000?oooooooooo01So
oooo00?o0000ooooooooool01?ooool2o`0000gooooo00?o0000ooooooooool09oooool01Ol0003o
oooooooooooooooo000001kooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0Aoooo
o`?o00003oooool00ol0003oooooooooo`0Dooooo`04o`000?ooooooooooo`0000Kooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`0Iooooo`005Oooool3o`0000_ooooo1Ol0000Koooo
o`03o`000?oooooooooo00;ooooo0_l0000?ooooo`03o`000?oooooooooo02Oooooo00Go0000oooo
ooooooooooooo`00000Mooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool03oooool3
o`0001?ooooo00?o0000ooooooooool04oooool01?l0003oooooooooool00007ooooo`03o`000?oo
oooooooo00?ooooo00?o0000ooooooooool06Oooool001Sooooo00?o0000ooooooooool02oooool0
0ol0003oooooo`000004o`0001Gooooo00Co0000oooooooooooooooo0_l0000Aooooo`03o`000?oo
oooooooo02Oooooo00Co0000oooooooooooo00007?ooool2o`0000oooooo00?o0000ooooooooool0
3_ooool2o`0001Oooooo00?o0000ooooooooool04Oooool01Ol0003ooooooooooooooooo000000So
oooo0_l00004ooooo`03o`000?oooooooooo01Sooooo000Hooooo`03o`000?oooooooooo00[ooooo
00?o0000ooooooooool01Oooool5o`00013ooooo00Co0000ooooool0003o00004_ooool00ol0003o
ooooooooo`0Yooooo`03o`000?oooooo000001_ooooo00?o0000ooooooooool03_ooool00ol0003o
ooooooooo`0<ooooo`?o00006_ooool00ol0003oooooooooo`0?ooooo`05o`000?oooooooooooooo
ool000002oooool01Ol0003ooooooooooooooooo000001[ooooo000Hooooo`03o`000?oooooooooo
00Wooooo00?o0000ooooooooool02oooool4o`0000_ooooo0ol0000Dooooo`03o`000?oooooooooo
02Wooooo00?o0000ooooool000006Oooool2o`00013ooooo00?o0000ooooooooool02_ooool3o`00
01kooooo00?o0000ooooooooool03_ooool01Ol0003ooooooooooooooooo000000cooooo00Go0000
ooooooooooooooooo`00000Iooooo`006Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooo
oooo00oooooo1Ol00005ooooo`;o00005_ooool00ol0003oooooooooo`0Zooooo`03o`000?oooooo
oooo01Oooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`08ooooo`?o00008_ooool0
0ol0003oooooooooo`0<ooooo`05o`000?ooooooooooooooool000003_ooool01Ol0003ooooooooo
oooooooo000001Sooooo000Iooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool05Ooo
ool6o`0001Oooooo00?o0000ooooooooool0:_ooool00ol0003oooooooooo`0Fooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool01_ooool3o`0002Kooooo00?o0000ooooooooool02ooo
ool01Ol0003ooooooooooooooooo000000oooooo00Co0000oooooooooooo00006?ooool001[ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`0Gooooo`;o00000_ooool5o`0001?ooooo
00?o0000ooooooooool0:Oooool00ol0003oooooo`00000Eooooo`;o00004Oooool00ol0003ooooo
ooooo`05ooooo`;o0000:_ooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool0
00004Oooool01?l0003oooooooooool0000Gooooo`006_ooool00ol0003oooooooooo`04ooooo`03
o`000?oooooooooo01Oooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02ooooo`Go
00003_ooool00ol0003oooooooooo`0Yooooo`03o`000?oooooo000001Cooooo00?o0000oooooooo
ool04?ooool00ol0003oooooooooo`03ooooo`?o0000;Oooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool04?ooool00ol0003oooooo`00000Gooooo`00
6oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01Kooooo0_l00004ooooo`03o`00
0?oooooooooo00Sooooo1Ol00009ooooo`03o`000?oooooooooo02Sooooo00Co0000oooooooooooo
00004_ooool2o`0001;ooooo00Co0000oooooooooooooooo0ol0000aooooo`03o`000?oooooooooo
00Kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooo0000
01Kooooo000Kooooo`05o`000?ooooooooooooooool000005oooool2o`0000Kooooo00?o0000oooo
ooooool03Oooool4o`0000Gooooo00?o0000ooooooooool0:?ooool01Ol0003ooooooooooooooooo
0000013ooooo00?o0000ooooooooool04Oooool00ol0003oooooo`000002o`0003Gooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01?ooooo00?o0000
ooooool000005Oooool001cooooo00Co0000oooooooooooo00005Oooool2o`0000Oooooo00?o0000
ooooooooool04_ooool5o`0002_ooooo00Go0000ooooooooooooooooo`00000>ooooo`;o00004ooo
ool3o`0003Wooooo0_l00004ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool05?oo
ool2o`0001Gooooo000Looooo`03o`000?oooooo000001Cooooo0_l00009ooooo`03o`000?oooooo
oooo01Kooooo1_l0000Uooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool02_ooool0
0ol0003oooooooooo`0Aooooo`;o0000?_ooool01?l0003oooooooooool00004ooooo`03o`000?oo
oooooooo01Kooooo1?l0000Booooo`007?ooool2o`0001?ooooo0_l0000:ooooo`03o`000?oooooo
oooo01Oooooo00?o0000ooooooooool00oooool5o`00023ooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`09ooooo`03o`000?oooooooooo00oooooo1?l00010ooooo`;o00001Oooool0
0ol0003oooooooooo`0Fooooo`Go00004Oooool001cooooo0_l0000Aooooo`;o00003?ooool00ol0
003oooooooooo`0Gooooo`03o`000?oooooooooo00Sooooo1Ol0000Jooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool01_ooool2o`0000oooooo0ol000000ooooooo0000o`000012oooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool05_ooool7o`00013ooooo000Kooooo`03
o`000?oooooo000000oooooo0_l0000=ooooo`03o`000?oooooooooo01Sooooo00?o0000oooooooo
ool03Oooool4o`0001Kooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03
o`000?oooooooooo00cooooo0ol00003ooooo`03o`000?oooooooooo047ooooo00?o0000ooooool0
00000oooool00ol0003oooooooooo`0Fooooo`Oo00004?ooool001_ooooo00Co0000oooooooooooo
00003?ooool2o`0000oooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`0Aooooo`Go
00004?ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00;ooooo0_l0000<ooooo`?o
00001Oooool00ol0003oooooooooo`12ooooo`06o`000?ooooooooooo`000?oooooo00006Oooool7
o`00013ooooo000Jooooo`05o`000?ooooooooooooooool000002_ooool2o`00017ooooo00?o0000
ooooooooool06?ooool00ol0003oooooooooo`0Fooooo`Go00002oooool00ol0003oooooooooo`07
ooooo`05o`000?ooooooooooooooool000003?ooool2o`0000Oooooo00?o0000ooooooooool0@_oo
ool00ol0003oooooooooo`02ooooo`;o00006_ooool5o`00017ooooo000Iooooo`03o`000?oooooo
oooo00?ooooo00?o0000ooooooooool01Oooool2o`0001;ooooo00?o0000ooooooooool06?ooool0
0ol0003oooooooooo`0Looooo`Go00001_ooool00ol0003oooooooooo`07ooooo`04o`000?oooooo
0000o`0000[ooooo0ol00008ooooo`03o`000?oooooooooo04;ooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`0Gooooo`03o`000?oooooo000000;o00004_ooool001Sooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`03ooooo`;o00005?ooool00ol0003oooooooooo`0H
ooooo`03o`000?oooooooooo027ooooo1_l0000;ooooo`03o`000?oooooooooo00Oooooo0ol0000:
ooooo`03o`000?oooooooooo04?ooooo00?o0000ooooooooool00oooool00ol0003oooooo`00000G
ooooo`03o`000?oooooooooo01Gooooo000Gooooo`03o`000?oooooooooo00Gooooo00Co0000oooo
oooooooooooo0_l0000Eooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool09_ooool5
o`0000Kooooo0_l00006ooooo`?o00003?ooool00ol0003oooooooooo`13ooooo`03o`000?oooooo
oooo00Cooooo00Co0000oooooooooooo00005Oooool00ol0003oooooooooo`0Fooooo`005_ooool0
0ol0003oooooooooo`07ooooo`?o00005oooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo02Gooooo00?o0000ooooooooool00oooool6o`0000;ooooo00?o0000ooooooooool00ol0000>
ooooo`03o`000?oooooooooo04?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo017ooooo00?o0000ooooooooool05oooool001Kooooo00?o0000oooo
ooooool01_ooool2o`0001Sooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Uoooo
o`03o`000?oooooooooo00Kooooo2?l0000@ooooo`03o`000?oooooooooo04Cooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00oooooo00?o0000oooo
ooooool06?ooool001Gooooo00?o0000ooooooooool01Oooool2o`0000;ooooo00?o0000oooooooo
ool05Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo02Cooooo00?o0000oooooooo
ool01Oooool2o`0000?ooooo2?l0000<ooooo`03o`000?oooooooooo04Cooooo00?o0000oooooooo
ool01?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00gooooo00?o0000oooooooo
ool06Oooool001Cooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`02ooooo`03o`00
0?oooooooooo01Gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Tooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool00ol00003ooooo`03o`000?oooooooooo00;ooooo
1Ol00006ooooo`03o`000?oooooooooo04Gooooo00?o0000ooooooooool01?ooool00ol0003ooooo
ooooo`06ooooo`;o00003?ooool00ol0003oooooooooo`0Jooooo`004oooool00ol0003ooooooooo
o`04ooooo`;o00001_ooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool09Oooool00ol0003oooooooooo`03ooooo`Co00001_ooool00ol0003ooooooooo
o`07ooooo`Ko0000Aoooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00?o
0000ooooooooool02?ooool00ol0003oooooooooo`0Kooooo`004_ooool00ol0003oooooooooo`03
ooooo`;o00002?ooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool09?ooool00ol0003oooooooooo`02ooooo`?o00002_ooool00ol0003oooooooooo`0:
ooooo`Ko0000@_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooooooool01_ooool00ol0003oooooooooo`0Looooo`004_ooool01?l0003oooooooooooooool2
o`0000_ooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`0Kooooo`03o`000?oooooo
oooo02Cooooo00?o0000ooooool000000_l0000=ooooo`03o`000?oooooooooo00Wooooo00?o0000
ooooooooool01?ooool4o`0003kooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0;
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool07Oooool0017ooooo00?o0000oooo
ooooool00_l0000=ooooo`03o`000?oooooooooo017ooooo00?o0000ooooooooool06oooool00ol0
003oooooooooo`0Sooooo`Go00003oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo
00Wooooo1Ol0000hooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool01?l0
003oooooooooooooool2o`00023ooooo000<ooooo`?o000000?oooooo`000?ooool00_l0000?oooo
o`03o`000?oooooooooo013ooooo00?o0000ooooooooool07?ooool00ol0003oooooooooo`0Qoooo
o`?o000000?oooooo`000?ooool04?ooool00ol0003oooooooooo`05ooooo`;o00004Oooool5o`00
03?ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooo0000
02;ooooo000;ooooo`Oo00004_ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo01co
oooo00?o0000ooooooooool07_ooool3o`000004ooooool0003o0000o`0001?ooooo00?o0000oooo
ooooool00oooool00ol0003oooooooooo`0Fooooo`Go0000;Oooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo013ooooo00?o0000ooooooooool08Oooool000[ooooo1ol0000Cooooo`03
o`000?oooooooooo00kooooo00?o0000ooooooooool07Oooool00ol0003oooooooooo`0Kooooo`?o
00001?ooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00;ooooo00?o0000oooooooo
ool07?ooool5o`0002Oooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0?ooooo`03
o`000?oooooo000002;ooooo000:ooooo`Oo00005?ooool00ol0003oooooooooo`0=ooooo`03o`00
0?oooooooooo01cooooo00?o0000ooooooooool06Oooool3o`0000Kooooo00?o0000ooooooooool0
5?ooool01Ol0003ooooooooooooooooo000002Cooooo1Ol0000Rooooo`03o`000?oooooooooo00Ko
oooo00?o0000ooooooooool03_ooool01Ol0003ooooooooooooooooo0000027ooooo000:ooooo`_o
00004?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo01cooooo00?o0000oooooooo
ool05oooool2o`0000Oooooo0ol0000Gooooo`03o`000?oooooo000002[ooooo1?l0000Mooooo`03
o`000?oooooooooo00Kooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo01kooooo000;ooooo`Go000000?oooooo`000?l000000_ooool5o`0000cooooo
00?o0000ooooooooool02oooool00ol0003oooooooooo`0Mooooo`03o`000?oooooooooo01Cooooo
0ol00008ooooo`04o`000?ooooooooooo`0001Oooooo0_l0000_ooooo`Go00006?ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`0Mooooo`003?ooool3o`0000Cooooo0_l00005ooooo`Go00001oooool00ol0003ooooo
ooooo`0;ooooo`03o`000?oooooooooo01gooooo00?o0000ooooooooool04Oooool3o`0000Wooooo
0_l00002ooooo`03o`000?oooooooooo01Kooooo0_l0000dooooo`Go00004_ooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01oooool00ol0003ooooo
ooooo`0Looooo`003Oooool00ol0003oooooooooo`05ooooo`;o00002?ooool5o`0000?ooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`0Nooooo`03o`000?oooooooooo00kooooo0ol0
000;ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool05Oooool00ol0003oooooo`00
000iooooo`Go00003?ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00cooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`0Kooooo`003Oooool00ol0003oooooooooo`07
ooooo`;o00002oooool5o`0000[ooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`0;
ooooo`?o00003Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01Gooooo00Go0000
ooooooooooooooooo`00000mooooo`Go00001oooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00cooooo00?o0000ooooooooool02oooool2o`0001_ooooo000=ooooo`03o`000?oooooo
oooo00Wooooo00?o0000ooooooooool02oooool00ol0003oooooo`000003o`0000Gooooo00?o0000
ooooooooool07oooool00ol0003oooooooooo`09ooooo`;o00003_ooool2o`0000Gooooo00?o0000
ooooooooool05?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo043ooooo1?l00002
ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo01Sooooo000=ooooo`03o`000?oooooooooo00[ooooo0_l0000<oooo
o`03o`000?oooooooooo00;ooooo1_l0000Pooooo`03o`000?oooooooooo00Oooooo0ol0000?oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool05?ooool00ol0003oooooooooo`04oooo
o`03o`000?oooooooooo04?ooooo1Ol00006ooooo`03o`000?oooooooooo00_ooooo00?o0000oooo
ooooool04?ooool00ol0003oooooooooo`0Gooooo`003Oooool00ol0003oooooooooo`0<ooooo`;o
00002_ooool00ol0003oooooooooo`06ooooo`Ko00007?ooool00ol0003oooooooooo`04ooooo`?o
00004?ooool2o`0000Oooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`05ooooo`03
o`000?oooooooooo04Cooooo00Co0000oooooooooooooooo1Ol000000ooooooo0000ooooo`0;oooo
o`03o`000?oooooooooo01;ooooo00?o0000ooooooooool05_ooool000gooooo00?o0000oooooooo
ool03_ooool2o`0000Wooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`03ooooo`Go
00005oooool01?l0003oooooooooooooool3o`0001;ooooo00?o0000ooooooooool01oooool00ol0
003oooooooooo`0Booooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool0@oooool00ol0
003oooooooooo`07ooooo`Go00001oooool2o`0001Kooooo00?o0000ooooooooool05Oooool000go
oooo00?o0000ooooooooool04?ooool2o`0000Oooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`08ooooo`Go00004_ooool4o`0001?ooooo0_l00009ooooo`03o`000?oooooooooo01;o
oooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`12ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool00_ooool5o`000003ooooool0003ooooo01Sooooo00?o0000oooooooo
ool05?ooool000gooooo00?o0000ooooooooool04_ooool2o`0000Kooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`0>ooooo`Go00003?ooool2o`0001Gooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`0@ooooo`;o00002oooool00ol0003oooooooooo`11ooooo`03o`00
0?oooooooooo00Sooooo00?o0000ooooooooool01oooool5o`0001Oooooo00?o0000ooooooooool0
4oooool000gooooo00?o0000ooooooooool05?ooool2o`0000Cooooo00?o0000ooooooooool00ooo
ool00ol0003oooooooooo`0Cooooo`Co00001Oooool3o`000003ooooool0003ooooo01?ooooo00?o
0000ooooooooool02Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo00cooooo00?o
0000ooooooooool0?oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool00oooool4o`0001Cooooo00?o0000ooooooooool04_ooool000gooooo00?o0000
ooooooooool05_ooool2o`0000;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0H
ooooo`Go00001?ooool00ol0003oooooooooo`0@ooooo`;o00003?ooool00ol0003oooooooooo`0?
ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`08
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02?ooool5o`00013ooooo00?o0000
ooooooooool04Oooool000gooooo00?o0000ooooooooool06?ooool2o`000003ooooool0003ooooo
00;ooooo00?o0000ooooooooool05oooool3o`0000?ooooo1Ol0000Aooooo`03o`000?oooooooooo
00_ooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo
03gooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo
00kooooo1Ol0000<ooooo`03o`000?oooooooooo013ooooo000=ooooo`03o`000?oooooooooo01[o
oooo0_l00002ooooo`03o`000?oooooooooo01Gooooo0ol00009ooooo`03o`000?oooooo000000Co
00002_ooool2o`0000kooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`0@ooooo`03
o`000?oooooooooo03gooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`05ooooo`03
o`000?oooooooooo01Cooooo1Ol00008ooooo`03o`000?oooooo000000;o00003Oooool000gooooo
00?o0000ooooooooool07?ooool3o`0001Gooooo0_l0000<ooooo`03o`000?oooooooooo00Cooooo
1Ol00004ooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`0Aooooo`03o`000?oooooooooo03cooooo00?o0000ooooooooool02Oooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo01[ooooo1Ol00004ooooo`Go00003?ooool000gooooo00?o
0000ooooooooool07?ooool01?l0003oooooo`000?l0000Aooooo`?o00003_ooool00ol0003ooooo
ooooo`09ooooo`Go00003oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo01?ooooo
00?o0000ooooooooool0>_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool08?ooool:o`0000_ooooo000=ooooo`03o`000?oooooooooo01gooooo00?o
0000ooooooooool00_l0000<ooooo`?o00004Oooool00ol0003oooooooooo`0:ooooo`04o`000?oo
ooooooooooooo`Co00002oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo01Cooooo
00?o0000ooooooooool0>_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool09?ooool7o`0000_ooooo000=ooooo`03o`000?oooooooooo01gooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`06ooooo`?o00005?ooool00ol0003ooooooooo
o`09ooooo`03o`000?oooooooooo00Kooooo1Ol00005ooooo`03o`000?oooooooooo00cooooo00?o
0000ooooooooool05_ooool00ol0003oooooooooo`0hooooo`03o`000?oooooooooo00Wooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`0Qooooo`_o00002oooool000gooooo00?o0000
ooooooooool07?ooool00ol0003oooooo`000004ooooo`;o00000oooool3o`0001Oooooo00?o0000
ooooooooool01oooool2o`0000kooooo1_l0000=ooooo`03o`000?oooooooooo01Oooooo00?o0000
ooooooooool0>?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool07Oooool5o`0000Cooooo1_l0000<ooooo`003Oooool00ol0003oooooooooo`0Loooo
o`03o`000?oooooo000000Kooooo0ol0000Jooooo`03o`000?oooooooooo00Kooooo00?o0000oooo
ooooool04_ooool6o`0000Sooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0foooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Ioooo
o`Go00002?ooool00ol0003oooooooooo`03o`0000gooooo000=ooooo`03o`000?oooooooooo01_o
oooo00Co0000oooooooooooo00001?ooool3o`000003ooooool0003o000001Sooooo00?o0000oooo
ooooool01Oooool2o`0001Gooooo00?o0000ooooooooool00oooool5o`0000;ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0eooooo`03o`000?oooooooooo00[ooooo00?o0000oooo
ooooool00_l0000Gooooo`Go00003?ooool01Ol0003ooooooooooooooooo000000oooooo000=oooo
o`03o`000?oooooooooo01_ooooo00Co0000oooooooooooooooo1?l00006ooooo`;o00005_ooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool02Oooool5
o`0001[ooooo00?o0000ooooooooool0=Oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooo
000001Cooooo1Ol0000@ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool03Oooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`03o`0000_ooooo0_l0000Doooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`09oooo
o`03o`000?oooooooooo00;ooooo1?l0000Gooooo`03o`000?oooooooooo03?ooooo00?o0000oooo
ooooool02_ooool00ol0003oooooo`00000Aooooo`Co00005?ooool00ol0003oooooooooo`02oooo
o`03o`000?oooooooooo00kooooo000=ooooo`03o`000?oooooooooo01[ooooo0ol00003ooooo`03
o`000?oooooooooo00[ooooo0_l0000Booooo`04o`000?ooooooooooooooo`;o00006?ooool00ol0
003oooooooooo`08ooooo`03o`000?oooooooooo00Oooooo1Ol0000Booooo`03o`000?oooooooooo
03;ooooo00?o0000ooooooooool02oooool2o`0000gooooo1Ol0000Gooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool03_ooool000gooooo00?o0000ooooooooool06?ooool3o`0000Go
oooo00?o0000ooooooooool03?ooool2o`00013ooooo00Co0000oooooooooooo00006Oooool00ol0
003oooooooooo`07ooooo`;o00003oooool5o`0000kooooo00?o0000ooooooooool0<Oooool00ol0
003oooooooooo`0:ooooo`;o00002Oooool5o`0001_ooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`0>ooooo`003Oooool00ol0003oooooooooo`0Eooooo`?o000000?oooooo`000?oo
ool01_ooool00ol0003oooooooooo`0=ooooo`;o00003_ooool3o`0001[ooooo00?o0000oooooooo
ool01_ooool00ol0003oooooooooo`0Dooooo`Go00002Oooool00ol0003oooooooooo`0`ooooo`03
o`000?oooooooooo00_ooooo00?o0000ooooooooool00oooool5o`0001oooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`0>ooooo`003Oooool00ol0003oooooooooo`0Booooo`?o0000
1?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00oooooo0_l0000<ooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Jooooo`Go0000
1Oooool00ol0003oooooooooo`0_ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooool00000
1?l0000Sooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03oooool000gooooo00?o
0000ooooooooool03oooool3o`0000Kooooo00?o0000ooooooooool01oooool00ol0003ooooooooo
o`0@ooooo`;o00002?ooool2o`0001cooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`0Pooooo`Co000000?oooooo`000?ooool0;oooool00ol0003oooooooooo`08ooooo`Go00009ooo
ool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00oooooo000=ooooo`03o`000?oooooo
oooo00cooooo0ol00009ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool04_ooool2
o`0000Gooooo00?o0000ooooool000006oooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo02Gooooo1Ol0000/ooooo`03o`000?oooooooooo00Gooooo1?l000001?oooooo0000ooooool0
000Wooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool03oooool000gooooo00?o0000
ooooooooool02_ooool2o`0000_ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0D
ooooo`;o00000_ooool01?l0003oooooooooool0000Kooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool0:?ooool00ol0003oooooooooo`05o`0002Oooooo00?o0000ooooooooool01Ol0
0004ooooo`03o`000?oooooo000002Kooooo0_l0000:ooooo`03o`000?oooooooooo00oooooo000=
ooooo`03o`000?oooooooooo00Oooooo0ol0000=ooooo`03o`000?oooooooooo00Wooooo00?o0000
ooooooooool05Oooool2o`0000?ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo02Wooooo00?o0000ooooooooool01Oooool5o`00023ooooo1Ol00008
ooooo`04o`000?ooooooooooo`0002Gooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`0@ooooo`003?ooool3o`0000Gooooo0ol0000?ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooooooool05?ooool01?l0003oooooo`000?l00002ooooo`03o`000?oooooooooo01Sooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`0[ooooo`03o`000?oooooooooo00Wooooo1Ol0
000Fooooo`Go000000?oooooo`000?ooool02Oooool01?l0003oooooooooool0000Uooooo`03o`00
0?oooooooooo00[ooooo00?o0000ooooooooool04?ooool000_ooooo1Ol000001?oooooo0000o`00
0?l0000Booooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool04Oooool2o`0000Cooooo
0ol0000Jooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0;?ooool00ol0003ooooo
ooooo`0>ooooo`Go00003?ooool5o`0000Gooooo00?o0000ooooooooool02?ooool01Ol0003ooooo
oooooooooooo000002Cooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0@ooooo`00
2_ooool7o`0001Gooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0@ooooo`03o`00
0?oooooooooo00Kooooo0_l0000Hooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0
;_ooool00ol0003oooooooooo`0Booooo`Co00001?ooool4o`0000Wooooo00?o0000ooooooooool0
2?ooool01Ol0003ooooooooooooooooo000002Cooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`0Aooooo`002_ooool7o`0001Cooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`0>ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool00_l0000Fooooo`05o`00
0?ooooooooooooooool00000<Oooool00ol0003oooooooooo`0Eooooo`Ko00003?ooool00ol0003o
ooooooooo`06ooooo`;o00001?ooool00ol0003oooooooooo`0Qooooo`03o`000?oooooooooo00co
oooo00?o0000ooooooooool04Oooool000[ooooo2?l0000Cooooo`03o`000?oooooooooo00gooooo
00?o0000ooooooooool03?ooool2o`0000Wooooo00?o0000ooooooooool00_ooool2o`0001?ooooo
00Go0000ooooooooooooooooo`00000cooooo`03o`000?oooooooooo00oooooo1Ol00006ooooo`Go
00001_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool08Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo017ooooo000;ooooo`Go0000
0_ooool3o`0000oooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`0;ooooo`03o`00
0?oooooooooo00Wooooo00?o0000ooooooooool01?ooool2o`00017ooooo00Co0000oooooooooooo
0000=?ooool00ol0003oooooooooo`0:ooooo`Go00004?ooool5o`000003ooooool0003ooooo00Ko
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Pooooo`03o`000?oooooooooo00ko
oooo00?o0000ooooooooool04Oooool000cooooo1?l00005ooooo`?o00003?ooool00ol0003ooooo
ooooo`0?ooooo`03o`000?oooooooooo00Sooooo0_l0000<ooooo`03o`000?oooooooooo00Kooooo
0_l0000?ooooo`03o`000?oooooo000003Gooooo00?o0000ooooooooool01Oooool5o`0001[ooooo
1Ol00003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool07oooool00ol0003ooooo
ooooo`0>ooooo`03o`000?oooooooooo01;ooooo000?ooooo`03o`000?oooooooooo00Kooooo0ol0
0008ooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool01oooool00ol0003ooooooooo
o`0<ooooo`03o`000?oooooooooo00Sooooo0_l0000<ooooo`03o`000?oooooo000003Oooooo00?o
0000ooooool000001?l0000Nooooo`03o`000?oooooooooo00?ooooo1Ol00005ooooo`03o`000?oo
oooooooo01oooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0Booooo`004?ooool0
0ol0003oooooooooo`08ooooo`;o00001_ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooo
oooo00Cooooo0_l0000?ooooo`03o`000?oooooooooo00[ooooo0_l0000:ooooo`;o0000=_ooool4
o`0002?ooooo00?o0000ooooooooool01?ooool01?l0003oooooooooooooool4o`000003ooooool0
003ooooo01oooooo00?o0000ooooooooool04?ooool00ol0003oooooooooo`0Booooo`004?ooool0
0ol0003oooooooooo`0:ooooo`?o00000oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooo
oooo00?ooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0<ooooo`;o00001oooool0
0ol0003oooooooooo`0aooooo`Go00000oooool00ol0003oooooooooo`0Pooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool01_ooool5o`0001cooooo00?o0000ooooooooool04?ooool0
0ol0003oooooooooo`0Cooooo`004Oooool00ol0003oooooooooo`0<ooooo`?o00005Oooool01Ol0
003ooooooooooooooooo0000017ooooo00?o0000ooooooooool03oooool2o`0000Cooooo0_l0000^
ooooo`Go00002?ooool00ol0003oooooooooo`0Oooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`02ooooo`Go00005_ooool00ol0003oooooooooo`0A
ooooo`03o`000?oooooooooo01?ooooo000Booooo`03o`000?oooooooooo00gooooo1?l0000Boooo
o`04o`000?oooooo0000o`0001;ooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`02
o`0002[ooooo1Ol0000>ooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Sooooo1Ol0000?ooooo`;o00005?ooool0
0ol0003oooooooooo`0Cooooo`004_ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo
00;ooooo0ol0000@ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool04_ooool2o`00
0003ooooool0003ooooo02Cooooo1Ol0000Cooooo`03o`000?oooooooooo01gooooo00?o0000oooo
ooooool00oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00gooooo1Ol00009oooo
o`03o`000?oooooooooo01Cooooo00?o0000ooooooooool04oooool001?ooooo00?o0000oooooooo
ool02oooool00ol0003oooooooooo`05ooooo`;o00003?ooool3o`0001Cooooo00?o0000oooooooo
ool04_ooool01?l0003oooooo`000?l0000Pooooo`Go00006?ooool00ol0003oooooooooo`0Moooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Boooo
o`Go00000oooool00ol0003oooooooooo`0Dooooo`03o`000?oooooooooo01Cooooo000Cooooo`03
o`000?oooooooooo00[ooooo00?o0000ooooooooool02?ooool3o`0000Sooooo00Co0000oooooooo
oooo00005?ooool00ol0003oooooooooo`0Aooooo`05o`000?ooooooooooo`000?ooool00_l0000J
ooooo`Co00007_ooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool02Oooool00ol0003oooooooooo`0Hooooo`Co00005_ooool00ol0003oooooooooo`0D
ooooo`005?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00_ooooo0ol00004oooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`0@oooo
o`05o`000?ooooooooooooooool000000oooool2o`0001?ooooo1Ol0000Rooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool01Ol0000Aooooo`03o`000?oooooooooo01Cooooo000Eoooo
o`03o`000?oooooooooo00Oooooo00?o0000ooooooooool03oooool4o`0000Gooooo00?o0000oooo
ooooool04Oooool00ol0003oooooooooo`0?ooooo`05o`000?ooooooooooooooool000001_ooool2
o`0000cooooo1Ol0000Xooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool00_l0000<
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool01_ooool5o`0000_ooooo00?o0000
ooooooooool05Oooool001Gooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0@oooo
o`03o`000?oooooo000000;o00001?ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo
00kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`06ooooo`;o00001Oooool5o`00
02gooooo00?o0000ooooooooool06?ooool01?l0003oooooooooool0000>ooooo`03o`000?oooooo
oooo01Sooooo00?o0000ooooooooool03?ooool5o`0000Kooooo00?o0000ooooooooool00ol0000B
ooooo`005_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00kooooo0_l00005oooo
o`?o000000?oooooo`000?ooool04?ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo
00;ooooo00?o0000ooooooooool02Oooool5o`0003?ooooo00?o0000ooooooooool05oooool00ol0
003oooooo`00000>ooooo`03o`000?oooooooooo01Sooooo00?o0000ooooooooool04_ooool5o`00
0003ooooool0003ooooo00Go00004Oooool001Kooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`0>ooooo`03o`000?oooooooooo00Sooooo0ol0000@ooooo`03o`000?oooooooooo00go
oooo00?o0000ooooooooool00oooool00ol0003oooooooooo`04ooooo`Go00000_ooool2o`0003Co
oooo00?o0000ooooooooool05_ooool00ol0003oooooo`00000?ooooo`03o`000?oooooooooo01Oo
oooo00?o0000ooooooooool06?ooool9o`00013ooooo000Gooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool03?ooool2o`0000gooooo0ol0000>ooooo`03o`000?oooooooooo00cooooo
00?o0000ooooooooool00oooool01?l0003oooooooooooooool4o`0000Wooooo0_l0000cooooo`03
o`000?oooooooooo01Cooooo00?o0000ooooool000004?ooool00ol0003oooooooooo`0Fooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooool000001_l0000@ooooo`006?ooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool03_ooool00ol0003ooooo
o`000002o`0000_ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`03ooooo`Go0000
3oooool2o`00037ooooo00?o0000ooooooooool05?ooool2o`00013ooooo00?o0000ooooooooool0
5_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo000000Ko00004?ooool001Sooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00oooooo00?o
0000ooooooooool00_ooool3o`0000Sooooo00?o0000ooooooooool02_ooool00ol0003oooooo`00
0004o`000003ooooool0003ooooo01?ooooo0_l0000_ooooo`03o`000?oooooooooo01?ooooo0_l0
000Aooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool06oooool8o`00017ooooo000I
ooooo`04o`000?ooooooooooo`0000cooooo0_l0000Booooo`03o`000?oooooooooo00Gooooo0ol0
0005ooooo`03o`000?oooooooooo00Oooooo1Ol00005ooooo`03o`000?oooooooooo01Gooooo0_l0
000^ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool03oooool00ol0003ooooooooo
o`0Eooooo`03o`000?oooooooooo01_ooooo0_l00003ooooo`?o00004_ooool001Wooooo00Co0000
oooooooooooo00002oooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00Oooooo0_l0
0003ooooo`03o`000?oooooooooo00;ooooo1_l00009ooooo`03o`000?oooooooooo01Oooooo00?o
0000ooooooooool0:oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo013ooooo00?o
0000ooooooooool04oooool2o`0001cooooo0ol0000Iooooo`006_ooool2o`0000[ooooo0_l0000F
ooooo`03o`000?oooooooooo00Wooooo0ol000000ooooooo0000o`000002o`0000Cooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`0Iooooo`;o0000;?ooool00ol0003oooooooooo`0?
ooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`0J
ooooo`;o00000_ooool00ol0003oooooooooo`0Gooooo`006oooool00ol0003oooooooooo`07oooo
o`03o`000?oooooooooo01Oooooo00?o0000ooooooooool01oooool7o`0000Gooooo00?o0000oooo
ooooool02?ooool00ol0003oooooooooo`0Kooooo`;o0000:_ooool00ol0003oooooooooo`0>oooo
o`;o00004_ooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo01Wooooo0_l00004oooo
o`03o`000?oooooooooo01Oooooo000Kooooo`03o`000?oooooooooo00Gooooo0_l0000Jooooo`03
o`000?oooooooooo00;ooooo1Ol00003ooooo`04o`000?ooooooooooooooo`?o000000?oooooo`00
0?ooool02Oooool00ol0003oooooooooo`0Nooooo`;o0000:Oooool00ol0003oooooooooo`0<oooo
o`;o00004oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool00oooool00ol0003oooooooooo`0Hooooo`006_ooool00ol0003oooooo`000005ooooo`03
o`000?oooooooooo01[ooooo1Ol00008ooooo`03o`000?oooooooooo00Cooooo0ol00009ooooo`03
o`000?oooooooooo023ooooo0_l0000Wooooo`03o`000?oooooooooo00_ooooo0_l0000Cooooo`03
o`000?oooooooooo017ooooo00?o0000ooooooooool06?ooool2o`0000Kooooo00?o0000oooooooo
ool06?ooool001[ooooo00?o0000ooooool000001?ooool00ol0003oooooooooo`0Fooooo`Go0000
00?oooooo`000?ooool02_ooool00ol0003oooooooooo`03ooooo`04o`000?ooooooooooooooo`;o
00001oooool00ol0003oooooooooo`0Rooooo`;o00009Oooool00ol0003oooooooooo`09ooooo`;o
000000?oooooo`000?ooool04_ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo01Oo
oooo0_l00008ooooo`03o`000?oooooooooo01Sooooo000Iooooo`06o`000?ooooooooooooooool0
003ooooo0_l0000Dooooo`Go00001_ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo
00;ooooo00?o0000ooooooooool01?ooool3o`0000?ooooo00?o0000ooooooooool09Oooool2o`00
02Cooooo00?o0000ooooooooool01oooool01?l0003oooooooooool0000Cooooo`03o`000?oooooo
oooo013ooooo00?o0000ooooooooool05_ooool2o`0000[ooooo00?o0000ooooooooool06?ooool0
01Wooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0@ooooo`Co00003?ooool00ol0
003oooooooooo`08ooooo`05o`000?ooooooooooooooool000002_ooool4o`0002Wooooo0_l0000R
ooooo`03o`000?oooooooooo00Kooooo00Go0000ooooooooooooooooo`00000Cooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool05Oooool2o`0000_ooooo00?o0000ooooooooool06Ooo
ool001Sooooo00Co0000oooooooooooooooo0ol0000=ooooo`Go00004?ooool00ol0003ooooooooo
o`07ooooo`05o`000?ooooooooooooooool000003Oooool4o`0002Wooooo0_l0000Qooooo`03o`00
0?oooooooooo00Cooooo00Go0000ooooooooooooooooo`00000Cooooo`03o`000?oooooooooo00oo
oooo00?o0000ooooooooool05?ooool2o`0000gooooo00?o0000ooooooooool06Oooool001Sooooo
00Co0000oooooooooooo00000oooool00ol0003oooooooooo`05ooooo`Go00005_ooool00ol0003o
ooooooooo`06ooooo`04o`000?ooooooooooo`0000kooooo00Co0000oooooooooooooooo0ol0000X
ooooo`;o00007oooool00ol0003oooooooooo`03ooooo`05o`000?ooooooooooooooool000005?oo
ool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo01?ooooo0_l0000?ooooo`03o`000?oo
oooooooo01Wooooo000Eooooo`Co000000?oooooo`000?ooool01?ooool00ol0003oooooo`000004
o`0001_ooooo00?o0000ooooooooool01_ooool00ol0003oooooo`00000>ooooo`03o`000?oooooo
oooo00Gooooo0_l0000Xooooo`;o00007_ooool01Ol0003ooooooooooooooooo000000Cooooo00?o
0000ooooooooool04Oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo01;ooooo0_l0
000@ooooo`03o`000?oooooooooo01[ooooo000Dooooo`Ko00000oooool5o`00027ooooo00?o0000
ooooooooool01Oooool2o`0000oooooo00?o0000ooooooooool01oooool3o`0002Oooooo0_l0000L
ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000ooooooooool04_ooool00ol0003ooooooooo
o`0=ooooo`03o`000?oooooooooo017ooooo0_l0000Booooo`03o`000?oooooooooo01[ooooo000C
ooooo`[o00001?ooool00ol0003oooooooooo`0Oooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool03_ooool00ol0003oooooooooo`0:ooooo`?o00009_ooool2o`0001[ooooo00?o0000
ooooool000001Oooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo00cooooo00?o0000
ooooooooool04?ooool2o`0001Cooooo00?o0000ooooooooool06_ooool001?ooooo1ol00007oooo
o`03o`000?oooooooooo023ooooo00?o0000ooooooooool00oooool2o`0000oooooo00?o0000oooo
ooooool03_ooool3o`0002Gooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo01;ooooo00?o0000ooooooooool02oooool2o`00017ooooo0_l0000Foooo
o`03o`000?oooooooooo01[ooooo000Cooooo`Oo00002?ooool00ol0003oooooooooo`0Oooooo`04
o`000?ooooooooooooooo`;o000000?oooooo`000?ooool03_ooool00ol0003oooooooooo`0Aoooo
o`?o00008oooool2o`0001Oooooo0_l00004ooooo`03o`000?oooooooooo01?ooooo00?o0000oooo
ooooool02_ooool00ol0003oooooooooo`0?ooooo`;o00005oooool00ol0003oooooooooo`0Koooo
o`005?ooool5o`0000[ooooo00?o0000ooooooooool07_ooool01?l0003oooooooooool00002oooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool05Oooool2o`0002?ooooo0_l0000Doooo
o`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`0:
ooooo`03o`000?oooooooooo00kooooo0_l0000Iooooo`03o`000?oooooooooo01_ooooo000Eoooo
o`?o00002oooool00ol0003oooooooooo`0Oooooo`;o00000oooool00ol0003oooooooooo`0=oooo
o`03o`000?oooooooooo01Oooooo0ol0000Rooooo`;o00004Oooool01Ol0003ooooooooooooooooo
000000;ooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`09ooooo`03o`000?oooooo
oooo00gooooo0_l0000Kooooo`03o`000?oooooooooo01_ooooo000Tooooo`03o`000?oooooooooo
01kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo
01_ooooo0ol0000Qooooo`;o00003_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooo0000
01Cooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0<ooooo`;o00007?ooool00ol0
003oooooooooo`0Looooo`009?ooool00ol0003oooooooooo`0Mooooo`03o`000?oooooo000000?o
oooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0Nooooo`?o00008?ooool2o`0000[o
oooo0_l00006ooooo`;o00005Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00co
oooo00?o0000ooooooooool07?ooool00ol0003oooooooooo`0Looooo`009Oooool00ol0003ooooo
ooooo`0Kooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`0Rooooo`?o00007oooool2o`0000Oooooo00?o0000ooooooooool01_ooool2o`0001Go
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0;ooooo`;o00007oooool00ol0003o
ooooooooo`0Looooo`009_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo00;ooooo
00Co0000oooooooooooo00003Oooool00ol0003oooooooooo`0Uooooo`?o00007_ooool2o`0000Co
oooo00?o0000ooooooooool01oooool2o`0001Cooooo00?o0000ooooooooool01oooool00ol0003o
ooooooooo`0:ooooo`;o00008Oooool00ol0003oooooooooo`0Looooo`009_ooool00ol0003ooooo
ooooo`0Hooooo`03o`000?oooooooooo00?ooooo00Co0000oooooooooooo00003Oooool00ol0003o
ooooooooo`0Xooooo`;o00007_ooool2o`000003ooooool0003ooooo00Sooooo00?o0000ooooool0
00005?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Wooooo0_l0000Rooooo`03
o`000?oooooooooo01gooooo000Wooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool0
1Oooool2o`0000gooooo00?o0000ooooooooool0:oooool3o`0001gooooo0_l00008ooooo`05o`00
0?ooooooooooooooool000004_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00So
oooo0_l0000Tooooo`03o`000?oooooooooo01gooooo000Wooooo`03o`000?oooooooooo01Gooooo
00?o0000ooooooooool01_ooool2o`0000gooooo00?o0000ooooooooool0;_ooool3o`0001Wooooo
00?o0000ooooooooool00_l00006ooooo`05o`000?ooooooooooooooool000004_ooool00ol0003o
ooooooooo`05ooooo`03o`000?oooooooooo00Oooooo0_l0000Vooooo`03o`000?oooooooooo01go
oooo000Xooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool01oooool2o`0000cooooo
00?o0000ooooooooool0<_ooool3o`0001Gooooo00?o0000ooooooooool00oooool2o`0000?ooooo
00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool01_ooool2o`0002Sooooo00?o0000ooooooooool07Oooool002Wooooo00?o
0000ooooooooool04Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00[ooooo00?o
0000ooooooooool0=Oooool3o`00017ooooo00?o0000ooooooooool01_ooool01?l0003ooooooooo
ool00005ooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool01?ooool00ol0003ooooo
ooooo`05ooooo`;o0000:Oooool00ol0003oooooooooo`0Nooooo`00:Oooool00ol0003ooooooooo
o`0@ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`0iooooo`;o00003_ooool00ol0003oooooooooo`08ooooo`;o00001oooool00ol0003ooooooooo
o`0=ooooo`03o`000?oooooooooo00;ooooo0_l00006ooooo`;o0000:Oooool3o`00023ooooo000Z
ooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool02oooool2o`0000[ooooo00?o0000
ooooooooool0>oooool3o`0000[ooooo00?o0000ooooooooool02Oooool01?l0003oooooo`000?l0
0005ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00_ooool00ol0003ooooooooo
o`04ooooo`;o0000:_ooool5o`0001oooooo000Zooooo`03o`000?oooooooooo00gooooo00?o0000
ooooooooool03?ooool2o`0000[ooooo00?o0000ooooooooool0?_ooool3o`0000Kooooo00?o0000
ooooooooool02_ooool01?l0003oooooooooooooool2o`0000?ooooo00?o0000ooooooooool03?oo
ool01Ol0003ooooooooooooooooo000000Gooooo0_l0000[ooooo`Oo00007_ooool002_ooooo00?o
0000ooooooooool02_ooool2o`0000kooooo00Co0000oooooooooooo00002?ooool00ol0003ooooo
ooooo`12ooooo`?o00000_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Cooooo
0_l00002ooooo`03o`000?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000004ooooo`;o
0000;Oooool7o`0001kooooo000/ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0
3_ooool01?l0003oooooooooool00008ooooo`03o`000?oooooooooo04Gooooo0ol0000<ooooo`03
o`000?oooooooooo00Kooooo0ol0000<ooooo`04o`000?ooooooooooo`0000?ooooo0_l0000[oooo
o`_o00007_ooool002cooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0?ooooo`04
o`000?ooooooooooo`0000Oooooo00?o0000ooooooooool0AOooool2o`0000;ooooo0_l00009oooo
o`03o`000?oooooooooo00Wooooo0_l0000:ooooo`04o`000?ooooooooooo`0000;ooooo0_l0000X
ooooo`Go00000_ooool2o`000003ooooool0003o000000?o00007oooool002gooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`0@ooooo`05o`000?ooooooooooooooool000001_ooool0
0ol0003oooooooooo`14ooooo`03o`000?oooooooooo00Cooooo0ol00005ooooo`03o`000?oooooo
oooo00_ooooo0ol00008ooooo`04o`000?oooooo0000ooooo`;o00009Oooool5o`0000Gooooo0_l0
0004ooooo`?o00008?ooool002gooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0A
ooooo`05o`000?ooooooooooooooool000001Oooool00ol0003oooooooooo`14ooooo`03o`000?oo
oooooooo00Sooooo0ol00002ooooo`03o`000?oooooooooo00cooooo00Co0000ooooool0003o0000
1_ooool2o`000003ooooool0003ooooo027ooooo1Ol00008ooooo`;o0000:Oooool002kooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo00;ooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`13ooooo`03o`000?oooooooooo00cooooo0ol0
000>ooooo`04o`000?ooooooooooooooo`;o00000oooool4o`0001oooooo1?l0000;ooooo`;o0000
:oooool002oooooo00Co0000oooooooooooo00005Oooool00ol0003oooooooooo`02ooooo`03o`00
0?oooooooooo00;ooooo00?o0000ooooooooool0@_ooool00ol0003oooooooooo`0>ooooo`03o`00
0?oooooo000000;o00003?ooool00ol0003oooooooooo`02ooooo`;o000000?oooooo`000?l00000
7?ooool5o`0000gooooo0_l0000]ooooo`00;oooool00ol0003oooooo`00000Fooooo`03o`000?oo
oooooooo00?ooooo00Co0000oooooooooooo0000A?ooool00ol0003oooooooooo`0>ooooo`03o`00
0?oooooooooo00?ooooo0_l0000:ooooo`03o`000?oooooooooo00?ooooo0ol0000Hooooo`Go0000
4?ooool2o`0002oooooo000`ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool01?oo
ool01?l0003oooooooooool00013ooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool0
1_ooool3o`0000Oooooo00Co0000oooooooooooooooo1?l000000ooooooo0000o`00000Aooooo`Go
00004oooool2o`00037ooooo000_ooooo`;o00005_ooool00ol0003oooooooooo`05ooooo`;o0000
@oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00Wooooo0ol00005ooooo`?o0000
00?oooooo`000?l000001?ooool00ol0003oooooooooo`09ooooo`Go00005_ooool2o`0003?ooooo
000^ooooo`04o`000?ooooooooooo`0001Gooooo00?o0000ooooooooool01Oooool2o`0004;ooooo
00?o0000ooooooooool03oooool00ol0003oooooooooo`0=ooooo`?o000000Coooooo`000?l0003o
oooo0_l000000ooooooo0000ooooo`04ooooo`;o00001Oooool4o`0001[ooooo00?o0000oooooooo
ool0<oooool002gooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Booooo`03o`00
0?oooooooooo00Gooooo00?o0000ooooooooool0@?ooool00ol0003oooooooooo`0?ooooo`03o`00
0?oooooooooo013ooooo1Ol00002ooooo`03o`000?oooooooooo00Kooooo1Ol0000Looooo`;o0000
=_ooool002Sooooo0ol000000ooooooo0000ooooo`04ooooo`03o`000?oooooooooo01;ooooo00?o
0000ooooooooool01Oooool2o`00043ooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`0>ooooo`;o00000oooool3o`000003ooooool0003ooooo00;ooooo1Ol00002ooooo`;o00006ooo
ool2o`0003Sooooo000Wooooo`Go00001oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooo
oooo00Gooooo0_l0000oooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool03Oooool2
o`0000Cooooo00?o0000ooooooooool01_l00009ooooo`;o00005oooool2o`0003[ooooo000Voooo
o`Oo00001oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo00Cooooo00Co0000oooo
oooooooo0000?Oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00_ooooo0_l00005
ooooo`Ko00000_ooool3o`0000[ooooo0_l0000Cooooo`;o0000??ooool002Kooooo2ol00003oooo
o`03o`000?oooooooooo00oooooo00?o0000ooooooooool01Oooool01?l0003oooooooooool0000k
ooooo`;o00004oooool00ol0003oooooooooo`0:ooooo`;o00000oooool4o`0000Cooooo0_l00005
ooooo`?o00002Oooool2o`0000oooooo0_l0000nooooo`009_ooool7o`0000Cooooo3ol00006oooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0goooo
o`03o`000?oooooooooo01;ooooo00?o0000ooooooooool02Oooool7o`0000;ooooo00?o0000oooo
ooooool00oooool2o`0000Sooooo0ol00008ooooo`;o00002oooool2o`00043ooooo000Wooooo`Go
00002Oooool00ol0003oooooooooo`08ooooo`oo00000oooool00ol0003oooooooooo`0fooooo`03
o`000?oooooooooo01?ooooo00?o0000ooooooooool01_ooool5o`0000Kooooo00?o0000oooooooo
ool01?ooool2o`0000_ooooo0ol00007ooooo`;o00001oooool2o`0004;ooooo000Xooooo`?o0000
2oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0
3_l0000]ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool00_ooool6o`0000Wooooo
00?o0000ooooooooool01?ooool01?l0003oooooooooool0000=ooooo`;o00001oooool2o`0000?o
oooo0_l00014ooooo`00=oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`07ooooo`oo00007Oooool00ol0003ooooo
ooooo`0Dooooo`Go00000oooool2o`0000Wooooo00?o0000ooooooooool01Oooool01?l0003ooooo
ooooool0000?ooooo`?o00001_ooool3o`0004Kooooo000gooooo`03o`000?oooooooooo00cooooo
00?o0000ooooooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01Kooooo
3ol0000=ooooo`03o`000?oooooooooo017ooooo1?l00006ooooo`;o00002_ooool00ol0003ooooo
ooooo`06ooooo`05o`000?ooooooooooooooool000004Oooool3o`0000;ooooo0_l000000ooooooo
0000o`000015ooooo`00>?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00;ooooo
00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Tooooo`oo00003Oooool5o`0000;ooooo
00?o0000ooooooooool00oooool2o`0000_ooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo01;ooooo0ol00004ooooo`;o0000@oooool003Sooooo00?o
0000ooooooooool02_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool0;oooool01?l0003oooooooooooooool>o`0000Kooooo00Co0000oooooooooooo
oooo0_l0000<ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`0@ooooo`;o00000oooool3o`0000?ooooo00?o0000ooooooooool0@?ooool003Wooooo
00?o0000ooooooooool02Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Sooooo
00?o0000ooooooooool0;Oooool00ol0003oooooooooo`05ooooo`Go00001_ooool?o`0000Oooooo
0_l00009ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03Oooool2o`0000Sooooo
0_l00002ooooo`;o0000@?ooool003[ooooo00?o0000ooooooooool02?ooool00ol0003ooooooooo
o`02ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0;?ooool00ol0003ooooooooo
o`02ooooo`Co00004?ooool3o`0000Oooooo3ol00003ooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool02oooool2o`0000cooooo0ol000000ooooooo0000o`00000nooooo`00>_ooool0
0ol0003oooooooooo`08ooooo`05o`000?ooooooooooooooool000003?ooool00ol0003ooooooooo
o`0Zooooo`Ko00004oooool2o`0000kooooo00?o0000ooooooooool01oooool>o`0000Sooooo0_l0
000Aooooo`Go00000oooool3o`0003Kooooo000kooooo`03o`000?oooooooooo00Oooooo00Go0000
ooooooooooooooooo`00000<ooooo`03o`000?oooooooooo02Kooooo1Ol0000Fooooo`;o00003ooo
ool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool03ol0
000?ooooo`Wo0000=Oooool003_ooooo00?o0000ooooooooool01oooool01?l0003oooooooooool0
000=ooooo`03o`000?oooooooooo027ooooo1Ol00002ooooo`03o`000?oooooooooo01Cooooo0_l0
00000ooooooo0000ooooo`0=ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool01ooo
ool00ol0003oooooooooo`03ooooo`;o00002Oooool?o`0000?ooooo1ol0000dooooo`00??ooool0
0ol0003oooooooooo`06ooooo`04o`000?ooooooooooo`0000kooooo00?o0000ooooooooool06ooo
ool5o`0000Kooooo00?o0000ooooooooool04oooool2o`0000;ooooo00?o0000ooooooooool03?oo
ool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Sooooo00Co0000oooooooooooooooo
0_l0000Jooooo`[o0000=?ooool003gooooo00?o0000ooooooooool01?ooool01Ol0003ooooooooo
oooooooo000000kooooo00?o0000ooooooooool05oooool4o`0000Wooooo0_l0000Dooooo`;o0000
0oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0
2Oooool3o`0001kooooo2?l0000dooooo`00?Oooool00ol0003oooooooooo`04ooooo`04o`000?oo
ooooooooo`00013ooooo00?o0000ooooooooool04Oooool5o`0000cooooo00?o0000ooooooooool0
4_ooool2o`0000Gooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0<ooooo`03o`00
0?oooooooooo00Wooooo0_l0000Mooooo`[o0000=Oooool003kooooo00?o0000ooooooooool00ooo
ool01?l0003oooooooooool0000@ooooo`03o`000?oooooooooo00cooooo1Ol0000@ooooo`03o`00
0?oooooooooo017ooooo0_l00006ooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool0
3Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooo000001[ooooo0ol00002ooooo`04o`00
0?ooooooooooooooo`?o0000=_ooool003kooooo00?o0000ooooooooool00oooool00ol0003ooooo
o`00000Booooo`03o`000?oooooooooo00Kooooo1Ol0000Dooooo`03o`000?oooooooooo013ooooo
0_l00008ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`07ooooo`;o00000oooool00ol0003oooooooooo`0Dooooo`?o00000oooool2o`0003gooooo
000oooooo`03o`000?oooooooooo00;ooooo00?o0000ooooool000004_ooool01?l0003ooooooooo
ooooool5o`0001Sooooo00?o0000ooooooooool04?ooool00ol0003oooooooooo`07ooooo`03o`00
0?oooooooooo00[ooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`05ooooo`;o0000
1Oooool00ol0003oooooooooo`0Aooooo`?o00001Oooool00ol0003oooooooooo`0mooooo`00@?oo
ool01?l0003oooooooooooooool2o`0001?ooooo1?l0000Looooo`03o`000?oooooooooo00oooooo
0_l00009ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool03_ooool00ol0003ooooo
ooooo`04ooooo`;o00002?ooool00ol0003oooooooooo`0=ooooo`?o00001oooool00ol0003ooooo
ooooo`0nooooo`00@?ooool01?l0003oooooooooooooool2o`0000kooooo1Ol000000ooooooo0000
ooooo`0Looooo`03o`000?oooooooooo00kooooo0_l0000;ooooo`03o`000?oooooooooo00Wooooo
00?o0000ooooooooool03oooool00ol0003oooooooooo`02ooooo`;o00002_ooool00ol0003ooooo
ooooo`0:ooooo`?o00002?ooool2o`00047ooooo0011ooooo`Go00002Oooool5o`0000Kooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0=ooooo`;o00003?ooool00ol0003ooooooooo
o`08ooooo`;o00004_ooool00ol0003oooooooooo`02o`0000gooooo00?o0000ooooooooool01_oo
ool3o`0000[ooooo00?o0000ooooooooool0@Oooool0047ooooo1Ol00004ooooo`Go00003?ooool0
0ol0003oooooooooo`0Hooooo`03o`000?oooooooooo00cooooo0_l0000>ooooo`03o`000?oooooo
oooo00Oooooo00?o0000ooooooooool04Oooool01?l0003oooooo`000?l0000?ooooo`03o`000?oo
oooooooo00?ooooo0ol0000;ooooo`;o0000A?ooool0043ooooo2_l0000Aooooo`03o`000?oooooo
oooo01Oooooo00?o0000ooooooooool02oooool2o`0000oooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`0Booooo`;o00004Oooool00ol0003oooooooooo`03o`0000gooooo00?o0000
ooooooooool0A?ooool0043ooooo1ol0000Eooooo`03o`000?oooooooooo01Gooooo00?o0000oooo
ooooool02_ooool2o`00013ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0Aoooo
o`;o00005?ooool2o`0000kooooo0_l00017ooooo`00@?ooool7o`0001Gooooo00?o0000oooooooo
ool05?ooool00ol0003oooooooooo`09ooooo`;o00004_ooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo013ooooo0_l000000ooooooo0000ooooo`0@ooooo`Co00003_ooool00ol0003o
ooooooooo`17ooooo`00@Oooool5o`000003ooooool0003o000001Cooooo00?o0000ooooooooool0
4_ooool00ol0003oooooooooo`08ooooo`;o00004oooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00oooooo0_l00002ooooo`03o`000?oooooooooo00gooooo0ol00004ooooo`03o`00
0?oooooooooo00Wooooo0_l0001:ooooo`00@_ooool3o`0000Cooooo0_l0000Booooo`03o`000?oo
oooooooo013ooooo0_l00009ooooo`;o00005Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00kooooo0_l00004ooooo`03o`000?oooooooooo00[ooooo0ol00007ooooo`03o`000?oo
oooooooo00Sooooo00?o0000ooooooooool0B_ooool004_ooooo0_l0000Aooooo`03o`000?oooooo
oooo00kooooo00?o0000ooooooooool01oooool2o`0001Kooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`0=ooooo`;o00001_ooool00ol0003oooooooooo`07ooooo`?o00002oooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo04_ooooo001=ooooo`;o00003oooool00ol0
003oooooooooo`0=ooooo`03o`000?oooooooooo00Kooooo0_l0000Gooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool03?ooool2o`0000Oooooo00?o0000ooooooooool01Oooool3o`00
00kooooo00?o0000ooooooooool01?ooool2o`0004kooooo001?ooooo`;o00003Oooool00ol0003o
ooooooooo`0<ooooo`03o`000?oooooooooo00Gooooo0_l0000Iooooo`03o`000?oooooooooo00Co
oooo00?o0000ooooooooool02oooool2o`0000Wooooo00?o0000ooooooooool00_ooool3o`00017o
oooo1?l00002ooooo`03o`000?oooooooooo04kooooo001Aooooo`;o00003?ooool00ol0003ooooo
ooooo`0:ooooo`03o`000?oooooooooo00Cooooo0_l0000Jooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool02_ooool2o`0000[ooooo00?o0000ooooooooool00ol0000Dooooo`Ko0000
DOooool005?ooooo0_l0000:ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00ooo
ool2o`0001_ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09ooooo`;o00003?oo
ool3o`0001Kooooo1ol0001Aooooo`00EOooool2o`0000Wooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`03ooooo`03o`000?oooooooooo01_ooooo00?o0000ooooooooool00oooool0
0ol0003oooooooooo`08ooooo`;o00002oooool3o`0001Wooooo1ol0001Aooooo`00Eoooool2o`00
00Oooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`02ooooo`;o00007Oooool00ol0
003oooooooooo`02ooooo`;o00002Oooool2o`0000[ooooo0ol00002ooooo`03o`000?oooooooooo
01Oooooo1ol0001Aooooo`00FOooool2o`0000Kooooo00?o0000ooooooooool01?ooool01?l0003o
ooooooooooooool2o`0001oooooo00Go0000ooooooooooooooooo`000009ooooo`;o00002Oooool3
o`0000Gooooo00?o0000ooooooooool06?ooool5o`0005;ooooo001Kooooo`;o00001?ooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo00;o00008?ooool01Ol0003ooooooooooooooooo
000000Sooooo0_l00009ooooo`;o00001oooool00ol0003oooooooooo`0Jooooo`?o0000Doooool0
05gooooo0_l00002ooooo`?o00000_ooool01?l0003oooooo`000?l0000Qooooo`05o`000?oooooo
ooooooooool000002?ooool00ol0003oooooooooo`06ooooo`?o00002Oooool00ol0003ooooooooo
o`1`ooooo`00Goooool9o`0002?ooooo00Co0000oooooooooooo00001oooool2o`0000Kooooo0ol0
000;ooooo`03o`000?oooooooooo077ooooo001Oooooo`Oo00009?ooool01?l0003oooooooooool0
0006ooooo`;o00001Oooool3o`0000kooooo00?o0000ooooooooool0LOooool005oooooo3?l0000O
ooooo`03o`000?oooooo000000Gooooo0_l00004ooooo`?o00004?ooool00ol0003oooooooooo`1b
ooooo`00Goooool7o`0000Gooooo3_l0000@ooooo`03o`000?oooooo000000Cooooo0_l00003oooo
o`?o00004oooool00ol0003oooooooooo`1booooo`00H?ooool5o`0001Cooooo3_l000001?oooooo
0000ooooool00003ooooo`;o00000_ooool3o`0001?ooooo1?l0001dooooo`00HOooool3o`0002?o
oooo3_l0000Booooo`Go0000M?ooool008Oooooo0_l000000ooooooo0000o`000003o`0000Kooooo
3_l00003ooooo`Oo0000Loooool008Cooooo2?l0000Gooooo`[o0000Loooool008?ooooo1_l0000M
ooooo`Oo0000Loooool008;ooooo1ol0000Nooooo`Go0000M?ooool008;ooooo1ol0000Oooooo`?o
0000MOooool008;ooooo1ol0002Gooooo`00Poooool5o`0009Sooooo0024ooooo`?o0000VOooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.04651, -0.0500055, \
0.00382061, 0.00382873}},
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]],

Cell["\<\
It is impossible to create a graph with an odd number of vertices \
such that each vertex has degree one.\
\>", "Text",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(BoundedDegreeNetwork[15, 1, 1]\)], "Input",
  CellLabel->"In[11]:=",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[BoxData[
    \(BoundedDegreeNetwork::"failure" \(\(:\)\(\ \)\) 
      "No suitable degree sequence satisfying the specified constraints was \
found after 50 iterations."\)], "Message",
  CellLabel->"From In[11]:=",
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]],

Cell["\<\
However, graphs with an even number of vertices can be constructed \
under this constraint.\
\>", "Text",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(BoundedDegreeNetwork[14, 1, 1] // ShowGraph;\)\)], "Input",
  CellLabel->"In[13]:=",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0348525 0.930295 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.90953 .70182 m
.2166 .13633 L
s
.7834 .86367 m
.09047 .29818 L
s
.60115 .95349 m
.39885 .04651 L
s
.39885 .95349 m
.7834 .13633 L
s
.2166 .86367 m
.90953 .29818 L
s
.09047 .70182 m
.95455 .5 L
s
.04545 .5 m
.60115 .04651 L
s
.025 w
.90953 .70182 Mdot
.7834 .86367 Mdot
.60115 .95349 Mdot
.39885 .95349 Mdot
.2166 .86367 Mdot
.09047 .70182 Mdot
.04545 .5 Mdot
.09047 .29818 Mdot
.2166 .13633 Mdot
.39885 .04651 Mdot
.60115 .04651 Mdot
.7834 .13633 Mdot
.90953 .29818 Mdot
.95455 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[13]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00LOooool3o`0003Oooooo0ol0001booooo`00L?ooool5o`0003Gooooo1Ol0001aooooo`00
Koooool7o`0003?ooooo1ol0001`ooooo`00Koooool7o`0003?ooooo1ol0001`ooooo`00Koooool7
o`0003?ooooo1ol0001`ooooo`00L?ooool5o`0003Cooooo1_l0001aooooo`00LOooool3o`0003Co
oooo00?o0000ooooooooool00ol0001booooo`00Loooool00ol0003oooooooooo`0`ooooo`;o0000
N?ooool007?ooooo00?o0000ooooooooool0;oooool00ol0003oooooooooo`1hooooo`00Loooool0
0ol0003oooooooooo`0^ooooo`03o`000?oooooooooo07Wooooo001cooooo`03o`000?oooooooooo
02gooooo00?o0000ooooooooool0N_ooool007Cooooo00?o0000ooooooooool0:oooool00ol0003o
ooooooooo`1kooooo`00M?ooool00ol0003oooooooooo`0Yooooo`;o0000O_ooool007Cooooo00?o
0000ooooooooool0:?ooool00ol0003oooooooooo`1nooooo`00M?ooool00ol0003oooooooooo`0W
ooooo`03o`000?oooooooooo07oooooo001eooooo`03o`000?oooooooooo02Gooooo00?o0000oooo
ooooool0P?ooool007Gooooo00?o0000ooooooooool08oooool2o`0008?ooooo001eooooo`03o`00
0?oooooooooo02;ooooo00?o0000ooooooooool0Poooool007Gooooo00?o0000ooooooooool08Ooo
ool00ol0003oooooooooo`24ooooo`00MOooool00ol0003oooooooooo`0Pooooo`03o`000?oooooo
oooo08Gooooo001fooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool0Q_ooool007Ko
oooo00?o0000ooooooooool07?ooool2o`0008Wooooo001fooooo`03o`000?oooooooooo01_ooooo
00?o0000ooooooooool0ROooool007Kooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`2:ooooo`00Moooool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo08_ooooo001goooo
o`03o`000?oooooooooo01Kooooo0_l0002>ooooo`00?Oooool3o`0003Oooooo00?o0000oooooooo
ool05Oooool00ol0003oooooooooo`1=ooooo`?o0000?_ooool003cooooo1Ol0000fooooo`03o`00
0?oooooooooo01Cooooo00?o0000ooooooooool0COooool5o`0003gooooo000kooooo`Oo0000=Ooo
ool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo04gooooo1ol0000looooo`00>oooool7
o`0003Kooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`1>ooooo`Oo0000??ooool0
03_ooooo1ol0000fooooo`03o`000?oooooooooo00oooooo0_l0001Aooooo`Oo0000??ooool003co
oooo1_l0000fooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool0D_ooool5o`0003go
oooo000mooooo`?o00000_ooool00ol0003oooooooooo`0cooooo`03o`000?oooooooooo00gooooo
00?o0000ooooooooool0E?ooool3o`0003kooooo0013ooooo`;o0000=?ooool00ol0003ooooooooo
o`0;ooooo`03o`000?oooooooooo05Cooooo00?o0000ooooooooool0?oooool004Gooooo00?o0000
ooooooooool0<Oooool00ol0003oooooooooo`09ooooo`;o0000Eoooool00ol0003oooooooooo`0o
ooooo`00A_ooool00ol0003oooooooooo`0`ooooo`03o`000?oooooooooo00Sooooo00?o0000oooo
ooooool0E_ooool00ol0003oooooooooo`10ooooo`00Aoooool00ol0003oooooooooo`0_ooooo`03
o`000?oooooooooo00Oooooo00?o0000ooooooooool0Eoooool00ol0003oooooooooo`10ooooo`00
B?ooool00ol0003oooooooooo`0^ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool0
Eoooool00ol0003oooooooooo`11ooooo`00BOooool2o`0002oooooo00?o0000ooooooooool01?oo
ool00ol0003oooooooooo`1Hooooo`03o`000?oooooooooo047ooooo001;ooooo`03o`000?oooooo
oooo02cooooo00?o0000ooooooooool00_ooool2o`0005_ooooo00?o0000ooooooooool0@Oooool0
04cooooo00?o0000ooooooooool0:oooool01Ol0003ooooooooooooooooo000005cooooo00?o0000
ooooooooool0@_ooool004gooooo00?o0000ooooooooool0:_ooool01?l0003oooooooooool0001M
ooooo`03o`000?oooooooooo04;ooooo001>ooooo`;o0000:oooool2o`0005gooooo00?o0000oooo
ooooool0@oooool0053ooooo00?o0000ooooooooool09oooool2o`0005kooooo00?o0000oooooooo
ool0@oooool0057ooooo00?o0000ooooooooool09Oooool00ol0003oooooo`00001Mooooo`03o`00
0?oooooooooo04Cooooo001Booooo`03o`000?oooooooooo02?ooooo00Co0000oooooooooooo0000
GOooool00ol0003oooooooooo`14ooooo`00Doooool00ol0003oooooooooo`0Qooooo`05o`000?oo
ooooooooooooool00000G?ooool00ol0003oooooooooo`15ooooo`00E?ooool2o`00023ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`1Iooooo`03o`000?oooooooooo04Gooooo001F
ooooo`03o`000?oooooooooo01_ooooo0_l00006ooooo`03o`000?oooooooooo05Sooooo00?o0000
ooooooooool0A_ooool005Oooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`06oooo
o`03o`000?oooooooooo05Sooooo00?o0000ooooooooool0A_ooool005Sooooo00?o0000oooooooo
ool05oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo05Oooooo00?o0000oooooooo
ool0Aoooool005Wooooo0_l0000Fooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool0
E_ooool00ol0003oooooooooo`17ooooo`00Foooool00ol0003oooooooooo`0Aooooo`;o00003?oo
ool00ol0003oooooooooo`1Eooooo`03o`000?oooooooooo04Sooooo001Looooo`03o`000?oooooo
oooo00oooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`1Eooooo`03o`000?oooooo
oooo04Sooooo001Mooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool03Oooool00ol0
003oooooooooo`1Dooooo`03o`000?oooooooooo04Wooooo001Nooooo`;o00003?ooool00ol0003o
ooooooooo`0>ooooo`03o`000?oooooooooo05Cooooo00?o0000ooooooooool0BOooool0063ooooo
00?o0000ooooooooool02?ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo05?ooooo
00?o0000ooooooooool0BOooool0067ooooo00?o0000ooooooooool01Oooool2o`0001?ooooo00?o
0000ooooooooool0D_ooool00ol0003oooooooooo`1:ooooo`00H_ooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool0D_ooool00ol0003oooooooooo`1:
ooooo`00Hoooool01Ol0003ooooooooooooooooo000001Kooooo00?o0000ooooooooool0DOooool0
0ol0003oooooooooo`1;ooooo`00I?ooool3o`0001Sooooo00?o0000ooooooooool0D?ooool00ol0
003oooooooooo`1;ooooo`00I?ooool3o`0001Sooooo00?o0000ooooooooool0Coooool00ol0003o
ooooooooo`1<ooooo`00Hoooool01Ol0003ooooooooooooooooo000001Oooooo00?o0000oooooooo
ool0Coooool00ol0003oooooooooo`1<ooooo`00H_ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo01Cooooo00?o0000ooooooooool0C_ooool00ol0003oooooooooo`1=ooooo`00HOoo
ool00ol0003oooooooooo`05ooooo`;o00005?ooool00ol0003oooooooooo`1>ooooo`03o`000?oo
oooooooo04gooooo001Pooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool04_ooool0
0ol0003oooooooooo`1<ooooo`03o`000?oooooooooo04kooooo001Nooooo`;o00003?ooool00ol0
003oooooooooo`0Aooooo`03o`000?oooooooooo04cooooo00?o0000ooooooooool0C_ooool005go
oooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo04_o
oooo00?o0000ooooooooool0Coooool005cooooo00?o0000ooooooooool03oooool2o`00013ooooo
00?o0000ooooooooool0Boooool00ol0003oooooooooo`1?ooooo`00Foooool00ol0003ooooooooo
o`0Booooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool0BOooool00ol0003ooooooooo
o`1@ooooo`00FOooool2o`0001Kooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`19
ooooo`03o`000?oooooooooo053ooooo001Hooooo`03o`000?oooooooooo01Oooooo00?o0000oooo
ooooool03?ooool00ol0003oooooooooo`18ooooo`03o`000?oooooooooo057ooooo000Hooooo`?o
0000??ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo00_ooooo00?o0000oooooooo
ool0B?ooool00ol0003oooooooooo`0dooooo`?o00006_ooool001Oooooo1Ol0000jooooo`03o`00
0?oooooooooo01_ooooo0_l0000<ooooo`03o`000?oooooooooo04Oooooo00?o0000ooooooooool0
<oooool5o`0001Wooooo000Fooooo`Oo0000>?ooool00ol0003oooooooooo`0Nooooo`03o`000?oo
oooooooo00Wooooo00?o0000ooooooooool0A_ooool00ol0003oooooooooo`0cooooo`Oo00006?oo
ool001Kooooo1ol0000fooooo`;o00008_ooool00ol0003oooooooooo`08ooooo`03o`000?oooooo
oooo04Kooooo00?o0000ooooooooool0<oooool7o`0001Sooooo000Fooooo`Oo0000=Oooool00ol0
003oooooooooo`0Sooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool0AOooool00ol0
003oooooooooo`0dooooo`Oo00006?ooool001Oooooo1_l0000dooooo`03o`000?oooooooooo02Go
oooo0_l00007ooooo`03o`000?oooooooooo04Gooooo00?o0000ooooooooool0=Oooool5o`0001Wo
oooo000Hooooo`?o00000_ooool00ol0003oooooooooo`0`ooooo`03o`000?oooooooooo02Sooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`13ooooo`03o`000?oooooooooo03Gooooo
00?o0000ooooool000000_l0000Jooooo`007_ooool2o`0002kooooo0_l0000/ooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool0@oooool00ol0003oooooooooo`0cooooo`;o00007ooo
ool0023ooooo00?o0000ooooooooool0:_ooool00ol0003oooooooooo`0]ooooo`03o`000?oooooo
oooo00?ooooo00?o0000ooooooooool0@_ooool00ol0003oooooooooo`0cooooo`03o`000?oooooo
oooo01oooooo000Qooooo`03o`000?oooooooooo02Sooooo00?o0000ooooooooool0;oooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo04;ooooo00?o0000ooooooooool0<_ooool00ol0
003oooooooooo`0Pooooo`008_ooool00ol0003oooooooooo`0Vooooo`03o`000?oooooooooo037o
oooo0_l00003ooooo`03o`000?oooooooooo043ooooo00?o0000ooooooooool0<_ooool00ol0003o
ooooooooo`0Qooooo`008oooool00ol0003oooooooooo`0Tooooo`03o`000?oooooooooo03Cooooo
00Co0000oooooooooooo0000@_ooool00ol0003oooooooooo`0aooooo`03o`000?oooooooooo02;o
oooo000Tooooo`;o00008_ooool2o`0003Sooooo00?o0000ooooool00000@Oooool00ol0003ooooo
ooooo`0`ooooo`;o00009Oooool002Kooooo00?o0000ooooooooool07_ooool00ol0003ooooooooo
o`0iooooo`;o0000@Oooool00ol0003oooooooooo`0_ooooo`03o`000?oooooooooo02Gooooo000W
ooooo`03o`000?oooooooooo01cooooo00?o0000ooooooooool0>oooool2o`0003oooooo00?o0000
ooooooooool0;oooool00ol0003oooooooooo`0Vooooo`00:?ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo03gooooo0_l0000nooooo`03o`000?oooooooooo02kooooo00?o0000oooo
ooooool09oooool002Wooooo0_l0000Hooooo`;o0000@?ooool00ol0003oooooo`00000looooo`03
o`000?oooooooooo02gooooo0_l0000Zooooo`00:oooool00ol0003oooooooooo`0Dooooo`03o`00
0?oooooooooo043ooooo00Co0000oooooooooooo0000>oooool00ol0003oooooooooo`0/ooooo`03
o`000?oooooooooo02[ooooo000/ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool0
@Oooool01?l0003oooooooooooooool2o`0003Wooooo00?o0000ooooooooool0:oooool00ol0003o
ooooooooo`0[ooooo`00;Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo04?ooooo
00?o0000ooooooooool00_ooool00ol0003oooooooooo`0eooooo`03o`000?oooooooooo02_ooooo
00?o0000ooooooooool0;?ooool002kooooo00?o0000ooooooooool03_ooool00ol0003ooooooooo
o`14ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0=?ooool00ol0003ooooooooo
o`0Zooooo`03o`000?oooooooooo02gooooo000_ooooo`;o00003?ooool2o`0004Oooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`0booooo`03o`000?oooooooooo02Wooooo0_l0000`
ooooo`00<Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo04Oooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`0aooooo`03o`000?oooooooooo02Sooooo00?o0000oooo
ooooool0<?ooool003;ooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`18ooooo`03
o`000?oooooooooo00Kooooo0_l0000`ooooo`03o`000?oooooooooo02Sooooo00?o0000oooooooo
ool0<Oooool003?ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`1:ooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool0;Oooool00ol0003oooooooooo`0Wooooo`03o`00
0?oooooooooo03;ooooo000dooooo`;o00000_ooool2o`0004gooooo00?o0000ooooooooool02?oo
ool00ol0003oooooooooo`0[ooooo`03o`000?oooooooooo02Kooooo0_l0000eooooo`00=_ooool2
o`0004oooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Zooooo`03o`000?oooooo
oooo02Gooooo00?o0000ooooooooool0=Oooool003Kooooo0_l0001?ooooo`03o`000?oooooooooo
00[ooooo0_l0000Yooooo`03o`000?oooooooooo02Gooooo00?o0000ooooooooool0=_ooool003Go
oooo00Co0000oooooooooooo0000Coooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo
02Kooooo00?o0000ooooooooool09?ooool00ol0003oooooooooo`0gooooo`00=?ooool00ol0003o
ooooooooo`02ooooo`;o0000COooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo02Co
oooo00?o0000ooooooooool08oooool2o`0003[ooooo000booooo`;o00001oooool00ol0003ooooo
ooooo`1:ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool08oooool00ol0003ooooo
ooooo`0Rooooo`03o`000?oooooooooo03[ooooo000aooooo`03o`000?oooooooooo00Sooooo00?o
0000ooooooooool0BOooool00ol0003oooooooooo`0>ooooo`;o00008_ooool00ol0003ooooooooo
o`0Rooooo`03o`000?oooooooooo03_ooooo000`ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooooooool0B?ooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo01oooooo00?o0000
ooooooooool08Oooool00ol0003oooooooooo`0looooo`00;oooool00ol0003oooooooooo`0<oooo
o`03o`000?oooooooooo04Sooooo00?o0000ooooooooool04?ooool00ol0003oooooooooo`0Noooo
o`03o`000?oooooooooo023ooooo00?o0000ooooooooool0?Oooool002gooooo0_l0000@ooooo`;o
0000B?ooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo01cooooo00?o0000oooooooo
ool07oooool2o`00043ooooo000/ooooo`03o`000?oooooooooo01;ooooo00?o0000ooooooooool0
AOooool00ol0003oooooooooo`0Booooo`03o`000?oooooooooo01_ooooo00?o0000ooooooooool0
7_ooool00ol0003oooooooooo`10ooooo`00:oooool00ol0003oooooooooo`0Dooooo`03o`000?oo
oooooooo04Cooooo00?o0000ooooooooool04oooool2o`0001[ooooo00?o0000ooooooooool07_oo
ool00ol0003oooooooooo`11ooooo`00:_ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooo
oooo04Cooooo00?o0000ooooooooool05?ooool00ol0003oooooooooo`0Gooooo`03o`000?oooooo
oooo01gooooo00?o0000ooooooooool0@_ooool002Wooooo00?o0000ooooooooool06?ooool2o`00
04Cooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`0Eooooo`03o`000?oooooooooo
01cooooo0_l00015ooooo`009oooool2o`0001gooooo00?o0000ooooooooool0@Oooool00ol0003o
ooooooooo`0Fooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool06oooool00ol0003o
ooooooooo`15ooooo`009_ooool00ol0003oooooooooo`0Nooooo`03o`000?oooooooooo043ooooo
00?o0000ooooooooool05oooool2o`0001?ooooo00?o0000ooooooooool06oooool00ol0003ooooo
ooooo`16ooooo`009Oooool00ol0003oooooooooo`0Pooooo`03o`000?oooooooooo03oooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool0Aoooool002Cooooo00?o0000ooooooooool08_ooool2o`00043ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo01Wooooo0_l0001:
ooooo`008_ooool2o`0002Oooooo00?o0000ooooooooool0?Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00gooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`1:oooo
o`008Oooool00ol0003oooooooooo`0Xooooo`03o`000?oooooooooo03cooooo00?o0000oooooooo
ool06oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo01Sooooo00?o0000oooooooo
ool0Boooool0023ooooo00?o0000ooooooooool0:_ooool00ol0003oooooooooo`0kooooo`03o`00
0?oooooooooo01cooooo0_l0000;ooooo`03o`000?oooooooooo01Oooooo00?o0000ooooooooool0
C?ooool001oooooo00?o0000ooooooooool0;?ooool00ol0003oooooooooo`0kooooo`03o`000?oo
oooooooo01gooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0Gooooo`03o`000?oo
oooooooo04gooooo000Nooooo`03o`000?oooooooooo02kooooo0_l0000kooooo`03o`000?oooooo
oooo01kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Eooooo`;o0000D?ooool0
01cooooo0_l0000cooooo`03o`000?oooooooooo03Sooooo00?o0000ooooooooool07oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool0D?ooool001_o
oooo00?o0000ooooooooool0=?ooool00ol0003oooooooooo`0gooooo`03o`000?oooooooooo023o
oooo0_l00004ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool0DOooool001[ooooo
00?o0000ooooooooool0=_ooool00ol0003oooooooooo`0fooooo`03o`000?oooooooooo02;ooooo
00Go0000ooooooooooooooooo`00000Eooooo`03o`000?oooooooooo05;ooooo000Iooooo`03o`00
0?oooooooooo03Sooooo0_l0000gooooo`03o`000?oooooooooo02;ooooo00?o0000ooooool00000
5?ooool2o`0005Gooooo000Gooooo`;o0000?Oooool00ol0003oooooooooo`0dooooo`03o`000?oo
oooooooo02?ooooo0_l0000Cooooo`03o`000?oooooooooo05Gooooo000Fooooo`03o`000?oooooo
oooo03kooooo00?o0000ooooooooool0<oooool00ol0003oooooooooo`0Sooooo`?o00004Oooool0
0ol0003oooooooooo`1Fooooo`005Oooool00ol0003oooooooooo`10ooooo`03o`000?oooooooooo
03;ooooo00?o0000ooooooooool08oooool01?l0003oooooooooool0000?ooooo`03o`000?oooooo
oooo05Oooooo000Dooooo`03o`000?oooooooooo04;ooooo00?o0000ooooooooool0<_ooool00ol0
003oooooooooo`0Qooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool02oooool00ol0
003oooooooooo`1Hooooo`004oooool00ol0003oooooooooo`14ooooo`;o0000<_ooool00ol0003o
ooooooooo`0Qooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool02?ooool2o`0005_o
oooo000Aooooo`;o0000BOooool00ol0003oooooooooo`0_ooooo`03o`000?oooooooooo023ooooo
00?o0000ooooooooool01Oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo05_ooooo
000;ooooo`?o00000_ooool00ol0003oooooooooo`1:ooooo`03o`000?oooooooooo02kooooo00?o
0000ooooooooool08?ooool00ol0003oooooooooo`06ooooo`;o00001Oooool00ol0003ooooooooo
o`1<ooooo`?o00003Oooool000[ooooo1_l0001>ooooo`03o`000?oooooooooo02gooooo00?o0000
ooooooooool07oooool00ol0003oooooooooo`09ooooo`05o`000?ooooooooooooooool00000C_oo
ool5o`0000cooooo0009ooooo`Oo0000Coooool2o`0002kooooo00?o0000ooooooooool07_ooool0
0ol0003oooooooooo`0:ooooo`?o0000C_ooool7o`0000_ooooo0009ooooo`Oo0000DOooool00ol0
003oooooooooo`0[ooooo`03o`000?oooooooooo01gooooo00?o0000ooooooooool02oooool2o`00
04oooooo1ol0000;ooooo`002Oooool7o`0005;ooooo00?o0000ooooooooool0:_ooool00ol0003o
ooooooooo`0Mooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool00_l0001:ooooo`[o
00002oooool000[ooooo1Ol0001Dooooo`03o`000?oooooooooo02Wooooo00?o0000ooooooooool0
7?ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0
@_ooool5o`0000Cooooo1Ol0000<ooooo`002oooool3o`0005Kooooo0_l0000Zooooo`03o`000?oo
oooooooo01_ooooo00?o0000ooooooooool02?ooool2o`0000Oooooo00?o0000ooooooooool0?Ooo
ool4o`0000[ooooo0ol0000=ooooo`00I_ooool00ol0003oooooooooo`0Wooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`08ooooo`03o`000?oooooo
oooo03Sooooo1?l0000Nooooo`00Ioooool00ol0003oooooooooo`0Vooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0:ooooo`;o0000<oooool5o`00
02;ooooo001Xooooo`03o`000?oooooooooo02Gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0;?ooool4o`0002Oo
oooo001Yooooo`03o`000?oooooooooo02Cooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool09oooool4o`0002_ooooo
001Zooooo`;o00009Oooool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo00Cooooo0_l0
000Cooooo`03o`000?oooooooooo02;ooooo1?l0000_ooooo`00K?ooool00ol0003oooooooooo`0R
ooooo`03o`000?oooooooooo01Oooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0D
ooooo`03o`000?oooooooooo01cooooo1Ol0000cooooo`00KOooool00ol0003oooooooooo`0Qoooo
o`03o`000?oooooooooo01Oooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0Foooo
o`;o00006?ooool4o`0003Sooooo001^ooooo`03o`000?oooooooooo023ooooo00?o0000oooooooo
ool05_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool04Oooool4o`0003cooooo001_ooooo`;o00008Oooool00ol0003oooooooooo`0Eooooo`04o`00
0?ooooooooooooooo`;o00007Oooool00ol0003oooooooooo`0<ooooo`Co0000@?ooool0077ooooo
00?o0000ooooooooool07_ooool00ol0003oooooooooo`0Dooooo`05o`000?ooooooooooooooool0
00008?ooool00ol0003oooooooooo`06ooooo`Go0000A?ooool007;ooooo00?o0000ooooooooool0
7Oooool00ol0003oooooooooo`0Dooooo`04o`000?ooooooooooo`0002;ooooo0_l00002ooooo`Co
0000BOooool007?ooooo00?o0000ooooooooool07?ooool00ol0003oooooooooo`0Cooooo`04o`00
0?ooooooooooo`0002?ooooo1?l0001=ooooo`00M?ooool2o`0001cooooo00?o0000ooooooooool0
4oooool3o`0001oooooo1Ol00003ooooo`03o`000?oooooooooo04_ooooo001fooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool04Oooool2o`0001gooooo1?l00009ooooo`03o`000?oo
oooooooo04[ooooo001gooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool04Oooool0
0ol0003oooooooooo`0Hooooo`Co00003_ooool00ol0003oooooooooo`19ooooo`00N?ooool00ol0
003oooooooooo`0Hooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool05Oooool4o`00
01?ooooo0_l00019ooooo`00NOooool00ol0003oooooooooo`0Gooooo`03o`000?oooooooooo00oo
oooo0_l0000Booooo`Go00006Oooool00ol0003oooooooooo`16ooooo`00N_ooool2o`0001Sooooo
00?o0000ooooooooool03?ooool3o`0000oooooo1?l0000Oooooo`03o`000?oooooooooo04Gooooo
001looooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool02oooool01?l0003ooooooooo
ool0000;ooooo`Co00009?ooool00ol0003oooooooooo`14ooooo`00OOooool00ol0003ooooooooo
o`0Dooooo`03o`000?oooooooooo00[ooooo00Go0000ooooooooooooooooo`000007ooooo`Co0000
:Oooool2o`0004Cooooo001nooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool02Ooo
ool01Ol0003ooooooooooooooooo000000?ooooo1Ol0000_ooooo`03o`000?oooooooooo047ooooo
001oooooo`;o00004oooool00ol0003oooooooooo`07ooooo`;o00001?ooool4o`0003Gooooo00?o
0000ooooooooool0@?ooool0087ooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo00Co0000>_ooool00ol0003oooooooooo`0oooooo`00P_ooool00ol0
003oooooooooo`0@ooooo`03o`000?oooooooooo00?ooooo1Ol00003ooooo`03o`000?oooooooooo
03Wooooo0_l0000oooooo`00Poooool00ol0003oooooooooo`0?ooooo`03o`000?oooooo000000Co
00001_ooool00ol0003oooooooooo`0looooo`03o`000?oooooooooo03cooooo0024ooooo`03o`00
0?oooooooooo00cooooo1?l00003ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0
?Oooool00ol0003oooooooooo`0kooooo`00QOooool2o`0000Sooooo1?l00003ooooo`04o`000?oo
oooo0000o`0000Oooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`0jooooo`00Qooo
ool00ol0003oooooooooo`05o`0000Oooooo0_l00009ooooo`03o`000?oooooooooo043ooooo00?o
0000ooooooooool0>Oooool008Kooooo1?l0000<ooooo`03o`000?oooooooooo00Oooooo00?o0000
ooooooooool0@_ooool2o`0003Wooooo0022ooooo`Co00000oooool00ol0003oooooooooo`09oooo
o`;o00002Oooool00ol0003oooooooooo`14ooooo`03o`000?oooooooooo03Kooooo001mooooo`Go
00002?ooool2o`0000Oooooo0_l000000ooooooo0000ooooo`07ooooo`03o`000?oooooooooo04Ko
oooo00?o0000ooooooooool0=Oooool007Wooooo1?l0000?ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo04Oooooo
00?o0000ooooooooool0=?ooool007Gooooo1?l0000Dooooo`05o`000?ooooooooooooooool00000
1Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo04Wooooo0_l0000dooooo`00LOoo
ool4o`0001Wooooo00?o0000ooooool000001_ooool00ol0003oooooooooo`04ooooo`03o`000?oo
oooooooo04_ooooo00?o0000ooooooooool0<Oooool006cooooo1Ol0000Mooooo`?o00001_ooool0
0ol0003oooooooooo`03ooooo`03o`000?oooooooooo04gooooo00?o0000ooooooooool0<?ooool0
06Sooooo1?l0000Qooooo`05o`000?ooooooooooooooool000001_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo04kooooo00?o0000ooooooooool0;oooool006Cooooo1?l0000Toooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02oooo
o`03o`000?oooooooooo04oooooo0_l0000_ooooo`00H?ooool4o`0002Oooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`02ooooo`05o`000?ooooooooooooooool00000E?ooool00ol0
003oooooooooo`0/ooooo`00Foooool5o`0002[ooooo00?o0000ooooooooool01oooool01Ol0003o
oooooooooooooooo000000?ooooo00?o0000ooooooooool0Doooool00ol0003oooooooooo`0[oooo
o`00Eoooool4o`0002gooooo0_l0000;ooooo`;o000000Goooooo`000?ooooooooooo`00001Goooo
o`03o`000?oooooooooo02[ooooo001Cooooo`Co0000<?ooool00ol0003oooooooooo`0=ooooo`05
o`000?oooooo0000ooooool00000F?ooool00ol0003oooooooooo`0Yooooo`00C_ooool5o`0003?o
oooo00?o0000ooooooooool03oooool3o`0005[ooooo0_l0000Yooooo`00B_ooool4o`0003Oooooo
00?o0000ooooooooool04Oooool2o`0005cooooo00?o0000ooooooooool09_ooool004Kooooo1?l0
000iooooo`;o00005?ooool3o`0005cooooo00?o0000ooooooooool09Oooool004;ooooo1?l0000l
ooooo`03o`000?oooooooooo01Cooooo0_l000000ooooooo0000ooooo`1Kooooo`03o`000?oooooo
oooo02Cooooo000mooooo`Go0000?oooool00ol0003oooooooooo`0Dooooo`03o`000?oooooo0000
00;ooooo00?o0000ooooooooool0F_ooool2o`0002Cooooo000iooooo`Co0000@oooool00ol0003o
ooooooooo`0Eooooo`03o`000?oooooo000000?ooooo00?o0000ooooooooool0Foooool00ol0003o
ooooooooo`0Qooooo`00=Oooool4o`0004Gooooo0_l0000Gooooo`04o`000?ooooooooooo`0000Co
oooo0_l0001Looooo`03o`000?oooooooooo023ooooo000aooooo`Co0000B?ooool00ol0003ooooo
ooooo`0Gooooo`04o`000?ooooooooooo`0000Kooooo00?o0000ooooooooool0F_ooool00ol0003o
ooooooooo`0Oooooo`00;?ooool5o`0004_ooooo00?o0000ooooooooool05oooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0F_ooool00ol0003ooooo
ooooo`0Nooooo`00:?ooool4o`0004oooooo00?o0000ooooooooool06?ooool00ol0003ooooooooo
o`02ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0F_ooool2o`0001kooooo000H
ooooo`?o00002Oooool4o`0005;ooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool0Foooool4o`0001[ooooo000Goooo
o`Go00000oooool5o`0005Cooooo0_l0000Kooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool01oooool2o`0005_ooooo1Ol0000Iooooo`005_ooool9o`0005Sooooo00?o0000oooooooo
ool06oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Sooooo00?o0000oooooooo
ool0Eoooool7o`0001Sooooo000Fooooo`Oo0000FOooool00ol0003oooooooooo`0Kooooo`03o`00
0?oooooooooo00Gooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`1Fooooo`Oo0000
6?ooool001Kooooo1ol0001Hooooo`03o`000?oooooooooo01cooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo05Gooooo1ol0000Hooooo`005oooool5
o`0005Oooooo0_l0000Nooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02oooool2
o`0005Kooooo1Ol0000Iooooo`006?ooool3o`0005Oooooo00?o0000ooooooooool07_ooool00ol0
003oooooooooo`06ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0E?ooool3o`00
01[ooooo001aooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool02?ooool00ol0003o
ooooooooo`0=ooooo`03o`000?oooooooooo073ooooo001`ooooo`03o`000?oooooooooo01oooooo
00?o0000ooooooooool02?ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo06oooooo
001_ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`0?ooooo`03o`000?oooooooooo06kooooo001]ooooo`;o00008_ooool00ol0003oooooooooo`09
ooooo`03o`000?oooooooooo013ooooo0_l0001^ooooo`00K?ooool00ol0003oooooooooo`0Qoooo
o`03o`000?oooooooooo00_ooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`1[oooo
o`00Joooool00ol0003oooooooooo`0Rooooo`03o`000?oooooooooo00_ooooo00?o0000oooooooo
ool04_ooool00ol0003oooooooooo`1Zooooo`00J_ooool00ol0003oooooooooo`0Rooooo`03o`00
0?oooooooooo00cooooo00?o0000ooooooooool04oooool00ol0003oooooooooo`1Yooooo`00J?oo
ool2o`0002Gooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0Dooooo`;o0000JOoo
ool006Oooooo00?o0000ooooooooool09?ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooo
oooo01Gooooo00?o0000ooooooooool0I_ooool006Kooooo00?o0000ooooooooool09Oooool00ol0
003oooooooooo`0>ooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool0IOooool006Go
oooo00?o0000ooooooooool09Oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo01Oo
oooo00?o0000ooooooooool0I?ooool006?ooooo0_l0000Xooooo`03o`000?oooooooooo00oooooo
00?o0000ooooooooool06?ooool2o`0006Cooooo001Rooooo`03o`000?oooooooooo02Sooooo00?o
0000ooooooooool03oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo067ooooo001Q
ooooo`03o`000?oooooooooo02Sooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo063ooooo001Pooooo`03o`000?oooooooooo02Wooooo00?o0000oooo
ooooool04Oooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo05oooooo001Oooooo`03
o`000?oooooooooo02Wooooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`0Looooo`03
o`000?oooooooooo05kooooo001Mooooo`;o0000;?ooool00ol0003oooooooooo`0Booooo`03o`00
0?oooooooooo01gooooo0_l0001Nooooo`00G?ooool00ol0003oooooooooo`0[ooooo`03o`000?oo
oooooooo01Cooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`1Kooooo`00Foooool0
0ol0003oooooooooo`0/ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool07oooool0
0ol0003oooooooooo`1Jooooo`00F_ooool00ol0003oooooooooo`0/ooooo`03o`000?oooooooooo
01Gooooo00?o0000ooooooooool08?ooool00ol0003oooooooooo`1Iooooo`00F?ooool2o`0002oo
oooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`0Qooooo`;o0000FOooool005Oooooo
00?o0000ooooooooool0;_ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo02?ooooo
00?o0000ooooooooool0E_ooool005Kooooo00?o0000ooooooooool0;oooool00ol0003ooooooooo
o`0Gooooo`03o`000?oooooooooo02?ooooo00?o0000ooooooooool0EOooool005Gooooo00?o0000
ooooooooool0;oooool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo02Cooooo00?o0000
ooooooooool0E?ooool005?ooooo0_l0000booooo`03o`000?oooooooooo01Sooooo00?o0000oooo
ooooool09Oooool2o`0005Cooooo001Booooo`03o`000?oooooooooo037ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo057ooooo001Aooooo`03o`00
0?oooooooooo03;ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Wooooo`03o`00
0?oooooooooo053ooooo001@ooooo`03o`000?oooooooooo03?ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Xooooo`03o`000?oooooooooo04oooooo001?ooooo`03o`000?oooooo
oooo03?ooooo00?o0000ooooooooool06oooool00ol0003oooooooooo`0Yooooo`03o`000?oooooo
oooo04kooooo001=ooooo`;o0000=_ooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo
02[ooooo0_l0001>ooooo`00C?ooool00ol0003oooooooooo`0eooooo`03o`000?oooooooooo01co
oooo00?o0000ooooooooool0;?ooool00ol0003oooooooooo`1;ooooo`00Boooool00ol0003ooooo
ooooo`0fooooo`03o`000?oooooooooo01gooooo00?o0000ooooooooool0;?ooool00ol0003ooooo
ooooo`1:ooooo`00B_ooool00ol0003oooooooooo`0fooooo`03o`000?oooooooooo01kooooo00?o
0000ooooooooool0;Oooool00ol0003oooooooooo`19ooooo`00B?ooool2o`0003Wooooo00?o0000
ooooooooool07_ooool00ol0003oooooooooo`0^ooooo`;o0000BOooool004Oooooo00?o0000oooo
ooooool0>?ooool00ol0003oooooooooo`0Oooooo`03o`000?oooooooooo033ooooo00?o0000oooo
ooooool0A_ooool004Kooooo00?o0000ooooooooool0>Oooool00ol0003oooooooooo`0Pooooo`03
o`000?oooooooooo033ooooo00?o0000ooooooooool0AOooool004Gooooo00?o0000ooooooooool0
>Oooool00ol0003oooooooooo`0Qooooo`03o`000?oooooooooo037ooooo00?o0000ooooooooool0
A?ooool004Cooooo00?o0000ooooooooool0>_ooool00ol0003oooooooooo`0Qooooo`03o`000?oo
oooooooo03;ooooo00?o0000ooooooooool0@oooool004;ooooo0_l0000looooo`03o`000?oooooo
oooo02;ooooo00?o0000ooooooooool0<oooool2o`0004?ooooo000mooooo`?o000000?oooooo`00
0?ooool0?Oooool00ol0003oooooooooo`0Rooooo`03o`000?oooooooooo03Gooooo00?o0000oooo
ool000000_l0000nooooo`00??ooool5o`0003kooooo00?o0000ooooooooool09?ooool00ol0003o
ooooooooo`0eooooo`Go0000?Oooool003_ooooo1ol0000mooooo`03o`000?oooooooooo02Cooooo
00?o0000ooooooooool0=?ooool7o`0003cooooo000kooooo`Oo0000??ooool00ol0003ooooooooo
o`0Uooooo`03o`000?oooooooooo03Cooooo1ol0000looooo`00>oooool7o`0003cooooo00?o0000
ooooooooool09Oooool00ol0003oooooooooo`0dooooo`Oo0000??ooool003cooooo1Ol0000moooo
o`03o`000?oooooooooo02Kooooo00?o0000ooooooooool0=?ooool5o`0003gooooo000mooooo`?o
0000?Oooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo03Gooooo0ol0000nooooo`00
OOooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo07Kooooo001looooo`03o`000?oo
oooooooo02Sooooo00?o0000ooooooooool0M_ooool007cooooo00?o0000ooooooooool0:?ooool0
0ol0003oooooooooo`1fooooo`00Noooool00ol0003oooooooooo`0Zooooo`03o`000?oooooooooo
07Gooooo001kooooo`03o`000?oooooooooo02[ooooo00?o0000ooooooooool0MOooool007[ooooo
00?o0000ooooooooool0:oooool00ol0003oooooooooo`1eooooo`00N_ooool00ol0003ooooooooo
o`0[ooooo`03o`000?oooooooooo07Gooooo001iooooo`03o`000?oooooooooo02gooooo00?o0000
ooooooooool0M?ooool007Wooooo00?o0000ooooooooool0;Oooool00ol0003oooooooooo`1doooo
o`00N?ooool00ol0003oooooooooo`0^ooooo`03o`000?oooooooooo07Cooooo001hooooo`03o`00
0?oooooooooo02kooooo00?o0000ooooooooool0M?ooool007Oooooo00?o0000ooooooooool0;ooo
ool00ol0003oooooooooo`1dooooo`00Moooool00ol0003oooooooooo`0`ooooo`03o`000?oooooo
oooo07?ooooo001fooooo`03o`000?oooooooooo037ooooo00?o0000ooooooooool0Loooool007Ko
oooo00?o0000ooooooooool0<Oooool00ol0003oooooooooo`1cooooo`00M_ooool00ol0003ooooo
ooooo`0aooooo`03o`000?oooooooooo07?ooooo001eooooo`03o`000?oooooooooo03?ooooo00?o
0000ooooooooool0L_ooool007Gooooo00?o0000ooooooooool0<oooool00ol0003oooooooooo`1b
ooooo`00M?ooool00ol0003oooooooooo`0dooooo`03o`000?oooooooooo07;ooooo001aooooo`Co
0000=_ooool3o`0007;ooooo001`ooooo`Go0000=Oooool5o`00077ooooo001_ooooo`Oo0000<ooo
ool7o`00073ooooo001_ooooo`Oo0000<oooool7o`00073ooooo001_ooooo`Oo0000<oooool7o`00
073ooooo001`ooooo`Go0000=Oooool5o`00077ooooo001aooooo`?o0000=oooool3o`0007;ooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo
0000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500054, -0.0374693, \
0.00383279, 0.00374543}},
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1152}, {0, 701}},
WindowSize->{520, 578},
WindowMargins->{{Automatic, 171}, {36, Automatic}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:1"->{
    Cell[1776, 53, 86, 1, 38, "ObjectName",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[1865, 56, 384, 10, 48, "Usage",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2252, 68, 528, 14, 56, "Notes",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2805, 86, 78, 1, 38, "ExampleSection",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2908, 91, 150, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[3061, 96, 76354, 1110, 294, 2427, 192, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79430, 1209, 173, 4, 40, "Text",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79628, 1217, 132, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79763, 1222, 263, 5, 47, "Message",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80041, 1230, 159, 4, 24, "Text",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80225, 1238, 150, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80378, 1243, 31624, 458, 294, 1094, 75, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:1", "LocalInteractions:3"}]},
  "LocalInteractions:3"->{
    Cell[1776, 53, 86, 1, 38, "ObjectName",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[1865, 56, 384, 10, 48, "Usage",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2252, 68, 528, 14, 56, "Notes",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2805, 86, 78, 1, 38, "ExampleSection",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[2908, 91, 150, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[3061, 96, 76354, 1110, 294, 2427, 192, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79430, 1209, 173, 4, 40, "Text",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79628, 1217, 132, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[79763, 1222, 263, 5, 47, "Message",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80041, 1230, 159, 4, 24, "Text",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80225, 1238, 150, 3, 26, "Input",
      CellTags->{"b:1", "LocalInteractions:3"}],
    Cell[80378, 1243, 31624, 458, 294, 1094, 75, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:1", "LocalInteractions:3"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:1", 112662, 1722},
  {"LocalInteractions:3", 113898, 1749}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 86, 1, 38, "ObjectName",
  CellTags->{"b:1", "LocalInteractions:3"}],
Cell[1865, 56, 384, 10, 48, "Usage",
  CellTags->{"b:1", "LocalInteractions:3"}],
Cell[2252, 68, 528, 14, 56, "Notes",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{
Cell[2805, 86, 78, 1, 38, "ExampleSection",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{
Cell[2908, 91, 150, 3, 26, "Input",
  CellTags->{"b:1", "LocalInteractions:3"}],
Cell[3061, 96, 76354, 1110, 294, 2427, 192, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]],
Cell[79430, 1209, 173, 4, 40, "Text",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{
Cell[79628, 1217, 132, 3, 26, "Input",
  CellTags->{"b:1", "LocalInteractions:3"}],
Cell[79763, 1222, 263, 5, 47, "Message",
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]],
Cell[80041, 1230, 159, 4, 24, "Text",
  CellTags->{"b:1", "LocalInteractions:3"}],

Cell[CellGroupData[{
Cell[80225, 1238, 150, 3, 26, "Input",
  CellTags->{"b:1", "LocalInteractions:3"}],
Cell[80378, 1243, 31624, 458, 294, 1094, 75, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:1", "LocalInteractions:3"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

