(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    149449,       2767]*)
(*NotebookOutlinePosition[    152757,       2850]*)
(*  CellTagsIndexPosition[    152645,       2843]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["ConnectedCaveman", "ObjectName",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[TextData[{
  "\[FilledSmallSquare] ",
  Cell[BoxData[
      FormBox[
        RowBox[{"ConnectedCaveman", "[", Cell[TextData[Cell[BoxData[
              FormBox[
                StyleBox[\(n\_i, \[Ellipsis], n\_k\),
                  FontFamily->"Times"], TraditionalForm]]]]]}], 
        TraditionalForm]],
    FontFamily->"Courier"],
  StyleBox["]",
    FontFamily->"Courier"],
  " constructs a \"connected caveman\" graph with cave sizes ",
  Cell[BoxData[
      \(TraditionalForm\`n\_1, \[Ellipsis], n\_k\)]],
  ".\n\[FilledSmallSquare] ",
  StyleBox["ConnectedCaveman[",
    FontFamily->"Courier"],
  Cell[BoxData[
      \(TraditionalForm\`{n\_i, \[Ellipsis], n\_k}\)]],
  StyleBox["]",
    FontFamily->"Courier"],
  " is an alternate form that also constructs a \"connected caveman\" graph \
with cave sizes ",
  Cell[BoxData[
      \(TraditionalForm\`n\_1, \[Ellipsis], n\_k\)]],
  ". "
}], "Usage",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[TextData[{
  "\[FilledSmallSquare] The ",
  Cell[BoxData[
      \(TraditionalForm\`i\^th\)]],
  "cave is a clique of size ",
  Cell[BoxData[
      \(TraditionalForm\`n\_i\)]],
  ". Each cave has a unique vertex which is connected to the unique vertex \
from other caves, forming a cycle of size ",
  Cell[BoxData[
      \(TraditionalForm\`k\)]],
  ".\n\[FilledSmallSquare] The option ",
  StyleBox["Radius",
    FontFamily->"Courier"],
  " specifies the size of the central cycle. "
}], "Notes",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{

Cell["Examples", "ExampleSection",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ConnectedCaveman[3, 7, 5, 3, 6] // ShowGraph;\)\)], "Input",
  CellLabel->"In[23]:=",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0324019 0.921522 0.0479513 0.959027 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.53129 .70664 m
.71044 .87452 L
s
.53129 .70664 m
.48117 .95205 L
s
.53129 .70664 m
.29787 .62771 L
s
.53129 .70664 m
.67555 .5 L
s
.71044 .87452 m
.48117 .95205 L
s
.29787 .62771 m
.31943 .75139 L
s
.29787 .62771 m
.23997 .84605 L
s
.29787 .62771 m
.11931 .84041 L
s
.29787 .62771 m
.04831 .73872 L
s
.29787 .62771 m
.08045 .61755 L
s
.29787 .62771 m
.19151 .56815 L
s
.29787 .62771 m
.29787 .37229 L
s
.31943 .75139 m
.23997 .84605 L
s
.31943 .75139 m
.11931 .84041 L
s
.31943 .75139 m
.04831 .73872 L
s
.31943 .75139 m
.08045 .61755 L
s
.31943 .75139 m
.19151 .56815 L
s
.23997 .84605 m
.11931 .84041 L
s
.23997 .84605 m
.04831 .73872 L
s
.23997 .84605 m
.08045 .61755 L
s
.23997 .84605 m
.19151 .56815 L
s
.11931 .84041 m
.04831 .73872 L
s
.11931 .84041 m
.08045 .61755 L
s
.11931 .84041 m
.19151 .56815 L
s
.04831 .73872 m
.08045 .61755 L
s
.04831 .73872 m
.19151 .56815 L
s
.08045 .61755 m
.19151 .56815 L
s
.29787 .37229 m
.14225 .42491 L
s
.29787 .37229 m
.04608 .28715 L
s
.29787 .37229 m
.14225 .14939 L
s
.29787 .37229 m
.29787 .20201 L
s
.29787 .37229 m
.53129 .29336 L
s
.14225 .42491 m
.04608 .28715 L
s
.14225 .42491 m
.14225 .14939 L
s
.14225 .42491 m
.29787 .20201 L
s
.04608 .28715 m
.14225 .14939 L
s
.04608 .28715 m
.29787 .20201 L
s
.14225 .14939 m
.29787 .20201 L
s
.53129 .29336 m
.48117 .04795 L
s
.53129 .29336 m
.71044 .12548 L
s
.53129 .29336 m
.67555 .5 L
s
.48117 .04795 m
.71044 .12548 L
s
.67555 .5 m
.74514 .37456 L
s
.67555 .5 m
.88433 .37456 L
s
.67555 .5 m
.95392 .5 L
s
.67555 .5 m
.88433 .62544 L
s
.67555 .5 m
.74514 .62544 L
s
.74514 .37456 m
.88433 .37456 L
s
.74514 .37456 m
.95392 .5 L
s
.74514 .37456 m
.88433 .62544 L
s
.74514 .37456 m
.74514 .62544 L
s
.88433 .37456 m
.95392 .5 L
s
.88433 .37456 m
.88433 .62544 L
s
.88433 .37456 m
.74514 .62544 L
s
.95392 .5 m
.88433 .62544 L
s
.95392 .5 m
.74514 .62544 L
s
.88433 .62544 m
.74514 .62544 L
s
.025 w
.53129 .70664 Mdot
.71044 .87452 Mdot
.48117 .95205 Mdot
.29787 .62771 Mdot
.31943 .75139 Mdot
.23997 .84605 Mdot
.11931 .84041 Mdot
.04831 .73872 Mdot
.08045 .61755 Mdot
.19151 .56815 Mdot
.29787 .37229 Mdot
.14225 .42491 Mdot
.04608 .28715 Mdot
.14225 .14939 Mdot
.29787 .20201 Mdot
.53129 .29336 Mdot
.48117 .04795 Mdot
.71044 .12548 Mdot
.67555 .5 Mdot
.74514 .37456 Mdot
.88433 .37456 Mdot
.95392 .5 Mdot
.88433 .62544 Mdot
.74514 .62544 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[23]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00ROooool3o`0009Cooooo0028ooooo`Go0000Toooool008Oooooo1ol0002Booooo`00Qooo
ool7o`0009;ooooo0027ooooo`So0000TOooool008Sooooo1Ol00002ooooo`?o0000S_ooool008Wo
oooo0ol00006ooooo`?o0000Roooool008_ooooo00?o0000ooooooooool01oooool3o`0008Sooooo
002;ooooo`03o`000?oooooooooo00[ooooo0ol00025ooooo`00Roooool00ol0003oooooooooo`0=
ooooo`;o0000Poooool008_ooooo00?o0000ooooooooool03oooool3o`00083ooooo002;ooooo`03
o`000?oooooooooo01;ooooo0ol0001mooooo`00S?ooool00ol0003oooooooooo`0Dooooo`?o0000
N_ooool008cooooo00?o0000ooooooooool05oooool3o`0007Oooooo002<ooooo`03o`000?oooooo
oooo01[ooooo0ol0001dooooo`00S?ooool00ol0003oooooooooo`0Mooooo`;o0000L_ooool008co
oooo00?o0000ooooooooool07oooool3o`0006oooooo002=ooooo`03o`000?oooooooooo027ooooo
0ol0001/ooooo`00SOooool00ol0003oooooooooo`0Tooooo`?o0000JOooool008gooooo00?o0000
ooooooooool09oooool3o`0006Kooooo002=ooooo`03o`000?oooooooooo02[ooooo0_l0001Toooo
o`00SOooool00ol0003oooooooooo`0/ooooo`?o0000HOooool008kooooo00?o0000ooooooooool0
;_ooool3o`0005kooooo002>ooooo`03o`000?oooooooooo037ooooo0ol00005ooooo`?o0000Dooo
ool008kooooo00?o0000ooooooooool0=?ooool3o`000003ooooool0003o000000?o0000D_ooool0
08kooooo00?o0000ooooooooool0=oooool7o`00057ooooo002>ooooo`03o`000?oooooooooo03Oo
oooo1ol0001Aooooo`00Soooool00ol0003oooooooooo`0fooooo`Oo0000DOooool008oooooo00?o
0000ooooooooool0=_ooool6o`0005;ooooo000Wooooo`?o0000IOooool00ol0003oooooooooo`0e
ooooo`03o`000?oooooooooo00?o0000Doooool002Kooooo1Ol0001Tooooo`03o`000?oooooooooo
03Cooooo00?o0000ooooooooool0Eoooool002Gooooo1ol0001Sooooo`03o`000?oooooooooo03?o
oooo00?o0000ooooooooool0F?ooool002Gooooo1ol0001Tooooo`03o`000?oooooooooo037ooooo
00?o0000ooooooooool0FOooool002Gooooo2Ol0001Rooooo`03o`000?oooooooooo033ooooo00?o
0000ooooooooool0F_ooool002Kooooo1Ol00003ooooo`?o0000Goooool00ol0003oooooooooo`0_
ooooo`03o`000?oooooooooo05_ooooo000Vooooo`Go00001_ooool3o`0005cooooo00?o0000oooo
ooooool0;_ooool00ol0003oooooooooo`1Looooo`009_ooool01Ol0003oooooo`000?oooooo0000
00Wooooo0ol0001Iooooo`03o`000?oooooooooo02gooooo00?o0000ooooooooool0GOooool002Go
oooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`09ooooo`?o0000Eoooool00ol0
003oooooooooo`0[ooooo`03o`000?oooooooooo05kooooo000Tooooo`05o`000?oooooooooooooo
ool000000oooool00ol0003oooooooooo`0;ooooo`?o0000E?ooool00ol0003oooooooooo`0Zoooo
o`03o`000?oooooooooo05oooooo000Tooooo`05o`000?ooooooooooooooool000001?ooool00ol0
003oooooooooo`0=ooooo`?o0000DOooool00ol0003oooooooooo`0Yooooo`03o`000?oooooooooo
063ooooo000Sooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`0@ooooo`?o0000C_ooool00ol0003oooooooooo`0Xooooo`03o`000?oooooooooo067o
oooo000Rooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003ooooo
ooooo`0Booooo`?o0000Boooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo06;ooooo
000Rooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`0Dooooo`?o0000BOooool00ol0003oooooooooo`0Tooooo`;o0000IOooool0027ooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01Oooooo0ol00016
ooooo`03o`000?oooooooooo02?ooooo00?o0000ooooooooool0IOooool0023ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01Wooooo0ol00013oooo
o`03o`000?oooooooooo02;ooooo00?o0000ooooooooool0I_ooool0023ooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo01_ooooo0ol00002ooooo`?o
0000>oooool00ol0003oooooooooo`0Qooooo`03o`000?oooooooooo06Oooooo000Oooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Nooooo`Ko0000
>_ooool00ol0003oooooooooo`0Pooooo`03o`000?oooooooooo06Sooooo000Nooooo`03o`000?oo
oooooooo00Oooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0Mooooo`Oo0000>_oo
ool00ol0003oooooooooo`0Nooooo`03o`000?oooooooooo06Wooooo000Nooooo`03o`000?oooooo
oooo00Oooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Kooooo`So0000>_ooool0
0ol0003oooooooooo`0Mooooo`03o`000?oooooooooo06[ooooo000Mooooo`03o`000?oooooooooo
00Sooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Gooooo`?o000000?oooooo`00
0?l000001Ol0000jooooo`03o`000?oooooooooo01cooooo00?o0000ooooooooool0Joooool001co
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo01Co
oooo0ol00005ooooo`Go0000>oooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo06co
oooo000Looooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool02_ooool00ol0003ooooo
ooooo`0@ooooo`?o00002?ooool4o`0003cooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`1]ooooo`006oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00_ooooo00?o
0000ooooooooool03?ooool3o`0000[ooooo00Co0000oooooooooooo0000?_ooool00ol0003ooooo
ooooo`0Hooooo`03o`000?oooooooooo06kooooo000Jooooo`03o`000?oooooooooo00_ooooo00?o
0000ooooooooool02oooool00ol0003oooooooooo`09ooooo`?o00003?ooool01Ol0003ooooooooo
oooooooo000003kooooo00?o0000ooooooooool05oooool00ol0003oooooooooo`1_ooooo`006Ooo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool01Ooo
ool3o`0000kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0looooo`03o`000?oo
oooooooo01Kooooo00?o0000ooooooooool0L?ooool001Wooooo00?o0000ooooooooool03?ooool0
0ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00;ooooo0_l0000Aooooo`03o`000?oooooo
oooo00;ooooo00?o0000ooooooooool0??ooool00ol0003oooooooooo`0Eooooo`03o`000?oooooo
oooo077ooooo000Hooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool03_ooool4o`00
01;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0looooo`03o`000?oooooooooo
01Cooooo00?o0000ooooooooool0L_ooool001Oooooo00?o0000ooooooooool03_ooool00ol0003o
ooooooooo`0<ooooo`?o00005?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo03go
oooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`1cooooo`005oooool00ol0003ooooo
ooooo`0>ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`03o`000?oooooooooo017ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0mooooo`03o`000?oooooooooo013ooooo
0_l0001fooooo`005_ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00Kooooo0ol0
0007ooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`0mooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool0M_ooool001Gooooo00?o0000
ooooooooool04?ooool00ol0003oooooooooo`03ooooo`?o00002_ooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool0?Oooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo07Oooooo000Eooooo`03o`000?oooooooooo013ooooo00?o0000oooo
ooooool00ol0000>ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01_ooool00ol0
003oooooooooo`0mooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0N?ooool001Co
oooo00?o0000ooooooooool04Oooool3o`0001;ooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo03kooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`1iooooo`004oooool00ol0003oooooooooo`0?ooooo`Co00005?ooool00ol0003ooooo
ooooo`0:ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0?_ooool00ol0003ooooo
ooooo`0:ooooo`03o`000?oooooooooo07[ooooo000Cooooo`03o`000?oooooooooo00gooooo0_l0
0003ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool02?ooool00ol0003ooooooooo
o`09ooooo`03o`000?oooooooooo03kooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`1kooooo`004_ooool00ol0003oooooooooo`0;ooooo`?o00001Oooool00ol0003oooooooooo`0D
ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0n
ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0O?ooool0017ooooo00?o0000oooo
ooooool02Oooool3o`0000Sooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00[ooooo00?o0000ooooooooool0?_ooool00ol0003oooooooooo`07oooo
o`03o`000?oooooooooo07gooooo000Aooooo`03o`000?oooooooooo00Kooooo0ol0000;ooooo`03
o`000?oooooooooo01Gooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0;ooooo`03
o`000?oooooooooo03oooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`1nooooo`00
4?ooool00ol0003oooooooooo`04ooooo`?o00003_ooool00ol0003oooooooooo`0Fooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0oooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool0Ooooool000cooooo1?l00004ooooo`?o00004Ooo
ool00ol0003oooooooooo`0Gooooo`05o`000?ooooooooooooooool000003_ooool00ol0003ooooo
ooooo`0oooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0P?ooool000_ooooo1Ol0
00001?oooooo0000o`000?l0000Dooooo`03o`000?oooooooooo01Oooooo00Co0000oooooooooooo
00003oooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo00;ooooo00?o0000oooooooo
ool0POooool000[ooooo1ol0000Gooooo`03o`000?oooooooooo01Sooooo0_l0000@ooooo`03o`00
0?oooooooooo03oooooo0ol000000ooooooo0000ooooo`23ooooo`002_ooool7o`0001Oooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo03kooooo1Ol0
0025ooooo`002_ooool9o`0001Gooooo00?o0000ooooooooool06?ooool2o`00013ooooo00?o0000
ooooooooool0?Oooool7o`0008Cooooo000;ooooo`Go00000oooool3o`0001;ooooo00?o0000oooo
ooooool05oooool01?l0003oooooooooool0000?ooooo`03o`000?oooooooooo03gooooo1ol00024
ooooo`003?ooool4o`0000Kooooo0ol0000?ooooo`03o`000?oooooooooo01Oooooo00Go0000oooo
ooooooooooooo`00000>ooooo`03o`000?oooooooooo03_ooooo2Ol00024ooooo`003oooool00ol0
003oooooooooo`07ooooo`?o00003?ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0hooooo`?o00000oooool5o`00
08Gooooo000@ooooo`03o`000?oooooooooo00Wooooo0ol00009ooooo`03o`000?oooooooooo01Go
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo03Go
oooo0ol00007ooooo`Co0000QOooool0017ooooo00?o0000ooooooooool02oooool3o`0000Kooooo
00?o0000ooooooooool05Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[ooooo
00?o0000ooooooooool0<_ooool3o`0000gooooo00?o0000ooooooooool0Poooool001;ooooo00?o
0000ooooooooool03Oooool3o`0000?ooooo00?o0000ooooooooool05?ooool00ol0003ooooooooo
o`07ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool0;oooool3o`00017ooooo00?o
0000ooooooooool0P_ooool001;ooooo00?o0000ooooooooool04?ooool4o`0001Gooooo00?o0000
ooooooooool02?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo02cooooo0ol0000E
ooooo`03o`000?oooooooooo087ooooo000Cooooo`03o`000?oooooooooo01;ooooo0ol0000Boooo
o`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Yoooo
o`?o00006Oooool00ol0003oooooooooo`20ooooo`005?ooool00ol0003oooooooooo`0Aooooo`03
o`000?oooooooooo00?o00003oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool09_ooool3o`0001cooooo00?o0000ooooooooool0P?ooool001Cooooo
00?o0000ooooooooool04Oooool00ol0003oooooooooo`03ooooo`?o00002oooool00ol0003ooooo
ooooo`0=ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool08oooool3o`00023ooooo
00?o0000ooooooooool0Ooooool001Gooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`06ooooo`?o00001oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool08?ooool3o`0002Cooooo00?o0000ooooooooool0O_ooool001Kooooo00?o0000
ooooooooool03oooool00ol0003oooooooooo`09ooooo`?o00001?ooool00ol0003oooooooooo`0?
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool07Oooool3o`0002Oooooo00?o0000
ooooooooool0O_ooool001Kooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0<oooo
o`Co00004oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01[ooooo0ol0000[oooo
o`03o`000?oooooooooo07gooooo000Gooooo`03o`000?oooooooooo00kooooo00?o0000oooooooo
ool03_ooool4o`00017ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Gooooo`?o
0000;oooool00ol0003oooooooooo`1looooo`006?ooool00ol0003oooooooooo`0=ooooo`03o`00
0?oooooooooo00kooooo00Co0000oooooooooooooooo0ol0000?ooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool05?ooool3o`0003;ooooo00?o0000ooooooooool0O?ooool001Sooooo
00?o0000ooooooooool03Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00Gooooo
0ol0000=ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool04Oooool3o`0003Kooooo
00?o0000ooooooooool0Noooool001Wooooo00?o0000ooooooooool03?ooool00ol0003ooooooooo
o`0<ooooo`03o`000?oooooooooo00Wooooo0ol0000:ooooo`03o`000?oooooooooo00;ooooo00?o
0000ooooooooool03_ooool3o`0003[ooooo00?o0000ooooooooool0N_ooool001[ooooo00?o0000
ooooooooool02oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00gooooo0ol00008
ooooo`05o`000?ooooooooooooooool000003Oooool3o`0003gooooo00?o0000ooooooooool0N_oo
ool001_ooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooo
oooo013ooooo0ol00006ooooo`04o`000?ooooooooooo`0000[ooooo0ol00011ooooo`03o`000?oo
oooooooo07Wooooo000Kooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02_ooool0
0ol0003oooooooooo`0Dooooo`?o00001?ooool4o`0000Kooooo0ol00015ooooo`03o`000?oooooo
oooo07Sooooo000Looooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool02Oooool00ol0
003oooooooooo`0Hooooo`?o000000?oooooo`000?l000000ol00002ooooo`?o0000BOooool00ol0
003oooooooooo`0[ooooo`?o00009Oooool3o`00027ooooo000Mooooo`03o`000?oooooooooo00So
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Kooooo`So0000C?ooool00ol0003o
ooooooooo`0Zooooo`Go00008oooool5o`00023ooooo000Mooooo`03o`000?oooooooooo00Sooooo
00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Looooo`Oo0000C_ooool00ol0003ooooo
ooooo`0Xooooo`Oo00008Oooool7o`0001oooooo000Nooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool01oooool00ol0003oooooooooo`0Looooo`So0000Coooool00ol0003ooooooooo
o`0Woooooboo00007oooool001oooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo01[ooooo0ol00002ooooo`Go0000D?ooool00ol0003oooooooooo`0W
ooooo`Oo00008Oooool7o`0001oooooo000Oooooo`03o`000?oooooooooo00Kooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`0Gooooo`?o00001_ooool3o`0005;ooooo00?o0000oooo
ooooool09oooool5o`000003ooooool0003ooooo01kooooo00?o0000ooooool000001?l0000Poooo
o`008?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool05Oooool3o`0000[ooooo00?o0000ooooooooool0D_ooool00ol0003oooooooooo`0Wooooo`?o
00000oooool2o`0001_ooooo0_l00003ooooo`?o00008Oooool0027ooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01?ooooo0ol0000=ooooo`03o`00
0?oooooooooo05;ooooo00?o0000ooooooooool09_ooool01Ol0003oooooo`000?oooooo000000Co
oooo0_l0000Gooooo`;o00001?ooool01Ol0003oooooo`000?oooooo0000023ooooo000Qooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0@ooooo`?o
00004?ooool00ol0003oooooooooo`1Cooooo`03o`000?oooooooooo02Cooooo00Co0000oooooooo
oooo00000_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo01?ooooo00?o0000oooo
ooooool00oooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo01gooooo000Roooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>oooo
o`?o00004oooool00ol0003oooooooooo`1Dooooo`03o`000?oooooooooo02?ooooo00Co0000oooo
oooooooo00000_ooool00ol0003oooooooooo`04ooooo`;o00004Oooool2o`0000Kooooo00Co0000
oooooooooooo00000_ooool00ol0003oooooooooo`0Mooooo`008oooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool03?ooool3o`0001Kooooo00?o0000
ooooooooool0EOooool00ol0003oooooooooo`0Qooooo`05o`000?ooooooooooooooool000000ooo
ool00ol0003oooooooooo`05ooooo`;o00003Oooool2o`0000Oooooo00Go0000oooooooooooooooo
o`000003ooooo`03o`000?oooooooooo01cooooo000Tooooo`05o`000?ooooooooooooooool00000
1?ooool00ol0003oooooooooo`09ooooo`?o00006Oooool00ol0003oooooooooo`1Eooooo`03o`00
0?oooooooooo027ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`07ooooo`05o`000?oooooooooooooo
ool000000oooool00ol0003oooooooooo`0Looooo`009?ooool01Ol0003ooooooooooooooooo0000
00?ooooo00?o0000ooooooooool01oooool3o`0001cooooo00?o0000ooooooooool0E_ooool00ol0
003oooooooooo`0Oooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`07ooooo`;o00001oooool2o`0000Wooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo01_ooooo000Uooooo`04o`000?ooooooooooo`00
00;ooooo00?o0000ooooooooool01Oooool3o`0001oooooo00?o0000ooooooooool0Eoooool00ol0
003oooooooooo`0Nooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`09ooooo`;o00000oooool2o`0000_ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo01_ooooo000Vooooo`05o`000?oooooo0000oooo
ool000001Oooool3o`0002;ooooo00?o0000ooooooooool0Eoooool00ol0003oooooooooo`0Moooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooo000000cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo01[ooooo000Vooooo`Go00000_ooool3o`0002Gooooo00?o0000oooooooo
ool0F?ooool00ol0003oooooooooo`0Looooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool00oooool00ol0003oooooooooo`0:ooooo`?o00003?ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo00?ooooo00?o0000ooooooooool06_ooool002Kooooo1ol0000Xooooo`03o`00
0?oooooooooo05Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`04ooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool01oooool2o`0000?ooooo0_l00009ooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Iooooo`009Ooo
ool7o`0002Wooooo00?o0000ooooooooool0FOooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool06?ooool002Gooooo1ol0000Yooooo`03o`000?oooooo
oooo05[ooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool00oooool2o`0000Wooooo0_l00005ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0Hooooo`009Oooool7o`00
02Wooooo00?o0000ooooooooool0Foooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo
00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`02o`0000gooooo0_l00002oooo
o`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Goooo
o`009_ooool5o`0002[ooooo00?o0000ooooooooool0Foooool00ol0003oooooooooo`0Fooooo`03
o`000?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooo`00000Aooooo`03
o`000?oooooo000000Sooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Gooooo`00
9oooool3o`0002_ooooo00?o0000ooooooooool0G?ooool00ol0003oooooooooo`0Dooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool01_ooool2o`0001?ooooo0_l00008ooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool05_ooool005Gooooo00?o0000ooooooooool0GOoo
ool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01?oo
ool2o`000003ooooool0003ooooo01;ooooo00Co0000ooooool0003o00001_ooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo01Kooooo001Eooooo`03o`000?oooooooooo05kooooo00?o
0000ooooooooool04Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00?ooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00;ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo01Gooooo001E
ooooo`03o`000?oooooooooo05kooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`08
ooooo`04o`000?ooooooooooooooo`;o00001Oooool00ol0003oooooooooo`0?ooooo`03o`000?oo
oooooooo00?ooooo0_l00003ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool05Ooo
ool005Gooooo00?o0000ooooooooool0Goooool00ol0003oooooooooo`0?ooooo`03o`000?oooooo
oooo00Wooooo00Co0000ooooool0003o00002?ooool00ol0003oooooooooo`0=ooooo`03o`000?oo
oooooooo00Kooooo0_l000000ooooooo0000ooooo`0:ooooo`03o`000?oooooooooo01Cooooo001E
ooooo`03o`000?oooooooooo063ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0:
ooooo`;o00002oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Wooooo0_l0000<
ooooo`03o`000?oooooooooo01?ooooo001Eooooo`03o`000?oooooooooo063ooooo00?o0000oooo
ooooool03Oooool00ol0003oooooooooo`09ooooo`;o00003?ooool00ol0003oooooooooo`0;oooo
o`03o`000?oooooooooo00[ooooo0_l0000;ooooo`03o`000?oooooooooo01?ooooo001Eooooo`03
o`000?oooooooooo067ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`08ooooo`;o
000000?oooooo`000?ooool03?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00_o
oooo00Co0000ooooool0003o00002_ooool00ol0003oooooooooo`0Booooo`00EOooool00ol0003o
ooooooooo`1Rooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool01oooool01Ol0003o
oooooooooooooooo000000gooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0;oooo
o`05o`000?ooooooooooooooool000002Oooool00ol0003oooooooooo`0Booooo`00EOooool00ol0
003oooooooooo`1Rooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01_ooool2o`00
00Cooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo
00cooooo00?o0000ooooooooool00_ooool2o`0000Sooooo00?o0000ooooooooool04Oooool005Go
oooo00?o0000ooooooooool0Hoooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Co
oooo0_l00006ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool01oooool00ol0003o
ooooooooo`0<ooooo`03o`000?oooooooooo00Cooooo0_l00006ooooo`03o`000?oooooooooo017o
oooo001Eooooo`03o`000?oooooooooo06Cooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`05ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo013ooooo001Eooooo`03o`000?oooooooooo06Cooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`02ooooo`;o00002Oooool00ol0003ooooooooo
o`0=ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool00ol0003ooooooooo
o`07ooooo`;o00001?ooool00ol0003oooooooooo`0@ooooo`00EOooool00ol0003oooooooooo`1U
ooooo`03o`000?oooooooooo00Cooooo00Co0000oooooooooooooooo0_l0000;ooooo`03o`000?oo
oooooooo00kooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Wooooo0_l00003ooooo`03o`000?oooooooooo00oooooo001Eooooo`03o`000?oooooo
oooo06Kooooo00?o0000ooooooooool00_ooool01Ol0003ooooooooooooooooo000000gooooo00?o
0000ooooooooool03oooool01Ol0003ooooooooooooooooo0000017ooooo00?o0000ooooooooool0
2oooool01Ol0003ooooooooooooooooo0000013ooooo001Eooooo`03o`000?oooooooooo06Oooooo
1Ol000000ooooooo0000o`00000>ooooo`03o`000?oooooooooo00oooooo00Go0000oooooooooooo
ooooo`00000Aooooo`03o`000?oooooooooo00cooooo0_l000000ooooooo0000o`000002o`0000go
oooo001Eooooo`03o`000?oooooooooo06Oooooo1_l0000@ooooo`03o`000?oooooooooo013ooooo
00?o0000ooooool000004_ooool00ol0003oooooooooo`0>ooooo`Ko00003?ooool005Gooooo00?o
0000ooooooooool0I_ooool7o`00013ooooo00?o0000ooooooooool04?ooool00ol0003oooooo`00
000Booooo`03o`000?oooooooooo00kooooo1ol0000;ooooo`00EOooool00ol0003oooooooooo`1V
oooooeOo00002oooool005Gooooo00?o0000ooooooooool0I_ooool7o`00013ooooo00?o0000oooo
ooooool04Oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00kooooo1ol0000;oooo
o`00EOooool00ol0003oooooooooo`1Wooooo`Go000000?oooooo`000?ooool03_ooool00ol0003o
ooooooooo`0@ooooo`03o`000?oooooo000001;ooooo00?o0000ooooooooool03Oooool00ol0003o
ooooo`000004o`0000cooooo001Eooooo`03o`000?oooooooooo06Oooooo1?l00003ooooo`;o0000
3Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooo000001;ooooo00?o0000ooooooooool0
2oooool2o`0000?ooooo0ol0000=ooooo`00EOooool00ol0003oooooooooo`1Wooooo`05o`000?oo
ooooooooooooool000001?ooool2o`0000_ooooo00?o0000ooooooooool03oooool01Ol0003ooooo
oooooooooooo0000017ooooo00?o0000ooooooooool02Oooool2o`0000Cooooo00?o0000oooooooo
ool03_ooool005Gooooo00?o0000ooooooooool0I_ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0>ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`08ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool03oooool005Gooooo00?o0000ooooooooool0IOoo
ool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo0_l00008ooooo`03o`000?oo
oooooooo00kooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Kooooo0_l00006ooooo`03o`000?oooooooooo00oooooo001Eooooo`03o`000?oooooo
oooo06Cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`05ooooo`;o00001_ooool0
0ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool0
0ol0003oooooooooo`04ooooo`;o00001oooool00ol0003oooooooooo`0@ooooo`00EOooool00ol0
003oooooooooo`1Tooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01oooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`0=ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01oooool00ol0
003oooooooooo`0@ooooo`00EOooool00ol0003oooooooooo`1Sooooo`03o`000?oooooooooo00So
oooo00?o0000ooooooooool01oooool2o`0000?ooooo00?o0000ooooooooool03?ooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Co0000oooooooooooooooo0_l00009oooo
o`03o`000?oooooooooo017ooooo001Eooooo`03o`000?oooooooooo06;ooooo00?o0000oooooooo
ool02Oooool00ol0003oooooooooo`09ooooo`;o000000?oooooo`000?ooool03Oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Co0000ooooool0003o00002oooool00ol0
003oooooooooo`0Aooooo`00EOooool00ol0003oooooooooo`1Rooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool02_ooool2o`0000gooooo00?o0000ooooooooool02Oooool00ol0003o
ooooooooo`0;ooooo`;o00003?ooool00ol0003oooooooooo`0Booooo`00EOooool00ol0003ooooo
ooooo`1Qooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool02oooool2o`0000cooooo
00?o0000ooooooooool02Oooool00ol0003oooooooooo`0:ooooo`;o00003Oooool00ol0003ooooo
ooooo`0Booooo`00EOooool00ol0003oooooooooo`1Pooooo`03o`000?oooooooooo00gooooo00?o
0000ooooooooool02_ooool01?l0003oooooo`000?l00009ooooo`03o`000?oooooooooo00_ooooo
00?o0000ooooooooool01oooool2o`000003ooooool0003ooooo00_ooooo00?o0000ooooooooool0
4oooool005Gooooo00?o0000ooooooooool0H?ooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Wooooo00Go0000ooooooooooooooooo`000007ooooo`03o`000?oooooooooo00gooooo
00?o0000ooooooooool01Oooool01Ol0003ooooooooooooooooo000000_ooooo00?o0000oooooooo
ool05?ooool005Gooooo00?o0000ooooooooool0Goooool00ol0003oooooooooo`0?ooooo`03o`00
0?oooooooooo00Wooooo00?o0000ooooooooool00_ooool2o`0000Gooooo00?o0000ooooooooool0
3Oooool00ol0003oooooooooo`03ooooo`;o00001?ooool00ol0003oooooooooo`09ooooo`03o`00
0?oooooooooo01Cooooo001Eooooo`03o`000?oooooooooo05kooooo00?o0000ooooooooool04Ooo
ool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Cooooo0_l00002ooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool00_l00006ooooo`03o`000?oooooooooo00Sooooo00?o
0000ooooooooool05Oooool003Gooooo0ol0000Mooooo`03o`000?oooooooooo05kooooo00?o0000
ooooooooool04Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooool000004Oooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00Sooooo00?o0000
ooooooooool05Oooool003Cooooo1Ol0000Looooo`03o`000?oooooooooo05gooooo00?o0000oooo
ooooool04oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Oooooo0_l0000Aoooo
o`;o00002Oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo01Kooooo000cooooo`Oo
00006oooool00ol0003oooooooooo`1Looooo`03o`000?oooooooooo01Cooooo00?o0000oooooooo
ool01oooool00ol0003oooooooooo`07ooooo`04o`000?oooooo0000o`0000gooooo0_l000000ooo
oooo0000ooooo`08ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool05_ooool003?o
oooo1ol0000Kooooo`03o`000?oooooooooo05_ooooo00?o0000ooooooooool05_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`09ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo01Oooooo000cooooo`Oo00006oooool00ol0003ooooo
ooooo`1Kooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00?ooooo0_l00007ooooo`;o00001Oooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool05oooool0033ooooo0ol0
00000ooooooo0000o`000003o`000003ooooool0003o000001Wooooo00?o0000ooooooooool0F_oo
ool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`06ooooo`;o00000oooool2o`0000Sooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01Sooooo000^ooooo`;o00001?ooool5
o`0000?ooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`1Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00?o0000ooooool000002oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool06Oooool002cooooo0_l00005ooooo`04o`000?oooooo0000
ooooo`;o00001?ooool2o`0001Kooooo00?o0000ooooooooool0FOooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09
ooooo`?o00002oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool06Oooool002[ooooo0_l00006ooooo`08o`000?ooooooooooo`000?oooooo0000oooo
ool00005ooooo`;o00005?ooool00ol0003oooooooooo`1Hooooo`03o`000?oooooooooo01cooooo
00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Sooooo
0_l00003ooooo`;o00002_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool06_ooool002Sooooo0_l00007ooooo`07o`000?ooooooooooooooool0003o
ooooo`000002ooooo`03o`000?oooooooooo00Cooooo0_l0000Booooo`03o`000?oooooooooo05Oo
oooo00?o0000ooooooooool07Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo01[o
oooo000Uooooo`?o00002Oooool01?l0003oooooooooool00002ooooo`05o`000?oooooooooooooo
ool000001oooool2o`00013ooooo00?o0000ooooooooool0Eoooool00ol0003oooooooooo`0Noooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`06oooo
o`;o00002Oooool2o`0000Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02oooo
o`03o`000?oooooooooo01_ooooo000Sooooo`;o00002oooool01Ol0003ooooooooooooooooo0000
00;ooooo00Go0000ooooooooooooooooo`000009ooooo`;o00003_ooool00ol0003oooooooooo`1F
ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00Cooooo0_l0000=ooooo`;o00001_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool06oooool0027ooooo0_l0000<oooo
o`03o`000?oooooooooo00;ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooo
oooo00Sooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`1Eooooo`03o`000?oooooo
oooo027ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool03oooool00ol0003oooooooooo`04ooooo`05o`000?ooooooooooooooool00000
0oooool00ol0003oooooooooo`0Looooo`007oooool2o`0000gooooo00?o0000ooooooooool00ooo
ool01Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooooooool02?ooool2o`0000_ooooo
00?o0000ooooooooool0E?ooool00ol0003oooooooooo`0Rooooo`05o`000?ooooooooooooooool0
00000oooool00ol0003oooooooooo`02ooooo`;o00004oooool2o`0000Cooooo00Go0000oooooooo
ooooooooo`000003ooooo`03o`000?oooooooooo01cooooo000Fooooo`?o00001?ooool2o`0000ko
oooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00;o
oooo00?o0000ooooooooool02_ooool2o`0000Wooooo00?o0000ooooooooool0E?ooool00ol0003o
ooooooooo`0Sooooo`04o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooooooo0_l0000G
ooooo`;o00000oooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo01gooooo000E
ooooo`So00003oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`0;ooooo`;o00001oooool00ol0003oooooooooo`1C
ooooo`03o`000?oooooooooo02Gooooo00Go0000ooooool0003oooooo`000003ooooo`03o`000?oo
oooooooo01Wooooo00Wo0000ooooooooooooooooo`000?oooooo0000ooooool000008?ooool001Co
oooo1ol0000Aooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00cooooo0_l00005ooooo`03o`000?oooooooooo05;o
oooo00?o0000ooooooooool09_ooool5o`000003ooooool0003o000001gooooo0_l000000ooooooo
0000o`000003o`00023ooooo000Dooooo`ko00002Oooool00ol0003oooooooooo`05ooooo`03o`00
0?oooooooooo00;ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0=ooooo`04o`00
0?ooooooooooooooo`?o0000Doooool00ol0003oooooooooo`0Vooooo`Ko00008Oooool6o`00023o
oooo000Dooooo`Oo00001ooooolEo`0000;ooooo00?o0000ooooooooool01?ooool00ol0003ooooo
ooooo`0>ooooo`Oo0000DOooool00ol0003oooooooooo`0Vooooo`Oo00008Oooool7o`0001oooooo
000Eooooo`Go000000?oooooo`000?ooool03?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00;ooooo5Ol00006ooooo`Oo0000Coooool00ol0003oooooooooo`0Woooooboo00007ooo
ool001Kooooo1?l00002ooooo`;o00002_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`07ooooo`go0000Coooool0
0ol0003oooooooooo`0Wooooo`Oo00008Oooool7o`0001oooooo000Fooooo`04o`000?oooooo0000
o`0000Cooooo0_l00007ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01?ooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00gooooo2Ol0001<ooooo`03o`000?oooooo
oooo02Wooooo1Ol0000Sooooo`Go00008?ooool001Kooooo00Go0000ooooool0003oooooo`000005
ooooo`;o00001Oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`09ooooo`?o000000?oooooo`000?l000000ol00003
ooooo`?o0000B?ooool00ol0003oooooooooo`0[ooooo`?o00009Oooool3o`00027ooooo000Eoooo
o`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool2o`0000;ooooo00?o0000
ooooooooool02?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00?o0000
ooooooooool01_ooool2o`0000Cooooo1?l00007ooooo`?o0000AOooool00ol0003oooooooooo`1g
ooooo`005Oooool01?l0003oooooooooool00003ooooo`03o`000?oooooooooo00Gooooo0_l0000;
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`03
ooooo`;o00001Oooool01Ol0003oooooo`000?oooooo000000[ooooo0ol00011ooooo`03o`000?oo
oooooooo07Sooooo000Eooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool01Ooo
ool3o`0000Wooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`08ooooo`04o`000?oo
ooooooooooooo`;o00001_ooool01_l0003oooooooooool0003oooooo`0000gooooo0ol0000moooo
o`03o`000?oooooooooo07Wooooo000Eooooo`04o`000?ooooooooooo`0000Cooooo00?o0000oooo
ooooool00oooool01?l0003oooooooooooooool2o`0000Oooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`09ooooo`?o00001oooool01?l0003oooooooooool00002ooooo`03o`000?oo
oooooooo00kooooo0ol0000iooooo`03o`000?oooooooooo07[ooooo000Dooooo`03o`000?oooooo
oooo00;ooooo00?o0000ooooooooool00_ooool01Ol0003ooooooooooooooooo000000Kooooo0_l0
0005ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01oooool2o`000003ooooool0
003ooooo00Oooooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`0Aooooo`?o0000
=_ooool00ol0003oooooooooo`1jooooo`005?ooool00ol0003oooooooooo`02ooooo`03o`000?oo
oooooooo00;ooooo00Co0000oooooooooooo00002Oooool2o`0000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`05ooooo`;o00000oooool00ol0003oooooooooo`05ooooo`05o`00
0?ooooooooooooooool000000_ooool00ol0003oooooooooo`0Dooooo`?o0000<_ooool00ol0003o
ooooooooo`1kooooo`005?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooool000002oooool00ol0003oooooo`00000;ooooo`03o`000?oooooooooo00;ooooo
0_l00006ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00_ooool01?l0003ooooo
ooooool0000Iooooo`?o0000;_ooool00ol0003oooooooooo`1looooo`004oooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool2o`0000_ooooo
00?o0000ooooool000000_l00009ooooo`05o`000?ooooooooooooooool000001?ooool00ol0003o
ooooooooo`02ooooo`03o`000?oooooooooo01Wooooo0ol0000[ooooo`03o`000?oooooooooo07co
oooo000Cooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool2o`0000kooooo
0ol00009ooooo`;o00003Oooool00ol0003oooooo`000005ooooo`03o`000?oooooooooo00;ooooo
00?o0000ooooooooool07?ooool3o`0002Oooooo00?o0000ooooooooool0OOooool001?ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`0000gooooo
00?o0000ooooooooool00_l00005ooooo`?o00003_ooool2o`0000Kooooo00?o0000ooooooooool0
0_ooool00ol0003oooooooooo`0Oooooo`?o00008oooool00ol0003oooooooooo`1nooooo`004ooo
ool00ol0003oooooooooo`04ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`03ooooo`;o000000?oooooo`000?l000000_ooool00ol0003ooooo
ooooo`0=ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003ooooo
ooooo`0Rooooo`?o00008?ooool00ol0003oooooooooo`1nooooo`004_ooool00ol0003ooooooooo
o`05ooooo`03o`000?oooooo000000Gooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`03ooooo`?o00001?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooo000000Cooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`0Uooooo`?o00007?ooool00ol0003ooooooooo
o`1oooooo`004_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo000000Kooooo00?o0000
ooooooooool02?ooool01?l0003oooooooooooooool2o`0000?ooooo0_l00003ooooo`03o`000?oo
oooooooo00[ooooo00Co0000oooooooooooo00001?ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo02Sooooo0ol0000Hooooo`03o`000?oooooooooo083ooooo000Booooo`03o`000?oo
oooooooo00Gooooo0_l00008ooooo`03o`000?oooooooooo00Oooooo00Co0000ooooool0003o0000
1oooool2o`000003ooooool0003ooooo00[ooooo00?o0000ooooooooool00_ooool01Ol0003ooooo
oooooooooooo000000Kooooo00?o0000ooooooooool0:_ooool3o`0001Gooooo00?o0000oooooooo
ool0P?ooool0017ooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`08ooooo`03o`00
0?oooooooooo00Gooooo0ol0000;ooooo`;o00002_ooool00ol0003oooooooooo`04ooooo`03o`00
0?oooooo000000Oooooo00?o0000ooooooooool0;Oooool3o`00017ooooo00?o0000ooooooooool0
POooool0017ooooo00?o0000ooooooooool01Oooool00ol0003oooooo`000009ooooo`03o`000?oo
oooooooo00?ooooo0ol0000>ooooo`;o00002?ooool00ol0003oooooooooo`06ooooo`;o00001ooo
ool00ol0003oooooooooo`0`ooooo`?o00003Oooool00ol0003oooooooooo`22ooooo`004Oooool0
0ol0003oooooooooo`04ooooo`04o`000?ooooooooooo`0000[ooooo00?o0000ooooooooool00_l0
0002ooooo`03o`000?oooooooooo00cooooo00Co0000ooooool0003o00001_ooool00ol0003ooooo
ooooo`06ooooo`;o00001oooool00ol0003oooooooooo`0cooooo`?o00001_ooool4o`0008Gooooo
000Aooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`00000;ooooo`;o0000
1?ooool00ol0003oooooooooo`0<ooooo`04o`000?ooooooooooooooo`;o00000oooool00ol0003o
ooooooooo`08ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0=_ooool3o`0000;o
oooo1Ol00025ooooo`004?ooool00ol0003oooooooooo`04ooooo`05o`000?ooooooooooooooool0
00002Oooool3o`0000Cooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`02ooooo`?o
00002_ooool00ol0003oooooo`000007ooooo`03o`000?oooooooooo03Sooooo2?l00024ooooo`00
4?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0
1?ooool3o`0000?ooooo00Go0000ooooooooooooooooo`00000@ooooo`03o`000?oooooooooo00?o
oooo0ol00009ooooo`03o`000?oooooo000000Oooooo00?o0000ooooooooool0>Oooool7o`0008Co
oooo000@ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003ooooo
ooooo`02ooooo`;o00001oooool01?l0003oooooooooool0000@ooooo`03o`000?oooooooooo00;o
oooo00Go0000ooooooooooooooooo`000008ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000
ooooooooool0>Oooool7o`0008Cooooo000?ooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool01Oooool01?l0003oooooo`000?l00009ooooo`04o`000?ooooooooooo`00013ooooo00Go
0000ooooooooooooooooo`000005ooooo`;o00001_ooool01Ol0003ooooooooooooooooo000000Go
oooo00?o0000ooooooooool0>_ooool6o`0008Cooooo000?ooooo`05o`000?ooooooooooooooool0
00002?ooool2o`0000cooooo0_l0000Aooooo`05o`000?ooooooooooooooool000001oooool2o`00
00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo
03_ooooo0ol00002ooooo`03o`000?oooooooooo087ooooo000?ooooo`04o`000?ooooooooooo`00
00Kooooo1?l0000>ooooo`03o`000?oooooooooo013ooooo00?o0000ooooool000002_ooool2o`00
0003ooooool0003ooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0koooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0P?ooool000oooooo00Co0000oooooooo
oooo00001?ooool2o`0000?ooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0@oooo
o`;o00003Oooool2o`0000Kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0joooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0Ooooool000kooooo00Co0000oooooooo
oooo00000oooool2o`0000Gooooo00?o0000ooooooooool03?ooool2o`00017ooooo00?o0000oooo
ooooool03?ooool00ol0003oooooo`000006ooooo`05o`000?ooooooooooooooool00000??ooool0
0ol0003oooooooooo`05ooooo`03o`000?oooooooooo07kooooo000<ooooo`?o000000Coooooo`00
0?oooooooooo0_l00007ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo00003ooo
ool2o`0000kooooo00?o0000ooooooooool00_l00004ooooo`05o`000?ooooooooooooooool00000
??ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo07gooooo000;ooooo`So00002_oo
ool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000003Oooool00ol0003ooooo
o`00000>ooooo`03o`000?oooooooooo00;ooooo0_l00003ooooo`04o`000?ooooooooooo`0003co
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`1looooo`002_ooool7o`0000cooooo
00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000cooooo00Co0000oooooooo
oooo00003_ooool00ol0003oooooooooo`04ooooo`;o00000_ooool00ol0003oooooo`00000koooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool0Noooool000[ooooo4Ol00002ooooo`03
o`000?oooooooooo00[ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`09ooooo`05
o`000?ooooooooooooooool000003?ooool00ol0003oooooooooo`07ooooo`;o000000?oooooo`00
0?l000000_l0000iooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool0N_ooool000[o
oooo1ol0000:oooooa[o00001Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool02Oooool6o`0003Sooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`1iooooo`002oooool5o`000003ooooool0003o000000[ooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo6_l00009ooooo`Oo0000
=oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo07Sooooo000<ooooo`Co00000ooo
ool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02_oo
ool00ol0003oooooooooo`0@o`0003Oooooo00?o0000ooooooooool03Oooool00ol0003ooooooooo
o`1gooooo`004?ooool01?l0003oooooooooooooool2o`0000Sooooo00?o0000ooooooooool02?oo
ool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00[ooooo1ol0000fooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool0M_ooool0013ooooo00?o0000ooooooooool00oooool2
o`0000Kooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`07ooooo`05o`000?oooooo
ooooooooool000002?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Sooooo0_l0
00000ooooooo0000o`000003o`0003Oooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`1eooooo`004Oooool00ol0003oooooooooo`04ooooo`;o00001?ooool00ol0003oooooooooo`07
ooooo`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00002Oooool00ol0003ooooooooo
o`08ooooo`03o`000?oooooooooo00Gooooo0ol00003ooooo`Co0000>?ooool00ol0003ooooooooo
o`0Aooooo`;o0000MOooool001;ooooo00?o0000ooooooooool01Oooool2o`0000;ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`09ooooo`;o00002_ooool00ol0003oooooooooo`07
ooooo`03o`000?oooooooooo00Cooooo0_l00005ooooo`03o`000?oooooooooo03[ooooo00?o0000
ooooooooool04oooool00ol0003oooooooooo`1booooo`004_ooool00ol0003oooooooooo`07oooo
o`?o00002Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Wooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`02ooooo`;o00001_ooool00ol0003oooooooooo`0koooo
o`03o`000?oooooooooo01Cooooo00?o0000ooooooooool0LOooool001?ooooo00?o0000oooooooo
ool02?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooool0
00002_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00;o00001oooool00ol0003o
ooooooooo`0kooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool0L?ooool001Cooooo
00?o0000ooooooooool02?ooool2o`0000Kooooo00?o0000ooooooooool02?ooool01Ol0003ooooo
oooooooooooo000000Wooooo00?o0000ooooooooool01oooool3o`0000Sooooo00?o0000oooooooo
ool0??ooool00ol0003oooooooooo`0Gooooo`03o`000?oooooooooo06oooooo000Eooooo`03o`00
0?oooooooooo00Oooooo00Co0000ooooool0003o00001?ooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`04ooooo`?o
00002_ooool00ol0003oooooooooo`0looooo`03o`000?oooooooooo01Sooooo00?o0000oooooooo
ool0K_ooool001Gooooo00?o0000ooooooooool01oooool01?l0003oooooooooooooool2o`000003
ooooool0003ooooo00Wooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00;ooooo0_l000000ooooooo0000ooooo`09ooooo`03o`000?oooooooooo03go
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`1]ooooo`005_ooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00?ooooo0_l00009ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool01_ooool00ol0003oooooooooo`02o`0000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`0nooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
K?ooool001Oooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`04ooooo`?o00001_oo
ool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Gooooo0ol00005ooooo`03o`000?oo
oooooooo00Kooooo00?o0000ooooooooool0?_ooool00ol0003oooooooooo`0Looooo`03o`000?oo
oooooooo06_ooooo000Hooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool00oooool0
1?l0003oooooooooool00004ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01?oo
ool2o`0000Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0oooooo`03o`000?oo
oooooooo01gooooo00?o0000ooooooooool0J_ooool001Sooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`02ooooo`03o`000?oooooooooo00;ooooo0_l000000ooooooo0000ooooo`0;
ooooo`03o`000?oooooooooo00?o000000?oooooo`000?ooool01Oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo043ooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`1Y
ooooo`006Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool01?ooool2o`0000gooooo0_l00005ooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`11ooooo`03o`000?oooooooooo01oooooo00?o0000oooo
ooooool0J?ooool001[ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00Cooooo00Co0000ooooool0003o00002Oooool3o`0000Kooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo047ooooo00?o0000oooo
ooooool08?ooool00ol0003oooooooooo`1Wooooo`006_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`;o
00001Oooool2o`0000?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool0@Oooool00ol0003oooooooooo`0Rooooo`03
o`000?oooooooooo06Kooooo000Kooooo`03o`000?oooooooooo00;ooooo00Go0000oooooooooooo
ooooo`000005ooooo`03o`000?oooooooooo00Gooooo0_l000000ooooooo0000o`000006ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo04;ooooo00?o0000ooooooooool08oooool00ol0003oooooooooo`1Uooooo`00
7?ooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000oooooooo
ool01_ooool3o`0000Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool0@oooool00ol0003oooooooooo`0Tooooo`03
o`000?oooooooooo06Cooooo000Mooooo`05o`000?ooooooooooooooool000000_ooool01Ol0003o
oooooooooooooooo000000Oooooo0_l00003ooooo`;o00001oooool01Ol0003ooooooooooooooooo
000000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`14ooooo`03o`000?oooooo
oooo02Gooooo0_l0001Tooooo`007Oooool01Ol0003ooooooooooooooooo000000;ooooo00Co0000
oooooooooooo00001_ooool2o`0000Oooooo0_l00006ooooo`05o`000?ooooooooooooooool00000
0_ooool01Ol0003ooooooooooooooooo000004Oooooo00?o0000ooooooooool09oooool00ol0003o
ooooooooo`1Qooooo`007_ooool01_l0003oooooooooool0003oooooo`0000?ooooo00?o0000oooo
ooooool00_ooool2o`0000_ooooo0_l00005ooooo`04o`000?ooooooooooo`0000;ooooo00Go0000
ooooooooooooooooo`000016ooooo`03o`000?oooooooooo02Wooooo00?o0000ooooooooool0H?oo
ool001oooooo00Go0000ooooool0003oooooo`000002ooooo`04o`000?ooooooooooooooo`;o0000
3oooool2o`0000?ooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00017oooo
o`03o`000?oooooooooo02[ooooo00?o0000ooooooooool0Goooool001oooooo00?o0000ooooool0
00000_l000000ooooooo0000ooooo`03o`0001?ooooo0_l00002ooooo`05o`000?oooooo0000oooo
ool000000_ooool00ol0003oooooooooo`16ooooo`03o`000?oooooooooo02_ooooo00?o0000oooo
ooooool0G_ooool0023ooooo1ol0000Hooooo`03o`000?oooooooooo00Co000000?oooooo`000?oo
ool0B?ooool00ol0003oooooooooo`0/ooooo`03o`000?oooooooooo05gooooo000Oooooo`Oo0000
6_ooool7o`0004[ooooo00?o0000ooooooooool0;Oooool00ol0003oooooooooo`1Looooo`007ooo
oolEo`0000gooooo1ol00018ooooo`03o`000?oooooooooo02oooooo00?o0000ooooooooool0Fooo
ool001oooooo1ol0000>oooooaCo0000B?ooool00ol0003oooooooooo`0`ooooo`03o`000?oooooo
oooo05[ooooo000Pooooo`Go00007?ooool7o`0004Sooooo00?o0000ooooooooool0<Oooool00ol0
003oooooooooo`1Iooooo`008Oooool3o`0001kooooo1Ol00019ooooo`03o`000?oooooooooo03;o
oooo00?o0000ooooooooool0F?ooool004?ooooo0ol0001:ooooo`03o`000?oooooooooo03?ooooo
00?o0000ooooooooool0Eoooool008oooooo00?o0000ooooooooool0=Oooool00ol0003ooooooooo
o`1Fooooo`00Soooool00ol0003oooooooooo`0fooooo`03o`000?oooooooooo05Gooooo002?oooo
o`03o`000?oooooooooo03Oooooo1?l0001Cooooo`00Soooool00ol0003oooooooooo`0gooooo`Go
0000D_ooool008oooooo00?o0000ooooooooool0=_ooool7o`00057ooooo002>ooooo`03o`000?oo
oooooooo03Kooooo2?l0001Aooooo`00S_ooool00ol0003oooooooooo`0cooooo`?o000000?ooooo
o`000?l000001Ol0001Aooooo`00S_ooool00ol0003oooooooooo`0`ooooo`?o00001Oooool5o`00
05;ooooo002>ooooo`03o`000?oooooooooo02gooooo0ol00009ooooo`?o0000Doooool008kooooo
00?o0000ooooooooool0:_ooool3o`0006;ooooo002=ooooo`03o`000?oooooooooo02Wooooo0_l0
001Uooooo`00SOooool00ol0003oooooooooo`0Vooooo`?o0000Ioooool008gooooo00?o0000oooo
ooooool08oooool3o`0006[ooooo002=ooooo`03o`000?oooooooooo023ooooo0ol0001]ooooo`00
SOooool00ol0003oooooooooo`0Mooooo`?o0000L?ooool008cooooo00?o0000ooooooooool06ooo
ool3o`0007?ooooo002<ooooo`03o`000?oooooooooo01Wooooo0_l0001fooooo`00S?ooool00ol0
003oooooooooo`0Fooooo`?o0000N?ooool008cooooo00?o0000ooooooooool04oooool3o`0007_o
oooo002<ooooo`03o`000?oooooooooo013ooooo0ol0001nooooo`00Roooool00ol0003ooooooooo
o`0>ooooo`?o0000POooool008_ooooo00?o0000ooooooooool03?ooool2o`0008Cooooo002;oooo
o`03o`000?oooooooooo00Wooooo0ol00026ooooo`00Roooool00ol0003oooooooooo`06ooooo`?o
0000ROooool008Wooooo0ol00005ooooo`?o0000S?ooool008Sooooo1Ol000001?oooooo0000o`00
0?l0002?ooooo`00Qoooool7o`0009;ooooo0027ooooo`Oo0000T_ooool008Oooooo1ol0002Boooo
o`00R?ooool5o`0009?ooooo0029ooooo`?o0000U?ooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0351667, -0.0500052, \
0.00378109, 0.00363322}},
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]],

Cell["\<\
Setting the Radius option provides control over the size of the \
central cycle.\
\>", "Text",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowGraphArray[{\[IndentingNewLine]ConnectedCaveman[3, 7, 5, 3, 
            6, \[IndentingNewLine]Radius \[Rule] 
              1.75\[IndentingNewLine]], \[IndentingNewLine]ConnectedCaveman[
            3, 7, 5, 3, 
            6, \[IndentingNewLine]Radius \[Rule] 
              4\[IndentingNewLine]]}\[IndentingNewLine]];\)\)], "Input",
  CellLabel->"In[32]:=",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .47619 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0113379 0.47619 [
[ 0 0 0 0 ]
[ 1 .47619 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .47619 L
0 .47619 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0113379 0.477324 0.464853 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0468911 0.968905 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.51138 .63379 m
.7379 .85117 L
s
.51138 .63379 m
.448 .95155 L
s
.51138 .63379 m
.36381 .58268 L
s
.51138 .63379 m
.60258 .5 L
s
.7379 .85117 m
.448 .95155 L
s
.36381 .58268 m
.39108 .74283 L
s
.36381 .58268 m
.2906 .8654 L
s
.36381 .58268 m
.13804 .8581 L
s
.36381 .58268 m
.04828 .72642 L
s
.36381 .58268 m
.08891 .56953 L
s
.36381 .58268 m
.22933 .50556 L
s
.36381 .58268 m
.36381 .41732 L
s
.39108 .74283 m
.2906 .8654 L
s
.39108 .74283 m
.13804 .8581 L
s
.39108 .74283 m
.04828 .72642 L
s
.39108 .74283 m
.08891 .56953 L
s
.39108 .74283 m
.22933 .50556 L
s
.2906 .8654 m
.13804 .8581 L
s
.2906 .8654 m
.04828 .72642 L
s
.2906 .8654 m
.08891 .56953 L
s
.2906 .8654 m
.22933 .50556 L
s
.13804 .8581 m
.04828 .72642 L
s
.13804 .8581 m
.08891 .56953 L
s
.13804 .8581 m
.22933 .50556 L
s
.04828 .72642 m
.08891 .56953 L
s
.04828 .72642 m
.22933 .50556 L
s
.08891 .56953 m
.22933 .50556 L
s
.36381 .41732 m
.16706 .48545 L
s
.36381 .41732 m
.04545 .30707 L
s
.36381 .41732 m
.16706 .12869 L
s
.36381 .41732 m
.36381 .19682 L
s
.36381 .41732 m
.51138 .36621 L
s
.16706 .48545 m
.04545 .30707 L
s
.16706 .48545 m
.16706 .12869 L
s
.16706 .48545 m
.36381 .19682 L
s
.04545 .30707 m
.16706 .12869 L
s
.04545 .30707 m
.36381 .19682 L
s
.16706 .12869 m
.36381 .19682 L
s
.51138 .36621 m
.448 .04845 L
s
.51138 .36621 m
.7379 .14883 L
s
.51138 .36621 m
.60258 .5 L
s
.448 .04845 m
.7379 .14883 L
s
.60258 .5 m
.69057 .33757 L
s
.60258 .5 m
.86655 .33757 L
s
.60258 .5 m
.95455 .5 L
s
.60258 .5 m
.86655 .66243 L
s
.60258 .5 m
.69057 .66243 L
s
.69057 .33757 m
.86655 .33757 L
s
.69057 .33757 m
.95455 .5 L
s
.69057 .33757 m
.86655 .66243 L
s
.69057 .33757 m
.69057 .66243 L
s
.86655 .33757 m
.95455 .5 L
s
.86655 .33757 m
.86655 .66243 L
s
.86655 .33757 m
.69057 .66243 L
s
.95455 .5 m
.86655 .66243 L
s
.95455 .5 m
.69057 .66243 L
s
.86655 .66243 m
.69057 .66243 L
s
.025 w
.51138 .63379 Mdot
.7379 .85117 Mdot
.448 .95155 Mdot
.36381 .58268 Mdot
.39108 .74283 Mdot
.2906 .8654 Mdot
.13804 .8581 Mdot
.04828 .72642 Mdot
.08891 .56953 Mdot
.22933 .50556 Mdot
.36381 .41732 Mdot
.16706 .48545 Mdot
.04545 .30707 Mdot
.16706 .12869 Mdot
.36381 .19682 Mdot
.51138 .36621 Mdot
.448 .04845 Mdot
.7379 .14883 Mdot
.60258 .5 Mdot
.69057 .33757 Mdot
.86655 .33757 Mdot
.95455 .5 Mdot
.86655 .66243 Mdot
.69057 .66243 Mdot
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0113379 0.97619 0.464853 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0162956 0.936861 0.0475018 0.950036 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.55349 .78391 m
.67979 .89924 L
s
.55349 .78391 m
.51815 .9525 L
s
.55349 .78391 m
.22436 .67547 L
s
.55349 .78391 m
.7569 .5 L
s
.67979 .89924 m
.51815 .9525 L
s
.22436 .67547 m
.23956 .76043 L
s
.22436 .67547 m
.18354 .82546 L
s
.22436 .67547 m
.09847 .82158 L
s
.22436 .67547 m
.04842 .75172 L
s
.22436 .67547 m
.07107 .66849 L
s
.22436 .67547 m
.14937 .63455 L
s
.22436 .67547 m
.22436 .32453 L
s
.23956 .76043 m
.18354 .82546 L
s
.23956 .76043 m
.09847 .82158 L
s
.23956 .76043 m
.04842 .75172 L
s
.23956 .76043 m
.07107 .66849 L
s
.23956 .76043 m
.14937 .63455 L
s
.18354 .82546 m
.09847 .82158 L
s
.18354 .82546 m
.04842 .75172 L
s
.18354 .82546 m
.07107 .66849 L
s
.18354 .82546 m
.14937 .63455 L
s
.09847 .82158 m
.04842 .75172 L
s
.09847 .82158 m
.07107 .66849 L
s
.09847 .82158 m
.14937 .63455 L
s
.04842 .75172 m
.07107 .66849 L
s
.04842 .75172 m
.14937 .63455 L
s
.07107 .66849 m
.14937 .63455 L
s
.22436 .32453 m
.11465 .36068 L
s
.22436 .32453 m
.04684 .26604 L
s
.22436 .32453 m
.11465 .17141 L
s
.22436 .32453 m
.22436 .20755 L
s
.22436 .32453 m
.55349 .21609 L
s
.11465 .36068 m
.04684 .26604 L
s
.11465 .36068 m
.11465 .17141 L
s
.11465 .36068 m
.22436 .20755 L
s
.04684 .26604 m
.11465 .17141 L
s
.04684 .26604 m
.22436 .20755 L
s
.11465 .17141 m
.22436 .20755 L
s
.55349 .21609 m
.51815 .0475 L
s
.55349 .21609 m
.67979 .10076 L
s
.55349 .21609 m
.7569 .5 L
s
.51815 .0475 m
.67979 .10076 L
s
.7569 .5 m
.80597 .41382 L
s
.7569 .5 m
.90409 .41382 L
s
.7569 .5 m
.95316 .5 L
s
.7569 .5 m
.90409 .58618 L
s
.7569 .5 m
.80597 .58618 L
s
.80597 .41382 m
.90409 .41382 L
s
.80597 .41382 m
.95316 .5 L
s
.80597 .41382 m
.90409 .58618 L
s
.80597 .41382 m
.80597 .58618 L
s
.90409 .41382 m
.95316 .5 L
s
.90409 .41382 m
.90409 .58618 L
s
.90409 .41382 m
.80597 .58618 L
s
.95316 .5 m
.90409 .58618 L
s
.95316 .5 m
.80597 .58618 L
s
.90409 .58618 m
.80597 .58618 L
s
.025 w
.55349 .78391 Mdot
.67979 .89924 Mdot
.51815 .9525 Mdot
.22436 .67547 Mdot
.23956 .76043 Mdot
.18354 .82546 Mdot
.09847 .82158 Mdot
.04842 .75172 Mdot
.07107 .66849 Mdot
.14937 .63455 Mdot
.22436 .32453 Mdot
.11465 .36068 Mdot
.04684 .26604 Mdot
.11465 .17141 Mdot
.22436 .20755 Mdot
.55349 .21609 Mdot
.51815 .0475 Mdot
.67979 .10076 Mdot
.7569 .5 Mdot
.80597 .41382 Mdot
.90409 .41382 Mdot
.95316 .5 Mdot
.90409 .58618 Mdot
.80597 .58618 Mdot
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[32]:=",
  ImageSize->{288, 137.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000RB000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00@?ooool3o`0009Gooooo0ol0
0015ooooo`00@?ooool3o`0009Gooooo1?l00014ooooo`00@?ooool6o`0009;ooooo0ol000001?oo
oooo0000o`000?l00011ooooo`00@Oooool00ol0003oooooooooo`02ooooo`?o0000T?ooool00ol0
003oooooooooo`03ooooo`?o0000?_ooool0047ooooo00?o0000ooooooooool01Oooool3o`0008go
oooo00?o0000ooooooooool01_ooool3o`0003_ooooo0012ooooo`03o`000?oooooooooo00Oooooo
0ol0002;ooooo`03o`000?oooooooooo00Sooooo0ol0000hooooo`00@_ooool00ol0003ooooooooo
o`0:ooooo`?o0000R?ooool00ol0003oooooooooo`0;ooooo`?o0000=Oooool004;ooooo00?o0000
ooooooooool03Oooool3o`0008Gooooo00?o0000ooooooooool03_ooool5o`00033ooooo0012oooo
o`03o`000?oooooooooo013ooooo0ol00022ooooo`03o`000?oooooooooo013ooooo0ol0000`oooo
o`00@_ooool00ol0003oooooooooo`0Cooooo`?o0000Ooooool00ol0003oooooooooo`0?ooooo`Co
0000<?ooool001_ooooo0ol0000Uooooo`03o`000?oooooooooo01Gooooo0ol0001mooooo`03o`00
0?oooooooooo00gooooo00?o0000ooooooooool0<_ooool001_ooooo1?l0000Tooooo`03o`000?oo
oooooooo01Sooooo0ol0001jooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0<ooo
ool001_ooooo0ol000001?oooooo0000o`000?l0000Qooooo`03o`000?oooooooooo01_ooooo0ol0
001gooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool0=?ooool001cooooo0ol00003
ooooo`?o00007_ooool00ol0003oooooooooo`0Nooooo`Co0000Loooool00ol0003oooooooooo`0:
ooooo`03o`000?oooooooooo03Gooooo000Kooooo`05o`000?oooooo0000ooooool000001Oooool3
o`0001_ooooo00?o0000ooooooooool07oooool3o`0007?ooooo00?o0000ooooooooool02Oooool0
0ol0003oooooooooo`0fooooo`006oooool01Ol0003oooooo`000?oooooo000000Sooooo0ol0000I
ooooo`03o`000?oooooooooo01kooooo0ol0001cooooo`03o`000?oooooooooo00Oooooo0_l0000i
ooooo`006_ooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo00Sooooo0_l0000G
ooooo`03o`000?oooooooooo01gooooo00?o0000ooooooooool0??ooool3o`0003Kooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`0iooooo`006Oooool01Ol0003ooooooooooooooooo
000000?ooooo00?o0000ooooooooool02Oooool3o`0001Cooooo00?o0000ooooooooool07?ooool0
0ol0003oooooooooo`0mooooo`Co0000=Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo03[ooooo000Hooooo`03o`000?oooooooooo00;ooooo00Go0000ooooooooooooooooo`00000>
ooooo`?o00004Oooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo03kooooo0ol00000
1?oooooo0000o`000?l0000booooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0>ooo
ool001Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooo
oooo00kooooo1Ol0000<ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0?oooool3
o`0000Cooooo0ol0000_ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0??ooool0
01Oooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo
00oooooo0ol0000=ooooo`03o`000?oooooooooo01Sooooo00?o0000ooooooooool0?oooool01Ol0
003oooooo`000?oooooo000000Kooooo0ol0000/ooooo`05o`000?ooooooooooooooool00000?ooo
ool001Kooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooo
oooo00cooooo1_l0000=ooooo`03o`000?oooooooooo01Oooooo00?o0000ooooooooool0@?ooool0
0ol0003oooooo`000002ooooo`03o`000?oooooooooo00Kooooo1?l0000Yooooo`03o`000?oooooo
0000043ooooo000Fooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0
003oooooooooo`08ooooo`?o00000oooool2o`0000kooooo00?o0000ooooooooool05_ooool00ol0
003oooooooooo`10ooooo`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01oooool3
o`0002Sooooo0ol00011ooooo`005Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo
00Gooooo00?o0000ooooooooool01?ooool3o`0000Gooooo00?o0000ooooool000003_ooool00ol0
003oooooooooo`0Eooooo`03o`000?oooooooooo043ooooo00Go0000ooooooooooooooooo`000003
ooooo`03o`000?oooooooooo00?ooooo1_l0000Xooooo`?o0000@Oooool001Cooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`05ooooo`04o`000?ooooooooooooooo`?o00002?ooool0
0ol0003oooooo`00000>ooooo`03o`000?oooooooooo01Cooooo00?o0000ooooooooool0@Oooool0
1Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooool000000_l00004ooooo`;o00009Ooo
ool6o`00047ooooo000Dooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01_ooool3
o`0000[ooooo00Co0000oooooooooooo00003oooool00ol0003oooooooooo`0Booooo`03o`000?oo
oooooooo047ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`04o`0000Kooooo00?o
0000ooooool000008_ooool3o`0000Gooooo00?o0000ooooooooool0?oooool001?ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`03ooooo`?o000000?oooooo`000?ooool02Oooool0
1Ol0003ooooooooooooooooo000000oooooo00?o0000ooooooooool04Oooool00ol0003ooooooooo
o`11ooooo`03o`000?oooooooooo00?ooooo0ol00003ooooo`03o`000?oooooooooo00?ooooo00Co
0000oooooooooooo00007oooool3o`0000Wooooo00?o0000ooooooooool0?_ooool001;ooooo00?o
0000ooooooooool02?ooool00ol0003oooooooooo`03o`0000Cooooo00?o0000ooooooooool02?oo
ool01Ol0003ooooooooooooooooo000000oooooo00?o0000ooooooooool04?ooool00ol0003ooooo
ooooo`12ooooo`03o`000?oooooooooo00Co00001_ooool01Ol0003ooooooooooooooooo000000?o
oooo00?o0000ooooooooool06_ooool3o`0000gooooo00?o0000ooooooooool0?Oooool001;ooooo
00?o0000ooooooooool02?ooool3o`0000Sooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool03_ooool2o`00047ooooo
1ol00003ooooo`03o`000?oooooooooo00Gooooo00Co0000oooooooooooo00000oooool00ol0003o
ooooooooo`0Gooooo`?o00004?ooool00ol0003oooooooooo`0mooooo`004Oooool00ol0003ooooo
ooooo`06ooooo`Co00002oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool03Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo047ooooo
1?l00006ooooo`03o`000?oooooooooo00Kooooo0_l00004ooooo`03o`000?oooooooooo01Cooooo
0ol0000Dooooo`03o`000?oooooooooo03cooooo000@ooooo`03o`000?oooooooooo00Cooooo0ol0
0003ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`03ooooo`03o`000?oooooooooo00kooooo00?o0000ooooooooool02oooool00ol0003ooooooooo
o`12ooooo`?o000000Coooooo`000?l0003o00000oooool00ol0003oooooooooo`06ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool04Oooool3o`0001Sooooo00?o0000ooooooooool0
>oooool000oooooo00?o0000ooooooooool00_ooool3o`0000Kooooo00?o0000ooooooooool02_oo
ool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool03_oo
ool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo04Kooooo00Co0000oooooooooooooooo
1?l00007ooooo`03o`000?oooooo000000Cooooo00?o0000ooooooooool03_ooool3o`0001cooooo
00?o0000ooooooooool0>_ooool000oooooo00?o0000ooooool000000_l00009ooooo`03o`000?oo
oooooooo00[ooooo00Go0000ooooooooooooooooo`000007ooooo`03o`000?oooooooooo00kooooo
00?o0000ooooooooool02Oooool00ol0003oooooooooo`18ooooo`03o`000?oooooooooo00?ooooo
0ol00004ooooo`05o`000?ooooooooooooooool000000oooool00ol0003oooooooooo`0;ooooo`?o
00007oooool00ol0003oooooooooo`0jooooo`003_ooool3o`0000cooooo00?o0000ooooooooool0
2oooool01?l0003oooooooooool00007ooooo`03o`000?oooooooooo00kooooo00?o0000oooooooo
ool02?ooool00ol0003oooooooooo`19ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool00ol000000ooooooo0000ooooo`03ooooo`04o`000?ooooooooooo`0000[ooooo0ol0000Soooo
o`03o`000?oooooooooo03Wooooo000<ooooo`?o00003_ooool00ol0003oooooooooo`0<ooooo`;o
00002?ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo00Oooooo00?o0000oooooooo
ool0Boooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00?ooooo0ol00003ooooo`04
o`000?ooooooooooo`0000Oooooo0ol0000Wooooo`03o`000?oooooooooo03Sooooo000<ooooo`Ko
00002oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Oooooo00?o0000oooooooo
ool03oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo04gooooo00Go0000oooooooo
ooooooooo`000004ooooo`04o`000?ooooooooooooooo`?o000000Coooooo`000?oooooo00001?oo
ool3o`0002_ooooo00?o0000ooooooooool0=oooool000cooooo0ol00003ooooo`?o00002?ooool0
0ol0003oooooooooo`0<ooooo`;o00002?ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool0C_ooool01Ol0003ooooooooooooooooo000000?ooooo00?o
0000ooooooooool01Oooool4o`000004ooooool0003o0000o`0002kooooo00?o0000ooooooooool0
=oooool000oooooo00?o0000ooooooooool00oooool3o`0000Gooooo00?o0000ooooooooool02ooo
ool01?l0003oooooooooool00007ooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool0
0oooool00ol0003oooooooooo`0>ooooo`?o00005?ooool3o`0002Sooooo00Co0000oooooooooooo
00000oooool00ol0003oooooooooo`06ooooo`Co0000<_ooool00ol0003oooooooooo`0fooooo`00
3oooool00ol0003oooooooooo`06ooooo`;o00000oooool00ol0003oooooooooo`0:ooooo`05o`00
0?ooooooooooooooool000001oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00;o
oooo00?o0000ooooooooool03ooooolJo`0002Wooooo00?o0000ooooool000000_ooool00ol0003o
ooooooooo`03ooooo`Oo0000=?ooool00ol0003oooooooooo`0eooooo`004?ooool00ol0003ooooo
ooooo`07ooooo`Co00003?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool03oooool01Ol0003ooooooooooooooooo000001;ooooo1Ol0000Aooooo`Co
0000:Oooool01Ol0003oooooo`000?oooooo000000?ooooo0ol00006ooooo`03o`000?oooooooooo
03?ooooo00?o0000ooooooooool0=?ooool0017ooooo00?o0000ooooooooool02Oooool3o`0000Wo
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo013o
oooo00?o0000ooooool000004oooool3o`0000;ooooo00?o0000ooooooooool03?ooool2o`000004
ooooool0003o0000o`0002[ooooo1ol00009ooooo`03o`000?oooooooooo03?ooooo00?o0000oooo
ooooool0=?ooool0017ooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`03o`0000Ko
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00oo
oooo0ol0000Dooooo`?o00000oooool2o`0000[ooooo0_l00003ooooo`?o0000:_ooool3o`0000go
oooo00?o0000ooooooooool0=?ooool00ol0003oooooooooo`0cooooo`004_ooool00ol0003ooooo
ooooo`08ooooo`03o`000?oooooooooo00?ooooo0_l00003ooooo`03o`000?oooooooooo00Kooooo
00?o0000ooooooooool00_ooool00ol0003oooooooooo`0?ooooo`?o00004oooool01Ol0003ooooo
o`000?oooooo000000Cooooo0_l00007ooooo`03o`000?oooooooooo00;ooooo00Go0000ooooool0
003oooooo`00000Yooooo`?o00003Oooool00ol0003oooooooooo`0eooooo`03o`000?oooooooooo
03;ooooo000Cooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01Oooool3o`0000[o
oooo00Go0000ooooooooooooooooo`00000>ooooo`Ko00004oooool01Ol0003oooooo`000?oooooo
000000Kooooo00?o0000ooooooooool00_ooool2o`0000Gooooo00?o0000ooooool000000_ooool0
0ol0003oooooooooo`0fooooo`03o`000?oooooooooo03Kooooo00?o0000ooooooooool0<Oooool0
01?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`07ooooo`Co00001oooool01Ol0
003ooooooooooooooooo000000_ooooo0ol00005ooooo`03o`000?oooooooooo013ooooo00Co0000
oooooooooooo00000_ooool00ol0003oooooooooo`04ooooo`Co00001_ooool01?l0003ooooooooo
ool00002ooooo`03o`000?oooooooooo03Kooooo00?o0000ooooooooool0=_ooool00ol0003ooooo
ooooo`0aooooo`005?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool00_ooool2o`0000Kooooo00Co0000oooooooooooo00002?ooool3o`0000Wooooo
00?o0000ooooooooool03oooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo00Go
oooo0ol00006ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool0=Oooool00ol0
003oooooooooo`0gooooo`03o`000?oooooooooo033ooooo000Eooooo`03o`000?oooooooooo00Go
oooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`?o00001?ooool00ol0003o
ooooo`000006ooooo`;o00003?ooool00ol0003oooooooooo`0>ooooo`05o`000?oooooooooooooo
ool000000oooool00ol0003oooooooooo`02ooooo`;o00000oooool00ol0003oooooooooo`02oooo
o`05o`000?ooooooooooooooool000000oooool00ol0003oooooooooo`0eooooo`03o`000?oooooo
oooo03Sooooo00?o0000ooooooooool0;oooool001Gooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo0ol000001?oooooo0000ooooool00003
ooooo`?o00003oooool00ol0003oooooooooo`0=ooooo`05o`000?ooooooooooooooool000001?oo
ool01?l0003oooooooooool00006ooooo`;o000000?oooooo`000?ooool00oooool00ol0003ooooo
ooooo`02ooooo`03o`000?oooooooooo03Cooooo00?o0000ooooooooool0>Oooool00ol0003ooooo
ooooo`0;ooooo`?o00002_ooool3o`0001?ooooo000Fooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool01?ooool00ol0003oooooooooo`0<ooooo`Oo00004oooool00ol0003ooooooooo
o`0;ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool3o`0000Wooooo0_l0
0004ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0=?ooool00ol0003ooooooooo
o`0iooooo`03o`000?oooooooooo00_ooooo4?l0000Cooooo`005oooool00ol0003oooooooooo`03
ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool03oooool3o`0001Kooooo00?o0000
ooooooooool02Oooool00ol0003oooooooooo`03ooooo`04o`000?ooooooooooooooo`?o00002_oo
ool00ol0003oooooo`000003ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0<ooo
ool00ol0003oooooooooo`0jooooo`03o`000?oooooooooo00[ooooo1Ol00006ooooo`Go00004ooo
ool001Oooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooo
oooo00cooooo1_l0000Fooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00oooool0
1?l0003oooooooooool00002ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool00_l0
00000ooooooo0000ooooo`05ooooo`03o`000?oooooooooo03;ooooo00?o0000ooooooooool0>ooo
ool00ol0003oooooooooo`09ooooo`?o00000_ooool00ol0003oooooooooo`02ooooo`03o`000?oo
oooooooo00?o00004oooool001Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00[ooooo0ol00004ooooo`03o`000?oooooooooo01Kooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`04ooooo`?o00001?ooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo00?ooooo0_l00006ooooo`03o`000?oooooooooo03;ooooo00?o0000
ooooooooool0??ooool00ol0003oooooooooo`08ooooo`?o00000oooool4o`0000?ooooo0ol0000C
ooooo`006Oooool01Ol0003ooooooooooooooooo000000?ooooo00?o0000ooooooooool02?ooool3
o`0000Oooooo00?o0000ooooooooool05oooool00ol0003oooooooooo`06ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooo
oooo00Cooooo0_l00006ooooo`03o`000?oooooooooo037ooooo00?o0000ooooooooool0??ooool0
0ol0003oooooooooo`07ooooo`05o`000?oooooo0000ooooool000000_ooool4o`0000;ooooo00Go
0000ooooool0003oooooo`00000Booooo`006Oooool01Ol0003ooooooooooooooooo000000?ooooo
00?o0000ooooooooool01Oooool3o`0000[ooooo00?o0000ooooooooool05oooool00ol0003ooooo
ooooo`05ooooo`03o`000?oooooooooo00?ooooo0ol00007ooooo`03o`000?oooooooooo00Cooooo
00?o0000ooooooooool01Oooool01?l0003oooooo`000?l00004ooooo`03o`000?oooooooooo037o
oooo00?o0000ooooooooool0?Oooool00ol0003oooooooooo`06ooooo`03o`000?oooooo000000;o
oooo0_l00004ooooo`;o00000_ooool00ol0003oooooo`00000Booooo`006_ooool01?l0003ooooo
ooooool00002ooooo`03o`000?oooooooooo00Cooooo0_l0000=ooooo`03o`000?oooooooooo01So
oooo00?o0000ooooooooool01?ooool01?l0003oooooooooooooool2o`0000;ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo00Co0000oooo
oooooooooooo0_l00003ooooo`03o`000?oooooooooo033ooooo00?o0000ooooooooool0?_ooool0
0ol0003oooooooooo`04ooooo`05o`000?ooooooooooo`000?ooool00_l00006ooooo`;o000000Go
ooooo`000?ooooooooooo`00000Aooooo`006oooool01Ol0003oooooo`000?oooooo000000Cooooo
0ol0000?ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool00_ooool01Ol0003ooooo
oooooooooooo000000Cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00Kooooo00?o0000ooooooooool00oooool01?l0003oooooooooool0000boooo
o`03o`000?oooooooooo03oooooo00?o0000ooooooooool00_ooool01?l0003oooooooooooooool2
o`0000;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02o`0000?ooooo00?o0000
ooooooooool03_ooool001_ooooo00Ko0000ooooool0003oooooo`000?ooool3o`0001;ooooo00?o
0000ooooooooool06_ooool01_l0003ooooooooooooooooo0000ooooo`;o00001Oooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`04ooooo`;o000000?oooooo`000?ooool0<?ooool00ol0003oooooooooo`0oooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooool000000_l00003ooooo`03o`000?oooooooooo00;o
oooo00Co0000oooooooooooooooo0ol000000ooooooo0000ooooo`0?ooooo`006oooool6o`0001Go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`03o`0000Oooooo00?o0000oooooooo
ool01oooool01?l0003oooooooooool00009ooooo`03o`000?oooooooooo00Kooooo0ol0000`oooo
o`03o`000?oooooooooo043ooooo00Ko0000oooooooooooo0000ooooool00002ooooo`03o`000?oo
oooooooo00;ooooo00Co0000oooooooooooo00001?ooool01_l0003oooooooooool0003oooooo`00
00oooooo000Kooooo`?o00006?ooool00ol0003oooooooooo`0Kooooo`03o`000?oooooo000000Wo
oooo00?o0000ooooooooool01oooool01?l0003oooooooooool00009ooooo`03o`000?oooooooooo
00Sooooo00?o0000ooooooooool0;_ooool00ol0003oooooooooo`11ooooo`04o`000?oooooo0000
o`0000?ooooo00?o0000ooooooooool00_ooool01?l0003oooooooooool00004ooooo`04o`000?oo
ooooooooooooo`;o00003oooool001_ooooo0ol00006ooooo`?o00003oooool00ol0003ooooooooo
o`0Koooooc7o0000;_ooool00ol0003oooooooooo`11oooooaco00003Oooool002Cooooo0ol0000?
ooooo`03o`000?oooooooooo01_ooooo1?l00008ooooo`03o`000?oooooooooo00Sooooo0_l0000:
ooooo`03o`000?oooooooooo00Oooooo1?l0000^ooooo`03o`000?oooooooooo047ooooo1?l00003
ooooo`03o`000?oooooooooo00?ooooo0_l00005ooooo`04o`000?ooooooooooooooo`Co00003Ooo
ool0027ooooo2?l0000=ooooo`03o`000?oooooooooo01_ooooo0ol000000ooooooo0000o`000006
ooooo`03o`000?oooooooooo00Sooooo0_l0000:ooooo`03o`000?oooooooooo00Gooooo0_l00000
1?oooooo0000o`000?l0000^ooooo`03o`000?oooooooooo047ooooo0ol000001Ooooooo0000o`00
0?oooooo000000Gooooo0_l00005ooooo`05o`000?oooooo0000o`000?ooool00ol0000=ooooo`00
7oooool2o`0000?ooooo0ol00002ooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool0
6_ooool01?l0003oooooooooool00003ooooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool01oooool01?l0003oooooooooool00009ooooo`03o`000?oooooooooo00?ooooo0_l00003oooo
o`03o`000?oooooooooo02kooooo00?o0000ooooooooool0@?ooool01?l0003oooooooooool00003
ooooo`;o00001?ooool01?l0003oooooooooool00004ooooo`;o00000oooool00ol0003ooooooooo
o`0=ooooo`007Oooool2o`0000Cooooo0ol000001?oooooo0000ooooooooool2o`0000[ooooo00?o
0000ooooooooool06_ooool01Ol0003ooooooooooooooooo000000?ooooo0_l00003ooooo`03o`00
0?oooooooooo00Oooooo00Co0000oooooooooooo00002Oooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool0;oooool00ol0003oooooooooo`0oooooo`03
o`000?oooooooooo00;ooooo00Co0000oooooooooooooooo0_l00003ooooo`04o`000?oooooooooo
o`0000?ooooo0_l00003ooooo`03o`000?oooooooooo00kooooo000Jooooo`?o00001Oooool01?l0
003oooooo`000?ooool2o`0000Cooooo0_l00008ooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool00_ooool00ol0003oooooooooo`03ooooo`;o000000?oooooo`000?ooool01oooool0
0ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool00_l00005
ooooo`03o`000?oooooooooo02oooooo00?o0000ooooooooool0?oooool00ol0003oooooooooo`02
ooooo`06o`000?ooooooooooooooool0003ooooo0ol00004ooooo`?o000000?oooooo`000?ooool0
0_ooool00ol0003oooooooooo`0>ooooo`006?ooool2o`0000Oooooo00So0000oooooooooooo0000
ooooool0003oooooo`0000Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0Hoooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool2o`0000Sooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooo000000Kooooo00?o0000oooo
ooooool0<?ooool00ol0003oooooooooo`0nooooo`03o`000?oooooooooo00Cooooo00Co0000oooo
oooooooo00000oooool00ol0003oooooooooo`02ooooo`05o`000?ooooooooooooooool000000_oo
ool00ol0003oooooooooo`0?ooooo`005_ooool2o`0000Sooooo00Oo0000ooooooooooooooooo`00
0?oooooo000000;ooooo00?o0000ooooooooool00oooool2o`0000Gooooo00?o0000ooooooooool0
6?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Gooooo0_l00006ooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool01Oooool2o`0000Kooooo00?o0000ooooooooool0
<Oooool00ol0003oooooooooo`0mooooo`03o`000?oooooooooo00Gooooo00Co0000oooooooooooo
00000_ooool00ol0003oooooo`000003o`000005ooooool0003oooooooooool000000_ooool00ol0
003oooooooooo`0?ooooo`004oooool3o`0000[ooooo00Co0000oooooooooooo00000_ooool01?l0
003oooooooooool00007ooooo`;o00000oooool00ol0003oooooooooo`0Gooooo`03o`000?oooooo
oooo00Kooooo00?o0000ooooooooool01?ooool00ol0003oooooo`000005ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool00oooool3o`0000Oooooo00?o0000ooooooooool0<Oooool0
0ol0003oooooooooo`0mooooo`03o`000?oooooooooo00Kooooo00?o0000ooooool000000_ooool0
0ol0003oooooooooo`02o`0000;ooooo00Ko0000oooooooooooo0000ooooool0000Booooo`004Ooo
ool;o`0000?ooooo00Go0000ooooooooooooooooo`000002ooooo`05o`000?ooooooooooooooool0
00002?ooool01?l0003oooooooooool0000Hooooo`03o`000?oooooooooo00Oooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`02o`0000;ooooo00?o0000ooooooooool01_ooool01Ol0
003ooooooooooooooooo000000;ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0b
ooooo`03o`000?oooooooooo03cooooo00?o0000ooooooooool01oooool01_l0003oooooo`000?oo
oooo0000ooooo`;o00000_ooool2o`000006ooooool0003oooooo`000?oooooo00004_ooool0017o
oooo1Ol00006oooooa;o00001_ooool4o`0001Oooooo00?o0000ooooooooool02?ooool00ol0003o
ooooooooo`03ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooool000002?ooool01?l0003o
ooooo`000?l00003ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0<_ooool00ol0
003oooooooooo`0kooooo`03o`000?oooooooooo00Wooooo1Ol00006ooooo`Go00004oooool0017o
oooo0ol00002ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00oooool01Ol0003o
oooooooooooooooo000000?ooooo00?o0000ooooooooool02_l0000Fooooo`03o`000?oooooooooo
00Wooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`;o00002Oooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0<oooool00ol0
003oooooooooo`0jooooo`03o`000?oooooooooo00[ooooo0ol0000:ooooo`?o00004oooool0017o
oooo0_l000001?oooooo0000ooooooooool2o`0000?ooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01?ooool8o`00
01;ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooool000001_ooool3o`0000Gooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`0cooooo`03o`000?oooooooooo03[ooooo00?o0000ooooooooool02_ooool@o`00
01?ooooo000Aooooo`04o`000?oooooo0000o`0000Cooooo0ol00006ooooo`03o`000?oooooooooo
00;ooooo00?o0000ooooooooool00oooool01?l0003oooooooooooooool2o`0000;ooooo0_l00004
ooooo`?o00003oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool00_ooool01?l0003oooooooooooooool2o`0000?ooooo00Go0000oooooooooooooooo
o`000004ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0=?ooool00ol0003ooooo
ooooo`0iooooo`03o`000?oooooooooo00_ooooo0ol0000:ooooo`?o00004oooool0017ooooo00Go
0000ooooool0003oooooo`000005ooooo`;o00001Oooool00ol0003oooooooooo`02ooooo`03o`00
0?oooooooooo00?ooooo1?l00003ooooo`?o00001oooool3o`0000_ooooo00?o0000ooooooooool0
3Oooool01Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooooooool00oooool3o`0000Co
oooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo03Co
oooo00?o0000ooooooooool0>?ooool00ol0003oooooooooo`0_ooooo`004Oooool00ol0003ooooo
o`000002ooooo`05o`000?ooooooooooooooool000000_ooool01Ol0003ooooooooooooooooo0000
00Gooooo00?o0000ooooooooool00_ooool3o`0000Cooooo00Go0000ooooool0003oooooo`000009
ooooo`;o00002?ooool00ol0003oooooooooo`0>ooooo`05o`000?ooooooooooooooool000000ooo
ool00ol0003oooooooooo`03ooooo`Co00001Oooool01Ol0003ooooooooooooooooo000000?ooooo
00?o0000ooooooooool0=Oooool00ol0003oooooooooo`0gooooo`03o`000?oooooooooo033ooooo
000@ooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooool000001?ooool2o`000003oooo
ool0003ooooo00Cooooo00?o0000ooooooooool00_l00003ooooo`04o`000?ooooooooooo`0000;o
oooo00?o0000ooooool000002oooool3o`0000Gooooo00?o0000ooooooooool03oooool01?l0003o
ooooooooool00002ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00_ooool2o`00
00Cooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`0eooooo`03o`000?oooooo
oooo03Oooooo00?o0000ooooooooool0<?ooool0013ooooo00Co0000oooooooooooo00000oooool2
o`0000Oooooo0_l00006ooooo`;o00001_ooool00ol0003oooooo`000002ooooo`03o`000?oooooo
000000kooooo1Ol0000Booooo`04o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooooooo
0_l00007ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003oooooooooool0000]oooo
o`?o00002?ooool00ol0003oooooooooo`0fooooo`03o`000?oooooooooo037ooooo000@ooooo`05
o`000?ooooooooooooooool000000_ooool2o`0000Sooooo0ol00002ooooo`?o00001oooool2o`00
00;ooooo00Co0000oooooooooooo00004?ooool3o`0001?ooooo00Go0000ooooool0003oooooo`00
0003ooooo`03o`000?oooooooooo00Sooooo0_l00002ooooo`05o`000?oooooo0000ooooool00000
;_ooool3o`0000Sooooo00?o0000ooooooooool0=Oooool00ol0003oooooooooo`0booooo`004?oo
ool01Ol0003ooooooooooooooooo000000;ooooo00?o0000ooooool000001_ooool01?l0003ooooo
ooooooooool2o`0000;ooooo00?o0000ooooooooool01Oooool2o`0000;ooooo00Co0000oooooooo
oooo00004?ooool4o`0001;ooooo00Ko0000ooooool0003oooooo`000?ooool2o`0000gooooo00Oo
0000ooooool0003oooooo`000?oooooo000002_ooooo2?l00006ooooo`03o`000?oooooooooo03Co
oooo00?o0000ooooooooool0<oooool000oooooo00?o0000ooooooooool00_ooool00ol0003ooooo
o`000002ooooo`03o`000?oooooooooo00Cooooo1?l000001Ooooooo0000o`000?oooooo000000Ko
oooo00Ko0000oooooooooooo0000ooooool00003ooooo`03o`000?oooooooooo00kooooo00Co0000
oooooooooooo00004_ooool5o`00013ooooo1Ol0000Zooooo`;o00000oooool3o`0000;ooooo0_l0
0004ooooo`03o`000?oooooooooo03Cooooo00?o0000ooooooooool0<oooool000oooooo00?o0000
ooooooooool00_ooool2o`0000Cooooo00?o0000ooooooooool00_ooool2o`0000Kooooo0_l00005
ooooo`07o`000?ooooooooooooooool0003oooooo`000003ooooo`03o`000?oooooooooo00kooooo
00Go0000ooooooooooooooooo`00000Aooooo`?o00005?ooool3o`0002Gooooo1Ol00004ooooo`?o
000000?oooooo`000?ooool00_ooool2o`0000;ooooo00?o0000ooooooooool0<oooool00ol0003o
ooooooooo`0dooooo`003oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Cooooo
00Ko0000ooooool0003o0000ooooool00007ooooo`;o00001?ooool00ol0003oooooooooo`02oooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo00kooooo6_l0000Uooooo`go000000?oooooo`000?ooool00oooool3o`00
03Cooooo00?o0000ooooooooool0=Oooool000oooooo00Co0000oooooooooooooooo0_l00006oooo
o`;o00000_ooool00ol0003oooooooooo`07ooooo`?o000000?oooooo`000?ooool01?ooool2o`00
00?ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00gooooo0ol0000Dooooo`?o00009Oooool5o`0000;ooooo00Go0000oooooooooooo0000ooooo`0<
o`0003;ooooo00?o0000ooooooooool0=Oooool000oooooo00Co0000oooooooooooo00000_ooool0
1?l0003oooooooooooooool2o`000005ooooool0003oooooooooool000002Oooool00ol0003ooooo
ooooo`02o`0000Gooooo0_l00003ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0
1Oooool00ol0003oooooooooo`1<ooooo`;o00000_ooool2o`0000;ooooo00Co0000oooooooooooo
00000_ooool00ol0003oooooooooo`04o`0000;ooooo0ol0000^ooooo`03o`000?oooooooooo03Ko
oooo000>ooooo`05o`000?ooooooooooooooool000000_ooool01?l0003oooooo`000?l00004oooo
o`03o`000?oooooo000000Wooooo00?o0000ooooool000000_ooool2o`0000?ooooo00?o0000oooo
ool000000oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00Kooooo00?o0000oooo
ooooool0B_ooool2o`000008ooooool0003oooooo`000?oooooo0000ooooool00002ooooo`03o`00
0?oooooooooo00?o00000_ooool2o`0000Gooooo0ol0000Zooooo`03o`000?oooooooooo03Oooooo
000>ooooo`04o`000?ooooooooooo`0000?ooooo0_l00006ooooo`03o`000?oooooo000000Wooooo
0_l00005ooooo`03o`000?oooooo000000?ooooo00Co0000oooooooooooo00003_ooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo04Wooooo00?o0000ooooool000000_l00003ooooo`;o
00000_ooool3o`0000;ooooo00?o0000ooooool000000_l00008ooooo`?o00009_ooool00ol0003o
ooooooooo`0hooooo`003_ooool01?l0003oooooo`000?ooool4o`0000Sooooo00?o0000oooooooo
ool02?ooool2o`0000Kooooo0_l00003ooooo`04o`000?ooooooooooo`0000gooooo00?o0000oooo
ooooool02Oooool00ol0003oooooooooo`18ooooo`05o`000?oooooo0000ooooool000001?ooool4
o`0000Cooooo0_l000001?oooooo0000ooooool0000:ooooo`?o00008oooool00ol0003ooooooooo
o`0hooooo`003_ooool4o`0000?ooooo00?o0000ooooooooool01_ooool2o`0000Wooooo0_l00007
ooooo`?o00000_ooool00ol0003oooooo`00000=ooooo`03o`000?oooooooooo00[ooooo00?o0000
ooooooooool0A_ooool00ol0003oooooooooo`02o`000004ooooool0003oooooooooo`?o000000?o
ooooo`000?l000000oooool2o`000004ooooool0003oooooo`0000gooooo0ol0000Oooooo`03o`00
0?oooooooooo03Wooooo000<ooooo`Co00001_ooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooo000000Oooooo00?o0000ooooool000001oooool00ol0003oooooooooo`02o`000003ooooool0
003o000000gooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`15ooooo`04o`000?oo
ooooooooo`0000;ooooo1?l00004ooooo`Co00000_ooool01?l0003oooooooooool0000@ooooo`?o
00006oooool00ol0003oooooooooo`0jooooo`003?ooool=o`0000Gooooo00?o0000ooooool00000
1_ooool01?l0003oooooooooool00006ooooo`03o`000?oooooooooo00?ooooo0_l000000ooooooo
0000ooooo`0;ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0A?ooool01Ol0003o
ooooo`000?l0003ooooo00?o000000?oooooo`000?ooool00oooool00ol0003oooooo`000002o`00
0005ooooool0003o0000ooooool000004oooool3o`0001Oooooo00?o0000ooooooooool0>oooool0
00cooooo1Ol00005ooooo`03o`000?oooooooooo01Oo000000?oooooo`000?ooool01Oooool4o`00
00_ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`13ooooo`Go00000oooool00ol0
003oooooooooo`03ooooo`;o00000oooool3o`000003ooooool0003ooooo01Gooooo0ol0000Doooo
o`03o`000?oooooooooo03_ooooo000>ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000
ooooooooool00oooool<o`0000[ooooo00?o0000ooooooooool03oooool2o`00047ooooo1?l00000
0ooooooo0000ooooo`03ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00_ooool4
o`000003ooooool0003ooooo01Oooooo0ol0000@ooooo`03o`000?oooooooooo03cooooo000?oooo
o`03o`000?oooooooooo00;o00000_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo
00;ooooo00Co0000oooooooooooo00001?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00?ooooo1Ol0000:ooooo`03o`000?oooooooooo017ooooo00?o0000ooooooooool0?_oooolA
o`0000Cooooo00?o0000ooooooooool00ol0000Kooooo`?o00003?ooool00ol0003oooooooooo`0m
ooooo`003oooool00ol0003oooooooooo`02ooooo`;o000000?oooooo`000?ooool01?ooool00ol0
003oooooooooo`03ooooo`;o00001_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo
00?o00000_ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo01;ooooo00?o0000oooo
ooooool0?Oooool5o`000005ooooool0003oooooooooool000000_ooool00ol0003oooooooooo`0<
o`0001kooooo0ol00009ooooo`03o`000?oooooooooo03gooooo000@ooooo`03o`000?oooooooooo
00?ooooo0_l00004ooooo`03o`000?oooooooooo00Cooooo0_l00006ooooo`03o`000?oooooooooo
00;ooooo00Co0000ooooool0003o00001?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooo
oooo01?ooooo00?o0000ooooooooool0?_ooool00ol0003oooooooooo`02o`0000;ooooo00Ko0000
oooooooooooo0000ooooool00002ooooo`04o`000?ooooooooooooooo`Ko00008Oooool3o`0000Go
oooo00?o0000ooooooooool0?_ooool0017ooooo00?o0000ooooooooool00oooool3o`0000;ooooo
00?o0000ooooooooool00oooool00ol0003oooooo`000006ooooo`03o`000?oooooooooo00;ooooo
0_l00005ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool05?ooool00ol0003ooooo
ooooo`0nooooo`03o`000?oooooooooo00;o000000?oooooo`000?ooool00_ooool01Ol0003ooooo
oooooooooooo000000;ooooo0_l00003ooooo`03o`000?oooooooooo02?ooooo1Ol00011ooooo`00
4_ooool00ol0003oooooooooo`02ooooo`06o`000?ooooooooooo`000?oooooo00001?ooool01Ol0
003ooooooooooooooooo000000Gooooo00?o0000ooooooooool00ol00005ooooo`03o`000?oooooo
oooo00cooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`0mooooo`05o`000?oooooo
ooooo`000?ooool00_l00002ooooo`03o`000?oooooo000000;ooooo1?l00003ooooo`03o`000?oo
oooooooo02Kooooo0ol00011ooooo`004_ooool00ol0003oooooooooo`02ooooo`04o`000?oooooo
ooooooooo`;o00001?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00;ooooo0ol0
0002ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool03?ooool00ol0003ooooooooo
o`0Gooooo`03o`000?oooooooooo03gooooo00Ko0000ooooool0003oooooo`000?ooool2o`0000;o
oooo1?l00002ooooo`04o`000?ooooooooooo`0002Wooooo1?l00010ooooo`004oooool01Ol0003o
oooooooooooooooo000000?ooooo00Ko0000ooooool0003o0000ooooool00005ooooo`03o`000?oo
oooooooo00?o00000oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00gooooo00?o
0000ooooooooool06?ooool00ol0003oooooooooo`0mooooo`;o000000?oooooo`000?ooool01_l0
00001?oooooo0000ooooool00002ooooo`03o`000?oooooooooo02Wooooo00Co0000oooooooooooo
0000?oooool001Cooooo00Go0000ooooooooooooooooo`000002ooooo`04o`000?oooooooooooooo
o`;o00001_ooool2o`0000?ooooo00Co0000oooooooooooo00001?ooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0??ooool00ol0003oooooo`000002
o`000003ooooool0003o000000;o000000Ooooooo`000?l0003oooooo`000?oooooo000002cooooo
00Go0000ooooooooooooooooo`00000nooooo`005?ooool01Ol0003ooooooooooooooooo000000;o
oooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooooooool2o`000003ooooool0003o
oooo00;ooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`0?ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool0??ooool5o`0000Cooooo0_l000001?oooooo0000o`00
0?l0000/ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0>oooool001Gooooo00Ko
0000oooooooooooo0000ooooool00002ooooo`03o`000?oooooooooo00;ooooo0ol00004ooooo`04
o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooo00004Oooool00ol0003oooooooooo`0L
ooooo`03o`000?oooooooooo03_ooooo2?l00002ooooo`Co0000;Oooool00ol0003oooooooooo`04
ooooo`03o`000?oooooooooo03[ooooo000Fooooo`05o`000?oooooo0000ooooool000000_ooool0
1?l0003oooooooooooooool2o`000003ooooool0003o000000?ooooo00Ko0000oooooooooooo0000
ooooool00002ooooo`03o`000?oooooooooo013ooooo00?o0000ooooooooool07Oooool00ol0003o
ooooooooo`0jooooo`?o00001Oooool6o`0002gooooo00?o0000ooooooooool01Oooool00ol0003o
ooooooooo`0iooooo`005oooool2o`000005ooooool0003oooooo`000?ooool00ol00005ooooo`;o
00000_ooool01Ol0003oooooo`000?oooooo000000;ooooo00?o0000ooooooooool04?ooool00ol0
003oooooooooo`0Nooooo`03o`000?oooooooooo04Cooooo0ol0000]ooooo`03o`000?oooooooooo
00Kooooo0_l0000iooooo`005oooool00ol0003oooooo`000004o`0000[ooooo0_l000001ooooooo
0000o`000?oooooo0000ooooool000004oooool00ol0003oooooooooo`0Nooooo`?o0000M?ooool0
0ol0003oooooooooo`08ooooo`03o`000?oooooooooo03Kooooo000Hooooo`Co00003_ooool2o`00
0004ooooool0003o0000o`0001Cooooo00?o0000ooooooooool07_ooool3o`0007?ooooo00?o0000
ooooooooool02_ooool00ol0003oooooooooo`0eooooo`006?ooool<o`0000Oooooo1?l0000Doooo
o`03o`000?oooooooooo01gooooo1Ol0001cooooo`03o`000?oooooooooo00_ooooo00?o0000oooo
ooooool0=?ooool001Sooooo0ol00009ooooo`[o00005Oooool00ol0003oooooooooo`0Jooooo`?o
0000N?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo03?ooooo000[ooooo`?o0000
5Oooool00ol0003oooooooooo`0Gooooo`?o0000Noooool00ol0003oooooooooo`0=ooooo`03o`00
0?oooooooooo03;ooooo0013ooooo`03o`000?oooooooooo01Cooooo0ol0001nooooo`03o`000?oo
oooooooo00kooooo00?o0000ooooooooool0<Oooool004?ooooo00?o0000ooooooooool04Oooool3
o`00087ooooo00?o0000ooooooooool03oooool3o`00033ooooo0012ooooo`03o`000?oooooooooo
00oooooo0ol00023ooooo`03o`000?oooooooooo013ooooo0ol0000`ooooo`00@_ooool00ol0003o
ooooooooo`0<ooooo`?o0000Q_ooool00ol0003oooooooooo`0=ooooo`Ko0000<?ooool004;ooooo
00?o0000ooooooooool02Oooool3o`0008Wooooo00?o0000ooooooooool02_ooool3o`0003Kooooo
0012ooooo`03o`000?oooooooooo00Kooooo0ol0002<ooooo`03o`000?oooooooooo00Oooooo0ol0
000iooooo`00@_ooool00ol0003oooooooooo`03ooooo`?o0000Soooool00ol0003oooooooooo`04
ooooo`?o0000??ooool0047ooooo00Co0000oooooooooooooooo0ol0002Aooooo`03o`000?oooooo
oooo00;ooooo0ol0000oooooo`00@?ooool5o`0009?ooooo1_l00012ooooo`00@?ooool3o`0009Go
oooo0ol00015ooooo`00@?ooool3o`0009Gooooo0ol00015ooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {136.125, 0}} -> {-0.0541876, -0.0238146, \
0.00734626, 0.00734626}, {{7.375, 137}, {132.875, 3.1875}} -> {-0.11259, \
-0.0745272, 0.0084861, 0.00796222}, {{149.938, 279.625}, {132.875, 3.1875}} -> \
{-1.25147, -0.0763837, 0.00823059, 0.00811645}},
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]],

Cell["A list of integers can also be given as the first argument.", "Text",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ConnectedCaveman[{3, 7, 5, 3, 6}] // ShowGraph;\)\)], "Input",
  CellLabel->"In[33]:=",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0324019 0.921522 0.0479513 0.959027 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.53129 .70664 m
.71044 .87452 L
s
.53129 .70664 m
.48117 .95205 L
s
.53129 .70664 m
.29787 .62771 L
s
.53129 .70664 m
.67555 .5 L
s
.71044 .87452 m
.48117 .95205 L
s
.29787 .62771 m
.31943 .75139 L
s
.29787 .62771 m
.23997 .84605 L
s
.29787 .62771 m
.11931 .84041 L
s
.29787 .62771 m
.04831 .73872 L
s
.29787 .62771 m
.08045 .61755 L
s
.29787 .62771 m
.19151 .56815 L
s
.29787 .62771 m
.29787 .37229 L
s
.31943 .75139 m
.23997 .84605 L
s
.31943 .75139 m
.11931 .84041 L
s
.31943 .75139 m
.04831 .73872 L
s
.31943 .75139 m
.08045 .61755 L
s
.31943 .75139 m
.19151 .56815 L
s
.23997 .84605 m
.11931 .84041 L
s
.23997 .84605 m
.04831 .73872 L
s
.23997 .84605 m
.08045 .61755 L
s
.23997 .84605 m
.19151 .56815 L
s
.11931 .84041 m
.04831 .73872 L
s
.11931 .84041 m
.08045 .61755 L
s
.11931 .84041 m
.19151 .56815 L
s
.04831 .73872 m
.08045 .61755 L
s
.04831 .73872 m
.19151 .56815 L
s
.08045 .61755 m
.19151 .56815 L
s
.29787 .37229 m
.14225 .42491 L
s
.29787 .37229 m
.04608 .28715 L
s
.29787 .37229 m
.14225 .14939 L
s
.29787 .37229 m
.29787 .20201 L
s
.29787 .37229 m
.53129 .29336 L
s
.14225 .42491 m
.04608 .28715 L
s
.14225 .42491 m
.14225 .14939 L
s
.14225 .42491 m
.29787 .20201 L
s
.04608 .28715 m
.14225 .14939 L
s
.04608 .28715 m
.29787 .20201 L
s
.14225 .14939 m
.29787 .20201 L
s
.53129 .29336 m
.48117 .04795 L
s
.53129 .29336 m
.71044 .12548 L
s
.53129 .29336 m
.67555 .5 L
s
.48117 .04795 m
.71044 .12548 L
s
.67555 .5 m
.74514 .37456 L
s
.67555 .5 m
.88433 .37456 L
s
.67555 .5 m
.95392 .5 L
s
.67555 .5 m
.88433 .62544 L
s
.67555 .5 m
.74514 .62544 L
s
.74514 .37456 m
.88433 .37456 L
s
.74514 .37456 m
.95392 .5 L
s
.74514 .37456 m
.88433 .62544 L
s
.74514 .37456 m
.74514 .62544 L
s
.88433 .37456 m
.95392 .5 L
s
.88433 .37456 m
.88433 .62544 L
s
.88433 .37456 m
.74514 .62544 L
s
.95392 .5 m
.88433 .62544 L
s
.95392 .5 m
.74514 .62544 L
s
.88433 .62544 m
.74514 .62544 L
s
.025 w
.53129 .70664 Mdot
.71044 .87452 Mdot
.48117 .95205 Mdot
.29787 .62771 Mdot
.31943 .75139 Mdot
.23997 .84605 Mdot
.11931 .84041 Mdot
.04831 .73872 Mdot
.08045 .61755 Mdot
.19151 .56815 Mdot
.29787 .37229 Mdot
.14225 .42491 Mdot
.04608 .28715 Mdot
.14225 .14939 Mdot
.29787 .20201 Mdot
.53129 .29336 Mdot
.48117 .04795 Mdot
.71044 .12548 Mdot
.67555 .5 Mdot
.74514 .37456 Mdot
.88433 .37456 Mdot
.95392 .5 Mdot
.88433 .62544 Mdot
.74514 .62544 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[33]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00ROooool3o`0009Cooooo0028ooooo`Go0000Toooool008Oooooo1ol0002Booooo`00Qooo
ool7o`0009;ooooo0027ooooo`So0000TOooool008Sooooo1Ol00002ooooo`?o0000S_ooool008Wo
oooo0ol00006ooooo`?o0000Roooool008_ooooo00?o0000ooooooooool01oooool3o`0008Sooooo
002;ooooo`03o`000?oooooooooo00[ooooo0ol00025ooooo`00Roooool00ol0003oooooooooo`0=
ooooo`;o0000Poooool008_ooooo00?o0000ooooooooool03oooool3o`00083ooooo002;ooooo`03
o`000?oooooooooo01;ooooo0ol0001mooooo`00S?ooool00ol0003oooooooooo`0Dooooo`?o0000
N_ooool008cooooo00?o0000ooooooooool05oooool3o`0007Oooooo002<ooooo`03o`000?oooooo
oooo01[ooooo0ol0001dooooo`00S?ooool00ol0003oooooooooo`0Mooooo`;o0000L_ooool008co
oooo00?o0000ooooooooool07oooool3o`0006oooooo002=ooooo`03o`000?oooooooooo027ooooo
0ol0001/ooooo`00SOooool00ol0003oooooooooo`0Tooooo`?o0000JOooool008gooooo00?o0000
ooooooooool09oooool3o`0006Kooooo002=ooooo`03o`000?oooooooooo02[ooooo0_l0001Toooo
o`00SOooool00ol0003oooooooooo`0/ooooo`?o0000HOooool008kooooo00?o0000ooooooooool0
;_ooool3o`0005kooooo002>ooooo`03o`000?oooooooooo037ooooo0ol00005ooooo`?o0000Dooo
ool008kooooo00?o0000ooooooooool0=?ooool3o`000003ooooool0003o000000?o0000D_ooool0
08kooooo00?o0000ooooooooool0=oooool7o`00057ooooo002>ooooo`03o`000?oooooooooo03Oo
oooo1ol0001Aooooo`00Soooool00ol0003oooooooooo`0fooooo`Oo0000DOooool008oooooo00?o
0000ooooooooool0=_ooool6o`0005;ooooo000Wooooo`?o0000IOooool00ol0003oooooooooo`0e
ooooo`03o`000?oooooooooo00?o0000Doooool002Kooooo1Ol0001Tooooo`03o`000?oooooooooo
03Cooooo00?o0000ooooooooool0Eoooool002Gooooo1ol0001Sooooo`03o`000?oooooooooo03?o
oooo00?o0000ooooooooool0F?ooool002Gooooo1ol0001Tooooo`03o`000?oooooooooo037ooooo
00?o0000ooooooooool0FOooool002Gooooo2Ol0001Rooooo`03o`000?oooooooooo033ooooo00?o
0000ooooooooool0F_ooool002Kooooo1Ol00003ooooo`?o0000Goooool00ol0003oooooooooo`0_
ooooo`03o`000?oooooooooo05_ooooo000Vooooo`Go00001_ooool3o`0005cooooo00?o0000oooo
ooooool0;_ooool00ol0003oooooooooo`1Looooo`009_ooool01Ol0003oooooo`000?oooooo0000
00Wooooo0ol0001Iooooo`03o`000?oooooooooo02gooooo00?o0000ooooooooool0GOooool002Go
oooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`09ooooo`?o0000Eoooool00ol0
003oooooooooo`0[ooooo`03o`000?oooooooooo05kooooo000Tooooo`05o`000?oooooooooooooo
ool000000oooool00ol0003oooooooooo`0;ooooo`?o0000E?ooool00ol0003oooooooooo`0Zoooo
o`03o`000?oooooooooo05oooooo000Tooooo`05o`000?ooooooooooooooool000001?ooool00ol0
003oooooooooo`0=ooooo`?o0000DOooool00ol0003oooooooooo`0Yooooo`03o`000?oooooooooo
063ooooo000Sooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`0@ooooo`?o0000C_ooool00ol0003oooooooooo`0Xooooo`03o`000?oooooooooo067o
oooo000Rooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003ooooo
ooooo`0Booooo`?o0000Boooool00ol0003oooooooooo`0Wooooo`03o`000?oooooooooo06;ooooo
000Rooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01?ooool00ol0003ooooooooo
o`0Dooooo`?o0000BOooool00ol0003oooooooooo`0Tooooo`;o0000IOooool0027ooooo00?o0000
ooooooooool01?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01Oooooo0ol00016
ooooo`03o`000?oooooooooo02?ooooo00?o0000ooooooooool0IOooool0023ooooo00?o0000oooo
ooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01Wooooo0ol00013oooo
o`03o`000?oooooooooo02;ooooo00?o0000ooooooooool0I_ooool0023ooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo01_ooooo0ol00002ooooo`?o
0000>oooool00ol0003oooooooooo`0Qooooo`03o`000?oooooooooo06Oooooo000Oooooo`03o`00
0?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Nooooo`Ko0000
>_ooool00ol0003oooooooooo`0Pooooo`03o`000?oooooooooo06Sooooo000Nooooo`03o`000?oo
oooooooo00Oooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`0Mooooo`Oo0000>_oo
ool00ol0003oooooooooo`0Nooooo`03o`000?oooooooooo06Wooooo000Nooooo`03o`000?oooooo
oooo00Oooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Kooooo`So0000>_ooool0
0ol0003oooooooooo`0Mooooo`03o`000?oooooooooo06[ooooo000Mooooo`03o`000?oooooooooo
00Sooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Gooooo`?o000000?oooooo`00
0?l000001Ol0000jooooo`03o`000?oooooooooo01cooooo00?o0000ooooooooool0Joooool001co
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo01Co
oooo0ol00005ooooo`Go0000>oooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo06co
oooo000Looooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool02_ooool00ol0003ooooo
ooooo`0@ooooo`?o00002?ooool4o`0003cooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`1]ooooo`006oooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00_ooooo00?o
0000ooooooooool03?ooool3o`0000[ooooo00Co0000oooooooooooo0000?_ooool00ol0003ooooo
ooooo`0Hooooo`03o`000?oooooooooo06kooooo000Jooooo`03o`000?oooooooooo00_ooooo00?o
0000ooooooooool02oooool00ol0003oooooooooo`09ooooo`?o00003?ooool01Ol0003ooooooooo
oooooooo000003kooooo00?o0000ooooooooool05oooool00ol0003oooooooooo`1_ooooo`006Ooo
ool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool01Ooo
ool3o`0000kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0looooo`03o`000?oo
oooooooo01Kooooo00?o0000ooooooooool0L?ooool001Wooooo00?o0000ooooooooool03?ooool0
0ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00;ooooo0_l0000Aooooo`03o`000?oooooo
oooo00;ooooo00?o0000ooooooooool0??ooool00ol0003oooooooooo`0Eooooo`03o`000?oooooo
oooo077ooooo000Hooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool03_ooool4o`00
01;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0looooo`03o`000?oooooooooo
01Cooooo00?o0000ooooooooool0L_ooool001Oooooo00?o0000ooooooooool03_ooool00ol0003o
ooooooooo`0<ooooo`?o00005?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo03go
oooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`1cooooo`005oooool00ol0003ooooo
ooooo`0>ooooo`03o`000?oooooooooo00Wooooo0ol00003ooooo`03o`000?oooooooooo017ooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0mooooo`03o`000?oooooooooo013ooooo
0_l0001fooooo`005_ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00Kooooo0ol0
0007ooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool01Oooool00ol0003ooooooooo
o`0mooooo`03o`000?oooooooooo00oooooo00?o0000ooooooooool0M_ooool001Gooooo00?o0000
ooooooooool04?ooool00ol0003oooooooooo`03ooooo`?o00002_ooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool0?Oooool00ol0003oooooooooo`0>
ooooo`03o`000?oooooooooo07Oooooo000Eooooo`03o`000?oooooooooo013ooooo00?o0000oooo
ooooool00ol0000>ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01_ooool00ol0
003oooooooooo`0mooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool0N?ooool001Co
oooo00?o0000ooooooooool04Oooool3o`0001;ooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo03kooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`1iooooo`004oooool00ol0003oooooooooo`0?ooooo`Co00005?ooool00ol0003ooooo
ooooo`0:ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0?_ooool00ol0003ooooo
ooooo`0:ooooo`03o`000?oooooooooo07[ooooo000Cooooo`03o`000?oooooooooo00gooooo0_l0
0003ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool02?ooool00ol0003ooooooooo
o`09ooooo`03o`000?oooooooooo03kooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`1kooooo`004_ooool00ol0003oooooooooo`0;ooooo`?o00001Oooool00ol0003oooooooooo`0D
ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0n
ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool0O?ooool0017ooooo00?o0000oooo
ooooool02Oooool3o`0000Sooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`05oooo
o`03o`000?oooooooooo00[ooooo00?o0000ooooooooool0?_ooool00ol0003oooooooooo`07oooo
o`03o`000?oooooooooo07gooooo000Aooooo`03o`000?oooooooooo00Kooooo0ol0000;ooooo`03
o`000?oooooooooo01Gooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0;ooooo`03
o`000?oooooooooo03oooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`1nooooo`00
4?ooool00ol0003oooooooooo`04ooooo`?o00003_ooool00ol0003oooooooooo`0Fooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0oooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool0Ooooool000cooooo1?l00004ooooo`?o00004Ooo
ool00ol0003oooooooooo`0Gooooo`05o`000?ooooooooooooooool000003_ooool00ol0003ooooo
ooooo`0oooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0P?ooool000_ooooo1Ol0
00001?oooooo0000o`000?l0000Dooooo`03o`000?oooooooooo01Oooooo00Co0000oooooooooooo
00003oooool00ol0003oooooooooo`0oooooo`03o`000?oooooooooo00;ooooo00?o0000oooooooo
ool0POooool000[ooooo1ol0000Gooooo`03o`000?oooooooooo01Sooooo0_l0000@ooooo`03o`00
0?oooooooooo03oooooo0ol000000ooooooo0000ooooo`23ooooo`002_ooool7o`0001Oooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo03kooooo1Ol0
0025ooooo`002_ooool9o`0001Gooooo00?o0000ooooooooool06?ooool2o`00013ooooo00?o0000
ooooooooool0?Oooool7o`0008Cooooo000;ooooo`Go00000oooool3o`0001;ooooo00?o0000oooo
ooooool05oooool01?l0003oooooooooool0000?ooooo`03o`000?oooooooooo03gooooo1ol00024
ooooo`003?ooool4o`0000Kooooo0ol0000?ooooo`03o`000?oooooooooo01Oooooo00Go0000oooo
ooooooooooooo`00000>ooooo`03o`000?oooooooooo03_ooooo2Ol00024ooooo`003oooool00ol0
003oooooooooo`07ooooo`?o00003?ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`0hooooo`?o00000oooool5o`00
08Gooooo000@ooooo`03o`000?oooooooooo00Wooooo0ol00009ooooo`03o`000?oooooooooo01Go
oooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo03Go
oooo0ol00007ooooo`Co0000QOooool0017ooooo00?o0000ooooooooool02oooool3o`0000Kooooo
00?o0000ooooooooool05Oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00[ooooo
00?o0000ooooooooool0<_ooool3o`0000gooooo00?o0000ooooooooool0Poooool001;ooooo00?o
0000ooooooooool03Oooool3o`0000?ooooo00?o0000ooooooooool05?ooool00ol0003ooooooooo
o`07ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool0;oooool3o`00017ooooo00?o
0000ooooooooool0P_ooool001;ooooo00?o0000ooooooooool04?ooool4o`0001Gooooo00?o0000
ooooooooool02?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo02cooooo0ol0000E
ooooo`03o`000?oooooooooo087ooooo000Cooooo`03o`000?oooooooooo01;ooooo0ol0000Boooo
o`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Yoooo
o`?o00006Oooool00ol0003oooooooooo`20ooooo`005?ooool00ol0003oooooooooo`0Aooooo`03
o`000?oooooooooo00?o00003oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool09_ooool3o`0001cooooo00?o0000ooooooooool0P?ooool001Cooooo
00?o0000ooooooooool04Oooool00ol0003oooooooooo`03ooooo`?o00002oooool00ol0003ooooo
ooooo`0=ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool08oooool3o`00023ooooo
00?o0000ooooooooool0Ooooool001Gooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`06ooooo`?o00001oooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo00Kooooo00?o
0000ooooooooool08?ooool3o`0002Cooooo00?o0000ooooooooool0O_ooool001Kooooo00?o0000
ooooooooool03oooool00ol0003oooooooooo`09ooooo`?o00001?ooool00ol0003oooooooooo`0?
ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool07Oooool3o`0002Oooooo00?o0000
ooooooooool0O_ooool001Kooooo00?o0000ooooooooool03oooool00ol0003oooooooooo`0<oooo
o`Co00004oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01[ooooo0ol0000[oooo
o`03o`000?oooooooooo07gooooo000Gooooo`03o`000?oooooooooo00kooooo00?o0000oooooooo
ool03_ooool4o`00017ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Gooooo`?o
0000;oooool00ol0003oooooooooo`1looooo`006?ooool00ol0003oooooooooo`0=ooooo`03o`00
0?oooooooooo00kooooo00Co0000oooooooooooooooo0ol0000?ooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool05?ooool3o`0003;ooooo00?o0000ooooooooool0O?ooool001Sooooo
00?o0000ooooooooool03Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00Gooooo
0ol0000=ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool04Oooool3o`0003Kooooo
00?o0000ooooooooool0Noooool001Wooooo00?o0000ooooooooool03?ooool00ol0003ooooooooo
o`0<ooooo`03o`000?oooooooooo00Wooooo0ol0000:ooooo`03o`000?oooooooooo00;ooooo00?o
0000ooooooooool03_ooool3o`0003[ooooo00?o0000ooooooooool0N_ooool001[ooooo00?o0000
ooooooooool02oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00gooooo0ol00008
ooooo`05o`000?ooooooooooooooool000003Oooool3o`0003gooooo00?o0000ooooooooool0N_oo
ool001_ooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooo
oooo013ooooo0ol00006ooooo`04o`000?ooooooooooo`0000[ooooo0ol00011ooooo`03o`000?oo
oooooooo07Wooooo000Kooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool02_ooool0
0ol0003oooooooooo`0Dooooo`?o00001?ooool4o`0000Kooooo0ol00015ooooo`03o`000?oooooo
oooo07Sooooo000Looooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool02Oooool00ol0
003oooooooooo`0Hooooo`?o000000?oooooo`000?l000000ol00002ooooo`?o0000BOooool00ol0
003oooooooooo`0[ooooo`?o00009Oooool3o`00027ooooo000Mooooo`03o`000?oooooooooo00So
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0Kooooo`So0000C?ooool00ol0003o
ooooooooo`0Zooooo`Go00008oooool5o`00023ooooo000Mooooo`03o`000?oooooooooo00Sooooo
00?o0000ooooooooool02?ooool00ol0003oooooooooo`0Looooo`Oo0000C_ooool00ol0003ooooo
ooooo`0Xooooo`Oo00008Oooool7o`0001oooooo000Nooooo`03o`000?oooooooooo00Oooooo00?o
0000ooooooooool01oooool00ol0003oooooooooo`0Looooo`So0000Coooool00ol0003ooooooooo
o`0Woooooboo00007oooool001oooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo01[ooooo0ol00002ooooo`Go0000D?ooool00ol0003oooooooooo`0W
ooooo`Oo00008Oooool7o`0001oooooo000Oooooo`03o`000?oooooooooo00Kooooo00?o0000oooo
ooooool01_ooool00ol0003oooooooooo`0Gooooo`?o00001_ooool3o`0005;ooooo00?o0000oooo
ooooool09oooool5o`000003ooooool0003ooooo01kooooo00?o0000ooooool000001?l0000Poooo
o`008?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Gooooo00?o0000oooooooo
ool05Oooool3o`0000[ooooo00?o0000ooooooooool0D_ooool00ol0003oooooooooo`0Wooooo`?o
00000oooool2o`0001_ooooo0_l00003ooooo`?o00008Oooool0027ooooo00?o0000ooooooooool0
1?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01?ooooo0ol0000=ooooo`03o`00
0?oooooooooo05;ooooo00?o0000ooooooooool09_ooool01Ol0003oooooo`000?oooooo000000Co
oooo0_l0000Gooooo`;o00001?ooool01Ol0003oooooo`000?oooooo0000023ooooo000Qooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0@ooooo`?o
00004?ooool00ol0003oooooooooo`1Cooooo`03o`000?oooooooooo02Cooooo00Co0000oooooooo
oooo00000_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo01?ooooo00?o0000oooo
ooooool00oooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo01gooooo000Roooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>oooo
o`?o00004oooool00ol0003oooooooooo`1Dooooo`03o`000?oooooooooo02?ooooo00Co0000oooo
oooooooo00000_ooool00ol0003oooooooooo`04ooooo`;o00004Oooool2o`0000Kooooo00Co0000
oooooooooooo00000_ooool00ol0003oooooooooo`0Mooooo`008oooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool03?ooool3o`0001Kooooo00?o0000
ooooooooool0EOooool00ol0003oooooooooo`0Qooooo`05o`000?ooooooooooooooool000000ooo
ool00ol0003oooooooooo`05ooooo`;o00003Oooool2o`0000Oooooo00Go0000oooooooooooooooo
o`000003ooooo`03o`000?oooooooooo01cooooo000Tooooo`05o`000?ooooooooooooooool00000
1?ooool00ol0003oooooooooo`09ooooo`?o00006Oooool00ol0003oooooooooo`1Eooooo`03o`00
0?oooooooooo027ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`07ooooo`05o`000?oooooooooooooo
ool000000oooool00ol0003oooooooooo`0Looooo`009?ooool01Ol0003ooooooooooooooooo0000
00?ooooo00?o0000ooooooooool01oooool3o`0001cooooo00?o0000ooooooooool0E_ooool00ol0
003oooooooooo`0Oooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`07ooooo`;o00001oooool2o`0000Wooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo01_ooooo000Uooooo`04o`000?ooooooooooo`00
00;ooooo00?o0000ooooooooool01Oooool3o`0001oooooo00?o0000ooooooooool0Eoooool00ol0
003oooooooooo`0Nooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`09ooooo`;o00000oooool2o`0000_ooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`02ooooo`03o`000?oooooooooo01_ooooo000Vooooo`05o`000?oooooo0000oooo
ool000001Oooool3o`0002;ooooo00?o0000ooooooooool0Eoooool00ol0003oooooooooo`0Moooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0:oooo
o`03o`000?oooooo000000cooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03oooo
o`03o`000?oooooooooo01[ooooo000Vooooo`Go00000_ooool3o`0002Gooooo00?o0000oooooooo
ool0F?ooool00ol0003oooooooooo`0Looooo`03o`000?oooooooooo00?ooooo00?o0000oooooooo
ool00oooool00ol0003oooooooooo`0:ooooo`?o00003?ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo00?ooooo00?o0000ooooooooool06_ooool002Kooooo1ol0000Xooooo`03o`00
0?oooooooooo05Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`04ooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool01oooool2o`0000?ooooo0_l00009ooooo`03o`00
0?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Iooooo`009Ooo
ool7o`0002Wooooo00?o0000ooooooooool0FOooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`05ooooo`03o`000?oo
oooooooo00Gooooo00?o0000ooooooooool06?ooool002Gooooo1ol0000Yooooo`03o`000?oooooo
oooo05[ooooo00?o0000ooooooooool06?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool00oooool2o`0000Wooooo0_l00005ooooo`03o`000?oooooo
oooo00Gooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0Hooooo`009Oooool7o`00
02Wooooo00?o0000ooooooooool0Foooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo
00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`02o`0000gooooo0_l00002oooo
o`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Goooo
o`009_ooool5o`0002[ooooo00?o0000ooooooooool0Foooool00ol0003oooooooooo`0Fooooo`03
o`000?oooooooooo00Kooooo00?o0000ooooooooool01_ooool00ol0003oooooo`00000Aooooo`03
o`000?oooooo000000Sooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Gooooo`00
9oooool3o`0002_ooooo00?o0000ooooooooool0G?ooool00ol0003oooooooooo`0Dooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool01_ooool2o`0001?ooooo0_l00008ooooo`03o`00
0?oooooooooo00Oooooo00?o0000ooooooooool05_ooool005Gooooo00?o0000ooooooooool0GOoo
ool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01?oo
ool2o`000003ooooool0003ooooo01;ooooo00Co0000ooooool0003o00001_ooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo01Kooooo001Eooooo`03o`000?oooooooooo05kooooo00?o
0000ooooooooool04Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00?ooooo00?o
0000ooooooooool00_ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00;ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo01Gooooo001E
ooooo`03o`000?oooooooooo05kooooo00?o0000ooooooooool04Oooool00ol0003oooooooooo`08
ooooo`04o`000?ooooooooooooooo`;o00001Oooool00ol0003oooooooooo`0?ooooo`03o`000?oo
oooooooo00?ooooo0_l00003ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool05Ooo
ool005Gooooo00?o0000ooooooooool0Goooool00ol0003oooooooooo`0?ooooo`03o`000?oooooo
oooo00Wooooo00Co0000ooooool0003o00002?ooool00ol0003oooooooooo`0=ooooo`03o`000?oo
oooooooo00Kooooo0_l000000ooooooo0000ooooo`0:ooooo`03o`000?oooooooooo01Cooooo001E
ooooo`03o`000?oooooooooo063ooooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0:
ooooo`;o00002oooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00Wooooo0_l0000<
ooooo`03o`000?oooooooooo01?ooooo001Eooooo`03o`000?oooooooooo063ooooo00?o0000oooo
ooooool03Oooool00ol0003oooooooooo`09ooooo`;o00003?ooool00ol0003oooooooooo`0;oooo
o`03o`000?oooooooooo00[ooooo0_l0000;ooooo`03o`000?oooooooooo01?ooooo001Eooooo`03
o`000?oooooooooo067ooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`08ooooo`;o
000000?oooooo`000?ooool03?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00_o
oooo00Co0000ooooool0003o00002_ooool00ol0003oooooooooo`0Booooo`00EOooool00ol0003o
ooooooooo`1Rooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool01oooool01Ol0003o
oooooooooooooooo000000gooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0;oooo
o`05o`000?ooooooooooooooool000002Oooool00ol0003oooooooooo`0Booooo`00EOooool00ol0
003oooooooooo`1Rooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01_ooool2o`00
00Cooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo
00cooooo00?o0000ooooooooool00_ooool2o`0000Sooooo00?o0000ooooooooool04Oooool005Go
oooo00?o0000ooooooooool0Hoooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Co
oooo0_l00006ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool01oooool00ol0003o
ooooooooo`0<ooooo`03o`000?oooooooooo00Cooooo0_l00006ooooo`03o`000?oooooooooo017o
oooo001Eooooo`03o`000?oooooooooo06Cooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool03Oooool00ol0003ooooo
ooooo`05ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`04ooooo`03o`000?oooooooooo013ooooo001Eooooo`03o`000?oooooooooo06Cooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`02ooooo`;o00002Oooool00ol0003ooooooooo
o`0=ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool00ol0003ooooooooo
o`07ooooo`;o00001?ooool00ol0003oooooooooo`0@ooooo`00EOooool00ol0003oooooooooo`1U
ooooo`03o`000?oooooooooo00Cooooo00Co0000oooooooooooooooo0_l0000;ooooo`03o`000?oo
oooooooo00kooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Wooooo0_l00003ooooo`03o`000?oooooooooo00oooooo001Eooooo`03o`000?oooooo
oooo06Kooooo00?o0000ooooooooool00_ooool01Ol0003ooooooooooooooooo000000gooooo00?o
0000ooooooooool03oooool01Ol0003ooooooooooooooooo0000017ooooo00?o0000ooooooooool0
2oooool01Ol0003ooooooooooooooooo0000013ooooo001Eooooo`03o`000?oooooooooo06Oooooo
1Ol000000ooooooo0000o`00000>ooooo`03o`000?oooooooooo00oooooo00Go0000oooooooooooo
ooooo`00000Aooooo`03o`000?oooooooooo00cooooo0_l000000ooooooo0000o`000002o`0000go
oooo001Eooooo`03o`000?oooooooooo06Oooooo1_l0000@ooooo`03o`000?oooooooooo013ooooo
00?o0000ooooool000004_ooool00ol0003oooooooooo`0>ooooo`Ko00003?ooool005Gooooo00?o
0000ooooooooool0I_ooool7o`00013ooooo00?o0000ooooooooool04?ooool00ol0003oooooo`00
000Booooo`03o`000?oooooooooo00kooooo1ol0000;ooooo`00EOooool00ol0003oooooooooo`1V
oooooeOo00002oooool005Gooooo00?o0000ooooooooool0I_ooool7o`00013ooooo00?o0000oooo
ooooool04Oooool00ol0003oooooooooo`0Aooooo`03o`000?oooooooooo00kooooo1ol0000;oooo
o`00EOooool00ol0003oooooooooo`1Wooooo`Go000000?oooooo`000?ooool03_ooool00ol0003o
ooooooooo`0@ooooo`03o`000?oooooo000001;ooooo00?o0000ooooooooool03Oooool00ol0003o
ooooo`000004o`0000cooooo001Eooooo`03o`000?oooooooooo06Oooooo1?l00003ooooo`;o0000
3Oooool00ol0003oooooooooo`0@ooooo`03o`000?oooooo000001;ooooo00?o0000ooooooooool0
2oooool2o`0000?ooooo0ol0000=ooooo`00EOooool00ol0003oooooooooo`1Wooooo`05o`000?oo
ooooooooooooool000001?ooool2o`0000_ooooo00?o0000ooooooooool03oooool01Ol0003ooooo
oooooooooooo0000017ooooo00?o0000ooooooooool02Oooool2o`0000Cooooo00?o0000oooooooo
ool03_ooool005Gooooo00?o0000ooooooooool0I_ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0>ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`08ooooo`03o`00
0?oooooooooo00?ooooo00?o0000ooooooooool03oooool005Gooooo00?o0000ooooooooool0IOoo
ool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo0_l00008ooooo`03o`000?oo
oooooooo00kooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Kooooo0_l00006ooooo`03o`000?oooooooooo00oooooo001Eooooo`03o`000?oooooo
oooo06Cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`05ooooo`;o00001_ooool0
0ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool03Oooool0
0ol0003oooooooooo`04ooooo`;o00001oooool00ol0003oooooooooo`0@ooooo`00EOooool00ol0
003oooooooooo`1Tooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool01oooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo00gooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`0=ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool01oooool00ol0
003oooooooooo`0@ooooo`00EOooool00ol0003oooooooooo`1Sooooo`03o`000?oooooooooo00So
oooo00?o0000ooooooooool01oooool2o`0000?ooooo00?o0000ooooooooool03?ooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Co0000oooooooooooooooo0_l00009oooo
o`03o`000?oooooooooo017ooooo001Eooooo`03o`000?oooooooooo06;ooooo00?o0000oooooooo
ool02Oooool00ol0003oooooooooo`09ooooo`;o000000?oooooo`000?ooool03Oooool00ol0003o
ooooooooo`07ooooo`03o`000?oooooooooo00cooooo00Co0000ooooool0003o00002oooool00ol0
003oooooooooo`0Aooooo`00EOooool00ol0003oooooooooo`1Rooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool02_ooool2o`0000gooooo00?o0000ooooooooool02Oooool00ol0003o
ooooooooo`0;ooooo`;o00003?ooool00ol0003oooooooooo`0Booooo`00EOooool00ol0003ooooo
ooooo`1Qooooo`03o`000?oooooooooo00_ooooo00?o0000ooooooooool02oooool2o`0000cooooo
00?o0000ooooooooool02Oooool00ol0003oooooooooo`0:ooooo`;o00003Oooool00ol0003ooooo
ooooo`0Booooo`00EOooool00ol0003oooooooooo`1Pooooo`03o`000?oooooooooo00gooooo00?o
0000ooooooooool02_ooool01?l0003oooooo`000?l00009ooooo`03o`000?oooooooooo00_ooooo
00?o0000ooooooooool01oooool2o`000003ooooool0003ooooo00_ooooo00?o0000ooooooooool0
4oooool005Gooooo00?o0000ooooooooool0H?ooool00ol0003oooooooooo`0>ooooo`03o`000?oo
oooooooo00Wooooo00Go0000ooooooooooooooooo`000007ooooo`03o`000?oooooooooo00gooooo
00?o0000ooooooooool01Oooool01Ol0003ooooooooooooooooo000000_ooooo00?o0000oooooooo
ool05?ooool005Gooooo00?o0000ooooooooool0Goooool00ol0003oooooooooo`0?ooooo`03o`00
0?oooooooooo00Wooooo00?o0000ooooooooool00_ooool2o`0000Gooooo00?o0000ooooooooool0
3Oooool00ol0003oooooooooo`03ooooo`;o00001?ooool00ol0003oooooooooo`09ooooo`03o`00
0?oooooooooo01Cooooo001Eooooo`03o`000?oooooooooo05kooooo00?o0000ooooooooool04Ooo
ool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Cooooo0_l00002ooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool00_l00006ooooo`03o`000?oooooooooo00Sooooo00?o
0000ooooooooool05Oooool003Gooooo0ol0000Mooooo`03o`000?oooooooooo05kooooo00?o0000
ooooooooool04Oooool00ol0003oooooooooo`08ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooool000004Oooool00ol0003oooooo`000008ooooo`03o`000?oooooooooo00Sooooo00?o0000
ooooooooool05Oooool003Cooooo1Ol0000Looooo`03o`000?oooooooooo05gooooo00?o0000oooo
ooooool04oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Oooooo0_l0000Aoooo
o`;o00002Oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo01Kooooo000cooooo`Oo
00006oooool00ol0003oooooooooo`1Looooo`03o`000?oooooooooo01Cooooo00?o0000oooooooo
ool01oooool00ol0003oooooooooo`07ooooo`04o`000?oooooo0000o`0000gooooo0_l000000ooo
oooo0000ooooo`08ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool05_ooool003?o
oooo1ol0000Kooooo`03o`000?oooooooooo05_ooooo00?o0000ooooooooool05_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`09ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo01Oooooo000cooooo`Oo00006oooool00ol0003ooooo
ooooo`1Kooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool01_ooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00?ooooo0_l00007ooooo`;o00001Oooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool05oooool0033ooooo0ol0
00000ooooooo0000o`000003o`000003ooooool0003o000001Wooooo00?o0000ooooooooool0F_oo
ool00ol0003oooooooooo`0Hooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`06ooooo`;o00000oooool2o`0000Sooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01Sooooo000^ooooo`;o00001?ooool5
o`0000?ooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`1Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Wooooo00?o0000ooooool000002oooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool06Oooool002cooooo0_l00005ooooo`04o`000?oooooo0000
ooooo`;o00001?ooool2o`0001Kooooo00?o0000ooooooooool0FOooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool01?ooool00ol0003oooooooooo`09
ooooo`?o00002oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo00?o0000
ooooooooool06Oooool002[ooooo0_l00006ooooo`08o`000?ooooooooooo`000?oooooo0000oooo
ool00005ooooo`;o00005?ooool00ol0003oooooooooo`1Hooooo`03o`000?oooooooooo01cooooo
00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00Sooooo
0_l00003ooooo`;o00002_ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool06_ooool002Sooooo0_l00007ooooo`07o`000?ooooooooooooooool0003o
ooooo`000002ooooo`03o`000?oooooooooo00Cooooo0_l0000Booooo`03o`000?oooooooooo05Oo
oooo00?o0000ooooooooool07Oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo01[o
oooo000Uooooo`?o00002Oooool01?l0003oooooooooool00002ooooo`05o`000?oooooooooooooo
ool000001oooool2o`00013ooooo00?o0000ooooooooool0Eoooool00ol0003oooooooooo`0Noooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`06oooo
o`;o00002Oooool2o`0000Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02oooo
o`03o`000?oooooooooo01_ooooo000Sooooo`;o00002oooool01Ol0003ooooooooooooooooo0000
00;ooooo00Go0000ooooooooooooooooo`000009ooooo`;o00003_ooool00ol0003oooooooooo`1F
ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00Cooooo0_l0000=ooooo`;o00001_ooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool06oooool0027ooooo0_l0000<oooo
o`03o`000?oooooooooo00;ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooo
oooo00Sooooo00?o0000ooooooooool02oooool00ol0003oooooooooo`1Eooooo`03o`000?oooooo
oooo027ooooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Cooooo00?o
0000ooooooooool03oooool00ol0003oooooooooo`04ooooo`05o`000?ooooooooooooooool00000
0oooool00ol0003oooooooooo`0Looooo`007oooool2o`0000gooooo00?o0000ooooooooool00ooo
ool01Ol0003ooooooooooooooooo000000Cooooo00?o0000ooooooooool02?ooool2o`0000_ooooo
00?o0000ooooooooool0E?ooool00ol0003oooooooooo`0Rooooo`05o`000?ooooooooooooooool0
00000oooool00ol0003oooooooooo`02ooooo`;o00004oooool2o`0000Cooooo00Go0000oooooooo
ooooooooo`000003ooooo`03o`000?oooooooooo01cooooo000Fooooo`?o00001?ooool2o`0000ko
oooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00;o
oooo00?o0000ooooooooool02_ooool2o`0000Wooooo00?o0000ooooooooool0E?ooool00ol0003o
ooooooooo`0Sooooo`04o`000?ooooooooooo`0000;ooooo00Co0000oooooooooooooooo0_l0000G
ooooo`;o00000oooool01?l0003oooooooooool00002ooooo`03o`000?oooooooooo01gooooo000E
ooooo`So00003oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo00?o0000
ooooooooool00oooool00ol0003oooooooooo`0;ooooo`;o00001oooool00ol0003oooooooooo`1C
ooooo`03o`000?oooooooooo02Gooooo00Go0000ooooool0003oooooo`000003ooooo`03o`000?oo
oooooooo01Wooooo00Wo0000ooooooooooooooooo`000?oooooo0000ooooool000008?ooool001Co
oooo1ol0000Aooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00cooooo0_l00005ooooo`03o`000?oooooooooo05;o
oooo00?o0000ooooooooool09_ooool5o`000003ooooool0003o000001gooooo0_l000000ooooooo
0000o`000003o`00023ooooo000Dooooo`ko00002Oooool00ol0003oooooooooo`05ooooo`03o`00
0?oooooooooo00;ooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`0=ooooo`04o`00
0?ooooooooooooooo`?o0000Doooool00ol0003oooooooooo`0Vooooo`Ko00008Oooool6o`00023o
oooo000Dooooo`Oo00001ooooolEo`0000;ooooo00?o0000ooooooooool01?ooool00ol0003ooooo
ooooo`0>ooooo`Oo0000DOooool00ol0003oooooooooo`0Vooooo`Oo00008Oooool7o`0001oooooo
000Eooooo`Go000000?oooooo`000?ooool03?ooool00ol0003oooooooooo`06ooooo`03o`000?oo
oooooooo00;ooooo5Ol00006ooooo`Oo0000Coooool00ol0003oooooooooo`0Woooooboo00007ooo
ool001Kooooo1?l00002ooooo`;o00002_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooo
oooo00Cooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`07ooooo`go0000Coooool0
0ol0003oooooooooo`0Wooooo`Oo00008Oooool7o`0001oooooo000Fooooo`04o`000?oooooo0000
o`0000Cooooo0_l00007ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01?ooool0
0ol0003oooooooooo`06ooooo`03o`000?oooooooooo00gooooo2Ol0001<ooooo`03o`000?oooooo
oooo02Wooooo1Ol0000Sooooo`Go00008?ooool001Kooooo00Go0000ooooool0003oooooo`000005
ooooo`;o00001Oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Gooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`09ooooo`?o000000?oooooo`000?l000000ol00003
ooooo`?o0000B?ooool00ol0003oooooooooo`0[ooooo`?o00009Oooool3o`00027ooooo000Eoooo
o`04o`000?ooooooooooo`0000;ooooo00?o0000ooooooooool01?ooool2o`0000;ooooo00?o0000
ooooooooool02?ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Sooooo00?o0000
ooooooooool01_ooool2o`0000Cooooo1?l00007ooooo`?o0000AOooool00ol0003oooooooooo`1g
ooooo`005Oooool01?l0003oooooooooool00003ooooo`03o`000?oooooooooo00Gooooo0_l0000;
ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`03
ooooo`;o00001Oooool01Ol0003oooooo`000?oooooo000000[ooooo0ol00011ooooo`03o`000?oo
oooooooo07Sooooo000Eooooo`04o`000?ooooooooooo`0000?ooooo00?o0000ooooooooool01Ooo
ool3o`0000Wooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`08ooooo`04o`000?oo
ooooooooooooo`;o00001_ooool01_l0003oooooooooool0003oooooo`0000gooooo0ol0000moooo
o`03o`000?oooooooooo07Wooooo000Eooooo`04o`000?ooooooooooo`0000Cooooo00?o0000oooo
ooooool00oooool01?l0003oooooooooooooool2o`0000Oooooo00?o0000ooooooooool01oooool0
0ol0003oooooooooo`09ooooo`?o00001oooool01?l0003oooooooooool00002ooooo`03o`000?oo
oooooooo00kooooo0ol0000iooooo`03o`000?oooooooooo07[ooooo000Dooooo`03o`000?oooooo
oooo00;ooooo00?o0000ooooooooool00_ooool01Ol0003ooooooooooooooooo000000Kooooo0_l0
0005ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool01oooool2o`000003ooooool0
003ooooo00Oooooo00Co0000oooooooooooo00000_ooool00ol0003oooooooooo`0Aooooo`?o0000
=_ooool00ol0003oooooooooo`1jooooo`005?ooool00ol0003oooooooooo`02ooooo`03o`000?oo
oooooooo00;ooooo00Co0000oooooooooooo00002Oooool2o`0000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`05ooooo`;o00000oooool00ol0003oooooooooo`05ooooo`05o`00
0?ooooooooooooooool000000_ooool00ol0003oooooooooo`0Dooooo`?o0000<_ooool00ol0003o
ooooooooo`1kooooo`005?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooool000002oooool00ol0003oooooo`00000;ooooo`03o`000?oooooooooo00;ooooo
0_l00006ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00_ooool01?l0003ooooo
ooooool0000Iooooo`?o0000;_ooool00ol0003oooooooooo`1looooo`004oooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02oooool2o`0000_ooooo
00?o0000ooooool000000_l00009ooooo`05o`000?ooooooooooooooool000001?ooool00ol0003o
ooooooooo`02ooooo`03o`000?oooooooooo01Wooooo0ol0000[ooooo`03o`000?oooooooooo07co
oooo000Cooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool00oooool2o`0000kooooo
0ol00009ooooo`;o00003Oooool00ol0003oooooo`000005ooooo`03o`000?oooooooooo00;ooooo
00?o0000ooooooooool07?ooool3o`0002Oooooo00?o0000ooooooooool0OOooool001?ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`0000gooooo
00?o0000ooooooooool00_l00005ooooo`?o00003_ooool2o`0000Kooooo00?o0000ooooooooool0
0_ooool00ol0003oooooooooo`0Oooooo`?o00008oooool00ol0003oooooooooo`1nooooo`004ooo
ool00ol0003oooooooooo`04ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`03ooooo`;o000000?oooooo`000?l000000_ooool00ol0003ooooo
ooooo`0=ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003ooooo
ooooo`0Rooooo`?o00008?ooool00ol0003oooooooooo`1nooooo`004_ooool00ol0003ooooooooo
o`05ooooo`03o`000?oooooo000000Gooooo00?o0000ooooooooool02Oooool00ol0003ooooooooo
o`03ooooo`?o00001?ooool00ol0003oooooooooo`0<ooooo`03o`000?oooooo000000Cooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`0Uooooo`?o00007?ooool00ol0003ooooooooo
o`1oooooo`004_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooo000000Kooooo00?o0000
ooooooooool02?ooool01?l0003oooooooooooooool2o`0000?ooooo0_l00003ooooo`03o`000?oo
oooooooo00[ooooo00Co0000oooooooooooo00001?ooool00ol0003oooooooooo`03ooooo`03o`00
0?oooooooooo02Sooooo0ol0000Hooooo`03o`000?oooooooooo083ooooo000Booooo`03o`000?oo
oooooooo00Gooooo0_l00008ooooo`03o`000?oooooooooo00Oooooo00Co0000ooooool0003o0000
1oooool2o`000003ooooool0003ooooo00[ooooo00?o0000ooooooooool00_ooool01Ol0003ooooo
oooooooooooo000000Kooooo00?o0000ooooooooool0:_ooool3o`0001Gooooo00?o0000oooooooo
ool0P?ooool0017ooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`08ooooo`03o`00
0?oooooooooo00Gooooo0ol0000;ooooo`;o00002_ooool00ol0003oooooooooo`04ooooo`03o`00
0?oooooo000000Oooooo00?o0000ooooooooool0;Oooool3o`00017ooooo00?o0000ooooooooool0
POooool0017ooooo00?o0000ooooooooool01Oooool00ol0003oooooo`000009ooooo`03o`000?oo
oooooooo00?ooooo0ol0000>ooooo`;o00002?ooool00ol0003oooooooooo`06ooooo`;o00001ooo
ool00ol0003oooooooooo`0`ooooo`?o00003Oooool00ol0003oooooooooo`22ooooo`004Oooool0
0ol0003oooooooooo`04ooooo`04o`000?ooooooooooo`0000[ooooo00?o0000ooooooooool00_l0
0002ooooo`03o`000?oooooooooo00cooooo00Co0000ooooool0003o00001_ooool00ol0003ooooo
ooooo`06ooooo`;o00001oooool00ol0003oooooooooo`0cooooo`?o00001_ooool4o`0008Gooooo
000Aooooo`03o`000?oooooooooo00?ooooo00Go0000ooooooooooooooooo`00000;ooooo`;o0000
1?ooool00ol0003oooooooooo`0<ooooo`04o`000?ooooooooooooooo`;o00000oooool00ol0003o
ooooooooo`08ooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0=_ooool3o`0000;o
oooo1Ol00025ooooo`004?ooool00ol0003oooooooooo`04ooooo`05o`000?ooooooooooooooool0
00002Oooool3o`0000Cooooo00?o0000ooooooooool03_ooool00ol0003oooooooooo`02ooooo`?o
00002_ooool00ol0003oooooo`000007ooooo`03o`000?oooooooooo03Sooooo2?l00024ooooo`00
4?ooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool0
1?ooool3o`0000?ooooo00Go0000ooooooooooooooooo`00000@ooooo`03o`000?oooooooooo00?o
oooo0ol00009ooooo`03o`000?oooooo000000Oooooo00?o0000ooooooooool0>Oooool7o`0008Co
oooo000@ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003ooooo
ooooo`02ooooo`;o00001oooool01?l0003oooooooooool0000@ooooo`03o`000?oooooooooo00;o
oooo00Go0000ooooooooooooooooo`000008ooooo`04o`000?ooooooooooo`0000Kooooo00?o0000
ooooooooool0>Oooool7o`0008Cooooo000?ooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool01Oooool01?l0003oooooo`000?l00009ooooo`04o`000?ooooooooooo`00013ooooo00Go
0000ooooooooooooooooo`000005ooooo`;o00001_ooool01Ol0003ooooooooooooooooo000000Go
oooo00?o0000ooooooooool0>_ooool6o`0008Cooooo000?ooooo`05o`000?ooooooooooooooool0
00002?ooool2o`0000cooooo0_l0000Aooooo`05o`000?ooooooooooooooool000001oooool2o`00
00?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo
03_ooooo0ol00002ooooo`03o`000?oooooooooo087ooooo000?ooooo`04o`000?ooooooooooo`00
00Kooooo1?l0000>ooooo`03o`000?oooooooooo013ooooo00?o0000ooooool000002_ooool2o`00
0003ooooool0003ooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0koooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool0P?ooool000oooooo00Co0000oooooooo
oooo00001?ooool2o`0000?ooooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0@oooo
o`;o00003Oooool2o`0000Kooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0joooo
o`03o`000?oooooooooo00Cooooo00?o0000ooooooooool0Ooooool000kooooo00Co0000oooooooo
oooo00000oooool2o`0000Gooooo00?o0000ooooooooool03?ooool2o`00017ooooo00?o0000oooo
ooooool03?ooool00ol0003oooooo`000006ooooo`05o`000?ooooooooooooooool00000??ooool0
0ol0003oooooooooo`05ooooo`03o`000?oooooooooo07kooooo000<ooooo`?o000000Coooooo`00
0?oooooooooo0_l00007ooooo`03o`000?oooooooooo00_ooooo00Co0000oooooooooooo00003ooo
ool2o`0000kooooo00?o0000ooooooooool00_l00004ooooo`05o`000?ooooooooooooooool00000
??ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo07gooooo000;ooooo`So00002_oo
ool00ol0003oooooooooo`0:ooooo`05o`000?ooooooooooooooool000003Oooool00ol0003ooooo
o`00000>ooooo`03o`000?oooooooooo00;ooooo0_l00003ooooo`04o`000?ooooooooooo`0003co
oooo00?o0000ooooooooool01oooool00ol0003oooooooooo`1looooo`002_ooool7o`0000cooooo
00?o0000ooooooooool02_ooool01Ol0003ooooooooooooooooo000000cooooo00Co0000oooooooo
oooo00003_ooool00ol0003oooooooooo`04ooooo`;o00000_ooool00ol0003oooooo`00000koooo
o`03o`000?oooooooooo00Wooooo00?o0000ooooooooool0Noooool000[ooooo4Ol00002ooooo`03
o`000?oooooooooo00[ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`09ooooo`05
o`000?ooooooooooooooool000003?ooool00ol0003oooooooooo`07ooooo`;o000000?oooooo`00
0?l000000_l0000iooooo`03o`000?oooooooooo00[ooooo00?o0000ooooooooool0N_ooool000[o
oooo1ol0000:oooooa[o00001Oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00[o
oooo00?o0000ooooooooool02Oooool6o`0003Sooooo00?o0000ooooooooool02oooool00ol0003o
ooooooooo`1iooooo`002oooool5o`000003ooooool0003o000000[ooooo00?o0000ooooooooool0
2Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo6_l00009ooooo`Oo0000
=oooool00ol0003oooooooooo`0<ooooo`03o`000?oooooooooo07Sooooo000<ooooo`Co00000ooo
ool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02_oo
ool00ol0003oooooooooo`0@o`0003Oooooo00?o0000ooooooooool03Oooool00ol0003ooooooooo
o`1gooooo`004?ooool01?l0003oooooooooooooool2o`0000Sooooo00?o0000ooooooooool02?oo
ool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01Ooo
ool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00[ooooo1ol0000fooooo`03o`000?oo
oooooooo00oooooo00?o0000ooooooooool0M_ooool0013ooooo00?o0000ooooooooool00oooool2
o`0000Kooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`07ooooo`05o`000?oooooo
ooooooooool000002?ooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Sooooo0_l0
00000ooooooo0000o`000003o`0003Oooooo00?o0000ooooooooool04?ooool00ol0003ooooooooo
o`1eooooo`004Oooool00ol0003oooooooooo`04ooooo`;o00001?ooool00ol0003oooooooooo`07
ooooo`03o`000?oooooooooo00Sooooo00Co0000oooooooooooo00002Oooool00ol0003ooooooooo
o`08ooooo`03o`000?oooooooooo00Gooooo0ol00003ooooo`Co0000>?ooool00ol0003ooooooooo
o`0Aooooo`;o0000MOooool001;ooooo00?o0000ooooooooool01Oooool2o`0000;ooooo00?o0000
ooooooooool01oooool00ol0003oooooooooo`09ooooo`;o00002_ooool00ol0003oooooooooo`07
ooooo`03o`000?oooooooooo00Cooooo0_l00005ooooo`03o`000?oooooooooo03[ooooo00?o0000
ooooooooool04oooool00ol0003oooooooooo`1booooo`004_ooool00ol0003oooooooooo`07oooo
o`?o00002Oooool00ol0003oooooooooo`09ooooo`03o`000?oooooooooo00Wooooo00?o0000oooo
ooooool01oooool00ol0003oooooooooo`02ooooo`;o00001_ooool00ol0003oooooooooo`0koooo
o`03o`000?oooooooooo01Cooooo00?o0000ooooooooool0LOooool001?ooooo00?o0000oooooooo
ool02?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Wooooo00?o0000ooooool0
00002_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00;o00001oooool00ol0003o
ooooooooo`0kooooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool0L?ooool001Cooooo
00?o0000ooooooooool02?ooool2o`0000Kooooo00?o0000ooooooooool02?ooool01Ol0003ooooo
oooooooooooo000000Wooooo00?o0000ooooooooool01oooool3o`0000Sooooo00?o0000oooooooo
ool0??ooool00ol0003oooooooooo`0Gooooo`03o`000?oooooooooo06oooooo000Eooooo`03o`00
0?oooooooooo00Oooooo00Co0000ooooool0003o00001?ooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00?ooooo00?o0000ooooooooool01oooool00ol0003oooooooooo`04ooooo`?o
00002_ooool00ol0003oooooooooo`0looooo`03o`000?oooooooooo01Sooooo00?o0000oooooooo
ool0K_ooool001Gooooo00?o0000ooooooooool01oooool01?l0003oooooooooooooool2o`000003
ooooool0003ooooo00Wooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`07ooooo`03
o`000?oooooooooo00;ooooo0_l000000ooooooo0000ooooo`09ooooo`03o`000?oooooooooo03go
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`1]ooooo`005_ooool00ol0003ooooo
ooooo`06ooooo`03o`000?oooooooooo00?ooooo0_l00009ooooo`03o`000?oooooooooo00Gooooo
00?o0000ooooooooool01_ooool00ol0003oooooooooo`02o`0000?ooooo00?o0000ooooooooool0
1oooool00ol0003oooooooooo`0nooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
K?ooool001Oooooo00?o0000ooooooooool01Oooool00ol0003oooooooooo`04ooooo`?o00001_oo
ool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo00Gooooo0ol00005ooooo`03o`000?oo
oooooooo00Kooooo00?o0000ooooooooool0?_ooool00ol0003oooooooooo`0Looooo`03o`000?oo
oooooooo06_ooooo000Hooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool00oooool0
1?l0003oooooooooool00004ooooo`03o`000?oooooooooo00Sooooo00?o0000ooooooooool01?oo
ool2o`0000Kooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0oooooo`03o`000?oo
oooooooo01gooooo00?o0000ooooooooool0J_ooool001Sooooo00?o0000ooooooooool01Oooool0
0ol0003oooooooooo`02ooooo`03o`000?oooooooooo00;ooooo0_l000000ooooooo0000ooooo`0;
ooooo`03o`000?oooooooooo00?o000000?oooooo`000?ooool01Oooool00ol0003oooooooooo`05
ooooo`03o`000?oooooooooo043ooooo00?o0000ooooooooool07_ooool00ol0003oooooooooo`1Y
ooooo`006Oooool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo00;ooooo00?o0000oooo
ooooool01?ooool2o`0000gooooo0_l00005ooooo`03o`000?oooooooooo00?ooooo00?o0000oooo
ooooool01?ooool00ol0003oooooooooo`11ooooo`03o`000?oooooooooo01oooooo00?o0000oooo
ooooool0J?ooool001[ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00Cooooo00Co0000ooooool0003o00002Oooool3o`0000Kooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo047ooooo00?o0000oooo
ooooool08?ooool00ol0003oooooooooo`1Wooooo`006_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`;o
00001Oooool2o`0000?ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00Cooooo00?o0000ooooooooool0@Oooool00ol0003oooooooooo`0Rooooo`03
o`000?oooooooooo06Kooooo000Kooooo`03o`000?oooooooooo00;ooooo00Go0000oooooooooooo
ooooo`000005ooooo`03o`000?oooooooooo00Gooooo0_l000000ooooooo0000o`000006ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo04;ooooo00?o0000ooooooooool08oooool00ol0003oooooooooo`1Uooooo`00
7?ooool00ol0003oooooooooo`02ooooo`04o`000?ooooooooooo`0000Cooooo00?o0000oooooooo
ool01_ooool3o`0000Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo00;ooooo00?o0000ooooooooool0@oooool00ol0003oooooooooo`0Tooooo`03
o`000?oooooooooo06Cooooo000Mooooo`05o`000?ooooooooooooooool000000_ooool01Ol0003o
oooooooooooooooo000000Oooooo0_l00003ooooo`;o00001oooool01Ol0003ooooooooooooooooo
000000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`14ooooo`03o`000?oooooo
oooo02Gooooo0_l0001Tooooo`007Oooool01Ol0003ooooooooooooooooo000000;ooooo00Co0000
oooooooooooo00001_ooool2o`0000Oooooo0_l00006ooooo`05o`000?ooooooooooooooool00000
0_ooool01Ol0003ooooooooooooooooo000004Oooooo00?o0000ooooooooool09oooool00ol0003o
ooooooooo`1Qooooo`007_ooool01_l0003oooooooooool0003oooooo`0000?ooooo00?o0000oooo
ooooool00_ooool2o`0000_ooooo0_l00005ooooo`04o`000?ooooooooooo`0000;ooooo00Go0000
ooooooooooooooooo`000016ooooo`03o`000?oooooooooo02Wooooo00?o0000ooooooooool0H?oo
ool001oooooo00Go0000ooooool0003oooooo`000002ooooo`04o`000?ooooooooooooooo`;o0000
3oooool2o`0000?ooooo00Co0000oooooooooooo00000_ooool01?l0003oooooooooool00017oooo
o`03o`000?oooooooooo02[ooooo00?o0000ooooooooool0Goooool001oooooo00?o0000ooooool0
00000_l000000ooooooo0000ooooo`03o`0001?ooooo0_l00002ooooo`05o`000?oooooo0000oooo
ool000000_ooool00ol0003oooooooooo`16ooooo`03o`000?oooooooooo02_ooooo00?o0000oooo
ooooool0G_ooool0023ooooo1ol0000Hooooo`03o`000?oooooooooo00Co000000?oooooo`000?oo
ool0B?ooool00ol0003oooooooooo`0/ooooo`03o`000?oooooooooo05gooooo000Oooooo`Oo0000
6_ooool7o`0004[ooooo00?o0000ooooooooool0;Oooool00ol0003oooooooooo`1Looooo`007ooo
oolEo`0000gooooo1ol00018ooooo`03o`000?oooooooooo02oooooo00?o0000ooooooooool0Fooo
ool001oooooo1ol0000>oooooaCo0000B?ooool00ol0003oooooooooo`0`ooooo`03o`000?oooooo
oooo05[ooooo000Pooooo`Go00007?ooool7o`0004Sooooo00?o0000ooooooooool0<Oooool00ol0
003oooooooooo`1Iooooo`008Oooool3o`0001kooooo1Ol00019ooooo`03o`000?oooooooooo03;o
oooo00?o0000ooooooooool0F?ooool004?ooooo0ol0001:ooooo`03o`000?oooooooooo03?ooooo
00?o0000ooooooooool0Eoooool008oooooo00?o0000ooooooooool0=Oooool00ol0003ooooooooo
o`1Fooooo`00Soooool00ol0003oooooooooo`0fooooo`03o`000?oooooooooo05Gooooo002?oooo
o`03o`000?oooooooooo03Oooooo1?l0001Cooooo`00Soooool00ol0003oooooooooo`0gooooo`Go
0000D_ooool008oooooo00?o0000ooooooooool0=_ooool7o`00057ooooo002>ooooo`03o`000?oo
oooooooo03Kooooo2?l0001Aooooo`00S_ooool00ol0003oooooooooo`0cooooo`?o000000?ooooo
o`000?l000001Ol0001Aooooo`00S_ooool00ol0003oooooooooo`0`ooooo`?o00001Oooool5o`00
05;ooooo002>ooooo`03o`000?oooooooooo02gooooo0ol00009ooooo`?o0000Doooool008kooooo
00?o0000ooooooooool0:_ooool3o`0006;ooooo002=ooooo`03o`000?oooooooooo02Wooooo0_l0
001Uooooo`00SOooool00ol0003oooooooooo`0Vooooo`?o0000Ioooool008gooooo00?o0000oooo
ooooool08oooool3o`0006[ooooo002=ooooo`03o`000?oooooooooo023ooooo0ol0001]ooooo`00
SOooool00ol0003oooooooooo`0Mooooo`?o0000L?ooool008cooooo00?o0000ooooooooool06ooo
ool3o`0007?ooooo002<ooooo`03o`000?oooooooooo01Wooooo0_l0001fooooo`00S?ooool00ol0
003oooooooooo`0Fooooo`?o0000N?ooool008cooooo00?o0000ooooooooool04oooool3o`0007_o
oooo002<ooooo`03o`000?oooooooooo013ooooo0ol0001nooooo`00Roooool00ol0003ooooooooo
o`0>ooooo`?o0000POooool008_ooooo00?o0000ooooooooool03?ooool2o`0008Cooooo002;oooo
o`03o`000?oooooooooo00Wooooo0ol00026ooooo`00Roooool00ol0003oooooooooo`06ooooo`?o
0000ROooool008Wooooo0ol00005ooooo`?o0000S?ooool008Sooooo1Ol000001?oooooo0000o`00
0?l0002?ooooo`00Qoooool7o`0009;ooooo0027ooooo`Oo0000T_ooool008Oooooo1ol0002Boooo
o`00R?ooool5o`0009?ooooo0029ooooo`?o0000U?ooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool0
0?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0351667, -0.0500052, \
0.00378109, 0.00363322}},
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1152}, {0, 701}},
WindowSize->{520, 578},
WindowMargins->{{Automatic, 281}, {31, Automatic}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:2"->{
    Cell[1776, 53, 82, 1, 38, "ObjectName",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[1861, 56, 954, 28, 86, "Usage",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[2818, 86, 544, 16, 62, "Notes",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3387, 106, 78, 1, 38, "ExampleSection",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3490, 111, 151, 3, 26, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3644, 116, 52079, 847, 294, 2952, 235, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[55738, 966, 148, 4, 24, "Text",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[55911, 974, 430, 8, 138, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[56344, 984, 40671, 918, 144, 6127, 486, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97030, 1905, 119, 1, 24, "Text",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97174, 1910, 153, 3, 26, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97330, 1915, 52079, 847, 294, 2952, 235, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}]},
  "LocalInteractions:5"->{
    Cell[1776, 53, 82, 1, 38, "ObjectName",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[1861, 56, 954, 28, 86, "Usage",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[2818, 86, 544, 16, 62, "Notes",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3387, 106, 78, 1, 38, "ExampleSection",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3490, 111, 151, 3, 26, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[3644, 116, 52079, 847, 294, 2952, 235, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[55738, 966, 148, 4, 24, "Text",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[55911, 974, 430, 8, 138, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[56344, 984, 40671, 918, 144, 6127, 486, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97030, 1905, 119, 1, 24, "Text",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97174, 1910, 153, 3, 26, "Input",
      CellTags->{"b:2", "LocalInteractions:5"}],
    Cell[97330, 1915, 52079, 847, 294, 2952, 235, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:2", "LocalInteractions:5"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:2", 150069, 2783},
  {"LocalInteractions:5", 151358, 2811}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 82, 1, 38, "ObjectName",
  CellTags->{"b:2", "LocalInteractions:5"}],
Cell[1861, 56, 954, 28, 86, "Usage",
  CellTags->{"b:2", "LocalInteractions:5"}],
Cell[2818, 86, 544, 16, 62, "Notes",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{
Cell[3387, 106, 78, 1, 38, "ExampleSection",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{
Cell[3490, 111, 151, 3, 26, "Input",
  CellTags->{"b:2", "LocalInteractions:5"}],
Cell[3644, 116, 52079, 847, 294, 2952, 235, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]],
Cell[55738, 966, 148, 4, 24, "Text",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{
Cell[55911, 974, 430, 8, 138, "Input",
  CellTags->{"b:2", "LocalInteractions:5"}],
Cell[56344, 984, 40671, 918, 144, 6127, 486, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]],
Cell[97030, 1905, 119, 1, 24, "Text",
  CellTags->{"b:2", "LocalInteractions:5"}],

Cell[CellGroupData[{
Cell[97174, 1910, 153, 3, 26, "Input",
  CellTags->{"b:2", "LocalInteractions:5"}],
Cell[97330, 1915, 52079, 847, 294, 2952, 235, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:2", "LocalInteractions:5"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

