(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    559546,      31936]*)
(*NotebookOutlinePosition[    572266,      32214]*)
(*  CellTagsIndexPosition[    572150,      32207]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["LocalInteractionModel", "ObjectName",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] ",
  Cell[BoxData[
      RowBox[{
        StyleBox["LocalInteractionModel",
          "MR"], 
        StyleBox["[",
          "MR"], 
        StyleBox[\(opts ... \),
          "TI"], 
        StyleBox["]",
          "MR"]}]]],
  " creates a local interaction model using the specified options"
}], "Usage",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
\[FilledSmallSquare] There are many options available\
\>", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[Cell[BoxData[GridBox[{
        {\(InteractionNeighborhood\), \(GridGraph[10, 10]\), 
          
          StyleBox[\(Graph\ specifying\ the\ interaction\ neighborhood\ for\ \
the\ local\ interaction\ model\),
            FontFamily->"Times"]},
        {"UpdateNeighborhood", \(GridGraph[10, 10]\), 
          
          StyleBox[\(Graph\ specifying\ the\ update\ neighborhood\ for\ the\ \
local\ interaction\ model\),
            FontFamily->"Times"]},
        {"PayoffMatrix", \({{{2, 1}, {0, 0}}, {{0, 0}, {1, 2}}}\), 
          StyleBox[\(Payoff\ matrix\ assigned\ to\ each\ agent\),
            FontFamily->"Times"]},
        {"LearningRule", "ImitateBestNeighbor", 
          StyleBox[\(Learning\ rule\ assigned\ to\ each\ agent\),
            FontFamily->"Times"]},
        {"StrategyColors", \({Red, Green}\), 
          StyleBox[\(Colors\ used\ to\ represent\ each\ strategy\),
            FontFamily->"Times"]},
        {\(InitialStrategyDistribution\), "Uniform", 
          RowBox[{
            
            StyleBox[\(A\ list\ of\ strategies\ to\ be\ assigned\ to\ each\ \
agent\),
              FontFamily->"Times"], 
            StyleBox[",",
              FontFamily->"Times"], 
            StyleBox[" ",
              FontFamily->"Times"], 
            RowBox[{
              StyleBox["or",
                FontFamily->"Times"], 
              StyleBox[" ",
                FontFamily->"Times"], 
              StyleBox["the",
                FontFamily->"Times"], 
              StyleBox[" ",
                FontFamily->"Times"], 
              StyleBox["symbol",
                FontFamily->"Times"], 
              StyleBox[" ",
                FontFamily->"Times"], 
              StyleBox["Uniform",
                FontFamily->"Courier"]}], 
            StyleBox[",",
              FontFamily->"Times"], 
            StyleBox[" ",
              FontFamily->"Times"], 
            
            StyleBox[\(indicating\ that\ strategies\ are\ to\ be\ drawn\ from\
\ a\ uniform\ random\ distribution\ over\ the\ strategy\ \(\(space\)\(.\)\)\),
              
              FontFamily->"Times"]}]},
        {"RandomGenerator", "Java", 
          StyleBox[\(Which\ random\ number\ generator\ to\ use\),
            FontFamily->"Times"]},
        {"RandomSeed", "None", 
          StyleBox[\(Initial\ seed\ for\ the\ random\ number\ generator\),
            FontFamily->"Times"]},
        {"GraphicsEngine", "Java", 
          
          StyleBox[\(The\ graphics\ engine\ to\ use\ when\ displaying\ graphs\
\ in\ the\ view\ window\),
            FontFamily->"Times"]},
        {"NeighborhoodsToDraw", "Automatic", 
          StyleBox[\(Neighborhoods\ to\ include\ in\ the\ view\ window\),
            FontFamily->"Times"]},
        {"NeighborhoodStyle", "Automatic", 
          
          StyleBox[\(How\ to\ decide\ if\ the\ specified\ neighborhoods\ are\ \
directed\ or\ undirected\),
            FontFamily->"Times"]},
        {\(InteractionNeighborhoodColor\), "Black", 
          StyleBox[\(Color\ of\ the\ interaction\ neighborhood\),
            FontFamily->"Times"]},
        {\(UpdateNeighborhoodColor\), "Blue", 
          StyleBox[\(Color\ of\ the\ update\ neighborhood\),
            FontFamily->"Times"]},
        {\(UseInteractionNeighborhoodWeights\), "False", 
          
          StyleBox[\(Interpret\ weights\ assigned\ to\ edges\ in\ the\ \
interaction\ neighborhood\ as\ probabilites\),
            FontFamily->"Times"]},
        {\(UseUpdateNeighborhoodWeights\), "False", 
          
          StyleBox[\(Interpret\ weights\ assigned\ to\ edges\ in\ the\ \
interaction\ neighborhood\ as\ probabilites\),
            FontFamily->"Times"]},
        {"ShowVertexNumbers", "False", 
          
          StyleBox[\(Whether\ the\ view\ window\ should\ display\ the\ vertex\
\ number\),
            FontFamily->"Times"]},
        {"ShowWindow", "True", 
          
          StyleBox[\(Whether\ to\ construct\ the\ view\ window\ for\ the\ \
model\),
            FontFamily->"Times"]},
        {"WindowLabel", Cell["\"Window Label\""], 
          StyleBox[\(String\ for\ the\ title\ of\ the\ view\ window\),
            FontFamily->"Times"]}
        }]]]], "3ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] The default settings construct a local interaction \
model on a ",
  Cell[BoxData[
      \(TraditionalForm\`10\[Cross]\ 10\)]],
  " von Neumann lattice with the payoff matrix for ",
  StyleBox["Battle of the Sexes",
    FontSlant->"Italic"],
  ", using Red to denote the strategy \"Watch Boxing\" and Green to denote \
the strategy \"Go to Ballet\"."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
\[FilledSmallSquare] For each edge in the interaction neighborhood, a toss of \
a fair coin determines which agent is Row and which agent is Column. See the \
discussion on NeighborhoodStyle below for more details.\
\>", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
\[FilledSmallSquare] If only the interaction neighborhood is specified, the \
update neighborhood will automatically be taken to be the same as the \
interaction neighborhood.\
\>", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\[FilledSmallSquare] See also Evolve.", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["PayoffMatrix", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
\[FilledSmallSquare] Setting PayoffMatrix using a rule insures all agents in \
the model have the same payoff matrix, since agents are assigned a payoff \
matrix using a technique similar to the following:\
\>", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    RowBox[{"r", "=", 
      RowBox[{"PayoffMatrix", "\[Rule]", 
        RowBox[{"(", GridBox[{
              {\({1, 2}\), \({0, 0}\)},
              {\({0, 0}\), \({2, 1}\)}
              }], ")"}]}]}]], "Input",
  CellLabel->"In[5]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \(PayoffMatrix \[Rule] {{{1, 2}, {0, 0}}, {{0, 0}, {2, 1}}}\)], "Output",
  CellLabel->"Out[5]=",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[PayoffMatrix, {5}] /. r\)], "Input",
  CellLabel->"In[11]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \({{{{1, 2}, {0, 0}}, {{0, 0}, {2, 1}}}, {{{1, 2}, {0, 0}}, {{0, 0}, {2, 
            1}}}, {{{1, 2}, {0, 0}}, {{0, 0}, {2, 1}}}, {{{1, 2}, {0, 
            0}}, {{0, 0}, {2, 1}}}, {{{1, 2}, {0, 0}}, {{0, 0}, {2, 
            1}}}}\)], "Output",
  CellLabel->"Out[11]=",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell["\<\
\[FilledSmallSquare] If PayoffMatrix is set using a deferred rule, though, \
each agent can be assigned a different payoff matrix if random functions are \
included in the payoffs..\
\>", "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[{\(\[Alpha] := 
        Random[];\), "\[IndentingNewLine]", \(\[Epsilon] :=  .1*
          Random[];\), "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{"r", "=", 
        RowBox[{"PayoffMatrix", "\[RuleDelayed]", 
          RowBox[{"(", GridBox[{
                {\({1 + \[Alpha], 
                    2 + \[Alpha]}\), \({\[Epsilon], \[Epsilon]}\)},
                {\({\[Epsilon], \[Epsilon]}\), \({2 + \[Alpha], 
                    1 + \[Alpha]}\)}
                }], ")"}]}]}], ";"}]}], "Input",
  CellLabel->"In[12]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[PayoffMatrix, {5}] /. r\)], "Input",
  CellLabel->"In[15]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \({{{{1.1730356203815646`, 2.501020643059333`}, {0.042989105356976226`, 
            0.03433457713112902`}}, {{0.0030008769455020495`, 
            0.022314035869498035`}, {2.365479623061673`, 
            1.0882710408314664`}}}, {{{1.0122211031970851`, 
            2.0840638009797976`}, {0.0059705862529670735`, 
            0.08511296125596876`}}, {{0.06192969826734355`, 
            0.06801282948633444`}, {2.910124088331551`, 
            1.5268837892059004`}}}, {{{1.3674507708452566`, 
            2.7683529356489993`}, {0.03850906082996764`, 
            0.07573438640410508`}}, {{0.035421572025979765`, 
            0.012400396991205972`}, {2.352065564566428`, 
            1.7529903642446136`}}}, {{{1.181180099878233`, 
            2.6229833268527267`}, {0.09221745109966656`, 
            0.04096445929333237`}}, {{0.015117133042321252`, 
            0.03998429681577464`}, {2.5566948879349924`, 
            1.3213735521018573`}}}, {{{1.1389502272261274`, 
            2.315779167177949`}, {0.04969890254053218`, 
            0.04702439395421697`}}, {{0.051965324455269184`, 
            0.06356508723146043`}, {2.586864937073771`, 
            1.9433601503362694`}}}}\)], "Output",
  CellLabel->"Out[15]=",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["LearningRule", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
The learning rule must initially be one of the following two options. Custom \
learning rules can be defined after the model has been initialized.\
\>", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[Cell[BoxData[
    FormBox[GridBox[{
          {"ImitateBestNeighbor", Cell["\<\
Adopt the strategy of the best performing individual in your update 
neighborhood, flipping a coin to decide which strategy to adopt in the case \
of ties.\
\>"]},
          {"BestResponse", Cell["\<\
Adopt the strategy which provides the best response to the current \
distribution of strategies held by individuals in your interaction \
neighborhood, flipping a coin to decide which strategy to adopt in the case \
of ties.\
\>"]}
          }], TraditionalForm]]]], "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["RandomGenerator", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[Cell[BoxData[
    FormBox[GridBox[{
          {
            StyleBox["Java",
              FontFamily->"Courier",
              FontSlant->"Plain"], Cell["\<\
Use the standard Java Random class for all randomization in the model.\
\>"]},
          {
            StyleBox["Mathematica",
              FontFamily->"Courier",
              FontSlant->"Plain"], Cell[TextData[{
              "Use callbacks into ",
              StyleBox["Mathematica",
                FontSlant->"Italic"],
              
              " for all randomization in the model. This can slow the model \
down significantly if many callbacks need to be made."
            }]]}
          }], TraditionalForm]]]], "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["RandomSeed", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] If ",
  StyleBox["None",
    FontFamily->"Courier"],
  ", no random seed is specified.\n\[FilledSmallSquare] If set to an integer, \
that integer is used to seed the random number generator. If  ",
  StyleBox["RandomGenerator",
    FontFamily->"Courier"],
  " was set to ",
  StyleBox["Mathematica",
    FontFamily->"Courier"],
  ",",
  " the model makes a callback into ",
  StyleBox["Mathematica, ",
    FontSlant->"Italic"],
  "invoking ",
  StyleBox["SeedRandom",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox["[",
    FontWeight->"Bold"],
  StyleBox["integer",
    FontSlant->"Italic"],
  StyleBox["].",
    FontWeight->"Bold"]
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["GraphicsEngine", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[Cell[BoxData[
    FormBox[GridBox[{
          {
            StyleBox["Java",
              FontFamily->"Courier",
              FontSlant->"Plain"], Cell[TextData[{
              
              "Use a Java class for drawing the view window. This can be much \
faster than using ",
              StyleBox["Mathematica",
                FontSlant->"Italic"],
              
              ". However, it does not provide the same degree of control over \
the output."
            }]]},
          {
            StyleBox["Mathematica",
              FontFamily->"Courier",
              FontSlant->"Plain"], Cell[TextData[{
              "Use ",
              StyleBox["Mathematica",
                FontSlant->"Italic"],
              
              " for displaying the view window. Although slower, the full \
range of display options for ",
              StyleBox["Combinatorica",
                FontSlant->"Italic"],
              " graphics can be used."
            }]]}
          }], TraditionalForm]]]], "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["NeighborhoodsToDraw", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] Valid settings include any subset of ",
  StyleBox["{Interaction,Update}",
    FontFamily->"Courier"],
  " (including the empty set), or ",
  StyleBox["Automatic",
    FontFamily->"Courier"],
  ".\n\[FilledSmallSquare] If ",
  StyleBox["Automatic,",
    FontFamily->"Courier"],
  " the drawn neighborhood depends on which neighborhoods were explicitly \
specified. If the only specified neighborhood is the interaction neighborhood \
(which means that the update neighborhood defaults to the same structure as \
well), then only the interaction neighborhood is drawn. If both an \
interaction and update neighborhood are given, then both interaction and \
update neighborhoods are drawn.\n\[FilledSmallSquare] If ",
  StyleBox["NeighborhoodsToDraw",
    FontFamily->"Courier"],
  " equals the empty set, only the vertices of the interaction neighborhood \
will be shown."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["NeighborhoodStyle", "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] If ",
  StyleBox["Automatic",
    FontFamily->"Courier"],
  ", neighborhoods will be treated as directed or undirected according to \
whether the specified graph is directed or undirected."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] If ",
  StyleBox["NeighborhoodStyle",
    FontFamily->"Courier"],
  " is set to ",
  StyleBox["Directed",
    FontFamily->"Courier"],
  " or ",
  StyleBox["Undirected",
    FontFamily->"Courier"],
  ", graphs will be interpreted in that way."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] A ",
  StyleBox["Directed",
    FontFamily->"Courier"],
  " interaction neighborhood allows for asymmetric interactions: i.e., for \
each edge connecting ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " to ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  ",  ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " interacts with ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " and only ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " receives a payoff. Roles of Row and Column are still assigned at random."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] A ",
  StyleBox["Undirected",
    FontFamily->"Courier"],
  " interaction neighborhood means all interactions are symmetric: i.e., an \
edge connecting ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " to ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " means that, for each edge in the interaction neighborhood, a coin flip \
determines who is Row and who is Column. The interaction occurs, and both ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " receive payoffs."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] A ",
  StyleBox["Directed",
    FontFamily->"Courier"],
  " update neighborhood means that asymmetric learning can occur: i.e., if \
the edge ",
  Cell[BoxData[
      \(TraditionalForm\`{A, B}\)]],
  " exists but ",
  Cell[BoxData[
      \(TraditionalForm\`{B, A}\)]],
  " does not, ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " can learn from ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " during the update phase but ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " cannot learn from ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  "."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[TextData[{
  "\[FilledSmallSquare] A ",
  StyleBox["Undirected",
    FontFamily->"Courier"],
  " update neighborhood means learning is symmetric i.e., if the edge ",
  Cell[BoxData[
      \(TraditionalForm\`{A, B}\)]],
  " exists, ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " can learn from ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " during the update phase and ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " can learn from ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " as well."
}], "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "SubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["The following contructs a basic local interaction model.", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[{
    \(<< LocalInteractions`LocalInteractions`\), "\[IndentingNewLine]", 
    \(\(m1 = LocalInteractionModel[];\)\)}], "Input",
  CellLabel->"In[1]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
The default model is not interesting as all strategies are initially assigned \
to 1. We can change this by explicitly specifying the \
InitialStrategyDistribution.\
\>", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \(\(m2 = 
        LocalInteractionModel[\[IndentingNewLine]InitialStrategyDistribution \
\[Rule] Table[
              Random[Integer, {1, 
                  2}], {100}]\[IndentingNewLine]];\)\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["Here is a picture of the resulting model:", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowGraph[m2];\)\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.04545 .04545 m
.14646 .04545 L
s
.04545 .04545 m
.04545 .14646 L
s
.14646 .04545 m
.24747 .04545 L
s
.14646 .04545 m
.14646 .14646 L
s
.24747 .04545 m
.34848 .04545 L
s
.24747 .04545 m
.24747 .14646 L
s
.34848 .04545 m
.44949 .04545 L
s
.34848 .04545 m
.34848 .14646 L
s
.44949 .04545 m
.55051 .04545 L
s
.44949 .04545 m
.44949 .14646 L
s
.55051 .04545 m
.65152 .04545 L
s
.55051 .04545 m
.55051 .14646 L
s
.65152 .04545 m
.75253 .04545 L
s
.65152 .04545 m
.65152 .14646 L
s
.75253 .04545 m
.85354 .04545 L
s
.75253 .04545 m
.75253 .14646 L
s
.85354 .04545 m
.95455 .04545 L
s
.85354 .04545 m
.85354 .14646 L
s
.95455 .04545 m
.95455 .14646 L
s
.04545 .14646 m
.14646 .14646 L
s
.04545 .14646 m
.04545 .24747 L
s
.14646 .14646 m
.24747 .14646 L
s
.14646 .14646 m
.14646 .24747 L
s
.24747 .14646 m
.34848 .14646 L
s
.24747 .14646 m
.24747 .24747 L
s
.34848 .14646 m
.44949 .14646 L
s
.34848 .14646 m
.34848 .24747 L
s
.44949 .14646 m
.55051 .14646 L
s
.44949 .14646 m
.44949 .24747 L
s
.55051 .14646 m
.65152 .14646 L
s
.55051 .14646 m
.55051 .24747 L
s
.65152 .14646 m
.75253 .14646 L
s
.65152 .14646 m
.65152 .24747 L
s
.75253 .14646 m
.85354 .14646 L
s
.75253 .14646 m
.75253 .24747 L
s
.85354 .14646 m
.95455 .14646 L
s
.85354 .14646 m
.85354 .24747 L
s
.95455 .14646 m
.95455 .24747 L
s
.04545 .24747 m
.14646 .24747 L
s
.04545 .24747 m
.04545 .34848 L
s
.14646 .24747 m
.24747 .24747 L
s
.14646 .24747 m
.14646 .34848 L
s
.24747 .24747 m
.34848 .24747 L
s
.24747 .24747 m
.24747 .34848 L
s
.34848 .24747 m
.44949 .24747 L
s
.34848 .24747 m
.34848 .34848 L
s
.44949 .24747 m
.55051 .24747 L
s
.44949 .24747 m
.44949 .34848 L
s
.55051 .24747 m
.65152 .24747 L
s
.55051 .24747 m
.55051 .34848 L
s
.65152 .24747 m
.75253 .24747 L
s
.65152 .24747 m
.65152 .34848 L
s
.75253 .24747 m
.85354 .24747 L
s
.75253 .24747 m
.75253 .34848 L
s
.85354 .24747 m
.95455 .24747 L
s
.85354 .24747 m
.85354 .34848 L
s
.95455 .24747 m
.95455 .34848 L
s
.04545 .34848 m
.14646 .34848 L
s
.04545 .34848 m
.04545 .44949 L
s
.14646 .34848 m
.24747 .34848 L
s
.14646 .34848 m
.14646 .44949 L
s
.24747 .34848 m
.34848 .34848 L
s
.24747 .34848 m
.24747 .44949 L
s
.34848 .34848 m
.44949 .34848 L
s
.34848 .34848 m
.34848 .44949 L
s
.44949 .34848 m
.55051 .34848 L
s
.44949 .34848 m
.44949 .44949 L
s
.55051 .34848 m
.65152 .34848 L
s
.55051 .34848 m
.55051 .44949 L
s
.65152 .34848 m
.75253 .34848 L
s
.65152 .34848 m
.65152 .44949 L
s
.75253 .34848 m
.85354 .34848 L
s
.75253 .34848 m
.75253 .44949 L
s
.85354 .34848 m
.95455 .34848 L
s
.85354 .34848 m
.85354 .44949 L
s
.95455 .34848 m
.95455 .44949 L
s
.04545 .44949 m
.14646 .44949 L
s
.04545 .44949 m
.04545 .55051 L
s
.14646 .44949 m
.24747 .44949 L
s
.14646 .44949 m
.14646 .55051 L
s
.24747 .44949 m
.34848 .44949 L
s
.24747 .44949 m
.24747 .55051 L
s
.34848 .44949 m
.44949 .44949 L
s
.34848 .44949 m
.34848 .55051 L
s
.44949 .44949 m
.55051 .44949 L
s
.44949 .44949 m
.44949 .55051 L
s
.55051 .44949 m
.65152 .44949 L
s
.55051 .44949 m
.55051 .55051 L
s
.65152 .44949 m
.75253 .44949 L
s
.65152 .44949 m
.65152 .55051 L
s
.75253 .44949 m
.85354 .44949 L
s
.75253 .44949 m
.75253 .55051 L
s
.85354 .44949 m
.95455 .44949 L
s
.85354 .44949 m
.85354 .55051 L
s
.95455 .44949 m
.95455 .55051 L
s
.04545 .55051 m
.14646 .55051 L
s
.04545 .55051 m
.04545 .65152 L
s
.14646 .55051 m
.24747 .55051 L
s
.14646 .55051 m
.14646 .65152 L
s
.24747 .55051 m
.34848 .55051 L
s
.24747 .55051 m
.24747 .65152 L
s
.34848 .55051 m
.44949 .55051 L
s
.34848 .55051 m
.34848 .65152 L
s
.44949 .55051 m
.55051 .55051 L
s
.44949 .55051 m
.44949 .65152 L
s
.55051 .55051 m
.65152 .55051 L
s
.55051 .55051 m
.55051 .65152 L
s
.65152 .55051 m
.75253 .55051 L
s
.65152 .55051 m
.65152 .65152 L
s
.75253 .55051 m
.85354 .55051 L
s
.75253 .55051 m
.75253 .65152 L
s
.85354 .55051 m
.95455 .55051 L
s
.85354 .55051 m
.85354 .65152 L
s
.95455 .55051 m
.95455 .65152 L
s
.04545 .65152 m
.14646 .65152 L
s
.04545 .65152 m
.04545 .75253 L
s
.14646 .65152 m
.24747 .65152 L
s
.14646 .65152 m
.14646 .75253 L
s
.24747 .65152 m
.34848 .65152 L
s
.24747 .65152 m
.24747 .75253 L
s
.34848 .65152 m
.44949 .65152 L
s
.34848 .65152 m
.34848 .75253 L
s
.44949 .65152 m
.55051 .65152 L
s
.44949 .65152 m
.44949 .75253 L
s
.55051 .65152 m
.65152 .65152 L
s
.55051 .65152 m
.55051 .75253 L
s
.65152 .65152 m
.75253 .65152 L
s
.65152 .65152 m
.65152 .75253 L
s
.75253 .65152 m
.85354 .65152 L
s
.75253 .65152 m
.75253 .75253 L
s
.85354 .65152 m
.95455 .65152 L
s
.85354 .65152 m
.85354 .75253 L
s
.95455 .65152 m
.95455 .75253 L
s
.04545 .75253 m
.14646 .75253 L
s
.04545 .75253 m
.04545 .85354 L
s
.14646 .75253 m
.24747 .75253 L
s
.14646 .75253 m
.14646 .85354 L
s
.24747 .75253 m
.34848 .75253 L
s
.24747 .75253 m
.24747 .85354 L
s
.34848 .75253 m
.44949 .75253 L
s
.34848 .75253 m
.34848 .85354 L
s
.44949 .75253 m
.55051 .75253 L
s
.44949 .75253 m
.44949 .85354 L
s
.55051 .75253 m
.65152 .75253 L
s
.55051 .75253 m
.55051 .85354 L
s
.65152 .75253 m
.75253 .75253 L
s
.65152 .75253 m
.65152 .85354 L
s
.75253 .75253 m
.85354 .75253 L
s
.75253 .75253 m
.75253 .85354 L
s
.85354 .75253 m
.95455 .75253 L
s
.85354 .75253 m
.85354 .85354 L
s
.95455 .75253 m
.95455 .85354 L
s
.04545 .85354 m
.14646 .85354 L
s
.04545 .85354 m
.04545 .95455 L
s
.14646 .85354 m
.24747 .85354 L
s
.14646 .85354 m
.14646 .95455 L
s
.24747 .85354 m
.34848 .85354 L
s
.24747 .85354 m
.24747 .95455 L
s
.34848 .85354 m
.44949 .85354 L
s
.34848 .85354 m
.34848 .95455 L
s
.44949 .85354 m
.55051 .85354 L
s
.44949 .85354 m
.44949 .95455 L
s
.55051 .85354 m
.65152 .85354 L
s
.55051 .85354 m
.55051 .95455 L
s
.65152 .85354 m
.75253 .85354 L
s
.65152 .85354 m
.65152 .95455 L
s
.75253 .85354 m
.85354 .85354 L
s
.75253 .85354 m
.75253 .95455 L
s
.85354 .85354 m
.95455 .85354 L
s
.85354 .85354 m
.85354 .95455 L
s
.95455 .85354 m
.95455 .95455 L
s
.04545 .95455 m
.14646 .95455 L
s
.14646 .95455 m
.24747 .95455 L
s
.24747 .95455 m
.34848 .95455 L
s
.34848 .95455 m
.44949 .95455 L
s
.44949 .95455 m
.55051 .95455 L
s
.55051 .95455 m
.65152 .95455 L
s
.65152 .95455 m
.75253 .95455 L
s
.75253 .95455 m
.85354 .95455 L
s
.85354 .95455 m
.95455 .95455 L
s
0 1 0 r
.025 w
.04545 .04545 Mdot
1 0 0 r
.14646 .04545 Mdot
0 1 0 r
.24747 .04545 Mdot
1 0 0 r
.34848 .04545 Mdot
0 1 0 r
.44949 .04545 Mdot
.55051 .04545 Mdot
.65152 .04545 Mdot
1 0 0 r
.75253 .04545 Mdot
0 1 0 r
.85354 .04545 Mdot
1 0 0 r
.95455 .04545 Mdot
.04545 .14646 Mdot
0 1 0 r
.14646 .14646 Mdot
1 0 0 r
.24747 .14646 Mdot
0 1 0 r
.34848 .14646 Mdot
.44949 .14646 Mdot
.55051 .14646 Mdot
1 0 0 r
.65152 .14646 Mdot
.75253 .14646 Mdot
0 1 0 r
.85354 .14646 Mdot
1 0 0 r
.95455 .14646 Mdot
.04545 .24747 Mdot
0 1 0 r
.14646 .24747 Mdot
1 0 0 r
.24747 .24747 Mdot
.34848 .24747 Mdot
.44949 .24747 Mdot
.55051 .24747 Mdot
.65152 .24747 Mdot
0 1 0 r
.75253 .24747 Mdot
.85354 .24747 Mdot
.95455 .24747 Mdot
1 0 0 r
.04545 .34848 Mdot
.14646 .34848 Mdot
0 1 0 r
.24747 .34848 Mdot
1 0 0 r
.34848 .34848 Mdot
0 1 0 r
.44949 .34848 Mdot
1 0 0 r
.55051 .34848 Mdot
0 1 0 r
.65152 .34848 Mdot
1 0 0 r
.75253 .34848 Mdot
0 1 0 r
.85354 .34848 Mdot
1 0 0 r
.95455 .34848 Mdot
0 1 0 r
.04545 .44949 Mdot
.14646 .44949 Mdot
.24747 .44949 Mdot
.34848 .44949 Mdot
1 0 0 r
.44949 .44949 Mdot
.55051 .44949 Mdot
0 1 0 r
.65152 .44949 Mdot
.75253 .44949 Mdot
.85354 .44949 Mdot
1 0 0 r
.95455 .44949 Mdot
0 1 0 r
.04545 .55051 Mdot
.14646 .55051 Mdot
1 0 0 r
.24747 .55051 Mdot
.34848 .55051 Mdot
.44949 .55051 Mdot
0 1 0 r
.55051 .55051 Mdot
1 0 0 r
.65152 .55051 Mdot
.75253 .55051 Mdot
0 1 0 r
.85354 .55051 Mdot
1 0 0 r
.95455 .55051 Mdot
0 1 0 r
.04545 .65152 Mdot
1 0 0 r
.14646 .65152 Mdot
.24747 .65152 Mdot
.34848 .65152 Mdot
.44949 .65152 Mdot
0 1 0 r
.55051 .65152 Mdot
.65152 .65152 Mdot
1 0 0 r
.75253 .65152 Mdot
0 1 0 r
.85354 .65152 Mdot
1 0 0 r
.95455 .65152 Mdot
0 1 0 r
.04545 .75253 Mdot
.14646 .75253 Mdot
1 0 0 r
.24747 .75253 Mdot
0 1 0 r
.34848 .75253 Mdot
.44949 .75253 Mdot
1 0 0 r
.55051 .75253 Mdot
.65152 .75253 Mdot
0 1 0 r
.75253 .75253 Mdot
1 0 0 r
.85354 .75253 Mdot
0 1 0 r
.95455 .75253 Mdot
1 0 0 r
.04545 .85354 Mdot
0 1 0 r
.14646 .85354 Mdot
.24747 .85354 Mdot
1 0 0 r
.34848 .85354 Mdot
0 1 0 r
.44949 .85354 Mdot
1 0 0 r
.55051 .85354 Mdot
0 1 0 r
.65152 .85354 Mdot
1 0 0 r
.75253 .85354 Mdot
0 1 0 r
.85354 .85354 Mdot
.95455 .85354 Mdot
1 0 0 r
.04545 .95455 Mdot
.14646 .95455 Mdot
.24747 .95455 Mdot
.34848 .95455 Mdot
0 1 0 r
.44949 .95455 Mdot
1 0 0 r
.55051 .95455 Mdot
0 1 0 r
.65152 .95455 Mdot
.75253 .95455 Mdot
.85354 .95455 Mdot
1 0 0 r
.95455 .95455 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[4]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`002oooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0o`0Joooo
o`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool000gooooo000:ooooo`Go
0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Go
0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`003?ooool000Wooooo1ol0o`0Fooooo`Ooo`00
5_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o01Kooooo1ol0o`0Fooooo`Ooo`00
5_ooool7o`3o01Kooooo1ooo000;ooooo`002Oooool7o`3o01Ko00001ooo000Fo`0000Oo0?l05_l0
0007ool001Ko00001ol0o`0Fo`0000Oo0?l05_l00007o`3o01Ko00001ooo000Fo`0000Oo0?l05_l0
0007ool000_ooooo0009ooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7
o`3o01Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`002oooool0
00[ooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Go0?l06?ooool5o`3o
01Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000<ooooo`002oooool3o`3o01[o
oooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0o`0Jooooo`?o0?l06_ooool3o`3o01[o
oooo0ooo000Jooooo`?o0?l06_ooool3ool000gooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00co
oooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000;ooooo`?oo`006_ooool3o`3o01[ooooo0ooo000J
ooooo`?o0?l06_ooool3o`3o01[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0J
ooooo`?oo`003Oooool000[ooooo1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hoooo
o`Go0?l06?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000<oooo
o`002Oooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o01Kooooo1ol0o`0Fooooo`Oo
0?l05_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool000_ooooo0009ooooo`Ooo`00
5_l00007o`3o01Ko00001ooo000Fo`0000Oo0?l05_l00007o`3o01Ko00001ol0o`0Fo`0000Ooo`00
5_l00007ool001Ko00001ol0o`0Fo`0000Ooo`002oooool000Wooooo1ooo000Fooooo`Oo0?l05_oo
ool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7o`3o01Kooooo1ooo000Fooooo`Ooo`005_oo
ool7o`3o01Kooooo1ooo000;ooooo`002_ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5
o`3o01Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5
ool000cooooo000;ooooo`?oo`006_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o
01[ooooo0ol0o`0Jooooo`?oo`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`003Oooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000_ooooo0ooo
000Jooooo`?o0?l06_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo
000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ol0o`0=ooooo`002_ooool5ool001Sooooo1Ol0o`0H
ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0H
ooooo`Go0?l06?ooool5o`3o00cooooo0009ooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Foooo
o`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ol0o`0Foooo
o`Oo0?l02oooool000Wooooo1ooo000Fo`0000Oo0?l05_l00007ool001Ko00001ooo000Fo`0000Oo
o`005_l00007ool001Ko00001ooo000Fo`0000Oo0?l05_l00007o`3o01Ko00001ol0o`0;ooooo`00
2Oooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Ooo`00
5_ooool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7o`3o00_ooooo000:ooooo`Goo`006?oo
ool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?oo
ool5o`3o01Sooooo1Ol0o`0Hooooo`Go0?l03?ooool000_ooooo0ooo000Jooooo`?o0?l06_ooool3
ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3
o`3o01[ooooo0ol0o`0=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo
0ol0o`0Jooooo`?oo`006_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool000gooooo
000:ooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo
000Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`003?ooool000Wooooo1ooo000F
ooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0F
ooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000;ooooo`002Oooool7ool001Ko00001ooo000Fo`00
00Oo0?l05_l00007ool001Ko00001ol0o`0Fo`0000Ooo`005_l00007o`3o01Ko00001ooo000Fo`00
00Oo0?l05_l00007ool000_ooooo0009ooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Fooooo`Oo
o`005_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Oo
o`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000Hooooo`Go0?l0
6?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000<ooooo`002ooo
ool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_oo
ool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool000gooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00co
oooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000;ooooo`?o0?l06_ooool3o`3o01[o
oooo0ol0o`0Jooooo`?o0?l06_ooool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o01[o
oooo0ol0o`0Jooooo`?oo`003Oooool000[ooooo1Ol0o`0Hooooo`Go0?l06?ooool5o`3o01Sooooo
1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5o`3o01Sooooo
1Ooo000<ooooo`002Oooool7o`3o01Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7o`3o01Kooooo1ooo
000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool000_ooooo0009
ooooo`Oo0?l05_l00007o`3o01Ko00001ol0o`0Fo`0000Oo0?l05_l00007ool001Ko00001ooo000F
o`0000Oo0?l05_l00007o`3o01Ko00001ol0o`0Fo`0000Ooo`002oooool000Wooooo1ol0o`0Foooo
o`Oo0?l05_ooool7o`3o01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Foooo
o`Oo0?l05_ooool7o`3o01Kooooo1ooo000;ooooo`002_ooool5o`3o01Sooooo1Ol0o`0Hooooo`Go
0?l06?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Go
0?l06?ooool5ool000cooooo000;ooooo`?o0?l06_ooool3o`3o01[ooooo0ol0o`0Jooooo`?o0?l0
6_ooool3ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ol0o`0Jooooo`?oo`00
3Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00_ooooo0ol0o`0Jooooo`?o0?l06_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3o`3o
01[ooooo0ooo000Jooooo`?oo`006_ooool3o`3o01[ooooo0ooo000=ooooo`002_ooool5o`3o01So
oooo1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5ool001So
oooo1Ooo000Hooooo`Go0?l06?ooool5ool000cooooo0009ooooo`Oo0?l05_l00007o`3o01Ko0000
1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ol0o`0Fo`0000Ooo`005_l00007ool001Ko0000
1ol0o`0Fo`0000Ooo`002oooool000Wooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo
000;ooooo`002Oooool7o`3o01Kooooo1ol0o`0Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000F
ooooo`Oo0?l05_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool000_ooooo000:oooo
o`Go0?l06?ooool5o`3o01Sooooo1Ooo000Hooooo`Goo`006?ooool5ool001Sooooo1Ol0o`0Hoooo
o`Goo`006?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`003?ooool000_ooooo0ol0o`0Jooooo`?o
0?l06_ooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3o`3o01[ooooo0ooo000Jooooo`?o
o`006_ooool3o`3o01[ooooo0ooo000=ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`002oooool3o`3o01[ooooo0ooo000Jooooo`?oo`006_ooool3
ool001[ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3
ool000gooooo000:ooooo`Go0?l06?ooool5ool001Sooooo1Ooo000Hooooo`Goo`006?ooool5ool0
01Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`003?ooool000Wo
oooo1ol0o`0Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Ooo`005_l00007o`3o01Ko
00001ol0o`0Fo`0000Ooo`005_l00007o`3o01Ko00001ooo000;ooooo`002Oooool7o`3o01Kooooo
1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o01Kooooo
1ooo000Fooooo`Oo0?l05_ooool7ool000_ooooo0009ooooo`Oo0?l05_ooool7ool001Kooooo1ooo
000Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0
o`0Fooooo`Ooo`002oooool000[ooooo1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ooo000H
ooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ooo000<
ooooo`002oooool3o`3o01[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo0ooo000Joooo
o`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool000gooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo
000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=oooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00co
oooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000=ooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo00cooooo000=ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo00cooooo000;ooooo`?o0?l06_oo
ool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?oo`006_oo
ool3o`3o01[ooooo0ooo000Jooooo`?o0?l03Oooool000[ooooo1Ol0o`0Hooooo`Go0?l06?ooool5
ool001Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5ool001Sooooo1Ooo000Hooooo`Go0?l06?ooool5
ool001Sooooo1Ol0o`0<ooooo`002Oooool7o`3o01Ko00001ol0o`0Fo`0000Ooo`005_l00007o`3o
01Ko00001ol0o`0Fo`0000Ooo`005_l00007ool001Ko00001ol0o`0Fo`0000Ooo`005_l00007o`3o
00_ooooo0009ooooo`Oo0?l05_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o01Ko
oooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l02oooool000Wooooo
1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Kooooo
1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0;ooooo`002_ooool5o`3o01Sooooo1Ol0
o`0Hooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Goo`006?ooool5ool001Sooooo1Ol0
o`0Hooooo`Goo`006?ooool5o`3o00cooooo000;ooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000J
ooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?oo`006_ooool3o`3o01[ooooo0ooo000J
ooooo`?o0?l03Oooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?oo
ool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000go
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0I
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
3?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool0
00gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool03?ooool000gooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Ioooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool03?ooool000gooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool03?ooool000_ooooo0ooo000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l0
6_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3o`3o01[ooooo0ol0o`0=ooooo`002_oo
ool5ool001Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Goo`006?oo
ool5o`3o01Sooooo1Ooo000Hooooo`Go0?l06?ooool5o`3o00cooooo0009ooooo`Ooo`005_l00007
o`3o01Ko00001ol0o`0Fo`0000Ooo`005_l00007o`3o01Ko00001ooo000Fo`0000Oo0?l05_l00007
ool001Ko00001ol0o`0Fo`0000Oo0?l02oooool000Wooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o
01Kooooo1ooo000Fooooo`Oo0?l05_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o
01Kooooo1ol0o`0;ooooo`002Oooool7ool001Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool001Ko
oooo1ol0o`0Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o00_o
oooo000:ooooo`Goo`006?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Goo`006?ooool5o`3o01Sooooo
1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Go0?l03?ooool000_ooooo0ooo
000Jooooo`?o0?l06_ooool3o`3o01[ooooo0ooo000Jooooo`?o0?l06_ooool3ool001[ooooo0ol0
o`0Jooooo`?oo`006_ooool3o`3o01[ooooo0ol0o`0=ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03
o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`00
3Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool0
6_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool0
0ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003o
ooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo
00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooooooo
o`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000
ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Joooo
o`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<oooo
o`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooooooo
ool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`00
0?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Ooo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_oo
ool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooo
oooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0
003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wo
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[o
oooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003ooooo
ooooo`0<ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o
0000ooooooooool06_ooool00ol0003oooooooooo`0<ooooo`003Oooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01Wooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0J
ooooo`03o`000?oooooooooo01[ooooo00?o0000ooooooooool06_ooool00ol0003oooooooooo`0<
ooooo`003Oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01Wooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo01[ooooo00?o0000oooo
ooooool06_ooool00ol0003oooooooooo`0<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?o
o`006_ooool3ool001[ooooo0ol0o`0Jooooo`?oo`006_ooool3o`3o01[ooooo0ol0o`0Jooooo`?o
0?l06_ooool3ool000gooooo000:ooooo`Goo`006?ooool5ool001Sooooo1Ooo000Hooooo`Goo`00
6?ooool5o`3o01Sooooo1Ooo000Hooooo`Go0?l06?ooool5o`3o01Sooooo1Ol0o`0Hooooo`Goo`00
3?ooool000Wooooo1ooo000Fo`0000Ooo`005_l00007ool001Ko00001ooo000Fo`0000Oo0?l05_l0
0007ool001Ko00001ol0o`0Fo`0000Oo0?l05_l00007o`3o01Ko00001ooo000;ooooo`002Oooool7
ool001Kooooo1ooo000Fooooo`Ooo`005_ooool7ool001Kooooo1ol0o`0Fooooo`Ooo`005_ooool7
o`3o01Kooooo1ol0o`0Fooooo`Oo0?l05_ooool7ool000_ooooo0009ooooo`Ooo`005_ooool7ool0
01Kooooo1ooo000Fooooo`Ooo`005_ooool7o`3o01Kooooo1ooo000Fooooo`Oo0?l05_ooool7o`3o
01Kooooo1ol0o`0Fooooo`Ooo`002oooool000[ooooo1Ooo000Hooooo`Goo`006?ooool5ool001So
oooo1Ooo000Hooooo`Go0?l06?ooool5ool001Sooooo1Ol0o`0Hooooo`Go0?l06?ooool5o`3o01So
oooo1Ooo000<ooooo`002oooool3ool001[ooooo0ooo000Jooooo`?oo`006_ooool3ool001[ooooo
0ol0o`0Jooooo`?oo`006_ooool3o`3o01[ooooo0ol0o`0Jooooo`?o0?l06_ooool3ool000gooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500054, -0.0500054, \
0.00383279, 0.00383279}},
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell["Alternative topologies can be used:", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \(\(m3 = 
        LocalInteractionModel[\[IndentingNewLine]InteractionNeighborhood \
\[Rule] RandomTree[
              100], \[IndentingNewLine]InitialStrategyDistribution \[Rule] 
            Table[Random[
                Integer, {1, 2}], {100}]\[IndentingNewLine]];\)\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowGraph[m3];\)\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0525226 1.05045 -0.418155 1.35062 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.63578 .13184 m
.58576 .13184 L
s
.42282 .43339 m
.4497 .37915 L
s
.4497 .37915 m
.47839 .32647 L
s
.4497 .37915 m
.41925 .43017 L
s
.51863 .70156 m
.50781 .76435 L
s
.51863 .70156 m
.52946 .63878 L
s
.67739 .24496 m
.72722 .25056 L
s
.67739 .24496 m
.62908 .22831 L
s
.67739 .24496 m
.72571 .2616 L
s
.67739 .24496 m
.72273 .27213 L
s
.90262 .38463 m
.94748 .41308 L
s
.90262 .38463 m
.85776 .35619 L
s
.34712 .28255 m
.39571 .26727 L
s
.90161 .38712 m
.94546 .41805 L
s
.90161 .38712 m
.85776 .35619 L
s
.58576 .19615 m
.62908 .22831 L
s
.58576 .19615 m
.58576 .26047 L
s
.58576 .19615 m
.54245 .22831 L
s
.58576 .19615 m
.58576 .13184 L
s
.50708 .27379 m
.47839 .32647 L
s
.50708 .27379 m
.54245 .22831 L
s
.50708 .27379 m
.47171 .31926 L
s
.50708 .27379 m
.4661 .31067 L
s
.36333 .45266 m
.39946 .40818 L
s
.3339 .50571 m
.36796 .45861 L
s
.76806 .29931 m
.72273 .27213 L
s
.76806 .29931 m
.81339 .32649 L
s
.15773 .73398 m
.19273 .68804 L
s
.58576 .06753 m
.58576 .13184 L
s
.62908 .22831 m
.66445 .27379 L
s
.29787 .55038 m
.26295 .59643 L
s
.29787 .55038 m
.33285 .50441 L
s
.29787 .55038 m
.26282 .59627 L
s
.61785 .382 m
.59871 .32259 L
s
.61785 .382 m
.63699 .44142 L
s
.54681 .51223 m
.55334 .44847 L
s
.54681 .51223 m
.54029 .57599 L
s
.12264 .77981 m
.15769 .73393 L
s
.12264 .77981 m
.08761 .82572 L
s
.12264 .77981 m
.0876 .8257 L
s
.12264 .77981 m
.08758 .82568 L
s
.35922 .53305 m
.3888 .48119 L
s
.71271 .74784 m
.72722 .80938 L
s
.71271 .74784 m
.69819 .6863 L
s
.59871 .32259 m
.58576 .26047 L
s
.59871 .32259 m
.60524 .38635 L
s
.5381 .64024 m
.54029 .57599 L
s
.55334 .44847 m
.55987 .38471 L
s
.36186 .45059 m
.39946 .40818 L
s
.3888 .48119 m
.41925 .43017 L
s
.3888 .48119 m
.35749 .53135 L
s
.58576 .26047 m
.57282 .32259 L
s
.74174 .87093 m
.75626 .93247 L
s
.74174 .87093 m
.72722 .80938 L
s
.3446 .25896 m
.29469 .26317 L
s
.3446 .25896 m
.29459 .26036 L
s
.3446 .25896 m
.39457 .25615 L
s
.57282 .32259 m
.55987 .38471 L
s
.61177 .45011 m
.60524 .38635 L
s
.49413 .24496 m
.54245 .22831 L
s
.49413 .24496 m
.44792 .26957 L
s
.49413 .24496 m
.44454 .25335 L
s
.15769 .73393 m
.19273 .68804 L
s
.12256 .7797 m
.15764 .73387 L
s
.39571 .26727 m
.34665 .27982 L
s
.39571 .26727 m
.44454 .25335 L
s
.53376 .63975 m
.54029 .57599 L
s
.36957 .4605 m
.40255 .41215 L
s
.22778 .64215 m
.19273 .68804 L
s
.22778 .64215 m
.26282 .59627 L
s
.52465 .50502 m
.50857 .56592 L
s
.52465 .50502 m
.54073 .44412 L
s
.73518 .36474 m
.69982 .31926 L
s
.47171 .31926 m
.43634 .36474 L
s
.39495 .26174 m
.44454 .25335 L
s
.66915 .56322 m
.65307 .50232 L
s
.66915 .56322 m
.68677 .6234 L
s
.66915 .56322 m
.68367 .62476 L
s
.33337 .50506 m
.36796 .45861 L
s
.51861 .5018 m
.54073 .44412 L
s
.54073 .44412 m
.55987 .38471 L
s
.8596 .3511 m
.81339 .32649 L
s
.42513 .34756 m
.4661 .31067 L
s
.24459 .26177 m
.29459 .26036 L
s
.36796 .45861 m
.40255 .41215 L
s
.36796 .45861 m
.33285 .50441 L
s
.65911 .4991 m
.63699 .44142 L
s
.63699 .44142 m
.65307 .50232 L
s
.15764 .73387 m
.19273 .68804 L
s
.69982 .31926 m
.66445 .27379 L
s
.40255 .41215 m
.43634 .36474 L
s
.29469 .26317 m
.24478 .26737 L
s
.69819 .6863 m
.68367 .62476 L
s
.05252 .87155 m
.08758 .82568 L
s
.39946 .40818 m
.43634 .36474 L
s
.54029 .57599 m
.52946 .63878 L
s
.33285 .50441 m
.29761 .55005 L
s
.81339 .32649 m
.85776 .35619 L
s
.39457 .25615 m
.44454 .25335 L
s
1 0 0 r
.025 w
.63578 .13184 Mdot
.42282 .43339 Mdot
0 1 0 r
.4497 .37915 Mdot
1 0 0 r
.51863 .70156 Mdot
.67739 .24496 Mdot
.90262 .38463 Mdot
.34712 .28255 Mdot
0 1 0 r
.90161 .38712 Mdot
.58576 .19615 Mdot
1 0 0 r
.50708 .27379 Mdot
0 1 0 r
.36333 .45266 Mdot
.3339 .50571 Mdot
1 0 0 r
.76806 .29931 Mdot
.15773 .73398 Mdot
0 1 0 r
.72722 .25056 Mdot
.47839 .32647 Mdot
1 0 0 r
.58576 .06753 Mdot
0 1 0 r
.62908 .22831 Mdot
.29787 .55038 Mdot
1 0 0 r
.61785 .382 Mdot
0 1 0 r
.54681 .51223 Mdot
1 0 0 r
.12264 .77981 Mdot
0 1 0 r
.35922 .53305 Mdot
1 0 0 r
.71271 .74784 Mdot
0 1 0 r
.59871 .32259 Mdot
1 0 0 r
.5381 .64024 Mdot
.72571 .2616 Mdot
.55334 .44847 Mdot
.36186 .45059 Mdot
.3888 .48119 Mdot
.58576 .26047 Mdot
.50781 .76435 Mdot
0 1 0 r
.74174 .87093 Mdot
.3446 .25896 Mdot
.57282 .32259 Mdot
.61177 .45011 Mdot
1 0 0 r
.49413 .24496 Mdot
.15769 .73393 Mdot
0 1 0 r
.12256 .7797 Mdot
1 0 0 r
.75626 .93247 Mdot
.39571 .26727 Mdot
.53376 .63975 Mdot
.36957 .4605 Mdot
.54245 .22831 Mdot
0 1 0 r
.22778 .64215 Mdot
1 0 0 r
.41925 .43017 Mdot
.52465 .50502 Mdot
0 1 0 r
.72273 .27213 Mdot
1 0 0 r
.73518 .36474 Mdot
.44792 .26957 Mdot
.34665 .27982 Mdot
0 1 0 r
.47171 .31926 Mdot
1 0 0 r
.39495 .26174 Mdot
.58576 .13184 Mdot
0 1 0 r
.66915 .56322 Mdot
.50857 .56592 Mdot
1 0 0 r
.33337 .50506 Mdot
0 1 0 r
.51861 .5018 Mdot
.54073 .44412 Mdot
1 0 0 r
.8596 .3511 Mdot
0 1 0 r
.35749 .53135 Mdot
1 0 0 r
.42513 .34756 Mdot
.08761 .82572 Mdot
0 1 0 r
.24459 .26177 Mdot
.72722 .80938 Mdot
.36796 .45861 Mdot
1 0 0 r
.65911 .4991 Mdot
.63699 .44142 Mdot
.15764 .73387 Mdot
0 1 0 r
.69982 .31926 Mdot
.40255 .41215 Mdot
1 0 0 r
.0876 .8257 Mdot
0 1 0 r
.94546 .41805 Mdot
1 0 0 r
.19273 .68804 Mdot
.29469 .26317 Mdot
.26295 .59643 Mdot
0 1 0 r
.24478 .26737 Mdot
.66445 .27379 Mdot
1 0 0 r
.65307 .50232 Mdot
0 1 0 r
.69819 .6863 Mdot
.05252 .87155 Mdot
1 0 0 r
.39946 .40818 Mdot
.54029 .57599 Mdot
.33285 .50441 Mdot
0 1 0 r
.68677 .6234 Mdot
1 0 0 r
.81339 .32649 Mdot
.55987 .38471 Mdot
.29459 .26036 Mdot
.08758 .82568 Mdot
.52946 .63878 Mdot
0 1 0 r
.43634 .36474 Mdot
1 0 0 r
.39457 .25615 Mdot
.94748 .41308 Mdot
.44454 .25335 Mdot
0 1 0 r
.26282 .59627 Mdot
1 0 0 r
.4661 .31067 Mdot
0 1 0 r
.85776 .35619 Mdot
1 0 0 r
.60524 .38635 Mdot
.68367 .62476 Mdot
0 1 0 r
.29761 .55005 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[7]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00Yoooool3ool007Kooooo002Vooooo`Goo`00MOooool00:Gooooo1ooo
001dooooo`00YOooool7ool007Cooooo002Uooooo`Ooo`00M?ooool00:Kooooo1Ooo001eooooo`00
Yoooool3ool007Kooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooo
oooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Go
oooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002X
ooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03
o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oo
oooooooo07Gooooo002Wooooo`?oo`002oooool3ool006Sooooo002Vooooo`Goo`002Oooool5ool0
06Oooooo002Uooooo`Ooo`001oooool7ool006Kooooo002Uooooo`Ooo`001ol00007ool006Kooooo
002Uooooo`Ooo`001oooool7ool006Kooooo002Vooooo`Goo`002Oooool5ool006Oooooo002Woooo
o`?oo`002oooool3ool006Sooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`00
0?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooo
oooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Go
oooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002X
ooooo`03o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Xooooo`03
o`000?oooooooooo07Gooooo002Xooooo`03o`000?oooooooooo07Gooooo002Wooooo`?o0?l0M_oo
ool00:Kooooo1Ol0o`1eooooo`00YOooool7o`3o07Cooooo002Uooooo`Oo0?l0M?ooool00:Gooooo
1ol0o`1dooooo`00Y?ooool2o`0000Go0?l000?o0000ooooooooool0L_ooool00:?ooooo00Co0000
oooooooooooooooo0ol0o`02ooooo`;o0000L_ooool00:7ooooo0_l00005ooooo`03o`000?oooooo
oooo00?ooooo00?o0000ooooooooool0Koooool00:3ooooo00?o0000ooooooooool01Oooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo06kooooo002Jooooo`?oo`000_ooool00ol0003o
ooooooooo`06ooooo`03o`000?oooooooooo00Gooooo0_l000001?oooooo0?l0o`3o0?l0o`1Zoooo
o`00VOooool5ool00003o`000?oooooooooo00Oooooo00?o0000ooooooooool01oooool5o`3o06Wo
oooo002Hooooo`Ooo`002Oooool00ol0003oooooooooo`06ooooo`Oo0?l0J?ooool009Sooooo1ooo
0009ooooo`03o`000?oooooooooo00Kooooo1ol0o`1Xooooo`00Uoooool00ol0003oo`00ool00005
ool000Wooooo00?o0000ooooooooool01_ooool7o`3o0003o`000?oooooooooo06Gooooo002<oooo
o`?oo`001Oooool3o`0000;ooooo1Ooo000:ooooo`03o`000?oooooooooo00Oooooo1Ol0o`02oooo
o`?o00001Oooool3ool005cooooo002;ooooo`Goo`0000Coooooo`000?l0003o00001Oooool01?l0
003oo`00ool00?oo000;ooooo`03o`000?oooooooooo00Sooooo0ol0o`000ol0003oooooooooo`03
ooooo`?o000000?oooooool00?oo00000ooo000:ooooo`?o0?l0C_ooool007kooooo0ooo0009oooo
o`Ooo`001oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00cooooo00?o0000oooo
ooooool01Oooool7ool000Sooooo1Ol0o`1=ooooo`00L?ooool3ool000[ooooo1Ooo0008ooooo`Oo
o`001oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00cooooo00?o0000oooooooo
ool01Oooool7ool000Co00000oooool7o`3o04cooooo001Cooooo`?oo`002oooool3o`3o00_ooooo
1Ooo0008ooooo`Ooo`001ol00007ool000Kooooo00?o0000ooooooooool03Oooool3ool000kooooo
00?o0000ooooooooool01?ooool7ool00004o`000?ooooooooooooooo`?o00001ol0o`1<ooooo`00
AOooool3o`3o00[ooooo1Ooo0009ooooo`Go0?l02Oooool7ool000?ooooo1?l00007ool000Gooooo
0_l000000oooooooo`00ool00003ool000Kooooo00?o0000ooooooooool03Oooool5ool000kooooo
00?o0000ooooooooool01?ooool5ool000Go00000oooool2o`3o00?oo`000_l0o`1<ooooo`00A?oo
ool5o`3o00Sooooo1ooo0007ooooo`Oo0?l01?ooool4o`0000Ooo`001ol00007ool000?ooooo0_l0
0004ooooo`?oo`001_ooool00ol0003oooooooooo`0=ooooo`Ooo`003_ooool00ol0003ooooooooo
o`04ooooo`?oo`000oooool2o`000005ooooool0003o0000o`000?ooool01Ooo001=ooooo`00@ooo
ool7o`3o00?ooooo1?l00007ool000Oo00001ol0o`04o`0000Cooooo1ooo0005o`0000?ooooo1Ooo
0003ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool03_ooool7ool000oooooo00?o
0000ooooooooool02oooool01?l0003oooooooooooooool7ool004cooooo0013ooooo`Oo0?l01ol0
0007ool000?o00001?ooool7o`3o00Sooooo1ooo00000ol0003oooooooooo`06ooooo`Goo`000_l0
000;ooooo`?oo`0000?oooooo`000?ooool03oooool7ool000oooooo00?o0000ooooool0o`000_l0
o`0:ooooo`;o000000?oooooool00?l0o`000_l0o`03ool004cooooo0013ooooo`Oo0?l00ol00004
ooooo`Ooo`002?ooool5o`3o00Wooooo1ooo0008ooooo`Ooo`002oooool5ool001;ooooo1Ooo000A
ooooo`Go0?l02oooool00ol0003o0?l0o`3o0003o`3o00;oo`00C?ooool004?ooooo1ol0o`08oooo
o`Goo`002_ooool01?l0o`3oo`00ool00?oo0007ooooo`;o00001ooo0008ooooo`Ooo`002_ooool7
ool001;ooooo0ooo000Aooooo`Oo0?l02_ooool7o`3o04gooooo0014ooooo`Go0?l02_ooool3ool0
00_ooooo1Ooo0002ooooo`Go00000_ooool5ool000Wooooo1ooo000:ooooo`Ooo`004_ooool00ol0
003oooooo`00000Aooooo`Oo0?l02_ooool7o`3o04gooooo0015ooooo`?o0?l06?ooool7ool000?o
00001_ooool3ool000_ooooo1Ooo000;ooooo`Ooo`004_ooool00ol0003oooooo`00000Aooooo`Oo
0?l02_ooool7o`3o04gooooo001Pooooo`Ooo`006?ooool3ool000cooooo00?o0000ool00?oo0000
0ooo000Cooooo`03o`000?oooooo000001;ooooo1Ol0o`0<ooooo`Go0?l000?oooooo`000?l00000
Boooool0063ooooo1ooo000Vooooo`03o`000?oooooo000000?oo`005?ooool00ol0003oooooo`00
000Cooooo`?o0?l000?o0000ooooooooool02oooool3o`3o00Cooooo0_l00019ooooo`00H?ooool7
ool002Gooooo00Co0000ooooool0003o00005_ooool01?l0003oooooooooool0000Gooooo`03o`00
0?oooooooooo01?ooooo00Co0000oooooooooooooooo0ooo0012ooooo`00HOooool5ool002Gooooo
00?o0000ooooooooool00_l0000Fooooo`04o`000?ooooooooooo`0001Oooooo00?o0000oooooooo
ool05?ooool2o`0000Goo`00@Oooool006;ooooo0ooo000Uooooo`03o`000?oooooooooo00;o0000
5oooool01Ol0003ooooooooooooooooo000001Oooooo00?o0000ooooooooool05?ooool7ool0043o
oooo0029ooooo`06o`000?ooooooooooo`000?oooooo00005oooool01Ol0003ooooooooooooooooo
000001Sooooo00?o0000ooooooooool04oooool7ool0043ooooo0024ooooo`?oo`0000Ooooooo`00
0?ooooooooooo`000?oooooo000001Oooooo00?o0000ooooooooool00_ooool00ol0003ooooooooo
o`0Gooooo`03o`000?oooooooooo01;ooooo1ooo0010ooooo`00Poooool5ool000;ooooo00?o0000
ooooool000006?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01Sooooo00?o0000
ooooooooool04_ooool5ool000;o0000?oooool008;ooooo1ooo00001?oooooo0000ooooool0000H
ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool06?ooool00ol0003oooooo`3o0002
o`3o017ooooo0ooo0003ooooo`;o0000?Oooool008;ooooo1ooo00000ol0o`3oooooo`00000Goooo
o`?o0?l01?ooool3o`3o01Wooooo1Ol0o`0Hooooo`03o`000?oooooooooo03[ooooo0022ooooo`Oo
o`000_l0o`000ol0003oooooooooo`0Dooooo`Go0?l00_ooool5o`3o01Oooooo1ol0o`0Hooooo`;o
00000_ooool3ool003Gooooo0022ooooo`03o`000?oo003oo`0000?oo`001?l0o`0Eooooo`ko0?l0
5_ooool7o`3o01[ooooo00?o0000ool00?oo00000ooo000dooooo`00POooool00ol0003ooooooooo
o`03ool000Ko0?l05?ooool>o`3o01Kooooo1ol0o`0Jooooo`Ooo`00<oooool0083ooooo00?o0000
ooooooooool00_ooool8o`3o01Cooooo3_l0o`0Gooooo`Go0?l06oooool7ool003?ooooo001ooooo
o`03o`000?oooooooooo00?ooooo00?o0000o`3o0?l0o`001Ol0o`0Eooooo`Go0?l00_ooool5o`3o
01Wooooo0ol0o`000ol0003oooooooooo`0Iooooo`Ooo`00<oooool007kooooo00?o0000oooooooo
ool00oooool00ol0003oooooooooo`05o`3o01Oooooo0ol0o`04ooooo`?o0?l07_ooool00ol0003o
ooooooooo`0Iooooo`Goo`000ol0000aooooo`00N?ooool3ool000;ooooo00?o0000ooooooooool0
1?ooool01?l0003oooooooooooooool3o`3o01Sooooo00?o0000ooooooooool01Oooool2o`0001ko
oooo00?o0000ooooooooool06_ooool3ool000?ooooo0ol0000_ooooo`00Moooool5ool00003o`00
0?oooooooooo00Cooooo00Go0000ooooooooooooooooo`00000Kooooo`03o`000?oooooooooo00Ko
oooo0_l0000Nooooo`03o`000?oooooooooo023ooooo0ol00003ooooo`?oo`00:?ooool007Kooooo
1ooo0005ooooo`03o`000?oooooooooo00;ooooo00?o0000ooooooooool06Oooool00ol0003ooooo
ooooo`06ooooo`;o00007oooool00ol0003oooooooooo`0Qooooo`?o00001Ooo000Wooooo`00M_oo
ool7ool000Cooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Jooooo`03o`000?oo
oooooooo00Kooooo0_l0000Pooooo`03o`000?oooooooooo027ooooo00?o0000ool00?oo00000ol0
o`02ool002Kooooo001fooooo`Ooo`000oooool00ol0003oooooooooo`03ooooo`03o`000?oooooo
oooo01Wooooo00?o0000ooooooooool01oooool00ol0003oooooo`00000Pooooo`03o`000?oooooo
oooo027ooooo00?oo`00o`3o0?l0o`000ol0o`000ooo003oooooooooo`0Tooooo`00Moooool5ool0
00?o0?l000?oooooo`000?ooool00oooool00ol0003oooooooooo`0Jooooo`03o`000?oooooooooo
00Oooooo00?o0000ooooool000008?ooool00ol0003oooooool00002ool001oooooo1ol0o`0Voooo
o`00N?ooool3ool000Go0?l01?ooool00ol0003oooooooooo`0Kooooo`03o`000?oooooooooo00Oo
oooo00?o0000ooooool000008Oooool5ool001kooooo1ol0o`0Vooooo`00N_ooool7o`3o00?ooooo
00?o0000ooooooooool06oooool00ol0003oooooooooo`07ooooo`04o`000?ooooooooooo`0001oo
oooo1ooo000Mooooo`Oo0?l09_ooool007[ooooo1ol0o`02ooooo`03o`000?oooooooooo01_ooooo
00?o0000ooooooooool02?ooool01?l0003oooooooooool0000Oooooo`Ooo`007_ooool5o`3o00;o
00009Oooool007[ooooo2Ol0o`000ol0003oooooooooo`0Kooooo`03o`000?oooooooooo00Sooooo
00Co0000oooooooooooo00007oooool7ool001oooooo0ol0o`02ooooo`?o00008oooool007_ooooo
2Ol0o`0Mooooo`03o`000?oooooooooo00Wooooo00?o0000ooooooooool00ooo000Mooooo`Goo`00
9oooool2o`0002;ooooo001kooooo`03o`000?l0o`3o0?l000Oo0?l06_ooool3ool000[ooooo2?oo
000Mooooo`?oo`00:Oooool3o`000004oooooooo003oo`00ool001cooooo001jooooo`04o`000?oo
ooooooooooooo`Oo0?l06Oooool5ool000Sooooo2_oo0019ooooo`;o000000Goo`00o`3o0?l0o`3o
0?l0ool0000Kooooo`00NOooool2o`0000?ooooo1ol0o`0Hooooo`Ooo`001_ooool;ool004[ooooo
00?oo`00o`3o0?l0o`000ol0o`000ooo003oooooooooo`0Hooooo`00N?ooool2o`0000Gooooo1Ol0
o`0Iooooo`Ooo`001_ooool;ool004[ooooo1ol0o`0Jooooo`00Moooool2o`0000Kooooo00Co0000
o`3o0?l0o`3o0?l06_ooool7ool000Kooooo2_oo001;ooooo`Oo0?l06_ooool007Oooooo0_l00006
ooooo`03o`000?oooooooooo01cooooo1Ooo0008ooooo`Soo`00C?ooool7o`3o00;o00006?ooool0
07Kooooo0_l00006ooooo`;o00007oooool3ool000[ooooo0ooo0003ooooo`03o`000?oooooooooo
04_ooooo1Ol0o`02o`000003ooooool0003ooooo01Kooooo001aooooo`?oo`0000?oooooo`000?l0
00001oooool00ol0003oooooooooo`0Nooooo`;o00003?ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo04cooooo0ol0o`03ooooo`Co00005Oooool0073ooooo1Ooo00000ol0003ooooo
ooooo`05ooooo`;o00008?ooool00ol0003oooooooooo`0;ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool0Doooool00ol0003oooooo`000002ooooo`?oo`003oooool006oooooo1ooo
0006ooooo`;o00008?ooool2o`0000kooooo00?o0000ooooooooool00_ooool00ol0003ooooooooo
o`1Dooooo`?o00001Ooo000>ooooo`00Koooool7ool00003o`3o0?oooooooooo00?ooooo0_l0000P
ooooo`;o00003_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo05Kooooo1ooo000=
ooooo`00Koooool7ool00003o`3o0?oooooooooo00;ooooo0_l0000Qooooo`;o00003_ooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo05Gooooo1ooo000=ooooo`00L?ooool5ool000;o
0?l01?ooool2o`00023ooooo00?o0000ooooool000003_ooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo05Gooooo1ooo000=ooooo`00L?ooool01?l0003oo`00ool00?oo0002o`3o0004
oooooooo003oo`00ool000;o00008Oooool00ol0003oooooo`00000>ooooo`03o`000?oooooooooo
00?ooooo00?o0000ooooooooool0EOooool00ol0o`3oo`00ool00003ool00003o`3o0?oooooooooo
00_ooooo001_ooooo`03o`000?oooooo000000?o0?l000?oooooool00?oo00000ooo00000ol0003o
ooooooooo`0Oooooo`03o`000?oooooo000000kooooo00?o0000ooooooooool00oooool00ol0003o
ooooooooo`1Eooooo`;o0?l00ooo0002o`3o00gooooo001^ooooo`04o`000?oooooo0000o`0000?o
oooo1ooo000Qooooo`03o`000?oooooo000000kooooo00?o0000ooooooooool01?ooool00ol0003o
ooooooooo`1Eooooo`Go0?l03_ooool006gooooo00Co0000oooooooooooo00001?ooool8ool001oo
oooo00Co0000oooooooooooo00003_ooool00ol0003oooooooooo`04ooooo`?oo`00E_ooool3o`3o
00oooooo001/ooooo`;o000000?oooooo`000?ooool01?ooool8ool001gooooo0ol0o`000ooooooo
0000ooooo`0>ooooo`03o`000?oooooooooo00?ooooo1Ooo001Wooooo`00Joooool2o`000003oooo
ool0003o000000Kooooo1ooo000Looooo`Go0?l00_oo000?ooooo`04o`000?ooooooooooooooo`Oo
o`00I_ooool006Kooooo0ooo00001?l0o`3o0000o`000?ooool2o`0000Oooooo00?o0000ool00?oo
00000ooo000Looooo`Oo0?l00_oo000=ooooo`?o0?l00_ooool7ool006Kooooo001Uooooo`Goo`00
00Go0?l0ooooool0003oooooo`000007ooooo`03o`000?ooooooo`0000;oo`007Oooool7o`3o00?o
o`002oooool5o`3o0003oooooooo003oo`0000Goo`00I_ooool006Cooooo1?oo0004o`3o00;o0000
1oooool00ol0003oooooooooo`0Pooooo`Oo0?l00ooo000:ooooo`Oo0?l000?oooooool00?oo0000
0ooo001Wooooo`00I?ooool3ool000Go0?l000?o0000ooooooooool01_ooool00ol0003ooooooooo
o`0Qooooo`Go0?l01?oo000:ooooo`Oo0?l00_ooool3ool006Sooooo001Tooooo`;oo`001ol0o`07
ooooo`03o`000?oooooooooo02?ooooo0ol0o`04ool000_ooooo1ol0o`04ooooo`03o`000?oooooo
oooo06Kooooo001Uooooo`03ool00?l0o`3o0?l000Go0?l000?oo`00ooooooooool00oooool00ol0
003oooooooooo`0Tooooo`03o`000?oooooooooo00?oo`003Oooool5o`3o00Gooooo0_l0001Woooo
o`00I_ooool7o`3o0003ool00?oooooooooo00?ooooo00?o0000ooooooooool08oooool2o`0000?o
oooo00?o0000ooooooooool03Oooool3o`3o00Oooooo00?o0000ooooooooool0IOooool006Oooooo
1Ol0o`02ool000Cooooo00?o0000ooooooooool09?ooool00ol0003oooooooooo`02ooooo`03o`00
0?oooooooooo01Oooooo0_l0001Vooooo`00Ioooool01?l0003o0?l0o`3o0?l0o`02ool000Gooooo
00?o0000ooooooooool08oooool2o`0000?ooooo00?o0000ooooooooool06?ooool2o`0006Kooooo
001Vooooo`03o`000?oooooooooo00?oo`000_ooool3ool00003o`000?oooooooooo02Cooooo0_l0
0003ooooo`03o`000?oooooooooo01Sooooo00?o0000ooooool00000IOooool006Kooooo00?o0000
ooooooooool01?ooool5ool002Gooooo0_l00004ooooo`03o`000?oooooooooo01Wooooo0_l0001U
ooooo`00IOooool00ol0003oooooooooo`04ooooo`Ooo`009?ooool2o`0000Cooooo00?o0000oooo
ooooool06Oooool2o`0006Gooooo001Tooooo`03o`000?oooooooooo00Gooooo1ooo000Tooooo`;o
00001?ooool00ol0003oooooooooo`0Iooooo`03o`000?oooooo000006Cooooo001Sooooo`;o0000
1oooool7ool002?ooooo00?o0000ooooool000001?ooool00ol0003oooooooooo`0Iooooo`03o`00
0?oooooo000006Cooooo001Rooooo`;o00002Oooool5ool002Cooooo0_l00005ooooo`03o`000?oo
oooooooo01[ooooo00?o0000oooooooo00000_oo001Qooooo`00G_ooool3ool00003ooooool0003o
oooo00Wooooo00Co0000ool00?oo003oo`008_ooool3o`3o0003ooooool0003ooooo00Cooooo00?o
0000ooooooooool06_ooool6ool0063ooooo001Mooooo`Goo`002oooool00ol0003oooooooooo`0R
ooooo`Go0?l000?oo`00ooooooooool00oooool00ol0003oooooooooo`0Iooooo`Soo`00Goooool0
05cooooo1ooo0009ooooo`03o`000?oooooooooo02;ooooo1ol0o`001Ooo003ooooooooooooooooo
000001_ooooo2Ooo001Oooooo`00G?ooool7ool000Sooooo0_l0000Tooooo`Oo0?l00_oo00001?oo
oooo0?l0o`3o0?l0o`0Jooooo`Woo`00Goooool005cooooo1ooo0008ooooo`03o`000?oooooooooo
02?ooooo1ol0o`02ool000Go0?l06Oooool8ool0063ooooo001Looooo`Ooo`001oooool2o`0002Ko
oooo1Ol0o`03ool000Ko0?l06Oooool6ool0067ooooo001Mooooo`Goo`002?ooool00ol0003ooooo
ooooo`0Vooooo`?o0?l00ooo0007o`3o01[ooooo0ooo001Sooooo`00G?ooool00ol0003oooooool0
0002ool000Sooooo0_l0000Zooooo`?oo`0000?oooooo`3o0?l0o`001Ol0o`0Looooo`03o`000?oo
oooooooo067ooooo001Looooo`03o`000?oooooooooo00Kooooo0ol0o`02o`0002_ooooo00?o0000
ooooooooool00_ooool5o`3o01gooooo00?o0000ooooooooool0HOooool005_ooooo00?o0000oooo
ooooool01_ooool5o`3o02_ooooo00?o0000ooooooooool01?ooool3o`3o01oooooo00?o0000oooo
ooooool0H?ooool005[ooooo00?o0000ooooooooool01_ooool7o`3o02[ooooo00?o0000oooooooo
ool01Oooool00ol0003oooooooooo`0Nooooo`03o`000?oooooooooo063ooooo001Iooooo`03o`00
0?oooooooooo00Oooooo2?l0o`0Yooooo`03o`000?oooooooooo00Gooooo00?o0000ooooooooool0
7_ooool00ol0003oooooooooo`1Pooooo`00F?ooool00ol0003oooooooooo`08ooooo`So0?l0:Ooo
ool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool0Gooo
ool005Cooooo0ol0o`000ooooooo0000ooooo`0:ooooo`Oo0?l0:?ooool00ol0003oooooooooo`06
ooooo`03o`000?oooooooooo01oooooo00?o0000ooooooooool0Goooool005?ooooo1Ol0o`0=oooo
o`Go0?l0:Oooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo01oooooo00?o0000oooo
ooooool0Goooool005;ooooo1ol0o`0=ooooo`?o0?l0:_ooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo01oooooo00?o0000ooooooooool0Goooool005;ooooo1ol0o`0jooooo`03o`00
0?oooooooooo00Gooooo00?o0000ooooooooool08Oooool00ol0003oooooooooo`1Nooooo`00D_oo
ool7o`3o03Wooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`0Pooooo`?o0?l0Gooo
ool005;ooooo1ol0o`0gooooo`?o0?l02?ooool00ol0003oooooooooo`0Oooooo`Go0?l0G_ooool0
05?ooooo1Ol0o`0gooooo`Go0?l01oooool00ol0003oooooooooo`0Nooooo`Oo0?l0GOooool005;o
oooo00?o0000ooooool0o`000_l0o`0gooooo`Oo0?l01_ooool00ol0003oooooooooo`0Nooooo`Oo
0?l0GOooool005;ooooo00?o0000ooooooooool0>Oooool7o`3o00Gooooo0ooo000Oooooo`Oo0?l0
GOooool0057ooooo00?o0000ooooooooool0>_ooool7o`3o00Cooooo1Ooo000Oooooo`Go0?l0G_oo
ool0053ooooo00?o0000ooooooooool0??ooool5o`3o00Cooooo1ooo000Oooooo`?o0?l000?o0000
ooooooooool0G?ooool0053ooooo00?o0000ooooooooool0?Oooool3o`3o00Gooooo1ooo000Roooo
o`03o`000?oooooooooo05cooooo001?ooooo`03o`000?oooooooooo04Kooooo1ooo000Rooooo`03
o`000?oooooooooo05cooooo001>ooooo`03o`000?oooooooooo04Sooooo1Ooo000Sooooo`;o0000
GOooool004[ooooo0ol0o`000ol0003oooooooooo`1:ooooo`?oo`009?ooool2o`0005gooooo0019
ooooo`Go0?l0C?ooool2o`0002Kooooo00?o0000ooooooooool0Foooool004Sooooo1ol0o`1;oooo
o`;o00009_ooool2o`0005cooooo0018ooooo`Oo0?l0Boooool2o`0002Kooooo0_l0001Looooo`00
B?ooool7o`3o04_ooooo0_l0000Vooooo`;o0000G?ooool004Wooooo1Ol0o`1;ooooo`?o00009ooo
ool00ol0003oooooooooo`1Jooooo`00BOooool01?l0003o0?l0o`3o0?l0o`1<ooooo`?o00009ooo
ool2o`0005_ooooo0018ooooo`03o`000?oooooooooo04kooooo0_l0000Xooooo`04o`000?l0o`3o
0?l0o`3o05Wooooo0018ooooo`03o`000?oooooooooo04kooooo0_l0000Xooooo`?oo`000_l0o`1H
ooooo`00Aoooool00ol0003oooooooooo`1>ooooo`03o`000?oooooo000002Oooooo1Ooo0002o`3o
05Oooooo0016ooooo`03o`000?oooooooooo04oooooo00?o0000ooooool000009_ooool7ool00003
o`3o0?oooooooooo05Gooooo0015ooooo`03o`000?oooooooooo053ooooo0ol0000Vooooo`Ooo`00
00?o0?l0ooooooooool0EOooool004Cooooo00?o0000ooooooooool0Coooool6ool002Gooooo1ooo
001Hooooo`00@?ooool3o`3o0003ooooool0003ooooo04oooooo2?oo000Uooooo`Goo`00FOooool0
03oooooo1Ol0o`1@ooooo`[oo`009Oooool3ool005[ooooo000nooooo`Oo0?l0Coooool:ool002Oo
oooo00?o0000ooooooooool0F?ooool003kooooo1ol0o`1?ooooo`[oo`009oooool00ol0003ooooo
ooooo`1Hooooo`00?_ooool7o`3o053ooooo2?oo000Xooooo`03o`000?oooooooooo05Sooooo000o
ooooo`Go0?l0D_ooool6ool002[ooooo00?o0000ooooooooool0Eoooool003oooooo00Co0000o`3o
0?l0o`3o0?l0Doooool00ol0003oooooooooo`0]ooooo`03o`000?oooooooooo05Oooooo000noooo
o`03o`000?oooooooooo05Gooooo00?o0000ooooooooool0;Oooool00ol0003oooooooooo`1Goooo
o`00?_ooool00ol0003oooooooooo`1Eooooo`03o`000?oooooooooo02gooooo00?o0000oooooooo
ool0Eoooool003gooooo00?o0000ooooooooool0E_ooool00ol0003oooooooooo`0]ooooo`03o`00
0?oooooooooo05Oooooo000looooo`03o`000?oooooooooo05Oooooo00?o0000ooooooooool0;_oo
ool00ol0003oooooooooo`1Fooooo`00>oooool00ol0003oooooooooo`1Hooooo`03o`000?oooooo
oooo02kooooo00?o0000ooooooooool0E_ooool003[ooooo00?o0000ooooooooool0F?ooool00ol0
003oooooooooo`0_ooooo`03o`000?oooooooooo05Kooooo000fooooo`?oo`0000?oooooo`000?oo
ool0FOooool00ol0003oooooooooo`0_ooooo`?o0?l0E_ooool003Gooooo1Ooo001Kooooo`03o`00
0?oooooooooo02kooooo1Ol0o`1Eooooo`00=?ooool7ool005[ooooo00?o0000ooooooooool0;Ooo
ool7o`3o05Cooooo000dooooo`Ooo`00F_ooool00ol0003oooooooooo`0]ooooo`Oo0?l0E?ooool0
03Cooooo1ooo001Hooooo`?oo`00;oooool7o`3o05Cooooo000eooooo`Goo`00F?ooool5ool002oo
oooo1Ol0o`1Eooooo`00=Oooool01?l0003oo`00ool00?oo001Hooooo`Ooo`00;oooool3o`3o05Ko
oooo000dooooo`03o`000?oooooooooo05[ooooo1ooo000aooooo`03o`000?oooooooooo05Cooooo
000dooooo`03o`000?oooooooooo05[ooooo1ooo000aooooo`03o`000?oooooooooo05Cooooo000c
ooooo`03o`000?oooooooooo05cooooo1Ooo000booooo`03o`000?oooooooooo05Cooooo000boooo
o`03o`000?oooooooooo05kooooo0ooo000dooooo`03o`000?oooooooooo05?ooooo000aooooo`03
o`000?oooooooooo05oooooo00?o0000ooooooooool0=?ooool00ol0003oooooooooo`1Cooooo`00
<?ooool00ol0003oooooooooo`1Pooooo`03o`000?oooooooooo03Cooooo00?o0000ooooooooool0
Doooool002cooooo0ooo00000ooooooo0000ooooo`1Qooooo`03o`000?oooooooooo03Cooooo00?o
0000ooooooooool0Doooool002_ooooo1Ooo001Sooooo`03o`000?oooooooooo03Gooooo00?o0000
ooooooooool0D_ooool002[ooooo1ooo001Rooooo`03o`000?oooooooooo03Gooooo00?o0000oooo
ooooool0D_ooool002[ooooo1ooo001Rooooo`03o`000?oooooooooo03Gooooo00?o0000oooooooo
ool0D_ooool002[ooooo1ooo001Qooooo`03o`000?oooooooooo03Kooooo0ooo001Booooo`00:ooo
ool5ool006;ooooo00?o0000ooooooooool0=Oooool5ool0057ooooo000[ooooo`04o`000?oo003o
o`00ool006?ooooo00?o0000ooooooooool0=?ooool7ool0053ooooo000Zooooo`03o`000?oooooo
oooo06Gooooo00?o0000ooooooooool0=?ooool7ool0053ooooo000Zooooo`03o`000?oooooooooo
06Gooooo00?o0000ooooooooool0=?ooool7ool0053ooooo000Yooooo`03o`000?oooooooooo06Co
oooo0ooo000gooooo`Goo`00DOooool002Sooooo00?o0000ooooooooool0I?ooool5ool003Oooooo
0ooo001Booooo`009oooool00ol0003oooooooooo`1Tooooo`Ooo`00>?ooool00ol0003ooooooooo
o`1@ooooo`009_ooool00ol0003oooooooooo`1Uooooo`Ooo`00>?ooool00ol0003oooooooooo`1@
ooooo`008_ooool3o`3o0003ooooool0003ooooo06Kooooo1ooo000hooooo`03o`000?oooooooooo
053ooooo000Qooooo`Go0?l0JOooool5ool003Wooooo00?o0000ooooooooool0D?ooool0023ooooo
1ol0o`1Yooooo`?oo`00>oooool00ol0003oooooooooo`1?ooooo`008?ooool7o`3o0:Oooooo00?o
0000ooooooooool0Coooool0023ooooo1ol0o`2Wooooo`03o`000?oooooooooo04oooooo000Qoooo
o`Go0?l0Z?ooool00ol0003oooooooooo`1?ooooo`008Oooool01?l0003o0?l0o`3o0?l0o`2Zoooo
o`03o`000?oooooooooo04kooooo000Pooooo`03o`000?oooooooooo0:cooooo00?o0000oooooooo
ool0C_ooool0023ooooo00?o0000ooooooooool0[?ooool00ol0003oooooooooo`1>ooooo`007ooo
ool00ol0003oooooooooo`2]ooooo`?o0?l0C_ooool001kooooo00?o0000ooooooooool0[Oooool5
o`3o04gooooo000Mooooo`03o`000?oooooooooo0:gooooo1ol0o`1<ooooo`007?ooool00ol0003o
ooooooooo`2^ooooo`Oo0?l0C?ooool001cooooo00?o0000ooooooooool0[_ooool7o`3o04cooooo
000Hooooo`?oo`0000?o0000ooooooooool0/?ooool5o`3o04gooooo000Gooooo`Goo`00/oooool3
o`3o04kooooo000Fooooo`Ooo`00]?ooool00ol0003oooooooooo`1<ooooo`005_ooool7ool00;Co
oooo00?o0000ooooooooool0C?ooool001Kooooo1ooo002dooooo`03o`000?oooooooooo04cooooo
000Gooooo`Goo`00]_ooool00ol0003oooooooooo`1;ooooo`005_ooool00ol0003oooooool00002
ool00;Oooooo00?o0000ooooooooool0Boooool001Kooooo00?o0000ooooooooool0^Oooool00ol0
003oooooooooo`1;ooooo`005Oooool00ol0003oooooooooo`2jooooo`03o`000?oooooooooo04_o
oooo000Dooooo`03o`000?oooooooooo0;_ooooo00?o0000ooooooooool0Boooool001Cooooo00?o
0000ooooooooool0_?ooool00ol0003oooooooooo`1:ooooo`004oooool00ol0003oooooooooo`2m
ooooo`03o`000?oooooooooo04[ooooo000Booooo`03o`000?oooooooooo0;kooooo00?o0000oooo
ooooool0B_ooool000kooooo0ol0o`000ol0003oooooooooo`2oooooo`?o0?l0B_ooool000gooooo
1Ol0o`30ooooo`Go0?l0BOooool000cooooo1ol0o`2nooooo`Oo0?l0B?ooool000cooooo1ol0o`2n
ooooo`Oo0?l0B?ooool000cooooo1ol0o`2nooooo`Oo0?l0B?ooool000gooooo1Ol0o`30ooooo`Go
0?l0BOooool000kooooo0ol0o`32ooooo`?o0?l0B_ooool00=Gooooo00?o0000ooooooooool0B?oo
ool00=Kooooo00?o0000ooooooooool0Aoooool00=Kooooo00?o0000ooooooooool0Aoooool00=Ko
oooo00?o0000ooooooooool0Aoooool00=Kooooo00?o0000ooooooooool0Aoooool00=Oooooo00?o
0000ooooooooool0A_ooool00=Oooooo00?o0000ooooooooool0A_ooool00=Oooooo00?o0000oooo
ooooool0A_ooool00=Oooooo00?o0000ooooooooool0A_ooool00=Sooooo00?o0000ooooooooool0
AOooool00=Oooooo0ooo0016ooooo`00e_ooool5ool004Gooooo003Eooooo`Ooo`00A?ooool00=Go
oooo1ooo0014ooooo`00eOooool7ool004Cooooo003Fooooo`Goo`00AOooool00=Oooooo0ooo0016
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0500049, 0.309599, \
0.00331701, 0.00257982}},
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell["\<\
Payoff matrices can be entered in a way which makes the structure easy to \
see. Here is Rock-Scissors-Paper:\
\>", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    RowBox[{
      RowBox[{"m4", "=", 
        RowBox[{"LocalInteractionModel", "[", "\[IndentingNewLine]", 
          
          RowBox[{\(InteractionNeighborhood \[Rule] GridGraph[20, 20]\), ",", 
            "\[IndentingNewLine]", 
            RowBox[{"PayoffMatrix", "\[Rule]", 
              RowBox[{"(", GridBox[{
                    {\({1, 1}\), \({2, 0}\), \({0, 2}\)},
                    {\({0, 2}\), \({1, 1}\), \({2, 0}\)},
                    {\({2, 0}\), \({0, 2}\), \({1, 1}\)}
                    }], ")"}]}], ",", 
            "\[IndentingNewLine]", \(InitialStrategyDistribution \[Rule] 
              Table[\[IndentingNewLine]Random[
                  Integer, {1, 3}], {400}\[IndentingNewLine]]\), ",", 
            "\[IndentingNewLine]", \(StrategyColors \[Rule] {Red, Green, 
                Blue}\), ",", 
            "\[IndentingNewLine]", \(NeighborhoodsToDraw \[Rule] {}\)}], 
          "\[IndentingNewLine]", "]"}]}], ";"}]], "Input",
  CellLabel->"In[9]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowGraph[m4, VertexStyle \[Rule] Box[Large]];\)\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 1 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
1 0 0 r
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
0 1 0 r
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
0 0 1 r
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
0 1 0 r
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
1 0 0 r
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
0 0 1 r
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
1 0 0 r
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
0 0 1 r
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
0 1 0 r
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
0 0 1 r
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
0 1 0 r
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
1 0 0 r
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
0 0 1 r
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
1 0 0 r
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
0 1 0 r
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
1 0 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
0 0 1 r
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
1 0 0 r
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
0 1 0 r
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
0 0 1 r
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
0 1 0 r
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
0 0 1 r
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
1 0 0 r
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
0 1 0 r
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
1 0 0 r
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
0 1 0 r
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
0 0 1 r
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
1 0 0 r
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
0 0 1 r
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
1 0 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
0 1 0 r
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
0 0 1 r
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
1 0 0 r
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
0 0 1 r
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
1 0 0 r
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
0 0 1 r
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
0 1 0 r
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
1 0 0 r
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
0 1 0 r
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
0 0 1 r
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
1 0 0 r
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
0 0 1 r
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
0 1 0 r
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
1 0 0 r
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 0 1 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
0 1 0 r
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
0 0 1 r
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
1 0 0 r
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
0 1 0 r
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
0 0 1 r
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
1 0 0 r
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
0 0 1 r
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
0 1 0 r
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
1 0 0 r
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
0 1 0 r
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
1 0 0 r
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
0 0 1 r
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
1 0 0 r
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
0 0 1 r
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
1 0 0 r
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 0 1 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
0 1 0 r
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
0 0 1 r
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
0 1 0 r
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
0 0 1 r
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
0 1 0 r
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
1 0 0 r
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
0 0 1 r
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
0 1 0 r
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
1 0 0 r
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
0 0 1 r
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
1 0 0 r
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
0 1 0 r
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 0 1 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
1 0 0 r
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
0 1 0 r
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
0 0 1 r
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
0 1 0 r
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
0 0 1 r
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
0 1 0 r
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
0 0 1 r
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
0 1 0 r
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
1 0 0 r
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
0 0 1 r
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
1 0 0 r
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
0 0 1 r
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
1 0 0 r
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
0 1 0 r
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
1 0 0 r
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 0 1 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
0 1 0 r
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
0 0 1 r
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
1 0 0 r
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
0 1 0 r
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
0 0 1 r
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
1 0 0 r
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
0 0 1 r
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
0 1 0 r
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
1 0 0 r
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
0 0 1 r
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
0 1 0 r
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
1 0 0 r
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
0 1 0 r
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
0 0 1 r
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
1 0 0 r
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 0 1 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
0 1 0 r
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
0 0 1 r
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
1 0 0 r
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
0 1 0 r
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
0 0 1 r
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
0 1 0 r
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
1 0 0 r
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
0 0 1 r
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
0 1 0 r
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
1 0 0 r
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
0 1 0 r
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
0 0 1 r
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
1 0 0 r
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 0 1 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
1 0 0 r
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
0 1 0 r
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
1 0 0 r
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
0 0 1 r
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
0 1 0 r
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
1 0 0 r
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
0 1 0 r
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
1 0 0 r
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
0 0 1 r
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
0 1 0 r
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
0 0 1 r
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
0 1 0 r
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
0 0 1 r
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
0 1 0 r
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
0 0 1 r
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
0 1 0 r
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
1 0 0 r
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
0 0 1 r
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
0 1 0 r
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
0 0 1 r
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
1 0 0 r
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
0 0 1 r
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
0 1 0 r
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 0 1 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
1 0 0 r
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
0 1 0 r
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
0 0 1 r
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
0 1 0 r
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
0 0 1 r
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
0 1 0 r
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
1 0 0 r
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
0 1 0 r
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
0 0 1 r
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
1 0 0 r
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
0 0 1 r
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
1 0 0 r
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
0 0 1 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
0 1 0 r
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
1 0 0 r
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
0 1 0 r
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
0 0 1 r
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
0 1 0 r
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
0 0 1 r
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
0 1 0 r
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
0 0 1 r
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
1 0 0 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
0 0 1 r
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
1 0 0 r
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
0 0 1 r
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
0 1 0 r
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
0 0 1 r
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
1 0 0 r
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
0 1 0 r
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
0 0 1 r
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
0 1 0 r
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
1 0 0 r
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
0 1 0 r
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
0 0 1 r
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
1 0 0 r
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
0 0 1 r
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
1 0 0 r
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
0 0 1 r
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
0 1 0 r
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
0 0 1 r
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
0 1 0 r
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
1 0 0 r
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
0 1 0 r
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
0 0 1 r
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
1 0 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
0 0 1 r
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
0 1 0 r
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
1 0 0 r
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
0 1 0 r
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
0 0 1 r
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
0 1 0 r
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
0 0 1 r
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
0 1 0 r
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
1 0 0 r
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
0 1 0 r
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
0 0 1 r
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
1 0 0 r
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
0 1 0 r
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
1 0 0 r
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 1 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
1 0 0 r
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
0 1 0 r
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
1 0 0 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
0 1 0 r
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
1 0 0 r
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
0 1 0 r
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
1 0 0 r
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
0 0 1 r
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
1 0 0 r
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
0 1 0 r
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
1 0 0 r
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
0 0 1 r
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
1 0 0 r
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 1 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
0 0 1 r
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
0 1 0 r
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
0 0 1 r
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
0 1 0 r
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
0 0 1 r
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
1 0 0 r
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
0 1 0 r
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
1 0 0 r
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
0 1 0 r
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
0 0 1 r
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
1 0 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
0 0 1 r
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
1 0 0 r
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
0 0 1 r
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
1 0 0 r
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
0 0 1 r
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
1 0 0 r
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
0 0 1 r
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
0 1 0 r
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
1 0 0 r
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
0 1 0 r
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 0 1 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
1 0 0 r
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
0 0 1 r
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
1 0 0 r
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
0 0 1 r
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
0 1 0 r
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
1 0 0 r
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
0 1 0 r
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
1 0 0 r
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
0 0 1 r
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
1 0 0 r
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
0 1 0 r
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
1 0 0 r
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 0 1 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
0 1 0 r
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
0 0 1 r
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
0 1 0 r
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
1 0 0 r
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
0 0 1 r
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
0 1 0 r
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
0 0 1 r
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
1 0 0 r
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
0 0 1 r
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
1 0 0 r
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
0 0 1 r
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
0 1 0 r
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
0 0 1 r
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
0 1 0 r
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
0 0 1 r
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[10]:=",
  ImageSize->{316, 316},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004l0001?2000`400?l00000o`00003ooooooolm
ooooo`00oooooolmooooo`00oooooolmooooo`00oooooolmooooo`00oooooolmooooo`00oooooolm
ooooo`001?ooool?o`3o00ooo`003ol0o`0?o`00o`oo0?l03ooo000?o`00o`ooo`007ol00?l?o`3o
01ko003o3ol0o`0?ool001ko003o7_oo000Do`3o00Kooooo0004ooooo`oo0?l03ooo000?o`3o00oo
003o3ol0o`0?ool000oo003o3ooo000Oo`00o`oo0?l07_l00?l?o`3o00ooo`007_l00?lNool001Co
0?l01_ooool000Cooooo3ol0o`0?ool000oo0?l03ol00?l?o`3o00ooo`003ol00?l?ool001oo003o
3ol0o`0No`00o`oo0?l03ooo000No`00oakoo`005?l0o`06ooooo`001?ooool?o`3o00ooo`003ol0
o`0?o`00o`oo0?l03ooo000?o`00o`ooo`007ol00?l?o`3o01ko003o3ol0o`0?ool001ko003o7_oo
000Do`3o00Kooooo0004ooooo`oo0?l03ooo000?o`3o00oo003o3ol0o`0?ool000oo003o3ooo000O
o`00o`oo0?l07_l00?l?o`3o00ooo`007_l00?lNool001Co0?l01_ooool000Cooooo3ol0o`0?ool0
00oo0?l03ol00?l?o`3o00ooo`003ol00?l?ool001oo003o3ol0o`0No`00o`oo0?l03ooo000No`00
oakoo`005?l0o`06ooooo`001?ooool?o`3o00ooo`003ol0o`0?o`00o`oo0?l03ooo000?o`00o`oo
o`007ol00?l?o`3o01ko003o3ol0o`0?ool001ko003o7_oo000Do`3o00Kooooo0004ooooo`oo0?l0
3ooo000?o`3o00oo003o3ol0o`0?ool000oo003o3ooo000Oo`00o`oo0?l07_l00?l?o`3o00ooo`00
7_l00?lNool001Co0?l01_ooool000Cooooo3ol0o`0?ool000oo0?l03ol00?l?o`3o00ooo`003ol0
0?l?ool001oo003o3ol0o`0No`00o`oo0?l03ooo000No`00oakoo`005?l0o`06ooooo`001?ooool?
o`3o00ooo`003ol0o`0?o`00o`oo0?l03ooo000?o`00o`ooo`007ol00?l?o`3o01ko003o3ol0o`0?
ool001ko003o7_oo000Do`3o00Kooooo0004ooooo`oo0?l03ooo000?o`3o00oo003o3ol0o`0?ool0
00oo003o3ooo000Oo`00o`oo0?l07_l00?l?o`3o00ooo`007_l00?lNool001Co0?l01_ooool000Co
oooo3ol0o`0?ool000oo0?l03ol00?l?o`3o00ooo`003ol00?l?ool001oo003o3ol0o`0No`00o`oo
0?l03ooo000No`00oakoo`005?l0o`06ooooo`001?ooool?o`3o00ooo`003ol0o`0?o`00o`oo0?l0
3ooo000?o`00o`ooo`007ol00?l?o`3o01ko003o3ol0o`0?ool001ko003o7_oo000Do`3o00Kooooo
0004ooooo`oo0?l03ooo000?o`3o00oo003o3ol0o`0?ool000oo003o3ooo000Oo`00o`oo0?l07_l0
0?l?o`3o00ooo`007_l00?lNool001Co0?l01_ooool000Cooooo3ol0o`0?ool000oo0?l03ol00?l?
o`3o00ooo`003ol00?l?ool001oo003o3ol0o`0No`00o`oo0?l03ooo000No`00oakoo`005?l0o`06
ooooo`001?ooool?ool000oo003o3ooo000?o`3o00oo003o3ol0o`0?o`00odcoo`003ol0o`0?ool0
01ko0?l03ol00?l?ool002?o003o1_ooool000Cooooo3ooo000?o`00o`ooo`003ol0o`0?o`00o`oo
0?l03ol00?m<ool000oo0?l03ooo000No`3o00oo003o3ooo000So`00o`Kooooo0004ooooo`ooo`00
3ol00?l?ool000oo0?l03ol00?l?o`3o00oo003oC?oo000?o`3o00ooo`007_l0o`0?o`00o`ooo`00
8ol00?l6ooooo`001?ooool?ool000oo003o3ooo000?o`3o00oo003o3ol0o`0?o`00odcoo`003ol0
o`0?ool001ko0?l03ol00?l?ool002?o003o1_ooool000Cooooo3ooo000?o`00o`ooo`003ol0o`0?
o`00o`oo0?l03ol00?m<ool000oo0?l03ooo000No`3o00oo003o3ooo000So`00o`Kooooo0004oooo
o`ooo`003ol00?l?ool000oo0?l03ol00?l?o`3o00oo003oC?oo000?o`3o00ooo`007_l0o`0?o`00
o`ooo`008ol00?l6ooooo`001?ooool?ool000oo003o3ooo000?o`3o00oo003o3ol0o`0?o`00odco
o`003ol0o`0?ool001ko0?l03ol00?l?ool002?o003o1_ooool000Cooooo3ooo000?o`00o`ooo`00
3ol0o`0?o`00o`oo0?l03ol00?m<ool000oo0?l03ooo000No`3o00oo003o3ooo000So`00o`Kooooo
0004ooooo`ooo`003ol00?l?ool000oo0?l03ol00?l?o`3o00oo003oC?oo000?o`3o00ooo`007_l0
o`0?o`00o`ooo`008ol00?l6ooooo`001?ooool?ool000oo003o3ooo000?o`3o00oo003o3ol0o`0?
o`00odcoo`003ol0o`0?ool001ko0?l03ol00?l?ool002?o003o1_ooool000Cooooo3ooo000?o`00
o`ooo`003ol0o`0?o`00o`oo0?l03ol00?m<ool000oo0?l03ooo000No`3o00oo003o3ooo000So`00
o`Kooooo0004ooooo`ooo`003ol00?l?ool000oo0?l03ol00?l?o`3o00oo003oC?oo000?o`3o00oo
o`007_l0o`0?o`00o`ooo`008ol00?l6ooooo`001?ooool?ool000oo003o3ooo000?o`3o00oo003o
3ol0o`0?o`00odcoo`003ol0o`0?ool001ko0?l03ol00?l?ool002?o003o1_ooool000Cooooo3ooo
000?o`00o`ooo`003ol0o`0?o`00o`oo0?l03ol00?m<ool000oo0?l03ooo000No`3o00oo003o3ooo
000So`00o`Kooooo0004ooooo`ooo`003ol00?l?ool000oo0?l03ol00?l?o`3o00oo003oC?oo000?
o`3o00ooo`007_l0o`0?o`00o`ooo`008ol00?l6ooooo`001?ooool?ool000oo0?l07_l00?lNool0
00oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006oooo
o`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko
003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o
7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?oo
ool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo
000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000O
o`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool0
00oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00
o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo
0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l0
7_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l0
5?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo
000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lN
ool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006
ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o
01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo
003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`00
1?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o
3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?ool000oo0?l07_l00?lNool000oo003o7_oo
000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?o`00o`oo0?l05?oo0006ooooo`001?ooool?
ool000oo0?l07_l00?lNool000oo003o7_oo000Oo`00o`oo0?l03ooo000?o`3o01ko003o3ooo000?
o`00o`oo0?l05?oo0006ooooo`001?ooool?o`00o`oo0?l03ol00?l?ool001ko0?l07_l00?lOool0
00oo003o3ol0o`0?ool000oo0?l03ooo000?o`00o`ooo`003ol00?lSool000Kooooo0004ooooo`oo
003o3ol0o`0?o`00o`ooo`007_l0o`0No`00oaooo`003ol00?l?o`3o00ooo`003ol0o`0?ool000oo
003o3ooo000?o`00ob?oo`001_ooool000Cooooo3ol00?l?o`3o00oo003o3ooo000No`3o01ko003o
7ooo000?o`00o`oo0?l03ooo000?o`3o00ooo`003ol00?l?ool000oo003o8ooo0006ooooo`001?oo
ool?o`00o`oo0?l03ol00?l?ool001ko0?l07_l00?lOool000oo003o3ol0o`0?ool000oo0?l03ooo
000?o`00o`ooo`003ol00?lSool000Kooooo0004ooooo`oo003o3ol0o`0?o`00o`ooo`007_l0o`0N
o`00oaooo`003ol00?l?o`3o00ooo`003ol0o`0?ool000oo003o3ooo000?o`00ob?oo`001_ooool0
00Cooooo3ol00?l?o`3o00oo003o3ooo000No`3o01ko003o7ooo000?o`00o`oo0?l03ooo000?o`3o
00ooo`003ol00?l?ool000oo003o8ooo0006ooooo`001?ooool?o`00o`oo0?l03ol00?l?ool001ko
0?l07_l00?lOool000oo003o3ol0o`0?ool000oo0?l03ooo000?o`00o`ooo`003ol00?lSool000Ko
oooo0004ooooo`oo003o3ol0o`0?o`00o`ooo`007_l0o`0No`00oaooo`003ol00?l?o`3o00ooo`00
3ol0o`0?ool000oo003o3ooo000?o`00ob?oo`001_ooool000Cooooo3ol00?l?o`3o00oo003o3ooo
000No`3o01ko003o7ooo000?o`00o`oo0?l03ooo000?o`3o00ooo`003ol00?l?ool000oo003o8ooo
0006ooooo`001?ooool?o`00o`oo0?l03ol00?l?ool001ko0?l07_l00?lOool000oo003o3ol0o`0?
ool000oo0?l03ooo000?o`00o`ooo`003ol00?lSool000Kooooo0004ooooo`oo003o3ol0o`0?o`00
o`ooo`007_l0o`0No`00oaooo`003ol00?l?o`3o00ooo`003ol0o`0?ool000oo003o3ooo000?o`00
ob?oo`001_ooool000Cooooo3ol00?l?o`3o00oo003o3ooo000No`3o01ko003o7ooo000?o`00o`oo
0?l03ooo000?o`3o00ooo`003ol00?l?ool000oo003o8ooo0006ooooo`001?ooool?o`00o`oo0?l0
3ol00?l?ool001ko0?l07_l00?lOool000oo003o3ol0o`0?ool000oo0?l03ooo000?o`00o`ooo`00
3ol00?lSool000Kooooo0004ooooo`oo003o3ol0o`0?o`00o`ooo`007_l0o`0No`00oaooo`003ol0
0?l?o`3o00ooo`003ol0o`0?ool000oo003o3ooo000?o`00ob?oo`001_ooool000Cooooo3ol00?l?
o`3o00oo003o3ooo000No`3o01ko003o7ooo000?o`00o`oo0?l03ooo000?o`3o00ooo`003ol00?l?
ool000oo003o8ooo0006ooooo`001?oooollo`00oako0?l03ol00?l?o`3o00oo003o4?l0o`0?ool0
03co003o3ol0o`0?ool000oo003o3ooo000Do`3o00Kooooo0004ooooocco003o7_l0o`0?o`00o`oo
0?l03ol00?l@o`3o00ooo`00??l00?l?o`3o00ooo`003ol00?l?ool001Co0?l01_ooool000Cooooo
??l00?lNo`3o00oo003o3ol0o`0?o`00oa3o0?l03ooo000lo`00o`oo0?l03ooo000?o`00o`ooo`00
5?l0o`06ooooo`001?oooollo`00oako0?l03ol00?l?o`3o00oo003o4?l0o`0?ool003co003o3ol0
o`0?ool000oo003o3ooo000Do`3o00Kooooo0004ooooocco003o7_l0o`0?o`00o`oo0?l03ol00?l@
o`3o00ooo`00??l00?l?o`3o00ooo`003ol00?l?ool001Co0?l01_ooool000Cooooo??l00?lNo`3o
00oo003o3ol0o`0?o`00oa3o0?l03ooo000lo`00o`oo0?l03ooo000?o`00o`ooo`005?l0o`06oooo
o`001?oooollo`00oako0?l03ol00?l?o`3o00oo003o4?l0o`0?ool003co003o3ol0o`0?ool000oo
003o3ooo000Do`3o00Kooooo0004ooooocco003o7_l0o`0?o`00o`oo0?l03ol00?l@o`3o00ooo`00
??l00?l?o`3o00ooo`003ol00?l?ool001Co0?l01_ooool000Cooooo??l00?lNo`3o00oo003o3ol0
o`0?o`00oa3o0?l03ooo000lo`00o`oo0?l03ooo000?o`00o`ooo`005?l0o`06ooooo`001?ooooll
o`00oako0?l03ol00?l?o`3o00oo003o4?l0o`0?ool003co003o3ol0o`0?ool000oo003o3ooo000D
o`3o00Kooooo0004ooooocco003o7_l0o`0?o`00o`oo0?l03ol00?l@o`3o00ooo`00??l00?l?o`3o
00ooo`003ol00?l?ool001Co0?l01_ooool000Cooooo??l00?lNo`3o00oo003o3ol0o`0?o`00oa3o
0?l03ooo000lo`00o`oo0?l03ooo000?o`00o`ooo`005?l0o`06ooooo`001?oooollo`00oako0?l0
3ol00?l?o`3o00oo003o4?l0o`0?ool003co003o3ol0o`0?ool000oo003o3ooo000Do`3o00Kooooo
0004ooooocco003o7_l0o`0?o`00o`oo0?l03ol00?l@o`3o00ooo`00??l00?l?o`3o00ooo`003ol0
0?l?ool001Co0?l01_ooool000Cooooo??l00?lNo`3o00oo003o3ol0o`0?o`00oa3o0?l03ooo000l
o`00o`oo0?l03ooo000?o`00o`ooo`005?l0o`06ooooo`001?ooool?o`00oakoo`003ol0o`0?o`00
oako0?l03ol00?lOo`3o00oo003o3ol0o`0Nool000oo003o3ooo000?o`00o`ooo`003ol0o`0Dool0
00Kooooo0004ooooo`oo003o7_oo000?o`3o00oo003o7_l0o`0?o`00oaoo0?l03ol00?l?o`3o01ko
o`003ol00?l?ool000oo003o3ooo000?o`3o01Coo`001_ooool000Cooooo3ol00?lNool000oo0?l0
3ol00?lNo`3o00oo003o7ol0o`0?o`00o`oo0?l07_oo000?o`00o`ooo`003ol00?l?ool000oo0?l0
5?oo0006ooooo`001?ooool?o`00oakoo`003ol0o`0?o`00oako0?l03ol00?lOo`3o00oo003o3ol0
o`0Nool000oo003o3ooo000?o`00o`ooo`003ol0o`0Dool000Kooooo0004ooooo`oo003o7_oo000?
o`3o00oo003o7_l0o`0?o`00oaoo0?l03ol00?l?o`3o01koo`003ol00?l?ool000oo003o3ooo000?
o`3o01Coo`001_ooool000Cooooo3ol00?lNool000oo0?l03ol00?lNo`3o00oo003o7ol0o`0?o`00
o`oo0?l07_oo000?o`00o`ooo`003ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?o`00oako
o`003ol0o`0?o`00oako0?l03ol00?lOo`3o00oo003o3ol0o`0Nool000oo003o3ooo000?o`00o`oo
o`003ol0o`0Dool000Kooooo0004ooooo`oo003o7_oo000?o`3o00oo003o7_l0o`0?o`00oaoo0?l0
3ol00?l?o`3o01koo`003ol00?l?ool000oo003o3ooo000?o`3o01Coo`001_ooool000Cooooo3ol0
0?lNool000oo0?l03ol00?lNo`3o00oo003o7ol0o`0?o`00o`oo0?l07_oo000?o`00o`ooo`003ol0
0?l?ool000oo0?l05?oo0006ooooo`001?ooool?o`00oakoo`003ol0o`0?o`00oako0?l03ol00?lO
o`3o00oo003o3ol0o`0Nool000oo003o3ooo000?o`00o`ooo`003ol0o`0Dool000Kooooo0004oooo
o`oo003o7_oo000?o`3o00oo003o7_l0o`0?o`00oaoo0?l03ol00?l?o`3o01koo`003ol00?l?ool0
00oo003o3ooo000?o`3o01Coo`001_ooool000Cooooo3ol00?lNool000oo0?l03ol00?lNo`3o00oo
003o7ol0o`0?o`00o`oo0?l07_oo000?o`00o`ooo`003ol00?l?ool000oo0?l05?oo0006ooooo`00
1?ooool?o`00oakoo`003ol0o`0?o`00oako0?l03ol00?lOo`3o00oo003o3ol0o`0Nool000oo003o
3ooo000?o`00o`ooo`003ol0o`0Dool000Kooooo0004ooooo`oo003o7_oo000?o`3o00oo003o7_l0
o`0?o`00oaoo0?l03ol00?l?o`3o01koo`003ol00?l?ool000oo003o3ooo000?o`3o01Coo`001_oo
ool000Cooooo3ol00?lNool000oo0?l03ol00?lNo`3o00oo003o7ol0o`0?o`00o`oo0?l07_oo000?
o`00o`ooo`003ol00?l?ool000oo0?l05?oo0006ooooo`001?oooolNo`00o`oo0?l03ol00?l?ool0
02go0?l03ol00?l@ool001ko003o3ol0o`0?ool000oo003o3ol0o`0?ool000oo0?l03ol00?lDool0
00Kooooo0004oooooako003o3ol0o`0?o`00o`ooo`00;Ol0o`0?o`00oa3oo`007_l00?l?o`3o00oo
o`003ol00?l?o`3o00ooo`003ol0o`0?o`00oaCoo`001_ooool000Cooooo7_l00?l?o`3o00oo003o
3ooo000]o`3o00oo003o4?oo000No`00o`oo0?l03ooo000?o`00o`oo0?l03ooo000?o`3o00oo003o
5?oo0006ooooo`001?oooolNo`00o`oo0?l03ol00?l?ool002go0?l03ol00?l@ool001ko003o3ol0
o`0?ool000oo003o3ol0o`0?ool000oo0?l03ol00?lDool000Kooooo0004oooooako003o3ol0o`0?
o`00o`ooo`00;Ol0o`0?o`00oa3oo`007_l00?l?o`3o00ooo`003ol00?l?o`3o00ooo`003ol0o`0?
o`00oaCoo`001_ooool000Cooooo7_l00?l?o`3o00oo003o3ooo000]o`3o00oo003o4?oo000No`00
o`oo0?l03ooo000?o`00o`oo0?l03ooo000?o`3o00oo003o5?oo0006ooooo`001?oooolNo`00o`oo
0?l03ol00?l?ool002go0?l03ol00?l@ool001ko003o3ol0o`0?ool000oo003o3ol0o`0?ool000oo
0?l03ol00?lDool000Kooooo0004oooooako003o3ol0o`0?o`00o`ooo`00;Ol0o`0?o`00oa3oo`00
7_l00?l?o`3o00ooo`003ol00?l?o`3o00ooo`003ol0o`0?o`00oaCoo`001_ooool000Cooooo7_l0
0?l?o`3o00oo003o3ooo000]o`3o00oo003o4?oo000No`00o`oo0?l03ooo000?o`00o`oo0?l03ooo
000?o`3o00oo003o5?oo0006ooooo`001?oooolNo`00o`oo0?l03ol00?l?ool002go0?l03ol00?l@
ool001ko003o3ol0o`0?ool000oo003o3ol0o`0?ool000oo0?l03ol00?lDool000Kooooo0004oooo
oako003o3ol0o`0?o`00o`ooo`00;Ol0o`0?o`00oa3oo`007_l00?l?o`3o00ooo`003ol00?l?o`3o
00ooo`003ol0o`0?o`00oaCoo`001_ooool000Cooooo7_l00?l?o`3o00oo003o3ooo000]o`3o00oo
003o4?oo000No`00o`oo0?l03ooo000?o`00o`oo0?l03ooo000?o`3o00oo003o5?oo0006ooooo`00
1?oooolNo`00o`oo0?l03ol00?l?ool002go0?l03ol00?l@ool001ko003o3ol0o`0?ool000oo003o
3ol0o`0?ool000oo0?l03ol00?lDool000Kooooo0004oooooako003o3ol0o`0?o`00o`ooo`00;Ol0
o`0?o`00oa3oo`007_l00?l?o`3o00ooo`003ol00?l?o`3o00ooo`003ol0o`0?o`00oaCoo`001_oo
ool000Cooooo7_l00?l?o`3o00oo003o3ooo000]o`3o00oo003o4?oo000No`00o`oo0?l03ooo000?
o`00o`oo0?l03ooo000?o`3o00oo003o5?oo0006ooooo`001?ooool?o`00o`oo0?l03ol00?lNool0
01ko0?l03ol00?l?o`3o013oo`007_l00?l?o`3o00ooo`007_l0o`0?o`00oc;oo`001_ooool000Co
oooo3ol00?l?o`3o00oo003o7_oo000No`3o00oo003o3ol0o`0@ool001ko003o3ol0o`0?ool001ko
0?l03ol00?lbool000Kooooo0004ooooo`oo003o3ol0o`0?o`00oakoo`007_l0o`0?o`00o`oo0?l0
4?oo000No`00o`oo0?l03ooo000No`3o00oo003o<_oo0006ooooo`001?ooool?o`00o`oo0?l03ol0
0?lNool001ko0?l03ol00?l?o`3o013oo`007_l00?l?o`3o00ooo`007_l0o`0?o`00oc;oo`001_oo
ool000Cooooo3ol00?l?o`3o00oo003o7_oo000No`3o00oo003o3ol0o`0@ool001ko003o3ol0o`0?
ool001ko0?l03ol00?lbool000Kooooo0004ooooo`oo003o3ol0o`0?o`00oakoo`007_l0o`0?o`00
o`oo0?l04?oo000No`00o`oo0?l03ooo000No`3o00oo003o<_oo0006ooooo`001?ooool?o`00o`oo
0?l03ol00?lNool001ko0?l03ol00?l?o`3o013oo`007_l00?l?o`3o00ooo`007_l0o`0?o`00oc;o
o`001_ooool000Cooooo3ol00?l?o`3o00oo003o7_oo000No`3o00oo003o3ol0o`0@ool001ko003o
3ol0o`0?ool001ko0?l03ol00?lbool000Kooooo0004ooooo`oo003o3ol0o`0?o`00oakoo`007_l0
o`0?o`00o`oo0?l04?oo000No`00o`oo0?l03ooo000No`3o00oo003o<_oo0006ooooo`001?ooool?
o`00o`oo0?l03ol00?lNool001ko0?l03ol00?l?o`3o013oo`007_l00?l?o`3o00ooo`007_l0o`0?
o`00oc;oo`001_ooool000Cooooo3ol00?l?o`3o00oo003o7_oo000No`3o00oo003o3ol0o`0@ool0
01ko003o3ol0o`0?ool001ko0?l03ol00?lbool000Kooooo0004ooooo`oo003o3ol0o`0?o`00oako
o`007_l0o`0?o`00o`oo0?l04?oo000No`00o`oo0?l03ooo000No`3o00oo003o<_oo0006ooooo`00
1?ooool?o`00o`oo0?l03ol00?lNool001ko0?l03ol00?l?o`3o013oo`007_l00?l?o`3o00ooo`00
7_l0o`0?o`00oc;oo`001_ooool000Cooooo3ol00?l?o`3o00oo003o7_oo000No`3o00oo003o3ol0
o`0@ool001ko003o3ol0o`0?ool001ko0?l03ol00?lbool000Kooooo0004ooooo`oo003o3ol0o`0?
o`00oakoo`007_l0o`0?o`00o`oo0?l04?oo000No`00o`oo0?l03ooo000No`3o00oo003o<_oo0006
ooooo`001?ooool?o`00oakoo`007_l0o`0Nool005_o003o3ol0o`0?ool000oo0?l07_oo000?o`00
oaCo0?l01_ooool000Cooooo3ol00?lNool001ko0?l07_oo001Ko`00o`oo0?l03ooo000?o`3o01ko
o`003ol00?lDo`3o00Kooooo0004ooooo`oo003o7_oo000No`3o01koo`00Fol00?l?o`3o00ooo`00
3ol0o`0Nool000oo003o5?l0o`06ooooo`001?ooool?o`00oakoo`007_l0o`0Nool005_o003o3ol0
o`0?ool000oo0?l07_oo000?o`00oaCo0?l01_ooool000Cooooo3ol00?lNool001ko0?l07_oo001K
o`00o`oo0?l03ooo000?o`3o01koo`003ol00?lDo`3o00Kooooo0004ooooo`oo003o7_oo000No`3o
01koo`00Fol00?l?o`3o00ooo`003ol0o`0Nool000oo003o5?l0o`06ooooo`001?ooool?o`00oako
o`007_l0o`0Nool005_o003o3ol0o`0?ool000oo0?l07_oo000?o`00oaCo0?l01_ooool000Cooooo
3ol00?lNool001ko0?l07_oo001Ko`00o`oo0?l03ooo000?o`3o01koo`003ol00?lDo`3o00Kooooo
0004ooooo`oo003o7_oo000No`3o01koo`00Fol00?l?o`3o00ooo`003ol0o`0Nool000oo003o5?l0
o`06ooooo`001?ooool?o`00oakoo`007_l0o`0Nool005_o003o3ol0o`0?ool000oo0?l07_oo000?
o`00oaCo0?l01_ooool000Cooooo3ol00?lNool001ko0?l07_oo001Ko`00o`oo0?l03ooo000?o`3o
01koo`003ol00?lDo`3o00Kooooo0004ooooo`oo003o7_oo000No`3o01koo`00Fol00?l?o`3o00oo
o`003ol0o`0Nool000oo003o5?l0o`06ooooo`001?ooool?o`00oakoo`007_l0o`0Nool005_o003o
3ol0o`0?ool000oo0?l07_oo000?o`00oaCo0?l01_ooool000Cooooo3ol00?lNool001ko0?l07_oo
001Ko`00o`oo0?l03ooo000?o`3o01koo`003ol00?lDo`3o00Kooooo0004ooooo`oo003o7_oo000N
o`3o01koo`00Fol00?l?o`3o00ooo`003ol0o`0Nool000oo003o5?l0o`06ooooo`001?ooool?o`3o
00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006oooo
o`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o
003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`00
3ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l0
0?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0N
o`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o
00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006oooo
o`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o
003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`00
3ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l0
0?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0N
o`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o
00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006oooo
o`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o
003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`00
3ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0No`00occo0?l04?l0
0?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o00oo003o3ol0o`0N
o`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006ooooo`001?ooool?o`3o
00oo003o3ol0o`0No`00occo0?l04?l00?l?o`3o00ooo`003ol00?l?o`3o04_o003o5?oo0006oooo
o`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02go
o`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l0
3ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0
o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0N
o`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool0
01ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06oooo
o`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02go
o`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l0
3ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0
o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0N
o`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool0
01ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06oooo
o`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02go
o`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l0
3ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0No`00oakoo`007ol0
o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool001ko003o3ol0o`0N
o`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06ooooo`001?ooool?ool0
01ko003o3ol0o`0No`00oakoo`007ol0o`0?o`00oako0?l03ol00?l?o`3o02goo`008ol0o`06oooo
o`001?ooool?o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00ooo`00;_l0o`0No`00oako
0?l07_l00?l?o`3o02?o003o1_ooool000Cooooo3ol0o`0?o`00o`ooo`003ol00?l?ool000oo003o
3ol0o`0?ool002ko0?l07_l00?lNo`3o01ko003o3ol0o`0So`00o`Kooooo0004ooooo`oo0?l03ol0
0?l?ool000oo003o3ooo000?o`00o`oo0?l03ooo000^o`3o01ko003o7_l0o`0No`00o`oo0?l08ol0
0?l6ooooo`001?ooool?o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00ooo`00;_l0o`0N
o`00oako0?l07_l00?l?o`3o02?o003o1_ooool000Cooooo3ol0o`0?o`00o`ooo`003ol00?l?ool0
00oo003o3ol0o`0?ool002ko0?l07_l00?lNo`3o01ko003o3ol0o`0So`00o`Kooooo0004ooooo`oo
0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ooo000^o`3o01ko003o7_l0o`0No`00o`oo
0?l08ol00?l6ooooo`001?ooool?o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00ooo`00
;_l0o`0No`00oako0?l07_l00?l?o`3o02?o003o1_ooool000Cooooo3ol0o`0?o`00o`ooo`003ol0
0?l?ool000oo003o3ol0o`0?ool002ko0?l07_l00?lNo`3o01ko003o3ol0o`0So`00o`Kooooo0004
ooooo`oo0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ooo000^o`3o01ko003o7_l0o`0N
o`00o`oo0?l08ol00?l6ooooo`001?ooool?o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o
00ooo`00;_l0o`0No`00oako0?l07_l00?l?o`3o02?o003o1_ooool000Cooooo3ol0o`0?o`00o`oo
o`003ol00?l?ool000oo003o3ol0o`0?ool002ko0?l07_l00?lNo`3o01ko003o3ol0o`0So`00o`Ko
oooo0004ooooo`oo0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ooo000^o`3o01ko003o
7_l0o`0No`00o`oo0?l08ol00?l6ooooo`001?ooool?o`3o00oo003o3ooo000?o`00o`ooo`003ol0
0?l?o`3o00ooo`00;_l0o`0No`00oako0?l07_l00?l?o`3o02?o003o1_ooool000Cooooo3ol0o`0?
o`00o`ooo`003ol00?l?ool000oo003o3ol0o`0?ool002ko0?l07_l00?lNo`3o01ko003o3ol0o`0S
o`00o`Kooooo0004ooooo`oo0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ooo000^o`3o
01ko003o7_l0o`0No`00o`oo0?l08ol00?l6ooooo`001?ooool?ool000oo003o7_oo000]o`00oako
0?l07ol00?lNool000oo0?l07_l00?l?o`3o03;oo`001_ooool000Cooooo3ooo000?o`00oakoo`00
;Ol00?lNo`3o01oo003o7_oo000?o`3o01ko003o3ol0o`0bool000Kooooo0004ooooo`ooo`003ol0
0?lNool002go003o7_l0o`0Oo`00oakoo`003ol0o`0No`00o`oo0?l0<_oo0006ooooo`001?ooool?
ool000oo003o7_oo000]o`00oako0?l07ol00?lNool000oo0?l07_l00?l?o`3o03;oo`001_ooool0
00Cooooo3ooo000?o`00oakoo`00;Ol00?lNo`3o01oo003o7_oo000?o`3o01ko003o3ol0o`0bool0
00Kooooo0004ooooo`ooo`003ol00?lNool002go003o7_l0o`0Oo`00oakoo`003ol0o`0No`00o`oo
0?l0<_oo0006ooooo`001?ooool?ool000oo003o7_oo000]o`00oako0?l07ol00?lNool000oo0?l0
7_l00?l?o`3o03;oo`001_ooool000Cooooo3ooo000?o`00oakoo`00;Ol00?lNo`3o01oo003o7_oo
000?o`3o01ko003o3ol0o`0bool000Kooooo0004ooooo`ooo`003ol00?lNool002go003o7_l0o`0O
o`00oakoo`003ol0o`0No`00o`oo0?l0<_oo0006ooooo`001?ooool?ool000oo003o7_oo000]o`00
oako0?l07ol00?lNool000oo0?l07_l00?l?o`3o03;oo`001_ooool000Cooooo3ooo000?o`00oako
o`00;Ol00?lNo`3o01oo003o7_oo000?o`3o01ko003o3ol0o`0bool000Kooooo0004ooooo`ooo`00
3ol00?lNool002go003o7_l0o`0Oo`00oakoo`003ol0o`0No`00o`oo0?l0<_oo0006ooooo`001?oo
ool?ool000oo003o7_oo000]o`00oako0?l07ol00?lNool000oo0?l07_l00?l?o`3o03;oo`001_oo
ool000Cooooo3ooo000?o`00oakoo`00;Ol00?lNo`3o01oo003o7_oo000?o`3o01ko003o3ol0o`0b
ool000Kooooo0004ooooo`ooo`003ol00?lNool002go003o7_l0o`0Oo`00oakoo`003ol0o`0No`00
o`oo0?l0<_oo0006ooooo`001?ooool?ool000oo0?l07_l00?l?ool000oo003o3ooo000^o`00o`oo
0?l07_l00?l?o`3o00ooo`00??l0o`0Do`00o`Kooooo0004ooooo`ooo`003ol0o`0No`00o`ooo`00
3ol00?l?ool002ko003o3ol0o`0No`00o`oo0?l03ooo000lo`3o01Co003o1_ooool000Cooooo3ooo
000?o`3o01ko003o3ooo000?o`00o`ooo`00;_l00?l?o`3o01ko003o3ol0o`0?ool003co0?l05?l0
0?l6ooooo`001?ooool?ool000oo0?l07_l00?l?ool000oo003o3ooo000^o`00o`oo0?l07_l00?l?
o`3o00ooo`00??l0o`0Do`00o`Kooooo0004ooooo`ooo`003ol0o`0No`00o`ooo`003ol00?l?ool0
02ko003o3ol0o`0No`00o`oo0?l03ooo000lo`3o01Co003o1_ooool000Cooooo3ooo000?o`3o01ko
003o3ooo000?o`00o`ooo`00;_l00?l?o`3o01ko003o3ol0o`0?ool003co0?l05?l00?l6ooooo`00
1?ooool?ool000oo0?l07_l00?l?ool000oo003o3ooo000^o`00o`oo0?l07_l00?l?o`3o00ooo`00
??l0o`0Do`00o`Kooooo0004ooooo`ooo`003ol0o`0No`00o`ooo`003ol00?l?ool002ko003o3ol0
o`0No`00o`oo0?l03ooo000lo`3o01Co003o1_ooool000Cooooo3ooo000?o`3o01ko003o3ooo000?
o`00o`ooo`00;_l00?l?o`3o01ko003o3ol0o`0?ool003co0?l05?l00?l6ooooo`001?ooool?ool0
00oo0?l07_l00?l?ool000oo003o3ooo000^o`00o`oo0?l07_l00?l?o`3o00ooo`00??l0o`0Do`00
o`Kooooo0004ooooo`ooo`003ol0o`0No`00o`ooo`003ol00?l?ool002ko003o3ol0o`0No`00o`oo
0?l03ooo000lo`3o01Co003o1_ooool000Cooooo3ooo000?o`3o01ko003o3ooo000?o`00o`ooo`00
;_l00?l?o`3o01ko003o3ol0o`0?ool003co0?l05?l00?l6ooooo`001?ooool?ool000oo0?l07_l0
0?l?ool000oo003o3ooo000^o`00o`oo0?l07_l00?l?o`3o00ooo`00??l0o`0Do`00o`Kooooo0004
ooooo`ooo`003ol0o`0No`00o`ooo`003ol00?l?ool002ko003o3ol0o`0No`00o`oo0?l03ooo000l
o`3o01Co003o1_ooool000Cooooo3ooo000?o`3o01ko003o3ooo000?o`00o`ooo`00;_l00?l?o`3o
01ko003o3ol0o`0?ool003co0?l05?l00?l6ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o
0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`00
1?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l0
3ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l0
0?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?
ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?
ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o
00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo
003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo
0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o
3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0
o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo
0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?
ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool0
04_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006oooo
o`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko
0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l0
4?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?oo
ool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol0
0?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool000oo003o3ol0o`0?ool004_o0?l04?l00?l?
o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool000oo0?l05?oo0006ooooo`001?ooool?ool0
00oo003o3ol0o`0?ool004_o0?l04?l00?l?o`3o00oo003o3ol0o`0?ool001ko0?l03ol00?l?ool0
00oo0?l05?oo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00oo
o`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l0
7_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0
o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?
o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006
ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool0
01ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00oo
o`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l0
7_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0
o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?
o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006
ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool0
01ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00oo
o`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l0
7_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?o`3o01koo`003ol0
o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006ooooo`001?ooool?
o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool001ko0?l08ooo0006
ooooo`001?ooool?o`3o01koo`003ol0o`0Nool000oo0?l07_oo000^o`3o00ooo`007_l00?l?ool0
01ko0?l08ooo0006ooooo`001?ooool?ool000oo003o7_oo000?o`3o00oo003o3ol0o`0?o`00obko
0?l03ol00?l?ool000oo0?l07_oo000No`3o02?o003o1_ooool000Cooooo3ooo000?o`00oakoo`00
3ol0o`0?o`00o`oo0?l03ol00?l^o`3o00oo003o3ooo000?o`3o01koo`007_l0o`0So`00o`Kooooo
0004ooooo`ooo`003ol00?lNool000oo0?l03ol00?l?o`3o00oo003o;_l0o`0?o`00o`ooo`003ol0
o`0Nool001ko0?l08ol00?l6ooooo`001?ooool?ool000oo003o7_oo000?o`3o00oo003o3ol0o`0?
o`00obko0?l03ol00?l?ool000oo0?l07_oo000No`3o02?o003o1_ooool000Cooooo3ooo000?o`00
oakoo`003ol0o`0?o`00o`oo0?l03ol00?l^o`3o00oo003o3ooo000?o`3o01koo`007_l0o`0So`00
o`Kooooo0004ooooo`ooo`003ol00?lNool000oo0?l03ol00?l?o`3o00oo003o;_l0o`0?o`00o`oo
o`003ol0o`0Nool001ko0?l08ol00?l6ooooo`001?ooool?ool000oo003o7_oo000?o`3o00oo003o
3ol0o`0?o`00obko0?l03ol00?l?ool000oo0?l07_oo000No`3o02?o003o1_ooool000Cooooo3ooo
000?o`00oakoo`003ol0o`0?o`00o`oo0?l03ol00?l^o`3o00oo003o3ooo000?o`3o01koo`007_l0
o`0So`00o`Kooooo0004ooooo`ooo`003ol00?lNool000oo0?l03ol00?l?o`3o00oo003o;_l0o`0?
o`00o`ooo`003ol0o`0Nool001ko0?l08ol00?l6ooooo`001?ooool?ool000oo003o7_oo000?o`3o
00oo003o3ol0o`0?o`00obko0?l03ol00?l?ool000oo0?l07_oo000No`3o02?o003o1_ooool000Co
oooo3ooo000?o`00oakoo`003ol0o`0?o`00o`oo0?l03ol00?l^o`3o00oo003o3ooo000?o`3o01ko
o`007_l0o`0So`00o`Kooooo0004ooooo`ooo`003ol00?lNool000oo0?l03ol00?l?o`3o00oo003o
;_l0o`0?o`00o`ooo`003ol0o`0Nool001ko0?l08ol00?l6ooooo`001?ooool?ool000oo003o7_oo
000?o`3o00oo003o3ol0o`0?o`00obko0?l03ol00?l?ool000oo0?l07_oo000No`3o02?o003o1_oo
ool000Cooooo3ooo000?o`00oakoo`003ol0o`0?o`00o`oo0?l03ol00?l^o`3o00oo003o3ooo000?
o`3o01koo`007_l0o`0So`00o`Kooooo0004ooooo`ooo`003ol00?lNool000oo0?l03ol00?l?o`3o
00oo003o;_l0o`0?o`00o`ooo`003ol0o`0Nool001ko0?l08ol00?l6ooooo`001?ooool?ool002go
003o7_oo000?o`00obkoo`003ol00?lNool001ko003o7_l0o`0Nool001Co0?l01_ooool000Cooooo
3ooo000]o`00oakoo`003ol00?l^ool000oo003o7_oo000No`00oako0?l07_oo000Do`3o00Kooooo
0004ooooo`ooo`00;Ol00?lNool000oo003o;_oo000?o`00oakoo`007_l00?lNo`3o01koo`005?l0
o`06ooooo`001?ooool?ool002go003o7_oo000?o`00obkoo`003ol00?lNool001ko003o7_l0o`0N
ool001Co0?l01_ooool000Cooooo3ooo000]o`00oakoo`003ol00?l^ool000oo003o7_oo000No`00
oako0?l07_oo000Do`3o00Kooooo0004ooooo`ooo`00;Ol00?lNool000oo003o;_oo000?o`00oako
o`007_l00?lNo`3o01koo`005?l0o`06ooooo`001?ooool?ool002go003o7_oo000?o`00obkoo`00
3ol00?lNool001ko003o7_l0o`0Nool001Co0?l01_ooool000Cooooo3ooo000]o`00oakoo`003ol0
0?l^ool000oo003o7_oo000No`00oako0?l07_oo000Do`3o00Kooooo0004ooooo`ooo`00;Ol00?lN
ool000oo003o;_oo000?o`00oakoo`007_l00?lNo`3o01koo`005?l0o`06ooooo`001?ooool?ool0
02go003o7_oo000?o`00obkoo`003ol00?lNool001ko003o7_l0o`0Nool001Co0?l01_ooool000Co
oooo3ooo000]o`00oakoo`003ol00?l^ool000oo003o7_oo000No`00oako0?l07_oo000Do`3o00Ko
oooo0004ooooo`ooo`00;Ol00?lNool000oo003o;_oo000?o`00oakoo`007_l00?lNo`3o01koo`00
5?l0o`06ooooo`001?ooool?ool002go003o7_oo000?o`00obkoo`003ol00?lNool001ko003o7_l0
o`0Nool001Co0?l01_ooool000Cooooo3ooo000]o`00oakoo`003ol00?l^ool000oo003o7_oo000N
o`00oako0?l07_oo000Do`3o00Kooooo0004ooooo`ooo`00;Ol00?lNool000oo003o;_oo000?o`00
oakoo`007_l00?lNo`3o01koo`005?l0o`06ooooo`001?oooolNo`00o`ooo`003ol00?l?ool000oo
003o3ol0o`0^ool002go0?l03ooo000?o`00o`ooo`007_l0o`0Sool000Kooooo0004oooooako003o
3ooo000?o`00o`ooo`003ol00?l?o`3o02koo`00;Ol0o`0?ool000oo003o3ooo000No`3o02?oo`00
1_ooool000Cooooo7_l00?l?ool000oo003o3ooo000?o`00o`oo0?l0;_oo000]o`3o00ooo`003ol0
0?l?ool001ko0?l08ooo0006ooooo`001?oooolNo`00o`ooo`003ol00?l?ool000oo003o3ol0o`0^
ool002go0?l03ooo000?o`00o`ooo`007_l0o`0Sool000Kooooo0004oooooako003o3ooo000?o`00
o`ooo`003ol00?l?o`3o02koo`00;Ol0o`0?ool000oo003o3ooo000No`3o02?oo`001_ooool000Co
oooo7_l00?l?ool000oo003o3ooo000?o`00o`oo0?l0;_oo000]o`3o00ooo`003ol00?l?ool001ko
0?l08ooo0006ooooo`001?oooolNo`00o`ooo`003ol00?l?ool000oo003o3ol0o`0^ool002go0?l0
3ooo000?o`00o`ooo`007_l0o`0Sool000Kooooo0004oooooako003o3ooo000?o`00o`ooo`003ol0
0?l?o`3o02koo`00;Ol0o`0?ool000oo003o3ooo000No`3o02?oo`001_ooool000Cooooo7_l00?l?
ool000oo003o3ooo000?o`00o`oo0?l0;_oo000]o`3o00ooo`003ol00?l?ool001ko0?l08ooo0006
ooooo`001?oooolNo`00o`ooo`003ol00?l?ool000oo003o3ol0o`0^ool002go0?l03ooo000?o`00
o`ooo`007_l0o`0Sool000Kooooo0004oooooako003o3ooo000?o`00o`ooo`003ol00?l?o`3o02ko
o`00;Ol0o`0?ool000oo003o3ooo000No`3o02?oo`001_ooool000Cooooo7_l00?l?ool000oo003o
3ooo000?o`00o`oo0?l0;_oo000]o`3o00ooo`003ol00?l?ool001ko0?l08ooo0006ooooo`001?oo
oolNo`00o`ooo`003ol00?l?ool000oo003o3ol0o`0^ool002go0?l03ooo000?o`00o`ooo`007_l0
o`0Sool000Kooooo0004oooooako003o3ooo000?o`00o`ooo`003ol00?l?o`3o02koo`00;Ol0o`0?
ool000oo003o3ooo000No`3o02?oo`001_ooool000Cooooo7_l00?l?ool000oo003o3ooo000?o`00
o`oo0?l0;_oo000]o`3o00ooo`003ol00?l?ool001ko0?l08ooo0006ooooo`001?oooolNo`00oako
0?l03ol00?lNo`3o00ooo`003ol00?l@o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00oo
003o7_l0o`0Do`00o`Kooooo0004oooooako003o7_l0o`0?o`00oako0?l03ooo000?o`00oa3o0?l0
3ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ol00?lNo`3o01Co003o1_ooool000Cooooo7_l0
0?lNo`3o00oo003o7_l0o`0?ool000oo003o4?l0o`0?o`00o`ooo`003ol00?l?ool000oo003o3ol0
o`0?o`00oako0?l05?l00?l6ooooo`001?oooolNo`00oako0?l03ol00?lNo`3o00ooo`003ol00?l@
o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00oo003o7_l0o`0Do`00o`Kooooo0004oooo
oako003o7_l0o`0?o`00oako0?l03ooo000?o`00oa3o0?l03ol00?l?ool000oo003o3ooo000?o`00
o`oo0?l03ol00?lNo`3o01Co003o1_ooool000Cooooo7_l00?lNo`3o00oo003o7_l0o`0?ool000oo
003o4?l0o`0?o`00o`ooo`003ol00?l?ool000oo003o3ol0o`0?o`00oako0?l05?l00?l6ooooo`00
1?oooolNo`00oako0?l03ol00?lNo`3o00ooo`003ol00?l@o`3o00oo003o3ooo000?o`00o`ooo`00
3ol00?l?o`3o00oo003o7_l0o`0Do`00o`Kooooo0004oooooako003o7_l0o`0?o`00oako0?l03ooo
000?o`00oa3o0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ol00?lNo`3o01Co003o1_oo
ool000Cooooo7_l00?lNo`3o00oo003o7_l0o`0?ool000oo003o4?l0o`0?o`00o`ooo`003ol00?l?
ool000oo003o3ol0o`0?o`00oako0?l05?l00?l6ooooo`001?oooolNo`00oako0?l03ol00?lNo`3o
00ooo`003ol00?l@o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00oo003o7_l0o`0Do`00
o`Kooooo0004oooooako003o7_l0o`0?o`00oako0?l03ooo000?o`00oa3o0?l03ol00?l?ool000oo
003o3ooo000?o`00o`oo0?l03ol00?lNo`3o01Co003o1_ooool000Cooooo7_l00?lNo`3o00oo003o
7_l0o`0?ool000oo003o4?l0o`0?o`00o`ooo`003ol00?l?ool000oo003o3ol0o`0?o`00oako0?l0
5?l00?l6ooooo`001?oooolNo`00oako0?l03ol00?lNo`3o00ooo`003ol00?l@o`3o00oo003o3ooo
000?o`00o`ooo`003ol00?l?o`3o00oo003o7_l0o`0Do`00o`Kooooo0004oooooako003o7_l0o`0?
o`00oako0?l03ooo000?o`00oa3o0?l03ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ol00?lN
o`3o01Co003o1_ooool000Cooooo7_l00?lNo`3o00oo003o7_l0o`0?ool000oo003o4?l0o`0?o`00
o`ooo`003ol00?l?ool000oo003o3ol0o`0?o`00oako0?l05?l00?l6ooooo`001?oooolNo`00oako
0?l03ol00?lNo`3o00ooo`003ol00?l@o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00oo
003o7_l0o`0Do`00o`Kooooo0004oooooako003o7_l0o`0?o`00oako0?l03ooo000?o`00oa3o0?l0
3ol00?l?ool000oo003o3ooo000?o`00o`oo0?l03ol00?lNo`3o01Co003o1_ooool000Cooooo7_l0
0?lNo`3o00oo003o7_l0o`0?ool000oo003o4?l0o`0?o`00o`ooo`003ol00?l?ool000oo003o3ol0
o`0?o`00oako0?l05?l00?l6ooooo`001?oooolNo`00oako0?l03ol00?lNo`3o00ooo`003ol00?l@
o`3o00oo003o3ooo000?o`00o`ooo`003ol00?l?o`3o00oo003o7_l0o`0Do`00o`Kooooo0004oooo
oako003o7_l0o`0?o`00oako0?l03ooo000?o`00oa3o0?l03ol00?l?ool000oo003o3ooo000?o`00
o`oo0?l03ol00?lNo`3o01Co003o1_ooool00?oooooo?Oooool00?oooooo?Oooool00?oooooo?Ooo
ool00?oooooo?Oooool00001\
\>"],
  ImageRangeCache->{{{0, 315}, {315, 0}} -> {-0.0500054, -0.0500054, \
0.0034921, 0.0034921}},
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell["Run the model forward 100 generations", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \(Evolve[m4, Generations \[Rule] 100]\)], "Input",
  CellLabel->"In[11]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell["\<\
The model settles into cycles where the naturally expected invasions occur:\
\>", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[{
    \(\(graphlist = {};\)\), "\[IndentingNewLine]", 
    \(\(Do[\[IndentingNewLine]AppendTo[graphlist, 
          ExtractGraph[m4]]; \[IndentingNewLine]Evolve[
          m4], {12}\[IndentingNewLine]];\)\)}], "Input",
  CellLabel->"In[17]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowGraphArray[\[IndentingNewLine]Partition[graphlist, 4], 
        VertexStyle \[Rule] Box[Large]\[IndentingNewLine]];\)\)], "Input",
  CellLabel->"In[19]:=",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .74419 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.0177187 0.238095 [
[ 0 0 0 0 ]
[ 1 .74419 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .74419 L
0 .74419 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0177187 0.245293 0.239203 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
1 0 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
0 0 1 r
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
0 1 0 r
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
0 0 1 r
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
1 0 0 r
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
0 1 0 r
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
0 0 1 r
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
0 1 0 r
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
0 0 1 r
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
1 0 0 r
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
0 1 0 r
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 0 1 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
0 1 0 r
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
0 0 1 r
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
1 0 0 r
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
0 1 0 r
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 0 1 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
0 1 0 r
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
0 0 1 r
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
1 0 0 r
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
0 1 0 r
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 0 1 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
0 1 0 r
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
1 0 0 r
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
0 1 0 r
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
0 0 1 r
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
1 0 0 r
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
0 1 0 r
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
1 0 0 r
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
0 1 0 r
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
0 0 1 r
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
1 0 0 r
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 1 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
1 0 0 r
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
0 1 0 r
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
0 0 1 r
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
1 0 0 r
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 1 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
1 0 0 r
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
0 1 0 r
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
0 0 1 r
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
1 0 0 r
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 1 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
1 0 0 r
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
0 0 1 r
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
1 0 0 r
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
0 1 0 r
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
0 0 1 r
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
1 0 0 r
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 1 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
1 0 0 r
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
0 0 1 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
1 0 0 r
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
0 1 0 r
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
0 0 1 r
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 1 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
1 0 0 r
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 1 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 0 1 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
1 0 0 r
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
0 1 0 r
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
0 0 1 r
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 1 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
0 0 1 r
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
1 0 0 r
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
0 1 0 r
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
0 0 1 r
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
0 1 0 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
0 0 1 r
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
1 0 0 r
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
0 1 0 r
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
0 0 1 r
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
0 1 0 r
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
0 0 1 r
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
1 0 0 r
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
0 1 0 r
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
0 0 1 r
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 1 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
0 0 1 r
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
1 0 0 r
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
0 1 0 r
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
0 0 1 r
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 1 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
0 0 1 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
1 0 0 r
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
0 1 0 r
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
0 0 1 r
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 1 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
1 0 0 r
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 0 1 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
1 0 0 r
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
0 1 0 r
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
0 0 1 r
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
1 0 0 r
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 1 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
1 0 0 r
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
0 0 1 r
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
1 0 0 r
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
0 1 0 r
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
0 0 1 r
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
1 0 0 r
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 1 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
1 0 0 r
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
0 1 0 r
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
0 0 1 r
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
1 0 0 r
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 1 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
1 0 0 r
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
0 1 0 r
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
0 0 1 r
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
1 0 0 r
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.267442 0.0177187 0.488926 0.239203 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 1 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
0 1 0 r
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
0 0 1 r
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
1 0 0 r
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
0 1 0 r
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 0 1 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
0 1 0 r
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
0 0 1 r
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
1 0 0 r
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
0 1 0 r
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 0 1 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
0 1 0 r
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
0 0 1 r
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
1 0 0 r
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
0 1 0 r
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 0 1 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
0 1 0 r
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
1 0 0 r
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
0 1 0 r
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
0 0 1 r
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
1 0 0 r
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
0 1 0 r
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
1 0 0 r
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
0 1 0 r
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
0 0 1 r
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
1 0 0 r
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 1 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
1 0 0 r
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
0 1 0 r
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
0 0 1 r
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
1 0 0 r
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 1 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
1 0 0 r
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
0 1 0 r
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
0 0 1 r
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
1 0 0 r
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 1 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
1 0 0 r
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
0 0 1 r
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
1 0 0 r
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
0 1 0 r
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
0 0 1 r
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
1 0 0 r
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
0 0 1 r
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
1 0 0 r
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
0 1 0 r
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
0 0 1 r
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
1 0 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
0 0 1 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
1 0 0 r
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
0 1 0 r
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
0 0 1 r
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 1 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
1 0 0 r
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 1 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
0 0 1 r
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
1 0 0 r
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
0 1 0 r
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
0 0 1 r
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 1 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
1 0 0 r
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
0 1 0 r
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
0 0 1 r
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
1 0 0 r
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
0 1 0 r
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
0 0 1 r
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
0 1 0 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
0 0 1 r
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
1 0 0 r
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
0 1 0 r
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
0 0 1 r
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
0 1 0 r
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
0 0 1 r
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
1 0 0 r
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
0 1 0 r
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
0 0 1 r
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 1 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
0 0 1 r
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
1 0 0 r
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
0 1 0 r
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
0 0 1 r
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 1 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
0 0 1 r
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
1 0 0 r
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
0 1 0 r
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
0 0 1 r
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 1 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
1 0 0 r
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 1 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 0 1 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
1 0 0 r
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
0 1 0 r
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
0 0 1 r
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
1 0 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
0 0 1 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
1 0 0 r
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
0 1 0 r
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
0 0 1 r
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
1 0 0 r
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
0 0 1 r
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
1 0 0 r
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
0 1 0 r
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
0 0 1 r
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
1 0 0 r
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 1 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
1 0 0 r
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
0 1 0 r
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
0 0 1 r
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
1 0 0 r
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.511074 0.0177187 0.732558 0.239203 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 1 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
0 1 0 r
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
0 0 1 r
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
1 0 0 r
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
0 1 0 r
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 0 1 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
0 1 0 r
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
0 0 1 r
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
1 0 0 r
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
0 1 0 r
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 0 1 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
0 1 0 r
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
1 0 0 r
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
0 1 0 r
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
0 0 1 r
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
1 0 0 r
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
0 1 0 r
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
1 0 0 r
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
0 1 0 r
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
0 0 1 r
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
1 0 0 r
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 1 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
1 0 0 r
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
0 1 0 r
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
0 0 1 r
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
1 0 0 r
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 1 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
1 0 0 r
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
0 1 0 r
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
0 0 1 r
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
1 0 0 r
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 1 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
1 0 0 r
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
0 0 1 r
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
1 0 0 r
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
0 1 0 r
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
0 0 1 r
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
1 0 0 r
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
0 0 1 r
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
1 0 0 r
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
0 1 0 r
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
0 0 1 r
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
1 0 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
0 0 1 r
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
1 0 0 r
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
0 1 0 r
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
0 0 1 r
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
1 0 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
0 0 1 r
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
1 0 0 r
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
0 1 0 r
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
0 0 1 r
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
1 0 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 0 1 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 1 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
0 0 1 r
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
1 0 0 r
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
0 1 0 r
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
0 0 1 r
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 1 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
1 0 0 r
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
0 1 0 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
0 0 1 r
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
1 0 0 r
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
0 1 0 r
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
1 0 0 r
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
0 1 0 r
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
0 0 1 r
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
1 0 0 r
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
0 1 0 r
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
0 0 1 r
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
1 0 0 r
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
0 1 0 r
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
0 0 1 r
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 1 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
0 0 1 r
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
1 0 0 r
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
0 1 0 r
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
0 0 1 r
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 1 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
1 0 0 r
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
0 1 0 r
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
0 0 1 r
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
1 0 0 r
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
0 1 0 r
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
0 0 1 r
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
1 0 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 1 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
0 0 1 r
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
1 0 0 r
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
0 1 0 r
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
0 0 1 r
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
1 0 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
0 0 1 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
1 0 0 r
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
0 1 0 r
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
0 0 1 r
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
1 0 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
0 0 1 r
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
1 0 0 r
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
0 1 0 r
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
0 0 1 r
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
1 0 0 r
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
0 0 1 r
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
1 0 0 r
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
0 1 0 r
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
0 0 1 r
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
1 0 0 r
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.754707 0.0177187 0.97619 0.239203 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 1 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
0 1 0 r
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
0 0 1 r
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
1 0 0 r
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
0 1 0 r
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 0 1 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
0 1 0 r
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
1 0 0 r
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
0 1 0 r
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
0 0 1 r
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
1 0 0 r
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
0 1 0 r
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
1 0 0 r
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
0 1 0 r
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
0 0 1 r
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
1 0 0 r
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 1 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
1 0 0 r
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
0 1 0 r
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
0 0 1 r
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
1 0 0 r
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 1 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
1 0 0 r
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
0 1 0 r
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
0 0 1 r
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
1 0 0 r
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 1 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
1 0 0 r
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
0 0 1 r
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
1 0 0 r
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
0 1 0 r
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
0 0 1 r
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
1 0 0 r
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
0 0 1 r
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
1 0 0 r
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
0 1 0 r
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
0 0 1 r
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
1 0 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
0 0 1 r
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
1 0 0 r
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
0 1 0 r
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
0 0 1 r
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
1 0 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
0 0 1 r
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
1 0 0 r
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
0 1 0 r
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
0 0 1 r
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
1 0 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
0 0 1 r
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
0 1 0 r
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
0 0 1 r
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
1 0 0 r
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
0 1 0 r
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
0 0 1 r
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
1 0 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
0 0 1 r
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 1 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
0 0 1 r
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
1 0 0 r
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
0 1 0 r
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
1 0 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
0 1 0 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
0 0 1 r
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
1 0 0 r
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
0 1 0 r
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
1 0 0 r
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
0 1 0 r
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
0 0 1 r
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
1 0 0 r
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
0 1 0 r
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
1 0 0 r
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
0 1 0 r
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
0 0 1 r
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
1 0 0 r
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
0 1 0 r
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
1 0 0 r
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
0 1 0 r
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
0 0 1 r
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
1 0 0 r
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
0 1 0 r
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
0 0 1 r
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
1 0 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
0 1 0 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
0 0 1 r
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
1 0 0 r
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
0 1 0 r
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
0 0 1 r
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
1 0 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 0 1 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 1 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
0 0 1 r
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
1 0 0 r
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
0 1 0 r
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
0 0 1 r
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
1 0 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
0 0 1 r
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
1 0 0 r
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
0 1 0 r
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
0 0 1 r
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
1 0 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
0 0 1 r
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
1 0 0 r
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
0 1 0 r
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
0 0 1 r
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
1 0 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
0 0 1 r
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
1 0 0 r
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
0 1 0 r
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
0 0 1 r
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
1 0 0 r
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.261351 0.245293 0.482835 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 1 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
1 0 0 r
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
0 0 1 r
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
1 0 0 r
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
0 1 0 r
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
0 0 1 r
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
1 0 0 r
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
0 0 1 r
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
1 0 0 r
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
0 1 0 r
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
0 0 1 r
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
1 0 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
0 0 1 r
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
1 0 0 r
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
0 1 0 r
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
0 0 1 r
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
1 0 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
0 0 1 r
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
1 0 0 r
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
0 1 0 r
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
0 0 1 r
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
1 0 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
0 0 1 r
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
0 1 0 r
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
0 0 1 r
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
1 0 0 r
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
0 1 0 r
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
0 0 1 r
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
0 1 0 r
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
0 0 1 r
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
1 0 0 r
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
0 1 0 r
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 0 1 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
0 1 0 r
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
0 0 1 r
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
1 0 0 r
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
0 1 0 r
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 0 1 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
0 1 0 r
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
0 0 1 r
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
1 0 0 r
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
0 1 0 r
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 0 1 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
0 1 0 r
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
1 0 0 r
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
0 1 0 r
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
0 0 1 r
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
1 0 0 r
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
0 1 0 r
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 0 1 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
0 1 0 r
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
1 0 0 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
0 1 0 r
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
0 0 1 r
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
1 0 0 r
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 0 1 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
0 1 0 r
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 0 1 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
1 0 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
0 1 0 r
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
0 0 1 r
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
1 0 0 r
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 0 1 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
1 0 0 r
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
0 1 0 r
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
0 0 1 r
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
1 0 0 r
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
0 0 1 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
1 0 0 r
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
0 1 0 r
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
0 0 1 r
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
1 0 0 r
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
0 0 1 r
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
1 0 0 r
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
0 1 0 r
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
0 0 1 r
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
1 0 0 r
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 0 1 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
1 0 0 r
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
0 1 0 r
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
0 0 1 r
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
1 0 0 r
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 0 1 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
1 0 0 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
0 1 0 r
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
0 0 1 r
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
1 0 0 r
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 0 1 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
0 1 0 r
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
1 0 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
0 1 0 r
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
0 0 1 r
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
1 0 0 r
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
0 1 0 r
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 0 1 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
0 1 0 r
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
1 0 0 r
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
0 1 0 r
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
0 0 1 r
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
1 0 0 r
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
0 1 0 r
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 0 1 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
0 1 0 r
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
0 0 1 r
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
1 0 0 r
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
0 1 0 r
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 0 1 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
0 1 0 r
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
0 0 1 r
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
1 0 0 r
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
0 1 0 r
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.267442 0.261351 0.488926 0.482835 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
1 0 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
0 0 1 r
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
1 0 0 r
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
0 1 0 r
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
0 0 1 r
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
1 0 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
0 0 1 r
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
1 0 0 r
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
0 1 0 r
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
0 0 1 r
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
1 0 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
0 0 1 r
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
1 0 0 r
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
0 1 0 r
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
0 0 1 r
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
1 0 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
0 0 1 r
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
0 1 0 r
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
0 0 1 r
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
1 0 0 r
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
0 1 0 r
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
0 0 1 r
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
0 1 0 r
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
0 0 1 r
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
1 0 0 r
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
0 1 0 r
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 0 1 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
0 1 0 r
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
0 0 1 r
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
1 0 0 r
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
0 1 0 r
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 0 1 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
0 1 0 r
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
0 0 1 r
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
1 0 0 r
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
0 1 0 r
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 0 1 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
0 1 0 r
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
1 0 0 r
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
0 1 0 r
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
0 0 1 r
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
1 0 0 r
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
0 1 0 r
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
1 0 0 r
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
0 1 0 r
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
0 0 1 r
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
1 0 0 r
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 1 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
1 0 0 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
0 1 0 r
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
0 0 1 r
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
1 0 0 r
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 0 1 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
0 1 0 r
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 0 1 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
1 0 0 r
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
0 1 0 r
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
0 0 1 r
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
1 0 0 r
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 0 1 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
0 1 0 r
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
0 0 1 r
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
1 0 0 r
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
0 1 0 r
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
0 0 1 r
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
1 0 0 r
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
0 0 1 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
1 0 0 r
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
0 1 0 r
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
0 0 1 r
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
1 0 0 r
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
0 0 1 r
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
1 0 0 r
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
0 1 0 r
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
0 0 1 r
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
1 0 0 r
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 0 1 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
1 0 0 r
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
0 1 0 r
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
0 0 1 r
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
1 0 0 r
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 0 1 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
1 0 0 r
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
0 1 0 r
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
0 0 1 r
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
1 0 0 r
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 0 1 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
0 1 0 r
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 0 1 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
1 0 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
0 1 0 r
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
0 0 1 r
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
1 0 0 r
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 1 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
1 0 0 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
0 1 0 r
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
0 0 1 r
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
1 0 0 r
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
0 1 0 r
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
1 0 0 r
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
0 1 0 r
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
0 0 1 r
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
1 0 0 r
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
0 1 0 r
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 0 1 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
0 1 0 r
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
0 0 1 r
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
1 0 0 r
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
0 1 0 r
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.511074 0.261351 0.732558 0.482835 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
1 0 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
0 0 1 r
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
1 0 0 r
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
0 1 0 r
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
0 0 1 r
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
1 0 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
0 0 1 r
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
1 0 0 r
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
0 1 0 r
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
0 0 1 r
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
1 0 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
0 0 1 r
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
0 1 0 r
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
0 0 1 r
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
1 0 0 r
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
0 1 0 r
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
0 0 1 r
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
0 1 0 r
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
0 0 1 r
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
1 0 0 r
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
0 1 0 r
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 0 1 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
0 1 0 r
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
0 0 1 r
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
1 0 0 r
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
0 1 0 r
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 0 1 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
0 1 0 r
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
0 0 1 r
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
1 0 0 r
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
0 1 0 r
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
0 0 1 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
0 1 0 r
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
1 0 0 r
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
0 1 0 r
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
0 0 1 r
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
1 0 0 r
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
0 1 0 r
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
1 0 0 r
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
0 1 0 r
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
0 0 1 r
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
1 0 0 r
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 1 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
1 0 0 r
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
0 1 0 r
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
0 0 1 r
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
1 0 0 r
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 1 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
1 0 0 r
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
0 1 0 r
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
0 0 1 r
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
1 0 0 r
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 1 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
1 0 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 0 1 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
1 0 0 r
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
0 1 0 r
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
0 0 1 r
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
1 0 0 r
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 0 1 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
0 1 0 r
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
0 0 1 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
1 0 0 r
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
0 1 0 r
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
0 0 1 r
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
0 1 0 r
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
0 0 1 r
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
1 0 0 r
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
0 1 0 r
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
0 0 1 r
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
1 0 0 r
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
0 1 0 r
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
0 0 1 r
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
1 0 0 r
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
0 0 1 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
1 0 0 r
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
0 1 0 r
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
0 0 1 r
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
1 0 0 r
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 0 1 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
0 1 0 r
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
0 0 1 r
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
1 0 0 r
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
0 1 0 r
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
0 0 1 r
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
1 0 0 r
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 1 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 0 1 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
1 0 0 r
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
0 1 0 r
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
0 0 1 r
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
1 0 0 r
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 1 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
1 0 0 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
0 1 0 r
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
0 0 1 r
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
1 0 0 r
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 1 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
1 0 0 r
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
0 1 0 r
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
0 0 1 r
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
1 0 0 r
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
0 1 0 r
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
1 0 0 r
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
0 1 0 r
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
0 0 1 r
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
1 0 0 r
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
0 1 0 r
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.754707 0.261351 0.97619 0.482835 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
1 0 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
0 0 1 r
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
1 0 0 r
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
0 1 0 r
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
0 0 1 r
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
1 0 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
0 0 1 r
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
0 1 0 r
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
0 0 1 r
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
1 0 0 r
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
0 1 0 r
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
0 0 1 r
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
0 1 0 r
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
0 0 1 r
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
1 0 0 r
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
0 1 0 r
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 0 1 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
0 1 0 r
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
0 0 1 r
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
1 0 0 r
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
0 1 0 r
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 0 1 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
0 1 0 r
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
0 0 1 r
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
1 0 0 r
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
0 1 0 r
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
0 0 1 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
0 1 0 r
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
1 0 0 r
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
0 1 0 r
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
0 0 1 r
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
1 0 0 r
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
0 1 0 r
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
1 0 0 r
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
0 1 0 r
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
0 0 1 r
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
1 0 0 r
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 1 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
1 0 0 r
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
0 1 0 r
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
0 0 1 r
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
1 0 0 r
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 1 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
1 0 0 r
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
0 1 0 r
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
0 0 1 r
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
1 0 0 r
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 1 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
1 0 0 r
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
0 0 1 r
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
1 0 0 r
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
0 1 0 r
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
0 0 1 r
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
1 0 0 r
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 1 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
1 0 0 r
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 0 1 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
1 0 0 r
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
0 1 0 r
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
0 0 1 r
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 1 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
0 0 1 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
1 0 0 r
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
0 1 0 r
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
0 0 1 r
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
0 1 0 r
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
0 0 1 r
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
1 0 0 r
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
0 1 0 r
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
0 0 1 r
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
0 1 0 r
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
0 0 1 r
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
1 0 0 r
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
0 1 0 r
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
0 0 1 r
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
0 1 0 r
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
0 0 1 r
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
1 0 0 r
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
0 1 0 r
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
0 0 1 r
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
1 0 0 r
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 1 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
0 0 1 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
1 0 0 r
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
0 1 0 r
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
0 0 1 r
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
1 0 0 r
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 1 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
1 0 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 0 1 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
1 0 0 r
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
0 1 0 r
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
0 0 1 r
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
1 0 0 r
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 1 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
1 0 0 r
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
0 1 0 r
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
0 0 1 r
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
1 0 0 r
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 1 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
1 0 0 r
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
0 1 0 r
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
0 0 1 r
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
1 0 0 r
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 1 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
1 0 0 r
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
0 1 0 r
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
0 0 1 r
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
1 0 0 r
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
0 1 0 r
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.504983 0.245293 0.726467 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 1 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
0 1 0 r
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
1 0 0 r
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
0 1 0 r
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
0 0 1 r
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
1 0 0 r
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
0 1 0 r
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
1 0 0 r
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
0 1 0 r
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
0 0 1 r
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
1 0 0 r
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 1 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
1 0 0 r
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
0 1 0 r
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
0 0 1 r
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
1 0 0 r
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 1 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
1 0 0 r
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
0 1 0 r
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
0 0 1 r
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
1 0 0 r
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
0 1 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
1 0 0 r
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
0 0 1 r
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
1 0 0 r
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
0 1 0 r
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
0 0 1 r
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
1 0 0 r
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
0 0 1 r
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
1 0 0 r
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
0 1 0 r
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
0 0 1 r
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
1 0 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
0 0 1 r
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
1 0 0 r
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
0 1 0 r
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
0 0 1 r
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
1 0 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
0 0 1 r
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
1 0 0 r
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
0 1 0 r
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
0 0 1 r
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
1 0 0 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
0 0 1 r
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
0 1 0 r
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
0 0 1 r
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
1 0 0 r
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
0 1 0 r
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
0 0 1 r
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
1 0 0 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
0 0 1 r
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
0 1 0 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
0 0 1 r
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
1 0 0 r
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
0 1 0 r
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
1 0 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
0 0 1 r
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
1 0 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
0 1 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
0 0 1 r
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
1 0 0 r
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
0 1 0 r
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
1 0 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
0 1 0 r
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
0 0 1 r
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
1 0 0 r
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
0 1 0 r
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
1 0 0 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
0 1 0 r
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
0 0 1 r
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
1 0 0 r
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
0 1 0 r
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
1 0 0 r
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
0 1 0 r
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
0 0 1 r
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
1 0 0 r
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
0 1 0 r
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
1 0 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
0 1 0 r
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
0 0 1 r
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
1 0 0 r
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
0 1 0 r
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
1 0 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
0 1 0 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
0 0 1 r
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
1 0 0 r
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
0 1 0 r
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
1 0 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
0 0 1 r
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 1 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
0 0 1 r
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
1 0 0 r
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
0 1 0 r
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
0 0 1 r
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
1 0 0 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
0 0 1 r
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
0 1 0 r
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
0 0 1 r
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
1 0 0 r
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
0 1 0 r
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
0 0 1 r
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
1 0 0 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
0 0 1 r
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
1 0 0 r
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
0 1 0 r
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
0 0 1 r
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
1 0 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
0 0 1 r
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
1 0 0 r
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
0 1 0 r
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
0 0 1 r
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.267442 0.504983 0.488926 0.726467 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 1 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
1 0 0 r
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
0 1 0 r
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
0 0 1 r
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
1 0 0 r
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 1 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
1 0 0 r
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
0 1 0 r
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
0 0 1 r
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
1 0 0 r
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 1 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
1 0 0 r
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
0 1 0 r
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
0 0 1 r
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
1 0 0 r
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
0 1 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
1 0 0 r
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
0 0 1 r
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
1 0 0 r
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
0 1 0 r
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
0 0 1 r
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
1 0 0 r
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
0 0 1 r
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
1 0 0 r
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
0 1 0 r
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
0 0 1 r
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
1 0 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
0 0 1 r
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
1 0 0 r
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
0 1 0 r
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
0 0 1 r
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
1 0 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
0 0 1 r
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
1 0 0 r
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
0 1 0 r
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
0 0 1 r
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
1 0 0 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
0 0 1 r
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
0 1 0 r
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
0 0 1 r
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
1 0 0 r
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
0 1 0 r
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
0 0 1 r
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
0 1 0 r
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
0 0 1 r
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
1 0 0 r
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
0 1 0 r
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 0 1 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
0 1 0 r
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
0 0 1 r
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
1 0 0 r
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
0 1 0 r
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
1 0 0 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
0 0 1 r
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
1 0 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
0 1 0 r
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
0 0 1 r
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
1 0 0 r
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
0 1 0 r
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
1 0 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
0 0 1 r
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
1 0 0 r
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
0 1 0 r
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
0 0 1 r
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
1 0 0 r
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
0 1 0 r
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
1 0 0 r
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
0 1 0 r
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
0 0 1 r
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
1 0 0 r
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
0 1 0 r
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
1 0 0 r
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
0 1 0 r
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
0 0 1 r
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
1 0 0 r
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
0 1 0 r
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
1 0 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
0 1 0 r
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
0 0 1 r
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
1 0 0 r
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
0 1 0 r
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
1 0 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
0 1 0 r
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
0 0 1 r
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
1 0 0 r
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
0 1 0 r
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
1 0 0 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
0 0 1 r
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
1 0 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
0 1 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
0 0 1 r
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
1 0 0 r
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
0 1 0 r
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 0 1 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
0 1 0 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
0 0 1 r
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
1 0 0 r
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
0 1 0 r
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
0 0 1 r
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
0 1 0 r
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
0 0 1 r
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
1 0 0 r
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
0 1 0 r
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
0 0 1 r
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
1 0 0 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
0 0 1 r
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
1 0 0 r
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
0 1 0 r
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
0 0 1 r
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.511074 0.504983 0.732558 0.726467 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 1 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
1 0 0 r
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
0 1 0 r
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
0 0 1 r
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
1 0 0 r
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 1 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
1 0 0 r
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
0 1 0 r
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
0 0 1 r
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
1 0 0 r
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
0 1 0 r
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
1 0 0 r
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
0 0 1 r
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
1 0 0 r
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
0 1 0 r
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
0 0 1 r
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
1 0 0 r
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
0 0 1 r
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
1 0 0 r
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
0 1 0 r
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
0 0 1 r
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
1 0 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
0 0 1 r
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
1 0 0 r
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
0 1 0 r
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
0 0 1 r
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
1 0 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
0 0 1 r
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
1 0 0 r
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
0 1 0 r
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
0 0 1 r
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
1 0 0 r
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
0 0 1 r
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
0 1 0 r
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
0 0 1 r
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
1 0 0 r
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
0 1 0 r
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
0 0 1 r
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
0 1 0 r
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
0 0 1 r
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
1 0 0 r
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
0 1 0 r
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 0 1 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
0 1 0 r
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
0 0 1 r
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
1 0 0 r
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
0 1 0 r
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 0 1 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
0 1 0 r
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
0 0 1 r
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
1 0 0 r
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
0 1 0 r
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 0 1 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
0 1 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
1 0 0 r
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
0 1 0 r
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
0 0 1 r
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
1 0 0 r
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
0 1 0 r
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
1 0 0 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
0 0 1 r
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
1 0 0 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
0 1 0 r
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
0 0 1 r
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
1 0 0 r
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
0 0 1 r
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
1 0 0 r
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
0 1 0 r
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
0 0 1 r
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
1 0 0 r
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
0 1 0 r
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
0 0 1 r
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
1 0 0 r
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
0 1 0 r
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
1 0 0 r
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
0 1 0 r
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
0 0 1 r
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
1 0 0 r
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
0 1 0 r
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
1 0 0 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
0 0 1 r
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
1 0 0 r
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
0 1 0 r
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
0 0 1 r
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
1 0 0 r
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
0 1 0 r
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 0 1 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
1 0 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
0 1 0 r
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
0 0 1 r
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
1 0 0 r
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
0 1 0 r
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 0 1 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
0 1 0 r
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
0 0 1 r
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
1 0 0 r
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
0 1 0 r
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 0 1 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
0 1 0 r
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
0 0 1 r
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
1 0 0 r
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
0 1 0 r
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
0 0 1 r
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
0 1 0 r
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
0 0 1 r
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
1 0 0 r
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
0 1 0 r
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
0 0 1 r
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.754707 0.504983 0.97619 0.726467 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 1 0 r
.01364 .01364 m
.01364 .07727 L
.07727 .07727 L
.07727 .01364 L
F
.06148 .01364 m
.06148 .07727 L
.12512 .07727 L
.12512 .01364 L
F
1 0 0 r
.10933 .01364 m
.10933 .07727 L
.17297 .07727 L
.17297 .01364 L
F
.15718 .01364 m
.15718 .07727 L
.22081 .07727 L
.22081 .01364 L
F
.20502 .01364 m
.20502 .07727 L
.26866 .07727 L
.26866 .01364 L
F
.25287 .01364 m
.25287 .07727 L
.31651 .07727 L
.31651 .01364 L
F
.30072 .01364 m
.30072 .07727 L
.36435 .07727 L
.36435 .01364 L
F
.34856 .01364 m
.34856 .07727 L
.4122 .07727 L
.4122 .01364 L
F
.39641 .01364 m
.39641 .07727 L
.46005 .07727 L
.46005 .01364 L
F
.44426 .01364 m
.44426 .07727 L
.50789 .07727 L
.50789 .01364 L
F
0 1 0 r
.49211 .01364 m
.49211 .07727 L
.55574 .07727 L
.55574 .01364 L
F
.53995 .01364 m
.53995 .07727 L
.60359 .07727 L
.60359 .01364 L
F
.5878 .01364 m
.5878 .07727 L
.65144 .07727 L
.65144 .01364 L
F
.63565 .01364 m
.63565 .07727 L
.69928 .07727 L
.69928 .01364 L
F
0 0 1 r
.68349 .01364 m
.68349 .07727 L
.74713 .07727 L
.74713 .01364 L
F
.73134 .01364 m
.73134 .07727 L
.79498 .07727 L
.79498 .01364 L
F
.77919 .01364 m
.77919 .07727 L
.84282 .07727 L
.84282 .01364 L
F
.82703 .01364 m
.82703 .07727 L
.89067 .07727 L
.89067 .01364 L
F
1 0 0 r
.87488 .01364 m
.87488 .07727 L
.93852 .07727 L
.93852 .01364 L
F
.92273 .01364 m
.92273 .07727 L
.98636 .07727 L
.98636 .01364 L
F
0 1 0 r
.01364 .06148 m
.01364 .12512 L
.07727 .12512 L
.07727 .06148 L
F
1 0 0 r
.06148 .06148 m
.06148 .12512 L
.12512 .12512 L
.12512 .06148 L
F
.10933 .06148 m
.10933 .12512 L
.17297 .12512 L
.17297 .06148 L
F
.15718 .06148 m
.15718 .12512 L
.22081 .12512 L
.22081 .06148 L
F
.20502 .06148 m
.20502 .12512 L
.26866 .12512 L
.26866 .06148 L
F
0 0 1 r
.25287 .06148 m
.25287 .12512 L
.31651 .12512 L
.31651 .06148 L
F
.30072 .06148 m
.30072 .12512 L
.36435 .12512 L
.36435 .06148 L
F
1 0 0 r
.34856 .06148 m
.34856 .12512 L
.4122 .12512 L
.4122 .06148 L
F
.39641 .06148 m
.39641 .12512 L
.46005 .12512 L
.46005 .06148 L
F
.44426 .06148 m
.44426 .12512 L
.50789 .12512 L
.50789 .06148 L
F
.49211 .06148 m
.49211 .12512 L
.55574 .12512 L
.55574 .06148 L
F
0 1 0 r
.53995 .06148 m
.53995 .12512 L
.60359 .12512 L
.60359 .06148 L
F
.5878 .06148 m
.5878 .12512 L
.65144 .12512 L
.65144 .06148 L
F
.63565 .06148 m
.63565 .12512 L
.69928 .12512 L
.69928 .06148 L
F
.68349 .06148 m
.68349 .12512 L
.74713 .12512 L
.74713 .06148 L
F
0 0 1 r
.73134 .06148 m
.73134 .12512 L
.79498 .12512 L
.79498 .06148 L
F
.77919 .06148 m
.77919 .12512 L
.84282 .12512 L
.84282 .06148 L
F
.82703 .06148 m
.82703 .12512 L
.89067 .12512 L
.89067 .06148 L
F
.87488 .06148 m
.87488 .12512 L
.93852 .12512 L
.93852 .06148 L
F
1 0 0 r
.92273 .06148 m
.92273 .12512 L
.98636 .12512 L
.98636 .06148 L
F
.01364 .10933 m
.01364 .17297 L
.07727 .17297 L
.07727 .10933 L
F
.06148 .10933 m
.06148 .17297 L
.12512 .17297 L
.12512 .10933 L
F
.10933 .10933 m
.10933 .17297 L
.17297 .17297 L
.17297 .10933 L
F
.15718 .10933 m
.15718 .17297 L
.22081 .17297 L
.22081 .10933 L
F
0 0 1 r
.20502 .10933 m
.20502 .17297 L
.26866 .17297 L
.26866 .10933 L
F
.25287 .10933 m
.25287 .17297 L
.31651 .17297 L
.31651 .10933 L
F
.30072 .10933 m
.30072 .17297 L
.36435 .17297 L
.36435 .10933 L
F
.34856 .10933 m
.34856 .17297 L
.4122 .17297 L
.4122 .10933 L
F
1 0 0 r
.39641 .10933 m
.39641 .17297 L
.46005 .17297 L
.46005 .10933 L
F
.44426 .10933 m
.44426 .17297 L
.50789 .17297 L
.50789 .10933 L
F
.49211 .10933 m
.49211 .17297 L
.55574 .17297 L
.55574 .10933 L
F
.53995 .10933 m
.53995 .17297 L
.60359 .17297 L
.60359 .10933 L
F
0 1 0 r
.5878 .10933 m
.5878 .17297 L
.65144 .17297 L
.65144 .10933 L
F
.63565 .10933 m
.63565 .17297 L
.69928 .17297 L
.69928 .10933 L
F
.68349 .10933 m
.68349 .17297 L
.74713 .17297 L
.74713 .10933 L
F
.73134 .10933 m
.73134 .17297 L
.79498 .17297 L
.79498 .10933 L
F
0 0 1 r
.77919 .10933 m
.77919 .17297 L
.84282 .17297 L
.84282 .10933 L
F
.82703 .10933 m
.82703 .17297 L
.89067 .17297 L
.89067 .10933 L
F
.87488 .10933 m
.87488 .17297 L
.93852 .17297 L
.93852 .10933 L
F
.92273 .10933 m
.92273 .17297 L
.98636 .17297 L
.98636 .10933 L
F
1 0 0 r
.01364 .15718 m
.01364 .22081 L
.07727 .22081 L
.07727 .15718 L
F
.06148 .15718 m
.06148 .22081 L
.12512 .22081 L
.12512 .15718 L
F
.10933 .15718 m
.10933 .22081 L
.17297 .22081 L
.17297 .15718 L
F
0 0 1 r
.15718 .15718 m
.15718 .22081 L
.22081 .22081 L
.22081 .15718 L
F
.20502 .15718 m
.20502 .22081 L
.26866 .22081 L
.26866 .15718 L
F
.25287 .15718 m
.25287 .22081 L
.31651 .22081 L
.31651 .15718 L
F
.30072 .15718 m
.30072 .22081 L
.36435 .22081 L
.36435 .15718 L
F
.34856 .15718 m
.34856 .22081 L
.4122 .22081 L
.4122 .15718 L
F
.39641 .15718 m
.39641 .22081 L
.46005 .22081 L
.46005 .15718 L
F
1 0 0 r
.44426 .15718 m
.44426 .22081 L
.50789 .22081 L
.50789 .15718 L
F
.49211 .15718 m
.49211 .22081 L
.55574 .22081 L
.55574 .15718 L
F
.53995 .15718 m
.53995 .22081 L
.60359 .22081 L
.60359 .15718 L
F
.5878 .15718 m
.5878 .22081 L
.65144 .22081 L
.65144 .15718 L
F
0 1 0 r
.63565 .15718 m
.63565 .22081 L
.69928 .22081 L
.69928 .15718 L
F
.68349 .15718 m
.68349 .22081 L
.74713 .22081 L
.74713 .15718 L
F
.73134 .15718 m
.73134 .22081 L
.79498 .22081 L
.79498 .15718 L
F
.77919 .15718 m
.77919 .22081 L
.84282 .22081 L
.84282 .15718 L
F
0 0 1 r
.82703 .15718 m
.82703 .22081 L
.89067 .22081 L
.89067 .15718 L
F
.87488 .15718 m
.87488 .22081 L
.93852 .22081 L
.93852 .15718 L
F
.92273 .15718 m
.92273 .22081 L
.98636 .22081 L
.98636 .15718 L
F
1 0 0 r
.01364 .20502 m
.01364 .26866 L
.07727 .26866 L
.07727 .20502 L
F
.06148 .20502 m
.06148 .26866 L
.12512 .26866 L
.12512 .20502 L
F
0 0 1 r
.10933 .20502 m
.10933 .26866 L
.17297 .26866 L
.17297 .20502 L
F
.15718 .20502 m
.15718 .26866 L
.22081 .26866 L
.22081 .20502 L
F
.20502 .20502 m
.20502 .26866 L
.26866 .26866 L
.26866 .20502 L
F
.25287 .20502 m
.25287 .26866 L
.31651 .26866 L
.31651 .20502 L
F
.30072 .20502 m
.30072 .26866 L
.36435 .26866 L
.36435 .20502 L
F
.34856 .20502 m
.34856 .26866 L
.4122 .26866 L
.4122 .20502 L
F
.39641 .20502 m
.39641 .26866 L
.46005 .26866 L
.46005 .20502 L
F
.44426 .20502 m
.44426 .26866 L
.50789 .26866 L
.50789 .20502 L
F
1 0 0 r
.49211 .20502 m
.49211 .26866 L
.55574 .26866 L
.55574 .20502 L
F
.53995 .20502 m
.53995 .26866 L
.60359 .26866 L
.60359 .20502 L
F
.5878 .20502 m
.5878 .26866 L
.65144 .26866 L
.65144 .20502 L
F
.63565 .20502 m
.63565 .26866 L
.69928 .26866 L
.69928 .20502 L
F
0 1 0 r
.68349 .20502 m
.68349 .26866 L
.74713 .26866 L
.74713 .20502 L
F
.73134 .20502 m
.73134 .26866 L
.79498 .26866 L
.79498 .20502 L
F
.77919 .20502 m
.77919 .26866 L
.84282 .26866 L
.84282 .20502 L
F
.82703 .20502 m
.82703 .26866 L
.89067 .26866 L
.89067 .20502 L
F
0 0 1 r
.87488 .20502 m
.87488 .26866 L
.93852 .26866 L
.93852 .20502 L
F
.92273 .20502 m
.92273 .26866 L
.98636 .26866 L
.98636 .20502 L
F
1 0 0 r
.01364 .25287 m
.01364 .31651 L
.07727 .31651 L
.07727 .25287 L
F
0 0 1 r
.06148 .25287 m
.06148 .31651 L
.12512 .31651 L
.12512 .25287 L
F
.10933 .25287 m
.10933 .31651 L
.17297 .31651 L
.17297 .25287 L
F
.15718 .25287 m
.15718 .31651 L
.22081 .31651 L
.22081 .25287 L
F
.20502 .25287 m
.20502 .31651 L
.26866 .31651 L
.26866 .25287 L
F
0 1 0 r
.25287 .25287 m
.25287 .31651 L
.31651 .31651 L
.31651 .25287 L
F
.30072 .25287 m
.30072 .31651 L
.36435 .31651 L
.36435 .25287 L
F
0 0 1 r
.34856 .25287 m
.34856 .31651 L
.4122 .31651 L
.4122 .25287 L
F
.39641 .25287 m
.39641 .31651 L
.46005 .31651 L
.46005 .25287 L
F
.44426 .25287 m
.44426 .31651 L
.50789 .31651 L
.50789 .25287 L
F
.49211 .25287 m
.49211 .31651 L
.55574 .31651 L
.55574 .25287 L
F
1 0 0 r
.53995 .25287 m
.53995 .31651 L
.60359 .31651 L
.60359 .25287 L
F
.5878 .25287 m
.5878 .31651 L
.65144 .31651 L
.65144 .25287 L
F
.63565 .25287 m
.63565 .31651 L
.69928 .31651 L
.69928 .25287 L
F
.68349 .25287 m
.68349 .31651 L
.74713 .31651 L
.74713 .25287 L
F
0 1 0 r
.73134 .25287 m
.73134 .31651 L
.79498 .31651 L
.79498 .25287 L
F
.77919 .25287 m
.77919 .31651 L
.84282 .31651 L
.84282 .25287 L
F
.82703 .25287 m
.82703 .31651 L
.89067 .31651 L
.89067 .25287 L
F
.87488 .25287 m
.87488 .31651 L
.93852 .31651 L
.93852 .25287 L
F
0 0 1 r
.92273 .25287 m
.92273 .31651 L
.98636 .31651 L
.98636 .25287 L
F
.01364 .30072 m
.01364 .36435 L
.07727 .36435 L
.07727 .30072 L
F
.06148 .30072 m
.06148 .36435 L
.12512 .36435 L
.12512 .30072 L
F
.10933 .30072 m
.10933 .36435 L
.17297 .36435 L
.17297 .30072 L
F
.15718 .30072 m
.15718 .36435 L
.22081 .36435 L
.22081 .30072 L
F
0 1 0 r
.20502 .30072 m
.20502 .36435 L
.26866 .36435 L
.26866 .30072 L
F
.25287 .30072 m
.25287 .36435 L
.31651 .36435 L
.31651 .30072 L
F
.30072 .30072 m
.30072 .36435 L
.36435 .36435 L
.36435 .30072 L
F
.34856 .30072 m
.34856 .36435 L
.4122 .36435 L
.4122 .30072 L
F
0 0 1 r
.39641 .30072 m
.39641 .36435 L
.46005 .36435 L
.46005 .30072 L
F
.44426 .30072 m
.44426 .36435 L
.50789 .36435 L
.50789 .30072 L
F
.49211 .30072 m
.49211 .36435 L
.55574 .36435 L
.55574 .30072 L
F
.53995 .30072 m
.53995 .36435 L
.60359 .36435 L
.60359 .30072 L
F
1 0 0 r
.5878 .30072 m
.5878 .36435 L
.65144 .36435 L
.65144 .30072 L
F
.63565 .30072 m
.63565 .36435 L
.69928 .36435 L
.69928 .30072 L
F
.68349 .30072 m
.68349 .36435 L
.74713 .36435 L
.74713 .30072 L
F
.73134 .30072 m
.73134 .36435 L
.79498 .36435 L
.79498 .30072 L
F
0 1 0 r
.77919 .30072 m
.77919 .36435 L
.84282 .36435 L
.84282 .30072 L
F
.82703 .30072 m
.82703 .36435 L
.89067 .36435 L
.89067 .30072 L
F
.87488 .30072 m
.87488 .36435 L
.93852 .36435 L
.93852 .30072 L
F
.92273 .30072 m
.92273 .36435 L
.98636 .36435 L
.98636 .30072 L
F
0 0 1 r
.01364 .34856 m
.01364 .4122 L
.07727 .4122 L
.07727 .34856 L
F
.06148 .34856 m
.06148 .4122 L
.12512 .4122 L
.12512 .34856 L
F
.10933 .34856 m
.10933 .4122 L
.17297 .4122 L
.17297 .34856 L
F
0 1 0 r
.15718 .34856 m
.15718 .4122 L
.22081 .4122 L
.22081 .34856 L
F
.20502 .34856 m
.20502 .4122 L
.26866 .4122 L
.26866 .34856 L
F
.25287 .34856 m
.25287 .4122 L
.31651 .4122 L
.31651 .34856 L
F
.30072 .34856 m
.30072 .4122 L
.36435 .4122 L
.36435 .34856 L
F
.34856 .34856 m
.34856 .4122 L
.4122 .4122 L
.4122 .34856 L
F
.39641 .34856 m
.39641 .4122 L
.46005 .4122 L
.46005 .34856 L
F
0 0 1 r
.44426 .34856 m
.44426 .4122 L
.50789 .4122 L
.50789 .34856 L
F
.49211 .34856 m
.49211 .4122 L
.55574 .4122 L
.55574 .34856 L
F
.53995 .34856 m
.53995 .4122 L
.60359 .4122 L
.60359 .34856 L
F
.5878 .34856 m
.5878 .4122 L
.65144 .4122 L
.65144 .34856 L
F
1 0 0 r
.63565 .34856 m
.63565 .4122 L
.69928 .4122 L
.69928 .34856 L
F
.68349 .34856 m
.68349 .4122 L
.74713 .4122 L
.74713 .34856 L
F
.73134 .34856 m
.73134 .4122 L
.79498 .4122 L
.79498 .34856 L
F
.77919 .34856 m
.77919 .4122 L
.84282 .4122 L
.84282 .34856 L
F
0 1 0 r
.82703 .34856 m
.82703 .4122 L
.89067 .4122 L
.89067 .34856 L
F
.87488 .34856 m
.87488 .4122 L
.93852 .4122 L
.93852 .34856 L
F
.92273 .34856 m
.92273 .4122 L
.98636 .4122 L
.98636 .34856 L
F
0 0 1 r
.01364 .39641 m
.01364 .46005 L
.07727 .46005 L
.07727 .39641 L
F
.06148 .39641 m
.06148 .46005 L
.12512 .46005 L
.12512 .39641 L
F
0 1 0 r
.10933 .39641 m
.10933 .46005 L
.17297 .46005 L
.17297 .39641 L
F
.15718 .39641 m
.15718 .46005 L
.22081 .46005 L
.22081 .39641 L
F
.20502 .39641 m
.20502 .46005 L
.26866 .46005 L
.26866 .39641 L
F
.25287 .39641 m
.25287 .46005 L
.31651 .46005 L
.31651 .39641 L
F
.30072 .39641 m
.30072 .46005 L
.36435 .46005 L
.36435 .39641 L
F
.34856 .39641 m
.34856 .46005 L
.4122 .46005 L
.4122 .39641 L
F
.39641 .39641 m
.39641 .46005 L
.46005 .46005 L
.46005 .39641 L
F
.44426 .39641 m
.44426 .46005 L
.50789 .46005 L
.50789 .39641 L
F
0 0 1 r
.49211 .39641 m
.49211 .46005 L
.55574 .46005 L
.55574 .39641 L
F
.53995 .39641 m
.53995 .46005 L
.60359 .46005 L
.60359 .39641 L
F
.5878 .39641 m
.5878 .46005 L
.65144 .46005 L
.65144 .39641 L
F
.63565 .39641 m
.63565 .46005 L
.69928 .46005 L
.69928 .39641 L
F
1 0 0 r
.68349 .39641 m
.68349 .46005 L
.74713 .46005 L
.74713 .39641 L
F
.73134 .39641 m
.73134 .46005 L
.79498 .46005 L
.79498 .39641 L
F
.77919 .39641 m
.77919 .46005 L
.84282 .46005 L
.84282 .39641 L
F
.82703 .39641 m
.82703 .46005 L
.89067 .46005 L
.89067 .39641 L
F
0 1 0 r
.87488 .39641 m
.87488 .46005 L
.93852 .46005 L
.93852 .39641 L
F
.92273 .39641 m
.92273 .46005 L
.98636 .46005 L
.98636 .39641 L
F
0 0 1 r
.01364 .44426 m
.01364 .50789 L
.07727 .50789 L
.07727 .44426 L
F
.06148 .44426 m
.06148 .50789 L
.12512 .50789 L
.12512 .44426 L
F
0 1 0 r
.10933 .44426 m
.10933 .50789 L
.17297 .50789 L
.17297 .44426 L
F
.15718 .44426 m
.15718 .50789 L
.22081 .50789 L
.22081 .44426 L
F
.20502 .44426 m
.20502 .50789 L
.26866 .50789 L
.26866 .44426 L
F
1 0 0 r
.25287 .44426 m
.25287 .50789 L
.31651 .50789 L
.31651 .44426 L
F
.30072 .44426 m
.30072 .50789 L
.36435 .50789 L
.36435 .44426 L
F
0 1 0 r
.34856 .44426 m
.34856 .50789 L
.4122 .50789 L
.4122 .44426 L
F
.39641 .44426 m
.39641 .50789 L
.46005 .50789 L
.46005 .44426 L
F
.44426 .44426 m
.44426 .50789 L
.50789 .50789 L
.50789 .44426 L
F
.49211 .44426 m
.49211 .50789 L
.55574 .50789 L
.55574 .44426 L
F
0 0 1 r
.53995 .44426 m
.53995 .50789 L
.60359 .50789 L
.60359 .44426 L
F
.5878 .44426 m
.5878 .50789 L
.65144 .50789 L
.65144 .44426 L
F
.63565 .44426 m
.63565 .50789 L
.69928 .50789 L
.69928 .44426 L
F
.68349 .44426 m
.68349 .50789 L
.74713 .50789 L
.74713 .44426 L
F
1 0 0 r
.73134 .44426 m
.73134 .50789 L
.79498 .50789 L
.79498 .44426 L
F
.77919 .44426 m
.77919 .50789 L
.84282 .50789 L
.84282 .44426 L
F
.82703 .44426 m
.82703 .50789 L
.89067 .50789 L
.89067 .44426 L
F
.87488 .44426 m
.87488 .50789 L
.93852 .50789 L
.93852 .44426 L
F
0 1 0 r
.92273 .44426 m
.92273 .50789 L
.98636 .50789 L
.98636 .44426 L
F
0 0 1 r
.01364 .49211 m
.01364 .55574 L
.07727 .55574 L
.07727 .49211 L
F
.06148 .49211 m
.06148 .55574 L
.12512 .55574 L
.12512 .49211 L
F
.10933 .49211 m
.10933 .55574 L
.17297 .55574 L
.17297 .49211 L
F
0 1 0 r
.15718 .49211 m
.15718 .55574 L
.22081 .55574 L
.22081 .49211 L
F
1 0 0 r
.20502 .49211 m
.20502 .55574 L
.26866 .55574 L
.26866 .49211 L
F
.25287 .49211 m
.25287 .55574 L
.31651 .55574 L
.31651 .49211 L
F
.30072 .49211 m
.30072 .55574 L
.36435 .55574 L
.36435 .49211 L
F
.34856 .49211 m
.34856 .55574 L
.4122 .55574 L
.4122 .49211 L
F
0 1 0 r
.39641 .49211 m
.39641 .55574 L
.46005 .55574 L
.46005 .49211 L
F
.44426 .49211 m
.44426 .55574 L
.50789 .55574 L
.50789 .49211 L
F
.49211 .49211 m
.49211 .55574 L
.55574 .55574 L
.55574 .49211 L
F
.53995 .49211 m
.53995 .55574 L
.60359 .55574 L
.60359 .49211 L
F
0 0 1 r
.5878 .49211 m
.5878 .55574 L
.65144 .55574 L
.65144 .49211 L
F
.63565 .49211 m
.63565 .55574 L
.69928 .55574 L
.69928 .49211 L
F
.68349 .49211 m
.68349 .55574 L
.74713 .55574 L
.74713 .49211 L
F
.73134 .49211 m
.73134 .55574 L
.79498 .55574 L
.79498 .49211 L
F
1 0 0 r
.77919 .49211 m
.77919 .55574 L
.84282 .55574 L
.84282 .49211 L
F
.82703 .49211 m
.82703 .55574 L
.89067 .55574 L
.89067 .49211 L
F
.87488 .49211 m
.87488 .55574 L
.93852 .55574 L
.93852 .49211 L
F
.92273 .49211 m
.92273 .55574 L
.98636 .55574 L
.98636 .49211 L
F
0 0 1 r
.01364 .53995 m
.01364 .60359 L
.07727 .60359 L
.07727 .53995 L
F
.06148 .53995 m
.06148 .60359 L
.12512 .60359 L
.12512 .53995 L
F
.10933 .53995 m
.10933 .60359 L
.17297 .60359 L
.17297 .53995 L
F
.15718 .53995 m
.15718 .60359 L
.22081 .60359 L
.22081 .53995 L
F
.20502 .53995 m
.20502 .60359 L
.26866 .60359 L
.26866 .53995 L
F
1 0 0 r
.25287 .53995 m
.25287 .60359 L
.31651 .60359 L
.31651 .53995 L
F
.30072 .53995 m
.30072 .60359 L
.36435 .60359 L
.36435 .53995 L
F
.34856 .53995 m
.34856 .60359 L
.4122 .60359 L
.4122 .53995 L
F
.39641 .53995 m
.39641 .60359 L
.46005 .60359 L
.46005 .53995 L
F
0 1 0 r
.44426 .53995 m
.44426 .60359 L
.50789 .60359 L
.50789 .53995 L
F
.49211 .53995 m
.49211 .60359 L
.55574 .60359 L
.55574 .53995 L
F
.53995 .53995 m
.53995 .60359 L
.60359 .60359 L
.60359 .53995 L
F
.5878 .53995 m
.5878 .60359 L
.65144 .60359 L
.65144 .53995 L
F
0 0 1 r
.63565 .53995 m
.63565 .60359 L
.69928 .60359 L
.69928 .53995 L
F
.68349 .53995 m
.68349 .60359 L
.74713 .60359 L
.74713 .53995 L
F
.73134 .53995 m
.73134 .60359 L
.79498 .60359 L
.79498 .53995 L
F
.77919 .53995 m
.77919 .60359 L
.84282 .60359 L
.84282 .53995 L
F
1 0 0 r
.82703 .53995 m
.82703 .60359 L
.89067 .60359 L
.89067 .53995 L
F
.87488 .53995 m
.87488 .60359 L
.93852 .60359 L
.93852 .53995 L
F
.92273 .53995 m
.92273 .60359 L
.98636 .60359 L
.98636 .53995 L
F
.01364 .5878 m
.01364 .65144 L
.07727 .65144 L
.07727 .5878 L
F
0 0 1 r
.06148 .5878 m
.06148 .65144 L
.12512 .65144 L
.12512 .5878 L
F
.10933 .5878 m
.10933 .65144 L
.17297 .65144 L
.17297 .5878 L
F
.15718 .5878 m
.15718 .65144 L
.22081 .65144 L
.22081 .5878 L
F
.20502 .5878 m
.20502 .65144 L
.26866 .65144 L
.26866 .5878 L
F
1 0 0 r
.25287 .5878 m
.25287 .65144 L
.31651 .65144 L
.31651 .5878 L
F
.30072 .5878 m
.30072 .65144 L
.36435 .65144 L
.36435 .5878 L
F
.34856 .5878 m
.34856 .65144 L
.4122 .65144 L
.4122 .5878 L
F
.39641 .5878 m
.39641 .65144 L
.46005 .65144 L
.46005 .5878 L
F
0 1 0 r
.44426 .5878 m
.44426 .65144 L
.50789 .65144 L
.50789 .5878 L
F
.49211 .5878 m
.49211 .65144 L
.55574 .65144 L
.55574 .5878 L
F
.53995 .5878 m
.53995 .65144 L
.60359 .65144 L
.60359 .5878 L
F
.5878 .5878 m
.5878 .65144 L
.65144 .65144 L
.65144 .5878 L
F
0 0 1 r
.63565 .5878 m
.63565 .65144 L
.69928 .65144 L
.69928 .5878 L
F
.68349 .5878 m
.68349 .65144 L
.74713 .65144 L
.74713 .5878 L
F
.73134 .5878 m
.73134 .65144 L
.79498 .65144 L
.79498 .5878 L
F
.77919 .5878 m
.77919 .65144 L
.84282 .65144 L
.84282 .5878 L
F
1 0 0 r
.82703 .5878 m
.82703 .65144 L
.89067 .65144 L
.89067 .5878 L
F
.87488 .5878 m
.87488 .65144 L
.93852 .65144 L
.93852 .5878 L
F
.92273 .5878 m
.92273 .65144 L
.98636 .65144 L
.98636 .5878 L
F
.01364 .63565 m
.01364 .69928 L
.07727 .69928 L
.07727 .63565 L
F
.06148 .63565 m
.06148 .69928 L
.12512 .69928 L
.12512 .63565 L
F
0 0 1 r
.10933 .63565 m
.10933 .69928 L
.17297 .69928 L
.17297 .63565 L
F
.15718 .63565 m
.15718 .69928 L
.22081 .69928 L
.22081 .63565 L
F
1 0 0 r
.20502 .63565 m
.20502 .69928 L
.26866 .69928 L
.26866 .63565 L
F
.25287 .63565 m
.25287 .69928 L
.31651 .69928 L
.31651 .63565 L
F
.30072 .63565 m
.30072 .69928 L
.36435 .69928 L
.36435 .63565 L
F
.34856 .63565 m
.34856 .69928 L
.4122 .69928 L
.4122 .63565 L
F
0 1 0 r
.39641 .63565 m
.39641 .69928 L
.46005 .69928 L
.46005 .63565 L
F
.44426 .63565 m
.44426 .69928 L
.50789 .69928 L
.50789 .63565 L
F
.49211 .63565 m
.49211 .69928 L
.55574 .69928 L
.55574 .63565 L
F
.53995 .63565 m
.53995 .69928 L
.60359 .69928 L
.60359 .63565 L
F
0 0 1 r
.5878 .63565 m
.5878 .69928 L
.65144 .69928 L
.65144 .63565 L
F
.63565 .63565 m
.63565 .69928 L
.69928 .69928 L
.69928 .63565 L
F
.68349 .63565 m
.68349 .69928 L
.74713 .69928 L
.74713 .63565 L
F
.73134 .63565 m
.73134 .69928 L
.79498 .69928 L
.79498 .63565 L
F
1 0 0 r
.77919 .63565 m
.77919 .69928 L
.84282 .69928 L
.84282 .63565 L
F
.82703 .63565 m
.82703 .69928 L
.89067 .69928 L
.89067 .63565 L
F
.87488 .63565 m
.87488 .69928 L
.93852 .69928 L
.93852 .63565 L
F
.92273 .63565 m
.92273 .69928 L
.98636 .69928 L
.98636 .63565 L
F
.01364 .68349 m
.01364 .74713 L
.07727 .74713 L
.07727 .68349 L
F
0 0 1 r
.06148 .68349 m
.06148 .74713 L
.12512 .74713 L
.12512 .68349 L
F
.10933 .68349 m
.10933 .74713 L
.17297 .74713 L
.17297 .68349 L
F
1 0 0 r
.15718 .68349 m
.15718 .74713 L
.22081 .74713 L
.22081 .68349 L
F
.20502 .68349 m
.20502 .74713 L
.26866 .74713 L
.26866 .68349 L
F
.25287 .68349 m
.25287 .74713 L
.31651 .74713 L
.31651 .68349 L
F
.30072 .68349 m
.30072 .74713 L
.36435 .74713 L
.36435 .68349 L
F
0 1 0 r
.34856 .68349 m
.34856 .74713 L
.4122 .74713 L
.4122 .68349 L
F
.39641 .68349 m
.39641 .74713 L
.46005 .74713 L
.46005 .68349 L
F
.44426 .68349 m
.44426 .74713 L
.50789 .74713 L
.50789 .68349 L
F
.49211 .68349 m
.49211 .74713 L
.55574 .74713 L
.55574 .68349 L
F
0 0 1 r
.53995 .68349 m
.53995 .74713 L
.60359 .74713 L
.60359 .68349 L
F
.5878 .68349 m
.5878 .74713 L
.65144 .74713 L
.65144 .68349 L
F
.63565 .68349 m
.63565 .74713 L
.69928 .74713 L
.69928 .68349 L
F
.68349 .68349 m
.68349 .74713 L
.74713 .74713 L
.74713 .68349 L
F
1 0 0 r
.73134 .68349 m
.73134 .74713 L
.79498 .74713 L
.79498 .68349 L
F
.77919 .68349 m
.77919 .74713 L
.84282 .74713 L
.84282 .68349 L
F
.82703 .68349 m
.82703 .74713 L
.89067 .74713 L
.89067 .68349 L
F
.87488 .68349 m
.87488 .74713 L
.93852 .74713 L
.93852 .68349 L
F
0 1 0 r
.92273 .68349 m
.92273 .74713 L
.98636 .74713 L
.98636 .68349 L
F
0 0 1 r
.01364 .73134 m
.01364 .79498 L
.07727 .79498 L
.07727 .73134 L
F
.06148 .73134 m
.06148 .79498 L
.12512 .79498 L
.12512 .73134 L
F
.10933 .73134 m
.10933 .79498 L
.17297 .79498 L
.17297 .73134 L
F
.15718 .73134 m
.15718 .79498 L
.22081 .79498 L
.22081 .73134 L
F
1 0 0 r
.20502 .73134 m
.20502 .79498 L
.26866 .79498 L
.26866 .73134 L
F
.25287 .73134 m
.25287 .79498 L
.31651 .79498 L
.31651 .73134 L
F
.30072 .73134 m
.30072 .79498 L
.36435 .79498 L
.36435 .73134 L
F
0 1 0 r
.34856 .73134 m
.34856 .79498 L
.4122 .79498 L
.4122 .73134 L
F
.39641 .73134 m
.39641 .79498 L
.46005 .79498 L
.46005 .73134 L
F
.44426 .73134 m
.44426 .79498 L
.50789 .79498 L
.50789 .73134 L
F
.49211 .73134 m
.49211 .79498 L
.55574 .79498 L
.55574 .73134 L
F
0 0 1 r
.53995 .73134 m
.53995 .79498 L
.60359 .79498 L
.60359 .73134 L
F
.5878 .73134 m
.5878 .79498 L
.65144 .79498 L
.65144 .73134 L
F
.63565 .73134 m
.63565 .79498 L
.69928 .79498 L
.69928 .73134 L
F
.68349 .73134 m
.68349 .79498 L
.74713 .79498 L
.74713 .73134 L
F
1 0 0 r
.73134 .73134 m
.73134 .79498 L
.79498 .79498 L
.79498 .73134 L
F
.77919 .73134 m
.77919 .79498 L
.84282 .79498 L
.84282 .73134 L
F
.82703 .73134 m
.82703 .79498 L
.89067 .79498 L
.89067 .73134 L
F
.87488 .73134 m
.87488 .79498 L
.93852 .79498 L
.93852 .73134 L
F
0 1 0 r
.92273 .73134 m
.92273 .79498 L
.98636 .79498 L
.98636 .73134 L
F
0 0 1 r
.01364 .77919 m
.01364 .84282 L
.07727 .84282 L
.07727 .77919 L
F
.06148 .77919 m
.06148 .84282 L
.12512 .84282 L
.12512 .77919 L
F
.10933 .77919 m
.10933 .84282 L
.17297 .84282 L
.17297 .77919 L
F
0 1 0 r
.15718 .77919 m
.15718 .84282 L
.22081 .84282 L
.22081 .77919 L
F
1 0 0 r
.20502 .77919 m
.20502 .84282 L
.26866 .84282 L
.26866 .77919 L
F
.25287 .77919 m
.25287 .84282 L
.31651 .84282 L
.31651 .77919 L
F
0 1 0 r
.30072 .77919 m
.30072 .84282 L
.36435 .84282 L
.36435 .77919 L
F
.34856 .77919 m
.34856 .84282 L
.4122 .84282 L
.4122 .77919 L
F
.39641 .77919 m
.39641 .84282 L
.46005 .84282 L
.46005 .77919 L
F
.44426 .77919 m
.44426 .84282 L
.50789 .84282 L
.50789 .77919 L
F
0 0 1 r
.49211 .77919 m
.49211 .84282 L
.55574 .84282 L
.55574 .77919 L
F
.53995 .77919 m
.53995 .84282 L
.60359 .84282 L
.60359 .77919 L
F
.5878 .77919 m
.5878 .84282 L
.65144 .84282 L
.65144 .77919 L
F
.63565 .77919 m
.63565 .84282 L
.69928 .84282 L
.69928 .77919 L
F
1 0 0 r
.68349 .77919 m
.68349 .84282 L
.74713 .84282 L
.74713 .77919 L
F
.73134 .77919 m
.73134 .84282 L
.79498 .84282 L
.79498 .77919 L
F
.77919 .77919 m
.77919 .84282 L
.84282 .84282 L
.84282 .77919 L
F
.82703 .77919 m
.82703 .84282 L
.89067 .84282 L
.89067 .77919 L
F
0 1 0 r
.87488 .77919 m
.87488 .84282 L
.93852 .84282 L
.93852 .77919 L
F
.92273 .77919 m
.92273 .84282 L
.98636 .84282 L
.98636 .77919 L
F
0 0 1 r
.01364 .82703 m
.01364 .89067 L
.07727 .89067 L
.07727 .82703 L
F
.06148 .82703 m
.06148 .89067 L
.12512 .89067 L
.12512 .82703 L
F
0 1 0 r
.10933 .82703 m
.10933 .89067 L
.17297 .89067 L
.17297 .82703 L
F
.15718 .82703 m
.15718 .89067 L
.22081 .89067 L
.22081 .82703 L
F
.20502 .82703 m
.20502 .89067 L
.26866 .89067 L
.26866 .82703 L
F
.25287 .82703 m
.25287 .89067 L
.31651 .89067 L
.31651 .82703 L
F
.30072 .82703 m
.30072 .89067 L
.36435 .89067 L
.36435 .82703 L
F
.34856 .82703 m
.34856 .89067 L
.4122 .89067 L
.4122 .82703 L
F
.39641 .82703 m
.39641 .89067 L
.46005 .89067 L
.46005 .82703 L
F
0 0 1 r
.44426 .82703 m
.44426 .89067 L
.50789 .89067 L
.50789 .82703 L
F
.49211 .82703 m
.49211 .89067 L
.55574 .89067 L
.55574 .82703 L
F
.53995 .82703 m
.53995 .89067 L
.60359 .89067 L
.60359 .82703 L
F
.5878 .82703 m
.5878 .89067 L
.65144 .89067 L
.65144 .82703 L
F
1 0 0 r
.63565 .82703 m
.63565 .89067 L
.69928 .89067 L
.69928 .82703 L
F
.68349 .82703 m
.68349 .89067 L
.74713 .89067 L
.74713 .82703 L
F
.73134 .82703 m
.73134 .89067 L
.79498 .89067 L
.79498 .82703 L
F
.77919 .82703 m
.77919 .89067 L
.84282 .89067 L
.84282 .82703 L
F
0 1 0 r
.82703 .82703 m
.82703 .89067 L
.89067 .89067 L
.89067 .82703 L
F
.87488 .82703 m
.87488 .89067 L
.93852 .89067 L
.93852 .82703 L
F
.92273 .82703 m
.92273 .89067 L
.98636 .89067 L
.98636 .82703 L
F
0 0 1 r
.01364 .87488 m
.01364 .93852 L
.07727 .93852 L
.07727 .87488 L
F
.06148 .87488 m
.06148 .93852 L
.12512 .93852 L
.12512 .87488 L
F
0 1 0 r
.10933 .87488 m
.10933 .93852 L
.17297 .93852 L
.17297 .87488 L
F
.15718 .87488 m
.15718 .93852 L
.22081 .93852 L
.22081 .87488 L
F
.20502 .87488 m
.20502 .93852 L
.26866 .93852 L
.26866 .87488 L
F
.25287 .87488 m
.25287 .93852 L
.31651 .93852 L
.31651 .87488 L
F
.30072 .87488 m
.30072 .93852 L
.36435 .93852 L
.36435 .87488 L
F
.34856 .87488 m
.34856 .93852 L
.4122 .93852 L
.4122 .87488 L
F
0 0 1 r
.39641 .87488 m
.39641 .93852 L
.46005 .93852 L
.46005 .87488 L
F
.44426 .87488 m
.44426 .93852 L
.50789 .93852 L
.50789 .87488 L
F
.49211 .87488 m
.49211 .93852 L
.55574 .93852 L
.55574 .87488 L
F
.53995 .87488 m
.53995 .93852 L
.60359 .93852 L
.60359 .87488 L
F
1 0 0 r
.5878 .87488 m
.5878 .93852 L
.65144 .93852 L
.65144 .87488 L
F
.63565 .87488 m
.63565 .93852 L
.69928 .93852 L
.69928 .87488 L
F
.68349 .87488 m
.68349 .93852 L
.74713 .93852 L
.74713 .87488 L
F
.73134 .87488 m
.73134 .93852 L
.79498 .93852 L
.79498 .87488 L
F
0 1 0 r
.77919 .87488 m
.77919 .93852 L
.84282 .93852 L
.84282 .87488 L
F
.82703 .87488 m
.82703 .93852 L
.89067 .93852 L
.89067 .87488 L
F
.87488 .87488 m
.87488 .93852 L
.93852 .93852 L
.93852 .87488 L
F
.92273 .87488 m
.92273 .93852 L
.98636 .93852 L
.98636 .87488 L
F
0 0 1 r
.01364 .92273 m
.01364 .98636 L
.07727 .98636 L
.07727 .92273 L
F
.06148 .92273 m
.06148 .98636 L
.12512 .98636 L
.12512 .92273 L
F
.10933 .92273 m
.10933 .98636 L
.17297 .98636 L
.17297 .92273 L
F
0 1 0 r
.15718 .92273 m
.15718 .98636 L
.22081 .98636 L
.22081 .92273 L
F
.20502 .92273 m
.20502 .98636 L
.26866 .98636 L
.26866 .92273 L
F
.25287 .92273 m
.25287 .98636 L
.31651 .98636 L
.31651 .92273 L
F
.30072 .92273 m
.30072 .98636 L
.36435 .98636 L
.36435 .92273 L
F
0 0 1 r
.34856 .92273 m
.34856 .98636 L
.4122 .98636 L
.4122 .92273 L
F
.39641 .92273 m
.39641 .98636 L
.46005 .98636 L
.46005 .92273 L
F
.44426 .92273 m
.44426 .98636 L
.50789 .98636 L
.50789 .92273 L
F
.49211 .92273 m
.49211 .98636 L
.55574 .98636 L
.55574 .92273 L
F
1 0 0 r
.53995 .92273 m
.53995 .98636 L
.60359 .98636 L
.60359 .92273 L
F
.5878 .92273 m
.5878 .98636 L
.65144 .98636 L
.65144 .92273 L
F
.63565 .92273 m
.63565 .98636 L
.69928 .98636 L
.69928 .92273 L
F
.68349 .92273 m
.68349 .98636 L
.74713 .98636 L
.74713 .92273 L
F
0 1 0 r
.73134 .92273 m
.73134 .98636 L
.79498 .98636 L
.79498 .92273 L
F
.77919 .92273 m
.77919 .98636 L
.84282 .98636 L
.84282 .92273 L
F
.82703 .92273 m
.82703 .98636 L
.89067 .98636 L
.89067 .92273 L
F
.87488 .92273 m
.87488 .98636 L
.93852 .98636 L
.93852 .92273 L
F
0 0 1 r
.92273 .92273 m
.92273 .98636 L
.98636 .98636 L
.98636 .92273 L
F
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[19]:=",
  ImageSize->{288, 214.312},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000eR000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`001oooool3ool000go003o1_l0o`0<o`00o`coo`003?l0o`04o`00o`So
oooo3?l00?l<o`3o00go003o3?oo000=o`3o00Sooooo2Ol00?lBo`3o00co003o3Ooo000:o`3o00So
oooo1_l00?lHo`3o00co003o3Ooo0007o`3o00Wooooo0007ooooo`?oo`003Ol00?l6o`3o00co003o
3?oo000<o`3o00Co003o2?ooool<o`00o`co0?l03Ol00?l<ool000go0?l02?ooool9o`00oa;o0?l0
3?l00?l=ool000[o0?l02?ooool6o`00oaSo0?l03?l00?l=ool000Oo0?l02Oooool000Oooooo0ooo
000=o`00o`Ko0?l03?l00?l<ool000co0?l01?l00?l8ooooo`co003o3?l0o`0=o`00o`coo`003Ol0
o`08ooooo`Wo003o4_l0o`0<o`00o`goo`002_l0o`08ooooo`Ko003o6?l0o`0<o`00o`goo`001ol0
o`09ooooo`001oooool=o`00o`co0?l03?l00?l<ool000go0?l02?ooool9o`00oa?o0?l03?l00?l<
ool000[o0?l02?ooool6o`00oaSo0?l03Ol00?l<ool000Oo0?l02?ooool3o`00o`co0?l01_oo000<
o`3o00co003o3Ooo0004o`3o00Wooooo0007ooooo`go003o3?l0o`0<o`00o`coo`003Ol0o`08oooo
o`Wo003o4ol0o`0<o`00o`coo`002_l0o`08ooooo`Ko003o6?l0o`0=o`00o`coo`001ol0o`08oooo
o`?o003o3?l0o`06ool000co0?l03?l00?l=ool000Co0?l02Oooool000Oooooo3Ol00?l<o`3o00co
003o3?oo000=o`3o00Sooooo2Ol00?lCo`3o00co003o3?oo000:o`3o00Sooooo1_l00?lHo`3o00go
003o3?oo0007o`3o00Sooooo0ol00?l<o`3o00Koo`003?l0o`0<o`00o`goo`001?l0o`09ooooo`00
1oooool9o`00oa?o0?l03?l00?l<ool000[o0?l02?ooool6o`00oaWo0?l03?l00?l<ool000Oo0?l0
2?ooool3o`00o`co0?l01_oo000<o`3o00go003o3?oo0004o`3o00Sooooo3?l0o`0<ool000co0?l0
3?l00?l>ool000Wooooo0007ooooo`Wo003o4ol0o`0<o`00o`coo`002_l0o`08ooooo`Ko003o6Ol0
o`0<o`00o`coo`001ol0o`08ooooo`?o003o3?l0o`06ool000co0?l03Ol00?l<ool000Co0?l02?oo
ool<o`3o00coo`003?l0o`0<o`00o`koo`002Oooool000Oooooo2Ol00?lCo`3o00co003o3?oo000:
o`3o00Sooooo1_l00?lIo`3o00co003o3?oo0007o`3o00Sooooo0ol00?l<o`3o00Koo`003?l0o`0=
o`00o`coo`001?l0o`08ooooo`co0?l03?oo000<o`3o00co003o3_oo0009ooooo`001oooool6o`00
oaWo0?l03?l00?l<ool000Oo0?l02?ooool3o`00o`co0?l01_oo000=o`3o00co003o3?oo0004o`3o
00Sooooo3?l0o`0<ool000co0?l03Ol00?l=ool000Sooooo2Ol0o`0Bool000co0?l03?l00?l;ool0
00Wooooo0007ooooo`Ko003o6Ol0o`0<o`00o`coo`001ol0o`08ooooo`?o003o3?l0o`06ool000go
0?l03?l00?l<ool000Co0?l02?ooool<o`3o00coo`003?l0o`0=o`00o`goo`002?ooool9o`3o01;o
o`003?l0o`0<o`00o`_oo`002Oooool000Oooooo1_l00?lIo`3o00co003o3?oo0007o`3o00Sooooo
0ol00?l<o`3o00Koo`003Ol0o`0<o`00o`coo`001?l0o`08ooooo`co0?l03?oo000<o`3o00go003o
3Ooo0008ooooo`Wo0?l04_oo000<o`3o00co003o2ooo0009ooooo`001oooool3o`00o`go0?l01_oo
000<o`3o00co003o3?oo0004o`3o00Sooooo3?l0o`0<ool000go0?l03?l00?l=ool000Sooooo2Ol0
o`0Bool000co0?l03Ol00?l:ool000Sooooo1_l0o`0Hool000co0?l03Ol00?l7ool000Wooooo0007
ooooo`?o003o3Ol0o`06ool000co0?l03?l00?l<ool000Co0?l02?ooool<o`3o00coo`003Ol0o`0<
o`00o`goo`002?ooool9o`3o01;oo`003?l0o`0=o`00o`[oo`002?ooool6o`3o01Soo`003?l0o`0=
o`00o`Ooo`002Oooool000Oooooo0ol00?l=o`3o00Koo`003?l0o`0<o`00o`coo`001?l0o`08oooo
o`co0?l03?oo000=o`3o00co003o3Ooo0008ooooo`Wo0?l04_oo000<o`3o00go003o2_oo0008oooo
o`Ko0?l06?oo000<o`3o00go003o1ooo0009ooooo`001oooool=o`3o00coo`003?l0o`0<o`00o`go
o`002?ooool9o`3o01?oo`003?l0o`0<o`00o`[oo`002?ooool6o`3o01Soo`003Ol0o`0<o`00o`Oo
o`002?ooool3o`3o00coo`001_l00?l<ool000co0?l03Ol00?l4ool000Wooooo0007ooooo`go0?l0
3?oo000<o`3o00co003o3Ooo0008ooooo`Wo0?l04ooo000<o`3o00co003o2_oo0008ooooo`Ko0?l0
6?oo000=o`3o00co003o1ooo0008ooooo`?o0?l03?oo0006o`00o`coo`003?l0o`0=o`00o`Coo`00
2Oooool000Oooooo3Ol0o`0<ool000co0?l03?l00?l=ool000Sooooo2Ol0o`0Cool000co0?l03?l0
0?l:ool000Sooooo1_l0o`0Hool000go0?l03?l00?l7ool000Sooooo0ol0o`0<ool000Ko003o3?oo
000<o`3o00go003o1?oo0009ooooo`001oooool9o`3o01?oo`003?l0o`0<o`00o`[oo`002?ooool6
o`3o01Woo`003?l0o`0<o`00o`Ooo`002?ooool3o`3o00coo`001_l00?l<ool000go0?l03?l00?l4
ool000Sooooo3?oo000<o`00o`coo`003?l0o`0>o`00o`Wooooo0007ooooo`Wo0?l04ooo000<o`3o
00co003o2_oo0008ooooo`Ko0?l06Ooo000<o`3o00co003o1ooo0008ooooo`?o0?l03?oo0006o`00
o`coo`003Ol0o`0<o`00o`Coo`002?ooool<ool000co003o3?oo000<o`3o00ko003o2Oooool000Oo
oooo2Ol0o`0Cool000co0?l03?l00?l:ool000Sooooo1_l0o`0Iool000co0?l03?l00?l7ool000So
oooo0ol0o`0<ool000Ko003o3?oo000=o`3o00co003o1?oo0008ooooo`coo`003?l00?l<ool000co
0?l03_l00?l9ooooo`001oooool6o`3o01Woo`003?l0o`0<o`00o`Ooo`002?ooool3o`3o00coo`00
1_l00?l=ool000co0?l03?l00?l4ool000Sooooo3?oo000<o`00o`coo`003Ol0o`0=o`00o`Sooooo
2Ooo000Bo`00o`coo`003?l0o`0;o`00o`Wooooo0007ooooo`Ko0?l06Ooo000<o`3o00co003o1ooo
0008ooooo`?o0?l03?oo0006o`00o`goo`003?l0o`0<o`00o`Coo`002?ooool<ool000co003o3?oo
000=o`3o00go003o2?ooool9ool001;o003o3?oo000<o`3o00_o003o2Oooool000Oooooo1_l0o`0I
ool000co0?l03?l00?l7ool000Sooooo0ol0o`0<ool000Ko003o3Ooo000<o`3o00co003o1?oo0008
ooooo`coo`003?l00?l<ool000go0?l03Ol00?l8ooooo`Woo`004_l00?l<ool000co0?l02ol00?l9
ooooo`001oooool3o`3o00goo`001_l00?l<ool000co0?l03?l00?l4ool000Sooooo3?oo000<o`00
o`goo`003?l0o`0=o`00o`Sooooo2Ooo000Bo`00o`coo`003Ol0o`0:o`00o`Sooooo1_oo000Ho`00
o`coo`003Ol0o`07o`00o`Wooooo0007ooooo`?o0?l03Ooo0006o`00o`coo`003?l0o`0<o`00o`Co
o`002?ooool<ool000co003o3Ooo000<o`3o00go003o2?ooool9ool001;o003o3?oo000=o`3o00[o
003o2?ooool6ool001So003o3?oo000=o`3o00Oo003o2Oooool000Oooooo0ol0o`0=ool000Ko003o
3?oo000<o`3o00co003o1?oo0008ooooo`coo`003?l00?l=ool000co0?l03Ol00?l8ooooo`Woo`00
4_l00?l<ool000go0?l02_l00?l8ooooo`Koo`006?l00?l<ool000go0?l01ol00?l9ooooo`001ooo
ool3o`3o00goo`001_l00?l<ool000co0?l03?l00?l4ool000Sooooo3?oo000<o`00o`goo`003?l0
o`0=o`00o`Sooooo2Ooo000Bo`00o`coo`003Ol0o`0:o`00o`Sooooo1_oo000Ho`00o`coo`003Ol0
o`07o`00o`Wooooo0007ooooo`?o0?l02_oo000<o`00o`coo`003?l0o`0=o`00o`Sooooo3?oo000@
o`00o`coo`003?l0o`0:o`00o`Sooooo2Ooo000Eo`00o`goo`003?l0o`07o`00o`Sooooo1_oo0009
o`00o`Ko0?l03?l00?l<ool000go0?l01?l00?l9ooooo`001oooool3o`3o00[oo`003?l00?l<ool0
00co0?l03Ol00?l8ooooo`coo`004?l00?l<ool000co0?l02_l00?l8ooooo`Woo`005Ol00?l=ool0
00co0?l01ol00?l8ooooo`Koo`002Ol00?l6o`3o00co003o3?oo000=o`3o00Co003o2Oooool000Oo
oooo0ol0o`0:ool000co003o3?oo000<o`3o00go003o2?ooool<ool0013o003o3?oo000<o`3o00[o
003o2?ooool9ool001Go003o3Ooo000<o`3o00Oo003o2?ooool6ool000Wo003o1_l0o`0<o`00o`co
o`003Ol0o`04o`00o`Wooooo0007ooooo`Ko0?l01ooo0003o`3o00co003o3?oo000<o`3o00[o003o
2?ooool3o`3o00coo`000ol0o`0=o`00o`coo`003?l0o`07o`00o`Sooooo3?oo0003o`00o`Ko0?l0
3?l00?l=ool000co0?l01?l00?l8ooooo`Woo`000ol00?l<o`3o00co003o3?oo000>o`3o00Wooooo
0007ooooo`Ko0?l01ooo0003o`3o00co003o3?oo000<o`3o00[o003o2?ooool3o`3o00coo`000ol0
o`0=o`00o`coo`003?l0o`07o`00o`Sooooo3?oo0003o`00o`Ko0?l03?l00?l=ool000co0?l01?l0
0?l8ooooo`Woo`000ol00?l<o`3o00co003o3?oo000>o`3o00Wooooo0007ooooo`Ko0?l01ooo0003
o`3o00co003o3?oo000<o`3o00[o003o2?ooool3o`3o00coo`000ol0o`0=o`00o`coo`003?l0o`07
o`00o`Sooooo3?oo0003o`00o`Ko0?l03?l00?l=ool000co0?l01?l00?l8ooooo`Woo`000ol00?l<
o`3o00co003o3?oo000>o`3o00Wooooo0007oooooa?o0?l03?l00?l<ool000co0?l01ol00?l8oooo
o`Ko0?l01_oo0009o`3o00go003o3?oo000<o`3o00Co003o2?ooool3o`3o00coo`002Ol0o`0<o`00
o`goo`003Ol0o`08ooooo`ooo`003?l0o`0<o`00o`coo`002ol0o`09ooooo`001ooooolCo`3o00co
003o3?oo000<o`3o00Oo003o2?ooool6o`3o00Koo`002Ol0o`0=o`00o`coo`003?l0o`04o`00o`So
oooo0ol0o`0<ool000Wo0?l03?l00?l=ool000go0?l02?ooool?ool000co0?l03?l00?l<ool000_o
0?l02Oooool000Oooooo4ol0o`0<o`00o`coo`003?l0o`07o`00o`Sooooo1_l0o`06ool000Wo0?l0
3Ol00?l<ool000co0?l01?l00?l8ooooo`?o0?l03?oo0009o`3o00co003o3Ooo000=o`3o00Sooooo
3ooo000<o`3o00co003o3?oo000;o`3o00Wooooo0007oooooa?o0?l03?l00?l<ool000co0?l01ol0
0?l8oooooaGo0?l03Ol00?l<ool000co0?l01?l00?l8ooooo`Ko0?l01_oo000<o`3o00co003o3Ooo
000=o`3o00Sooooo0ol0o`0<ool000co0?l03?l00?l<ool000_o0?l02Oooool000Oooooo4ol0o`0<
o`00o`coo`003?l0o`07o`00o`Sooooo5Ol0o`0=o`00o`coo`003?l0o`04o`00o`Sooooo1_l0o`06
ool000co0?l03?l00?l=ool000go0?l02?ooool3o`3o00coo`003?l0o`0<o`00o`coo`002ol0o`09
ooooo`001ooooolCo`3o00co003o3?oo000<o`3o00Oo003o2?oooolEo`3o00go003o3?oo000<o`3o
00Co003o2?ooool6o`3o00Koo`003?l0o`0<o`00o`goo`003Ol0o`08ooooo`?o0?l03?oo000<o`3o
00co003o3?oo000;o`3o00Wooooo0007ooooo`?o003o3Ol0o`0<o`00o`coo`003?l0o`0:o`00o`So
oooo4_l0o`0=o`00o`coo`003?l0o`07o`00o`Sooooo5Ol0o`0<o`00o`goo`003?l0o`04o`00o`So
oooo1_l0o`06ool000co0?l03?l00?l<ool000ko0?l02Oooool000Oooooo0ol00?l=o`3o00co003o
3?oo000<o`3o00[o003o2?oooolBo`3o00go003o3?oo000<o`3o00Oo003o2?oooolEo`3o00co003o
3Ooo000<o`3o00Co003o2?ooool6o`3o00Koo`003?l0o`0<o`00o`coo`003_l0o`09ooooo`001ooo
ool3o`00o`go0?l03?l00?l<ool000co0?l02_l00?l8oooooa;o0?l03Ol00?l<ool000co0?l01ol0
0?l8oooooaGo0?l03?l00?l=ool000co0?l01?l00?l8ooooo`Ko0?l01_oo000<o`3o00co003o3?oo
000>o`3o00Wooooo0007ooooo`go0?l03?l00?l<ool000co0?l03Ol00?l8ooooo`oo0?l03Ol00?l<
ool000co0?l02_l00?l8oooooa;o0?l03?l00?l=ool000co0?l01ol00?l8ooooo`?o0?l01_oo000<
o`3o00co003o3?oo000=o`3o00Co003o2Oooool000Oooooo3Ol0o`0<o`00o`coo`003?l0o`0=o`00
o`Sooooo3ol0o`0=o`00o`coo`003?l0o`0:o`00o`Sooooo4_l0o`0<o`00o`goo`003?l0o`07o`00
o`Sooooo0ol0o`06ool000co0?l03?l00?l<ool000go0?l01?l00?l9ooooo`001oooool=o`3o00co
003o3?oo000<o`3o00go003o2?ooool?o`3o00go003o3?oo000<o`3o00[o003o2?oooolBo`3o00co
003o3Ooo000<o`3o00Oo003o2?ooool3o`3o00Koo`003?l0o`0<o`00o`coo`003Ol0o`04o`00o`Wo
oooo0007ooooo`go0?l03?l00?l<ool000co0?l03Ol00?l8ooooo`oo0?l03Ol00?l<ool000co0?l0
2_l00?l8ooooo`?o0?l01_oo0009o`3o00co003o3Ooo000<o`3o00Oo003o2?ooool<ool000Wo0?l0
3?l00?l<ool000go0?l01?l00?l9ooooo`001oooool=o`3o00co003o3?oo000<o`3o00go003o2?oo
ool?o`3o00go003o3?oo000<o`3o00[o003o2?ooool3o`3o00Koo`002Ol0o`0<o`00o`goo`003?l0
o`07o`00o`Sooooo3?oo0009o`3o00co003o3?oo000=o`3o00Co003o2Oooool000Oooooo3Ol0o`0<
o`00o`coo`003?l0o`0=o`00o`Sooooo3ol0o`0=o`00o`coo`003?l0o`0:o`00o`Sooooo0ol0o`06
ool000Wo0?l03?l00?l=ool000co0?l01ol00?l8ooooo`coo`002Ol0o`0<o`00o`coo`003Ol0o`04
o`00o`Wooooo0007ooooo`Ko0?l00ooo000=o`00o`coo`003?l0o`0<o`00o`Coo`002?ooool3o`3o
00Koo`000ol0o`0<o`00o`goo`003?l0o`0=o`00o`Sooooo3?oo0003o`3o00co003o3?oo000=o`3o
00[o003o2?ooool9ool000?o003o1_l0o`0<o`00o`coo`003Ol0o`07o`00o`Wooooo0007ooooo`Ko
0?l00ooo000=o`00o`coo`003?l0o`0<o`00o`Coo`002?ooool3o`3o00Koo`000ol0o`0<o`00o`go
o`003?l0o`0=o`00o`Sooooo3?oo0003o`3o00co003o3?oo000=o`3o00[o003o2?ooool9ool000?o
003o1_l0o`0<o`00o`coo`003Ol0o`07o`00o`Wooooo0007ooooo`Ko0?l00ooo000=o`00o`coo`00
3?l0o`0<o`00o`Coo`002?ooool3o`3o00Koo`000ol0o`0<o`00o`goo`003?l0o`0=o`00o`Sooooo
3?oo0003o`3o00co003o3?oo000=o`3o00[o003o2?ooool9ool000?o003o1_l0o`0<o`00o`coo`00
3Ol0o`07o`00o`Wooooo0007ooooo`?o0?l02_oo0006o`00o`coo`003?l0o`0<o`00o`Ooo`002?oo
ool9ool000co003o3Ooo000<o`3o00co003o1?oo0008ooooo`Woo`003ol00?l<ool000go0?l03Ol0
0?l8ooooo`Koo`005Ol00?l<ool000co0?l02ol00?l9ooooo`001oooool3o`3o00[oo`001_l00?l<
ool000co0?l03?l00?l7ool000Sooooo2Ooo000<o`00o`goo`003?l0o`0<o`00o`Coo`002?ooool9
ool000oo003o3?oo000=o`3o00go003o2?ooool6ool001Go003o3?oo000<o`3o00_o003o2Oooool0
00Oooooo0ol0o`0:ool000Ko003o3?oo000<o`3o00co003o1ooo0008ooooo`Woo`003?l00?l=ool0
00co0?l03?l00?l4ool000Sooooo2Ooo000?o`00o`coo`003Ol0o`0=o`00o`Sooooo1_oo000Eo`00
o`coo`003?l0o`0;o`00o`Wooooo0007ooooo`?o0?l06Ooo000<o`3o00co003o2_oo0008ooooo`co
o`001_l00?l=ool000co0?l03?l00?l7ool000Sooooo2Ooo000<o`00o`coo`003Ol0o`0<o`00o`Co
o`002?ooool6ool001;o003o3?oo000<o`3o00ko003o2Oooool000Oooooo0ol0o`0Iool000co0?l0
3?l00?l:ool000Sooooo3?oo0006o`00o`goo`003?l0o`0<o`00o`Ooo`002?ooool9ool000co003o
3?oo000=o`3o00co003o1?oo0008ooooo`Koo`004_l00?l<ool000co0?l03_l00?l9ooooo`001ooo
ool3o`3o01Woo`003?l0o`0<o`00o`[oo`002?ooool<ool000Ko003o3Ooo000<o`3o00co003o1ooo
0008ooooo`Woo`003?l00?l<ool000go0?l03?l00?l4ool000Sooooo1_oo000Bo`00o`coo`003?l0
o`0>o`00o`Wooooo0007ooooo`Ko0?l04ooo000<o`3o00co003o3Ooo0008ooooo`?o0?l06Ooo000<
o`3o00co003o2_oo0008ooooo`coo`001_l00?l<ool000go0?l03?l00?l7ool000Sooooo2Ooo000<
o`00o`coo`003?l0o`0=o`00o`Coo`002Oooool000Oooooo1_l0o`0Cool000co0?l03?l00?l=ool0
00Sooooo0ol0o`0Iool000co0?l03?l00?l:ool000Sooooo3?oo0006o`00o`coo`003Ol0o`0<o`00
o`Ooo`002?ooool9ool000co003o3?oo000<o`3o00go003o1?oo0009ooooo`001oooool6o`3o01?o
o`003?l0o`0<o`00o`goo`002?ooool3o`3o01Woo`003?l0o`0<o`00o`[oo`002?ooool<ool000Ko
003o3?oo000=o`3o00co003o1ooo0008ooooo`Woo`003?l00?l<ool000co0?l03Ol00?l4ool000Wo
oooo0007ooooo`Ko0?l04ooo000<o`3o00co003o3Ooo0008ooooo`?o0?l06Ooo000<o`3o00co003o
2_oo0008ooooo`coo`001_l00?l<ool000go0?l03?l00?l7ool000Sooooo2Ooo000<o`00o`coo`00
3?l0o`0=o`00o`Coo`002Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oo
oooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool000Oo
oooo0ol0o`0=ool000Ko003o3?oo000<o`3o00co003o1?oo0008ooooo`coo`003?l00?l=ool000co
0?l03Ol00?l8ooooo`Woo`004_l00?l<ool000go0?l02_l00?l8ooooo`Koo`006?l00?l<ool000go
0?l01ol00?l9ooooo`001oooool3o`3o00goo`001_l00?l<ool000co0?l03?l00?l4ool000Sooooo
3?oo000<o`00o`goo`003?l0o`0=o`00o`Sooooo2Ooo000Bo`00o`coo`003Ol0o`0:o`00o`Sooooo
1_oo000Ho`00o`coo`003Ol0o`07o`00o`Wooooo0007ooooo`?o0?l03Ooo0006o`00o`coo`003?l0
o`0<o`00o`Coo`002?ooool<ool000co003o3Ooo000<o`3o00go003o2?ooool9ool001;o003o3?oo
000=o`3o00[o003o2?ooool6ool001So003o3?oo000=o`3o00Oo003o2Oooool000Oooooo3Ooo000<
o`00o`coo`003?l0o`0=o`00o`Sooooo2Ooo000Co`00o`coo`003?l0o`0:o`00o`Sooooo1_oo000H
o`00o`goo`003?l0o`07o`00o`Sooooo0ooo000<o`00o`Ko0?l03?l00?l<ool000go0?l01?l00?l9
ooooo`001oooool=ool000co003o3?oo000<o`3o00go003o2?ooool9ool001?o003o3?oo000<o`3o
00[o003o2?ooool6ool001So003o3Ooo000<o`3o00Oo003o2?ooool3ool000co003o1_l0o`0<o`00
o`coo`003Ol0o`04o`00o`Wooooo0007ooooo`goo`003?l00?l<ool000co0?l03Ol00?l8ooooo`Wo
o`004ol00?l<ool000co0?l02_l00?l8ooooo`Koo`006?l00?l=ool000co0?l01ol00?l8ooooo`?o
o`003?l00?l6o`3o00co003o3?oo000=o`3o00Co003o2Oooool000Oooooo2Ooo000Co`00o`coo`00
3?l0o`0:o`00o`Sooooo1_oo000Io`00o`coo`003?l0o`07o`00o`Sooooo0ooo000<o`00o`Ko0?l0
3?l00?l=ool000co0?l01?l00?l8ooooo`co003o3?l0o`0<o`00o`coo`003_l0o`09ooooo`001ooo
ool9ool001?o003o3?oo000<o`3o00[o003o2?ooool6ool001Wo003o3?oo000<o`3o00Oo003o2?oo
ool3ool000co003o1_l0o`0<o`00o`goo`003?l0o`04o`00o`Sooooo3?l00?l<o`3o00co003o3?oo
000>o`3o00Wooooo0007ooooo`Woo`004ol00?l<ool000co0?l02_l00?l8ooooo`Koo`006Ol00?l<
ool000co0?l01ol00?l8ooooo`?oo`003?l00?l6o`3o00co003o3Ooo000<o`3o00Co003o2?ooool<
o`00o`co0?l03?l00?l<ool000ko0?l02Oooool000Oooooo1_oo000Io`00o`coo`003?l0o`07o`00
o`Sooooo0ooo000<o`00o`Ko0?l03Ol00?l<ool000co0?l01?l00?l8ooooo`co003o3?l0o`0<o`00
o`goo`003Ol0o`08ooooo`Wo003o4_l0o`0<o`00o`coo`002ol0o`09ooooo`001oooool6ool001Wo
003o3?oo000<o`3o00Oo003o2?ooool3ool000co003o1_l0o`0=o`00o`coo`003?l0o`04o`00o`So
oooo3?l00?l<o`3o00co003o3Ooo000=o`3o00Sooooo2Ol00?lBo`3o00co003o3?oo000;o`3o00Wo
oooo0007ooooo`Koo`006Ol00?l<ool000co0?l01ol00?l8ooooo`?oo`003?l00?l6o`3o00go003o
3?oo000<o`3o00Co003o2?ooool<o`00o`co0?l03?l00?l=ool000go0?l02?ooool9o`00oa;o0?l0
3?l00?l<ool000_o0?l02Oooool000Oooooo0ooo000=o`00o`Ko0?l03?l00?l<ool000co0?l01?l0
0?l8ooooo`co003o3?l0o`0=o`00o`coo`003Ol0o`08ooooo`Wo003o4_l0o`0<o`00o`goo`002_l0
o`08ooooo`Ko003o6?l0o`0<o`00o`goo`001ol0o`09ooooo`001oooool3ool000go003o1_l0o`0<
o`00o`coo`003?l0o`04o`00o`Sooooo3?l00?l<o`3o00go003o3?oo000=o`3o00Sooooo2Ol00?lB
o`3o00co003o3Ooo000:o`3o00Sooooo1_l00?lHo`3o00co003o3Ooo0007o`3o00Wooooo0007oooo
o`?oo`003Ol00?l6o`3o00co003o3?oo000<o`3o00Co003o2?ooool<o`00o`co0?l03Ol00?l<ool0
00go0?l02?ooool9o`00oa;o0?l03?l00?l=ool000[o0?l02?ooool6o`00oaSo0?l03?l00?l=ool0
00Oo0?l02Oooool000Oooooo3Ol00?l<o`3o00co003o3?oo000=o`3o00Sooooo2Ol00?lCo`3o00co
003o3?oo000:o`3o00Sooooo1_l00?lHo`3o00go003o3?oo0007o`3o00Sooooo0ol00?l<o`3o00Ko
o`003?l0o`0<o`00o`goo`001?l0o`09ooooo`001oooool=o`00o`co0?l03?l00?l<ool000go0?l0
2?ooool9o`00oa?o0?l03?l00?l<ool000[o0?l02?ooool6o`00oaSo0?l03Ol00?l<ool000Oo0?l0
2?ooool3o`00o`co0?l01_oo000<o`3o00co003o3Ooo0004o`3o00Wooooo0007ooooo`go003o3?l0
o`0<o`00o`coo`003Ol0o`08ooooo`Wo003o4ol0o`0<o`00o`coo`002_l0o`08ooooo`Ko003o6?l0
o`0=o`00o`coo`001ol0o`08ooooo`?o003o3?l0o`06ool000co0?l03?l00?l=ool000Co0?l02Ooo
ool000Oooooo2Ol00?lCo`3o00co003o3?oo000:o`3o00Sooooo1_l00?lIo`3o00co003o3?oo0007
o`3o00Sooooo0ol00?l<o`3o00Koo`003?l0o`0=o`00o`coo`001?l0o`08ooooo`co0?l03?oo000<
o`3o00co003o3_oo0009ooooo`001oooool9o`00oa?o0?l03?l00?l<ool000[o0?l02?ooool6o`00
oaWo0?l03?l00?l<ool000Oo0?l02?ooool3o`00o`co0?l01_oo000<o`3o00go003o3?oo0004o`3o
00Sooooo3?l0o`0<ool000co0?l03?l00?l>ool000Wooooo0007ooooo`Wo003o4ol0o`0<o`00o`co
o`002_l0o`08ooooo`Ko003o6Ol0o`0<o`00o`coo`001ol0o`08ooooo`?o003o3?l0o`06ool000co
0?l03Ol00?l<ool000Co0?l02?ooool<o`3o00coo`003?l0o`0<o`00o`koo`002Oooool000Oooooo
1_l00?lIo`3o00co003o3?oo0007o`3o00Sooooo0ol00?l<o`3o00Koo`003Ol0o`0<o`00o`coo`00
1?l0o`08ooooo`co0?l03?oo000<o`3o00go003o3Ooo0008ooooo`Wo0?l04_oo000<o`3o00co003o
2ooo0009ooooo`001oooool6o`00oaWo0?l03?l00?l<ool000Oo0?l02?ooool3o`00o`co0?l01_oo
000=o`3o00co003o3?oo0004o`3o00Sooooo3?l0o`0<ool000co0?l03Ol00?l=ool000Sooooo2Ol0
o`0Bool000co0?l03?l00?l;ool000Wooooo0007ooooo`Ko003o6Ol0o`0<o`00o`coo`001ol0o`08
ooooo`?o003o3?l0o`06ool000go0?l03?l00?l<ool000Co0?l02?ooool<o`3o00coo`003?l0o`0=
o`00o`goo`002?ooool9o`3o01;oo`003?l0o`0<o`00o`_oo`002Oooool000Oooooo0ol00?l=o`3o
00Koo`003?l0o`0<o`00o`coo`001?l0o`08ooooo`co0?l03?oo000=o`3o00co003o3Ooo0008oooo
o`Wo0?l04_oo000<o`3o00go003o2_oo0008ooooo`Ko0?l06?oo000<o`3o00go003o1ooo0009oooo
o`001oooool3o`00o`go0?l01_oo000<o`3o00co003o3?oo0004o`3o00Sooooo3?l0o`0<ool000go
0?l03?l00?l=ool000Sooooo2Ol0o`0Bool000co0?l03Ol00?l:ool000Sooooo1_l0o`0Hool000co
0?l03Ol00?l7ool000Wooooo0007ooooo`?o003o3Ol0o`06ool000co0?l03?l00?l<ool000Co0?l0
2?ooool<o`3o00coo`003Ol0o`0<o`00o`goo`002?ooool9o`3o01;oo`003?l0o`0=o`00o`[oo`00
2?ooool6o`3o01Soo`003?l0o`0=o`00o`Ooo`002Oooool000Oooooo0ol00?l:o`3o00coo`003?l0
o`0<o`00o`goo`002?ooool<o`3o013oo`003?l0o`0<o`00o`[oo`002?ooool9o`3o01Goo`003Ol0
o`0<o`00o`Ooo`002?ooool6o`3o00Woo`001_l00?l<ool000co0?l03Ol00?l4ool000Wooooo0007
ooooo`?o003o2_l0o`0<ool000co0?l03?l00?l=ool000Sooooo3?l0o`0@ool000co0?l03?l00?l:
ool000Sooooo2Ol0o`0Eool000go0?l03?l00?l7ool000Sooooo1_l0o`09ool000Ko003o3?oo000<
o`3o00go003o1?oo0009ooooo`001oooool3o`00o`[o0?l03?oo000<o`3o00co003o3Ooo0008oooo
o`co0?l04?oo000<o`3o00co003o2_oo0008ooooo`Wo0?l05Ooo000=o`3o00co003o1ooo0008oooo
o`Ko0?l02Ooo0006o`00o`coo`003?l0o`0=o`00o`Coo`002Oooool000Oooooo1_l00?l7o`3o00?o
003o3?oo000<o`3o00co003o2_oo0008ooooo`?o003o3?l0o`03o`00o`goo`003?l0o`0<o`00o`Oo
o`002?ooool<o`3o00?oo`001_l00?l<ool000go0?l03?l00?l4ool000Sooooo2Ol0o`03ool000co
003o3?oo000<o`3o00ko003o2Oooool000Oooooo1_l00?l7o`3o00?o003o3?oo000<o`3o00co003o
2_oo0008ooooo`?o003o3?l0o`03o`00o`goo`003?l0o`0<o`00o`Ooo`002?ooool<o`3o00?oo`00
1_l00?l<ool000go0?l03?l00?l4ool000Sooooo2Ol0o`03ool000co003o3?oo000<o`3o00ko003o
2Oooool000Oooooo1_l00?l7o`3o00?o003o3?oo000<o`3o00co003o2_oo0008ooooo`?o003o3?l0
o`03o`00o`goo`003?l0o`0<o`00o`Ooo`002?ooool<o`3o00?oo`001_l00?l<ool000go0?l03?l0
0?l4ool000Sooooo2Ol0o`03ool000co003o3?oo000<o`3o00ko003o2Oooool000Oooooo4ol00?l<
ool000co0?l03?l00?l7ool000Sooooo1_l00?l6o`3o00Wo003o3Ooo000<o`3o00co003o1?oo0008
ooooo`?o003o3?l0o`09o`00o`coo`003Ol0o`0=o`00o`Sooooo3ol0o`0<o`00o`coo`003?l0o`0;
o`00o`Wooooo0007oooooa?o003o3?oo000<o`3o00co003o1ooo0008ooooo`Ko003o1_l0o`09o`00
o`goo`003?l0o`0<o`00o`Coo`002?ooool3o`00o`co0?l02Ol00?l<ool000go0?l03Ol00?l8oooo
o`oo0?l03?l00?l<ool000co0?l02ol00?l9ooooo`001ooooolCo`00o`coo`003?l0o`0<o`00o`Oo
o`002?ooool6o`00o`Ko0?l02Ol00?l=ool000co0?l03?l00?l4ool000Sooooo0ol00?l<o`3o00Wo
003o3?oo000=o`3o00go003o2?ooool?o`3o00co003o3?oo000<o`3o00_o003o2Oooool000Oooooo
4ol00?l<ool000co0?l03?l00?l7ool000Sooooo5Ol00?l=ool000co0?l03?l00?l4ool000Sooooo
1_l00?l6o`3o00co003o3?oo000=o`3o00go003o2?ooool3o`00o`co0?l03?l00?l<ool000co0?l0
2ol00?l9ooooo`001ooooolCo`00o`coo`003?l0o`0<o`00o`Ooo`002?oooolEo`00o`goo`003?l0
o`0<o`00o`Coo`002?ooool6o`00o`Ko0?l03?l00?l<ool000go0?l03Ol00?l8ooooo`?o003o3?l0
o`0<o`00o`coo`003?l0o`0;o`00o`Wooooo0007oooooa?o003o3?oo000<o`3o00co003o1ooo0008
oooooaGo003o3Ooo000<o`3o00co003o1?oo0008ooooo`Ko003o1_l0o`0<o`00o`coo`003Ol0o`0=
o`00o`Sooooo0ol00?l<o`3o00co003o3?oo000<o`3o00_o003o2Oooool000Oooooo0ooo000=o`00
o`coo`003?l0o`0<o`00o`[oo`002?oooolBo`00o`goo`003?l0o`0<o`00o`Ooo`002?oooolEo`00
o`coo`003Ol0o`0<o`00o`Coo`002?ooool6o`00o`Ko0?l03?l00?l<ool000co0?l03_l00?l9oooo
o`001oooool3ool000go003o3?oo000<o`3o00co003o2_oo0008oooooa;o003o3Ooo000<o`3o00co
003o1ooo0008oooooaGo003o3?oo000=o`3o00co003o1?oo0008ooooo`Ko003o1_l0o`0<o`00o`co
o`003?l0o`0>o`00o`Wooooo0007ooooo`?oo`003Ol00?l<ool000co0?l03?l00?l:ool000Sooooo
4_l00?l=ool000co0?l03?l00?l7ool000Sooooo5Ol00?l<ool000go0?l03?l00?l4ool000Sooooo
1_l00?l6o`3o00co003o3?oo000<o`3o00ko003o2Oooool000Oooooo0ooo000=o`00o`coo`003?l0
o`0<o`00o`[oo`002?oooolBo`00o`goo`003?l0o`0<o`00o`Ooo`002?oooolEo`00o`coo`003Ol0
o`0<o`00o`Coo`002?ooool6o`00o`Ko0?l03?l00?l<ool000co0?l03_l00?l9ooooo`001oooool=
o`00o`coo`003?l0o`0<o`00o`goo`002?ooool?o`00o`goo`003?l0o`0<o`00o`[oo`002?oooolB
o`00o`coo`003Ol0o`0<o`00o`Ooo`002?ooool3o`00o`Ko0?l03?l00?l<ool000co0?l03Ol00?l4
ool000Wooooo0007ooooo`go003o3?oo000<o`3o00co003o3Ooo0008ooooo`oo003o3Ooo000<o`3o
00co003o2_oo0008oooooa;o003o3?oo000=o`3o00co003o1ooo0008ooooo`?o003o1_l0o`0<o`00
o`coo`003?l0o`0=o`00o`Coo`002Oooool000Oooooo3Ol00?l<ool000co0?l03?l00?l=ool000So
oooo3ol00?l=ool000co0?l03?l00?l:ool000Sooooo4_l00?l<ool000go0?l03?l00?l7ool000So
oooo0ol00?l6o`3o00co003o3?oo000<o`3o00go003o1?oo0009ooooo`001oooool=o`00o`coo`00
3?l0o`0<o`00o`goo`002?ooool?o`00o`goo`003?l0o`0<o`00o`[oo`002?ooool3o`00o`Ko0?l0
2Ol00?l<ool000go0?l03?l00?l7ool000Sooooo3?l0o`09o`00o`coo`003?l0o`0=o`00o`Coo`00
2Oooool000Oooooo3Ol00?l<ool000co0?l03?l00?l=ool000Sooooo3ol00?l=ool000co0?l03?l0
0?l:ool000Sooooo0ol00?l6o`3o00Wo003o3?oo000=o`3o00co003o1ooo0008ooooo`co0?l02Ol0
0?l<ool000co0?l03Ol00?l4ool000Wooooo0007ooooo`go003o3?oo000<o`3o00co003o3Ooo0008
ooooo`oo003o3Ooo000<o`3o00co003o2_oo0008ooooo`?o003o1_l0o`09o`00o`coo`003Ol0o`0<
o`00o`Ooo`002?ooool<o`3o00Wo003o3?oo000<o`3o00go003o1?oo0009ooooo`001oooool6o`00
o`?o0?l03Ooo000<o`3o00co003o3?oo0004o`3o00Sooooo0ol00?l6o`3o00?o003o3?oo000=o`3o
00co003o3Ooo0008ooooo`co0?l00ol00?l<ool000co0?l03Ol00?l:ool000Sooooo2Ol0o`03ool0
00Ko003o3?oo000<o`3o00go003o1ooo0009ooooo`001oooool6o`00o`?o0?l03Ooo000<o`3o00co
003o3?oo0004o`3o00Sooooo0ol00?l6o`3o00?o003o3?oo000=o`3o00co003o3Ooo0008ooooo`co
0?l00ol00?l<ool000co0?l03Ol00?l:ool000Sooooo2Ol0o`03ool000Ko003o3?oo000<o`3o00go
003o1ooo0009ooooo`001oooool6o`00o`?o0?l03Ooo000<o`3o00co003o3?oo0004o`3o00Sooooo
0ol00?l6o`3o00?o003o3?oo000=o`3o00co003o3Ooo0008ooooo`co0?l00ol00?l<ool000co0?l0
3Ol00?l:ool000Sooooo2Ol0o`03ool000Ko003o3?oo000<o`3o00go003o1ooo0009ooooo`001ooo
ool3o`00o`[o0?l01_oo000<o`3o00co003o3?oo0007o`3o00Sooooo2Ol0o`0<ool000go0?l03?l0
0?l<ool000Co0?l02?ooool9o`3o00ooo`003?l0o`0=o`00o`goo`002?ooool6o`3o01Goo`003?l0
o`0<o`00o`_oo`002Oooool000Oooooo0ol00?l:o`3o00Koo`003?l0o`0<o`00o`coo`001ol0o`08
ooooo`Wo0?l03?oo000=o`3o00co003o3?oo0004o`3o00Sooooo2Ol0o`0?ool000co0?l03Ol00?l=
ool000Sooooo1_l0o`0Eool000co0?l03?l00?l;ool000Wooooo0007ooooo`?o003o2_l0o`06ool0
00co0?l03?l00?l<ool000Oo0?l02?ooool9o`3o00coo`003Ol0o`0<o`00o`coo`001?l0o`08oooo
o`Wo0?l03ooo000<o`3o00go003o3Ooo0008ooooo`Ko0?l05Ooo000<o`3o00co003o2ooo0009oooo
o`001oooool3o`00oaWo0?l03?l00?l<ool000[o0?l02?ooool<o`3o00Koo`003Ol0o`0<o`00o`co
o`001ol0o`08ooooo`Wo0?l03?oo000<o`3o00go003o3?oo0004o`3o00Sooooo1_l0o`0Bool000co
0?l03?l00?l>ool000Wooooo0007ooooo`?o003o6Ol0o`0<o`00o`coo`002_l0o`08ooooo`co0?l0
1_oo000=o`3o00co003o3?oo0007o`3o00Sooooo2Ol0o`0<ool000co0?l03Ol00?l<ool000Co0?l0
2?ooool6o`3o01;oo`003?l0o`0<o`00o`koo`002Oooool000Oooooo0ol00?lIo`3o00co003o3?oo
000:o`3o00Sooooo3?l0o`06ool000go0?l03?l00?l<ool000Oo0?l02?ooool9o`3o00coo`003?l0
o`0=o`00o`coo`001?l0o`08ooooo`Ko0?l04_oo000<o`3o00co003o3_oo0009ooooo`001oooool6
o`00oa?o0?l03?l00?l<ool000go0?l02?ooool3o`00oaWo0?l03?l00?l<ool000[o0?l02?ooool<
o`3o00Koo`003?l0o`0=o`00o`coo`001ol0o`08ooooo`Wo0?l03?oo000<o`3o00co003o3Ooo0004
o`3o00Wooooo0007ooooo`Ko003o4ol0o`0<o`00o`coo`003Ol0o`08ooooo`?o003o6Ol0o`0<o`00
o`coo`002_l0o`08ooooo`co0?l01_oo000<o`3o00go003o3?oo0007o`3o00Sooooo2Ol0o`0<ool0
00co0?l03?l00?l=ool000Co0?l02Oooool000Oooooo1_l00?lCo`3o00co003o3?oo000=o`3o00So
oooo0ol00?lIo`3o00co003o3?oo000:o`3o00Sooooo3?l0o`06ool000co0?l03Ol00?l<ool000Oo
0?l02?ooool9o`3o00coo`003?l0o`0<o`00o`goo`001?l0o`09ooooo`001oooool6o`00oa?o0?l0
3?l00?l<ool000go0?l02?ooool3o`00oaWo0?l03?l00?l<ool000[o0?l02?ooool<o`3o00Koo`00
3?l0o`0=o`00o`coo`001ol0o`08ooooo`Wo0?l03?oo000<o`3o00co003o3Ooo0004o`3o00Wooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo
003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo0007ooooo`?o003o3Ol0o`06ool000co
0?l03?l00?l<ool000Co0?l02?ooool<o`3o00coo`003Ol0o`0<o`00o`goo`002?ooool9o`3o01;o
o`003?l0o`0=o`00o`[oo`002?ooool6o`3o01Soo`003?l0o`0=o`00o`Ooo`002Oooool000Oooooo
0ol00?l=o`3o00Koo`003?l0o`0<o`00o`coo`001?l0o`08ooooo`co0?l03?oo000=o`3o00co003o
3Ooo0008ooooo`Wo0?l04_oo000<o`3o00go003o2_oo0008ooooo`Ko0?l06?oo000<o`3o00go003o
1ooo0009ooooo`001oooool3o`00o`go0?l01_oo000<o`3o00co003o3?oo0004o`3o00Sooooo3?l0
o`0<ool000go0?l03?l00?l=ool000Sooooo2Ol0o`0Bool000co0?l03Ol00?l:ool000Sooooo1_l0
o`0Hool000co0?l03Ol00?l7ool000Wooooo0007ooooo`go0?l03?oo000<o`3o00co003o3Ooo0008
ooooo`Wo0?l04ooo000<o`3o00co003o2_oo0008ooooo`Ko0?l06?oo000=o`3o00co003o1ooo0008
ooooo`?o0?l03?oo0006o`00o`coo`003?l0o`0=o`00o`Coo`002Oooool000Oooooo3Ol0o`0<ool0
00co0?l03?l00?l=ool000Sooooo2Ol0o`0Cool000co0?l03?l00?l:ool000Sooooo1_l0o`0Hool0
00go0?l03?l00?l7ool000Sooooo0ol0o`0<ool000Ko003o3?oo000<o`3o00go003o1?oo0009oooo
o`001oooool=o`3o00coo`003?l0o`0<o`00o`goo`002?ooool9o`3o01?oo`003?l0o`0<o`00o`[o
o`002?ooool6o`3o01Soo`003Ol0o`0<o`00o`Ooo`002?ooool3o`3o00coo`001_l00?l<ool000co
0?l03Ol00?l4ool000Wooooo0007ooooo`Wo0?l04ooo000<o`3o00co003o2_oo0008ooooo`Ko0?l0
6Ooo000<o`3o00co003o1ooo0008ooooo`?o0?l03?oo0006o`00o`coo`003Ol0o`0<o`00o`Coo`00
2?ooool<ool000co003o3?oo000<o`3o00ko003o2Oooool000Oooooo2Ol0o`0Cool000co0?l03?l0
0?l:ool000Sooooo1_l0o`0Iool000co0?l03?l00?l7ool000Sooooo0ol0o`0<ool000Ko003o3?oo
000=o`3o00co003o1?oo0008ooooo`coo`003?l00?l<ool000co0?l03_l00?l9ooooo`001oooool9
o`3o01?oo`003?l0o`0<o`00o`[oo`002?ooool6o`3o01Woo`003?l0o`0<o`00o`Ooo`002?ooool3
o`3o00coo`001_l00?l<ool000go0?l03?l00?l4ool000Sooooo3?oo000<o`00o`coo`003?l0o`0>
o`00o`Wooooo0007ooooo`Ko0?l06Ooo000<o`3o00co003o1ooo0008ooooo`?o0?l03?oo0006o`00
o`goo`003?l0o`0<o`00o`Coo`002?ooool<ool000co003o3?oo000=o`3o00go003o2?ooool9ool0
01;o003o3?oo000<o`3o00_o003o2Oooool000Oooooo1_l0o`0Iool000co0?l03?l00?l7ool000So
oooo0ol0o`0<ool000Ko003o3Ooo000<o`3o00co003o1?oo0008ooooo`coo`003?l00?l<ool000go
0?l03Ol00?l8ooooo`Woo`004_l00?l<ool000co0?l02ol00?l9ooooo`001oooool6o`3o01Woo`00
3?l0o`0<o`00o`Ooo`002?ooool3o`3o00coo`001_l00?l=ool000co0?l03?l00?l4ool000Sooooo
3?oo000<o`00o`coo`003Ol0o`0=o`00o`Sooooo2Ooo000Bo`00o`coo`003?l0o`0;o`00o`Wooooo
0007ooooo`?o0?l03Ooo0006o`00o`coo`003?l0o`0<o`00o`Coo`002?ooool<ool000co003o3Ooo
000<o`3o00go003o2?ooool9ool001;o003o3?oo000=o`3o00[o003o2?ooool6ool001So003o3?oo
000=o`3o00Oo003o2Oooool000Oooooo0ol0o`0=ool000Ko003o3?oo000<o`3o00co003o1?oo0008
ooooo`coo`003?l00?l=ool000co0?l03Ol00?l8ooooo`Woo`004_l00?l<ool000go0?l02_l00?l8
ooooo`Koo`006?l00?l<ool000go0?l01ol00?l9ooooo`001oooool3o`3o00goo`001_l00?l<ool0
00co0?l03?l00?l4ool000Sooooo3?oo000<o`00o`goo`003?l0o`0=o`00o`Sooooo2Ooo000Bo`00
o`coo`003Ol0o`0:o`00o`Sooooo1_oo000Ho`00o`coo`003Ol0o`07o`00o`Wooooo0007ooooo`go
o`003?l00?l<ool000co0?l03Ol00?l8ooooo`Woo`004ol00?l<ool000co0?l02_l00?l8ooooo`Ko
o`006?l00?l=ool000co0?l01ol00?l8ooooo`?oo`003?l00?l6o`3o00co003o3?oo000=o`3o00Co
003o2Oooool000Oooooo3Ooo000<o`00o`coo`003?l0o`0=o`00o`Sooooo2Ooo000Co`00o`coo`00
3?l0o`0:o`00o`Sooooo1_oo000Ho`00o`goo`003?l0o`07o`00o`Sooooo0ooo000<o`00o`Ko0?l0
3?l00?l<ool000go0?l01?l00?l9ooooo`001oooool=ool000co003o3?oo000<o`3o00go003o2?oo
ool9ool001?o003o3?oo000<o`3o00[o003o2?ooool6ool001So003o3Ooo000<o`3o00Oo003o2?oo
ool3ool000co003o1_l0o`0<o`00o`coo`003Ol0o`04o`00o`Wooooo0007ooooo`Woo`004ol00?l<
ool000co0?l02_l00?l8ooooo`Koo`006Ol00?l<ool000co0?l01ol00?l8ooooo`?oo`003?l00?l6
o`3o00co003o3Ooo000<o`3o00Co003o2?ooool<o`00o`co0?l03?l00?l<ool000ko0?l02Oooool0
00Oooooo2Ooo000Co`00o`coo`003?l0o`0:o`00o`Sooooo1_oo000Io`00o`coo`003?l0o`07o`00
o`Sooooo0ooo000<o`00o`Ko0?l03?l00?l=ool000co0?l01?l00?l8ooooo`co003o3?l0o`0<o`00
o`coo`003_l0o`09ooooo`001oooool9ool001?o003o3?oo000<o`3o00[o003o2?ooool6ool001Wo
003o3?oo000<o`3o00Oo003o2?ooool3ool000co003o1_l0o`0<o`00o`goo`003?l0o`04o`00o`So
oooo3?l00?l<o`3o00co003o3?oo000>o`3o00Wooooo0007ooooo`Koo`006Ol00?l<ool000co0?l0
1ol00?l8ooooo`?oo`003?l00?l6o`3o00go003o3?oo000<o`3o00Co003o2?ooool<o`00o`co0?l0
3?l00?l=ool000go0?l02?ooool9o`00oa;o0?l03?l00?l<ool000_o0?l02Oooool000Oooooo1_oo
000Io`00o`coo`003?l0o`07o`00o`Sooooo0ooo000<o`00o`Ko0?l03Ol00?l<ool000co0?l01?l0
0?l8ooooo`co003o3?l0o`0<o`00o`goo`003Ol0o`08ooooo`Wo003o4_l0o`0<o`00o`coo`002ol0
o`09ooooo`001oooool6ool001Wo003o3?oo000<o`3o00Oo003o2?ooool3ool000co003o1_l0o`0=
o`00o`coo`003?l0o`04o`00o`Sooooo3?l00?l<o`3o00co003o3Ooo000=o`3o00Sooooo2Ol00?lB
o`3o00co003o3?oo000;o`3o00Wooooo0007ooooo`?oo`003Ol00?l6o`3o00co003o3?oo000<o`3o
00Co003o2?ooool<o`00o`co0?l03Ol00?l<ool000go0?l02?ooool9o`00oa;o0?l03?l00?l=ool0
00[o0?l02?ooool6o`00oaSo0?l03?l00?l=ool000Oo0?l02Oooool000Oooooo0ooo000=o`00o`Ko
0?l03?l00?l<ool000co0?l01?l00?l8ooooo`co003o3?l0o`0=o`00o`coo`003Ol0o`08ooooo`Wo
003o4_l0o`0<o`00o`goo`002_l0o`08ooooo`Ko003o6?l0o`0<o`00o`goo`001ol0o`09ooooo`00
1oooool3ool000go003o1_l0o`0<o`00o`coo`003?l0o`04o`00o`Sooooo3?l00?l<o`3o00go003o
3?oo000=o`3o00Sooooo2Ol00?lBo`3o00co003o3Ooo000:o`3o00Sooooo1_l00?lHo`3o00co003o
3Ooo0007o`3o00Wooooo0007ooooo`?oo`002_l00?l<o`3o00co003o3?oo000=o`3o00Sooooo3?l0
0?l@o`3o00co003o3?oo000:o`3o00Sooooo2Ol00?lEo`3o00go003o3?oo0007o`3o00Sooooo1_l0
0?l9o`3o00Koo`003?l0o`0<o`00o`goo`001?l0o`09ooooo`001oooool3ool000[o003o3?l0o`0<
o`00o`coo`003Ol0o`08ooooo`co003o4?l0o`0<o`00o`coo`002_l0o`08ooooo`Wo003o5Ol0o`0=
o`00o`coo`001ol0o`08ooooo`Ko003o2Ol0o`06ool000co0?l03?l00?l=ool000Co0?l02Oooool0
00Oooooo0ooo000:o`00o`co0?l03?l00?l<ool000go0?l02?ooool<o`00oa3o0?l03?l00?l<ool0
00[o0?l02?ooool9o`00oaGo0?l03Ol00?l<ool000Oo0?l02?ooool6o`00o`Wo0?l01_oo000<o`3o
00co003o3Ooo0004o`3o00Wooooo0007ooooo`Koo`001ol00?l3ool000co0?l03?l00?l<ool000[o
0?l02?ooool3ool000co003o0ooo000=o`3o00co003o3?oo0007o`3o00Sooooo3?l00?l3o`3o00Ko
o`003?l0o`0=o`00o`coo`001?l0o`08ooooo`Wo003o0ol0o`0<ool000co0?l03?l00?l>ool000Wo
oooo0007ooooo`Koo`001ol00?l3ool000co0?l03?l00?l<ool000[o0?l02?ooool3ool000co003o
0ooo000=o`3o00co003o3?oo0007o`3o00Sooooo3?l00?l3o`3o00Koo`003?l0o`0=o`00o`coo`00
1?l0o`08ooooo`Wo003o0ol0o`0<ool000co0?l03?l00?l>ool000Wooooo0007ooooo`Koo`001ol0
0?l3ool000co0?l03?l00?l<ool000[o0?l02?ooool3ool000co003o0ooo000=o`3o00co003o3?oo
0007o`3o00Sooooo3?l00?l3o`3o00Koo`003?l0o`0=o`00o`coo`001?l0o`08ooooo`Wo003o0ol0
o`0<ool000co0?l03?l00?l>ool000Wooooo0007oooooa?oo`003?l0o`0<o`00o`coo`001ol0o`08
ooooo`Koo`001_l00?l9ool000go0?l03?l00?l<ool000Co0?l02?ooool3ool000co003o2Ooo000<
o`3o00go003o3Ooo0008ooooo`oo003o3?oo000<o`3o00co003o2ooo0009ooooo`001ooooolCool0
00co0?l03?l00?l<ool000Oo0?l02?ooool6ool000Ko003o2Ooo000=o`3o00co003o3?oo0004o`3o
00Sooooo0ooo000<o`00o`Woo`003?l0o`0=o`00o`goo`002?ooool?o`00o`coo`003?l0o`0<o`00
o`_oo`002Oooool000Oooooo4ooo000<o`3o00co003o3?oo0007o`3o00Sooooo1_oo0006o`00o`Wo
o`003Ol0o`0<o`00o`coo`001?l0o`08ooooo`?oo`003?l00?l9ool000co0?l03Ol00?l=ool000So
oooo3ol00?l<ool000co0?l03?l00?l;ool000Wooooo0007oooooa?oo`003?l0o`0<o`00o`coo`00
1ol0o`08oooooaGoo`003Ol0o`0<o`00o`coo`001?l0o`08ooooo`Koo`001_l00?l<ool000co0?l0
3Ol00?l=ool000Sooooo0ooo000<o`00o`coo`003?l0o`0<o`00o`_oo`002Oooool000Oooooo4ooo
000<o`3o00co003o3?oo0007o`3o00Sooooo5Ooo000=o`3o00co003o3?oo0004o`3o00Sooooo1_oo
0006o`00o`coo`003?l0o`0=o`00o`goo`002?ooool3ool000co003o3?oo000<o`3o00co003o2ooo
0009ooooo`001ooooolCool000co0?l03?l00?l<ool000Oo0?l02?oooolEool000go0?l03?l00?l<
ool000Co0?l02?ooool6ool000Ko003o3?oo000<o`3o00go003o3Ooo0008ooooo`?oo`003?l00?l<
ool000co0?l03?l00?l;ool000Wooooo0007ooooo`?o0?l03Ooo000<o`3o00co003o3?oo000:o`3o
00Sooooo4_oo000=o`3o00co003o3?oo0007o`3o00Sooooo5Ooo000<o`3o00go003o3?oo0004o`3o
00Sooooo1_oo0006o`00o`coo`003?l0o`0<o`00o`koo`002Oooool000Oooooo0ol0o`0=ool000co
0?l03?l00?l<ool000[o0?l02?oooolBool000go0?l03?l00?l<ool000Oo0?l02?oooolEool000co
0?l03Ol00?l<ool000Co0?l02?ooool6ool000Ko003o3?oo000<o`3o00co003o3_oo0009ooooo`00
1oooool3o`3o00goo`003?l0o`0<o`00o`coo`002_l0o`08oooooa;oo`003Ol0o`0<o`00o`coo`00
1ol0o`08oooooaGoo`003?l0o`0=o`00o`coo`001?l0o`08ooooo`Koo`001_l00?l<ool000co0?l0
3?l00?l>ool000Wooooo0007ooooo`goo`003?l0o`0<o`00o`coo`003Ol0o`08ooooo`ooo`003Ol0
o`0<o`00o`coo`002_l0o`08oooooa;oo`003?l0o`0=o`00o`coo`001ol0o`08ooooo`?oo`001_l0
0?l<ool000co0?l03?l00?l=ool000Co0?l02Oooool000Oooooo3Ooo000<o`3o00co003o3?oo000=
o`3o00Sooooo3ooo000=o`3o00co003o3?oo000:o`3o00Sooooo4_oo000<o`3o00go003o3?oo0007
o`3o00Sooooo0ooo0006o`00o`coo`003?l0o`0<o`00o`goo`001?l0o`09ooooo`001oooool=ool0
00co0?l03?l00?l<ool000go0?l02?ooool?ool000go0?l03?l00?l<ool000[o0?l02?oooolBool0
00co0?l03Ol00?l<ool000Oo0?l02?ooool3ool000Ko003o3?oo000<o`3o00co003o3Ooo0004o`3o
00Wooooo0007ooooo`goo`003?l0o`0<o`00o`coo`003Ol0o`08ooooo`ooo`003Ol0o`0<o`00o`co
o`002_l0o`08ooooo`?oo`001_l00?l9ool000co0?l03Ol00?l<ool000Oo0?l02?ooool<o`00o`Wo
o`003?l0o`0<o`00o`goo`001?l0o`09ooooo`001oooool=ool000co0?l03?l00?l<ool000go0?l0
2?ooool?ool000go0?l03?l00?l<ool000[o0?l02?ooool3ool000Ko003o2Ooo000<o`3o00go003o
3?oo0007o`3o00Sooooo3?l00?l9ool000co0?l03?l00?l=ool000Co0?l02Oooool000Oooooo3Ooo
000<o`3o00co003o3?oo000=o`3o00Sooooo3ooo000=o`3o00co003o3?oo000:o`3o00Sooooo0ooo
0006o`00o`Woo`003?l0o`0=o`00o`coo`001ol0o`08ooooo`co003o2Ooo000<o`3o00co003o3Ooo
0004o`3o00Wooooo0007ooooo`Koo`000ol00?l=o`3o00co003o3?oo000<o`3o00Co003o2?ooool3
ool000Ko003o0ooo000<o`3o00go003o3?oo000=o`3o00Sooooo3?l00?l3ool000co0?l03?l00?l=
ool000[o0?l02?ooool9o`00o`?o0?l01_oo000<o`3o00co003o3Ooo0007o`3o00Wooooo0007oooo
o`Koo`000ol00?l=o`3o00co003o3?oo000<o`3o00Co003o2?ooool3ool000Ko003o0ooo000<o`3o
00go003o3?oo000=o`3o00Sooooo3?l00?l3ool000co0?l03?l00?l=ool000[o0?l02?ooool9o`00
o`?o0?l01_oo000<o`3o00co003o3Ooo0007o`3o00Wooooo0007ooooo`Koo`000ol00?l=o`3o00co
003o3?oo000<o`3o00Co003o2?ooool3ool000Ko003o0ooo000<o`3o00go003o3?oo000=o`3o00So
oooo3?l00?l3ool000co0?l03?l00?l=ool000[o0?l02?ooool9o`00o`?o0?l01_oo000<o`3o00co
003o3Ooo0007o`3o00Wooooo0007ooooo`?oo`002_l00?l6o`3o00co003o3?oo000<o`3o00Oo003o
2?ooool9o`00o`co0?l03Ol00?l<ool000co0?l01?l00?l8ooooo`Wo003o3ol0o`0<o`00o`goo`00
3Ol0o`08ooooo`Ko003o5Ol0o`0<o`00o`coo`002ol0o`09ooooo`001oooool3ool000[o003o1_l0
o`0<o`00o`coo`003?l0o`07o`00o`Sooooo2Ol00?l<o`3o00go003o3?oo000<o`3o00Co003o2?oo
ool9o`00o`oo0?l03?l00?l=ool000go0?l02?ooool6o`00oaGo0?l03?l00?l<ool000_o0?l02Ooo
ool000Oooooo0ooo000:o`00o`Ko0?l03?l00?l<ool000co0?l01ol00?l8ooooo`Wo003o3?l0o`0=
o`00o`coo`003?l0o`04o`00o`Sooooo2Ol00?l?o`3o00co003o3Ooo000=o`3o00Sooooo1_l00?lE
o`3o00co003o3?oo000;o`3o00Wooooo0007ooooo`?oo`006Ol00?l<ool000co0?l02_l00?l8oooo
o`co003o1_l0o`0=o`00o`coo`003?l0o`07o`00o`Sooooo2Ol00?l<o`3o00co003o3Ooo000<o`3o
00Co003o2?ooool6o`00oa;o0?l03?l00?l<ool000ko0?l02Oooool000Oooooo0ooo000Io`00o`co
o`003?l0o`0:o`00o`Sooooo3?l00?l6o`3o00go003o3?oo000<o`3o00Oo003o2?ooool9o`00o`co
0?l03?l00?l=ool000co0?l01?l00?l8ooooo`Ko003o4_l0o`0<o`00o`coo`003_l0o`09ooooo`00
1oooool3ool001Wo003o3?oo000<o`3o00[o003o2?ooool<o`00o`Ko0?l03Ol00?l<ool000co0?l0
1ol00?l8ooooo`Wo003o3?l0o`0<o`00o`goo`003?l0o`04o`00o`Sooooo1_l00?lBo`3o00co003o
3?oo000>o`3o00Wooooo0007ooooo`?oo`006Ol00?l<ool000co0?l02_l00?l8ooooo`co003o1_l0
o`0=o`00o`coo`003?l0o`07o`00o`Sooooo2Ol00?l<o`3o00co003o3Ooo000<o`3o00Co003o2?oo
ool6o`00oa;o0?l03?l00?l<ool000ko0?l02Oooool000Oooooo1_oo000Co`00o`coo`003?l0o`0=
o`00o`Sooooo0ooo000Io`00o`coo`003?l0o`0:o`00o`Sooooo3?l00?l6o`3o00co003o3Ooo000<
o`3o00Oo003o2?ooool9o`00o`co0?l03?l00?l<ool000go0?l01?l00?l9ooooo`001oooool6ool0
01?o003o3?oo000<o`3o00go003o2?ooool3ool001Wo003o3?oo000<o`3o00[o003o2?ooool<o`00
o`Ko0?l03?l00?l=ool000co0?l01ol00?l8ooooo`Wo003o3?l0o`0<o`00o`coo`003Ol0o`04o`00
o`Wooooo0007ooooo`Koo`004ol00?l<ool000co0?l03Ol00?l8ooooo`?oo`006Ol00?l<ool000co
0?l02_l00?l8ooooo`co003o1_l0o`0<o`00o`goo`003?l0o`07o`00o`Sooooo2Ol00?l<o`3o00co
003o3?oo000=o`3o00Co003o2Oooool000Oooooo1_oo000Co`00o`coo`003?l0o`0=o`00o`Sooooo
0ooo000Io`00o`coo`003?l0o`0:o`00o`Sooooo3?l00?l6o`3o00co003o3Ooo000<o`3o00Oo003o
2?ooool9o`00o`co0?l03?l00?l<ool000go0?l01?l00?l9ooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {213.312, 0}} -> {-0.102679, -0.0744342, \
0.0146528, 0.0146528}, {{7, 70.4375}, {208.188, 144.688}} -> {-0.171386, \
-0.139415, 0.0173401, 0.0173401}, {{76.8125, 140.312}, {208.188, 144.688}} -> \
{-1.38063, -0.138786, 0.017323, 0.017323}, {{146.625, 210.125}, {208.188, \
144.688}} -> {-2.58999, -0.138786, 0.017323, 0.017323}, {{216.5, 279.938}, \
{208.188, 144.688}} -> {-3.80413, -0.139415, 0.0173401, 0.0173401}, {{7, \
70.4375}, {138.375, 74.875}} -> {-0.171386, -1.34997, 0.0173401, 0.0173401}, \
{{76.8125, 140.312}, {138.375, 74.875}} -> {-1.38063, -1.34815, 0.017323, \
0.017323}, {{146.625, 210.125}, {138.375, 74.875}} -> {-2.58999, -1.34815, \
0.017323, 0.017323}, {{216.5, 279.938}, {138.375, 74.875}} -> {-3.80413, \
-1.34997, 0.0173401, 0.0173401}, {{7, 70.4375}, {68.5625, 5.0625}} -> \
{-0.171386, -2.56052, 0.0173401, 0.0173401}, {{76.8125, 140.312}, {68.5625, \
5.0625}} -> {-1.38063, -2.55751, 0.017323, 0.017323}, {{146.625, 210.125}, \
{68.5625, 5.0625}} -> {-2.58999, -2.55751, 0.017323, 0.017323}, {{216.5, \
279.938}, {68.5625, 5.0625}} -> {-3.80413, -2.56052, 0.0173401, 0.0173401}},
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell["Models should be deleted to free resources when done.", "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[BoxData[
    \(DeleteModel[m1, m2, m3, m4]\)], "Input",
  CellLabel->"In[21]:=",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1152}, {0, 700}},
WindowToolbars->{},
WindowSize->{587, 632},
WindowMargins->{{Automatic, 212}, {Automatic, 11}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:12"->{
    Cell[1776, 53, 89, 1, 38, "ObjectName",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[1868, 56, 392, 14, 32, "Usage",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[2263, 72, 124, 3, 22, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[2390, 77, 4263, 103, 569, "3ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[6656, 182, 449, 11, 52, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7108, 195, 285, 5, 36, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7396, 202, 246, 5, 36, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7645, 209, 100, 1, 22, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7782, 215, 87, 1, 24, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7872, 218, 276, 5, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8173, 227, 297, 8, 40, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8473, 237, 161, 3, 25, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8671, 245, 133, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8807, 250, 334, 6, 89, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[9156, 259, 252, 5, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[9411, 266, 587, 13, 73, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[10023, 283, 133, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[10159, 288, 1285, 22, 169, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11493, 316, 87, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11583, 319, 216, 4, 40, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11802, 325, 623, 14, 141, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[12462, 344, 90, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[12555, 347, 762, 20, 105, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[13354, 372, 85, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[13442, 375, 748, 26, 62, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[14227, 406, 89, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[14319, 409, 1084, 30, 123, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[15440, 444, 94, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[15537, 447, 970, 21, 164, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16544, 473, 92, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16639, 476, 288, 7, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16930, 485, 341, 12, 24, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[17274, 499, 630, 22, 56, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[17907, 523, 642, 20, 56, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[18552, 545, 664, 25, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19219, 572, 582, 21, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19838, 598, 80, 1, 26, "SubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19921, 601, 118, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20042, 604, 211, 4, 42, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20256, 610, 234, 5, 40, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20493, 617, 292, 7, 74, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20788, 626, 103, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20916, 631, 121, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[21040, 636, 75284, 1564, 294, 9130, 742, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96339, 2203, 97, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96439, 2206, 371, 8, 90, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96835, 2218, 121, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96959, 2223, 28938, 775, 294, 6246, 489, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[125912, 3001, 179, 4, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[126094, 3007, 1045, 21, 196, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[127164, 3032, 154, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[127321, 3037, 62841, 2729, 322, 28970, 2305, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190177, 5769, 99, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190279, 5772, 139, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190421, 5777, 145, 3, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190569, 5782, 302, 6, 90, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190896, 5792, 227, 4, 58, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[191126, 5798, 368140, 26126, 221, 330998, 25662, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[559281, 31927, 115, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[559399, 31930, 131, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}]},
  "LocalInteractions:21"->{
    Cell[1776, 53, 89, 1, 38, "ObjectName",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[1868, 56, 392, 14, 32, "Usage",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[2263, 72, 124, 3, 22, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[2390, 77, 4263, 103, 569, "3ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[6656, 182, 449, 11, 52, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7108, 195, 285, 5, 36, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7396, 202, 246, 5, 36, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7645, 209, 100, 1, 22, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7782, 215, 87, 1, 24, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[7872, 218, 276, 5, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8173, 227, 297, 8, 40, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8473, 237, 161, 3, 25, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8671, 245, 133, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[8807, 250, 334, 6, 89, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[9156, 259, 252, 5, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[9411, 266, 587, 13, 73, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[10023, 283, 133, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[10159, 288, 1285, 22, 169, "Output",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11493, 316, 87, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11583, 319, 216, 4, 40, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[11802, 325, 623, 14, 141, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[12462, 344, 90, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[12555, 347, 762, 20, 105, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[13354, 372, 85, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[13442, 375, 748, 26, 62, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[14227, 406, 89, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[14319, 409, 1084, 30, 123, "2ColumnTable",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[15440, 444, 94, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[15537, 447, 970, 21, 164, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16544, 473, 92, 1, 20, "SubsubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16639, 476, 288, 7, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[16930, 485, 341, 12, 24, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[17274, 499, 630, 22, 56, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[17907, 523, 642, 20, 56, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[18552, 545, 664, 25, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19219, 572, 582, 21, 40, "Notes",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19838, 598, 80, 1, 26, "SubsectionIcon",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[19921, 601, 118, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20042, 604, 211, 4, 42, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20256, 610, 234, 5, 40, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20493, 617, 292, 7, 74, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20788, 626, 103, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[20916, 631, 121, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[21040, 636, 75284, 1564, 294, 9130, 742, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96339, 2203, 97, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96439, 2206, 371, 8, 90, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96835, 2218, 121, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[96959, 2223, 28938, 775, 294, 6246, 489, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[125912, 3001, 179, 4, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[126094, 3007, 1045, 21, 196, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[127164, 3032, 154, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[127321, 3037, 62841, 2729, 322, 28970, 2305, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190177, 5769, 99, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190279, 5772, 139, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190421, 5777, 145, 3, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190569, 5782, 302, 6, 90, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[190896, 5792, 227, 4, 58, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[191126, 5798, 368140, 26126, 221, 330998, 25662, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[559281, 31927, 115, 1, 24, "Text",
      CellTags->{"b:12", "LocalInteractions:21"}],
    Cell[559399, 31930, 131, 3, 26, "Input",
      CellTags->{"b:12", "LocalInteractions:21"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:12", 560187, 31953},
  {"LocalInteractions:21", 566170, 32078}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 89, 1, 38, "ObjectName",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[1868, 56, 392, 14, 32, "Usage",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[2263, 72, 124, 3, 22, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[2390, 77, 4263, 103, 569, "3ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[6656, 182, 449, 11, 52, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[7108, 195, 285, 5, 36, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[7396, 202, 246, 5, 36, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[7645, 209, 100, 1, 22, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[7782, 215, 87, 1, 24, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[7872, 218, 276, 5, 40, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[8173, 227, 297, 8, 40, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[8473, 237, 161, 3, 25, "Output",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[8671, 245, 133, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[8807, 250, 334, 6, 89, "Output",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],
Cell[9156, 259, 252, 5, 40, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[9411, 266, 587, 13, 73, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[10023, 283, 133, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[10159, 288, 1285, 22, 169, "Output",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[11493, 316, 87, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[11583, 319, 216, 4, 40, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[11802, 325, 623, 14, 141, "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[12462, 344, 90, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[12555, 347, 762, 20, 105, "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[13354, 372, 85, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[13442, 375, 748, 26, 62, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[14227, 406, 89, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[14319, 409, 1084, 30, 123, "2ColumnTable",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[15440, 444, 94, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[15537, 447, 970, 21, 164, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[16544, 473, 92, 1, 20, "SubsubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[16639, 476, 288, 7, 40, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[16930, 485, 341, 12, 24, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[17274, 499, 630, 22, 56, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[17907, 523, 642, 20, 56, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[18552, 545, 664, 25, 40, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[19219, 572, 582, 21, 40, "Notes",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]],

Cell[CellGroupData[{
Cell[19838, 598, 80, 1, 26, "SubsectionIcon",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[19921, 601, 118, 1, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[20042, 604, 211, 4, 42, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[20256, 610, 234, 5, 40, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[20493, 617, 292, 7, 74, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[20788, 626, 103, 1, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[20916, 631, 121, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[21040, 636, 75284, 1564, 294, 9130, 742, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],
Cell[96339, 2203, 97, 1, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[96439, 2206, 371, 8, 90, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[96835, 2218, 121, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[96959, 2223, 28938, 775, 294, 6246, 489, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],
Cell[125912, 3001, 179, 4, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[126094, 3007, 1045, 21, 196, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[127164, 3032, 154, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[127321, 3037, 62841, 2729, 322, 28970, 2305, "GraphicsData", \
"PostScript", "Graphics",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],
Cell[190177, 5769, 99, 1, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[190279, 5772, 139, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[190421, 5777, 145, 3, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[190569, 5782, 302, 6, 90, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],

Cell[CellGroupData[{
Cell[190896, 5792, 227, 4, 58, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[191126, 5798, 368140, 26126, 221, 330998, 25662, "GraphicsData", \
"PostScript", "Graphics",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Open  ]],
Cell[559281, 31927, 115, 1, 24, "Text",
  CellTags->{"b:12", "LocalInteractions:21"}],
Cell[559399, 31930, 131, 3, 26, "Input",
  CellTags->{"b:12", "LocalInteractions:21"}]
}, Closed]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

