(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    102239,       2666]*)
(*NotebookOutlinePosition[    106829,       2775]*)
(*  CellTagsIndexPosition[    106715,       2768]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["WuenscheTree", "ObjectName",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[TextData[{
  "\[FilledSmallSquare] ",
  StyleBox["WuenscheTree[",
    FontFamily->"Courier"],
  Cell[BoxData[
      \(TraditionalForm\`n\_1, \[Ellipsis], n\_k\)]],
  StyleBox["]",
    FontFamily->"Courier"],
  " creates a set of random trees of size ",
  Cell[BoxData[
      \(TraditionalForm\`n\_1\)]],
  " to ",
  Cell[BoxData[
      \(TraditionalForm\`n\_k\)]],
  ", linking the roots of the ",
  Cell[BoxData[
      \(TraditionalForm\`k\)]],
  " trees to a unique central vertex.\n\[FilledSmallSquare] ",
  StyleBox["WuenscheTree[",
    FontFamily->"Courier"],
  Cell[BoxData[
      \(TraditionalForm\`g\_1, \[Ellipsis], g\_k\)]],
  StyleBox["]",
    FontFamily->"Courier"],
  " creates a set of random trees of size ",
  Cell[BoxData[
      \(TraditionalForm\`n\_1\)]],
  " to ",
  Cell[BoxData[
      \(TraditionalForm\`n\_k\)]],
  ", linking the roots of the ",
  Cell[BoxData[
      \(TraditionalForm\`k\)]],
  " trees to a unique central vertex."
}], "Usage",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell["\<\
\[FilledSmallSquare] The command accepts the following \
option:\
\>", "Notes",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[BoxData[GridBox[{
        {"UniqueCenter", "True", 
          
          StyleBox[\(include\ a\ unique\ central\ vertex\ to\ which\ the\ \
roots\ of\ all\ the\ trees\ connect\),
            FontFamily->"Times"]}
        }]], "3ColumnTable",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[TextData[{
  "\[FilledSmallSquare] ",
  StyleBox["WuenscheTree",
    FontFamily->"Courier"],
  " uses the default setting ",
  StyleBox["UniqueCenter->True",
    FontFamily->"Courier"],
  "."
}], "Notes",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{

Cell["Examples", "ExampleSection",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell["Here is a simple Wuensche tree with a unique central vertex.", "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(WuenscheTree[21, 17, 13, 24] // ShowGraph;\)\)], "Input",
  CellLabel->"In[15]:=",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.256041 1.19623 0.0454932 0.909863 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.51153 .62931 m
.56002 .6662 L
s
.51153 .62931 m
.46303 .6662 L
s
.51153 .62931 m
.46303 .59243 L
s
.51153 .62931 m
.56002 .59243 L
s
.66247 .80114 m
.62437 .75776 L
s
.79556 .76963 m
.84997 .80139 L
s
.79556 .76963 m
.74115 .73788 L
s
.66257 .9145 m
.63946 .86539 L
s
.65613 .91599 m
.63946 .86539 L
s
.84997 .80139 m
.90438 .83314 L
s
.62437 .75776 m
.58627 .71439 L
s
.75474 .71293 m
.68674 .70612 L
s
.61955 .81547 m
.59964 .76555 L
s
.61955 .81547 m
.63946 .86539 L
s
.7501 .72608 m
.68674 .70612 L
s
.7501 .72608 m
.81347 .74605 L
s
.68674 .70612 m
.62338 .68616 L
s
.68674 .70612 m
.74115 .73788 L
s
.62338 .68616 m
.56002 .6662 L
s
.60637 .81746 m
.59964 .76555 L
s
.59964 .76555 m
.58627 .71439 L
s
.58627 .71439 m
.56002 .6662 L
s
.94019 .78597 m
.87683 .76601 L
s
.81347 .74605 m
.87683 .76601 L
s
.2608 .69332 m
.32879 .68651 L
s
.36604 .73997 m
.41454 .70308 L
s
.36604 .73997 m
.31755 .77685 L
s
.06709 .95451 m
.11765 .91927 L
s
.33342 .69966 m
.39679 .6797 L
s
.32879 .68651 m
.39679 .6797 L
s
.06324 .95102 m
.11765 .91927 L
s
.44528 .71658 m
.46303 .6662 L
s
.41454 .70308 m
.46303 .6662 L
s
.39679 .6797 m
.46303 .6662 L
s
.26905 .81374 m
.22056 .85062 L
s
.26905 .81374 m
.31755 .77685 L
s
.22056 .85062 m
.17206 .88751 L
s
.13031 .92889 m
.17206 .88751 L
s
.05981 .94729 m
.11765 .91927 L
s
.11765 .91927 m
.17206 .88751 L
s
.31755 .48177 m
.26314 .45001 L
s
.31755 .48177 m
.36604 .51865 L
s
.31755 .48177 m
.2758 .44038 L
s
.26314 .45001 m
.2049 .42247 L
s
.26314 .45001 m
.20739 .41963 L
s
.26314 .45001 m
.21012 .41692 L
s
.26314 .45001 m
.21308 .41435 L
s
.34829 .54204 m
.28205 .52854 L
s
.34829 .54204 m
.41454 .55554 L
s
.36604 .51865 m
.41454 .55554 L
s
.39679 .50515 m
.41454 .55554 L
s
.41454 .55554 m
.46303 .59243 L
s
.7055 .48177 m
.754 .44488 L
s
.7055 .48177 m
.65701 .51865 L
s
.60851 .55554 m
.62627 .50515 L
s
.60851 .55554 m
.65701 .51865 L
s
.60851 .55554 m
.67476 .54204 L
s
.60851 .55554 m
.56002 .59243 L
s
.72878 .09567 m
.7154 .14683 L
s
.65297 .40305 m
.66192 .35133 L
s
.65297 .40305 m
.64402 .45477 L
s
.67386 .2475 m
.67087 .29961 L
s
.754 .44488 m
.80249 .408 L
s
.89948 .33423 m
.85099 .37111 L
s
.85099 .37111 m
.80249 .408 L
s
.67982 .2479 m
.67087 .29961 L
s
.67982 .2479 m
.68877 .19618 L
s
.70056 .19776 m
.68571 .24869 L
s
.70056 .19776 m
.7154 .14683 L
s
.62627 .50515 m
.64402 .45477 L
s
.68571 .24869 m
.67087 .29961 L
s
.66192 .35133 m
.67087 .29961 L
s
.67026 .40657 m
.64402 .45477 L
s
.7317 .09616 m
.748 .04549 L
s
.7317 .09616 m
.7154 .14683 L
s
.025 w
.51153 .62931 Mdot
.66247 .80114 Mdot
.79556 .76963 Mdot
.66257 .9145 Mdot
.65613 .91599 Mdot
.84997 .80139 Mdot
.62437 .75776 Mdot
.75474 .71293 Mdot
.61955 .81547 Mdot
.90438 .83314 Mdot
.7501 .72608 Mdot
.68674 .70612 Mdot
.62338 .68616 Mdot
.60637 .81746 Mdot
.59964 .76555 Mdot
.58627 .71439 Mdot
.94019 .78597 Mdot
.81347 .74605 Mdot
.87683 .76601 Mdot
.63946 .86539 Mdot
.74115 .73788 Mdot
.56002 .6662 Mdot
.2608 .69332 Mdot
.36604 .73997 Mdot
.06709 .95451 Mdot
.33342 .69966 Mdot
.32879 .68651 Mdot
.06324 .95102 Mdot
.44528 .71658 Mdot
.41454 .70308 Mdot
.39679 .6797 Mdot
.26905 .81374 Mdot
.22056 .85062 Mdot
.13031 .92889 Mdot
.05981 .94729 Mdot
.31755 .77685 Mdot
.11765 .91927 Mdot
.17206 .88751 Mdot
.46303 .6662 Mdot
.31755 .48177 Mdot
.26314 .45001 Mdot
.2049 .42247 Mdot
.20739 .41963 Mdot
.34829 .54204 Mdot
.36604 .51865 Mdot
.39679 .50515 Mdot
.21012 .41692 Mdot
.21308 .41435 Mdot
.28205 .52854 Mdot
.2758 .44038 Mdot
.41454 .55554 Mdot
.46303 .59243 Mdot
.7055 .48177 Mdot
.60851 .55554 Mdot
.72878 .09567 Mdot
.65297 .40305 Mdot
.67386 .2475 Mdot
.754 .44488 Mdot
.89948 .33423 Mdot
.85099 .37111 Mdot
.67982 .2479 Mdot
.70056 .19776 Mdot
.62627 .50515 Mdot
.68571 .24869 Mdot
.66192 .35133 Mdot
.80249 .408 Mdot
.65701 .51865 Mdot
.67026 .40657 Mdot
.7317 .09616 Mdot
.67087 .29961 Mdot
.748 .04549 Mdot
.64402 .45477 Mdot
.7154 .14683 Mdot
.67476 .54204 Mdot
.68877 .19618 Mdot
.56002 .59243 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[15]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00eOooool3o`0004Sooooo003Dooooo`Go0000Aoooool00=?ooooo1ol00016ooooo`00dooo
ool7o`0004Kooooo003Cooooo`Oo0000A_ooool00=Cooooo1Ol00017ooooo`00eOooool3o`0004So
oooo003Eooooo`03o`000?oooooooooo04Sooooo003Dooooo`03o`000?oooooooooo04Wooooo003D
ooooo`03o`000?oooooooooo04Wooooo003Dooooo`03o`000?oooooooooo04Wooooo003Cooooo`03
o`000?oooooooooo04[ooooo003Cooooo`03o`000?oooooooooo04[ooooo003Cooooo`03o`000?oo
oooooooo04[ooooo003@ooooo`?o0000COooool00<oooooo1Ol0001<ooooo`00c_ooool7o`0004_o
oooo003>ooooo`Oo0000Boooool00<kooooo1ol0001;ooooo`00coooool5o`0004cooooo003@oooo
o`?o0000COooool00=3ooooo00?o0000ooooooooool0COooool00=3ooooo00?o0000ooooooooool0
COooool00=3ooooo00?o0000ooooooooool0COooool00<oooooo00?o0000ooooooooool0C_ooool0
0<oooooo00?o0000ooooooooool0C_ooool00<oooooo00?o0000ooooooooool0C_ooool00<kooooo
00?o0000ooooooooool0Coooool00<kooooo00?o0000ooooooooool0Coooool00<cooooo0ol0001A
ooooo`00boooool5o`00053ooooo003:ooooo`Oo0000Coooool00<[ooooo1ol0001?ooooo`00b_oo
ool7o`0004oooooo003;ooooo`Go0000D?ooool00<cooooo0ol0001Aooooo`00c?ooool00ol0003o
ooooooooo`1Aooooo`00c?ooool00ol0003oooooooooo`1Aooooo`00boooool00ol0003ooooooooo
o`1Booooo`00boooool00ol0003oooooooooo`1Booooo`00boooool00ol0003oooooooooo`1Boooo
o`00boooool00ol0003oooooooooo`1Booooo`00b_ooool00ol0003oooooooooo`1Cooooo`00a?oo
ool7o`0005Gooooo0033ooooo`So0000EOooool00<;ooooo2_l0001Dooooo`00`_ooool:o`0005Co
oooo0032ooooo`[o0000E?ooool00<?ooooo2?l0001Eooooo`00a?ooool6o`0005Kooooo0035oooo
o`04o`000?ooooooooooo`0005Oooooo0034ooooo`04o`000?ooooooooooo`0005Sooooo0034oooo
o`04o`000?ooooooooooo`0005Sooooo0034ooooo`04o`000?ooooooooooo`0005Sooooo0034oooo
o`03o`000?oooooo000005Wooooo0034ooooo`03o`000?oooooo000005Wooooo0034ooooo`03o`00
0?oooooo000005Wooooo0034ooooo`;o0000F_ooool00<3ooooo1_l0001Jooooo`00_oooool8o`00
05Wooooo002nooooo`[o0000F?ooool00;kooooo2_l0001Hooooo`00__ooool:o`0005Sooooo002o
ooooo`So0000FOooool00<3ooooo1_l0001Jooooo`00`Oooool3o`0005cooooo0031ooooo`?o0000
G?ooool00<7ooooo0_l0001Mooooo`00`Oooool2o`0005gooooo0030ooooo`?o0000GOooool00<3o
oooo0ol0001Mooooo`00`?ooool2o`0005kooooo0030ooooo`;o0000G_ooool00;oooooo0ol0001N
ooooo`00__ooool5o`0005gooooo002mooooo`Oo0000G?ooool00;gooooo1ol0001Looooo`00_Ooo
ool7o`0005cooooo002nooooo`Go0000GOooool00;oooooo0ol0001Nooooo`00_oooool00ol0003o
ooooooooo`1Nooooo`00_oooool00ol0003oooooooooo`1Nooooo`00_oooool00ol0003ooooooooo
o`0oooooo`?o00007?ooool00;kooooo00?o0000ooooooooool0?oooool5o`0001_ooooo002noooo
o`03o`000?oooooooooo03kooooo1ol0000Jooooo`00__ooool00ol0003oooooooooo`0nooooo`Oo
00006_ooool00;kooooo00?o0000ooooooooool0?_ooool7o`0001[ooooo002looooo`?o0000@?oo
ool6o`0001_ooooo002kooooo`Go0000?_ooool00ol0003oooooooooo`03o`0001cooooo002joooo
o`Oo0000>oooool2o`0002;ooooo002jooooo`Oo0000>_ooool00ol0003oooooooooo`0Rooooo`00
^_ooool7o`0003Wooooo00?o0000ooooooooool08oooool00;_ooooo1Ol0000hooooo`;o00009_oo
ool00;cooooo0ol0000dooooo`?o000000?oooooo`000?ooool09oooool00;gooooo00?o0000oooo
ooooool0<_ooool5o`0002Wooooo002looooo`03o`000?oooooooooo03;ooooo1ol0000Xooooo`00
_?ooool00ol0003oooooooooo`0booooo`Oo0000:?ooool00;cooooo00?o0000ooooooooool0<_oo
ool7o`0002Sooooo002looooo`03o`000?oooooooooo03;ooooo1_l0000Yooooo`00_?ooool00ol0
003oooooooooo`0aooooo`03o`000?oooooooooo00?o0000:_ooool00;cooooo00?o0000oooooooo
ool0;oooool2o`00033ooooo002looooo`03o`000?oooooooooo02kooooo00?o0000ooooooooool0
<?ooool00;[ooooo0ol0000_ooooo`03o`000?oooooooooo037ooooo002iooooo`Go000000Cooooo
o`000?l0003o0000:?ooool2o`0003Cooooo002hooooo`_o00008_ooool3o`000003ooooool0003o
oooo03Gooooo000looooo`?o0000NOooool<o`00023ooooo1Ol0000gooooo`00>oooool5o`0007So
oooo3?l0000Oooooo`Oo0000=_ooool003[ooooo1ol0001hooooo`_o00007oooool7o`0003Kooooo
000iooooo`So0000NOooool3o`000003ooooool0003o000000?o00008?ooool7o`0003Kooooo000h
ooooo`Wo0000NOooool00ol0003oooooooooo`02ooooo`?o00008?ooool7o`0003Oooooo000goooo
o`[o0000NOooool01Ol0003ooooooooooooooooo000002;ooooo00Co0000oooooooooooooooo0ol0
000hooooo`00=oooool;o`0007Sooooo00Co0000oooooooooooo00008_ooool00ol0003ooooooooo
o`0mooooo`00=oooool7o`000003ooooool0003ooooo00?o0000M_ooool01?l0003oooooooooool0
000Pooooo`;o0000@?ooool003Sooooo1Ol000000ooooooo0000o`000005o`0000Wooooo0ol0001X
ooooo`04o`000?ooooooooooo`00023ooooo00?o0000ooooooooool0@?ooool003Wooooo0ol00004
ooooo`Ko00001oooool5o`0006Oooooo00Co0000oooooooooooo00006_ooool3o`000003ooooool0
003o000004?ooooo0012ooooo`Ko00001?ooool7o`0006Kooooo00?o0000ooooool000006_ooool5
o`0004Gooooo0015ooooo`Co000000?oooooo`000?l000001ol0001Vooooo`03o`000?oooooo0000
01Wooooo1ol00014ooooo`00Aoooool<o`0006Cooooo1?l0000Jooooo`Oo0000A?ooool004Sooooo
2_l0001Tooooo`Go00006_ooool7o`0004Cooooo0018ooooo`Wo000000?oooooo`000?ooool0HOoo
ool7o`0001Wooooo1_l00015ooooo`00B?ooool7o`0000Cooooo00?o0000ooooooooool0Goooool7
o`0001Sooooo00?o0000ooooooooool00ol00016ooooo`00BOooool5o`000003ooooool0003ooooo
00?ooooo00?o0000ooooooooool0G_ooool7o`0001Kooooo0_l0001<ooooo`00B_ooool3o`0000?o
oooo0_l00003ooooo`03o`000?oooooooooo05kooooo1Ol0000Fooooo`03o`000?oooooooooo04co
oooo001Booooo`;o00000_ooool00ol0003oooooooooo`1Nooooo`?o00005_ooool00ol0003ooooo
ooooo`1=ooooo`00E?ooool2o`000003ooooool0003ooooo05kooooo00?o0000ooooooooool05?oo
ool2o`00053ooooo001Fooooo`?o000000Coooooo`000?l0003o0000FOooool00ol0003ooooooooo
o`0@ooooo`?o000000?oooooo`000?ooool0DOooool005Sooooo1_l0001Hooooo`03o`000?oooooo
oooo00oooooo1Ol0001Cooooo`00F?ooool7o`0005Oooooo00?o0000ooooooooool03_ooool7o`00
05;ooooo001Hooooo`Oo0000E_ooool00ol0003oooooooooo`0?ooooo`Oo0000D_ooool005Sooooo
1ol0001Fooooo`03o`000?oooooooooo00oooooo1ol0001Booooo`00FOooool7o`0005Gooooo00?o
0000ooooooooool03_ooool7o`0005?ooooo001Jooooo`?o00000oooool00ol0003oooooooooo`1A
ooooo`03o`000?oooooooooo00kooooo00Co0000oooooooooooooooo0ol0001Dooooo`00HOooool0
0ol0003oooooooooo`0<ooooo`?o0000?oooool3o`0000oooooo00?o0000ooooooooool0FOooool0
06;ooooo0_l0000;ooooo`Go0000?Oooool5o`0000cooooo0_l0001Looooo`00I?ooool00ol0003o
ooooooooo`07ooooo`Oo0000>oooool7o`0000[ooooo00?o0000ooooooooool0G?ooool006Gooooo
1Ol00004ooooo`Oo0000>oooool7o`0000Cooooo0ol000000ooooooo0000o`00001Oooooo`00I_oo
ool5o`0000?ooooo1ol0000kooooo`Oo00000oooool5o`00067ooooo001Uooooo`Oo00000oooool5
o`0003gooooo1Ol00003ooooo`Oo0000H?ooool004oooooo0ol0000Cooooo`Oo00001?ooool3o`00
03oooooo0ol00004ooooo`Oo0000H?ooool004kooooo1Ol0000Booooo`Oo00001oooool00ol0003o
ooooooooo`0kooooo`03o`000?oooooooooo00Gooooo1ol0001Pooooo`00COooool7o`0001;ooooo
1Ol000000ooooooo0000ooooo`05ooooo`03o`000?oooooooooo03_ooooo00?o0000ooooooooool0
1Oooool6o`00067ooooo001=ooooo`Oo00004oooool3o`0000?ooooo00?o0000ooooooooool00ooo
ool00ol0003oooooooooo`0kooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00ol0
001Rooooo`00COooool;o`0000[ooooo0ol00009ooooo`03o`000?oooooooooo00?ooooo00?o0000
ooooooooool0>Oooool00ol0003oooooooooo`03ooooo`;o00002?ooool3o`0005gooooo001>oooo
o`Go00001Oooool5o`0000Cooooo1Ol00009ooooo`;o00000oooool00ol0003oooooooooo`0ioooo
o`03o`000?oooooooooo00;ooooo00?o0000ooooooooool01oooool5o`0005cooooo001?ooooo`?o
00002oooool:o`0000[ooooo00Co0000oooooooooooo0000>oooool01Ol0003ooooooooooooooooo
000000Wooooo1ol0001Kooooo`00H?ooool7o`0000_ooooo00Co0000oooooooooooo0000>Oooool0
0ol0003oooooooooo`02o`0000[ooooo1ol0001Kooooo`00H?ooool;o`0000Sooooo00?o0000oooo
ool000000_l0000eooooo`?o000000?oooooo`000?ooool01oooool;o`0005_ooooo001Qooooo`Go
00001Oooool5o`0000Cooooo1Ol0000cooooo`Go00001?ooool5o`0000Gooooo1Ol0001Looooo`00
H_ooool3o`0000_ooooo2_l0000aooooo`[o00002oooool3o`0005gooooo001cooooo`Oo0000<Ooo
ool7o`0006kooooo001cooooo`Oo0000<Oooool7o`0006kooooo001dooooo`Ko0000<Oooool6o`00
06oooooo001eooooo`?o00000_ooool00ol0003oooooooooo`0]ooooo`03o`000?oooooooooo00?o
0000L?ooool007_ooooo0_l0000[ooooo`;o0000M_ooool007gooooo00?o0000ooooooooool09ooo
ool00ol0003oooooooooo`1fooooo`00O_ooool00ol0003oooooooooo`0Uooooo`03o`000?oooooo
oooo07Oooooo001oooooo`;o00008oooool2o`0007[ooooo0021ooooo`03o`000?oooooo000000;o
00006Oooool3o`000003ooooool0003ooooo07_ooooo0022ooooo`Go00005oooool5o`0007gooooo
0021ooooo`Oo00005Oooool7o`0007cooooo0021ooooo`Oo00005Oooool7o`0007cooooo0021oooo
o`Oo00005Oooool7o`0007cooooo0022ooooo`Oo00004oooool7o`0007gooooo0023ooooo`?o0000
0oooool00ol0003oooooooooo`0?ooooo`04o`000?ooooooooooooooo`?o0000O_ooool008[ooooo
00?o0000ooooooooool03Oooool00ol0003oooooooooo`23ooooo`00Roooool2o`0000_ooooo0_l0
0026ooooo`00SOooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo08Kooooo002>oooo
o`;o000000Goooooo`000?l0003o0000ooooo`02o`0008Wooooo002@ooooo`Go0000Roooool008oo
oooo1ol0002:ooooo`00Soooool7o`0008[ooooo002?ooooo`Oo0000R_ooool008oooooo1ol0002:
ooooo`00S_ooool00ol0003oooooooooo`03o`0000;ooooo00?o0000ooooooooool0Qoooool008co
oooo0_l00009ooooo`;o0000Qoooool008_ooooo00?o0000ooooooooool02oooool00ol0003ooooo
ooooo`24ooooo`00R_ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo08?ooooo0028
ooooo`;o00004Oooool2o`0008?ooooo0023ooooo`?o000000?oooooo`000?ooool05?ooool00ol0
003oooooo`000002o`0007kooooo0022ooooo`Go00005oooool5o`0007gooooo0021ooooo`Oo0000
5Oooool7o`0007cooooo0021ooooo`Oo00005Oooool7o`0007cooooo001`ooooo`?o00002_ooool;
o`0001Gooooo2Ol0001jooooo`00Koooool5o`0000Cooooo1Ol00003ooooo`Oo00005oooool5o`00
00?ooooo1?l00007ooooo`?o0000K?ooool005gooooo0ol0000>ooooo`[o00001oooool01?l0003o
ooooooooooooool3o`0001Wooooo0ol00008ooooo`Co00000_ooool5o`0006_ooooo001Looooo`Go
00002oooool9o`0000Wooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0Looooo`03
o`000?oooooooooo00Wooooo2?l0001Zooooo`00BOooool3o`0000oooooo1ol000000ooooooo0000
o`00000@o`0000Oooooo0_l00004ooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool0
2Oooool7o`0006[ooooo0018ooooo`Go00001oooool?o`0000Gooooo1?l00003ooooo`Go00001ooo
ool00ol0003oooooooooo`04ooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool02Ooo
ool9o`0006Sooooo0017ooooo`go00001oooool7o`0000;ooooo1?l00008ooooo`?o00000_ooool3
o`000003ooooool0003o000000Oooooo00?o0000ooooooooool07oooool00ol0003oooooooooo`09
ooooo`Go00000oooool3o`0006Gooooo0017ooooo`Oo00003_ooool8o`00013ooooo1Ol00008oooo
o`03o`000?oooooooooo023ooooo00?o0000ooooooooool02_ooool3o`0000Oooooo0ol00006oooo
o`?o0000FOooool004Oooooo1ol0000>ooooo`Oo00004?ooool7o`0000Oooooo00?o0000oooooooo
ool08Oooool00ol0003oooooooooo`0Fooooo`?o00000_ooool5o`0005Sooooo0018ooooo`Go0000
3oooool7o`00013ooooo1ol00007ooooo`03o`000?oooooooooo027ooooo00?o0000ooooooooool0
6Oooool8o`0000kooooo0ol00016ooooo`00BOooool3o`00017ooooo1Ol0000Aooooo`Oo00001?oo
ool3o`0002Gooooo1?l0000Hooooo`Wo00002oooool5o`0004Gooooo001Nooooo`?o00004_ooool6
o`0000Cooooo1Ol0000Tooooo`Go00005ooooolBo`000003ooooool0003o000000Go0000A?ooool0
07;ooooo00?o0000ooooooooool00ol00004ooooo`Oo00008_ooool7o`0001Oooooo1Ol000000ooo
oooo0000ooooo`03o`0000Kooooo2?l00014ooooo`00LOooool00ol0003oooooooooo`08ooooo`Oo
00008_ooool7o`0001Sooooo0ol00003ooooo`;o00000_ooool3o`0000Cooooo1ol00014ooooo`00
Koooool2o`0000_ooooo1ol0000Rooooo`Oo00008?ooool2o`0000?ooooo0ol000000ooooooo0000
o`000004o`0004Gooooo001^ooooo`03o`000?oooooooooo00cooooo1Ol0000Tooooo`Go00008ooo
ool00ol0003oooooooooo`03ooooo`Oo0000AOooool006gooooo00?o0000ooooooooool03_ooool3
o`0002Kooooo1?l0000Tooooo`;o00000_ooool8o`0004Gooooo001Wooooo`?o00000_ooool00ol0
003oooooooooo`0jooooo`03o`000?oooooo000002Gooooo3?l00013ooooo`00I_ooool6o`0003go
oooo00Co0000oooooooooooo00009?ooool9o`0000?ooooo0ol00010ooooo`00IOooool7o`0003ko
oooo00Co0000oooooooooooo00008oooool8o`0000Oooooo0ol00005ooooo`?o0000=Oooool006Go
oooo1ol0000nooooo`05o`000?ooooooooooooooool000008_ooool7o`0000_ooooo0ol000000ooo
oooo0000o`000003o`0003Cooooo001Uooooo`Oo0000?_ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo023ooooo1Ol000000ooooooo0000ooooo`0<ooooo`Oo0000<oooool006Gooooo
1_l0000oooooo`03o`000?oooooooooo00?ooooo00?o0000ooooool000000_l0000Nooooo`?o0000
0oooool2o`0000_ooooo1ol0000cooooo`00I?ooool00ol0003oooooooooo`03o`00047ooooo00?o
0000ooooooooool00oooool5o`0002Gooooo0_l00009ooooo`Wo0000<Oooool006?ooooo00?o0000
ooooooooool0AOooool3o`0000;ooooo1ol0000Vooooo`;o00002?ooool5o`0000?ooooo1?l00007
ooooo`?o00008oooool0067ooooo0_l00017ooooo`Go000000?oooooo`000?l000001Ol0000Xoooo
o`;o00000_ooool3o`0000;ooooo0ol00008ooooo`Co00000_ooool5o`0002;ooooo001Pooooo`03
o`000?oooooooooo04Kooooo3_l0000Zooooo`Ko00004?ooool8o`00027ooooo001Nooooo`;o0000
BOooool7o`000003ooooool0003o000000?o0000:oooool7o`00013ooooo1ol0000Qooooo`00F_oo
ool4o`0004_ooooo1ol00002ooooo`Co0000:oooool7o`00013ooooo2Ol0000Oooooo`00FOooool5
o`0004cooooo1Ol00007ooooo`03o`000?oooooooooo02Sooooo1ol0000Aooooo`Go00000oooool3
o`0001cooooo001Hooooo`Oo0000C?ooool3o`0000Wooooo00?o0000ooooooooool0:?ooool5o`00
0003ooooool0003ooooo013ooooo0ol00007ooooo`?o00001Oooool3o`00017ooooo001Hooooo`Oo
0000COooool2o`0000[ooooo00?o0000ooooooooool0:?ooool3o`0000?ooooo0_l0000Looooo`?o
000000?oooooo`000?l000000ol0000@ooooo`00F?ooool7o`0004kooooo0_l0000:ooooo`03o`00
0?oooooooooo02oooooo0_l0000Mooooo`Oo00003oooool005Oooooo00?o0000ooooool000001?l0
001?ooooo`;o00002oooool00ol0003oooooooooo`0`ooooo`03o`000?oooooooooo01[ooooo1ol0
000?ooooo`00EOooool2o`0000?ooooo0ol0001@ooooo`;o00003?ooool00ol0003oooooo`000002
o`0002kooooo0_l0000Jooooo`Oo00003oooool005Cooooo00?o0000ooooooooool0E_ooool00ol0
003oooooo`00000<ooooo`Go0000;oooool5o`0001Kooooo1Ol0000@ooooo`00Doooool00ol0003o
ooooooooo`1Gooooo`03o`000?oooooo000000_ooooo1ol0000_ooooo`Go00005_ooool3o`00017o
oooo001Aooooo`;o0000F_ooool00ol0003oooooo`00000;ooooo`Oo0000;_ooool7o`0002Wooooo
001<ooooo`?o000000?oooooo`000?ooool0Foooool01?l0003oooooooooool0000:ooooo`Oo0000
;_ooool7o`0002Wooooo001;ooooo`Go0000G?ooool3o`000004ooooool0003o0000o`0000Wooooo
1Ol0000_ooooo`So0000:?ooool004[ooooo1ol0001Jooooo`Wo00002Oooool3o`00037ooooo1Ol0
0002ooooo`;o00009_ooool004[ooooo1ol0001Iooooo`_o0000?Oooool3o`0000Gooooo00?o0000
ooooooooool08oooool004[ooooo1ol0001Iooooo`_o0000A_ooool2o`0002?ooooo001:ooooo`Ko
0000F_ooool;o`0004Sooooo00?o0000ooooooooool08?ooool004Wooooo00?o0000ooooooooool0
0ol0001Looooo`Wo0000B_ooool2o`0000;ooooo0ol0000Kooooo`00Aoooool2o`0006?ooooo0ol0
00001?oooooo0000o`000?l0001=ooooo`Ko00006_ooool004Kooooo00?o0000ooooooooool0J_oo
ool00ol0003oooooooooo`1:ooooo`Oo00006Oooool004Gooooo00?o0000ooooooooool0Joooool0
0ol0003oooooooooo`1:ooooo`Oo00006Oooool004?ooooo0_l0001^ooooo`03o`000?oooooooooo
04[ooooo1ol0000Iooooo`00?_ooool3o`000003ooooool0003ooooo073ooooo00?o0000oooooooo
ool0B_ooool5o`0001[ooooo000mooooo`Go0000L_ooool00ol0003oooooooooo`1;ooooo`?o0000
6oooool003cooooo1ol0001booooo`03o`000?oooooooooo06Sooooo000looooo`Oo0000L_ooool0
0ol0003oooooooooo`1Xooooo`00??ooool7o`0007?ooooo0ol0001Wooooo`00>oooool7o`0007?o
oooo1Ol0001Vooooo`00>_ooool01?l0003oooooooooooooool3o`0007?ooooo1ol0001Uooooo`00
>Oooool00ol0003oooooooooo`1hooooo`Oo0000IOooool003Oooooo0_l0001kooooo`Oo0000IOoo
ool003Kooooo00?o0000ooooooooool0O?ooool5o`0006Kooooo000`ooooo`?o000000?oooooo`00
0?l00000P?ooool3o`0006Oooooo000_ooooo`Go0000Q?ooool2o`0006Kooooo000^ooooo`Oo0000
Q?ooool00ol0003oooooooooo`1Tooooo`00;_ooool7o`0008Cooooo0_l0001Uooooo`00;_ooool7
o`0008Cooooo0_l0001Uooooo`00;Oooool00ol0003oooooo`000004o`0008Kooooo0_l0001Toooo
o`00:oooool2o`000003ooooool0003ooooo00?o0000Qoooool2o`0006Cooooo000Yooooo`;o0000
0_ooool00ol0003oooooooooo`2:ooooo`03o`000?oooooo000006?ooooo000Wooooo`;o00000ooo
ool00ol0003oooooooooo`2<ooooo`Go0000H?ooool0023ooooo0ol00002ooooo`;o00001?ooool0
0ol0003oooooooooo`2<ooooo`Oo0000Goooool001oooooo1_l00005ooooo`03o`000?oooooooooo
08cooooo2Ol0001Nooooo`007_ooool7o`0000Cooooo00?o0000ooooooooool0SOooool9o`0005ko
oooo000Nooooo`Wo000000?oooooo`000?ooool0Soooool9o`0005kooooo000Nooooo`[o0000T_oo
ool7o`0005oooooo000Looooo`go0000T_ooool5o`00063ooooo000Jooooo`?o00000oooool9o`00
0?Oooooo000Hooooo`Co00001_ooool7o`000?Oooooo000@ooooo`?o00000oooool5o`0000Sooooo
1Ol0003hooooo`003oooool:o`0000_ooooo0ol0003iooooo`003_ooool:o`000?oooooo2Oooool0
00kooooo2?l0003oooooo`_ooooo000>ooooo`So0000ooooool;ooooo`003oooool8o`000?oooooo
2_ooool000oooooo2?l0003oooooo`[ooooo000@ooooo`Oo0000ooooool:ooooo`004Oooool5o`00
0?oooooo2oooool001;ooooo0ol0003oooooo`cooooo003oooooob7ooooo003oooooob7ooooo003o
ooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003oooooob7ooooo003o
ooooob7ooooo003oooooob7ooooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {0.214036, -0.0500056, \
0.00291278, 0.00382954}},
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],

Cell["\<\
If no unique central vertex is requested, the roots of the trees \
are connected in a cycle.\
\>", "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(WuenscheTree[21, 17, 13, 17, 
          24, \[IndentingNewLine]UniqueCenter \[Rule] 
            False\[IndentingNewLine]] // ShowGraph;\)\)], "Input",
  CellLabel->"In[17]:=",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.119857 1.23433 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.8663 .90127 m
.83059 .85181 L
s
.89367 .95375 m
.86153 .90298 L
s
.7622 .57481 m
.83453 .59212 L
s
.7622 .57481 m
.68987 .5575 L
s
.89124 .95455 m
.86153 .90298 L
s
.72402 .70415 m
.76872 .74947 L
s
.72402 .70415 m
.67931 .65883 L
s
.86153 .90298 m
.83059 .85181 L
s
.82524 .78695 m
.88176 .82443 L
s
.82524 .78695 m
.76872 .74947 L
s
.88176 .82443 m
.93828 .86191 L
s
.76872 .74947 m
.81343 .79478 L
s
.76872 .74947 m
.79966 .80064 L
s
.67931 .65883 m
.63461 .61351 L
s
.61753 .54019 m
.68987 .5575 L
s
.61753 .54019 m
.5452 .52288 L
s
.79966 .80064 m
.83059 .85181 L
s
.5899 .5682 m
.63461 .61351 L
s
.5899 .5682 m
.5452 .52288 L
s
.83059 .85181 m
.8566 .90445 L
s
.5452 .52288 m
.46017 .54323 L
s
.5452 .52288 m
.5452 .45703 L
s
.46017 .65526 m
.47207 .71058 L
s
.46017 .65526 m
.46017 .59924 L
s
.46017 .65526 m
.44827 .71058 L
s
.4799 .87853 m
.48252 .93451 L
s
.4799 .87853 m
.47729 .82255 L
s
.48027 .82244 m
.47505 .76656 L
s
.47207 .71058 m
.49271 .7645 L
s
.47207 .71058 m
.4869 .76552 L
s
.47207 .71058 m
.481 .76621 L
s
.47207 .71058 m
.47505 .76656 L
s
.47916 .87855 m
.47729 .82255 L
s
.47878 .8225 m
.47505 .76656 L
s
.47729 .82255 m
.47505 .76656 L
s
.46017 .59924 m
.46017 .54323 L
s
.41546 .58855 m
.46017 .54323 L
s
.4758 .82257 m
.47505 .76656 L
s
.46017 .54323 m
.40761 .48995 L
s
.10826 .52189 m
.18421 .51896 L
s
.1867 .53052 m
.26016 .51603 L
s
.33528 .50726 m
.26752 .53269 L
s
.33528 .50726 m
.26016 .51603 L
s
.33528 .50726 m
.40761 .48995 L
s
.33528 .47264 m
.40761 .48995 L
s
.06713 .61282 m
.13198 .58356 L
s
.18504 .52479 m
.26016 .51603 L
s
.06172 .60499 m
.13198 .58356 L
s
.19975 .55812 m
.13198 .58356 L
s
.19975 .55812 m
.26752 .53269 L
s
.18421 .51896 m
.26016 .51603 L
s
.40761 .48995 m
.46017 .43668 L
s
.31518 .2416 m
.35743 .28818 L
s
.33818 .29848 m
.39196 .33809 L
s
.33818 .29848 m
.2844 .25887 L
s
.39196 .33809 m
.43667 .38341 L
s
.39196 .33809 m
.35743 .28818 L
s
.43667 .32739 m
.43667 .38341 L
s
.24465 .14211 m
.20619 .09378 L
s
.24465 .14211 m
.28311 .19043 L
s
.2844 .25887 m
.23062 .21926 L
s
.20619 .09378 m
.16773 .04545 L
s
.43667 .38341 m
.46017 .43668 L
s
.35743 .28818 m
.32027 .23931 L
s
.35743 .28818 m
.32559 .23731 L
s
.35743 .28818 m
.33111 .23563 L
s
.32027 .23931 m
.28311 .19043 L
s
.28311 .19043 m
.24725 .14103 L
s
.46017 .43668 m
.5452 .45703 L
s
.5899 .41171 m
.63461 .36639 L
s
.5899 .41171 m
.5452 .45703 L
s
.74604 .21985 m
.70888 .26872 L
s
.6853 .41429 m
.75307 .38886 L
s
.6853 .41429 m
.61753 .43972 L
s
.72457 .21218 m
.69824 .26473 L
s
.73008 .21386 m
.69824 .26473 L
s
.73008 .21386 m
.76193 .16299 L
s
.69824 .26473 m
.66914 .31648 L
s
.75113 .22215 m
.70888 .26872 L
s
.66914 .31648 m
.63461 .36639 L
s
.66914 .31648 m
.70888 .26872 L
s
.76193 .16299 m
.79377 .11212 L
s
.69265 .43096 m
.76661 .41788 L
s
.69265 .43096 m
.61753 .43972 L
s
.69265 .43096 m
.76847 .42656 L
s
.76661 .41788 m
.84056 .4048 L
s
.75307 .38886 m
.82083 .36343 L
s
.63461 .36639 m
.68839 .32679 L
s
.82561 .06125 m
.79377 .11212 L
s
.68839 .32679 m
.74217 .28718 L
s
.74217 .28718 m
.79595 .24757 L
s
.61753 .43972 m
.5452 .45703 L
s
.025 w
.8663 .90127 Mdot
.89367 .95375 Mdot
.7622 .57481 Mdot
.89124 .95455 Mdot
.72402 .70415 Mdot
.86153 .90298 Mdot
.82524 .78695 Mdot
.88176 .82443 Mdot
.76872 .74947 Mdot
.83453 .59212 Mdot
.67931 .65883 Mdot
.81343 .79478 Mdot
.61753 .54019 Mdot
.79966 .80064 Mdot
.5899 .5682 Mdot
.93828 .86191 Mdot
.83059 .85181 Mdot
.63461 .61351 Mdot
.8566 .90445 Mdot
.68987 .5575 Mdot
.5452 .52288 Mdot
.46017 .65526 Mdot
.4799 .87853 Mdot
.48027 .82244 Mdot
.47207 .71058 Mdot
.49271 .7645 Mdot
.47916 .87855 Mdot
.48252 .93451 Mdot
.47878 .8225 Mdot
.47729 .82255 Mdot
.46017 .59924 Mdot
.4869 .76552 Mdot
.41546 .58855 Mdot
.44827 .71058 Mdot
.481 .76621 Mdot
.4758 .82257 Mdot
.47505 .76656 Mdot
.46017 .54323 Mdot
.10826 .52189 Mdot
.1867 .53052 Mdot
.33528 .50726 Mdot
.33528 .47264 Mdot
.06713 .61282 Mdot
.18504 .52479 Mdot
.06172 .60499 Mdot
.19975 .55812 Mdot
.18421 .51896 Mdot
.13198 .58356 Mdot
.26752 .53269 Mdot
.26016 .51603 Mdot
.40761 .48995 Mdot
.31518 .2416 Mdot
.33818 .29848 Mdot
.39196 .33809 Mdot
.43667 .32739 Mdot
.24465 .14211 Mdot
.2844 .25887 Mdot
.20619 .09378 Mdot
.43667 .38341 Mdot
.35743 .28818 Mdot
.32027 .23931 Mdot
.28311 .19043 Mdot
.24725 .14103 Mdot
.32559 .23731 Mdot
.16773 .04545 Mdot
.33111 .23563 Mdot
.23062 .21926 Mdot
.46017 .43668 Mdot
.5899 .41171 Mdot
.74604 .21985 Mdot
.6853 .41429 Mdot
.72457 .21218 Mdot
.73008 .21386 Mdot
.69824 .26473 Mdot
.75113 .22215 Mdot
.66914 .31648 Mdot
.76193 .16299 Mdot
.69265 .43096 Mdot
.76661 .41788 Mdot
.75307 .38886 Mdot
.63461 .36639 Mdot
.82083 .36343 Mdot
.82561 .06125 Mdot
.68839 .32679 Mdot
.84056 .4048 Mdot
.79377 .11212 Mdot
.74217 .28718 Mdot
.61753 .43972 Mdot
.76847 .42656 Mdot
.70888 .26872 Mdot
.79595 .24757 Mdot
.5452 .45703 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[17]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00;oooool3o`000>kooooo000^ooooo`Go0000kOooool002gooooo1ol0003/ooooo`00;Ooo
ool7o`000>cooooo000]ooooo`Oo0000]oooool3o`0003;ooooo000^ooooo`Go0000]oooool5o`00
037ooooo000_ooooo`Co0000]_ooool7o`00033ooooo000cooooo`03o`000?oooooooooo0;?ooooo
1ol0000`ooooo`00=?ooool00ol0003oooooooooo`2booooo`Oo0000<?ooool003Gooooo00?o0000
ooooooooool0/_ooool5o`00037ooooo000fooooo`03o`000?oooooooooo0;7ooooo1?l0000boooo
o`00=_ooool00ol0003oooooooooo`2aooooo`03o`000?oooooooooo03?ooooo000gooooo`03o`00
0?oooooooooo0:oooooo00?o0000ooooooooool0=?ooool003Sooooo00?o0000ooooool000000_l0
002/ooooo`03o`000?oooooooooo03Cooooo000iooooo`Go0000Z_ooool00ol0003oooooooooo`0e
ooooo`00>?ooool7o`000:Sooooo00?o0000ooooooooool0=_ooool003Sooooo1ol0002Xooooo`03
o`000?oooooooooo03Kooooo000hooooo`Oo0000Yoooool00ol0003oooooooooo`0gooooo`00>Ooo
ool5o`000:Sooooo00?o0000ooooooooool0=oooool003[ooooo1?l0002Tooooo`Co0000>_ooool0
03kooooo00?o0000ooooooooool0X?ooool5o`0003[ooooo000oooooo`03o`000?oooooooooo09ko
oooo1ol0000iooooo`00?oooool00ol0003oooooooooo`2Nooooo`Oo0000>Oooool0043ooooo00?o
0000ooooooooool0WOooool7o`0003Wooooo0011ooooo`03o`000?oooooooooo09gooooo1Ol0000j
ooooo`00@_ooool00ol0003oooooooooo`2Looooo`Co0000>oooool004;ooooo00?o0000oooooooo
ool0W?ooool00ol0003oooooooooo`0looooo`00@oooool00ol0003oooooo`000002o`0009Sooooo
00?o0000ooooooooool0?Oooool004Cooooo1Ol0002Fooooo`03o`000?oooooooooo03kooooo0013
ooooo`Oo0000UOooool00ol0003oooooooooo`0nooooo`00@oooool7o`0009Cooooo00?o0000oooo
ooooool0?oooool004?ooooo1ol0002Dooooo`03o`000?oooooooooo03oooooo0014ooooo`Go0000
U?ooool00ol0003oooooooooo`10ooooo`00AOooool4o`00093ooooo1?l00013ooooo`00BOooool0
0ol0003oooooooooo`2<ooooo`Go0000@oooool004[ooooo00?o0000ooooooooool0R_ooool7o`00
04;ooooo001:ooooo`03o`000?oooooooooo08[ooooo1ol00012ooooo`00Boooool00ol0003ooooo
ooooo`29ooooo`Oo0000@_ooool004cooooo00?o0000ooooooooool0ROooool5o`0004?ooooo001=
ooooo`03o`000?oooooooooo08Sooooo1?l00014ooooo`00COooool00ol0003oooooooooo`28oooo
o`03o`000?oooooooooo04Gooooo001>ooooo`03o`000?oooooo000000;o0000Q?ooool00ol0003o
ooooooooo`16ooooo`00Coooool5o`0008?ooooo00?o0000ooooooooool0A_ooool004kooooo1ol0
0021ooooo`03o`000?oooooooooo04Oooooo001>ooooo`Oo0000P?ooool00ol0003oooooooooo`18
ooooo`00C_ooool7o`00083ooooo00?o0000ooooooooool0B?ooool004oooooo1Ol00020ooooo`03
o`000?oooooooooo04Wooooo001@ooooo`Co0000N_ooool3o`0000?ooooo00?o0000ooooooooool0
BOooool005Cooooo00?o0000ooooooooool0M_ooool7o`0004cooooo0011ooooo`?o00004?ooool0
0ol0003oooooooooo`1eooooo`So0000C?ooool0043ooooo1Ol0000@ooooo`03o`000?oooooooooo
07Cooooo3Ol00017ooooo`00?oooool7o`00013ooooo00?o0000ooooooooool0Loooool>o`0004Ko
oooo000oooooo`Oo00004?ooool00ol0003oooooooooo`1dooooo`ko0000AOooool003oooooo1ol0
000Aooooo`03o`000?oooooooooo07Cooooo3Ol00015ooooo`00@?ooool6o`0001;ooooo00?o0000
ooooooooool00_ooool3o`0006gooooo00?o0000ooooool000002ol00015ooooo`00@Oooool3o`00
00;ooooo00?o0000ooooooooool04?ooool8o`0006cooooo00?o0000ooooool000001?ooool6o`00
04Kooooo0017ooooo`;o00004?ooool9o`0006[ooooo00?o0000ooooool000001Oooool5o`0004Oo
oooo0019ooooo`03o`000?oooooooooo00cooooo2_l0001Zooooo`;o00001Oooool2o`0004_ooooo
001:ooooo`03o`000?oooooooooo00[ooooo2ol0001Yooooo`03o`000?oooooo000000Cooooo0_l0
000?ooooo`?o0000>_ooool004_ooooo00?o0000ooooooooool02Oooool:o`0006[ooooo0_l00005
ooooo`03o`000?oooooooooo00gooooo1Ol0000iooooo`00C?ooool2o`0000Wooooo2_l0001Zoooo
o`;o00001?ooool00ol0003oooooooooo`0=ooooo`Oo0000>?ooool004kooooo00?o0000ooooool0
00000_l00005ooooo`So000000?oooooo`000?ooool0Ioooool2o`0000Cooooo0_l0000?ooooo`Oo
0000>?ooool004oooooo1Ol00005ooooo`Co000000Goooooo`000?l0003oooooo`00001Xooooo`03
o`000?oooooooooo00;ooooo0_l0000@ooooo`Oo0000>?ooool004kooooo1ol00008ooooo`06o`00
0?oooooo0000o`000?oooooo0000I?ooool4o`0000Cooooo00?o0000ooooooooool03oooool6o`00
03Wooooo001>ooooo`Oo00002Oooool2o`000003ooooool0003o000006?ooooo1ol000000ooooooo
0000ooooo`0?ooooo`;o00000_ooool3o`0003[ooooo001>ooooo`Oo00002_ooool3o`000003oooo
ool0003ooooo063ooooo2Ol0000@ooooo`03o`000?oooooooooo03oooooo001?ooooo`Oo00002_oo
ool4o`00067ooooo2_l0000=ooooo`;o0000@_ooool0053ooooo0ol00003ooooo`03o`000?oooooo
oooo00Sooooo0ol0001Qooooo`[o00003?ooool00ol0003oooooooooo`12ooooo`00Eoooool2o`00
00Wooooo0ol0001Qooooo`Wo00002oooool00ol0003oooooooooo`13ooooo`00FOooool00ol0003o
ooooooooo`07ooooo`Go0000G_ooool8o`0000Gooooo0ol00002ooooo`;o0000A_ooool005[ooooo
00?o0000ooooooooool01oooool5o`0005gooooo00?o0000ooooooooool01?l00005ooooo`Ko0000
B?ooool005_ooooo0_l00006ooooo`Oo0000Foooool01?l0003oooooooooool00008ooooo`Oo0000
B?ooool005gooooo00?o0000ooooool000000_l000000ooooooo0000o`000005o`0005_ooooo00?o
0000ooooool000002Oooool7o`0004Sooooo001Nooooo`co0000F_ooool00ol0003oooooo`00000:
ooooo`Oo0000B?ooool005gooooo3?l0001Jooooo`03o`000?oooooo000000_ooooo1_l00019oooo
o`00GOooool7o`000003ooooool0003o000000;o0000F_ooool00ol0003oooooo`00000:ooooo`03
o`000?oooooooooo00?o0000B_ooool005gooooo1ol00004ooooo`03o`000?oooooooooo05Oooooo
00?o0000ooooool000002Oooool2o`00053ooooo001Nooooo`Go000000?oooooo`000?ooool00ooo
ool00ol0003oooooooooo`1Booooo`?o000000?oooooo`000?l000002Oooool00ol0003ooooooooo
o`1@ooooo`00Goooool3o`0000?ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`1@
ooooo`Ko00002Oooool00ol0003oooooooooo`1Aooooo`00I_ooool2o`0000;ooooo00?o0000oooo
ooooool0Coooool7o`0000Oooooo0_l0001Dooooo`00J?ooool01?l0003oooooooooool0001@oooo
o`Oo000000Coooooo`000?l0003o00000_ooool00ol0003oooooooooo`1Dooooo`00JOooool01?l0
003oooooooooool0000?ooooo`?o0000?Oooool=o`0005Oooooo001Zooooo`03o`000?oooooo0000
00kooooo1Ol0000mooooo`co0000Eoooool006_ooooo0ol0000<ooooo`Oo0000?Oooool3o`000003
ooooool0003o000000Go0000Eoooool006gooooo1Ol00008ooooo`Oo0000?Oooool01?l0003ooooo
ooooooooool7o`0005Oooooo001^ooooo`Go00001oooool7o`0003cooooo00?o0000ooooooooool0
0_ooool6o`0005Sooooo001]ooooo`Oo00001oooool5o`0003cooooo00?o0000ooooooooool00_oo
ool00ol0003oooooooooo`03o`0005Wooooo001]ooooo`Oo00002?ooool3o`0003gooooo00?o0000
ooooooooool00_l0001Oooooo`00KOooool7o`0000Wooooo00?o0000ooooooooool0>oooool01?l0
003oooooooooool0001Qooooo`00K_ooool5o`0000[ooooo00?o0000ooooooooool0>_ooool01?l0
003oooooooooool0001Rooooo`00Koooool3o`000003ooooool0003ooooo00Sooooo00?o0000oooo
ooooool0>_ooool00ol0003oooooo`00001Sooooo`00M?ooool00ol0003oooooooooo`06ooooo`03
o`000?oooooooooo03Wooooo0ol0000^ooooo`?o0000<oooool007Gooooo00?o0000ooooooooool0
1Oooool00ol0003oooooooooo`0eooooo`Go0000;oooool5o`0003;ooooo001fooooo`03o`000?oo
oooooooo00Cooooo00?o0000ooooooooool0=?ooool5o`0002oooooo1ol0000aooooo`00Moooool0
0ol0003oooooooooo`03ooooo`03o`000?oooooooooo03?ooooo1ol0000^ooooo`Oo0000<Oooool0
07Sooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`0cooooo`Oo0000;Oooool8o`00
037ooooo001iooooo`05o`000?ooooooooooooooool00000=Oooool7o`0002[ooooo0ol00002oooo
o`Go0000<_ooool007[ooooo00?o0000ooooool000000_l0000eooooo`Go0000:?ooool3o`0000Ko
oooo0ol0000cooooo`00Noooool5o`0003?ooooo00?o0000ooooool000000_l0000Oooooo`?o0000
1Oooool2o`0003oooooo001jooooo`Oo0000<Oooool00ol0003oooooooooo`0Qooooo`Go000000Co
ooooo`000?l0003o0000@Oooool007[ooooo1ol0000`ooooo`03o`000?oooooooooo027ooooo1ol0
0014ooooo`00N_ooool7o`0002oooooo00?o0000ooooooooool08_ooool7o`0004Cooooo001koooo
o`Go0000;oooool00ol0003oooooooooo`0Rooooo`So0000A?ooool007cooooo0ol0000_ooooo`03
o`000?oooooooooo023ooooo0ol00002ooooo`Go00005Oooool3o`0002gooooo001oooooo`03o`00
0?oooooooooo02_ooooo00?o0000ooooooooool07_ooool3o`0000Kooooo0ol0000Eooooo`Go0000
;?ooool007oooooo00?o0000ooooooooool09_ooool3o`000003ooooool0003ooooo01Gooooo0ol0
0005ooooo`?o00008?ooool7o`0002_ooooo0020ooooo`03o`000?oooooooooo02Cooooo1Ol0000F
ooooo`Go000000Coooooo`000?l0003o00003oooool3o`00017ooooo1ol0000[ooooo`00P?ooool0
0ol0003oooooooooo`0Sooooo`Oo00005?ooool7o`00017ooooo1Ol00009ooooo`ko0000:oooool0
087ooooo00?o0000ooooooooool08_ooool7o`0001Cooooo1ol0000@ooooo`Oo000000?oooooo`00
0?l000001Ol00008ooooo`Go0000;?ooool0087ooooo00?o0000ooooooooool08_ooool7o`0001?o
oooo2?l0000@ooooo`So00004?ooool3o`0002gooooo0021ooooo`03o`000?oooooooooo02?ooooo
1Ol0000Booooo`;o00000_ooool6o`0000cooooo2ol00011ooooo`00P_ooool00ol0003ooooooooo
o`0Qooooo`03o`000?oooooo000000;o00004Oooool2o`0000Gooooo1_l00005ooooo`Ko00001Ooo
ool7o`00043ooooo0022ooooo`Co00007oooool00ol0003oooooooooo`0Aooooo`?o00001oooool;
o`0000?ooooo3ol00010ooooo`00P_ooool5o`0001gooooo00?o0000ooooooooool02Oooool3o`00
00Cooooo0_l0000:ooooo`ko00002?ooool7o`00043ooooo0021ooooo`Oo00006oooool00ol0003o
ooooooooo`09ooooo`Go000000?oooooo`000?l000001Oooool>o`00013ooooo1Ol00011ooooo`00
POooool7o`0001[ooooo00?o0000ooooooooool02Oooool>o`0000Sooooo1Ol0000Booooo`?o0000
@_ooool0087ooooo2_l0000Fooooo`03o`000?oooooooooo00[ooooo1ol0000@ooooo`?o0000F?oo
ool0087ooooo1_l00004ooooo`Co00004Oooool00ol0003oooooooooo`08ooooo`[o0000Joooool0
083ooooo00?o0000ooooooooool00ol00009ooooo`Co00002?ooool3o`000003ooooool0003ooooo
00Gooooo1Ol00004ooooo`Go0000K?ooool007oooooo00?o0000ooooooooool04Oooool4o`0000?o
oooo1Ol00003ooooo`Co00002_ooool3o`0006gooooo001nooooo`03o`000?oooooooooo01Kooooo
2ol0001nooooo`00OOooool00ol0003oooooooooo`0Iooooo`Oo0000P?ooool005oooooo0ol0000J
ooooo`03o`000?oooooooooo01[ooooo1ol00020ooooo`00G_ooool5o`0001Sooooo00?o0000oooo
ooooool07?ooool5o`00087ooooo001Mooooo`Oo00005_ooool00ol0003oooooooooo`0Nooooo`?o
0000P_ooool005gooooo1ol0000Eooooo`03o`000?oooooooooo023ooooo00?o0000ooooooooool0
POooool005gooooo2_l0000Aooooo`03o`000?oooooooooo027ooooo00?o0000ooooooooool0POoo
ool005kooooo1Ol00004ooooo`Co00002?ooool3o`000003ooooool0003ooooo02?ooooo00?o0000
ooooooooool0POooool005oooooo0ol00009ooooo`Co00000oooool5o`0002Gooooo00?o0000oooo
ooooool0POooool006oooooo2Ol0000Tooooo`03o`000?oooooooooo087ooooo001aooooo`Oo0000
9?ooool00ol0003oooooooooo`21ooooo`00K_ooool:o`0002Cooooo00?o0000ooooooooool0POoo
ool005oooooo0ol00008ooooo`Co00001?ooool6o`0002Cooooo00?o0000ooooooooool0POooool0
05kooooo1Ol00003ooooo`Co00002Oooool3o`0000;ooooo00?o0000ooooooooool08Oooool00ol0
003oooooooooo`21ooooo`00GOooool9o`0001?ooooo00?o0000ooooooooool08?ooool00ol0003o
ooooooooo`21ooooo`00BOooool3o`00013ooooo2?l0000Fooooo`03o`000?oooooooooo01oooooo
00?o0000ooooooooool0POooool001kooooo0ol0000Booooo`?o00004_ooool5o`0000Sooooo3ol0
000Gooooo`03o`000?oooooooooo01kooooo00?o0000ooooooooool0POooool001gooooo1Ol0000@
ooooo`Go00004?ooool>o`0000Cooooo0ol00002ooooo`Go00006Oooool00ol0003oooooooooo`0L
ooooo`?o0000P_ooool001cooooo1ol00007ooooobCo00002?ooool3o`0000Kooooo0ol0000Koooo
o`03o`000?oooooooooo01[ooooo1Ol00021ooooo`007?ooool>o`0000Oooooo2?l000000ooooooo
0000o`00000Bo`0000Gooooo0ol0000Xooooo`03o`000?oooooooooo01Sooooo1ol00020ooooo`00
7?ooool7o`0000kooooo2Ol00003ooooo`Go00001_ooool7o`000004ooooool0003o0000o`0002co
oooo00?o0000ooooooooool05oooool7o`00083ooooo000Mooooo`Go00004?ooool;o`0000cooooo
1ol0000`ooooo`03o`000?oooooooooo01;ooooo3_l0001mooooo`007_ooool3o`00017ooooo1ol0
000@ooooo`Oo0000<Oooool00ol0003oooooo`000002o`0000[ooooo1Ol00005ooooo`Go00001?oo
ool5o`0000Sooooo0ol0001]ooooo`00<_ooool7o`0000oooooo2?l0000booooo`Go00001?ooool5
o`0000_ooooo0ol000000ooooooo0000ooooo`07ooooo`Co00000oooool5o`0006cooooo000coooo
o`Go00003_ooool2o`0000;ooooo1Ol0000booooo`[o00005Oooool00ol0003oooooooooo`09oooo
o`Wo0000Joooool003Cooooo0ol0000=ooooo`;o00001Oooool3o`0003?ooooo1ol0000Iooooo`03
o`000?oooooooooo00[ooooo1ol0001[ooooo`00@Oooool3o`0003gooooo1ol0000Jooooo`03o`00
0?oooooooooo00Wooooo2_l0001Xooooo`00>?ooool3o`0000Cooooo0_l00011ooooo`Go00007?oo
ool00ol0003oooooooooo`09ooooo`Go00001?ooool4o`0000Sooooo0ol0001Iooooo`00=oooool5
o`000003ooooool0003o000004;ooooo00?o0000ooooool000000_l0000Nooooo`03o`000?oooooo
oooo00Wooooo0ol00009ooooo`Co00000oooool5o`0005Sooooo000fooooo`Oo0000@oooool01Ol0
003ooooooooooooooooo0000023ooooo00?o0000ooooooooool06?ooool9o`0005Oooooo000foooo
o`Oo0000@_ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01oooooo00?o0000oooo
ool000000_l0000Gooooo`Oo0000Eoooool003Gooooo2?l00011ooooo`03o`000?oooooooooo00?o
oooo00?o0000ooooooooool08?ooool5o`0001Kooooo3?l0001Booooo`00<_ooool3o`0000;ooooo
1Ol00011ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool07oooool7o`0001Kooooo
1Ol00006ooooo`Go00001_ooool3o`0004Cooooo000_ooooo`?o00001_ooool3o`00047ooooo00?o
0000ooooooooool01Oooool00ol0003oooooooooo`0Oooooo`Oo00005oooool3o`0000cooooo2_l0
0013ooooo`009?ooool3o`0000Gooooo0ol0001<ooooo`03o`000?oooooooooo00Kooooo00?o0000
ooooooooool07oooool7o`0002[ooooo1ol00012ooooo`008oooool5o`000004ooooool0003o0000
o`0004[ooooo0ol000000ooooooo0000ooooo`08ooooo`03o`000?oooooooooo023ooooo1Ol0000[
ooooo`[o0000?oooool002;ooooo1ol0001<ooooo`Go00002_ooool00ol0003oooooooooo`0Qoooo
o`?o000000?oooooo`000?ooool0:Oooool7o`0000?ooooo1Ol00008ooooo`?o0000;oooool002;o
oooo1ol0001;ooooo`Oo00002Oooool00ol0003oooooooooo`0Vooooo`03o`000?oooooooooo02So
oooo1Ol00009ooooo`Co00000oooool5o`0002kooooo000Pooooo`Wo0000Boooool7o`0000Wooooo
00?o0000ooooooooool09oooool00ol0003oooooooooo`0Xooooo`?o00003_ooool9o`0002gooooo
000Mooooo`Go000000?oooooo`000?l000000ol0001<ooooo`Oo00002?ooool3o`0002Wooooo00?o
0000ooooooooool0>_ooool7o`0002gooooo000@ooooo`?o00001_ooool6o`0000Gooooo0ol0001>
ooooo`Go00002?ooool5o`0002Wooooo00?o0000ooooooooool0>Oooool7o`0002gooooo000?oooo
o`Go00000_ooool3o`0000;ooooo0_l0001Iooooo`?o00002?ooool7o`0002Wooooo00?o0000oooo
ooooool0>Oooool5o`0002kooooo000>ooooo`So00000oooool2o`0006Kooooo1ol0000Zooooo`03
o`000?oooooooooo03Wooooo0ol0000_ooooo`003_ooool;o`0006Sooooo1ol0000[ooooo`03o`00
0?oooooo000000;o0000J?ooool000kooooo2Ol0001[ooooo`Go0000;Oooool5o`0006Oooooo000?
ooooo`So0000K?ooool3o`0002gooooo1ol0001Vooooo`004?ooool7o`0006gooooo00?o0000oooo
ooooool0;?ooool7o`0006Kooooo000Aooooo`Go0000K_ooool00ol0003oooooooooo`0/ooooo`Oo
0000I_ooool001;ooooo0ol0001_ooooo`03o`000?oooooooooo02gooooo1Ol0001Wooooo`00Q?oo
ool00ol0003oooooooooo`0^ooooo`Co0000Ioooool008Cooooo00?o0000ooooooooool0<_ooool0
0ol0003oooooooooo`1Tooooo`00Q?ooool00ol0003oooooooooo`0cooooo`03o`000?oooooooooo
06?ooooo0024ooooo`03o`000?oooooooooo03Cooooo00?o0000ooooooooool0H_ooool008Cooooo
00?o0000ooooooooool0=Oooool00ol0003oooooooooo`1Qooooo`00Q?ooool00ol0003ooooooooo
o`0fooooo`03o`000?oooooooooo063ooooo0023ooooo`?o0000>?ooool00ol0003oooooooooo`1O
ooooo`00P_ooool5o`0003Sooooo00?o0000ooooool000000_l0001Looooo`00POooool7o`0003So
oooo1Ol0001Kooooo`00POooool7o`0003Oooooo1ol0001Jooooo`00POooool7o`0003Oooooo1ol0
001Jooooo`00P_ooool5o`0003Sooooo1ol0001Jooooo`00Poooool3o`0003[ooooo1Ol0001Koooo
o`00Poooool00ol0003oooooo`00000kooooo`?o000000?oooooo`000?ooool0FOooool008?ooooo
00?o0000ooooool00000@?ooool00ol0003oooooooooo`1Gooooo`00Poooool00ol0003oooooo`00
0011ooooo`03o`000?oooooooooo05Kooooo0022ooooo`04o`000?ooooooooooo`0004;ooooo00?o
0000ooooooooool0EOooool008;ooooo00Go0000ooooooooooooooooo`000012ooooo`03o`000?oo
oooooooo05Cooooo0022ooooo`05o`000?ooooooooooooooool00000@oooool00ol0003ooooooooo
o`1Cooooo`00POooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo04;ooooo00?o0000
ooooooooool0D_ooool0087ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`13oooo
o`03o`000?oooooo000000;o0000Coooool007oooooo0ol00004ooooo`?o0000A?ooool5o`0004ko
oooo001nooooo`Go00000_ooool5o`0004;ooooo1ol0001=ooooo`00OOooool>o`00047ooooo1ol0
001=ooooo`00OOooool>o`00047ooooo1ol0001=ooooo`00OOooool>o`0004;ooooo1Ol0001>oooo
o`00O_ooool5o`0000;ooooo1Ol00014ooooo`?o000000?oooooo`000?ooool0C?ooool007oooooo
0ol00004ooooo`?o0000B_ooool00ol0003oooooooooo`1:ooooo`00Qoooool2o`0004_ooooo00?o
0000ooooooooool0BOooool008Oooooo0ol0001;ooooo`03o`000?oooooooooo04Sooooo0027oooo
o`?o0000C?ooool00ol0003oooooooooo`17ooooo`00Qoooool3o`0004gooooo00?o0000oooooooo
ool0A_ooool008Oooooo1?l0001=ooooo`03o`000?oooooooooo04Gooooo0028ooooo`?o0000C_oo
ool00ol0003oooooo`000002o`0004;ooooo0028ooooo`;o000000?oooooo`000?ooool0COooool5
o`00047ooooo0028ooooo`Co0000COooool7o`00043ooooo0028ooooo`Co0000COooool7o`00043o
oooo0028ooooo`Oo0000B_ooool7o`00043ooooo0027ooooo`Wo0000B_ooool7o`0003oooooo0026
ooooo`_o0000B_ooool5o`000003ooooool0003o000003gooooo0025ooooo`co0000COooool00ol0
003oooooo`000002ooooo`03o`000?oooooooooo03[ooooo0025ooooo`co0000C_ooool00ol0003o
ooooo`000002ooooo`;o0000>_ooool008Gooooo2ol0001@ooooo`03o`000?oooooo000000?ooooo
00?o0000ooooooooool0=oooool008Kooooo2Ol0001Aooooo`04o`000?ooooooooooo`0000?ooooo
0_l00002ooooo`?o0000<_ooool008Oooooo0ol0001Gooooo`04o`000?ooooooooooo`0000Cooooo
1_l0000aooooo`00R?ooool00ol0003oooooooooo`1Fooooo`05o`000?ooooooooooooooool00000
0oooool7o`00033ooooo0028ooooo`03o`000?oooooooooo05Oooooo00Ko0000oooooooooooooooo
o`000?ooool8o`00033ooooo0028ooooo`03o`000?oooooooooo05Sooooo3Ol0000`ooooo`00R?oo
ool00ol0003oooooooooo`1Hooooo`ko0000;oooool008Sooooo0_l0001Hooooo`co00000oooool0
0ol0003oooooooooo`0/ooooo`00R?ooool2o`0005Sooooo2ol00005ooooo`;o0000;?ooool008So
oooo0_l0001Hooooo`[o00002?ooool2o`0002[ooooo0028ooooo`;o0000FOooool8o`0000_ooooo
00?o0000ooooooooool09oooool008Sooooo0_l0001Jooooo`Co00003oooool2o`0002Oooooo0027
ooooo`Co0000G?ooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo00?o00008Oooool0
08Kooooo1_l0001Looooo`03o`000?oooooooooo00oooooo1_l0000Pooooo`00QOooool8o`0005_o
oooo00?o0000ooooooooool03oooool7o`0001oooooo0025ooooo`So0000G?ooool00ol0003ooooo
ooooo`0>ooooo`Oo00007oooool008Gooooo2?l0001Mooooo`03o`000?oooooooooo00gooooo1ol0
000Oooooo`00Q_ooool6o`0005kooooo00?o0000ooooooooool03_ooool7o`0001kooooo0027oooo
o`Co0000H?ooool00ol0003oooooooooo`0>ooooo`?o00000oooool00ol0003oooooooooo`0Koooo
o`00R?ooool00ol0003oooooooooo`1Pooooo`03o`000?oooooooooo01Gooooo0_l0000Kooooo`00
R?ooool00ol0003oooooooooo`1Qooooo`Co00005Oooool00ol0003oooooooooo`0Hooooo`00R?oo
ool00ol0003oooooooooo`1Qooooo`Go00005Oooool00ol0003oooooooooo`0Gooooo`00R?ooool0
0ol0003oooooooooo`1Pooooo`Oo00005Oooool2o`0001Oooooo0028ooooo`03o`000?oooooooooo
063ooooo1ol0000Gooooo`03o`000?oooooooooo00?o00004Oooool008Wooooo00?o0000oooooooo
ool0Goooool7o`0001Sooooo1_l0000@ooooo`00ROooool00ol0003oooooooooo`1Pooooo`Go0000
6Oooool7o`0000oooooo0029ooooo`03o`000?oooooooooo067ooooo1?l0000Iooooo`Oo00003ooo
ool008Wooooo00?o0000ooooooooool0I?ooool2o`0001Sooooo1ol0000?ooooo`00R?ooool3o`00
06Gooooo0_l0000Iooooo`Go00004?ooool008Oooooo1Ol0001Uooooo`;o00006Oooool3o`00017o
oooo0026ooooo`Oo0000I?ooool3o`0002cooooo0026ooooo`Oo0000IOooool2o`0002cooooo0026
ooooo`Oo0000IOooool3o`0002_ooooo0027ooooo`Go0000I_ooool01?l0003oooooo`000?l0000Z
ooooo`00R?ooool3o`0006Sooooo0_l000000ooooooo0000o`000002o`0002Kooooo0029ooooo`03
o`000?oooooooooo06Oooooo2?l0000Uooooo`00ROooool00ol0003oooooooooo`1Wooooo`Wo0000
9?ooool008Wooooo00?o0000ooooooooool0I_ooool:o`0002Cooooo0029ooooo`03o`000?oooooo
oooo06Kooooo2_l0000Tooooo`00ROooool00ol0003oooooooooo`1Vooooo`Wo00009Oooool008[o
oooo00?o0000ooooooooool0I_ooool7o`0002Kooooo002:ooooo`03o`000?oooooooooo06Oooooo
1_l0000Vooooo`00R_ooool00ol0003oooooooooo`1/ooooo`03o`000?oooooooooo02Cooooo002:
ooooo`03o`000?oooooooooo06gooooo00?o0000ooooooooool08oooool008Wooooo0ol0001^oooo
o`;o00009?ooool008Sooooo1Ol0001^ooooo`03o`000?oooooooooo02;ooooo0027ooooo`Oo0000
KOooool2o`0002?ooooo0027ooooo`Oo0000K_ooool00ol0003oooooooooo`0Qooooo`00Qoooool7
o`0006kooooo0_l0000Rooooo`00R?ooool5o`00073ooooo1Ol0000Nooooo`00ROooool3o`0007;o
oooo1Ol0000Mooooo`00oOooool7o`0001cooooo003looooo`So00007?ooool00?cooooo2?l0000L
ooooo`00o?ooool7o`0001gooooo003mooooo`Go00007_ooool00?kooooo0ol0000Oooooo`00oooo
oolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooo
oolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {0.0970988, -0.0500054, \
0.00282287, 0.00383279}},
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],

Cell["\<\
Graphs can also be passed in, although the results will typically \
not look as elegant.\
\>", "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(WuenscheTree[\[IndentingNewLine]Cycle[15], RandomGraph[17, 0.2], 
          GridGraph[5, 5]\[IndentingNewLine]] // ShowGraph;\)\)], "Input",
  CellLabel->"In[18]:=",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.194108 1.23225 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.43633 .51056 m
.47458 .55944 L
s
.43633 .51056 m
.35982 .51056 L
s
.43633 .51056 m
.47458 .46168 L
s
.54084 .58766 m
.6071 .61588 L
s
.54084 .58766 m
.47458 .55944 L
s
.6071 .61588 m
.67336 .64411 L
s
.67336 .64411 m
.73961 .67233 L
s
.73961 .67233 m
.80587 .70055 L
s
.80587 .70055 m
.87213 .72877 L
s
.87213 .72877 m
.93839 .75699 L
s
.93839 .75699 m
.47458 .95455 L
s
.47458 .95455 m
.47458 .8981 L
s
.47458 .8981 m
.47458 .84166 L
s
.47458 .84166 m
.47458 .78522 L
s
.47458 .78522 m
.47458 .72877 L
s
.47458 .72877 m
.47458 .67233 L
s
.47458 .67233 m
.47458 .61588 L
s
.47458 .61588 m
.47458 .55944 L
s
.17512 .43074 m
.06161 .46673 L
s
.17512 .43074 m
.10782 .61919 L
s
.17512 .43074 m
.12102 .39083 L
s
.17512 .43074 m
.22922 .47065 L
s
.28331 .51056 m
.22922 .55047 L
s
.28331 .51056 m
.20941 .52517 L
s
.28331 .51056 m
.20941 .49595 L
s
.28331 .51056 m
.22922 .47065 L
s
.28331 .51056 m
.35982 .51056 L
s
.22922 .55047 m
.18264 .59525 L
s
.22922 .55047 m
.16852 .58483 L
s
.06161 .46673 m
.13551 .48134 L
s
.14153 .64285 m
.18264 .59525 L
s
.14153 .64285 m
.10043 .69046 L
s
.18264 .59525 m
.13607 .64003 L
s
.18264 .59525 m
.13095 .63687 L
s
.18264 .59525 m
.20941 .52517 L
s
.18264 .59525 m
.20941 .49595 L
s
.13607 .64003 m
.16852 .58483 L
s
.13607 .64003 m
.10782 .61919 L
s
.13607 .64003 m
.10043 .69046 L
s
.16852 .58483 m
.10782 .61919 L
s
.16852 .58483 m
.22922 .47065 L
s
.13551 .48134 m
.20941 .49595 L
s
.25631 .04545 m
.29015 .09608 L
s
.25631 .04545 m
.32713 .08478 L
s
.29015 .09608 m
.32399 .1467 L
s
.29015 .09608 m
.35173 .13822 L
s
.32399 .1467 m
.35783 .19732 L
s
.32399 .1467 m
.37632 .19167 L
s
.35783 .19732 m
.39167 .24794 L
s
.35783 .19732 m
.40091 .24512 L
s
.39167 .24794 m
.4255 .29857 L
s
.32713 .08478 m
.35173 .13822 L
s
.32713 .08478 m
.41484 .12687 L
s
.35173 .13822 m
.37632 .19167 L
s
.35173 .13822 m
.42482 .18283 L
s
.37632 .19167 m
.40091 .24512 L
s
.37632 .19167 m
.43481 .23879 L
s
.40091 .24512 m
.4255 .29857 L
s
.40091 .24512 m
.4448 .29475 L
s
.4255 .29857 m
.45478 .35071 L
s
.41484 .12687 m
.42482 .18283 L
s
.41484 .12687 m
.55379 .18715 L
s
.42482 .18283 m
.43481 .23879 L
s
.42482 .18283 m
.53399 .24167 L
s
.43481 .23879 m
.4448 .29475 L
s
.43481 .23879 m
.51419 .29619 L
s
.4448 .29475 m
.45478 .35071 L
s
.4448 .29475 m
.49439 .35071 L
s
.45478 .35071 m
.47458 .40523 L
s
.55379 .18715 m
.53399 .24167 L
s
.55379 .18715 m
.73961 .34879 L
s
.53399 .24167 m
.51419 .29619 L
s
.53399 .24167 m
.67336 .37701 L
s
.51419 .29619 m
.49439 .35071 L
s
.51419 .29619 m
.6071 .40523 L
s
.49439 .35071 m
.47458 .40523 L
s
.49439 .35071 m
.54084 .43346 L
s
.47458 .40523 m
.47458 .46168 L
s
.73961 .34879 m
.67336 .37701 L
s
.67336 .37701 m
.6071 .40523 L
s
.6071 .40523 m
.54084 .43346 L
s
.54084 .43346 m
.47458 .46168 L
s
.025 w
.43633 .51056 Mdot
.54084 .58766 Mdot
.6071 .61588 Mdot
.67336 .64411 Mdot
.73961 .67233 Mdot
.80587 .70055 Mdot
.87213 .72877 Mdot
.93839 .75699 Mdot
.47458 .95455 Mdot
.47458 .8981 Mdot
.47458 .84166 Mdot
.47458 .78522 Mdot
.47458 .72877 Mdot
.47458 .67233 Mdot
.47458 .61588 Mdot
.47458 .55944 Mdot
.17512 .43074 Mdot
.28331 .51056 Mdot
.22922 .55047 Mdot
.06161 .46673 Mdot
.14153 .64285 Mdot
.18264 .59525 Mdot
.13607 .64003 Mdot
.16852 .58483 Mdot
.13095 .63687 Mdot
.20941 .52517 Mdot
.10782 .61919 Mdot
.12102 .39083 Mdot
.13551 .48134 Mdot
.20941 .49595 Mdot
.22922 .47065 Mdot
.10043 .69046 Mdot
.35982 .51056 Mdot
.25631 .04545 Mdot
.29015 .09608 Mdot
.32399 .1467 Mdot
.35783 .19732 Mdot
.39167 .24794 Mdot
.32713 .08478 Mdot
.35173 .13822 Mdot
.37632 .19167 Mdot
.40091 .24512 Mdot
.4255 .29857 Mdot
.41484 .12687 Mdot
.42482 .18283 Mdot
.43481 .23879 Mdot
.4448 .29475 Mdot
.45478 .35071 Mdot
.55379 .18715 Mdot
.53399 .24167 Mdot
.51419 .29619 Mdot
.49439 .35071 Mdot
.47458 .40523 Mdot
.73961 .34879 Mdot
.67336 .37701 Mdot
.6071 .40523 Mdot
.54084 .43346 Mdot
.47458 .46168 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[18]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00B?ooool3o`000=Gooooo0017ooooo`Go0000e?ooool004Kooooo1ol0003Cooooo`00A_oo
ool7o`000=?ooooo0016ooooo`Oo0000doooool004Oooooo1Ol000000ooooooo0000o`00003Aoooo
o`00B?ooool4o`0000?ooooo00?o0000ooooooooool0c_ooool004cooooo00Co0000oooooooooooo
oooo0_l0003>ooooo`00C?ooool00ol0003oooooooooo`03ooooo`;o0000c?ooool004gooooo00?o
0000ooooooooool01?ooool2o`000<[ooooo001>ooooo`03o`000?oooooooooo00Gooooo0_l00038
ooooo`00C_ooool00ol0003oooooooooo`07ooooo`04o`000?ooooooooooooooo`?o0000`Oooool0
04oooooo00?o0000ooooooooool01oooool7o`000<3ooooo001@ooooo`03o`000?oooooooooo00Oo
oooo1ol0002oooooo`00DOooool4o`0000Gooooo1ol0002oooooo`00DOooool5o`0000Cooooo1ol0
002oooooo`00D?ooool7o`0000Cooooo1Ol000000ooooooo0000o`00002mooooo`00D?ooool7o`00
00Gooooo0ol00004ooooo`?o0000^_ooool0053ooooo1ol00008ooooo`03o`000?oooooooooo00Co
oooo0_l0002hooooo`00DOooool7o`0000Oooooo00?o0000ooooooooool01_ooool2o`000;Kooooo
001Booooo`Co00000_ooool00ol0003oooooooooo`05ooooo`03o`000?oooooooooo00Oooooo0_l0
002dooooo`00EOooool01Ol0003ooooooooooooooooo000000Kooooo00?o0000ooooooooool02Ooo
ool2o`000;;ooooo001Fooooo`04o`000?ooooooooooooooo`;o00001Oooool00ol0003ooooooooo
o`0:ooooo`;o0000/?ooool005Kooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`02
ooooo`03o`000?oooooooooo00cooooo0ol00002ooooo`?o0000Z?ooool005Oooooo00?o0000oooo
ooooool00oooool2o`0000;ooooo00?o0000ooooooooool03oooool6o`000:Oooooo001Hooooo`03
o`000?oooooooooo00Cooooo00Co0000oooooooooooo00004?ooool7o`000:Kooooo001Hooooo`03
o`000?oooooooooo00Gooooo00?o0000ooooool000000ol0000=ooooo`Oo0000Y_ooool005Wooooo
00?o0000ooooooooool01Oooool6o`0000cooooo2?l0002Uooooo`00FOooool00ol0003ooooooooo
o`05ooooo`Oo00003?ooool5o`0000;ooooo0ol0002Rooooo`00F_ooool4o`0000?ooooo1ol0000=
ooooo`?o00001_ooool2o`000:3ooooo001Jooooo`Go00000_ooool7o`0000oooooo00?o0000oooo
ooooool01_ooool2o`0009kooooo001Iooooo`Oo00000_ooool7o`0000oooooo00?o0000oooooooo
ool01oooool3o`0009_ooooo001Iooooo`Oo00000oooool3o`0000?ooooo0_l0000=ooooo`03o`00
0?oooooooooo00[ooooo0_l0002Iooooo`00FOooool7o`0000Kooooo00?o0000ooooooooool00_oo
ool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00cooooo0_l0002Gooooo`00F_ooool6
o`0000Kooooo00?o0000ooooooooool00oooool2o`0000[ooooo00?o0000ooooooooool03_ooool3
o`0009Cooooo001Kooooo`Co000000?oooooo`000?ooool01Oooool00ol0003oooooooooo`04oooo
o`;o00002?ooool00ol0003oooooooooo`0Aooooo`;o0000T_ooool005oooooo00Co0000ooooool0
003o00001?ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo00Gooooo00?o0000oooo
ooooool04oooool2o`00093ooooo001Oooooo`05o`000?ooooooooooooooool000001?ooool00ol0
003oooooooooo`06ooooo`;o00001Oooool00ol0003oooooooooo`0Eooooo`;o0000S_ooool0063o
oooo00Go0000ooooooooooooooooo`000003ooooo`03o`000?oooooooooo00Sooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`0Gooooo`?o0000Roooool0067ooooo00Go0000oooooooo
ooooooooo`000003ooooo`03o`000?oooooooooo00Sooooo0_l00002ooooo`?o00006_ooool2o`00
08Wooooo001Qooooo`03o`000?oooooooooo00;ooooo00Co0000oooooooooooo00003?ooool6o`00
01_ooooo0_l00004ooooo`?o0000P?ooool006;ooooo00?o0000ooooooooool00_ooool2o`000003
ooooool0003ooooo00[ooooo1ol0000Looooo`So0000Ooooool006?ooooo00?o0000ooooooooool0
0oooool4o`0000Wooooo1ol0000Nooooo`Oo0000O_ooool006Cooooo1?l000000ooooooo0000o`00
0003o`0000Sooooo1ol0000Nooooo`Oo0000O_ooool006Cooooo2ol00008ooooo`Go000000?ooooo
o`000?l000007?ooool7o`0007kooooo001Sooooo`co00002Oooool3o`0000Cooooo0_l0000Koooo
o`Ko0000O_ooool006?ooooo3?l0000;ooooo`03o`000?oooooooooo00Cooooo00?o0000oooooooo
ool06Oooool3o`0000;ooooo00?o0000ooooooooool0Noooool006?ooooo2ol000000ooooooo0000
ooooo`09ooooo`03o`000?oooooooooo00Gooooo0_l0000Iooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool0N_ooool006Cooooo1Ol000000ooooooo0000o`000002o`0000;ooooo00?o
0000ooooooooool01oooool00ol0003oooooooooo`07ooooo`;o00005oooool00ol0003ooooooooo
o`04ooooo`03o`000?oooooooooo07Wooooo001Uooooo`Co00001Oooool00ol0003oooooooooo`02
o`0000Oooooo00?o0000ooooooooool02Oooool2o`0001Cooooo00?o0000ooooooooool01_ooool0
0ol0003oooooooooo`1hooooo`00J?ooool2o`0000Cooooo00?o0000ooooooooool00_ooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo00_ooooo0_l0000Booooo`03o`000?oooooooooo
00Oooooo0_l0001hooooo`00JOooool2o`0000Cooooo00?o0000ooooooooool00_ooool00ol0003o
ooooooooo`04ooooo`03o`000?oooooooooo00cooooo0_l0000@ooooo`03o`000?oooooooooo00Wo
oooo00?o0000ooooooooool0MOooool006[ooooo0_l00003ooooo`03o`000?oooooooooo00?ooooo
00?o0000ooooooooool00oooool00ol0003oooooooooo`0>ooooo`;o00003Oooool00ol0003ooooo
ooooo`0;ooooo`03o`000?oooooooooo07Cooooo001Zooooo`03o`000?oooooo000000?ooooo00?o
0000ooooooooool00oooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo013ooooo0_l0
000;ooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool0Loooool006_ooooo00?o0000
ooooool000000_ooool00ol0003oooooooooo`04ooooo`;o00000_ooool00ol0003oooooooooo`0B
ooooo`;o00002Oooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo07;ooooo001/oooo
o`05o`000?oooooo0000ooooool000002?ooool00ol0003oooooo`000002o`0001Cooooo00?o0000
ooooooooool01Oooool00ol0003oooooooooo`0?ooooo`03o`000?oooooooooo077ooooo001/oooo
o`06o`000?ooooooooooo`000?oooooo00002?ooool5o`0001Cooooo0_l00003ooooo`?o00004_oo
ool2o`00077ooooo001]ooooo`03o`000?oooooooooo00Go00001?ooool7o`0001Gooooo1ol0000C
ooooo`03o`000?oooooooooo06kooooo001^ooooo`So00000oooool7o`0001Kooooo1ol0000Coooo
o`03o`000?oooooooooo06gooooo001^ooooo`Wo00000_ooool7o`0001Kooooo1ol0000Dooooo`03
o`000?oooooooooo06cooooo001]ooooo`[o00000oooool6o`0001Kooooo1ol0000Eooooo`03o`00
0?oooooooooo06_ooooo001]ooooo`[o00001?ooool3o`0000;ooooo0_l0000Eooooo`Ko00005_oo
ool00ol0003oooooooooo`1Zooooo`00KOooool9o`0000Oooooo00?o0000ooooooooool00_ooool0
0ol0003oooooooooo`0Cooooo`?o00000_ooool00ol0003oooooooooo`0Dooooo`03o`000?oooooo
oooo06Wooooo001^ooooo`So00001oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo
01;ooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0Dooooo`;o0000JOooool006oo
oooo1?l00002ooooo`;o00001_ooool00ol0003oooooooooo`04ooooo`;o00004Oooool00ol0003o
ooooooooo`05ooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool0I_ooool007?ooooo
00Go0000ooooool0003oooooo`000005ooooo`03o`000?oooooooooo00Kooooo00?o0000oooooooo
ool03_ooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo01Gooooo00?o0000oooooooo
ool0IOooool007?ooooo00Ko0000oooooooooooo0000ooooool00004ooooo`03o`000?oooooooooo
00Oooooo0_l0000>ooooo`03o`000?oooooooooo00Oooooo00?o0000ooooooooool05Oooool00ol0
003oooooooooo`1Tooooo`00M?ooool00ol0003oooooo`000002ooooo`03o`000?oooooooooo00;o
oooo00?o0000ooooooooool02?ooool00ol0003oooooooooo`0:ooooo`03o`000?oooooooooo00Wo
oooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`1Sooooo`00MOooool01Ol0003ooooo
o`000?oooooo000000Cooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`09ooooo`03
o`000?oooooooooo00[ooooo00?o0000ooooooooool05Oooool00ol0003oooooooooo`1Rooooo`00
MOooool00ol0003oooooo`000002ooooo`05o`000?ooooooooooooooool000003?ooool2o`0000So
oooo00?o0000ooooooooool03?ooool00ol0003oooooooooo`0Eooooo`03o`000?oooooooooo067o
oooo001fooooo`;o00000oooool01?l0003oooooooooool0000>ooooo`03o`000?oooooooooo00Go
oooo00?o0000ooooooooool03Oooool00ol0003oooooooooo`0Eooooo`;o0000HOooool007Oooooo
0_l00003ooooo`03o`000?oooooo000000;o00003Oooool00ol0003oooooooooo`04ooooo`03o`00
0?oooooooooo00kooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`1Nooooo`00N?oo
ool4o`000003ooooool0003o000000?o00003Oooool2o`000004ooooool0003o0000o`0001;ooooo
00?o0000ooooooooool05_ooool00ol0003oooooooooo`1Mooooo`00N?ooool;o`0000kooooo1Ol0
000Booooo`03o`000?oooooooooo01Kooooo00?o0000ooooooooool0G?ooool007Oooooo3?l0000=
ooooo`Oo00004_ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo05_ooooo001goooo
o`co00003Oooool7o`0001?ooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`1Joooo
o`00Moooool;o`0000kooooo1ol0000Dooooo`03o`000?oooooooooo01Kooooo00?o0000oooooooo
ool0FOooool007Sooooo1Ol000000ooooooo0000o`000002o`0000oooooo1Ol0000Fooooo`03o`00
0?oooooooooo01Kooooo0_l0001Iooooo`00NOooool3o`0000Cooooo00?o0000ooooool000003ooo
ool4o`0001Oooooo00?o0000ooooooooool05oooool00ol0003oooooooooo`1Fooooo`00O?ooool0
1Ol0003ooooooooooooooooo000000;ooooo00?o0000ooooooooool03?ooool01Ol0003ooooooooo
oooooooo000001Oooooo00?o0000ooooooooool05oooool00ol0003oooooooooo`1Eooooo`00O?oo
ool01Ol0003ooooooooooooooooo000000?ooooo00?o0000ooooooooool02_ooool00ol0003ooooo
ooooo`03ooooo`03o`000?oooooooooo01Gooooo00?o0000ooooooooool05oooool00ol0003ooooo
ooooo`1Dooooo`00OOooool01?l0003oooooooooool00004ooooo`03o`000?oooooooooo00Wooooo
00?o0000ooooooooool01?ooool00ol0003oooooooooo`0Eooooo`03o`000?oooooooooo01Oooooo
00?o0000ooooooooool0Doooool007gooooo00Co0000oooooooooooo00001Oooool00ol0003ooooo
ooooo`07ooooo`03o`000?oooooooooo00Kooooo00?o0000ooooooooool05Oooool00ol0003ooooo
ooooo`0Gooooo`03o`000?oooooooooo05;ooooo001nooooo`04o`000?ooooooooooo`0000Cooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo01Gooooo
00?o0000ooooooooool05oooool2o`0005;ooooo001oooooo`03o`000?oooooo000000Gooooo00?o
0000ooooooooool01_ooool00ol0003oooooooooo`07ooooo`03o`000?oooooooooo01Kooooo00?o
0000ooooooooool06?ooool00ol0003oooooooooo`1?ooooo`00Ooooool00ol0003oooooo`000006
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool02Oooool00ol0003oooooooooo`0F
ooooo`03o`000?oooooooooo01Sooooo00?o0000ooooooooool0C_ooool0083ooooo0_l00007oooo
o`03o`000?oooooooooo00?ooooo00?o0000ooooooooool02_ooool00ol0003oooooooooo`0Foooo
o`03o`000?oooooooooo01Sooooo00?o0000ooooooooool0COooool0083ooooo1?l00006ooooo`03
o`000?oooooo000000;o00003_ooool00ol0003oooooooooo`0Fooooo`03o`000?oooooooooo01So
oooo00?o0000ooooool000000_l0001:ooooo`00P?ooool5o`0000Kooooo1Ol0000>ooooo`03o`00
0?oooooooooo01Kooooo00?o0000ooooooooool06?ooool5o`0004Wooooo001oooooo`Oo00001?oo
ool7o`0000kooooo00?o0000ooooooooool05_ooool00ol0003oooooooooo`0Fooooo`Oo0000B?oo
ool007oooooo1ol00004ooooo`Oo00003oooool00ol0003oooooooooo`0Fooooo`03o`000?oooooo
oooo01Gooooo1ol00018ooooo`00Ooooool7o`0000Cooooo1ol0000@ooooo`03o`000?oooooooooo
01Kooooo00?o0000ooooooooool04oooool8o`0004Sooooo0020ooooo`Go00001_ooool5o`00017o
oooo00?o0000ooooooooool05oooool00ol0003oooooooooo`0@ooooo`;o00000_ooool5o`0004Wo
oooo0021ooooo`?o00002?ooool3o`0001?ooooo00?o0000ooooooooool05oooool00ol0003ooooo
ooooo`0=ooooo`;o00001Oooool3o`0004[ooooo0023ooooo`03o`000?oooooooooo00Kooooo00Co
0000oooooooooooo00004oooool00ol0003oooooooooo`0Gooooo`03o`000?oooooooooo00Wooooo
0ol0001Dooooo`00Q?ooool00ol0003oooooooooo`05ooooo`05o`000?ooooooooooooooool00000
4oooool00ol0003oooooooooo`0Gooooo`03o`000?oooooo000000;o00001?ooool2o`0005Oooooo
0024ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool00_ooool00ol0003ooooooooo
o`0Booooo`03o`000?oooooooooo01Oooooo1Ol000000ooooooo0000o`00001Iooooo`00Q?ooool0
0ol0003oooooooooo`04ooooo`03o`000?oooooooooo00?ooooo00?o0000ooooooooool04_ooool0
0ol0003oooooooooo`0Eooooo`Oo0000Foooool008Gooooo00?o0000ooooooooool00oooool00ol0
003oooooooooo`03ooooo`03o`000?oooooooooo01?ooooo00?o0000ooooooooool05?ooool7o`00
05_ooooo000Qooooo`?o0000HOooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo00Go
oooo00?o0000ooooooooool04_ooool00ol0003oooooooooo`0Cooooo`So0000Foooool0023ooooo
1Ol0001Qooooo`05o`000?ooooooooooooooool000002?ooool00ol0003oooooooooo`0Booooo`03
o`000?oooooooooo013ooooo0_l00002ooooo`Go0000G?ooool001oooooo1ol0001Pooooo`05o`00
0?ooooooooooooooool000002?ooool00ol0003oooooooooo`0Cooooo`03o`000?oooooooooo00go
oooo0_l00005ooooo`?o0000GOooool001oooooo1ol0001Pooooo`04o`000?ooooooooooo`0000[o
oooo00?o0000ooooooooool04oooool00ol0003oooooooooo`09ooooo`?o0000Ioooool001oooooo
1ol0001Qooooo`?o00002_ooool00ol0003oooooooooo`0Dooooo`03o`000?oooooo000000;o0000
1?ooool2o`0006[ooooo000Pooooo`Oo0000Goooool5o`0000[ooooo00?o0000ooooooooool05?oo
ool5o`000003ooooool0003o000006cooooo000Qooooo`?o00000oooool00ol0003oooooooooo`1K
ooooo`Oo00002Oooool00ol0003oooooooooo`0Cooooo`Oo0000K_ooool002Sooooo0_l0001Koooo
o`Oo00002_ooool00ol0003oooooooooo`0Booooo`Oo0000K_ooool002[ooooo00?o0000oooooooo
ool0F?ooool7o`0000_ooooo00?o0000ooooooooool04?ooool8o`0006kooooo000[ooooo`03o`00
0?oooooooooo05Sooooo1Ol0000<ooooo`03o`000?oooooooooo00kooooo0_l00002ooooo`Go0000
Koooool002cooooo0_l0001Iooooo`?o00003_ooool00ol0003oooooooooo`0;ooooo`;o00001Ooo
ool3o`00073ooooo000^ooooo`03o`000?oooooooooo00?o0000E?ooool00ol0003oooooooooo`0=
ooooo`03o`000?oooooooooo00Sooooo0ol0001jooooo`00;oooool6o`0005?ooooo00?o0000oooo
ooooool03_ooool4o`0000Cooooo0_l0001mooooo`00;oooool7o`0005;ooooo00?o0000oooooooo
ool03_ooool5o`000003ooooool0003o000007oooooo000_ooooo`Oo0000D_ooool00ol0003ooooo
ooooo`0=ooooo`Oo0000POooool002gooooo2Ol0001Booooo`03o`000?oooooooooo00gooooo1ol0
0021ooooo`00:_ooool3o`0000?ooooo1_l0001Booooo`03o`000?oooooooooo00cooooo2?l00021
ooooo`009_ooool4o`0000Oooooo0ol00002ooooo`03o`000?oooooooooo04oooooo00?o0000oooo
ooooool02_ooool2o`0000;ooooo1Ol00022ooooo`008oooool3o`0000_ooooo00?o0000oooooooo
ool00oooool2o`0004oooooo00?o0000ooooooooool02?ooool2o`0000Gooooo0ol00023ooooo`00
8?ooool3o`0000gooooo00?o0000ooooooooool01_ooool00ol0003oooooooooo`1<ooooo`03o`00
0?oooooooooo00Gooooo0ol0002=ooooo`007Oooool3o`00013ooooo00?o0000ooooooooool01ooo
ool00ol0003oooooooooo`1:ooooo`?o00001?ooool2o`00093ooooo000Iooooo`Co00004oooool0
0ol0003oooooooooo`08ooooo`03o`000?oooooooooo04Sooooo1Ol000000ooooooo0000o`00002B
ooooo`004?ooool3o`0000?ooooo0ol0000Fooooo`03o`000?oooooooooo00[ooooo0_l00017oooo
o`Oo0000U?ooool000oooooo1ol0000Iooooo`03o`000?oooooooooo00cooooo00?o0000ooooool0
00000_l00012ooooo`Oo0000U?ooool000kooooo1ol0000Iooooo`03o`000?oooooooooo00kooooo
1Ol00011ooooo`Oo0000U?ooool000kooooo2_l0000Fooooo`03o`000?oooooooooo00gooooo1ol0
0011ooooo`Go0000UOooool000kooooo1ol00003ooooo`Go00002?ooool3o`0000Kooooo00?o0000
ooooooooool03Oooool7o`00047ooooo1?l0002Fooooo`003oooool5o`0000Wooooo1?l00003oooo
o`Go00001?ooool00ol0003oooooooooo`0>ooooo`Oo0000@?ooool00ol0003oooooooooo`2Hoooo
o`004?ooool3o`0000kooooo2Ol00003ooooo`03o`000?oooooooooo00oooooo1ol0000nooooo`03
o`000?oooooooooo09Wooooo000Sooooo`Oo00000oooool00ol0003oooooooooo`0@ooooo`?o0000
0oooool00ol0003oooooooooo`0kooooo`03o`000?oooooooooo09Wooooo000Sooooo`co00003?oo
ool3o`000003ooooool0003ooooo00Kooooo0_l0000jooooo`03o`000?oooooooooo09[ooooo000T
ooooo`Go00000oooool00ol0003oooooooooo`06o`0000Gooooo1_l00009ooooo`03o`000?oooooo
oooo03Kooooo00?o0000ooooooooool0Voooool002Gooooo0ol00004ooooo`03o`000?oooooooooo
00Kooooo2ol0000:ooooo`03o`000?oooooooooo03Cooooo00?o0000ooooooooool0W?ooool002_o
oooo00?o0000ooooooooool02oooool7o`0000_ooooo0_l0000dooooo`03o`000?oooooooooo09co
oooo000[ooooo`03o`000?oooooooooo00_ooooo3?l00008ooooo`03o`000?oooooooooo00?o0000
4oooool3o`0001?ooooo0ol000000ooooooo0000ooooo`2Nooooo`00:_ooool00ol0003ooooooooo
o`0=ooooo`Go00001_ooool5o`0000Cooooo1_l0000Aooooo`Go00004Oooool5o`000:3ooooo000Z
ooooo`03o`000?oooooooooo00kooooo0ol0000<ooooo`_o00003oooool7o`0000oooooo1ol0002O
ooooo`00:_ooool00ol0003oooooooooo`0>ooooo`03o`000?oooooooooo013ooooo<ol0002Ooooo
o`00:Oooool00ol0003oooooooooo`0?ooooo`?o00002oooool<o`0000oooooo1ol0000?ooooo`Oo
0000Woooool002Wooooo00?o0000ooooooooool03_ooool5o`0000Gooooo1Ol00004ooooo`Oo0000
4Oooool5o`00017ooooo1Ol0002Pooooo`00:Oooool00ol0003oooooooooo`0=ooooo`_o00002?oo
ool01?l0003oooooooooooooool3o`0001?ooooo0ol0000Cooooo`Co0000X?ooool002Sooooo00?o
0000ooooooooool03_ooool7o`0000[ooooo0_l0000dooooo`03o`000?oooooooooo09gooooo000X
ooooo`03o`000?oooooooooo00kooooo1ol00009ooooo`03o`000?oooooooooo03Gooooo00?o0000
ooooooooool0W?ooool002Oooooo00?o0000ooooooooool03_ooool7o`0000Wooooo00?o0000oooo
ooooool0=_ooool00ol0003oooooooooo`2Looooo`009oooool00ol0003oooooooooo`0>ooooo`;o
000000Coooooo`000?l0003o00002?ooool2o`0003[ooooo00?o0000ooooooooool0Voooool002Oo
oooo00?o0000ooooooooool03Oooool01Ol0003oooooo`000?oooooo000000Wooooo00?o0000oooo
ooooool0>oooool00ol0003oooooooooo`2Jooooo`009_ooool00ol0003oooooooooo`0>ooooo`;o
000000?oooooo`000?ooool01?ooool5o`0003oooooo00?o0000ooooooooool0VOooool002Kooooo
00?o0000ooooooooool03Oooool01Ol0003oooooo`000?oooooo000000Cooooo1Ol00010ooooo`03
o`000?oooooooooo09Wooooo000Vooooo`03o`000?oooooooooo00gooooo00Co0000ooooool0003o
00001?ooool7o`00043ooooo00?o0000ooooool000000_l0002Fooooo`009Oooool00ol0003ooooo
ooooo`0=ooooo`03o`000?oooooooooo00;o00001?ooool7o`00047ooooo1Ol0002Eooooo`009Ooo
ool00ol0003oooooooooo`0=ooooo`05o`000?oooooo0000ooooool000000oooool8o`00043ooooo
1ol0002Dooooo`009?ooool00ol0003oooooooooo`0=ooooo`03o`000?oooooooooo00;o00000_oo
ool2o`000003ooooool0003o000000Co0000@Oooool7o`0009Cooooo000Tooooo`03o`000?oooooo
oooo00cooooo00Co0000oooooooooooooooo1?l00002ooooo`03o`000?oooooooooo00?o0000@_oo
ool8o`0009?ooooo000Tooooo`03o`000?oooooooooo00cooooo00?o0000ooooooooool00ol00003
ooooo`03o`000?oooooooooo04Oooooo1Ol00002ooooo`;o0000TOooool002?ooooo00?o0000oooo
ooooool03?ooool01?l0003oooooooooooooool2o`0000?ooooo00?o0000ooooooooool0BOooool3
o`0000Gooooo0_l0002?ooooo`008oooool00ol0003oooooooooo`09ooooo`Co000000?oooooo`00
0?l000000_l00002ooooo`03o`000?oooooooooo04_ooooo00?o0000ooooooooool01_ooool3o`00
08cooooo000Sooooo`03o`000?oooooooooo00Sooooo1_l00002ooooo`04o`000?ooooooooooo`00
04kooooo00?o0000ooooooooool02Oooool2o`0000Cooooo0ol00023ooooo`008_ooool00ol0003o
ooooooooo`08ooooo`Oo000000Goooooo`000?l0003oooooo`00001?ooooo`03o`000?oooooooooo
00_ooooo0_l000000ooooooo0000o`000003o`0008;ooooo000Rooooo`03o`000?oooooooooo00So
oooo2Ol000000ooooooo0000ooooo`1?ooooo`03o`000?oooooooooo00gooooo1ol00021ooooo`00
8_ooool00ol0003oooooooooo`08ooooo`[o0000DOooool00ol0003oooooooooo`0=ooooo`Oo0000
POooool0027ooooo00?o0000ooooooooool01oooool2o`000003ooooool0003o000000So0000D?oo
ool00ol0003oooooooooo`0=ooooo`So0000P?ooool0027ooooo00?o0000ooooooooool01_ooool0
0ol0003oooooooooo`02ooooo`Wo0000D?ooool00ol0003oooooooooo`0>ooooo`Go00000_ooool2
o`0007kooooo000Pooooo`03o`000?oooooooooo00Gooooo0_l00004ooooo`03o`000?oooooooooo
00Oo0000D?ooool00ol0003oooooooooo`0?ooooo`?o00001Oooool2o`0007cooooo000Pooooo`03
o`000?oooooooooo00?ooooo0_l00005ooooo`04o`000?ooooooooooooooo`Ko0000DOooool00ol0
003oooooooooo`0Iooooo`?o0000NOooool0023ooooo00Co0000oooooooooooooooo0_l00007oooo
o`03o`000?oooooooooo00Ko0000DOooool3o`0001gooooo0_l00004ooooo`?o0000L?ooool001go
oooo0ol00002ooooo`;o00002?ooool00ol0003oooooo`000003o`0005Cooooo1Ol0000Nooooo`;o
000000?oooooo`000?l000000ol0001_ooooo`007?ooool6o`0000[ooooo0_l000000ooooooo0000
o`00001Dooooo`Oo00007oooool7o`0006kooooo000Kooooo`Oo00002Oooool2o`000003ooooool0
003o000005Gooooo1ol0000Oooooo`Oo0000K_ooool001_ooooo1ol00009ooooo`04o`000?oooooo
0000o`0005Kooooo1ol0000Oooooo`So0000KOooool001_ooooo1ol00007ooooo`;o000000Cooooo
o`000?oooooo0000Eoooool5o`00027ooooo1Ol00002ooooo`;o0000Joooool001cooooo1ol00000
1ooooooo0000o`000?l0003oooooo`000?ooool00_l000000ooooooo0000ooooo`1Hooooo`?o0000
8oooool3o`0000Gooooo0_l0001Yooooo`007Oooool3o`0000?ooooo1Ol000001Ooooooo0000o`00
0?oooooo000005_ooooo00?o0000ooooooooool0;?ooool3o`0006Kooooo000Rooooo`So000000?o
ooooo`000?ooool0Foooool00ol0003oooooooooo`0_ooooo`;o00001?ooool3o`0005gooooo000R
ooooo`Wo0000GOooool00ol0003oooooooooo`0aooooo`;o000000?oooooo`000?l000000ol0001L
ooooo`008_ooool:o`0005cooooo00?o0000ooooooooool0<oooool7o`0005_ooooo000Sooooo`Wo
0000G?ooool00ol0003oooooooooo`0cooooo`Oo0000Foooool002Cooooo2?l0001Looooo`03o`00
0?oooooooooo03?ooooo2?l0001Jooooo`009Oooool6o`0005gooooo00?o0000ooooooooool0=?oo
ool5o`0000;ooooo0_l0001Hooooo`009?ooool00ol0003oooooo`000003o`0005kooooo00?o0000
ooooooooool0=Oooool3o`0000Gooooo0_l0001Fooooo`009?ooool2o`0006;ooooo00?o0000oooo
ooooool0?oooool3o`0005?ooooo000Sooooo`;o0000Hoooool00ol0003oooooooooo`12ooooo`;o
0000DOooool002;ooooo0_l0001Sooooo`?o0000AOooool2o`0000;ooooo0ol0001:ooooo`008Ooo
ool2o`0006?ooooo1Ol00016ooooo`Ko0000BOooool0027ooooo0_l0001Rooooo`Oo0000AOooool7
o`0004Sooooo000Pooooo`;o0000Hoooool7o`0004Gooooo1ol00018ooooo`007oooool2o`0006Co
oooo1ol00015ooooo`Wo0000A_ooool001_ooooo0ol000000ooooooo0000ooooo`1Uooooo`Go0000
Aoooool5o`0000?ooooo0_l00014ooooo`006_ooool5o`0006Sooooo0ol00019ooooo`?o00001_oo
ool3o`00047ooooo000Iooooo`Oo0000J?ooool00ol0003oooooooooo`1Dooooo`;o0000?oooool0
01Wooooo1ol0001Xooooo`03o`000?oooooooooo05Kooooo0_l00003ooooo`?o0000=oooool001Wo
oooo1ol0001Xooooo`03o`000?oooooooooo05Sooooo1ol0000fooooo`006_ooool5o`0006Wooooo
00?o0000ooooooooool0FOooool7o`0003Gooooo000Kooooo`?o0000J_ooool00ol0003ooooooooo
o`1Iooooo`Oo0000=Oooool008Sooooo00?o0000ooooooooool0FOooool8o`0003Cooooo0028oooo
o`03o`000?oooooooooo05[ooooo1Ol00002ooooo`;o0000<_ooool008Sooooo00?o0000oooooooo
ool0Foooool3o`0000Gooooo0_l0000`ooooo`00R?ooool00ol0003oooooooooo`1Uooooo`?o0000
;Oooool008Oooooo0ol0001Yooooo`;o00001?ooool3o`0002Cooooo0026ooooo`Go0000J_ooool2
o`000003ooooool0003o000000?o00008oooool008Gooooo1ol0001[ooooo`Oo00008_ooool008Go
oooo1ol0001[ooooo`Oo00008_ooool008Gooooo1ol0001[ooooo`So00008Oooool008Kooooo1Ol0
001]ooooo`Go00000_ooool2o`0001oooooo0027ooooo`?o0000Koooool3o`0000Gooooo0_l0000M
ooooo`00R?ooool00ol0003oooooooooo`1hooooo`?o00006_ooool008Sooooo00?o0000oooooooo
ool0Noooool2o`0000Cooooo0ol0000Aooooo`00R?ooool00ol0003oooooooooo`1mooooo`;o0000
00?oooooo`000?l000000ol0000@ooooo`00R?ooool00ol0003oooooooooo`1oooooo`Oo00003ooo
ool008Sooooo00?o0000ooooooooool0Ooooool7o`0000oooooo0028ooooo`03o`000?oooooooooo
07kooooo2?l0000?ooooo`00R?ooool00ol0003oooooooooo`1looooo`;o00000_ooool5o`00013o
oooo0028ooooo`03o`000?oooooooooo07[ooooo0_l00005ooooo`?o00004Oooool008Sooooo00?o
0000ooooooooool0Moooool3o`0001_ooooo0027ooooo`?o0000M_ooool2o`0001kooooo0026oooo
o`Go0000Loooool2o`00023ooooo0025ooooo`Oo0000Koooool3o`0002;ooooo0025ooooo`Oo0000
KOooool2o`0002Gooooo0025ooooo`Oo0000J_ooool3o`0002Oooooo0026ooooo`Go0000JOooool2
o`0002[ooooo0027ooooo`?o0000J?ooool2o`0002cooooo0028ooooo`03o`000?oooooooooo06Co
oooo0ol0000^ooooo`00R?ooool00ol0003oooooooooo`1Rooooo`;o0000<Oooool008Sooooo00?o
0000ooooooooool0H?ooool2o`0003?ooooo0028ooooo`03o`000?oooooooooo05gooooo0ol0000e
ooooo`00R?ooool00ol0003oooooooooo`1Kooooo`;o0000>?ooool008Sooooo00?o0000oooooooo
ool0F?ooool3o`0003[ooooo0028ooooo`03o`000?oooooooooo05Kooooo0_l0000mooooo`00R?oo
ool00ol0003oooooooooo`1Dooooo`;o0000?oooool008Sooooo00?o0000ooooooooool0DOooool3
o`00047ooooo0027ooooo`?o0000D?ooool2o`0004Cooooo0026ooooo`Go0000COooool2o`0004Ko
oooo0025ooooo`Oo0000BOooool3o`0004Sooooo0025ooooo`Oo0000Aoooool2o`0004_ooooo0025
ooooo`Oo0000A?ooool3o`0004gooooo0026ooooo`Go0000@oooool2o`00053ooooo0027ooooo`?o
0000@_ooool2o`0005;ooooo0028ooooo`03o`000?oooooooooo03kooooo0ol0001Dooooo`00R?oo
ool00ol0003oooooooooo`0looooo`;o0000Eoooool008Sooooo00?o0000ooooooooool0>_ooool2
o`0005Wooooo0028ooooo`03o`000?oooooooooo03Oooooo0ol0001Kooooo`00R?ooool00ol0003o
ooooooooo`0eooooo`;o0000G_ooool008Sooooo00?o0000ooooooooool0<_ooool3o`00063ooooo
0028ooooo`03o`000?oooooooooo033ooooo0_l0001Sooooo`00R?ooool00ol0003oooooooooo`0^
ooooo`;o0000IOooool008Sooooo00?o0000ooooooooool0:oooool3o`0006Oooooo0027ooooo`?o
0000:_ooool2o`0006[ooooo0026ooooo`Go00009oooool2o`0006cooooo0025ooooo`Oo00008ooo
ool3o`0006kooooo0025ooooo`Oo00008Oooool2o`00077ooooo0025ooooo`Oo00007_ooool3o`00
07?ooooo0026ooooo`Go00007Oooool2o`0007Kooooo0027ooooo`?o00007?ooool2o`0007Sooooo
0028ooooo`03o`000?oooooooooo01Sooooo0ol0001jooooo`00R?ooool00ol0003oooooooooo`0F
ooooo`;o0000OOooool008Sooooo00?o0000ooooooooool05?ooool2o`0007oooooo0028ooooo`03
o`000?oooooooooo017ooooo0ol00021ooooo`00R?ooool00ol0003oooooooooo`0?ooooo`;o0000
Q?ooool008Sooooo00?o0000ooooooooool03?ooool3o`0008Kooooo0028ooooo`03o`000?oooooo
oooo00[ooooo0_l00029ooooo`00R?ooool00ol0003oooooooooo`08ooooo`;o0000Roooool008So
oooo00?o0000ooooooooool01Oooool3o`0008gooooo0028ooooo`03o`000?oooooooooo00?ooooo
0_l0002@ooooo`00Qoooool3o`0000;ooooo0_l0002Booooo`00Q_ooool6o`0009Cooooo0025oooo
o`Oo0000U?ooool008Gooooo1ol0002Dooooo`00QOooool7o`0009Cooooo0026ooooo`Go0000UOoo
ool008Oooooo0ol0002Fooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooo
oolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooo
oolQooooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {0.157519, -0.0500054, \
0.00282764, 0.00383279}},
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],

Cell["\<\
An alternative form accepts a list of integers or graphs as the \
first argument. This form is easier to use when the list of tree sizes (or \
the list of graphs) is being generated programmatically.\
\>", "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(WuenscheTree[{21, 17, 13, 17, 24}] // ShowGraph;\)\)], "Input",
  CellLabel->"In[21]:=",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0288161 0.996776 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 0 0 r
.005 w
[ ] 0 setdash
.39974 .38375 m
.44692 .41501 L
s
.39974 .38375 m
.38172 .43433 L
s
.39974 .38375 m
.34142 .38375 L
s
.39974 .38375 m
.38172 .33317 L
s
.39974 .38375 m
.44692 .35249 L
s
.50396 .42607 m
.56101 .43713 L
s
.50396 .42607 m
.44692 .41501 L
s
.4941 .44627 m
.54128 .47754 L
s
.4941 .44627 m
.44692 .41501 L
s
.78918 .48136 m
.73213 .4703 L
s
.56101 .43713 m
.61805 .44819 L
s
.67622 .4519 m
.7344 .45561 L
s
.67622 .4519 m
.61805 .44819 L
s
.47608 .46107 m
.50524 .50713 L
s
.47608 .46107 m
.44692 .41501 L
s
.54128 .47754 m
.58846 .5088 L
s
.72831 .48281 m
.67285 .46638 L
s
.72831 .48281 m
.78377 .49925 L
s
.72692 .4863 m
.67285 .46638 L
s
.95016 .54856 m
.8947 .53212 L
s
.8947 .53212 m
.83924 .51568 L
s
.83924 .51568 m
.78377 .49925 L
s
.73213 .4703 m
.67509 .45924 L
s
.61805 .44819 m
.67509 .45924 L
s
.61805 .44819 m
.67285 .46638 L
s
.39084 .48687 m
.38172 .43433 L
s
.2975 .85259 m
.30888 .80043 L
s
.34523 .64446 m
.35435 .59193 L
s
.34523 .64446 m
.33611 .69699 L
s
.2636 .95455 m
.27832 .90308 L
s
.3726 .48687 m
.36347 .5394 L
s
.3726 .48687 m
.38172 .43433 L
s
.35524 .48172 m
.38172 .43433 L
s
.29305 .85162 m
.30888 .80043 L
s
.29305 .85162 m
.27832 .90308 L
s
.29305 .85162 m
.27612 .90251 L
s
.33153 .75002 m
.33611 .69699 L
s
.36347 .5394 m
.35435 .59193 L
s
.30888 .80043 m
.32249 .74871 L
s
.32249 .74871 m
.33611 .69699 L
s
.34048 .47194 m
.38172 .43433 L
s
.22479 .38375 m
.16647 .38375 L
s
.22479 .38375 m
.28311 .38375 L
s
.13983 .48991 m
.18874 .46094 L
s
.18874 .46094 m
.23764 .43198 L
s
.16647 .38375 m
.10816 .38375 L
s
.28815 .40538 m
.23764 .43198 L
s
.28815 .40538 m
.23269 .42182 L
s
.28815 .40538 m
.34142 .38375 L
s
.18568 .45612 m
.23764 .43198 L
s
.04984 .38375 m
.10816 .38375 L
s
.28311 .38375 m
.34142 .38375 L
s
.28815 .36212 m
.34142 .38375 L
s
.34744 .29014 m
.38172 .33317 L
s
.40894 .17656 m
.42077 .12447 L
s
.40894 .17656 m
.39622 .22846 L
s
.40894 .17656 m
.42256 .12484 L
s
.36722 .22846 m
.38172 .27998 L
s
.36529 .17468 m
.37441 .22721 L
s
.37989 .12044 m
.38172 .1736 L
s
.37441 .22721 m
.38172 .27998 L
s
.37441 .22721 m
.36892 .17426 L
s
.38172 .22679 m
.38172 .1736 L
s
.38172 .22679 m
.38172 .27998 L
s
.38172 .1736 m
.38355 .12044 L
s
.41249 .17739 m
.39622 .22846 L
s
.38172 .27998 m
.38903 .22721 L
s
.38172 .27998 m
.39622 .22846 L
s
.38172 .27998 m
.38172 .33317 L
s
.61831 .13734 m
.58403 .18037 L
s
.61831 .13734 m
.65007 .09273 L
s
.61831 .13734 m
.65501 .09601 L
s
.4812 .30946 m
.51548 .26643 L
s
.4812 .30946 m
.44692 .35249 L
s
.72311 .08104 m
.67977 .11663 L
s
.72311 .08104 m
.76644 .04545 L
s
.67977 .11663 m
.63643 .15222 L
s
.57347 .17481 m
.59719 .12622 L
s
.57347 .17481 m
.54975 .2234 L
s
.58403 .18037 m
.54975 .2234 L
s
.63643 .15222 m
.59309 .18781 L
s
.54975 .2234 m
.51548 .26643 L
s
.54975 .2234 m
.59309 .18781 L
s
.55289 .30946 m
.50238 .33605 L
s
.55289 .30946 m
.6018 .28049 L
s
.55289 .30946 m
.60485 .28531 L
s
.55785 .31962 m
.61331 .30318 L
s
.55785 .31962 m
.50238 .33605 L
s
.61331 .30318 m
.66877 .28674 L
s
.56038 .33049 m
.50238 .33605 L
s
.50238 .33605 m
.44692 .35249 L
s
.72423 .27031 m
.66877 .28674 L
s
.025 w
.39974 .38375 Mdot
.50396 .42607 Mdot
.4941 .44627 Mdot
.78918 .48136 Mdot
.56101 .43713 Mdot
.67622 .4519 Mdot
.47608 .46107 Mdot
.54128 .47754 Mdot
.72831 .48281 Mdot
.72692 .4863 Mdot
.58846 .5088 Mdot
.95016 .54856 Mdot
.8947 .53212 Mdot
.83924 .51568 Mdot
.7344 .45561 Mdot
.73213 .4703 Mdot
.61805 .44819 Mdot
.50524 .50713 Mdot
.67509 .45924 Mdot
.67285 .46638 Mdot
.78377 .49925 Mdot
.44692 .41501 Mdot
.39084 .48687 Mdot
.2975 .85259 Mdot
.34523 .64446 Mdot
.2636 .95455 Mdot
.3726 .48687 Mdot
.35524 .48172 Mdot
.29305 .85162 Mdot
.33153 .75002 Mdot
.36347 .5394 Mdot
.30888 .80043 Mdot
.32249 .74871 Mdot
.34048 .47194 Mdot
.27832 .90308 Mdot
.27612 .90251 Mdot
.35435 .59193 Mdot
.33611 .69699 Mdot
.38172 .43433 Mdot
.22479 .38375 Mdot
.13983 .48991 Mdot
.18874 .46094 Mdot
.16647 .38375 Mdot
.28815 .40538 Mdot
.18568 .45612 Mdot
.23764 .43198 Mdot
.04984 .38375 Mdot
.10816 .38375 Mdot
.28311 .38375 Mdot
.23269 .42182 Mdot
.28815 .36212 Mdot
.34142 .38375 Mdot
.34744 .29014 Mdot
.40894 .17656 Mdot
.36722 .22846 Mdot
.36529 .17468 Mdot
.37989 .12044 Mdot
.37441 .22721 Mdot
.42077 .12447 Mdot
.38172 .22679 Mdot
.38172 .1736 Mdot
.38355 .12044 Mdot
.41249 .17739 Mdot
.38172 .27998 Mdot
.38903 .22721 Mdot
.39622 .22846 Mdot
.42256 .12484 Mdot
.36892 .17426 Mdot
.38172 .33317 Mdot
.61831 .13734 Mdot
.4812 .30946 Mdot
.72311 .08104 Mdot
.67977 .11663 Mdot
.57347 .17481 Mdot
.58403 .18037 Mdot
.59719 .12622 Mdot
.76644 .04545 Mdot
.63643 .15222 Mdot
.54975 .2234 Mdot
.51548 .26643 Mdot
.55289 .30946 Mdot
.55785 .31962 Mdot
.61331 .30318 Mdot
.56038 .33049 Mdot
.50238 .33605 Mdot
.6018 .28049 Mdot
.59309 .18781 Mdot
.65007 .09273 Mdot
.72423 .27031 Mdot
.66877 .28674 Mdot
.65501 .09601 Mdot
.60485 .28531 Mdot
.44692 .35249 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[21]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000182000`400?l00000o`00003ooooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQooooo`00oooooolQ
ooooo`00f_ooool3o`0004?ooooo003Iooooo`Go0000@_ooool00=Sooooo1ol00011ooooo`00f?oo
ool7o`00047ooooo003Hooooo`Oo0000@Oooool00=Sooooo1_l00012ooooo`00eoooool00ol0003o
ooooooooo`03o`0004?ooooo003Fooooo`03o`000?oooooooooo04Oooooo003Dooooo`;o0000B_oo
ool00=?ooooo00?o0000ooooooooool0B_ooool00<kooooo0ol000000ooooooo0000ooooo`1<oooo
o`00cOooool5o`0004kooooo003<ooooo`Oo0000COooool00;Wooooo0ol0000@ooooo`Oo0000COoo
ool00;Sooooo1Ol0000?ooooo`Oo0000COooool00;Oooooo1ol0000>ooooo`Ko0000C_ooool00;Oo
oooo2?l0000<ooooo`03o`000?oooooooooo00?o0000Coooool00;Oooooo2?l0000;ooooo`03o`00
0?oooooooooo05?ooooo002hooooo`Oo00002Oooool2o`0005Kooooo002hooooo`Ko00002Oooool0
0ol0003oooooooooo`1Fooooo`00^?ooool5o`0000Gooooo0ol000000ooooooo0000ooooo`1Hoooo
o`00Joooool5o`0004Oooooo0_l00008ooooo`Go0000F_ooool006[ooooo1ol00006ooooo`Co0000
>oooool2o`0000Sooooo1ol0001Iooooo`00JOooool9o`0000Cooooo1_l0000^ooooo`?o00002?oo
ool2o`0000Wooooo1ol0001Iooooo`00JOooool9o`0000?ooooo2?l0000/ooooo`Go00001oooool2
o`0000Wooooo1ol0001Iooooo`00JOooool9o`0000?ooooo2?l0000[ooooo`Oo00001Oooool2o`00
00[ooooo1_l0001Jooooo`00J_ooool7o`0000Cooooo2?l0000[ooooo`Oo000000?oooooo`000?l0
00000ol00009ooooo`;o00000_ooool3o`0005_ooooo001[ooooo`Go00001_ooool6o`0002cooooo
3?l00009ooooo`03o`000?oooooooooo063ooooo001]ooooo`;o00002?ooool4o`0002kooooo3?l0
0007ooooo`03o`000?oooooooooo067ooooo001]ooooo`;o00002?ooool2o`00037ooooo0ol00000
0ooooooo0000o`000005o`0000Kooooo00?o0000ooooooooool0H_ooool006gooooo0_l00008oooo
o`;o0000<?ooool00ol0003oooooooooo`02ooooo`[o000000?oooooo`000?l00000IOooool006go
oooo0_l00008ooooo`;o0000<?ooool00ol0003oooooooooo`03ooooo`[o0000Ioooool006gooooo
00?o0000ooooooooool01oooool00ol0003oooooooooo`0^ooooo`03o`000?oooooooooo00Cooooo
2ol0001Vooooo`00KOooool00ol0003oooooooooo`06ooooo`;o0000<?ooool00ol0003ooooooooo
o`03ooooo`03o`000?oooooooooo00;ooooo1ol0001Vooooo`00KOooool00ol0003oooooooooo`06
ooooo`;o0000;oooool00ol0003oooooooooo`03ooooo`03o`000?oooooooooo00?ooooo1ol0001V
ooooo`00KOooool00ol0003oooooooooo`06ooooo`03o`000?oooooooooo02kooooo00?o0000oooo
ooooool00_ooool00ol0003oooooooooo`04ooooo`Ko0000Ioooool006cooooo0ol00007ooooo`03
o`000?oooooooooo02gooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`03ooooo`03
o`000?oooooooooo00?o0000J?ooool006Oooooo2Ol00004ooooo`Co0000:oooool4o`0000Cooooo
00?o0000ooooooooool00oooool00ol0003oooooooooo`1/ooooo`00I_ooool;o`0000;ooooo1_l0
000Yooooo`Oo000000?oooooo`000?ooool00oooool2o`0006oooooo001Uooooo`co000000?ooooo
o`000?l000001_l0000Wooooo`Wo00001?ooool00ol0003oooooooooo`1_ooooo`00IOooool<o`00
0003ooooool0003o000000Ko00009oooool;o`000003ooooool0003ooooo077ooooo001Uooooo`_o
00000_ooool8o`0002Oooooo3?l0001cooooo`00I_ooool9o`0000Cooooo1_l0000Yooooo`co0000
L_ooool006Oooooo1?l00002ooooo`03o`000?oooooooooo00Cooooo1?l0000[ooooo`_o0000L_oo
ool006Wooooo00Go0000ooooooooooooooooo`000006ooooo`;o0000;?ooool00ol0003ooooooooo
o`09o`0007;ooooo001Yooooo`;o00000_ooool00ol0003oooooooooo`04ooooo`03o`000?oooooo
oooo02_ooooo00?o0000ooooool000000_ooool6o`0007?ooooo001Yooooo`;o00000_ooool00ol0
003oooooooooo`04ooooo`03o`000?oooooooooo02[ooooo00Ko0000oooooooooooo0000ooooool0
0002ooooo`?o0000M?ooool006Wooooo0_l00002ooooo`03o`000?oooooooooo00?ooooo0_l0000/
ooooo`05o`000?oooooo0000ooooool00000N_ooool006[ooooo00Co0000oooooooooooo00001Ooo
ool00ol0003oooooooooo`0Zooooo`03o`000?oooooo000000;o0000Noooool006[ooooo00Co0000
oooooooooooo00001Oooool00ol0003oooooooooo`0Zooooo`?o0000OOooool006[ooooo00Co0000
oooooooooooo00001?ooool2o`0002_ooooo0ol0001nooooo`00J_ooool01?l0003oooooooooool0
0004ooooo`03o`000?oooooooooo02Oooooo1Ol0001oooooo`00J?ooool;o`0002Sooooo1Ol00020
ooooo`00Ioooool=o`0002Kooooo1ol0001oooooo`00I_ooool?o`0002Gooooo1ol0001oooooo`00
I_ooool?o`0002Gooooo1ol0001oooooo`00I_ooool?o`0002Kooooo1Ol00020ooooo`00Ioooool=
o`0002Oooooo1?l00021ooooo`00J?ooool;o`0002Oooooo00?o0000ooooooooool0Poooool006[o
oooo0_l000000ooooooo0000ooooo`02o`0002Sooooo00?o0000ooooooooool0Q?ooool006[ooooo
00?o0000ooooool000000_l000000ooooooo0000ooooo`0Vooooo`03o`000?oooooooooo08Gooooo
001Zooooo`03o`000?oooooo000000;o000000?oooooo`000?ooool09_ooool00ol0003ooooooooo
o`25ooooo`00Joooool5o`0002Oooooo00?o0000ooooooooool0Q_ooool006_ooooo1Ol0000Roooo
o`?o000000?oooooo`000?ooool0R?ooool006_ooooo1Ol0000Qooooo`Go0000>?ooool3o`0004oo
oooo001/ooooo`?o00008Oooool7o`0003Kooooo1Ol0001>ooooo`00K?ooool3o`00027ooooo1ol0
000eooooo`Oo0000COooool006cooooo0ol0000Qooooo`Oo00005?ooool3o`0001kooooo1ol0001=
ooooo`00Joooool5o`00027ooooo1Ol0000Dooooo`Go00006oooool9o`0004gooooo001Zooooo`Oo
00008?ooool4o`0001Cooooo1ol0000>ooooo`?o00001_ooool3o`0000?ooooo1Ol0001>ooooo`00
H_ooool3o`0000Gooooo1ol0000Oooooo`03o`000?oooooooooo01Kooooo2?l0000<ooooo`Go0000
0_ooool3o`0000Oooooo0ol0001?ooooo`00HOooool5o`0000Cooooo1ol0000Oooooo`03o`000?oo
oooooooo01Kooooo2?l0000;ooooo`So0000G?ooool0063ooooo1ol00004ooooo`Go00007oooool0
0ol0003oooooooooo`0Fooooo`03o`000?oooooo000000Ko00002oooool7o`0005gooooo001Poooo
o`Oo00001Oooool3o`0001oooooo00?o0000ooooooooool05Oooool4o`000003ooooool0003o0000
00?o00002oooool8o`0005gooooo001Pooooo`Oo00001_ooool00ol0003oooooooooo`0Mooooo`03
o`000?oooooooooo01Cooooo1?l00004ooooo`Go00001oooool3o`0000;ooooo1Ol0001Nooooo`00
HOooool5o`0000Oooooo00?o0000ooooooooool06Oooool3o`000003ooooool0003ooooo00kooooo
0ol00002ooooo`?o00002?ooool5o`0000?ooooo0ol00006ooooo`?o0000Goooool006;ooooo1?l0
0007ooooo`03o`000?oooooooooo01Sooooo1Ol0000?ooooo`Oo00002Oooool9o`0006_ooooo001V
ooooo`03o`000?oooooooooo00Cooooo00?o0000ooooooooool05oooool7o`0000gooooo1ol0000:
ooooo`Oo0000KOooool006Oooooo00?o0000ooooooooool00oooool00ol0003oooooooooo`0Goooo
o`Oo00003Oooool7o`0000Sooooo2Ol0001]ooooo`00J?ooool00ol0003oooooooooo`02ooooo`03
o`000?oooooooooo01Oooooo1ol0000=ooooo`So00000oooool4o`0000?ooooo1Ol0001^ooooo`00
J?ooool00ol0003oooooooooo`02ooooo`03o`000?oooooooooo01Sooooo1Ol0000=ooooo`03o`00
0?oooooo000000Wo00002?ooool3o`0006oooooo001Yooooo`05o`000?ooooooooooooooool00000
6_ooool4o`0000cooooo0_l00003ooooo`Oo0000O?ooool006[ooooo00?o0000ooooool000000_l0
000Hooooo`03o`000?oooooooooo00cooooo0_l00003ooooo`Wo0000O?ooool006_ooooo1Ol0000G
ooooo`03o`000?oooooooooo00Gooooo0ol00002ooooo`;o00000_ooool3o`0000;ooooo1ol0001l
ooooo`00J_ooool7o`0001Gooooo00?o0000ooooooooool01Oooool6o`000004ooooool0003o0000
o`0000Cooooo2?l0001looooo`00J_ooool7o`0001Cooooo00?o0000ooooooooool01Oooool?o`00
0003ooooool0003o000000Go0000O?ooool006[ooooo1ol0000Cooooo`03o`000?oooooooooo00Ko
oooo1ol0000:ooooo`Go0000OOooool006_ooooo1Ol0000Cooooo`03o`000?oooooooooo00Gooooo
2Ol0000;ooooo`?o0000O_ooool006cooooo0ol0000@ooooo`?o000000?oooooo`000?ooool00ooo
ool3o`0000?ooooo1Ol0002=ooooo`00K_ooool00ol0003oooooooooo`0=ooooo`Go00000_ooool3
o`0000Oooooo0ol0002>ooooo`00DOooool3o`0001_ooooo00?o0000ooooooooool02oooool8o`00
09_ooooo001@ooooo`Go00006_ooool00ol0003oooooooooo`0;ooooo`Oo0000W?ooool004oooooo
1ol0000Iooooo`03o`000?oooooooooo00_ooooo1ol0002Looooo`00Coooool7o`0001[ooooo00?o
0000ooooooooool02Oooool7o`0009gooooo001?ooooo`Oo00006_ooool00ol0003oooooooooo`08
ooooo`04o`000?ooooooooooooooo`?o0000W_ooool0053ooooo1Ol000000ooooooo0000o`00000H
ooooo`03o`000?oooooooooo00Kooooo0_l0002Uooooo`00DOooool3o`0000Cooooo0ol0000Foooo
o`03o`000?oooooooooo00?ooooo0_l0002Wooooo`003Oooool3o`0000gooooo0ol0000>ooooo`?o
00003_ooool3o`0000kooooo0ol00008ooooo`;o00000oooool3o`0000kooooo0ol00002ooooo`03
o`000?oooooooooo0:Oooooo000<ooooo`Go00002oooool5o`0000cooooo1Ol0000<ooooo`Go0000
3?ooool5o`0000Wooooo1ol0000<ooooo`Ko0000Z_ooool000_ooooo1ol00009ooooo`Oo00002_oo
ool7o`0000[ooooo1ol0000:ooooo`Oo00002Oooool7o`0000[ooooo1ol0002Zooooo`002ooooom[
o`000:[ooooo000;ooooo`Oo00002Oooool7o`0000[ooooo1ol0000:ooooo`Oo00002_ooool7o`00
00Sooooo2?l0000:ooooo`Oo0000Z_ooool000cooooo1Ol0000;ooooo`Go00003?ooool5o`0000co
oooo1Ol0000<ooooo`Go00001oooool2o`0000;ooooo1Ol0000<ooooo`Oo0000ZOooool000gooooo
0ol0000=ooooo`?o00003_ooool3o`0000kooooo0ol0000>ooooo`Co00001?ooool3o`0000Gooooo
0ol0000>ooooo`?o00000oooool00ol0003oooooooooo`2Vooooo`00D?ooool5o`000003ooooool0
003o000001Wooooo00?o0000ooooooooool01?ooool2o`000:Kooooo001?ooooo`Oo00006_ooool0
0ol0003oooooooooo`07ooooo`;o0000Y?ooool004oooooo1ol0000Jooooo`03o`000?oooooooooo
00Wooooo00?o0000ooooooooool00ol0002Nooooo`00COooool9o`0001[ooooo00?o0000oooooooo
ool02_ooool6o`0009gooooo0011ooooo`?o00001_ooool5o`000003ooooool0003o000000?o0000
6_ooool00ol0003oooooooooo`0;ooooo`Oo0000W?ooool0043ooooo1Ol00002ooooo`?o000000?o
ooooo`000?l000001?ooool3o`0001_ooooo00?o0000ooooooooool02oooool7o`0000_ooooo0ol0
002>ooooo`00?oooool8o`0000;ooooo0_l0000Tooooo`03o`000?oooooooooo00_ooooo3?l00005
ooooo`Go0000SOooool003oooooo1ol000000ooooooo0000o`00000Sooooo`?o00003oooool7o`00
00Cooooo2ol0002<ooooo`00?oooool8o`0002Cooooo1Ol0000?ooooo`Co00000_ooool00ol0003o
ooooooooo`05ooooo`Oo00002oooool3o`0007kooooo0010ooooo`So00008_ooool7o`00017ooooo
00Co0000oooooooooooooooo0_l00005ooooo`go00001?ooool5o`0007gooooo0011ooooo`Oo0000
8_ooool7o`0001;ooooo00?o0000ooooooooool00_ooool00ol0003oooooooooo`03ooooo`Go0000
1oooool:o`0007cooooo000oooooo`Wo00008_ooool7o`0001;ooooo00?o0000ooooooooool00ooo
ool00ol0003oooooooooo`06o`0000_ooooo2?l0000;ooooo`?o0000KOooool003cooooo1?l00002
ooooo`Go00008oooool6o`0001Cooooo00?o0000ooooooooool00oooool6o`0000gooooo1ol00000
0ooooooo0000o`000003o`0000Gooooo1Ol0000=ooooo`?o0000G?ooool003Cooooo0ol00003oooo
o`;o000000?oooooo`000?l000001?ooool3o`0002?ooooo00?o0000ooooooooool00ol0000Foooo
o`03o`000?oooooooooo00;ooooo1ol0000=ooooo`Go00001oooool;o`0000_ooooo1Ol0000<oooo
o`?o0000C?ooool003?ooooo1ol000000ooooooo0000o`00000[ooooo`03o`000?oooooooooo00;o
000000?oooooo`000?ooool05Oooool00ol0003oooooooooo`02ooooo`Oo00003_ooool3o`0000co
oooo3Ol00004ooooo`Oo00002_ooool5o`0004_ooooo000booooo`Oo000000?oooooo`000?ooool0
:oooool02?l0003oooooooooool0003oooooo`000?oooooo00005oooool;o`0001gooooo3?l00000
0ooooooo0000o`00000>o`0000Cooooo1ol0001:ooooo`00<_ooool8o`0002cooooo00Wo0000oooo
ooooooooooooo`000?oooooo0000ooooool000005oooool<o`0001gooooo1Ol00003ooooo`oo0000
1Oooool;o`0004[ooooo000booooo`So0000:oooool01Ol0003ooooooooooooooooo000000;ooooo
00?o0000ooooool000005_ooool:o`0000?ooooo00?o0000ooooooooool06oooool3o`0000Oooooo
0ol000000ooooooo0000o`000005o`0000[ooooo1ol0001:ooooo`00<oooool7o`0002Kooooo0ol0
00000ooooooo0000ooooo`03ooooo`03o`000?oooooo000000?ooooo00?o0000ooooooooool04ooo
ool7o`0000Oooooo0_l0000Xooooo`ko00001Oooool5o`0004_ooooo000cooooo`Oo00009Oooool5
o`0000Cooooo00Co0000oooooooooooo00000oooool00ol0003oooooooooo`0Cooooo`Oo00002Ooo
ool2o`0002Oooooo1ol00006ooooo`[o0000Boooool003;ooooo00?o0000ooooool000001?l0000U
ooooo`Oo00000oooool01?l0003oooooooooool00003ooooo`03o`000?oooooooooo01Cooooo1Ol0
000<ooooo`03o`000?oooooooooo00?o00008Oooool9o`0000Sooooo1ol0001:ooooo`00<?ooool2
o`0000?ooooo0ol0000Vooooo`[o00000oooool01Ol0003ooooooooooooooooo000001Oooooo1?l0
000=ooooo`Ko00008Oooool5o`0000;ooooo1?l00005ooooo`[o00002?ooool3o`0003cooooo000^
ooooo`;o0000;_ooool:o`000004ooooool0003o0000o`0000?ooooo0ol0000Hooooo`03o`000?oo
oooooooo00_ooooo1ol0000Qooooo`?o00001Oooool5o`000003ooooool0003o000000Ko00000ooo
ool4o`0000?ooooo1Ol0000kooooo`009oooool3o`0000;ooooo0_l0000aooooo`ko000000?ooooo
o`000?l000000ol0000Hooooo`03o`000?oooooooooo00[ooooo1ol0000/ooooo`[o00002?ooool9
o`0003[ooooo000Vooooo`Ko0000=?oooolDo`0001Sooooo00?o0000ooooooooool02Oooool7o`00
02kooooo1ol0000;ooooo`Oo0000>_ooool002Gooooo1ol0000foooooa;o00006Oooool00ol0003o
ooooooooo`09ooooo`Oo0000;Oooool:o`0000Sooooo1ol0000jooooo`009Oooool7o`0003Oooooo
4Ol0000Iooooo`03o`000?oooooooooo00[ooooo0ol00003ooooo`03o`000?oooooooooo02[ooooo
1ol00003ooooo`?o00001Oooool6o`0003_ooooo000Uooooo`Oo0000>?ooool3o`000003ooooool0
003o000000?o000000?oooooo`000?l000000ol0000Kooooo`03o`000?oooooooooo013ooooo0_l0
000[ooooo`Go00001oooool3o`000003ooooool0003o000000Co0000??ooool002Kooooo1Ol0000n
ooooo`?o00000oooool3o`0001gooooo0ol0000Aooooo`03o`000?oooooooooo02Wooooo0ol0000;
ooooo`Oo0000??ooool002Oooooo0ol00010ooooo`03o`000?oooooooooo027ooooo1Ol0000Aoooo
o`03o`000?oooooooooo00?o0000<oooool7o`0003cooooo001Yooooo`03o`000?oooooooooo027o
oooo1ol0000Aooooo`Ko0000<_ooool:o`0003Wooooo001Yooooo`03o`000?oooooooooo027ooooo
1ol0000Aooooo`Oo0000<_ooool5o`0000Cooooo1?l00004ooooo`?o0000;_ooool006Wooooo00?o
0000ooooooooool08Oooool7o`00017ooooo1ol0000cooooo`?o00002Oooool8o`0002gooooo001Y
ooooo`03o`000?oooooooooo02;ooooo1Ol0000Booooo`Oo0000@Oooool7o`0002cooooo001Yoooo
o`03o`000?oooooooooo02?ooooo0ol0000Dooooo`Go0000@_ooool9o`0002[ooooo001Yooooo`03
o`000?oooooooooo03_ooooo0ol00013ooooo`Oo00000_ooool3o`0000Kooooo0ol0000Nooooo`00
JOooool00ol0003oooooooooo`22ooooo`Go00001_ooool3o`0000;ooooo1Ol0000Mooooo`00Iooo
ool3o`0008Gooooo0ol0000:ooooo`So00007?ooool006Kooooo1Ol0002Booooo`Oo00007?ooool0
06Gooooo1ol0002Aooooo`Wo00006_ooool006Gooooo1ol0002Booooo`Go00000oooool3o`0000Ko
oooo0ol0000>ooooo`00IOooool7o`0009?ooooo0ol00007ooooo`?o00000_ooool5o`0000gooooo
001Vooooo`Go0000XOooool8o`0000cooooo001Wooooo`?o0000Xoooool7o`0000cooooo001Woooo
o`03o`000?oooooooooo0:?ooooo1ol0000<ooooo`00Ioooool00ol0003oooooooooo`2Tooooo`Go
00003Oooool006Oooooo00?o0000ooooooooool0YOooool3o`0000kooooo001Wooooo`03o`000?oo
oooooooo0;Kooooo001Vooooo`03o`000?oooooooooo0;Oooooo001Vooooo`03o`000?oooooooooo
0;Oooooo001Vooooo`03o`000?oooooooooo0;Oooooo001Vooooo`03o`000?oooooooooo0;Oooooo
001Tooooo`?o0000^Oooool006?ooooo1Ol0002hooooo`00H_ooool7o`000;Oooooo001Rooooo`Oo
0000]oooool006;ooooo1ol0002gooooo`00Hoooool5o`000;Sooooo001Tooooo`?o0000^Oooool0
06Gooooo00?o0000ooooooooool0^?ooool006Cooooo00?o0000ooooooooool0^Oooool006Cooooo
00?o0000ooooooooool0^Oooool006Cooooo00?o0000ooooooooool0^Oooool006Cooooo00?o0000
ooooooooool0^Oooool006Cooooo00?o0000ooooooooool0^Oooool006Cooooo00?o0000oooooooo
ool0^Oooool006Cooooo00?o0000ooooooooool0^Oooool006?ooooo00?o0000ooooooooool0^_oo
ool006;ooooo0ol0002kooooo`00HOooool5o`000;[ooooo001Pooooo`Oo0000^Oooool0063ooooo
1ol0002iooooo`00H?ooool7o`000;Wooooo001Qooooo`Go0000^_ooool006;ooooo0ol0002koooo
o`00H_ooool00ol0003oooooooooo`2kooooo`00H_ooool00ol0003oooooooooo`2kooooo`00H_oo
ool00ol0003oooooooooo`2kooooo`00H_ooool00ol0003oooooooooo`2kooooo`00HOooool00ol0
003oooooooooo`2looooo`00HOooool00ol0003oooooooooo`2looooo`00HOooool00ol0003ooooo
ooooo`2looooo`00HOooool00ol0003oooooooooo`2looooo`00Goooool3o`000;kooooo001Noooo
o`Go0000_Oooool005gooooo1ol0002looooo`00GOooool7o`000;cooooo001Mooooo`Oo0000_?oo
ool005kooooo1Ol0002mooooo`00Goooool3o`000;kooooo001Oooooo`;o0000_oooool005oooooo
0_l0002oooooo`00G_ooool00ol0003oooooo`00002oooooo`00G_ooool00ol0003oooooo`00002o
ooooo`00G_ooool2o`000<3ooooo001Nooooo`;o0000`?ooool005gooooo00?o0000ooooool00000
`?ooool005_ooooo0ol000000ooooooo0000ooooo`2oooooo`00F_ooool7o`000;oooooo001Ioooo
o`Wo0000__ooool005Wooooo2_l0002mooooo`00FOooool:o`000;gooooo001Jooooo`Wo0000_Ooo
ool005_ooooo1ol0002nooooo`00Foooool00ol0003oooooooooo`03o`000;oooooo001Kooooo`03
o`000?oooooooooo0<;ooooo001Kooooo`03o`000?oooooooooo0<;ooooo001Jooooo`03o`000?oo
oooooooo0<?ooooo001Jooooo`03o`000?oooooooooo0<?ooooo001Jooooo`03o`000?oooooooooo
0<?ooooo001Iooooo`03o`000?oooooooooo0<Cooooo001Iooooo`03o`000?oooooooooo0<Cooooo
001Gooooo`?o0000a_ooool005Kooooo1Ol00035ooooo`00EOooool7o`000<Cooooo001Eooooo`Oo
0000a?ooool005Gooooo1ol00034ooooo`00E_ooool5o`000<Gooooo001Gooooo`?o0000a_ooool0
05Oooooo00?o0000ooooooooool0a_ooool005Oooooo00?o0000ooooooooool0a_ooool005Oooooo
00?o0000ooooooooool0a_ooool005Kooooo0_l00038ooooo`00E_ooool00ol0003oooooooooo`37
ooooo`00E_ooool00ol0003oooooooooo`37ooooo`00EOooool2o`000<Wooooo001Eooooo`;o0000
bOooool005?ooooo1?l00039ooooo`00D_ooool6o`000<Sooooo001Aooooo`So0000aoooool0057o
oooo2?l00037ooooo`00DOooool8o`000<Oooooo001Booooo`Ko0000b?ooool005?ooooo1?l00039
ooooo`00Doooool00ol0003oooooooooo`3:ooooo`00D_ooool00ol0003oooooooooo`3;ooooo`00
D_ooool00ol0003oooooooooo`3;ooooo`00D_ooool00ol0003oooooooooo`3;ooooo`00DOooool0
0ol0003oooooooooo`3<ooooo`00DOooool00ol0003oooooooooo`3<ooooo`00DOooool00ol0003o
ooooooooo`3<ooooo`00D?ooool00ol0003oooooooooo`3=ooooo`00C_ooool3o`000<oooooo001=
ooooo`Go0000c_ooool004cooooo1ol0003=ooooo`00C?ooool7o`000<gooooo001<ooooo`Oo0000
cOooool004gooooo1Ol0003>ooooo`00C_ooool3o`000<oooooo001>ooooo`03o`000?oooooooooo
0<oooooo001>ooooo`03o`000?oooooooooo0<oooooo001=ooooo`03o`000?oooooooooo0=3ooooo
001=ooooo`03o`000?oooooooooo0=3ooooo001=ooooo`03o`000?oooooooooo0=3ooooo001=oooo
o`03o`000?oooooooooo0=3ooooo001<ooooo`03o`000?oooooooooo0=7ooooo001<ooooo`03o`00
0?oooooooooo0=7ooooo001:ooooo`?o0000doooool004Wooooo1Ol0003Booooo`00B?ooool7o`00
0=7ooooo0018ooooo`Oo0000dOooool004Sooooo1ol0003Aooooo`00BOooool5o`000=;ooooo001:
ooooo`?o0000doooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Ooo
ool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Oooool00?oooooo8Ooo
ool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-0.0289143, -0.0500054, \
0.00349563, 0.00383279}},
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1152}, {0, 701}},
WindowSize->{520, 578},
WindowMargins->{{Automatic, 108}, {34, Automatic}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:15"->{
    Cell[1776, 53, 80, 1, 38, "ObjectName",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[1859, 56, 1019, 35, 86, "Usage",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[2881, 93, 135, 4, 24, "Notes",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3019, 99, 291, 7, 43, "3ColumnTable",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3313, 108, 255, 9, 24, "Notes",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3593, 121, 80, 1, 38, "ExampleSection",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3676, 124, 122, 1, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3823, 129, 150, 3, 26, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3976, 134, 20935, 549, 294, 4552, 341, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[24926, 686, 162, 4, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[25113, 694, 244, 5, 58, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[25360, 701, 25004, 655, 294, 5433, 408, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50379, 1359, 158, 4, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50562, 1367, 233, 4, 58, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50798, 1373, 28457, 646, 294, 4480, 344, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79270, 2022, 269, 5, 40, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79564, 2031, 156, 3, 26, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79723, 2036, 22476, 625, 294, 5467, 409, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}]},
  "LocalInteractions:44"->{
    Cell[1776, 53, 80, 1, 38, "ObjectName",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[1859, 56, 1019, 35, 86, "Usage",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[2881, 93, 135, 4, 24, "Notes",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3019, 99, 291, 7, 43, "3ColumnTable",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3313, 108, 255, 9, 24, "Notes",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3593, 121, 80, 1, 38, "ExampleSection",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3676, 124, 122, 1, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3823, 129, 150, 3, 26, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[3976, 134, 20935, 549, 294, 4552, 341, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[24926, 686, 162, 4, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[25113, 694, 244, 5, 58, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[25360, 701, 25004, 655, 294, 5433, 408, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50379, 1359, 158, 4, 24, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50562, 1367, 233, 4, 58, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[50798, 1373, 28457, 646, 294, 4480, 344, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79270, 2022, 269, 5, 40, "Text",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79564, 2031, 156, 3, 26, "Input",
      CellTags->{"b:15", "LocalInteractions:44"}],
    Cell[79723, 2036, 22476, 625, 294, 5467, 409, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->{"b:15", "LocalInteractions:44"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:15", 102860, 2682},
  {"LocalInteractions:44", 104789, 2723}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 80, 1, 38, "ObjectName",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[1859, 56, 1019, 35, 86, "Usage",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[2881, 93, 135, 4, 24, "Notes",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[3019, 99, 291, 7, 43, "3ColumnTable",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[3313, 108, 255, 9, 24, "Notes",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{
Cell[3593, 121, 80, 1, 38, "ExampleSection",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[3676, 124, 122, 1, 24, "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{
Cell[3823, 129, 150, 3, 26, "Input",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[3976, 134, 20935, 549, 294, 4552, 341, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],
Cell[24926, 686, 162, 4, 24, "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{
Cell[25113, 694, 244, 5, 58, "Input",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[25360, 701, 25004, 655, 294, 5433, 408, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],
Cell[50379, 1359, 158, 4, 24, "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{
Cell[50562, 1367, 233, 4, 58, "Input",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[50798, 1373, 28457, 646, 294, 4480, 344, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]],
Cell[79270, 2022, 269, 5, 40, "Text",
  CellTags->{"b:15", "LocalInteractions:44"}],

Cell[CellGroupData[{
Cell[79564, 2031, 156, 3, 26, "Input",
  CellTags->{"b:15", "LocalInteractions:44"}],
Cell[79723, 2036, 22476, 625, 294, 5467, 409, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->{"b:15", "LocalInteractions:44"}]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

